/*
 *      CAPS: Computational Aircraft Prototype Syntheses
 *
 *             AVL AIM
 *
 *      Copyright 2014-2015, Massachusetts Institute of Technology
 *      Licensed under The GNU Lesser General Public License, version 2.1
 *      See http://www.opensource.org/licenses/lgpl-2.1.php
 *
 */

#include <string.h>
#include <math.h>
#include "capsTypes.h"

#ifdef WIN32
#define getcwd   _getcwd
#define PATH_MAX _MAX_PATH
#else
#include <unistd.h>
#include <limits.h>
#endif

#define NUMINPUT  8
#define NUMOUT    2
#define NUMPTS   47
#define PI        3.1415926535897931159979635
#define NINT(A)         (((A) < 0)   ? (int)(A-0.5) : (int)(A+0.5))

//#define DEBUG


  typedef struct {
    int nbody;
    ego *bodies;
  } aimStorage;


  typedef struct {
    char   *name;
    double Cspace;
    int    Nchord;
    int    nowake;
    int    noalbe;
    int    noload;
    int    compon;
    int    iYdup;
    int    nsec;
    double xyz[3];       /* per body */
    int    le;
    int    index;        /* per section */
  } aimSurface;


  static aimStorage *instances = NULL;
  static int        nInstance  = 0;

  extern /*@null@*/ /*@only@*/
         char       *EG_strdup(/*@null@*/ const char *str);


#ifdef WIN32
static int
getline(char **linep, size_t *linecapp, FILE *stream)
{
  int  i, cnt = -1;
  char buffer[1026];

  *linep    = buffer;
  *linecapp = 1024;
  while (cnt < 1024) {
    i = fread(&buffer[cnt+1], sizeof(char), 1, stream);
    if (i == 0) return cnt;
    if ((buffer[cnt+1] == 12) || (buffer[cnt+1] == 15) ||
        (buffer[cnt+1] == 0)) break;
    cnt++;
  }
  buffer[cnt+1] = 0;
  if (cnt == -1) cnt++;
  return cnt;
}
#endif

/* ********************** AVL AIM Helper Functions ************************** */

static void
findLeadingEdge(ego body, int *nle, double *xle)
{
  int    i, n, stat, nNode, oclass, mtype, *sens;
  double xmin, xyz[3];
  ego    ref, *nodes, *objs;
  
  *nle = 0;
  
  /* assume the LE position is the most forward Node in X */
  stat = EG_getBodyTopos(body, NULL, NODE, &nNode, &nodes);
  if (stat != EGADS_SUCCESS) {
    printf(" AIM Warning: LE getBodyTopos Nodes = %d\n", stat);
    return;
  }
  for (i = 0; i < nNode; i++) {
    stat = EG_getTopology(nodes[i], &ref, &oclass, &mtype, xyz,
                          &n, &objs, &sens);
    if (stat != EGADS_SUCCESS) continue;
    if (*nle == 0) {
      *nle = i+1;
      xmin = xyz[0];
    } else {
      if (xyz[0] < xmin) {
        *nle = i+1;
        xmin = xyz[0];
      }
    }
  }
  if (*nle == 0) return;
  
  EG_getTopology(nodes[*nle-1], &ref, &oclass, &mtype, xle, &n, &objs, &sens);
  EG_free(nodes);
}


static void
orderSections(int i, aimSurface *surfaces)
{
  int    i1, i2, j, k, hit;
  double box[6], lScale;
  
  j = surfaces[0].index;
  box[0] = box[3] = surfaces[j].xyz[0];
  box[1] = box[4] = surfaces[j].xyz[1];
  box[2] = box[5] = surfaces[j].xyz[2];
  for (k = 0; k < surfaces[i].nsec; k++) {
    j = surfaces[k].index;
    if (surfaces[j].xyz[0] < box[0]) box[0] = surfaces[j].xyz[0];
    if (surfaces[j].xyz[0] > box[3]) box[3] = surfaces[j].xyz[0];
    if (surfaces[j].xyz[1] < box[1]) box[1] = surfaces[j].xyz[1];
    if (surfaces[j].xyz[1] > box[4]) box[4] = surfaces[j].xyz[1];
    if (surfaces[j].xyz[2] < box[2]) box[2] = surfaces[j].xyz[2];
    if (surfaces[j].xyz[2] > box[5]) box[5] = surfaces[j].xyz[2];
  }
  lScale = box[3] - box[0];
  if (box[4] - box[1] > lScale) lScale = box[4] - box[1];
  if (box[5] - box[2] > lScale) lScale = box[5] - box[2];
  
  if ((box[4]-box[1])/lScale > 1.e-5) {
    /* Y ordering */
    do {
      for (hit = k = 0; k < surfaces[i].nsec-1; k++) {
        i1 = surfaces[k  ].index;
        i2 = surfaces[k+1].index;
        if (surfaces[i1].xyz[1] <= surfaces[i2].xyz[1]) continue;
        surfaces[k  ].index = i2;
        surfaces[k+1].index = i1;
        hit++;
      }
    } while (hit != 0);
  } else if ((box[5]-box[2])/lScale > 1.e-5) {
    /* assume Vertical Tail and do Z ordering */
    do {
      for (hit = k = 0; k < surfaces[i].nsec-1; k++) {
        i1 = surfaces[k  ].index;
        i2 = surfaces[k+1].index;
        if (surfaces[i1].xyz[2] <= surfaces[i2].xyz[2]) continue;
        surfaces[k  ].index = i2;
        surfaces[k+1].index = i1;
        hit++;
      }
    } while (hit != 0);
  } else {
    printf(" AIM Warning: No Y or Z Variation in Ssurface LE Bounding Box!\n");
    printf("              %lf %lf   %lf %lf   %lf %lf  - lScale = %lf\n",
           box[0], box[3], box[1], box[4], box[2], box[5], lScale);
  }
/*
   printf(" Sections =");
   for (j = 0; j < surfaces[i].nsec; j++) printf(" %d", surfaces[j].index);
   printf("\n");
 */
}


static void
writeSection(FILE *fp, ego body, int le, double *xle)
{
  int          i, j, n, stat, atype, alen, Nspan, nNode, nEdge, oclass, mtype;
  int          *sens, cnt, sense;
  double       Sspace, chord, ainc, xyz[3], xmin, trange[2], result[9];
  double       xdot[3], ydot[3];
  const int    *ints;
  const char   *string;
  const double *reals;
  ego          *nodes, *edges, ref, te, obj, *objs;
  
  stat = EG_attributeRet(body, "avlSecValues", &atype, &alen,
                         &ints, &reals, &string);
  if (stat != EGADS_SUCCESS) return;
  Nspan  = NINT(reals[2]);
  Sspace =      reals[3];
  
  /* get LE position, Chord and Ainc from the Body */
  stat = EG_getBodyTopos(body, NULL, NODE, &nNode, &nodes);
  if (stat != EGADS_SUCCESS) {
    printf(" AIM Warning: getBodyTopos Nodes = %d\n", stat);
    return;
  }
  stat = EG_getBodyTopos(body, NULL, EDGE, &nEdge, &edges);
  if (stat != EGADS_SUCCESS) {
    EG_free(nodes);
    printf(" AIM Warning: getBodyTopos Edges = %d\n", stat);
    return;
  }
  if ((nEdge != nNode) || (nNode < 2) || (nNode > 3)) {
    printf(" AIM Warning: Body has %d nodes and %d edges!\n", nNode, nEdge);
    return;
  }
  
  /* find TE Edge object, if any */
  te = NULL;
  for (i = 0; i < nEdge; i++) {
    stat = EG_getTopology(edges[i], &ref, &oclass, &mtype, trange,
                          &n, &objs, &sens);
    if (stat != EGADS_SUCCESS) {
      printf(" AIM Warning: Edge %d getTopology = %d!\n", i, stat);
      continue;
    }
    if (n == 1) {
      printf(" AIM Warning: Edge %d has 1 Node!\n", i);
      continue;
    }
    if ((objs[0] != nodes[le-1]) && (objs[1] != nodes[le-1])) {
      te = edges[i];
/*    printf(" TE Edge = %d\n", i);  */
      break;
    }
  }
  if ((nEdge == 3) && (te == NULL)) {
    printf(" AIM Warning: No TE Edge with nEdge = 3!\n");
    return;
  }
  if (te == NULL) {
    i = le;
    if (i == 2) i = 0;
    EG_getTopology(nodes[i], &ref, &oclass, &mtype, result, &n, &objs, &sens);
  } else {
    EG_getTopology(te, &ref, &oclass, &mtype, trange, &n, &objs, &sens);
    xmin = 0.5*(trange[0]+trange[1]);
    stat = EG_evaluate(te, &xmin, result);
    if (stat != EGADS_SUCCESS) {
      printf(" AIM Warning: evaluate = %d!\n", stat);
      return;
    }
  }
  xdot[0]  =  result[0] - xle[0];
  xdot[1]  =  result[1] - xle[1];
  xdot[2]  =  result[2] - xle[2];
  chord    =  sqrt(xdot[0]*xdot[0] + xdot[1]*xdot[1] + xdot[2]*xdot[2]);
  xdot[0] /=  chord;
  xdot[1] /=  chord;
  xdot[2] /=  chord;
  ainc     = -acos(xdot[0])*180./PI;
  ydot[0]  = -xdot[2];
  ydot[1]  =  xdot[1];
  ydot[2]  =  xdot[0];
  if (fabs(xdot[1]) > 1.e-5)
    printf(" AIM Warning: Section not in Y-plane!\n");
  
  fprintf(fp, "SECTION\n%lf %lf %lf  %lf %lf  %d %lf\n\n",
          xle[0], xle[1], xle[2], chord, ainc, Nspan, Sspace);
  
  fprintf(fp, "AIRFOIL\n");
  for (cnt = i = 0; i < nEdge; i++) {
    if (edges[i] == te) continue;
    EG_getTopology(edges[i], &ref, &oclass, &mtype, trange, &n, &objs, &sens);
    if (objs[0] == nodes[le-1]) {
      sense = -1;
      obj   = objs[1];
    } else {
      sense =  1;
      obj   = objs[0];
    }
    if (cnt == 1) sense = -sense;
    
    EG_getTopology(obj, &ref, &oclass, &mtype, xyz, &n, &objs, &sens);
    xyz[0] -= xle[0];
    xyz[1] -= xle[1];
    xyz[2] -= xle[2];
    if (cnt == 0) {
      fprintf(fp, "1.000000 %lf\n",
              (ydot[0]*xyz[0]+ydot[1]*xyz[1]+ydot[2]*xyz[2])/chord);
    } else {
      fprintf(fp, "0.000000 0.000000\n");
    }
    for (j = 1; j < NUMPTS; j++) {
      if (sense == 1) {
        xmin = trange[0] + j*(trange[1]-trange[0])/NUMPTS;
      } else {
        xmin = trange[1] - j*(trange[1]-trange[0])/NUMPTS;
      }
      EG_evaluate(edges[i], &xmin, result);
      result[0] -= xle[0];
      result[1] -= xle[1];
      result[2] -= xle[2];
      fprintf(fp, "%lf %lf\n",
              (xdot[0]*result[0]+xdot[1]*result[1]+xdot[2]*result[2])/chord,
              (ydot[0]*result[0]+ydot[1]*result[1]+ydot[2]*result[2])/chord);
    }
    cnt++;
    if (cnt == 2)
      fprintf(fp, "1.000000 %lf\n",
              (ydot[0]*xyz[0]+ydot[1]*xyz[1]+ydot[2]*xyz[2])/chord);
  }
  
  fprintf(fp, "\n");
}


/* ********************** Exposed AIM Functions ***************************** */

int
aimInitialize(/*@unused@*/ int ngIn, /*@unused@*/ /*@null@*/ capsValue *gIn,
              int *nFid, int **fids, int *nIn, int *nOut,
              int *nFields, char ***fnames, int **ranks)
{
  int        *ints, ret;
  aimStorage *tmp;

#ifdef DEBUG
  printf("\n avlAIM/aimInitialize   ngIn = %d!\n", ngIn);
#endif

  *nFid    = 1;
  ints     = (int *) EG_alloc(sizeof(int));
  if (ints == NULL) return EGADS_MALLOC;
  ints[0]  = 13;
  *fids    = ints;

  /* specify the number of analysis input and out "parameters" */
  *nIn     = NUMINPUT;
  *nOut    = NUMOUT;

  /* specify the field variables this analysis can generate */
  *nFields = 0;
  *ranks   = NULL;
  *fnames  = NULL;
  
  /* get the instance storage */
  if (instances == NULL) {
    instances = (aimStorage *) EG_alloc(sizeof(aimStorage));
    if (instances == NULL) {
      EG_free(*fids);
      *fids   = NULL;
      return EGADS_MALLOC;
    }
  } else {
    tmp = (aimStorage *) EG_reall(instances, (nInstance+1)*sizeof(aimStorage));
    if (tmp == NULL) {
      EG_free(*fids);
      *fids   = NULL;
      return EGADS_MALLOC;
    }
    instances = tmp;
  }
  instances[nInstance].nbody  = 0;
  instances[nInstance].bodies = NULL;
  
  /* return an index for the instance generated */
  ret = nInstance;
  nInstance++;
  return ret;
}


int
aimInputs(/*@unused@*/ int inst, int index, char **ainame, capsValue *defval)
{
#ifdef DEBUG
  printf(" avlAIM/aimInputs instance = %d  index = %d!\n", inst, index);
#endif
  
  if (index == 1) {
    *ainame           = EG_strdup("Mach");
    defval->type      = Double;
    defval->vals.real = 0.0;
    defval->units     = NULL;
  } else if (index == 2) {
    *ainame           = EG_strdup("Sref");
    defval->type      = Double;
    defval->vals.real = 1.0;
    defval->units     = NULL;
  } else if (index == 3) {
    *ainame           = EG_strdup("Cref");
    defval->type      = Double;
    defval->vals.real = 1.0;
    defval->units     = NULL;
  } else if (index == 4) {
    *ainame           = EG_strdup("Bref");
    defval->type      = Double;
    defval->vals.real = 1.0;
    defval->units     = NULL;
  } else if (index == 5) {
    *ainame           = EG_strdup("Xref");
    defval->type      = Double;
    defval->vals.real = 0.0;
    defval->units     = NULL;
  } else if (index == 6) {
    *ainame           = EG_strdup("Yref");
    defval->type      = Double;
    defval->vals.real = 0.0;
    defval->units     = NULL;
  } else if (index == 7) {
    *ainame           = EG_strdup("Zref");
    defval->type      = Double;
    defval->vals.real = 0.0;
    defval->units     = NULL;
  } else if (index == 8) {
    *ainame           = EG_strdup("CDp");
    defval->type      = Double;
    defval->vals.real = 0.0;
    defval->units     = NULL;
  }
  return CAPS_SUCCESS;
}


int
aimPreAnalysis(int inst, int nbody, /*@null@*/ ego *bodies, char *apath,
               /*@null@*/ capsValue *inputs, capsErrs **errs)
{
  int          i, j, k, l, n, stat, atype, alen, nattr, Nchord, iYdup;
/*@-unrecog@*/
  char         cpath[PATH_MAX];
/*@+unrecog@*/
  double       Cspace;
  const int    *ints;
  const char   *string, *aname;
  const double *reals;
  FILE         *fp;
  aimSurface   *surfaces;

#ifdef DEBUG
  printf(" avlAIM/aimPreAnalysis instance = %d  nbody = %d!\n", inst, nbody);
#endif
  
  *errs = NULL;
  /* save away the current geometry for the instance */
  if (instances[inst].bodies != NULL) EG_free(instances[inst].bodies);
  instances[inst].bodies = NULL;
  instances[inst].nbody  = 0;
  if ((nbody > 0) && (bodies != NULL)) {
    instances[inst].bodies = (ego *) EG_alloc(nbody*sizeof(ego));
    if (instances[inst].bodies == NULL) {
#ifdef DEBUG
      printf(" avlAIM/aimPreAnalysis Cannot allocate %d bodies!\n", nbody);
#endif
      return EGADS_MALLOC;
    }
    for (i = 0; i < nbody; i++) instances[inst].bodies[i] = bodies[i];
    instances[inst].nbody = nbody;
  }
  if (inputs == NULL) {
#ifdef DEBUG
    printf(" avlAIM/aimPreAnalysis inputs == NULL!\n");
#endif
    return CAPS_NULLVALUE;
  }
  if ((nbody <= 0) || (bodies == NULL)) {
#ifdef DEBUG
    printf(" avlAIM/aimPreAnalysis No Bodies!\n");
#endif
    return CAPS_SOURCEERR;
  }

  /* get our memory to store AVL Surface info */
  surfaces = (aimSurface *) malloc(nbody*sizeof(aimSurface));
  if (surfaces == NULL) {
#ifdef DEBUG
    printf(" avlAIM/aimPreAnalysis Cannot allocate %d surfaces!\n", nbody);
#endif
    return EGADS_MALLOC;
  }
  for (i = 0; i < nbody; i++) {
    surfaces[i].name   = NULL;
    surfaces[i].Nchord = 0;
    surfaces[i].Cspace = 0.0;
    surfaces[i].nowake = 0;
    surfaces[i].noalbe = 0;
    surfaces[i].noload = 0;
    surfaces[i].compon = 0;
    surfaces[i].iYdup  = 0;
    surfaces[i].nsec   = 0;
    surfaces[i].xyz[0] = surfaces[i].xyz[1] = surfaces[i].xyz[2] = 0.0;
    surfaces[i].le     = 0;
  }
  
  /* accumulate the Surface data */
  for (n = i = 0; i < nbody; i++) {
    stat = EG_attributeRet(bodies[i], "avlSecValues", &atype, &alen,
                           &ints, &reals, &string);
    if (stat  != EGADS_SUCCESS) continue;
    if (atype != ATTRREAL)      continue;
    if (alen  <  5) {
      printf(" Body %d: SecValues type = %d, len = %d\n", i+1, atype, alen);
      continue;
    }
    Nchord = NINT(reals[0]);
    Cspace =      reals[1];
    iYdup  = NINT(reals[4]);
    stat   = EG_attributeRet(bodies[i], "avlSurface", &atype, &alen,
                             &ints, &reals, &string);
    if (stat   != EGADS_SUCCESS) continue;
    if (atype  != ATTRSTRING)    continue;
    if (string == NULL)          continue;
    findLeadingEdge(bodies[i], &surfaces[i].le, surfaces[i].xyz);
    if (surfaces[i].le == 0) {
#ifdef DEBUG
      printf(" avlAIM/aimPreAnalysis Body %d has no LE!\n", i+1);
#endif
      continue;
    }
    for (j = 0; j < n; j++)
      if (strcmp(string, surfaces[j].name) == 0) break;
    if (j == n) {
      surfaces[n].name   = EG_strdup(string);
      surfaces[n].Nchord = Nchord;
      surfaces[n].Cspace = Cspace;
      surfaces[n].iYdup  = iYdup;
      n++;
    }
    surfaces[j].nsec++;
    stat = EG_attributeRet(bodies[i], "avlKeyword", &atype, &alen,
                           &ints, &reals, &string);
    if ((stat == EGADS_SUCCESS) && (atype == ATTRSTRING) && (string != NULL)) {
      if (strstr(string,"NOWAKE") != NULL) surfaces[j].nowake = 1;
      if (strstr(string,"NOALBE") != NULL) surfaces[j].noalbe = 1;
      if (strstr(string,"NOLOAD") != NULL) surfaces[j].noload = 1;
    }
    stat = EG_attributeRet(bodies[i], "avlComponent", &atype, &alen,
                           &ints, &reals, &string);
    if ((stat == EGADS_SUCCESS) && (atype != ATTRSTRING))
      if (atype == ATTRREAL) {
        surfaces[j].compon = NINT(reals[0]);
      } else {
        surfaces[j].compon = ints[0];
      }
  }
  if (n == 0) {
#ifdef DEBUG
    printf(" avlAIM/aimPreAnalysis No AVL Surface Attributes!\n");
#endif
    free(surfaces);
    return CAPS_SOURCEERR;
  }
  
  /* get where we are and set the path to our input */
  (void) getcwd(cpath, PATH_MAX);
  if (chdir(apath) != 0) {
#ifdef DEBUG
    printf(" avlAIM/aimPreAnalysis Cannot chdir to %s!\n", apath);
#endif
    for (i = 0; i < n; i++) EG_free(surfaces[i].name);
    free(surfaces);
    return CAPS_DIRERR;
  }
  
  /* open and write the input to control the AVL session */
  fp = fopen("avlInput.txt", "w");
  if (fp == NULL) {
#ifdef DEBUG
    printf(" avlAIM/aimPreAnalysis Cannot open file avlInput.txt!\n");
#endif
    chdir(cpath);
    for (i = 0; i < n; i++) EG_free(surfaces[i].name);
    free(surfaces);
    return CAPS_IOERR;
  }
  fprintf(fp, "OPER\nM\nMN\n%lf\n\nX\n\nQuit\n", inputs[0].vals.real);
  fclose(fp);

  /* open the file and write the avl data */
  fp = fopen("caps.avl", "w");
  if (fp == NULL) {
#ifdef DEBUG
    printf(" avlAIM/aimPreAnalysis Cannot open file caps.avl!\n");
#endif
    chdir(cpath);
    for (i = 0; i < n; i++) EG_free(surfaces[i].name);
    free(surfaces);
    return CAPS_IOERR;
  }
  fprintf(fp, "CAPS generated Configuration\n");
  fprintf(fp, "0.0\n");                              /* Mach */
  fprintf(fp, "0 0 0.0\n");                          /* IYsym   IZsym   Zsym */
  fprintf(fp, "%lf %lf %lf\n", inputs[1].vals.real, inputs[2].vals.real,
                               inputs[3].vals.real); /* Sref    Cref    Bref */
  fprintf(fp, "%lf %lf %lf\n", inputs[4].vals.real, inputs[5].vals.real,
                               inputs[6].vals.real); /* Xref    Yref    Zref */
  fprintf(fp, "%lf\n", inputs[7].vals.real);         /* CDp */
  
  /* write out the Surfaces, one at a time */
  for (i = 0; i < n; i++) {
    if (surfaces[i].nsec < 2) {
      printf(" avlAIM/aimPreAnalysis Surface %s has %d Sections!\n",
             surfaces[i].name, surfaces[i].nsec);
      continue;
    }
    
    /* order Sections */
    for (k = j = 0; j < nbody; j++) {
      stat = EG_attributeRet(bodies[j], "avlSecValues", &atype, &alen,
                             &ints, &reals, &string);
      if (stat  != EGADS_SUCCESS) continue;
      if (atype != ATTRREAL)      continue;
      if (alen  <  5)             continue;
      stat = EG_attributeRet(bodies[j], "avlSurface", &atype, &alen,
                             &ints, &reals, &string);
      if (stat   != EGADS_SUCCESS) continue;
      if (atype  != ATTRSTRING)    continue;
      if (string == NULL)          continue;
      if (surfaces[j].le == 0)     continue;
      if (strcmp(string, surfaces[i].name) != 0) continue;
      surfaces[k].index = j;
      k++;
    }
    orderSections(i, surfaces);
    
    fprintf(fp, "#\nSURFACE\n%s\n%d %lf\n\n",
            surfaces[i].name, surfaces[i].Nchord, surfaces[i].Cspace);
    if  (surfaces[i].compon != 0)
      fprintf(fp, "COMPONENT\n%d\n\n", surfaces[i].compon);
    if  (surfaces[i].iYdup  != 0)
      fprintf(fp, "YDUPLICATE\n0.0\n\n");
    if  (surfaces[i].nowake != 0) fprintf(fp, "NOWAKE\n");
    if  (surfaces[i].noalbe != 0) fprintf(fp, "NOALBE\n");
    if  (surfaces[i].noload != 0) fprintf(fp, "NOLOAD\n");
    if ((surfaces[i].nowake != 0) || (surfaces[i].noalbe != 0) ||
        (surfaces[i].noload != 0)) fprintf(fp,"\n");
  
    /* write the Sections for each Surface */
    for (l = 0; l < surfaces[i].nsec; l++) {
      j = surfaces[l].index;
      writeSection(fp, bodies[j], surfaces[j].le, surfaces[j].xyz);
      
      /* look for Control Surface markers */
      stat = EG_attributeNum(bodies[j], &nattr);
      if (stat == EGADS_SUCCESS)
        for (k = 0; k < nattr; k++) {
          stat = EG_attributeGet(bodies[j], k+1, &aname, &atype, &alen,
                                 &ints, &reals, &string);
          if (stat  != EGADS_SUCCESS) continue;
          if (atype != ATTRREAL)      continue;
          if (alen  <  6)             continue;
          if (strncmp(aname, "avlCntrl", 8) != 0) continue;
          fprintf(fp, "CONTROL\n");
          fprintf(fp, "%s  %lf %lf  %lf %lf %lf  %lf\n", &aname[8], reals[0],
                  reals[1], reals[2], reals[3], reals[4], reals[5]);
        }
    }
  }
  fprintf(fp, "\n");
  fclose(fp);
  
  /* restore the path we came in with and get out */
  chdir(cpath);
  for (i = 0; i < n; i++) EG_free(surfaces[i].name);
  free(surfaces);
  return CAPS_SUCCESS;
}


int
aimOutputs(/*@unused@*/ int inst, int index, char **aoname, capsValue *form)
{
#ifdef DEBUG
  printf(" avlAIM/aimOutputs instance = %d  index = %d!\n", inst, index);
#endif

  if (index == 1) {
    *aoname = EG_strdup("CLtot");
    form->type = Double;
  } else if (index == 2) {
    *aoname = EG_strdup("CDtot");
    form->type = Double;
  }
  return CAPS_SUCCESS;
}


int
aimCalcOutput(/*@unused@*/ int inst, char *apath, int index, capsValue *val,
              capsErrs **errors)
{
  size_t linecap = 0;
  char   cpath[PATH_MAX], *str, *valstr, *line = NULL;
  FILE   *fp;

#ifdef DEBUG
  printf(" avlAIM/aimCalcOutput instance = %d  index = %d!\n", inst, index);
#endif

  *errors        = NULL;
  val->vals.real = 0.0;

  /* set the "search" string */
  if (index == 1) {
    str = "CLtot =";
  } else {
    str = "CDtot =";
  }

  /* get where we are and set the path to our output */
  (void) getcwd(cpath, PATH_MAX);
  if (chdir(apath) != 0) {
#ifdef DEBUG
    printf(" avlAIM/aimCalcOutput Cannot chdir to %s!\n", apath);
#endif
    return CAPS_DIRERR;
  }

  /* open the AVL output file */
  fp = fopen("avlOutput.txt", "r");
  if (fp == NULL) {
#ifdef DEBUG
    printf(" avlAIM/aimCalcOutput Cannot open Output file!\n");
#endif
    chdir(cpath);
    return CAPS_DIRERR;
  }

  /* scan the file for the string */
  valstr = NULL;
/*@-unrecog@*/
  while (getline(&line, &linecap, fp) >= 0) {
/*@+unrecog@*/
    if (line == NULL) continue;
    valstr = strstr(line, str);
    if (valstr != NULL) {
      /* found it -- get the value */
      sscanf(&valstr[7], "%lf", &val->vals.real);
      break;
    }
  }
  
  /* restore the path we came in with */
  chdir(cpath);
  fclose(fp);
  if (line != NULL) free(line);
  
  if (valstr == NULL) {
#ifdef DEBUG
    printf(" avlAIM/aimCalcOutput Cannot find %s in Output file!\n", str);
#endif
    return CAPS_NOTFOUND;
  }

  return CAPS_SUCCESS;
}


/* 
 * since this AIM does not support field variables or CAPS bounds, the 
 * following functions are not required to be filled in except for aimDiscr 
 * which just stores away our bodies and aimFreeDiscr that does some cleanup 
 * when CAPS terminates
 */

int
aimFreeDiscr(/*@null@*/ capsDiscr *discr)
{
  int i;
  
#ifdef DEBUG
  if (discr == NULL) {
    printf(" avlAIM/aimFreeDiscr w/ NULL!\n");
  } else {
    printf(" avlAIM/aimFreeDiscr instance = %d!\n", discr->instance);
  }
#endif
  
  if (discr == NULL) {
    
    /* signal for complete AIM cleanup */
    for (i = 0; i < nInstance; i++) {
      if (instances[i].bodies != NULL) EG_free(instances[i].bodies);
    }
    EG_free(instances);
    nInstance = 0;
    instances = NULL;
    
  }
  
  return CAPS_SUCCESS;
}


int
aimDiscr(/*@unused@*/ char *bname, capsDiscr *discr)
{
  int stat, inst;
  
  inst = discr->instance;
#ifdef DEBUG
  printf(" avlAIM/aimDiscr: bname = %s, instance = %d!\n", bname, inst);
#endif
  if ((inst < 0) || (inst >= nInstance)) return CAPS_BADINDEX;
  
  stat = aimFreeDiscr(discr);
  if (stat != CAPS_SUCCESS) return stat;
  discr->nBodies = instances[inst].nbody;
  discr->bodies  = instances[inst].bodies;
  
  return CAPS_SUCCESS;
}


int
aimLocateElement(/*@unused@*/ capsDiscr *discr, /*@unused@*/ double *params,
                 /*@unused@*/ double *param,    /*@unused@*/ int *eIndex,
                 /*@unused@*/ double *bary)
{
#ifdef DEBUG
  printf(" avlAIM/aimLocateElement instance = %d!\n", discr->instance);
#endif
  
  return CAPS_SUCCESS;
}


int
aimTransfer(/*@unused@*/ capsDiscr *discr, /*@unused@*/ char *name,
            /*@unused@*/ int npts, /*@unused@*/ int rank,
            /*@unused@*/ double *data, /*@unused@*/ char **units)
{
#ifdef DEBUG
  printf(" avlAIM/aimTransfer name = %s  instance = %d  npts = %d/%d!\n",
         name, discr->instance, npts, rank);
#endif

  return CAPS_SUCCESS;
}


int
aimInterpolation(/*@unused@*/ capsDiscr *discr, /*@unused@*/ char *name,
                 /*@unused@*/ int eIndex, /*@unused@*/ double *bary,
                 /*@unused@*/ int rank, /*@unused@*/ double *data,
                 /*@unused@*/ double *result)
{
#ifdef DEBUG
  printf(" avlAIM/aimInterpolation  %s  instance = %d!\n",
         name, discr->instance);
#endif

  return CAPS_SUCCESS;
}


int
aimInterpolateBar(/*@unused@*/ capsDiscr *discr, /*@unused@*/ char *name,
                  /*@unused@*/ int eIndex, /*@unused@*/ double *bary,
                  /*@unused@*/ int rank, /*@unused@*/ double *r_bar,
                  /*@unused@*/ double *d_bar)
{
#ifdef DEBUG
  printf(" avlAIM/aimInterpolateBar  %s  instance = %d!\n",
         name, discr->instance);
#endif
  
  return CAPS_SUCCESS;
}


int
aimIntegration(/*@unused@*/ capsDiscr *discr, /*@unused@*/ char *name,
               /*@unused@*/ int eIndex, /*@unused@*/ int rank,
               /*@unused@*/ /*@null@*/ double *data, /*@unused@*/ double *result)
{
#ifdef DEBUG
  printf(" avlAIM/aimIntegration  %s  instance = %d!\n",
         name, discr->instance);
#endif
  
  return CAPS_SUCCESS;
}


int
aimIntegrateBar(/*@unused@*/ capsDiscr *discr, /*@unused@*/ char *name,
                /*@unused@*/ int eIndex, /*@unused@*/ int rank,
                /*@unused@*/ double *r_bar, /*@unused@*/ double *d_bar)
{
#ifdef DEBUG
  printf(" avlAIM/aimIntegrateBar  %s  instance = %d!\n",
         name, discr->instance);
#endif

  return CAPS_SUCCESS;
}
