/*
 *      CAPS: Computational Aircraft Prototype Syntheses
 *
 *             AVL AIM tester
 *
 *      Copyright 2014-2015, Massachusetts Institute of Technology
 *      Licensed under The GNU Lesser General Public License, version 2.1
 *      See http://www.opensource.org/licenses/lgpl-2.1.php
 *
 */

#include "caps.h"
#include <string.h>

#ifdef WIN32
#define unlink   _unlink
#define getcwd   _getcwd
#define PATH_MAX _MAX_PATH
#else
#include <limits.h>
#include <unistd.h>
#endif


extern void printObjects( capsObj object, int indent );


int main(int argc, char *argv[])
{
  int            i, j, n, stat, fidelity, nFields, *ranks, dirty, nErr, nLines;
  char           *apath, **fnames, **lines, *name, *lunits, endf[10];
  char           check[9] = "AVL   c>";
/*@-unrecog@*/
  char           cpath[PATH_MAX];
/*@+unrecog@*/
  double         val;
  capsObj        pobj, aobj, obj, link, parent, mobj;
  capsErrs       *errors;
  capsOwn        current;
  ego            body;
  enum capsoType type;
  enum capssType subtype;
  FILE           *fp;

  if ((argc < 2) || (argc > 4)) {
    printf(" usage: avlTest filename apath [capsname]!\n");
    return 1;
  }
  
  stat = caps_open(argv[1], "avlTest", &pobj);
  if (stat != CAPS_SUCCESS) {
    printf(" caps_open = %d\n", stat);
    return 1;
  }
  
  /* look at the bodies and report units */
  stat = caps_size(pobj, BODIES, NONE, &n);
  if (stat != CAPS_SUCCESS) {
    printf(" caps_size on Bodies = %d\n", stat);
  } else {
    for (i = 1; i <= n; i++) {
      stat = caps_bodyByIndex(pobj, i, &body, &lunits);
      if (stat != CAPS_SUCCESS) {
        printf(" caps_bodyByIndex = %d for Body %d!\n", stat, i);
      } else {
        printf(" Body %d has length units = %s\n", i, lunits);
      }
    }
  }
  
  stat = caps_load(pobj, "avlAIM", argv[2], 13, &aobj);
  if (stat != CAPS_SUCCESS) {
    printf(" caps_load = %d\n", stat);
    caps_close(pobj);
    return 1;
  }
  stat = caps_info(pobj, &name, &type, &subtype, &link, &parent, &current);
  if (stat != CAPS_SUCCESS) {
    printf(" caps_info on Problem = %d\n", stat);
    caps_close(pobj);
    return 1;
  }
  
  /* find & set Mach number input */
  stat = caps_childByName(aobj, VALUE, ANALYSISIN, "Mach", &mobj);
  if (stat != CAPS_SUCCESS) {
    printf(" caps_childByName = %d\n", stat);
  } else {
    val  = 0.4;
    stat = caps_setValue(mobj, 1, (void *) &val);
    if (stat != CAPS_SUCCESS)
      printf(" caps_setValue = %d\n", stat);
  }
  
  /* get analysis info */
  stat = caps_analysisInfo(aobj, &fidelity, &apath, &nFields, &fnames,
                           &ranks, &dirty);
  if (stat != CAPS_SUCCESS) {
    printf(" caps_analysisInfo = %d\n", stat);
    caps_close(pobj);
    return 1;
  }
  printf("\n Fidelity =  %d", fidelity);
  printf("\n APath    = %s", apath);
  printf("\n Fields   =");
  for (i = 0; i < nFields; i++) printf("  %s (%d)", fnames[i], ranks[i]);
  printf("\n Dirty    =  %d\n", dirty);

  /* do the analysis */
  if (dirty != 0) {
    stat = caps_preAnalysis(aobj, &nErr, &errors);
    if (stat != CAPS_SUCCESS)
      printf(" caps_preAnalysis = %d\n", stat);
    if (errors != NULL) {
      for (i = 0; i < nErr; i++) {
        stat = caps_errorInfo(errors, i+1, &obj, &nLines, &lines);
        if (stat != CAPS_SUCCESS) {
          printf(" CAPS Error: caps_errorInfo[%d] = %d!\n", i+1, stat);
          continue;
        }
        for (j = 0; j < nLines; j++) printf(" %s\n", lines[j]);
      }
      caps_freeError(errors);
    }
    
    dirty = 0;
    stat  = caps_analysisInfo(aobj, &fidelity, &apath, &nFields, &fnames,
                              &ranks, &dirty);
    if (stat == CAPS_SUCCESS) printf("\n Dirty    =  %d\n", dirty);
    
    /* execute avl and run the post */
    if (dirty == 5) {
      
      (void) getcwd(cpath, PATH_MAX);
      if (chdir(apath) != 0) {
        printf(" ERROR: Cannot change directory to -> %s\n", apath);
        caps_close(pobj);
        return 1;
      }
      unlink("avlOutput.txt");                     /* remove any old output */
      printf(" Running AVL!\n");
      /* note: both avlInout.txt & caps.avl were generated by preAnalysis */
      system("avl caps < avlInput.txt > avlOutput.txt");
      /* check if ran correctly before invoking postAnalysis */
      fp = fopen("avlOutput.txt", "r");
      if (fp == NULL) {
        printf(" ERROR: Cannot open avlOutput.txt!\n");
        dirty = -1;
      } else {
        fseek(fp, -10, SEEK_END);
        (void) fread(endf, sizeof(char), 10, fp);
        fclose(fp);
        endf[8] = 0;
/*      for (i = 0; i < 9; i++)
          printf(" %d: %d %c  %d %c\n", 
                 i, endf[i], endf[i], check[i], check[i]); */
        if (strcmp(check,endf) != 0) {
          printf(" ERROR: End of avlOutput.txt not correct!\n");
          dirty = -1;
        }
      }
      chdir(cpath);

      if (dirty != -1) {
        /* run post if we are OK */
        stat = caps_postAnalysis(aobj, current, &nErr, &errors);
        if (stat != CAPS_SUCCESS)
          printf(" caps_postAnalysis = %d\n", stat);
        if (errors != NULL) {
          for (i = 0; i < nErr; i++) {
            stat = caps_errorInfo(errors, i+1, &obj, &nLines, &lines);
            if (stat != CAPS_SUCCESS) {
              printf(" CAPS Error: caps_errorInfo[%d] = %d!\n", i+1, stat);
              continue;
            }
            for (j = 0; j < nLines; j++) printf(" %s\n", lines[j]);
          }
          caps_freeError(errors);
        }
      }
      dirty = 0;
      stat  = caps_analysisInfo(aobj, &fidelity, &apath, &nFields, &fnames,
                                &ranks, &dirty);
      if (stat == CAPS_SUCCESS) printf(" Dirty    =  %d\n", dirty);
    }
  }
  
  /* output what we have */
  printf("\n");
  printObjects(pobj, 0);
  printf("\n");
  
  /* checkpoint the Problem */
  if (argc == 4) {
    stat = caps_save(pobj, argv[3]);
    if (stat != CAPS_SUCCESS) printf(" caps_save = %d\n", stat);
  }

  stat = caps_close(pobj);
  if (stat != CAPS_SUCCESS) printf(" caps_close = %d\n", stat);
  
  return 0;
}