/*
 *      CAPS: Computational Aircraft Prototype Syntheses
 *
 *             AIM Dynamic Subsystem
 *
 *      Copyright 2014-2015, Massachusetts Institute of Technology
 *      Licensed under The GNU Lesser General Public License, version 2.1
 *      See http://www.opensource.org/licenses/lgpl-2.1.php
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "capsTypes.h"

#ifdef WIN32
#include <windows.h>
#define DLL HINSTANCE
#else
#include <dlfcn.h>
#define DLL void *
#endif

#define MAXANAL 64


typedef int (*DLLFunc) (void);
typedef int (*aimI)    (int, /*@null@*/ capsValue *, int *, int **, int *,
                        int *, int *, char ***, int **);
typedef int (*aimD)    (char *, capsDiscr *);
typedef int (*aimF)    (/*@null@*/ capsDiscr *);
typedef int (*aimL)    (capsDiscr *, double *, double *, int *, double *);
typedef int (*aimIn)   (int, int, char **, capsValue *);
typedef int (*aimA)    (int, int, /*@null@*/ ego *, char *,
                        /*@null@*/ capsValue *, capsErrs **);
typedef int (*aimO)    (int, int, char **, capsValue *);
typedef int (*aimC)    (int, char *, int, capsValue *, capsErrs **);
typedef int (*aimT)    (capsDiscr *, char *, int, int, double *, char **);
typedef int (*aimP)    (capsDiscr *, char *, int, double *, int, double *,
                        double *);
typedef int (*aimG)    (capsDiscr *, char *, int, int, /*@null@*/ double *,
                        double *);

static char *aimName[MAXANAL];
static DLL   aimDLL[MAXANAL];
static aimI  aimInit[MAXANAL];
static aimD  aimDiscr[MAXANAL];
static aimF  aimFreeD[MAXANAL];
static aimL  aimLoc[MAXANAL];
static aimIn aimInput[MAXANAL];
static aimA  aimPAnal[MAXANAL];
static aimO  aimOutput[MAXANAL];
static aimC  aimCalc[MAXANAL];
static aimT  aimXfer[MAXANAL];
static aimP  aimIntrp[MAXANAL];
static aimP  aimIntrpBar[MAXANAL];
static aimG  aimIntgr[MAXANAL];
static aimG  aimIntgrBar[MAXANAL];

static int aim_nAnal = 0;


/* ************************* Utility Functions ***************************** */

static /*@null@*/ DLL aimDLopen(const char *name)
{
    int  i, len;
    DLL  dll;
    char *full;

    if (name == NULL) {
        printf(" Information: Dynamic Loader invoked with NULL name!\n");
        return NULL;
    }
    len  = strlen(name);
    full = (char *) malloc((len+5)*sizeof(char));
    if (full == NULL) {
        printf(" Information: Dynamic Loader MALLOC Error!\n");
        return NULL;
    }
    for (i = 0; i < len; i++) full[i] = name[i];
    full[len  ] = '.';
#ifdef WIN32
    full[len+1] = 'D';
    full[len+2] = 'L';
    full[len+3] = 'L';
    full[len+4] =  0;
    dll = LoadLibrary(full);
#else
    full[len+1] = 's';
    full[len+2] = 'o';
    full[len+3] =  0;
    dll = dlopen(full, RTLD_NOW /* RTLD_LAZY */);
#endif

    if (!dll) {
        printf(" Information: Dynamic Loader Error for %s\n", full);
#ifndef WIN32
        /*@-mustfreefresh@*/
        printf("              %s\n", dlerror());
        /*@+mustfreefresh@*/
#endif
        free(full);
        return NULL;
    }

    free(full);
    return dll;
}


static void aimDLclose(/*@only@*/ DLL dll)
{
#ifdef WIN32
    FreeLibrary(dll);
#else
    dlclose(dll);
#endif
}


static DLLFunc aimDLget(DLL dll, const char *symname, const char *name)
{
    DLLFunc data;

#ifdef WIN32
    data = (DLLFunc) GetProcAddress(dll, symname);
#else
    /*@-castfcnptr@*/
    data = (DLLFunc) dlsym(dll, symname);
    /*@+castfcnptr@*/
#endif
    if ((data == NULL) && (name != NULL))
        printf(" Information: Couldn't get symbol %s in %s\n", symname, name);
    return data;
}


static int aimDLoaded(const char *name)
{
    int i;

    for (i = 0; i < aim_nAnal; i++)
        if (strcmp(name, aimName[i]) == 0) return i;

    return -1;
}


static int aimDYNload(const char *name)
{
    int i, len, ret;
    DLL dll;

    if (aim_nAnal >= MAXANAL) {
        printf(" Information: Number of AIMs > %d!\n", MAXANAL);
        return EGADS_INDEXERR;
    }
    dll = aimDLopen(name);
    if (dll == NULL) return EGADS_NULLOBJ;

    ret              = aim_nAnal;
    aimInit[ret]     = (aimI)  aimDLget(dll, "aimInitialize",     name);
    aimDiscr[ret]    = (aimD)  aimDLget(dll, "aimDiscr",          name);
    aimFreeD[ret]    = (aimF)  aimDLget(dll, "aimFreeDiscr",      name);
    aimLoc[ret]      = (aimL)  aimDLget(dll, "aimLocateElement",  name);
    aimInput[ret]    = (aimIn) aimDLget(dll, "aimInputs",         name);
    aimPAnal[ret]    = (aimA)  aimDLget(dll, "aimPreAnalysis",    name);
    aimOutput[ret]   = (aimO)  aimDLget(dll, "aimOutputs",        name);
    aimCalc[ret]     = (aimC)  aimDLget(dll, "aimCalcOutput",     name);
    aimXfer[ret]     = (aimT)  aimDLget(dll, "aimTransfer",       name);
    aimIntrp[ret]    = (aimP)  aimDLget(dll, "aimInterpolation",  name);
    aimIntrpBar[ret] = (aimP)  aimDLget(dll, "aimInterpolateBar", name);
    aimIntgr[ret]    = (aimG)  aimDLget(dll, "aimIntegration",    name);
    aimIntgrBar[ret] = (aimG)  aimDLget(dll, "aimIntegrateBar",   name);
    if ((aimInit[ret]     == NULL) || (aimDiscr[ret] == NULL) ||
        (aimFreeD[ret]    == NULL) || (aimLoc[ret]   == NULL) ||
        (aimInput[ret]    == NULL) || (aimPAnal[ret] == NULL) ||
        (aimOutput[ret]   == NULL) || (aimCalc[ret]  == NULL) ||
        (aimXfer[ret]     == NULL) || (aimIntrp[ret] == NULL) ||
        (aimIntrpBar[ret] == NULL) || (aimIntgr[ret] == NULL) ||
        (aimIntgrBar[ret] == NULL)) {
        aimDLclose(dll);
        return EGADS_EMPTY;
    }

    len  = strlen(name) + 1;
    aimName[ret] = (char *) malloc(len*sizeof(char));
    if (aimName[ret] == NULL) {
        aimDLclose(dll);
        return EGADS_MALLOC;
    }
    for (i = 0; i < len; i++) aimName[ret][i] = name[i];
    aimDLL[ret] = dll;
    aim_nAnal++;

    return ret;
}


/* ************************* Exposed Functions ***************************** */

int
aim_Initialize(const char *analysisName,
               int        ngeomIn,      /* number of GeometryIn Values */
               /*@null@*/
               capsValue  *geomIn,      /* pointer to GeometryIn structures */
               int        *nFid,        /* returned number of fidelities */
               int        **fidels,     /* returned pointer to fidelities */
               int        *nIn,         /* returned number of inputs */
               int        *nOut,        /* returned number of outputs */
               int        *nField,      /* returned number of DataSet fields */
               char       ***fnames,    /* returned pointer to field strings */
               int        **ranks)      /* returned pointer to field ranks */
{
    int i;

    i = aimDLoaded(analysisName);

    if (i == -1) {
        i = aimDYNload(analysisName);
        if (i < 0) return i;
    }

    return aimInit[i](ngeomIn, geomIn, nFid, fidels, nIn, nOut, nField, fnames,
                      ranks);
}


int
aim_Index(char *analysisName)
{
  return aimDLoaded(analysisName);
}


int
aim_Discr(char      *analysisName,
          char      *bname,             /* Bound name */
          capsDiscr *discr)             /* the structure to fill */
{
    int i;

    i = aimDLoaded(analysisName);
    if (i == -1) return EGADS_NOTFOUND;

    return aimDiscr[i](bname, discr);
}


int
aim_FreeDiscr(char      *analysisName,
              capsDiscr *discr)         /* the structure to free up */
{
    int i;
  
    i = aimDLoaded(analysisName);
    if (i == -1) return EGADS_NOTFOUND;
  
    return aimFreeD[i](discr);
}


int
aim_LocateElement(char      *analysisName,
                  capsDiscr *discr,     /* the input discrete structure */
                  double    *params,    /* the input global parametric space */
                  double    *param,     /* the requested position */
                  int       *eIndex,    /* the returned element index */
                  double    *bary)      /* the barycentric coordinates */
{
    int i;
  
    i = aimDLoaded(analysisName);
    if (i == -1) return EGADS_NOTFOUND;
  
    return aimLoc[i](discr, params, param, eIndex, bary);
}


int
aim_LocateElIndex(int       i,
                  capsDiscr *discr,     /* the input discrete structure */
                  double    *params,    /* the input global parametric space */
                  double    *param,     /* the requested position */
                  int       *eIndex,    /* the returned element index */
                  double    *bary)      /* the barycentric coordinates */
{
  if ((i < 0) || (i >= aim_nAnal)) return EGADS_RANGERR;
  
  return aimLoc[i](discr, params, param, eIndex, bary);
}


int
aim_Inputs(const char *analysisName,
           int        instance,         /* instance index */
           int        index,            /* the input index [1-nIn] */
           char       **ainame,         /* pointer to the returned name */
           capsValue  *defaultVal)      /* pointer to default value (filled) */
{
    int i;
  
    i = aimDLoaded(analysisName);
    if (i == -1) return EGADS_NOTFOUND;
  
    return aimInput[i](instance, index, ainame, defaultVal);
}


int
aim_PreAnalysis(char      *analysisName,
                int       instance,     /* instance index */
                int       nBodies,      /* the number of EGADS Bodies */
                /*@null@*/
                ego       *bodies,      /* pointer to the EGADS Body objects */
                char      *apath,       /* filesystem path to write file(s) */
                /*@null@*/
                capsValue *inputs,      /* complete suite of analysis inputs */
                capsErrs  **errors)     /* returned pointer to error info */
{
    int i;
  
    i = aimDLoaded(analysisName);
    if (i == -1) return EGADS_NOTFOUND;
  
    return aimPAnal[i](instance, nBodies, bodies, apath, inputs, errors);
}


int
aim_Outputs(const char *analysisName,
            int        instance,        /* instance index */
            int        index,           /* the output index [1-nOut] */
            char       **aoname,        /* pointer to the returned name */
            capsValue  *formVal)        /* pointer to form/units (filled) */
{
    int i;
  
    i = aimDLoaded(analysisName);
    if (i == -1) return EGADS_NOTFOUND;
  
    return aimOutput[i](instance, index, aoname, formVal);
}


int
aim_CalcOutput(char      *analysisName,
               int       instance,      /* instance index */
               char      *apath,        /* filesystem path to write file(s) */
               int       index,         /* the output index [1-nOut] */
               capsValue *value,        /* pointer to value struct to fill */
               capsErrs  **errors)      /* returned pointer to error info */
{
    int i;
  
    i = aimDLoaded(analysisName);
    if (i == -1) return EGADS_NOTFOUND;
  
    return aimCalc[i](instance, apath, index, value, errors);
}


int
aim_Transfer(char      *analysisName,
             capsDiscr *discr,          /* the input discrete structure */
             char      *name,           /* the field name */
             int       npts,            /* number of points to be filled */
             int       rank,            /* the rank of the data */
             double    *data,           /* pointer to the memory to be filled */
             char      **units)         /* the units string returned */
{
    int i;
  
    i = aimDLoaded(analysisName);
    if (i == -1) return EGADS_NOTFOUND;
  
    return aimXfer[i](discr, name, npts, rank, data, units);
}


int
aim_Interpolation(char      *analysisName,
                  capsDiscr *discr,     /* the input discrete structure */
                  char      *name,      /* the dataset name */
                  int       eIndex,     /* the input element (1-bias) */
                  double    *bary,      /* the barycentric coordinates */
                  int       rank,       /* the rank of the data */
                  double    *data,      /* global discrete support */
                  double    *result)    /* the result (rank in length) */
{
    int i;
  
    i = aimDLoaded(analysisName);
    if (i == -1) return EGADS_NOTFOUND;
  
    return aimIntrp[i](discr, name, eIndex, bary, rank, data, result);
}


int
aim_InterpolIndex(int       i,
                  capsDiscr *discr,     /* the input discrete structure */
                  char      *name,      /* the dataset name */
                  int       eIndex,     /* the input element (1-bias) */
                  double    *bary,      /* the barycentric coordinates */
                  int       rank,       /* the rank of the data */
                  double    *data,      /* global discrete support */
                  double    *result)    /* the result (rank in length) */
{
  if ((i < 0) || (i >= aim_nAnal)) return EGADS_RANGERR;
  
  return aimIntrp[i](discr, name, eIndex, bary, rank, data, result);
}


int
aim_InterpolateBar(char      *analysisName,
                   capsDiscr *discr,    /* the input discrete structure */
                   char      *name,     /* the dataset name */
                   int       eIndex,    /* the input element (1-bias) */
                   double    *bary,     /* the barycentric coordinates */
                   int       rank,      /* the rank of the data */
                   double    *r_bar,    /* input d(objective)/d(result) */
                   double    *d_bar)    /* returned d(objective)/d(data) */
{
    int i;
  
    i = aimDLoaded(analysisName);
    if (i == -1) return EGADS_NOTFOUND;
  
    return aimIntrpBar[i](discr, name, eIndex, bary, rank, r_bar, d_bar);
}


int
aim_InterpolIndBar(int       i,
                   capsDiscr *discr,    /* the input discrete structure */
                   char      *name,     /* the dataset name */
                   int       eIndex,    /* the input element (1-bias) */
                   double    *bary,     /* the barycentric coordinates */
                   int       rank,      /* the rank of the data */
                   double    *r_bar,    /* input d(objective)/d(result) */
                   double    *d_bar)    /* returned d(objective)/d(data) */
{
  if ((i < 0) || (i >= aim_nAnal)) return EGADS_RANGERR;
  
  return aimIntrpBar[i](discr, name, eIndex, bary, rank, r_bar, d_bar);
}


int
aim_Integration(char      *analysisName,
                capsDiscr *discr,       /* the input discrete structure */
                char      *name,        /* the dataset name */
                int       eIndex,       /* the input element (1-bias) */
                int       rank,         /* the rank of the data */
                /*@null@*/
                double    *data,        /* global discrete support */
                double    *result)      /* the result (rank in length) */
{
    int i;
  
    i = aimDLoaded(analysisName);
    if (i == -1) return EGADS_NOTFOUND;
  
    return aimIntgr[i](discr, name, eIndex, rank, data, result);
}


int
aim_IntegrIndex(int       i,
                capsDiscr *discr,       /* the input discrete structure */
                char      *name,        /* the dataset name */
                int       eIndex,       /* the input element (1-bias) */
                int       rank,         /* the rank of the data */
                /*@null@*/
                double    *data,        /* global discrete support */
                double    *result)      /* the result (rank in length) */
{
  if ((i < 0) || (i >= aim_nAnal)) return EGADS_RANGERR;
  
  return aimIntgr[i](discr, name, eIndex, rank, data, result);
}


int
aim_IntegrateBar(char      *analysisName,
                 capsDiscr *discr,      /* the input discrete structure */
                 char      *name,       /* the dataset name */
                 int       eIndex,      /* the input element (1-bias) */
                 int       rank,        /* the rank of the data */
                 double    *r_bar,      /* input d(objective)/d(result) */
                 double    *d_bar)      /* returned d(objective)/d(data) */
{
    int i;
  
    i = aimDLoaded(analysisName);
    if (i == -1) return EGADS_NOTFOUND;
  
    return aimIntgrBar[i](discr, name, eIndex, rank, r_bar, d_bar);
}


int
aim_IntegrIndBar(int       i,
                 capsDiscr *discr,      /* the input discrete structure */
                 char      *name,       /* the dataset name */
                 int       eIndex,      /* the input element (1-bias) */
                 int       rank,        /* the rank of the data */
                 double    *r_bar,      /* input d(objective)/d(result) */
                 double    *d_bar)      /* returned d(objective)/d(data) */
{
  if ((i < 0) || (i >= aim_nAnal)) return EGADS_RANGERR;
  
  return aimIntgrBar[i](discr, name, eIndex, rank, r_bar, d_bar);
}


void aim_cleanupAll()
{
    int i;

    if (aim_nAnal == 0) return;

    for (i = 0; i < aim_nAnal; i++) {
        aimFreeD[i](NULL);
        free(aimName[i]);
        aimDLclose(aimDLL[i]);
    }

    aim_nAnal = 0;
}
