/*
 *      CAPS: Computational Aircraft Prototype Syntheses
 *
 *             Simple AIM Example Code
 *
 *      Copyright 2014-2015, Massachusetts Institute of Technology
 *      Licensed under The GNU Lesser General Public License, version 2.1
 *      See http://www.opensource.org/licenses/lgpl-2.1.php
 *
 */

#include <string.h>
#include <math.h>
#include "capsTypes.h"

#define DEBUG

#define CROSS(a,b,c)      a[0] = (b[1]*c[2]) - (b[2]*c[1]);\
                          a[1] = (b[2]*c[0]) - (b[0]*c[2]);\
                          a[2] = (b[0]*c[1]) - (b[1]*c[0])
#define DOT(a,b)         (a[0]*b[0] + a[1]*b[1] + a[2]*b[2])


  typedef struct {
    int nbody;
    ego *bodies;
  } aimStorage;


  static aimStorage *instances = NULL;
  static int        nInstance  = 0;

  extern char *EG_strdup( /*@null@*/ const char *str );



int
aimInitialize(int ngIn, /*@unused@*/ /*@null@*/ capsValue *gIn,
              int *nFid, int **fids, int *nIn, int *nOut,
              int *nFields, char ***fnames, int **ranks)
{
  int        *ints, ret;
  char       **strs;
  aimStorage *tmp;

#ifdef DEBUG
  printf("\n simpleAIM/aimInitialize   ngIn = %d!\n", ngIn);
#endif
  *nFid    = 2;
  ints     = (int *) EG_alloc(2*sizeof(int));
  if (ints == NULL) return EGADS_MALLOC;
  ints[0]  =  5;
  ints[1]  = 14;
  *fids    = ints;

  /* specify the number of analysis input and out "parameters" */
  *nIn     = 1;
  *nOut    = 1;

  /* specify the field variables this analysis can generate */
  *nFields = 2;
  ints     = (int *) EG_alloc(2*sizeof(int));
  if (ints == NULL) {
    EG_free(*fids);
    *fids = NULL;
    return EGADS_MALLOC;
  }
  ints[0]  = 1;
  ints[1]  = 3;
  *ranks   = ints;
  strs     = (char **) EG_alloc(2*sizeof(char *));
  if (strs == NULL) {
    EG_free(*ranks);
    EG_free(*fids);
    *ranks = NULL;
    *fids  = NULL;
    return EGADS_MALLOC;
  }
  strs[0]  = EG_strdup("scalar");
  strs[1]  = EG_strdup("vector");
  *fnames  = strs;
  
  /* get the instance storage */
  if (instances == NULL) {
    instances = (aimStorage *) EG_alloc(sizeof(aimStorage));
    if (instances == NULL) {
      EG_free(*fnames);
      EG_free(*ranks);
      EG_free(*fids);
      *fnames = NULL;
      *ranks  = NULL;
      *fids   = NULL;
      return EGADS_MALLOC;
    }
  } else {
    tmp = (aimStorage *) EG_reall(instances, (nInstance+1)*sizeof(aimStorage));
    if (tmp == NULL) {
      EG_free(*fnames);
      EG_free(*ranks);
      EG_free(*fids);
      *fnames = NULL;
      *ranks  = NULL;
      *fids   = NULL;
      return EGADS_MALLOC;
    }
    instances = tmp;
  }
  instances[nInstance].nbody  = 0;
  instances[nInstance].bodies = NULL;
  
  /* return an index for the instance generated */
  ret = nInstance;
  nInstance++;
  return ret;
}


int
aimFreeDiscr(/*@null@*/ capsDiscr *discr)
{
  int i;

#ifdef DEBUG
  if (discr == NULL) {
    printf(" simpleAIM/aimFreeDiscr w/ NULL!\n");
  } else {
    printf(" simpleAIM/aimFreeDiscr instance = %d!\n", discr->instance);
  }
#endif

  if (discr == NULL) {
    
    /* signal for complete AIM cleanup */
    for (i = 0; i < nInstance; i++) {
      if (instances[i].bodies != NULL) EG_free(instances[i].bodies);
    }
    EG_free(instances);
    nInstance = 0;
    instances = NULL;

  } else {

    /* free up this capsDiscr */
    if (discr->pairs != NULL) EG_free(discr->pairs);
    if (discr->types != NULL) {
      for (i = 0; i < discr->nTypes; i++) {
        if (discr->types[i].gst  != NULL) EG_free(discr->types[i].gst);
        if (discr->types[i].tris != NULL) EG_free(discr->types[i].tris);
      }
      EG_free(discr->types);
    }
    if (discr->topos != NULL) {
      for (i = 0; i < discr->nTopos; i++)
        if (discr->topos[i].nFaces > 2)
          if (discr->topos[i].fIndices.multi != NULL)
            EG_free(discr->topos[i].fIndices.multi);
      EG_free(discr->topos);
    }
    if (discr->topoTop != NULL) EG_free(discr->topoTop);
    if (discr->points  != NULL) EG_free(discr->points);
    if (discr->pntsTop != NULL) EG_free(discr->pntsTop);
    if (discr->faceUVs != NULL) EG_free(discr->faceUVs);
    if (discr->faceTop != NULL) EG_free(discr->faceTop);
    if (discr->elems   != NULL) EG_free(discr->elems);
    if (discr->elmsTop != NULL) EG_free(discr->elmsTop);
    if (discr->ptrm    != NULL) EG_free(discr->ptrm);

    discr->nPairs   = 0;
    discr->pairs    = NULL;
    discr->nFaceUVs = 0;
    discr->faceUVs  = NULL;
    discr->faceTop  = NULL;
    discr->nTopos   = 0;
    discr->topos    = NULL;
    discr->topoTop  = NULL;
    discr->nPoints  = 0;
    discr->points   = NULL;
    discr->pntsTop  = NULL;
    discr->nVerts   = 0;
    discr->verts    = NULL;
    discr->vrtsTop  = NULL;
    discr->nTypes   = 0;
    discr->types    = NULL;
    discr->nElems   = 0;
    discr->elems    = NULL;
    discr->elmsTop  = NULL;
    discr->nDtris   = 0;
    discr->dtris    = NULL;
    discr->ptrm     = NULL;

  }

  return CAPS_SUCCESS;
}


static int
aimLookup(capsDiscr *discr, int *vid, int ptype, int pindex,
          int start, int k, int kk)
{
  int i, index, *ints;

  vid[3*k  ] = ptype;
  vid[3*k+1] = pindex;
  vid[3*k+2] = kk+1;

  /* interior Face vertex */
  if (ptype == -1) return 1;

  /* Edge or Node vertex */
  for (i = start; i < k; i++)
    if ((ptype == vid[3*i]) && (pindex == vid[3*i+1])) {
      /* found -- set index and add to the Face list */
      vid[3*k+2] = i+1;
      index      = vid[3*i+2] - 1;
      if (discr->topos[index].nFaces == 1) {
        discr->topos[index].fIndices.faces[1] = k+1;
      } else if (discr->topos[index].nFaces == 2) {
        ints = (int *) EG_alloc(3*sizeof(int));
        if (ints == NULL) return 1;
        ints[0] = discr->topos[index].fIndices.faces[0];
        ints[1] = discr->topos[index].fIndices.faces[1];
        ints[2] = k+1;
        discr->topos[index].fIndices.multi = ints;
      } else {
        ints = (int *) EG_reall(discr->topos[index].fIndices.multi,
                                (discr->topos[index].nFaces+1)*sizeof(int));
        if (ints == NULL) return 1;
        ints[discr->topos[index].nFaces] = k+1;
        discr->topos[index].fIndices.multi = ints;
      }
      discr->topos[index].nFaces++;
      return 0;
    }
  
  return 1;
}


int
aimDiscr(char *bname, capsDiscr *discr)
{
  int          i, j, k, kk, n, ibod, stat, inst, nFace, atype, alen, tlen, elen;
  int          npts, ntris, start, *storage, *vid = NULL;
  double       size, box[6], params[3];
  const int    *ints, *ptype, *pindex, *tris, *nei;
  const double *reals, *xyz, *uv, *exyz, *t;
  const char   *string;
  ego          *faces, *tess;

  inst = discr->instance;
#ifdef DEBUG
  printf(" simpleAIM/aimDiscr: bname = %s, instance = %d!\n", bname, inst);
#endif
  if ((inst < 0) || (inst >= nInstance)) return CAPS_BADINDEX;
  
  stat = aimFreeDiscr(discr);
  if (stat != CAPS_SUCCESS) return stat;
  discr->nBodies = instances[inst].nbody;
  discr->bodies  = instances[inst].bodies;

  tess = NULL;
  if (discr->nBodies > 0) {
    tess = (ego *) EG_alloc(discr->nBodies*sizeof(ego));
    if (tess == NULL) return EGADS_MALLOC;
    for (i = 0; i < discr->nBodies; i++) tess[i] = NULL;
  }

  /* find any faces with our boundary marker */
  for (n = i = 0; i < discr->nBodies; i++) {
    stat = EG_getBodyTopos(discr->bodies[i], NULL, FACE, &nFace, &faces);
    if (stat != EGADS_SUCCESS) {
      printf(" simpleAIM: getBodyTopos (Face) = %d for Body %d!\n", stat, i+1);
      if (tess != NULL) {
        for (j = 0; j < i; j++)
          if (tess[j] != NULL) EG_deleteObject(tess[j]);
        EG_free(tess);
      }
      return stat;
    }
    for (j = 0; j < nFace; j++) {
      stat = EG_attributeRet(faces[j], "capsBoundary", &atype, &alen,
                             &ints, &reals, &string);
      if (stat  != EGADS_SUCCESS)     continue;
      if (atype != ATTRSTRING)        continue;
      if (strcmp(string, bname) != 0) continue;
#ifdef DEBUG
      printf(" simpleAIM/aimDiscr: Body %d/Face %d matches %s!\n",
             i+1, j+1, bname);
#endif
      /* tessellate if necessary -- simulate the AIM meshing */
      if (tess[i] == NULL) {
        stat = EG_getBoundingBox(discr->bodies[i], box);
        if (stat != EGADS_SUCCESS) {
          printf(" simpleAIM: getBoundingBox = %d for Body %d!\n", stat, i+1);
          for (j = 0; j < i; j++)
            if (tess[j] != NULL) EG_deleteObject(tess[j]);
          EG_free(tess);
          return stat;
        }
                                  size = box[3]-box[0];
        if (size < box[4]-box[1]) size = box[4]-box[1];
        if (size < box[5]-box[2]) size = box[5]-box[2];
        params[0] =  0.025*size;
        params[1] =  0.001*size;
        params[2] = 15.0;
        stat = EG_makeTessBody(discr->bodies[i], params, &tess[i]);
        if (stat != EGADS_SUCCESS) {
          printf(" simpleAIM: makeTessBody = %d for Body %d!\n", stat, i+1);
          for (j = 0; j < i; j++)
            if (tess[j] != NULL) EG_deleteObject(tess[j]);
          EG_free(tess);
          return stat;
        }
      }
      n++;
    }
    EG_free(faces);
  }
  if (n == 0) {
    printf(" simpleAIM/aimDiscr: No Faces match %s!\n", bname);
    if (tess != NULL) EG_free(tess);
    return CAPS_SUCCESS;
  }
  
  /* store away the faces */
  stat = EGADS_MALLOC;
  discr->nPairs = n;
  discr->pairs  = (int *) EG_alloc(2*n*sizeof(int));
  if (discr->pairs == NULL) goto bail;
  if (discr->nBodies != 1) {
    discr->elmsTop = (int *) EG_alloc(discr->nBodies*sizeof(int));
    if (discr->elmsTop == NULL) goto bail;
    discr->pntsTop = (int *) EG_alloc(discr->nBodies*sizeof(int));
    if (discr->pntsTop == NULL) goto bail;
    discr->topoTop = (int *) EG_alloc(discr->nBodies*sizeof(int));
    if (discr->topoTop == NULL) goto bail;
    discr->faceTop = (int *) EG_alloc(discr->nBodies*sizeof(int));
    if (discr->faceTop == NULL) goto bail;
  }
  npts = ntris = 0;
  for (n = i = 0; i < discr->nBodies; i++) {
    stat = EG_getBodyTopos(discr->bodies[i], NULL, FACE, &nFace, &faces);
    if (stat != EGADS_SUCCESS) {
      printf(" simpleAIM: getBodyTopos (Face) = %d for Body %d!\n", stat, i+1);
      goto bail;
    }
    for (j = 0; j < nFace; j++) {
      stat = EG_attributeRet(faces[j], "capsBoundary", &atype, &alen,
                             &ints, &reals, &string);
      if (stat  != EGADS_SUCCESS)     continue;
      if (atype != ATTRSTRING)        continue;
      if (strcmp(string, bname) != 0) continue;
      discr->pairs[2*n  ] = i+1;
      discr->pairs[2*n+1] = j+1;
      n++;
      stat = EG_getTessFace(tess[i], j+1, &alen, &xyz, &uv, &ptype, &pindex,
                                          &tlen, &tris, &nei);
      if (stat != EGADS_SUCCESS) {
        printf(" simpleAIM: EG_getTessFace %d = %d for Body %d!\n",
               j+1, stat, i+1);
        continue;
      }
      npts  += alen;
      ntris += tlen;
    }
    EG_free(faces);
    if (discr->nBodies != 1) {
      discr->elmsTop[i] = ntris;
      discr->faceTop[i] = npts;
    }
  }
  if ((npts == 0) || (ntris == 0)) {
#ifdef DEBUG
    printf(" simpleAIM/aimDiscr: ntris = %d, npts = %d!\n", ntris, npts);
#endif
    stat = CAPS_SOURCEERR;
    goto bail;
  }
  discr->nElems   = ntris;
  discr->nFaceUVs = npts;

  /* specify our single element type */
  stat = EGADS_MALLOC;
  discr->nTypes = 1;
  discr->types  = (capsEleType *) EG_alloc(sizeof(capsEleType));
  if (discr->types == NULL) goto bail;
  discr->types[0].nref  = 3;
  discr->types[0].ndata = 0;            /* data at geom reference positions */
  discr->types[0].ntri  = 1;
  discr->types[0].nmat  = 0;            /* match points at geom ref positions */
  discr->types[0].tris  = NULL;
  discr->types[0].gst   = NULL;
  discr->types[0].dst   = NULL;
  discr->types[0].matst = NULL;
  
  discr->types[0].tris   = (int *) EG_alloc(3*sizeof(int));
  if (discr->types[0].tris == NULL) goto bail;
  discr->types[0].tris[0] = 1;
  discr->types[0].tris[1] = 2;
  discr->types[0].tris[2] = 3;
  
  discr->types[0].gst   = (double *) EG_alloc(6*sizeof(double));
  if (discr->types[0].gst == NULL) goto bail;
  discr->types[0].gst[0] = 0.0;
  discr->types[0].gst[1] = 0.0;
  discr->types[0].gst[2] = 1.0;
  discr->types[0].gst[3] = 0.0;
  discr->types[0].gst[4] = 0.0;
  discr->types[0].gst[5] = 1.0;
  
  /* get the tessellation and
     make up a simple linear continuous triangle discretization */
#ifdef DEBUG
  printf(" simpleAIM/aimDiscr: ntris = %d, npts = %d!\n", ntris, npts);
#endif
  discr->faceUVs = (capsFaceUV *) EG_alloc(npts*sizeof(capsFaceUV));
  if (discr->faceUVs == NULL) goto bail;
  discr->points = (capsPoints *) EG_alloc(npts*sizeof(capsPoints));
  if (discr->points == NULL) goto bail;
  discr->topos = (capsTopos *) EG_alloc(npts*sizeof(capsTopos));
  if (discr->topos == NULL) goto bail;
  for (i = 0; i < npts; i++) discr->topos[i].nFaces = 0;
  vid = (int *) EG_alloc(3*npts*sizeof(int));
  if (vid == NULL) goto bail;
  for (ibod = kk = k = i = 0; i < n; i++) {
    while (discr->pairs[2*i] != ibod+1) {
      if (discr->nBodies != 1) discr->pntsTop[ibod] = discr->topoTop[ibod] = kk;
      ibod++;
    }
    stat = EG_getTessFace(tess[discr->pairs[2*i]-1], discr->pairs[2*i+1], &alen,
                          &xyz, &uv, &ptype, &pindex, &tlen, &tris, &nei);
    if (stat != EGADS_SUCCESS) continue;
    start = 0;
    if ((discr->nBodies != 1) && (ibod != 0)) start = discr->faceTop[ibod-1];
    for (j = 0; j < alen; j++, k++) {
      discr->faceUVs[k].fIndex           = discr->pairs[2*i+1];
      discr->faceUVs[k].uv[0]            = uv[2*j  ];
      discr->faceUVs[k].uv[1]            = uv[2*j+1];
      if (aimLookup(discr, vid, ptype[j],pindex[j], start, k,kk) == 0) continue;
      discr->points[kk].xyz[0]           = xyz[3*j  ];
      discr->points[kk].xyz[1]           = xyz[3*j+1];
      discr->points[kk].xyz[2]           = xyz[3*j+2];
      discr->points[kk].tIndex           = kk+1;
      discr->topos[kk].t                 = 0.0;
      discr->topos[kk].lTopo             = 0;
      discr->topos[kk].nFaces            = 1;
      discr->topos[kk].fIndices.faces[0] = k+1;
      if (ptype[j] == 0) {
        discr->topos[kk].lTopo           = -pindex[j];
      } else if (ptype[j] > 0) {
        discr->topos[kk].lTopo           =  pindex[j];
        stat = EG_getTessEdge(tess[discr->pairs[2*i]-1], pindex[j], &elen,
                              &exyz, &t);
        if (stat != EGADS_SUCCESS) {
          printf(" simpleAIM: EG_getTessEdge %d = %d for Body %d!\n",
                 pindex[j], stat, discr->pairs[2*i]);
        } else {
          discr->topos[k].t              =  t[ptype[j]-1];
        }
      }
      kk++;
    }
  }
  if (discr->nBodies != 1)
    discr->pntsTop[discr->nBodies-1] = discr->topoTop[discr->nBodies-1] = kk;
  discr->nPoints = discr->nTopos = kk;
  
  discr->elems = (capsElement *) EG_alloc(ntris*sizeof(capsElement));
  if (discr->elems == NULL) goto bail;
  storage = (int *) EG_alloc(3*ntris*sizeof(int));
  if (storage == NULL) goto bail;
  discr->ptrm = storage;
  npts = ntris = 0;
  for (k = i = 0; i < n; i++) {
    stat = EG_getTessFace(tess[discr->pairs[2*i]-1], discr->pairs[2*i+1], &alen,
                          &xyz, &uv, &ptype, &pindex, &tlen, &tris, &nei);
    if (stat != EGADS_SUCCESS) continue;
    for (j = 0; j < tlen; j++, k++) {
      discr->elems[k].tIndex    = 1;
      discr->elems[k].eIndex    = discr->pairs[2*i+1];
/*@-immediatetrans@*/
      discr->elems[k].gIndices  = &storage[3*k];
/*@+immediatetrans@*/
      discr->elems[k].dIndices  = NULL;
      storage[3*k  ] = tris[3*j  ] + npts;
      storage[3*k+1] = tris[3*j+1] + npts;
      storage[3*k+2] = tris[3*j+2] + npts;
    }
    npts  += alen;
    ntris += tlen;
  }
  /* correct the point indexing */
  for (i = 0; i < 3*ntris; i++) {
    j          = storage[i]-1;
    storage[i] = vid[3*j+2];
  }

  /* free up our stuff */
  if (vid != NULL) EG_free(vid);
  if (tess != NULL) {
    for (j = 0; j < discr->nBodies; j++)
      if (tess[j] != NULL) EG_deleteObject(tess[j]);
    EG_free(tess);
  }

  return CAPS_SUCCESS;
  
bail:
  if (vid != NULL) EG_free(vid);
  aimFreeDiscr(discr);
  if (tess != NULL) {
    for (j = 0; j < discr->nBodies; j++)
      if (tess[j] != NULL) EG_deleteObject(tess[j]);
    EG_free(tess);
  }

  return stat;
}


int
aimLocateElement(/*@unused@*/ capsDiscr *discr, /*@unused@*/ double *params,
                 /*@unused@*/ double *param,    /*@unused@*/ int *eIndex,
                 /*@unused@*/ double *bary)
{
  /* eIndex and bary on input are consistent with linear triangles therefore
     nothing needs to be done! */
  
  return CAPS_SUCCESS;
}


int
aimInputs(int inst, int index, char **ainame, capsValue *defval)
{
#ifdef DEBUG
  printf(" simpleAIM/aimInputs instance = %d  index = %d!\n", inst, index);
#endif
  *ainame = EG_strdup("simpleAIMin");

  defval->type      = Double;
  defval->vals.real = 5.0;
  defval->units     = EG_strdup("cm");
  return CAPS_SUCCESS;
}


int
aimPreAnalysis(int inst, int nbody, /*@null@*/ ego *bodies,
               /*@unused@*/ char *apath,
               /*@unused@*/ /*@null@*/ capsValue *inputs, capsErrs **errs)
{
  int i;

#ifdef DEBUG
  printf(" simpleAIM/aimPreAnalysis instance = %d  nbody = %d!\n", inst, nbody);
#endif
  
  *errs = NULL;
  /* save away the current geometry for the instance */
  if (instances[inst].bodies != NULL) EG_free(instances[inst].bodies);
  instances[inst].bodies = NULL;
  instances[inst].nbody  = 0;
  if ((nbody > 0) && (bodies != NULL)) {
    instances[inst].bodies = (ego *) EG_alloc(nbody*sizeof(ego));
    if (instances[inst].bodies == NULL) return EGADS_MALLOC;
    for (i = 0; i < nbody; i++) instances[inst].bodies[i] = bodies[i];
    instances[inst].nbody = nbody;
  }

  return CAPS_SUCCESS;
}


int
aimOutputs(int inst, int index, char **aoname, capsValue *form)
{
#ifdef DEBUG
  printf(" simpleAIM/aimOutputs instance = %d  index = %d!\n", inst, index);
#endif
  *aoname = EG_strdup("simpleAIMout");

  form->type = Double;
  return CAPS_SUCCESS;
}


int
aimCalcOutput(int inst, /*@unused@*/ char *apath, int index, capsValue *val,
              capsErrs **errors)
{
#ifdef DEBUG
  printf(" simpleAIM/aimCalcOutput instance = %d  index = %d!\n", inst, index);
#endif

  *errors = NULL;

  val->vals.real = 5.5;
  return CAPS_SUCCESS;
}


int
aimTransfer(capsDiscr *discr, char *name, int npts, int rank, double *data,
            /*@unused@*/ char **units)
{
  int i, j;

#ifdef DEBUG
  printf(" simpleAIM/aimTransfer name = %s  instance = %d  npts = %d/%d!\n",
         name, discr->instance, npts, rank);
#endif
  
  /* fill in with our coordinates -- for now */
  for (i = 0; i < npts; i++)
    for (j = 0; j < rank; j++)
      data[rank*i+j] = discr->points[i].xyz[j];

  return CAPS_SUCCESS;
}


int
aimInterpolation(capsDiscr *discr, /*@unused@*/ char *name, int eIndex,
                 double *bary, int rank, double *data, double *result)
{
  int    in[3], i;
  double we[3];

#ifdef DEBUG
  printf(" simpleAIM/aimInterpolation  %s  instance = %d!\n",
         name, discr->instance);
#endif

  we[0] = 1.0 - bary[0] - bary[1];
  we[1] = bary[0];
  we[2] = bary[1];
  in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
  in[1] = discr->elems[eIndex-1].gIndices[1] - 1;
  in[2] = discr->elems[eIndex-1].gIndices[2] - 1;
  for (i = 0; i < rank; i++)
    result[i] = data[rank*in[0]+i]*we[0] + data[rank*in[1]+i]*we[1] +
                data[rank*in[2]+i]*we[2];

  return CAPS_SUCCESS;
}


int
aimInterpolateBar(capsDiscr *discr, /*@unused@*/ char *name, int eIndex,
                  double *bary, int rank, double *r_bar, double *d_bar)
{
  int    in[3], i;
  double we[3];
  
#ifdef DEBUG
  printf(" simpleAIM/aimInterpolateBar  %s  instance = %d!\n",
         name, discr->instance);
#endif

  we[0] = 1.0 - bary[0] - bary[1];
  we[1] = bary[0];
  we[2] = bary[1];
  in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
  in[1] = discr->elems[eIndex-1].gIndices[1] - 1;
  in[2] = discr->elems[eIndex-1].gIndices[2] - 1;
  for (i = 0; i < rank; i++) {
/*  result[i] = data[rank*in[0]+i]*we[0] + data[rank*in[1]+i]*we[1] +
                data[rank*in[2]+i]*we[2];  */
    d_bar[rank*in[0]+i] += we[0]*r_bar[i];
    d_bar[rank*in[1]+i] += we[1]*r_bar[i];
    d_bar[rank*in[2]+i] += we[2]*r_bar[i];
  }
  
  return CAPS_SUCCESS;
}


int
aimIntegration(capsDiscr *discr, /*@unused@*/ char *name, int eIndex, int rank,
               /*@null@*/ double *data, double *result)
{
  int    i, in[3];
  double x1[3], x2[3], x3[3], area;

#ifdef DEBUG
  printf(" simpleAIM/aimIntegration  %s  instance = %d!\n",
         name, discr->instance);
#endif
  
  /* element indices */
  
  in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
  in[1] = discr->elems[eIndex-1].gIndices[1] - 1;
  in[2] = discr->elems[eIndex-1].gIndices[2] - 1;
  
  x1[0] = discr->points[in[1]].xyz[0] - discr->points[in[0]].xyz[0];
  x2[0] = discr->points[in[2]].xyz[0] - discr->points[in[0]].xyz[0];
  x1[1] = discr->points[in[1]].xyz[1] - discr->points[in[0]].xyz[1];
  x2[1] = discr->points[in[2]].xyz[1] - discr->points[in[0]].xyz[1];
  x1[2] = discr->points[in[1]].xyz[2] - discr->points[in[0]].xyz[2];
  x2[2] = discr->points[in[2]].xyz[2] - discr->points[in[0]].xyz[2];
  CROSS(x3, x1, x2);
  area  = sqrt(DOT(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */
  if (data == NULL) {
    *result = 3.0*area;
    return CAPS_SUCCESS;
  }
  
  for (i = 0; i < rank; i++)
    result[i] = (data[rank*in[0]+i] + data[rank*in[1]+i] +
                 data[rank*in[2]+i])*area;
  
  return CAPS_SUCCESS;
}


int
aimIntegrateBar(capsDiscr *discr, /*@unused@*/ char *name, int eIndex, int rank,
                double *r_bar, double *d_bar)
{
  int    i, in[3];
  double x1[3], x2[3], x3[3], area;

#ifdef DEBUG
  printf(" simpleAIM/aimIntegrateBar  %s  instance = %d!\n",
         name, discr->instance);
#endif
  
  /* element indices */
  
  in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
  in[1] = discr->elems[eIndex-1].gIndices[1] - 1;
  in[2] = discr->elems[eIndex-1].gIndices[2] - 1;
  
  x1[0] = discr->points[in[1]].xyz[0] - discr->points[in[0]].xyz[0];
  x2[0] = discr->points[in[2]].xyz[0] - discr->points[in[0]].xyz[0];
  x1[1] = discr->points[in[1]].xyz[1] - discr->points[in[0]].xyz[1];
  x2[1] = discr->points[in[2]].xyz[1] - discr->points[in[0]].xyz[1];
  x1[2] = discr->points[in[1]].xyz[2] - discr->points[in[0]].xyz[2];
  x2[2] = discr->points[in[2]].xyz[2] - discr->points[in[0]].xyz[2];
  CROSS(x3, x1, x2);
  area  = sqrt(DOT(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */
  
  for (i = 0; i < rank; i++) {
    /*  result[i] = (data[rank*in[0]+i] + data[rank*in[1]+i] +
     data[rank*in[2]+i])*area;  */
    d_bar[rank*in[0]+i] += area*r_bar[i];
    d_bar[rank*in[1]+i] += area*r_bar[i];
    d_bar[rank*in[2]+i] += area*r_bar[i];
  }

  return CAPS_SUCCESS;
}