/*
 ************************************************************************
 *                                                                      *
 * OpenCSM.h -- header for OpenCSM.c                                    *
 *                                                                      *
 *              Written by John Dannenhoffer @ Syracuse University      *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2010/2015  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#ifndef _OPENCSM_H_
#define _OPENCSM_H_

#define OCSM_MAJOR_VERSION  1
#define OCSM_MINOR_VERSION 07

#include <time.h>

/*
 ************************************************************************
 *                                                                      *
 * Definitions                                                          *
 *                                                                      *
 ************************************************************************
 */

/*
Node
    - a location in 3D space that serves as the terminus for one or
      more Edges

Edge
    - is associated with a 3D curve (if not degenerate)
    - has a range of parametric coordinates, where tmin <= t <= tmax
    - the positive orientation is going from tmin to tmax
    - has a Node for tmin and for tmax
    - if the Nodes at tmin and tmax are the same, the Edge forms a
      closed Loop (that is, is periodic) or is degenerate (if tmin
      equals tmax); otherwise it is open

Loop
    - free standing collection of one or more connected Edges with
      associated senses
    - if the Loop is closed, each of the corresponding Nodes is
      associated with exactly two Edges
    - if the Loop is open, the intermediate Nodes are each associated
      with two Edges and the Nodes at the ends each correspond to one
      Edge
    - the sense of the Loop is associated with the order of the Edges
      in the Loop and their associated senses

Face
    - a surface bounded by one or more Loops with associated senses
    - there may be only one outer Loop (sense = 1) and any number
      of inner Loops (sense = -1)
    - all associated Loops must be closed

Shell
    - a collection of one of more connected Faces
    - if all the Edges associated with a Shell are used by exactly two
      Faces in the Shell, the Shell is closed (manifold) and it
      segregates regionds of 3D space; otherwise the Shell is open

NODE  (nodeBody)
    - a single Node (which can be used as the terminus in operations
      such as rule, blend, and loft)
    - formed by creating an empty Sketch

WIRE  (wireBody)
    - a single Loop

SHEET  (sheetBody)
    - a single Shell that can be either non-manifold (open) or manifold
      (closed)

SOLID  (solidBody)
    - a manifold collection of one or more closed Shells with associated
      senses
    - there may be only one outer Shell (sense = 1) and any number of
      inner Shells (sense = -1)
 */

/*
 ************************************************************************
 *                                                                      *
 * CSM file format                                                      *
 *                                                                      *
 ************************************************************************
 */

/*
The .csm file contains a series of statements.

If a line contains a hash (#), all characters starting at the hash
   are ignored.

If a line contains a backslash, all characters starting at the
   backslash are ignored and the next line is appended; spaces at
   the beginning of the next line are treated normally.

All statements begin with a keyword (described below) and must
   contain at least the indicated number of arguments.

Any CSM statement can be used except the interface statement.

Extra arguments in a statement are discarded.  If one wants to add
   a comment, it is recommended to begin it with a hash (#) in case
   optional arguments are added in future releases.

Any statements after an 'end' statement are ignored.

All arguments must not contain any spaces or must be enclosed
   in a pair of double quotes (for example, "a + b").

Parameters are evaluated in the order that they appear in the
   file, using MATLAB-like syntax (see 'Expression rules' below).

During the build process, OpenCSM maintains a LIFO 'Stack' that
   can contain Bodies and Sketches.

The csm statements are executed in a stack-like way, taking their
   inputs from the Stack and depositing their results onto the Stack.

The default name for each Branch is 'Brch_xxxxxx', where xxxxxx
   is a unique sequence number.

Special characters:
   #          introduces comment
   "          ignore spaces until following "
   \          ignore this and following characters and concatenate next line
   <space>    separates arguments in .csm file (except between " and ")

   0-9        digits used in numbers and in names
   A-Z a-z    letters used in names
   _ : @      characters used in names (see rule for names)
   .          decimal separator (used in numbers), introduces dot-suffixes
                 (in names)
   ,          separates function arguments and row/column in subscripts
   ;          multi-value item separator
   ( )        groups expressions and function arguments
   [ ]        specifies subscripts in form [row,column] or [index]
   + - * / ^  arithmetic operators
   $          as first character, forces argument not to be evaluated
                 (used internally)
   @          as first character, introduces @-parameters (see below)
   !          evaluate before setting attribute

   ~          not used
   %          not used
   &          not used
   =          not used
   { }        not used
   '          not used
   < >        not used
   ?          not used
*/

/*
 ************************************************************************
 *                                                                      *
 * UDC file format                                                      *
 *                                                                      *
 ************************************************************************
 */

/*
A .udc file follows the rules of a .csm file, EXCEPT:

Zero or more interface statements must preceed any other non-comment
    CSM statement.

Any CSM statement can be used except the dimension, despmtr, lbound,
    and ubound statements.

Set statements define parameters only within the .udc file (that is,
   parameters have local scope).

Parameters defined outside the .udc file are not available, except those
   passed in via interface statements.

.udc files can be nested to a depth of 10 levels.

.udc files are executed via a udprim statement.
*/

/*
 ************************************************************************
 *                                                                      *
 * Valid CSM statements                                                 *
 *                                                                      *
 ************************************************************************
 */

/*
arc       xend yend zend dist $plane=xy
          use:    create a new circular arc to the new point, with a specified
                     offset distance
          pops:   -
          pushes: -
          notes:  Sketch must be open
                  solver may not be open
                  $plane must be xy, yz, or zx
                  point on circle is dist from midpoint in plane specified
                  if dist>0, sweep is counterclockwise
                  sensitivity computed w.r.t. xend, yend, zend, dist

assert    arg1 arg2
          use:    return error if arg1 and arg2 differ
          pops:   -
          pushes: -
          notes:  if (abs(arg1-arg2) > 1.e-6) return an error
                  cannot be attributed

attribute $attrName attrValue
          use:    sets an attribute for the entry on top of Stack
          pops:   any
          pushes: any
          notes:  Sketch may not be open
                  if first char of attrValue is '$', then do not evaluate
                  if attrValue is a parameter name, all elements are stored
                     in attribute
                  does not create a Branch
                  if before first Branch, then defines a global Attribute
                  is applied to a single Edge or Face if after a select
                     statement

bezier    x y z
          use:    add a Bezier control point
          pops:   -
          pushes: -
          notes:  Sketch must be open
                  solver may not be open
                  sensitivity computed w.r.t. x, y, z

blend     noseList=0 tailList=0 reorder=0
          use:    create a Body by blending through Sketches since mark
          pops:   Sketch1 ... mark
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  all Sketches must have the same number of Segments
                  if all Sketches are WIRE Bodies, then a SHEET is created
                     otherwise a SOLID is created
                  the first and/or last Sketch can be a point
                  noseList is either a 0 or the name of an 8-valued parameter
                  noseList is only applied if first Sketch is a point
                  tailList is either a 0 or the name of an 8-valued parameter
                  tailList is only applied if last  Sketch is a point
                  noseList and tailList contain
                     rad1;dx1;dy1;dz1;rad2;dx2;dy2;dz2
                  if reorder!=0 then Sketches are reordered to minimize Edge
                     lengths
                  first Sketch is unaltered if reorder>0
                  last  Sketch is unaltered if reorder<0
                  sensitivity computed w.r.t. noseList, tailList
                  sets up @-parameters
                  face-order is: base, end, seg1, seg2, ...

box       xbase ybase zbase dx dy dz
          use:    create a box Body
          pops:   -
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  sensitivity computed w.r.t. xbase, ybase, zbase, dx, dy, dz
                  computes Face, Edge, and Node sensitivities analytically
                  sets up @-parameters
                  face-order is: xmin, xmax, ymin, ymax, zmin, zmax

chamfer   radius edgeList=0
          use:    apply a chamfer to a Body
          pops:   Body
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  if previous operation is boolean, apply to all new Edges
                  edgeList=0 is the same as edgeList=[0;0]
                  edgeList is a multi-value Parameter or a semicolon-separated
                     list
                  pairs of edgeList entries are processed in order
                  pairs of edgeList entries are interpreted as follows:
                    col1  col2   meaning
                     =0    =0    add all Edges
                     >0    >0    add    Edges between iford=+icol1
                                                  and iford=+icol2
                     <0    <0    remove Edges between iford=-icol1
                                                  and iford=-icol2
                     >0    =0    add    Edges adjacent to iford=+icol1
                     <0    =0    remove Edges adjacent to iford=-icol1
                  sensitivity computed w.r.t. radius
                  sets up @-parameters
                  face-order is based upon order that is returned from EGADS

cirarc    xon yon zon xend yend zend
          use:    create a new circular arc, using the previous point
                     as well as the two points specified
          pops:   -
          pushes: -
          notes:  Sketch must be open
                  solver may not be open
                  sensitivity computed w.r.t. xon, yon, zon, xend, yend, zend

combine
          use:    combine Bodys since mark into next higher type
          pops:   Body1 ... mark
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  mark must be set
                  if all Bodys since mark are SHEETs
                     create SOLID from closed Shell
                  elseif all Bodys since mark are WIREs and are co-planar
                     create SHEET from closed Loop
                  endif
                  sets up @-parameters

cone      xvrtx yvrtx zvrtx xbase ybase zbase radius
          use:    create a cone Body
          pops:   -
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  sensitivity computed w.r.t. xvrtx, yvrtx, zvrtz, xbase, ybase,
                     zbase, radius
                  computes Face, Edge, and Node sensitivities analytically
                  sets up @-parameters
                  face-order is: (empty), base, umin, umax

connect   faceList1 faceList2
          use:    connects two Bodys with bridging FAces
          pops:   Body1 Body2
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  faceList1 and faceList2 must have the same length
                  all Faces in the faceLists must have 4 Edges
                  Faces within each faceList must be contiguous
                  faceList1[i] corresponds to faceList2[i]

cylinder  xbeg ybeg zbeg xend yend zend radius
          use:    create a cylinder Body
          pops:   -
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  sensitivity computed w.r.t. xbeg, ybeg, zbeg, xend, yend,
                     zend, radius
                  computes Face, Edge, and Node sensitivities analytically
                  sets up @-parameters
                  face-order is: beg, end, umin, umax

despmtr   $pmtrName values
          use:    define a (constant) EXTERNAL design Parameter
          pops:   -
          pushes: -
          notes:  Sketch may not be open
                  solver may not be open
                  may not be used in a .udc file
                  pmtrName can be in form 'name' or 'name[irow,icol]'
                  pmtrName must not start with '@'
                  pmtrName must not refer to an INTERNAL Parameter
                  pmtrName will be marked as EXTERNAL
                  if pmtrName already exists, value is not overwritten
                  pmtrName is used directly (without evaluation)
                  irow and icol cannot contain a comma or open bracket
                  values cannot refer to any other Parameter
                  if values has multiple values (separated by ;), then
                     any subscripts in pmtrName are ignored
                  values are defined across rows
                  if values is longer than Parameter size, extra values
                     are lost
                  if values is shorter than Parameter size, last value
                     is repeated
                  old values are overwritten
                  does not create a Branch
                  cannot be attributed

dimension $pmtrName nrow ncol despmtr=0
          use:    set up an array Parameter
          pops:   -
          pushes: -
          notes:  Sketch may not be open
                  solver may not be open
                  may not be used in a .udc file
                  nrow >= 1
                  ncol >= 1
                  pmtrName must not start with '@'
                  if despmtr=0, then marked as INTERNAL
                  if despmtr=1, then marked as EXTERNAL
                  old values are not overwritten
                  does not create a Branch
                  cannot be attributed

dump      $filename remove=0
          use:    write a file that contains the Body
          pops:   Body1 (if remove=1)
          pushes: -
          notes:  solver may not be open
                  if file exists, it is overwritten
                  filename is used directly (without evaluation)
                  if filename starts with '$/', it is prepended with path of
                     the .csm file
                  if remove=1, then Body1 is removed after dumping

end
          use:    signifies end of .csm or .udc file
          pops:   -
          pushes: -
          notes:  Sketch may not be open
                  solver may not be open
                  Bodys on Stack are returned last-in-first-out
                  cannot be attributed

extract   index
          use:    extract a Face of Node from a Body
          pops:   Body1
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  if     Body1=SOLID and index>0
                     create SHEET from +index'th Face of Body1
                  elseif Body1=SOLID and index<0
                     create WIRE from -index'th Edge of Body1
                  elseif Body1=SOLID and index=0
                     create SHEET from outer Shell of Body1
                  elseif Body1=SHEET and index>0
                     create SHEET from +index'th Face of Body1
                  elseif Body1=SHEET and index<0
                     create WIRE from -index'th Edge of Body1
                  elseif Body=SHEET and index=0
                     create WIRE from outer Loop of Body1
                     CURRENTLY NOT IMPLEMENTED
                  endif
                  sets up @-parameters

extrude   dx dy dz
          use:    create a Body by extruding a Sketch
          pops:   Sketch
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  if Sketch is a SHEET, then a SOLID is created
                  if Sketch is a WIRE, then a SHEET is created
                  sensitivity computed w.r.t. dx, dy, dz
                  computes Face sensitivities analytically
                  sets up @-parameters
                  face-order is: base, end, feat1, ...

fillet    radius edgeList=0
          use:    apply a fillet to a Body
          pops:   Body
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  if previous operation is boolean, apply to all new Edges
                  edgeList=0 is the same as edgeList=[0;0]
                  edgeList is a multi-value Parameter or a semicolon-separated
                     list
                  pairs of edgeList entries are processed in order
                  pairs of edgeList entries are interpreted as follows:
                    col1  col2   meaning
                     =0    =0    add all Edges
                     >0    >0    add    Edges between iford=+icol1
                                                  and iford=+icol2
                     <0    <0    remove Edges between iford=-icol1
                                                  and iford=-icol2
                     >0    =0    add    Edges adjacent to iford=+icol1
                     <0    =0    remove Edges adjacent to iford=-icol1
                  sensitivity computed w.r.t. radius
                  sets up @-parameters
                  face-order is based upon order that is returned from EGADS

hollow    thick faceList=0
          use:    hollow out a SOLID
          pops:   Body
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  if thick=0, then SOLID is converted to SHEET or SHEET to WIRE
                  if faceList=0, then create an offset Body instead
                  faceList is a multi-value Parameter or a semicolon-separated
                     list
                  pairs of faceList entries are processed in order
                  the first  entry in a pair indicates the Body when Face was
                     generated
                  the second entry in a pair indicates the face-order
                  sensitivity computed w.r.t. thick
                  sets up @-parameters
                  face-order is based upon order that is returned from EGADS

import    $filename bodynumber=1
          use:    import from filename
          pops:   -
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  filename is used directly (without evaluation)
                  if filename starts with '$$/', use path relative to .csm file
                  sets up @-parameters
                  face-order is based upon order in file

interface $argName $argType default
          use:    defines an argument for a .udc file
          pops:   -
          pushes: -
          notes:  only allowed in a .udc file
                  must be placed before any executable statement
                  argType must be "in", "out", or "dim"
                  if argType=="dim", then default contains number of elements
                  if argType=="dim", the default values are zero
                  cannot be attributed

intersect $order=none index=1 maxtol=0
          use:    perform Boolean intersection (Body2 & Body1)
          pops:   Body1 Body2
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  if     Body1=SOLID and Body2=SOLID
                     create SOLID that is common part of Body1 and Body2
                  elseif Body1=SOLID and Body2=SHEET
                     create SHEET that is the part of Body2 that is inside Body1
                     CURRENTLY NOT IMPLEMENTED
                  elseif Body1=SOLID and Body2=WIRE
                     create WIRE that is the part of Body2 that is inside Body1
                     CURRENTLY NOT IMPLEMENTED
                  elseif Body1=SHEET and Body2=SOLID
                     create SHEET that is the part of Body1 that is inside Body2
                  elseif Body1=SHEET and Body2=SHEET and Bodys are co-planar
                     create SHEET that is common part of Body1 and Body2
                     CURRENTLY NOT IMPLEMENTED
                  elseif Body1=SHEET and Body2=SHEET and Bodys are not co-planar
                     create WIRE at the intersection of Body1 and Body2
                     CURRENTLY NOT IMPLEMENTED
                  elseif Body1=SHEET and Body2=WIRE
                     create WIRE that is the part of Body2 that is inside Body1
                     CURRENTLY NOT IMPLEMENTED
                  elseif Body1=WIRE and Body2=SOLID
                     create WIRE that is the part of Body1 that is inside Body2
                     CURRENTLY NOT IMPLEMENTED
                  elseif Body1=WIRE and Body2=SHEET
                     create WIRE that is the part of Body1 that is inside Body2
                     CURRENTLY NOT IMPLEMENTED
                  endif
                  if intersection does not produce at least index Bodies, an
                     error is returned
                  order may be one of:
                     none    same order as returned from geometry engine
                     xmin    minimum xmin   is first
                     xmax    maximum xmax   is first
                     ymin    minimum ymin   is first
                     ymax    maximum ymax   is first
                     zmin    minimum zmin   is first
                     zmax    maximum zmax   is first
                     amin    minimum area   is first
                     amax    maximum area   is first
                     vmin    minimum volume is first
                     vmax    maximum volume is first
                  order is used directly (without evaluation)
                  if maxtol>0, then tolernce can be relaxed until successful
                  sets up @-parameters

join      toler=0
          use:    join two Bodies at a common Face
          pops:   Body1 Body2
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  if     Body1=SOLID and Body2=SOLID
                     create SOLID formed by joining Body1 and Body2 at
                        common Faces
                  elseif Body1=SHEET and Body2=SHEET
                     create SHEET formed by joining Body1 and Body2 at
                        common Edges
                     CURRENTLY NOT IMPLEMENTED
                  elseif Body1=WIRE and Body2=WIRE
                     create WIRE formed by joining Body1 and Body2 at
                        common end Node
                     CURRENTLY NOT IMPLEMENTED
                  endif
                  sets up @-parameters


lbound    $pmtrName bounds
          use:    defines a lower bound for a design Parameter
          pops:   -
          pushes: -
          notes:  Sketch may not be open
                  solver may not be open
                  may not be used in a .udc file
                  pmtrName must have been defined previously by despmtr
                     statement
                  pmtrName can be in form 'name' or 'name[irow,icol]'
                  bound is set by last lbound statement
                  name is used directly (without evaluation)
                  irow and icol cannot contain a comma or open bracket
                  bounds cannot refer to any other Parameter
                  does not create a Branch
                  if value of Parameter is smaller than bounds, a warning is
                     generated
                  cannot be attributed

linseg    x y z
          use:    create a new line segment, connecting the previous
                     and specified points
          pops:   -
          pushes: -
          notes:  Sketch must be open
                  solver may not be open
                  sensitivity computed w.r.t. x, y, z

loft      smooth
          use:    create a Body by lofting through Sketches since mark
          pops:   Sketch1 ... mark
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  all Sketches must have the same number of Segments
                  if Sketch is a SHEET, then a SOLID is created
                  if Sketch is a WIRE, then a SHEET is created
                  face-order is: base, end, feat1, ...
                  if NINT(smooth)=1, then sections are smoothed
                  the first and/or last Sketch can be a point

                  loft (through OpenCASCADE) is not very robust
                  use blend or rule if possible
                  sets up @-parameters
                  MAY BE DEPRECATED (use rule or blend)

macbeg    imacro
          use:    marks the start of a macro
          pops:   -
          pushes: -
          notes:  Sketch may not be open
                  solver may not be open
                  imacro must be between 1 and 100
                  cannot overwrite a previous macro
                  cannot be attributed
                  MAY BE DEPRECATED (use udprim)

macend
          use:    ends a macro
          pops:   -
          pushes: -
          notes:  cannot be attributed
                  MAY BE DEPRECATED (use udprim)

mark
          use:    used to identify groups such as in rule, loft, or blend
          pops:   -
          pushes: -
          notes:  Sketch may not be open
                  solver may not be open
                  cannot be attributed

mirror    nx ny nz dist=0
          use:    mirrors entry on top of Stack
          pops:   any
          pushes: any
          notes:  Sketch may not be open
                  solver may not be open
                  normal of the mirror plane is given by nx,ny,nz
                  mirror plane is dist from origin
                  sensitivity computed w.r.t. nx, ny, nz, dist
                  sets up @-parameters

name      $branchName
          use:    names the entry on top of Stack
          pops:   any
          pushes: any
          notes:  Sketch may not be open
                  does not create a Branch

patbeg    $pmtrName ncopy
          use:    execute the pattern multiple times
          pops:   -
          pushes: -
          notes:  solver may not be open
                  pattern contains all statement up to the matching patend
                  pmtrName must not start with '@'
                  pmtrName takes values from 1 to ncopy (see below)
                  pmtrName is used directly (without evaluation)
                  cannot be attributed

patbreak  expr
          use:    break out of inner-most pattern
          pops:   -
          pushes: -
          notes:  solver may not be open
                  there must be a matching patbeg
                  skip to branch after matching patend if expr>0

patend
          use:    mark the end of a pattern
          pops:   -
          pushes: -
          notes:  solver may not be open
                  there must be a matching patbeg for each patend

project   x y z dx dy dz
          use:    find the first projection from given point in given direction
          pops:   -
          pushes: -
          notes:  Sketch may not be open
                  solver may not be open
                  over-writes the following @-parameters: @iface, @xcg,
                     @ycg, and @zcg
                  cannot be attributed

recall    imacro
          use:    recalls copy of macro from a storage location imacro
          pops:   -
          pushes: any
          notes:  Sketch may not be open
                  solver may not be open
                  storage location imacro must have been previously filled by
                     a macro statement
                  MAY BE DEPRECATED (use udprim)

reorder   ishift iflip=0
          use:    change the order of Edges in a Body
          pops:   Body1
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  Body1 must be either WIRE or SHEET Body
                  Body1 must contain 1 Loop
                  if the Loop is open, ishift must be 0

restore   $name index=0
          use:    restores Body that was previously stored
          pops:   -
          pushes: any
          notes:  Sketch may not be open
                  solver may not be open
                  $name is used directrly (without evaluation)
                  error results if nothing has been stored in name

revolve   xorig yorig zorig dxaxis dyaxis dzaxis angDeg
          use:    create a Body by revolving a Sketch around an axis
          pops:   Sketch
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  if Sketch is a SHEET, then a SOLID is created
                  if Sketch is a WIRE, then a SHEET is created
                  sensitivity computed w.r.t. xorig, yorig, zorig, dxaxis,
                     dyaxis, dzaxis, andDeg
                  sets up @-parameters
                  face-order is: (base), (end), feat1, ...

rotatex   angDeg yaxis zaxis
          use:    rotates entry on top of Stack around x-like axis
          pops:   any
          pushes: any
          notes:  Sketch may not be open
                  solver may not be open
                  sensitivity computed w.r.t. angDeg, yaxis, zaxis
                  sets up @-parameters

rotatey   angDeg zaxis xaxis
          use:    rotates entry on top of Stack around y-like axis
          pops:   any
          pushes: any
          notes:  Sketch may not be open
                  solver may not be open
                  sensitivity computed w.r.t. angDeg, zaxis, xaxis
                  sets up @-parameters

rotatez   angDeg xaxis yaxis
          use:    rotates entry on top of Stack around z-like axis
          pops:   any
          pushes: any
          notes:  Sketch may not be open
                  solver may not be open
                  sensitivity computed w.r.t. angDeg, xaxis, yaxis
                  sets up @-parameters

rule      reorder=0
          use:    create a Body by creating ruled surfaces through Sketches
                     since mark
          pops:   Sketch1 ... mark
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  if reorder!=0 then Sketches are reordered to minimize Edge
                     lengths
                  first Sketch is unaltered if reorder>0
                  last  Sketch is unaltered if reorder<0
                  all Sketches must have the same number of Segments
                  if all Sketches are WIRE Bodies, then a SHEET is created
                     otherwise a SOLID is created
                  the first and/or last Sketch can be a point
                  computes Face sensitivities analytically
                  sets up @-parameters
                  face-order is: base, end, seg1a, seg1b, ..., seg2a, ...

scale     fact
          use:    scales entry on top of Stack
          pops:   any
          pushes: any
          notes:  Sketch may not be open
                  solver may not be open
                  sensitivity computed w.r.t. fact
                  sets up @-parameters

select    $type arg1 ...
          use:    selects entity for which @-parameters are evaluated
          pops:   -
          pushes: -
          notes:  if     arguments are: "body"
                     sets @ibody to @nbody
                     sets @iface to -1
                     sets @iedge to -1
                     stes @inode to -1
                  elseif arguments are: "body ibody"
                     sets @ibody to ibody
                     sets @iface to -1
                     sets @iedge to -1
                     sets @inode to -1
                  elseif arguments are: "face iface"
                     uses @ibody
                     sets @iface to iface
                     sets @iedge to -1
                     sets @inode to -1
                  elseif arguments are: "face ibody1 iford1 iseq=1"
                     uses @ibody
                     sets @iface to Face in @ibody that matches ibody1/iford1
                     sets @iedge to -1
                     sets @inode to -1
                  elseif arguments are: "edge iedge"
                     uses @ibody
                     sets @iface to -1
                     sets @iedge to iedge
                     sets @inode to -1
                  elseif arguments are: "edge ibody1 iford1 ibody2 iford2 iseq=1"
                     uses @ibody
                     sets @iface to -1
                     sets @iedge to Edge in @ibody that adjoins Faces
                        ibody1/iford1 and ibody2/iford2
                     sets @inode to -1
//                elseif arguments are: "node inode"
//                   uses @ibody
//                   sets @iface to -1
//                   sets @iedge to -1
//                   sets @inode to inode
//                elseif arguments are: "node ibody1 iford1 ibody2 iford2
//                                                          ibody3 iford3 iseq=1"
//                   uses @ibody
//                   sets @iface to -1
//                   sets @iedge to -1
//                   sets @inode to Node in @ibody that adjoins
//                      Faces ibody1/iford1, ibody2/iford2, and ibody3/iford3
                  endif

                  Face specifications are stored in faceID attribute
                  Edge specifications are stored in edgeID attribute
//                Node specifications are stored in nodeID attribute
                  iseq selects from amongst multiple Faces/Edges/Nodes that
                     match the ibody/iford specifications
                  sets up @-parameters

set       $pmtrName exprs
          use:    define a (redefinable) INTERNAL Parameter
          pops:   -
          pushes: -
          notes:  solver may not be open
                  pmtrName can be in form 'name' or 'name[irow,icol]'
                  pmtrName must not start with '@'
                  pmtrName must not refer to an EXTERNAL Parameter
                  pmtrName will be marked as INTERNAL
                  pmtrName is used directly (without evaluation)
                  irow and icol cannot contain a comma or open bracket
                  if exprs has multiple values (separated by ;), then
                     any subscripts in pmtrName are ignored
                  exprs are defined across rows
                  if exprs is longer than Parameter size, extra exprs are lost
                  if exprs is shorter than Parameter size, last expr is repeated
                  cannot be attributed

skbeg     x y z
          use:    start a new Sketch with the given point
          pops:   -
          pushes: -
          notes:  opens Sketch
                  solver may not be open
                  sensitivity computed w.r.t. x, y, z
                  cannot be attributed

skcon     $type index1 index2=-1 $value=0
          use:    creates a Sketch constraint
          pops:   -
          pushes: -
          notes:  Sketch must be open
                  solver may not be open
                  may only follow skvar or another skcon statement
                  $type
                     X  ::x[index1]=value
                     Y  ::y[index1]=value
                     Z  ::z[index1]=value
                     P  segments adjacent to point index1 are perpendicular
                     T  segments adjacent to point index1 are tangent
                     A  segments adjacent to point index1 have
                                                           angle=$value (deg)
                     W  width:  ::x[index2]-::x[index1]=value  if plane==xy
                                ::y[index2]-::y[index1]=value  if plane==yz
                                ::z[index2]-::z[index1]=value  if plane==zx
                     D  depth:  ::y[index2]-::y[index1]=value  if plane==xy
                                ::z[index2]-::z[index1]=value  if plane==zx
                                ::x[index2]-::x[index1]=value  if plane==zx
                     H  segment from index1 and index2 is horizontal
                     V  segment from index1 and index2 is vertical
                     I  segment from index1 and index2 has
                                                     inclination=$value (deg)
                     L  segment from index1 and index2 has length=$value
                     R  cirarc  from index1 and index2 has radius=$value
                     S  cirarc  from index1 and index2 has sweep=$value (deg)
                  index=1 refers to point in skbeg statement
                  cannot be attributed

skend     wireonly=0
          use:    completes a Sketch
          pops:   -
          pushes: Sketch
          notes:  Sketch must be open
                  solver may not be open
                  if Sketch contains skvar/skcon, then Sketch variables are
                     updated first
                  if wireonly=0, all linsegs and cirarcs must be x-, y-, or
                     z-co-planar
                  if Sketch is     closed and wireonly=0, then a SHEET is created
                  if Sketch is     closed and wirebody=1, then a WIRE  is creates
                  if Sketch is not closed,                then a WIRE is created
                  if skend immediately follows skbeg, then a NODE is created
                     (which can be used at either end of a loft or blend)
                  closes Sketch

skvar     $type valList
          use:    create multi-valued Sketch variables and their initial values
          pops:   -
          pushes: -
          notes:  Sketch must be open
                  solver may not be open
                  may only follow skbeg statement
                  $type
                     xy valList contains ::x[1]; ::y[1]; ::d[1]; ::x[2]; ...
                     yz valList contains ::y[1]; ::z[1]; ::d[1]; ::y[2]; ...
                     zx valList contains ::z[1]; ::x[1]; ::d[1]; ::z[2]; ...
                  valList is a semicolon-separated list
                  valList must end with a semicolon
                  the number of entries in valList is taken from number of
                     semicolons
                  the number of entries in valList must be evenly divisible by 3
                  enter :d[i] as zero for linsegs
                  values of ::x[1], ::y[1], and ::z[1] are overwritten by
                     values in skbeg
                  cannot be attributed

solbeg    $varlist
          use:    starts a solver block
          pops:   -
          pushes: -
          notes:  solver must not be open
                  opens the solver
                  varlist is a list of semi-colon-separated INTERNAL parameters
                  varlist must end with a semi-colon
                  cannot be attributed

solcon    $expr
          use:    constraint used to set solver parameters
          pops:   -
          pushes: -
          notes:  Sketch must not be open
                  solver must be open
                  solend will drive expr to zero

solend
          use:    close a solver block
          pops:   -
          pushes: -
          notes:  Sketch must not be open
                  solver must be open
                  adjust parameters to drive constraints to zero
                  closes the solver

sphere    xcent ycent zcent radius
          use:    create a sphere Body
          pops:   -
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  sensitivity computed w.r.t. xcent, ycent, zcent, radius
                  computes Face, Edge, and Node sensitivities analytically
                  sets up @-parameters
                  face-order is: ymin, ymax

spline    x y z
          use:    add a point to a spline
          pops:   -
          pushes: -
          notes:  Sketch must be open
                  solver may not be open
                  sensitivity computed w.r.t. x, y, z

store     $name index=0 keep=0
          use:    stores Body on top of Stack
          pops:   any
          pushes: -
          notes:  Sketch may not be open
                  solver may not be open
                  $name is used directly (without evaluation)
                  previous Body in name/index is overwritten
                  if keep==1, the Body is not popped off stack
                  cannot be attributed

subtract  $order=none index=1 maxtol=0
          use:    perform Boolean subtraction (Body2 - Body1)
          pops:   Body1 Body2
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  if     Body1=SOLID and Body2=SOLID
                     create SOLID that is the part of Body1 that is
                        outside Body2
                  elseif Body1=SOLID and Body2=SHEET
                     create SOLID that is Body1 scribed with Edges at
                        intersection with Body2
                  elseif Body1=SHEET and Body2=SOLID
                     create SHEET that is part of Body1 that is outside Body2
                  elseif Body1=SHEET and Body2=SHEET
                     create SHEET that is Body1 scribed with Edges at
                        intersection with Body2
                     CURRENTLY NOT IMPLEMENTED
                  elseif Body1=WIRE and Body2=SOLID
                     create WIRE that is part of Body1 that is outside Body2
                     CURRENTLY NOT IMPLEMENTED
                  elseif Body1=WIRE and Body2=SHEET
                     create WIRE that is Body1 scribed with Nodes at
                        intersection with Body2
                     CURRENTLY NOT IMPLEMENTED
                  endif
                  if subtraction does not produce at least index Bodies,
                     an error is returned
                  order may be one of:
                     none    same order as returned from geometry engine
                     xmin    minimum xmin   is first
                     xmax    maximum xmax   is first
                     ymin    minimum ymin   is first
                     ymax    maximum ymax   is first
                     zmin    minimum zmin   is first
                     zmax    maximum zmax   is first
                     amin    minimum area   is first
                     amax    maximum area   is first
                     vmin    minimum volume is first
                     vmax    maximum volume is first
                  if maxtol>0, then tolernce can be relaxed until successful
                  sets up @-parameters
                  order is used directly (without evaluation)

sweep
          use:    create a Body by sweeping a Sketch along a Sketch
          pops:   Sketch1 Sketch2
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  Sketch1 must be either a SHEET or WIRE
                  Sketch2 must be a WIRE
                  if Sketch2 is not slope-continuous, result may not be
                     as expected
                  sets up @-parameters
                  face-order is: base, end, feat1a, feat1b, ...

torus     xcent ycent zcent dxaxis dyaxis dzaxis majorRad minorRad
          use:    create a torus Body
          pops:   -
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  sensitivity computed w.r.t. xcent, ycent, zcent, dxaxis,
                     dyaxis, dzaxis, majorRad, minorRad
                  sets up @-parameters
                  face-order is: xmin/ymin, xmin/ymax, xmax/ymax, xmax,ymax

translate dx dy dz
          use:    translates the entry on top of Stack
          pops:   any
          pushes: any
          notes:  Sketch may not be open
                  solver may not be open
                  sensitivity computed w.r.t. dx, dy, dz
                  sets up @-parameters

ubound    $pmtrName bounds
          use:    defines an upper bound for a design Parameter
          pops:   -
          pushes: -
          notes:  Sketch may not be open
                  solver may not be open
                  may not be used in a .udc file
                  pmtrName must have been defined previously by
                     despmtr statement
                  pmtrName can be in form 'name' or 'name[irow,icol]'
                  bound is set by last ubound statement
                  name is used directly (without evaluation)
                  irow and icol cannot contain a comma or open bracket
                  bounds cannot refer to any other Parameter
                  does not create a Branch
                  if value of Parameter is larger than bounds, a warning
                     is generated
                  cannot be attributed

updarg    $primtype $argName1 argValue1 $argName2 argValue2 $argName3 argValue3
                                                            $argName4 argValue4
          use:    pre-set arguments for next udprim statement
          pops:   -
          pushes: -
          notes:  Sketch may not be open
                  solver may not be open
                  there can be no statements except other udpargs before the
                     next matching udprim
                  primtype determines the type of primitive
                  primtype must match primtype of next udprim statement
                  primtype is used directly (without evaluation)
                  argName#  is used directly (without evaluation)
                  argValue# is used directly if it starts with '$', otherwise it
                     is evaluated
                  if argValue starts with '$$/', use path relative to .csm file
                  arguments for following udprim statement are evaluated
                     in the order they are encountered (udparg first)
                  sensitivity computed w.r.t. argValue1, argValue2, argValue3,
                     argValue4
                  cannot be attributed

udprim    $primtype $argName1 argValue1 $argName2 argValue2 $argName3 argValue3
                                                            $argName4 argValue4
          use:    create a Body by executing a user-defined primitive
          pops:   -
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  primtype  determines the type of primitive and the number of
                     argName/argValue pairs
                  if primtype begins with a letter
                     then a compiled udp whose name is primtype.so is used
                  if primtype starts with a /
                     then a .udc file in the current directory will be used
                  if primtype starts with $/
                     then a .udc file in the same directory of the .csm file
                        will be used
                  primtype  is used directly (without evaluation)
                  argName#  is used directly (without evaluation)
                  argValue# is used directly if it starts with '$', otherwise it
                     is evaluated
                  if argValue# is <<, use data to matching >> as inline file
                  if argValue# starts with '$$/', use path relative to .csm file
                  extra arguments can be set with udparg statement
                  when called to execute a .udc file:
                     the level is incremented
                     INTERNAL Parameters are created for all interface statements
                        for "in"  the value is set to its default
                        for "out" the value is set to its default
                        for "dim" an array is created (of size=value) with
                           value=dot=0
                     the associated udparg and udprim statements are processed
                        in order
                        if argName matches a Parameter created by an interface
                           statement
                           if argValueX matches the name of a Parameter at
                              level-1
                              the values are copied into the new Parameter
                           else
                              argValueX is evalued and stored in the new
                                 Parameter
                        else
                           an error is returned
                     the statements in the .udc are executed until an end
                        statement
                        a set statement either creates a new Parameter or
                           overwrites a value
                     during the execution of the end statement
                        for values associated with an interface "out" statement
                           the value is copied to the appropriate @@-parameter
                              (at level-1)
                        all Parameters at the current level are destroyed
                        the level is decremented
                  sensitivity computed w.r.t. argValue1, argValue2, argValue3,
                     argValue4
                  computes Face and Edge sensitivities analytically (if supplied
                     by the udp)
                  sets up @-parameters
                  face-order is based upon order returned from udprim
                  see udp documentation for full information

union     tomark=0 trimList=0 maxtol=0
          use:    perform Boolean union
          pops:   Body1 Body2  -or-  Body1 ... mark
          pushes: Body
          notes:  Sketch may not be open
                  solver may not be open
                  if     tomark=1
                     create SOLID that is combination of SOLIDs since mark
                  elseif Body1=SOLID and Body2=SOLID
                     if trimList=0
                        create SOLID that us combination of Body1 and Body2
                     else
                        create SOLID that is trimmed combination of Body1 and
                           Body2
                        trimList contains x;y;z;dx;dy;dz where
                           (x,y,z) is inside the Body to be trimmed
                           (dx,dy,dz) is direction toward the trimming Body
                     endif
                  elseif Body1=SHEET and Body2=SHEET
                     create SHEET that is the combination of Bodys with possible
                        new Edges
                  endif
                  if maxtol>0, then tolernce can be relaxed until successful
                  sets up @-parameters
*/

/*
 ************************************************************************
 *                                                                      *
 * Number rules                                                         *
 *                                                                      *
 ************************************************************************
 */

/*
Numbers:
    start with a digit or decimal (.)
    followed by zero or more digits and/or decimals (.)
    there can be at most one decimal in a number
    optionally followed by an e, e+, e-, E, E+, or E-
    if there is an e or E, it must be followed by one or more digits
*/

/*
 ************************************************************************
 *                                                                      *
 * Parameter rules                                                      *
 *                                                                      *
 ************************************************************************
 */

/*
Valid names:
    start with a letter, colon (:), or at-sign (@)
    contains letters, digits, at-signs (@), underscores (_), and colons (:)
    contains fewer than 32 characters
    names that start with an at-sign cannot be set by a despmtr or set statement
    if a name has a dot-suffix, a property of the name (and not its value) is
       returned
       x.nrow   number of rows     in x
       x.ncol   number of columns  in x
       x.size   number of elements in x (=x.nrow*x.ncol)
       x.sum    sum of elements    in x
       x.min    minimum value      in x
       x.max    maximum value      in x

Array names:
    basic format is: name[irow,icol] or name[ielem]
    name must follow rules above
    irow, icol, ans ielem must be valid expressions
    irow, icol, and ielem start counting at 1
    values are stored across rows ([1,1], [1,2], ..., [2,1], ...)

Types:
    EXTERNAL
        if a scalar, declared and defined by a despmtr statement
        if an array, declared by a dimension statement (with despmtr=1)
                     values defined by one or more despmtr statements
        each value can only be defined in one despmtr statement
        can have an optional lower bound
        can have an optional upper bound
        is only available at the .csm file level
        can be set  outside ocsmBuild by a call to ocsmSetValu
        can be read outside ocsmBuild by a call to ocsmGetValu
    INTERNAL
        if a scalar, declared and defined by a set statement
        if an array, declared by a dimension statement (with despmtr=0)
                     values defined by one or more set statements
        values can be overwritten by subsequent set statements
        are created by an interface statement in a .udc file
        see scope rules (below)
    SOLVER
        not implemented yet

    @-parameters depend on the last 'select' statement.
        each time a new Body is added to the Stack, 'select body' is
            implicitly called
        depending on last select statement, the values of the
             @-parameters are given by:

               body face edge node  <- last select

        @nbody   x    x    x    x   number of Bodys
        @ibody   x    x    x    x   current   Body
        @nface   x    x    x    x   number of Faces in @ibody
        @iface  -1    x   -1   -1   current   Face  in @ibody
        @nedge   x    x    x    x   number of Edges in @ibody
        @iedge  -1   -1    x   -1   current   Edge  in @ibody
        @nnode   x    x    x    x   number of Nodes in @ibody
        @inode  -1   -1   -1    x   current   Node  in @ibody

        @ibody1 -1    x    x   -1   first  element of 'Body' attribute in @ibody
        @ibody2 -1    x    x   -1   second element of 'Body' attribute in @ibody

        @xmin    x    x    *    x   x-min of bounding box or x at beg of edge
        @ymin    x    x    *    x   y-min of bounding box or y at beg of edge
        @zmin    x    x    *    x   z-min of bounding box or z at beg of edge
        @xmax    x    x    *    x   x-max of bounding box or x at end of edge
        @ymax    x    x    *    x   y-max of bounding box or y at end of edge
        @zmax    x    x    *    x   z-max of bounding box or z at end of edge

        @length  0    0    x    0   length of edge
        @area    x    x    0    0   area of face or surface area of body
        @volume  x    0    0    0   volume of body (if a solid)

        @xcg     x    x    x    x   location of center of gravity
        @ycg     x    x    x    x
        @zcg     x    x    x    x

        @Ixx     x    x    x    0   centroidal moment of inertia
        @Ixy     x    x    x    0
        @Ixz     x    x    x    0
        @Iyx     x    x    x    0
        @Iyy     x    x    x    0
        @Iyz     x    x    x    0
        @Izx     x    x    x    0
        @Izy     x    x    x    0
        @Izz     x    x    x    0

Scope:
    EXTERNAL parameters are only usable within the .csm file
    INTERNAL within a .csm file
                created by a dimension or set statement
                values are usable only within the .csm file
             within a .udc file
                created by an interface of set statament
                values are usable only with the current .udc file
    SOLVER   parameters are only accessible between SOLBEG and SOLEND statements
*/

/*
 ************************************************************************
 *                                                                      *
 * Expression rules                                                     *
 *                                                                      *
 ************************************************************************
 */

/*
Valid operators (in order of precedence):
    ( )            parentheses, inner-most evaluated first
    func(a,b)      function arguments, then function itself
    ^              exponentiation      (evaluated left to right)
    * /            multiply and divide (evaluated left to right)
    + -            add and subtract    (evaluated left to right)

Valid function calls:
    pi(x)                        3.14159...*x
    min(x,y)                     minimum of x and y
    max(x,y)                     maximum of x and y
    sqrt(x)                      square root of x
    abs(x)                       absolute value of x
    int(x)                       integer part of x  (3.5 -> 3, -3.5 -> -3)
                                     produces derivative=0
    nint(x)                      nearest integer to x
                                     produces derivative=0
    ceil(x)                      smallest integer not less than x
                                     produces derivative=0
    floor(x)                     largest integer not greater than x
                                     produces derivative=0
    sign(test)                   returns -1, 0, or +1
                                     produces derivative=0
    exp(x)                       exponential of x
    log(x)                       natural logarithm of x
    log10(x)                     common logarithm of x
    sin(x)                       sine of x          (in radians)
    sind(x)                      sine of x          (in degrees)
    asin(x)                      arc-sine of x      (in radians)
    asind(x)                     arc-sine of x      (in degrees)
    cos(x)                       cosine of x        (in radians)
    cosd(x)                      cosine of x        (in degrees)
    acos(x)                      arc-cosine of x    (in radians)
    acosd(x)                     arc-cosine of x    (in degrees)
    tan(x)                       tangent of x       (in radians)
    tand(x)                      tangent of x       (in degrees)
    atan(x)                      arc-tangent of x   (in radians)
    atand(x)                     arc-tangent of x   (in degrees)
    atan2(y,x)                   arc-tangent of y/x (in radians)
    atan2d(y,x)                  arc-tangent of y/x (in degrees)
    hypot(x,y)                   hypotenuse: sqrt(x^2+y^2)
    Xcent(xa,ya,dab,xb,yb)       X-center of circular arc
                                     produces derivative=0
    Ycent(xa,ya,dab,xb,yb)       Y-center of circular arc
                                     produces derivative=0
    Xmidl(xa,ya,dab,xb,yb)       X-point at midpoint of circular arc
                                     produces derivative=0
    Ymidl(xa,ya,dab,xb,yb)       Y-point at midpoint of circular arc
                                     produces derivative=0
    seglen(xa,ya,dab,xb,yb)      length of segment
                                     produces derivative=0
    incline(xa,ya,dab,xb,yb)     inclination of chord (in degrees)
                                     produces derivative=0
    radius(xa,ya,dab,xb,yb)      radius of curvature (or 0 for linseg)
                                     produces derivative=0
    sweep(xa,ya,dab,xb,yb)       sweep angle of circular arc (in degrees)
                                     produces derivative=0
    turnang(xa,ya,dab,xb,yb,...
                     dbc,xc,yc)  turnnig angle at b (in degrees)
                                     produces derivative=0
    dip(xa,ya,xb,yb,rad)         acute dip between arc and chord
                                     produces derivative=0
    ifzero(test,ifTrue,ifFalse)  if test=0, return ifTrue, else ifFalse
    ifpos(test,ifTrue,ifFalse)   if test>0, return ifTrue, else ifFalse
    ifneg(test,ifTrue,ifFalse)   if test<0, return ifTrue, else ifFalse
    ifnan(test,ifTrue,ifFalse)   if test is NaN, return ifTrue, else ifFalse
*/

/*
 ************************************************************************
 *                                                                      *
 * Attribute rules (accessible through EGADS)                           *
 *                                                                      *
 ************************************************************************
 */

/*
EGADS attributes assigned to Bodys:

    body        Body index (bias-1)

    brch        Branch index (bias-1)

    <any>       all global attributes

    <any>       all attributes associated with Branch that created Body

    <any>       all attributes associated with "select $body" statement

EGADS attributes assigned to Faces:

    body        non-unique 2-tuple associated with first Face creation
        [0]     Body index in which Face first existed (bias-1)
        [1]     face-order associated with creation (see above)

    brch        non-unique even-numbered list associated with Branches
                   that are active when the Face is created (most
                   recent Branch is listed first)
        [2*i  ] Branch index (bias-1)
        [2*i+1] (see below)

                Branches that contribute to brch attribute are
                   primitive  (for which brch[2*i+1] is face-order)
                   udprim.udc (for which brch[2*i+1] is 1)
                   grown      (for which brch[2*i+1] is face-order)
                   applied    (for which brch[2*i+1] is face-order)
                   sketch     (for which brch[2*i+1] is Sketch primitive if
                               making WIRE)
                   patbeg     (for which brch[2*i+1] is pattern index)
                   recall     (for which brch[2*i+1] is 1)
                   restore    (for which brch[2*i+1] is Body number stored)

    faceID      unique 3-tuple that is assigned automatically
          [0]   body[0]
          [1]   body[1]
          [2]   sequence number

                if multiple Faces have same faceID[0] and faceID[1],
                   then the sequence number is defined based upon the
                   first rule that applies:
                   * Face with smaller xcg  has lower sequence number
                   * Face with smaller ycg  has lower sequence number
                   * Face with smaller zcg  has lower sequence number
                   * Face with smaller area has lower sequence number

    <any>       all attributes associated with Branch that first created Face

    <any>       all attributes associated with "select $face" statement

EGADS attributes assigned to Edges:

    body        non-unique 2-tuple associated with first Edge creation
        [0]     Body index in which Edge first existed (bias-1)
        [1]     100 * min(body[1][ileft],body[1][irite])
                    + max(body[1][ileft],body[1][irite])
                (or -3 if non-manifold)

    edgeID      unique 5-tuple that is assigned automatically
          [0]   faceID[0] of Face 1 (or 0 if non-manifold)
          [1]   faceID[1] of Face 1 (or 0 if non-manifold)
          [2]   faceID[0] of Face 2 (or 0 if non-manifold)
          [3]   faceID[1] of Face 2 (or 0 if non-manifold)
          [4]   sequence number

                edgeID[0]/[1] swapped with edge[2]/[3]
                   100*edgeID[0]+edgeID[1] > 100*edgeID[2]+edgeID[3]
                if multiple Edges have same edgeID[0], edgeID[1],
                   edgeID[2], and edgeID[3], then the sequence number
                   is defined based upon the first rule that applies:
                   * Edge with smaller xcg    has lower sequence number
                   * Edge with smaller ycg    has lower sequence number
                   * Edge with smaller zcg    has lower sequence number
                   * Edge with smaller length has lower sequence number

    <any>       all attributes associated with "select $edge" statement

EGADS attributes assigned to Nodes:

    nodeID      not assigned at this time

    <any>       all attributes associated with "select $node" statement
*/

/*
 ************************************************************************
 *                                                                      *
 * Structures                                                           *
 *                                                                      *
 ************************************************************************
 */

/* "Attr" is a Branch attribute */
typedef struct {
    char          *name;                /* Attribute name */
    char          *value;               /* Attribute value */
} attr_T;

/* "GRatt" is a graphic attribute */
typedef struct {
    void          *object;              /* pointer to GvGraphic (or NULL) */
    int           active;               /* =1 if entity should be rendered */
    int           color;                /* entity color in form 0x00rrggbb */
    int           bcolor;               /* back   color in form 0x00rrggbb */
    int           mcolor;               /* mesh   color in form 0x00rrggbb */
    int           lwidth;               /* line width in pixels */
    int           ptsize;               /* point size in pixels */
    int           render;               /* render flags: */
                                        /*    2 GV_FOREGROUND */
                                        /*    4 GV_ORIENTATION */
                                        /*    8 GV_TRANSPARENT */
                                        /*   16 GV_FACETLIGHT */
                                        /*   32 GV_MESH */
                                        /*   64 GV_FORWARD */
    int           dirty;                /* =1 if attributes have been changed */
} grat_T;

/* "Varg" is the (multi-) value of an argument associated with a Body */
typedef struct {
    int           nval;                 /* number of values (or 0 if string) */
    int           nrow;                 /* number of rows */
    int           ncol;                 /* number of columns */
    union {
        char      *str;                 /* character array if nval==0 */
        double    *val;                 /* array of values if nval>0  */
    };
    double        *dot;                 /* array of velocities if nval>0 */
} varg_T;

/* "Node" is a 0-D topological entity in a Body */
typedef struct {
    int           nedge;                /* number of indicent Edges */

    ego           enode;                /* EGADS node object */

    double        x;                    /* x-coordinate */
    double        y;                    /* y-coordinate */
    double        z;                    /* z-coordinate */
    int           ibody;                /* Body index (1-nbody) */
} node_T;

/* "Edge" is a 1-D topological entity in a Body */
typedef struct {
    int           ibeg;                 /* Node at beginning */
    int           iend;                 /* Node at end */
    int           ileft;                /* Face on the left */
    int           irite;                /* Face on the rite */
    int           nface;                /* number of incident Faces */
    int           ibody;                /* Body index (1-nbody) */
    int           iford;                /* face-order */
    int           imark;                /* value of "mark" attribute (or -1) */
    grat_T        gratt;                /* GRatt of the Edge */

    ego           eedge;                /* EGADS edge object */
} edge_T;

/* "Face" is a 2-D topological entity in a Body */
typedef struct {
    int           ibody;                /* Body index (1-nbody) */
    int           iford;                /* face-order */
    int           imark;                /* value of "mark" attribute (or -1) */
    grat_T        gratt;                /* GRatt of the Face */

    ego           eface;                /* EGADS face object */
} face_T;

/* "Body" is a boundary rperesentation */
typedef struct {
    int           ibrch;                /* Branch associated with Body */
    int           brtype;               /* Branch type (see below) */
    int           ileft;                /* left parent Body (or 0) */
    int           irite;                /* rite parent Body (or 0) */
    int           ichld;                /* child Body (or 0 for root) */
    varg_T        arg[10];              /* array of evaluated arguments (actually use 1-9) */

    ego           ebody;                /* EGADS Body         object(s) */
    ego           etess;                /* EGADS Tessellation object(s) */

    int           onstack;              /* =1 if on stack (and returned); =0 otherwise */
    int           hasdots;              /* =1 if an argument has a dot; =0 otherwise */
    int           botype;               /* Body type (see below) */
    int           nnode;                /* number of Nodes */
    node_T        *node;                /* array  of Nodes */
    int           nedge;                /* number of Edges */
    edge_T        *edge;                /* array  of Edges */
    int           nface;                /* number of Faces */
    face_T        *face;                /* array  of Faces */
    grat_T        gratt;                /* GRatt of the Nodes */
} body_T;

/* "Brch" is a Branch in a feature tree */
typedef struct {
    char          *name;                /* name  of Branch */
    int           type;                 /* type  of Branch (see OpenCSM.h) */
    int           class;                /* class of Branch (see OpenCSM.h) */
    int           indent;               /* indentation */
    int           actv;                 /* activity of Branch (see OpenCSM.h) */
    int           dirty;                /* =1 if dirty */
    int           nattr;                /* number of Attributes */
    attr_T        *attr;                /* array  of Attributes */
    int           ileft;                /* left parent Branch (or 0)*/
    int           irite;                /* rite parent Branch (or 0)*/
    int           ichld;                /* child Branch (or 0 for root) */
    int           narg;                 /* number of arguments */
    char          *arg1;                /* definition for args[1] */
    char          *arg2;                /* definition for args[2] */
    char          *arg3;                /* definition for args[3] */
    char          *arg4;                /* definition for args[4] */
    char          *arg5;                /* definition for args[5] */
    char          *arg6;                /* definition for args[6] */
    char          *arg7;                /* definition for args[7] */
    char          *arg8;                /* definition for args[8] */
    char          *arg9;                /* definition for args[9] */
} brch_T;

/* "Pmtr" is a driving (EXTERNAL) or driven (INTERNAL) Parameter */
typedef struct {
    char          *name;                /* name of Parameter */
    int           type;                 /* Parameter type (see below) */
    int           level;                /* associated level (nominally 0) */
    int           nrow;                 /* number of rows */
    int           ncol;                 /* number of columns */
    double        *value;               /* current value(s) */
    double        *dot;                 /* current velocity(s) */
    double        *lbnd;                /* lower Bound(s) */
    double        *ubnd;                /* upper Bound(s) */
} pmtr_T;

/* "Stor" is the storage locations used by store/restore */
typedef struct {
    char          name[33];             /* name  of Storage */
    int           index;                /* index of Storare */
    int           ibody;                /* Body number stored */
    ego           ebody;                /* EGADS Body */
} stor_T;

/* "Prof" is prfile data */
typedef struct {
    int           ncall;                /* number of calls */
    clock_t       time;                 /* total time */
} prof_T;

/* "Modl" is a constructive solid model consisting of a tree of Branches
         and (possibly) a set of Parameters as well as the associated Bodys */
typedef struct modl_T {
    int           magic;                /* magic number to check for valid *modl */
    int           checked;              /* =1 if successfully passed checks */
    int           nextseq;              /* number of next automatcally-numbered item */
    int           recycle;              /* last Body to recycle */
    int           sel_body;             /* Body selected (or -1) */
    int           sel_face;             /* Face selected (or -1) */
    int           sel_edge;             /* Edge selected (or -1) */
    int           sel_node;             /* Node selected (or -1) */

    int           level;                /* level of file (=0 for .csm, >0 for .udc) */
    char          filename[11][256];    /* filename for ocsmLoad */

    int           nattr;                /* number of global Attributes */
    attr_T        *attr;                /* array  of global Attributes */

    int           nstor;                /* number of storages */
    stor_T        *stor;                /* array  of storages */

    int           nbrch;                /* number of Branches */
    int           mbrch;                /* maximum   Branches */
    brch_T        *brch;                /* array  of Branches */

    int           npmtr;                /* number of Parameters */
    int           mpmtr;                /* maximum   Parameters */
    pmtr_T        *pmtr;                /* array  of Parameters */

    int           nbody;                /* number of Bodys */
    int           mbody;                /* maximum   Bodys */
    body_T        *body;                /* array  of Bodys */

    struct modl_T *perturb;             /* model of perturbed body for sensitivty */
    struct modl_T *basemodl;            /* base MODL while creating perturbation */
    double        dtime;                /* time step in sensitivity */

    ego           context;              /* EGADS context */

    char          *errorBuf;            /* error messages */
    prof_T        profile[100];         /* profile data */
} modl_T;

/*
 ************************************************************************
 *                                                                      *
 * Callable routines                                                    *
 *                                                                      *
 ************************************************************************
 */

/* return current version */
int ocsmVersion(int   *imajor,          /* (out) major version number */
                int   *iminor);         /* (out) minor version number */

/* set output level */
int ocsmSetOutLevel(int    ilevel);     /* (in)  output level: */
                                        /*       =0 warnings and errors only */
                                        /*       =1 nominal (default) */
                                        /*       =2 debug */

/* create a MODL by reading a .csm file */
int ocsmLoad(char   filename[],         /* (in)  file to be read (with .csm) */
             void   **modl);            /* (out) pointer to MODL */

/* save a MODL to a file */
int ocsmSave(void   *modl,              /* (in)  pointer to MODL */
             char   filename[]);        /* (in)  file to be written (with .csm) */

/* copy a MODL */
int ocsmCopy(void   *srcModl,           /* (in)  pointer to source MODL */
             void   **newModl);         /* (out) pointer to new    MODL */

/* free up all storage associated with a MODL */
int ocsmFree(
   /*@null@*/void   *modl);             /* (in)  pointer to MODL (or NULL) */

/* get info about a MODL */
int ocsmInfo(void   *modl,              /* (in)  pointer to MODL */
             int    *nbrch,             /* (out) number of Branches */
             int    *npmtr,             /* (out) number of Parameters */
             int    *nbody);            /* (out) number of Bodys */


/* check that Branches are properly ordered */
int ocsmCheck(void   *modl);            /* (in)  pointer to MODL */

/* build Bodys by executing the MODL up to a given Branch */
int ocsmBuild(void   *modl,             /* (in)  pointer to MODL */
              int    buildTo,           /* (in)  last Branch to execute (or 0 for all) */
              int    *builtTo,          /* (out) last Branch executed successfully */
              int    *nbody,            /* (in)  number of entries allocated in body[] */
                                        /* (out) number of Bodys on the stack */
    /*@null@*/int    body[]);           /* (out) array  of Bodys on the stack (LIFO)
                                                 (at least nbody long) */

/* create a new Branch */
int ocsmNewBrch(void   *modl,           /* (in)  pointer to MODL */
                int    iafter,          /* (in)  Branch index (0-nbrch) after which to add */
                int    type,            /* (in)  Branch type (see below) */
      /*@null@*/char   arg1[],          /* (in)  Argument 1 (or NULL) */
      /*@null@*/char   arg2[],          /* (in)  Argument 2 (or NULL) */
      /*@null@*/char   arg3[],          /* (in)  Argument 3 (or NULL) */
      /*@null@*/char   arg4[],          /* (in)  Argument 4 (or NULL) */
      /*@null@*/char   arg5[],          /* (in)  Argument 5 (or NULL) */
      /*@null@*/char   arg6[],          /* (in)  Argument 6 (or NULL) */
      /*@null@*/char   arg7[],          /* (in)  Argument 7 (or NULL) */
      /*@null@*/char   arg8[],          /* (in)  Argument 8 (or NULL) */
      /*@null@*/char   arg9[]);         /* (in)  Argument 9 (or NULL) */

/* get info about a Branch */
int ocsmGetBrch(void   *modl,           /* (in)  pointer to MODL */
                int    ibrch,           /* (in)  Branch index (1-nbrch) */
                int    *type,           /* (out) Branch type (see below) */
                int    *class,          /* (out) Branch class (see below) */
                int    *actv,           /* (out) Branch Activity (see below) */
                int    *ichld,          /* (out) ibrch of child (or 0 if root) */
                int    *ileft,          /* (out) ibrch of left parent (or 0) */
                int    *irite,          /* (out) ibrch of rite parent (or 0) */
                int    *narg,           /* (out) number of Arguments */
                int    *nattr);         /* (out) number of Attributes */

/* set activity for a Branch */
int ocsmSetBrch(void   *modl,           /* (in)  pointer to MODL */
                int    ibrch,           /* (in)  Branch index (1-nbrch) */
                int    actv);           /* (in)  Branch activity (see below) */

/* delete the last Branch */
int ocsmDelBrch(void   *modl,           /* (in)  pointer to MODL */
                int    ibrch);          /* (in)  Branch index (1-nbrch) */

/* print Branches to file */
int ocsmPrintBrchs(void   *modl,        /* (in)  pointer to MODL */
                   FILE   *fp);         /* (in)  pointer to FILE */

/* get an Argument for a Branch */
int ocsmGetArg(void   *modl,            /* (in)  pointer to MODL */
               int    ibrch,            /* (in)  Branch index (1-nbrch) */
               int    iarg,             /* (in)  Argument index (1-narg) */
               char   defn[],           /* (out) Argument definition (at least 129 long) */
               double *value,           /* (out) Argument value */
               double *dot);            /* (out) Argument velocity */

/* set an Argument for a Branch */
int ocsmSetArg(void   *modl,            /* (in)  pointer to MODL */
               int    ibrch,            /* (in)  Branch index (1-nbrch) */
               int    iarg,             /* (in)  Argument index (1-narg) */
               char   defn[]);          /* (in)  Argument definition */

/* return an Attribute for a Branch by index */
int ocsmRetAttr(void   *modl,           /* (in)  pointer to MODL */
                int    ibrch,           /* (in)  Branch index (1-nbrch) */
                int    iattr,           /* (in)  Attribute index (1-nattr) */
                char   aname[],         /* (out) Attribute name  (at least 129 long) */
                char   avalue[]);       /* (out) Attribute value (at least 129 long) */

/* get an Attribute for a Branch by name*/
int ocsmGetAttr(void   *modl,           /* (in)  pointer to MODL */
                int    ibrch,           /* (in)  Branch index (1-nbrch) or 0 for global */
                char   aname[],         /* (in)  Attribute name */
                char   avalue[]);       /* (out) Attribute value (at least 129 long) */

/* set an Attribute for a Branch */
int ocsmSetAttr(void   *modl,           /* (in)  pointer to MODL */
                int    ibrch,           /* (in)  Branch index (1-nbrch) or 0 for global */
                char   aname[],         /* (in)  Attribute name */
                char   avalue[]);       /* (in)  Attribute value */

/* print global Attributes to file */
int ocsmPrintAttrs(void   *modl,        /* (in)  pointer to MODL */
                   FILE   *fp);         /* (in)  pointer to FILE */

/* get the name of a Branch */
int ocsmGetName(void   *modl,           /* (in)  pointer to MODL */
                int    ibrch,           /* (in)  Branch index (1-nbrch) */
                char   name[]);         /* (out) Branch name (at least 129 long) */

/* set the name for a Branch */
int ocsmSetName(void   *modl,           /* (in)  pointer to MODL */
                int    ibrch,           /* (in)  Branch index (1-nbrch) */
                char   name[]);         /* (in)  Branch name */

/* get string data associated with a Sketch */
int ocsmGetSketch(void   *modl,         /* (in)  pointer to MODL */
                  int    ibrch,         /* (in)  Branch index (1-nbrch) within Sketch */
                  char   begs[],        /* (out) string (1024 long) with SKBEG info */
                                        /*       "xarg;xval;yarg;yval;zarg;zval;" */
                  char   vars[],        /* (out) string (1024 long) with Sketch variables */
                                        /*       "x1;y1;d1;x2; ... dn;" */
                  char   cons[],        /* (out) string (1024 long) with Sketch constraints */
                                        /*       "type1;index1_1;index2_1;value1; ... valuen;" */
                                        /*       index1 and index2 are bias-1 */
                  char   segs[]);       /* (out) string (1024 long) with Sketch segments */
                                        /*       "type1;ibeg1;iend1; ... iendn;" */
                                        /*       ibeg and iend are bias-1 */

/* solve for new Sketch variables */
int ocsmSolveSketch(void   *modl,       /* (in)  pointer to MODL */
                    char   vars_in[],   /* (in)  string with Sketch variables */
                    char   cons[],      /* (in)  string with Sketch constraints */
                    char   vars_out[]); /* (out) string (1024 long) with new Sketch variables */

/* overwrite Branches associated with a Sketch */
int ocsmSaveSketch(void   *modl,        /* (in)  pointer to MODL */
                   int    ibrch,        /* (in)  Branch index (1-nbrch) within Sketch */
                   char   vars[],       /* (in)  string with Sketch variables */
                   char   cons[],       /* (in)  string with Sketch constraints */
                   char   segs[]);      /* (in)  string with Sketch segments */

/* create a new Parameter */
int ocsmNewPmtr(void   *modl,           /* (in)  pointer to MODL */
                char   name[],          /* (in)  Parameter name */
                int    type,            /* (in)  Parameter type */
                int    nrow,            /* (in)  number of rows */
                int    ncol);           /* (in)  number of columns */

/* find (or create) a Parameter */
int ocsmFindPmtr(void   *modl,          /* (in)  pointer to MODL */
                 char   name[],         /* (in)  Parameter name */
                 int    type,           /* (in)  Parameter type */
                 int    nrow,           /* (in)  number of rows */
                 int    ncol,           /* (in)  number of columns */
                 int    *ipmtr);        /* (in)  Parameter index */

/* get info about a Parameter */
int ocsmGetPmtr(void   *modl,           /* (in)  pointer to MODL */
                int    ipmtr,           /* (in)  Parameter index (1-npmtr) */
                int    *type,           /* (out) Parameter type */
                int    *nrow,           /* (out) number of rows */
                int    *ncol,           /* (out) number of columns */
                char   name[]);         /* (out) Parameter name (at least 33 long) */

/* print external Parameters to file */
int ocsmPrintPmtrs(void   *modl,        /* (in)  pointer to MODL */
                   FILE   *fp);         /* (in)  pointer to FILE */

/* get the Value of a Parameter */
int ocsmGetValu(void   *modl,           /* (in)  pointer to MODL */
                int    ipmtr,           /* (in)  Parameter index (1-npmtr) */
                int    irow,            /* (in)  row    index (1-nrow) */
                int    icol,            /* (in)  column index (1-ncol) */
                double *value,          /* (out) Parameter value */
                double *dot);           /* (out) Parameter velocity */

/* set a Value for a Parameter */
int ocsmSetValu(void   *modl,           /* (in)  pointer to MODL */
                int    ipmtr,           /* (in)  Parameter index (1-npmtr) */
                int    irow,            /* (in)  row    index (1-nrow) */
                int    icol,            /* (in)  column index (1-ncol) */
                char   defn[]);         /* (in)  definition of Value */

/* set a (double) Value for a Parameter */
int ocsmSetValuD(void   *modl,          /* (in)  pointer to MODL */
                 int    ipmtr,          /* (in)  Parameter index (1-npmtr) */
                 int    irow,           /* (in)  row    index (1-nrow) */
                 int    icol,           /* (in)  column index (1-ncol) */
                 double value);         /* (in)  value to set */

/* get the Bounds of a Parameter */
int ocsmGetBnds(void   *modl,           /* (in)  pointer to MODL */
                int    ipmtr,           /* (in)  Parameter index (1-npmtr) */
                int    irow,            /* (in)  row    index (1-nrow) */
                int    icol,            /* (in)  column index (1-ncol) */
                double *lbound,         /* (out) lower Bound */
                double *ubound);        /* (out) upper Bound */

/* set the velocity for a Parameter */
int ocsmSetVel(void   *modl,            /* (in)  pointer to MODL */
               int    ipmtr,            /* (in)  Parameter index (1-npmtr) or 0 for all */
               int    irow,             /* (in)  row    index (1-nrow)     or 0 for all */
               int    icol,             /* (in)  column index (1-ncol)     or 0 for all */
               char   defn[]);          /* (in)  definition of Velocity */

/* set the (double) velocity for a Parameter */
int ocsmSetVelD(void   *modl,           /* (in)  pointer to MODL */
                int    ipmtr,           /* (in)  Parameter index (1-npmtr) or 0 for all */
                int    irow,            /* (in)  row    index (1-nrow)     or 0 for all */
                int    icol,            /* (in)  column index (1-ncol)     or 0 for all */
                double dot);            /* (in)  velocity to set */

/* get the parametric coordinates on an Edge or Face */
int ocsmGetUV(void   *modl,             /* (in)  pointer to MODL */
              int    ibody,             /* (in)  Body index (bias-1) */
              int    seltype,           /* (in)  OCSM_EDGE, or OCSM_FACE */
              int    iselect,           /* (in)  Edge or Face index (bias-1) */
              int    npnt,              /* (in)  number of points */
              double xyz[],             /* (in)  coordinates (3*npnt in length) */
              double uv[]);             /* (out) para coords (1*npnt or 2*npnt in length) */

/* get the coordinates on a Node, Edge, or Face */
int ocsmGetXYZ(void   *modl,            /* (in)  pointer to MODL */
               int    ibody,            /* (in)  Body index (bias-1) */
               int    seltype,          /* (in)  OCSM_NODE, OCSM_EDGE, or OCSM_FACE */
               int    iselect,          /* (in)  Node, Edge, or Face index (bias-1) */
               int    npnt,             /* (in)  number of points */
     /*@null@*/double uv[],             /* (in)  para coords (NULL, 1*npnt, or 2*npnt) */
               double xyz[]);           /* (out) coordinates (3*npnt in length) */

/* get the velocities of coordinates on a Node, Edge, or face */
int ocsmGetVel(void   *modl,            /* (in)  pointer to MODL */
               int    ibody,            /* (in)  Body index (bias-1) */
               int    seltype,          /* (in)  OCSM_NODE, OCSM_EDGE, or OCSM_FACE */
               int    iselect,          /* (in)  Node, Edge, or Face index (bias-1) */
               int    npnt,             /* (in)  number of points */
     /*@null@*/double uv[],             /* (in)  para coords
                                                    NULL           for OCSM_NODE
                                                    NULL or 1*npnt for OCSM_EDGE
                                                    NULL or 2*npnt for OCSM_FACE */
               double vel[]);           /* (out) velocities (in pre-allocated array)
                                                    3      for OCSM_NODE
                                                    3*npnt for OCEM_EDGE
                                                      npnt for OCSM_FACE */

/* get info about a Body */
int ocsmGetBody(void   *modl,           /* (in)  pointer to MODL */
                int    ibody,           /* (in)  Body index (1-nbody) */
                int    *type,           /* (out) Branch type (see below) */
                int    *ichld,          /* (out) ibody of child (or 0 if root) */
                int    *ileft,          /* (out) ibody of left parent (or 0) */
                int    *irite,          /* (out) ibody of rite parent (or 0) */
                double vals[],          /* (out) array  of Argument values (at least 10 long) */
                int    *nnode,          /* (out) number of Nodes */
                int    *nedge,          /* (out) number of Edges */
                int    *nface);         /* (out) number of Faces */

/* print all Bodys to file */
int ocsmPrintBodys(void   *modl,        /* (in)  pointer to MODL */
                   FILE   *fp);         /* (in)  pointer to FILE */

/* print the BRep associated with a specific Body */
int ocsmPrintBrep(void   *modl,         /* (in)  pointer to MODL */
                  int    ibody,         /* (in)  Body index (1-nbody) */
                  FILE   *fp);          /* (in)  pointer to File */

/* convert an OCSM code to text */
/*@observer@*/
char *ocsmGetText(int    icode);        /* (in)  code to look up */

/* convert text to an OCSM code */
int ocsmGetCode(char   *text);          /* (in)  text to look up */

/*
 ************************************************************************
 *                                                                      *
 * Defined constants                                                    *
 *                                                                      *
 ************************************************************************
 */

#define           OCSM_DIMENSION  100   /* not branches */
#define           OCSM_DESPMTR    101
#define           OCSM_LBOUND     102
#define           OCSM_UBOUND     103
#define           OCSM_NAME       104
#define           OCSM_ATTRIBUTE  105

#define           OCSM_BOX        111   /* OCSM_PRIMITIVE */
#define           OCSM_SPHERE     112
#define           OCSM_CONE       113
#define           OCSM_CYLINDER   114
#define           OCSM_TORUS      115
#define           OCSM_IMPORT     116
#define           OCSM_UDPRIM     117
#define           OCSM_RESTORE    118

#define           OCSM_EXTRUDE    121   /* OCSM_GROWN */
#define           OCSM_RULE       122
#define           OCSM_LOFT       123
#define           OCSM_BLEND      124
#define           OCSM_REVOLVE    125
#define           OCSM_SWEEP      126

#define           OCSM_FILLET     131   /* OCSM_APPLIED */
#define           OCSM_CHAMFER    132
#define           OCSM_HOLLOW     133
#define           OCSM_CONNECT    134

#define           OCSM_INTERSECT  141   /* OCSM_BOOLEAN */
#define           OCSM_SUBTRACT   142
#define           OCSM_UNION      143
#define           OCSM_JOIN       144
#define           OCSM_EXTRACT    145
#define           OCSM_COMBINE    146

#define           OCSM_TRANSLATE  151   /* OCSM_TRANSFORM */
#define           OCSM_ROTATEX    152
#define           OCSM_ROTATEY    153
#define           OCSM_ROTATEZ    154
#define           OCSM_SCALE      155
#define           OCSM_MIRROR     156
#define           OCSM_REORDER    157

#define           OCSM_SKBEG      161   /* OCSM_SKETCH */
#define           OCSM_SKVAR      162
#define           OCSM_SKCON      163
#define           OCSM_LINSEG     164
#define           OCSM_CIRARC     165
#define           OCSM_ARC        166
#define           OCSM_ELLARC     167
#define           OCSM_SPLINE     168
#define           OCSM_BEZIER     169
#define           OCSM_SKEND      170

#define           OCSM_SOLBEG     171   /* OCSM_SOLVER */
#define           OCSM_SOLCON     172
#define           OCSM_SOLEND     173

#define           OCSM_INTERFACE  181   /* OCSM_UTILITY */
#define           OCSM_END        182
#define           OCSM_SET        183
#define           OCSM_UDPARG     184
#define           OCSM_SELECT     185
#define           OCSM_PROJECT    186
#define           OCSM_MACBEG     187
#define           OCSM_MACEND     188
#define           OCSM_RECALL     189
#define           OCSM_STORE      190
#define           OCSM_PATBEG     191
#define           OCSM_PATBREAK   192
#define           OCSM_PATEND     193
#define           OCSM_MARK       194
#define           OCSM_DUMP       195
#define           OCSM_ASSERT     196

#define           OCSM_PRIMITIVE  201   /* Branch classes */
#define           OCSM_GROWN      202
#define           OCSM_APPLIED    203
#define           OCSM_BOOLEAN    204
#define           OCSM_TRANSFORM  205
#define           OCSM_SKETCH     206
#define           OCSM_SOLVER     207
#define           OCSM_UTILITY    208

#define           OCSM_ACTIVE     300   /* Branch activities */
#define           OCSM_SUPPRESSED 301
#define           OCSM_INACTIVE   302
#define           OCSM_DEFERRED   303

#define           OCSM_SOLID_BODY 400   /* Body types */
#define           OCSM_SHEET_BODY 401
#define           OCSM_WIRE_BODY  402
#define           OCSM_NODE_BODY  403
#define           OCSM_NULL_BODY  404

#define           OCSM_EXTERNAL   500   /* Parameter types */
#define           OCSM_INTERNAL   501

#define           OCSM_NODE       600   /* Selector types */
#define           OCSM_EDGE       601
#define           OCSM_FACE       602
#define           OCSM_VOLUME     603

#define           OCSM_UNDEFINED  -123.456

/*
 ************************************************************************
 *                                                                      *
 * Return codes (errors are -201 to -299)                               *
 *                                                                      *
 ************************************************************************
 */

#define SUCCESS                                 0

#define OCSM_FILE_NOT_FOUND                  -201
#define OCSM_ILLEGAL_STATEMENT               -202
#define OCSM_NOT_ENOUGH_ARGS                 -203
#define OCSM_NAME_ALREADY_DEFINED            -204
#define OCSM_NESTED_TOO_DEEPLY               -205
#define OCSM_PATBEG_WITHOUT_PATEND           -206
#define OCSM_PATEND_WITHOUT_PATBEG           -207
#define OCSM_NOT_MODL_STRUCTURE              -208
#define OCSM_PROBLEM_CREATING_PERTURB        -209

#define OCSM_MISSING_MARK                    -211
#define OCSM_INSUFFICIENT_BODYS_ON_STACK     -212
#define OCSM_WRONG_TYPES_ON_STACK            -213
#define OCSM_DID_NOT_CREATE_BODY             -214
#define OCSM_CREATED_TOO_MANY_BODYS          -215
#define OCSM_TOO_MANY_BODYS_ON_STACK         -216
#define OCSM_ERROR_IN_BODYS_ON_STACK         -217
#define OCSM_MODL_NOT_CHECKED                -218
#define OCSM_NEED_TESSELLATION               -219

#define OCSM_BODY_NOT_FOUND                  -221
#define OCSM_FACE_NOT_FOUND                  -222
#define OCSM_EDGE_NOT_FOUND                  -223
#define OCSM_NODE_NOT_FOUND                  -224
#define OCSM_ILLEGAL_VALUE                   -225
#define OCSM_ILLEGAL_ATTRIBUTE               -226

#define OCSM_SKETCH_IS_OPEN                  -231
#define OCSM_SKETCH_IS_NOT_OPEN              -232
#define OCSM_COLINEAR_SKETCH_POINTS          -233
#define OCSM_NON_COPLANAR_SKETCH_POINTS      -234
#define OCSM_TOO_MANY_SKETCH_POINTS          -235
#define OCSM_TOO_FEW_SPLINE_POINTS           -236
#define OCSM_SKETCH_DOES_NOT_CLOSE           -237
#define OCSM_CLOSED_SKETCH_NOT_PLANAR        -238
#define OCSM_ASSERT_FAILED                   -239

#define OCSM_ILLEGAL_CHAR_IN_EXPR            -241
#define OCSM_CLOSE_BEFORE_OPEN               -242
#define OCSM_MISSING_CLOSE                   -243
#define OCSM_IMPROPER_NESTING                -244
#define OCSM_ILLEGAL_TOKEN_SEQUENCE          -245
#define OCSM_ILLEGAL_NUMBER                  -246
#define OCSM_ILLEGAL_PMTR_NAME               -247
#define OCSM_ILLEGAL_FUNC_NAME               -248
#define OCSM_ILLEGAL_TYPE                    -249
#define OCSM_ILLEGAL_NARG                    -250

#define OCSM_NAME_NOT_FOUND                  -251
#define OCSM_NAME_NOT_UNIQUE                 -252
#define OCSM_PMTR_IS_EXTERNAL                -253
#define OCSM_PMTR_IS_INTERNAL                -254
#define OCSM_FUNC_ARG_OUT_OF_BOUNDS          -255
#define OCSM_VAL_STACK_UNDERFLOW             -256  /* probably not enough args to func */
#define OCSM_VAL_STACK_OVERFLOW              -257  /* probably too many   args to func */

#define OCSM_ILLEGAL_BRCH_INDEX              -261  /* should be from 1 to nbrch */
#define OCSM_ILLEGAL_PMTR_INDEX              -262  /* should be from 1 to npmtr */
#define OCSM_ILLEGAL_BODY_INDEX              -263  /* should be from 1 to nbody */
#define OCSM_ILLEGAL_ARG_INDEX               -264  /* should be from 1 to narg  */
#define OCSM_ILLEGAL_ACTIVITY                -265  /* should OCSM_ACTIVE or OCSM_SUPPRESSED */
#define OCSM_ILLEGAL_MACRO_INDEX             -266  /* should be between 1 and 100 */
#define OCSM_ILLEGAL_ARGUMENT                -267
#define OCSM_CANNOT_BE_SUPPRESSED            -268
#define OCSM_STORAGE_ALREADY_USED            -269
#define OCSM_NOTHING_PREVIOUSLY_STORED       -270

#define OCSM_SOLVER_IS_OPEN                  -271
#define OCSM_SOLVER_IS_NOT_OPEN              -272
#define OCSM_TOO_MANY_SOLVER_VARS            -273
#define OCSM_UNDERCONSTRAINED                -274
#define OCSM_OVERCONSTRAINED                 -275
#define OCSM_SINGULAR_MATRIX                 -276
#define OCSM_NOT_CONVERGED                   -277

#define OCSM_UDP_ERROR1                      -281
#define OCSM_UDP_ERROR2                      -282
#define OCSM_UDP_ERROR3                      -283
#define OCSM_UDP_ERROR4                      -284
#define OCSM_UDP_ERROR5                      -285
#define OCSM_UDP_ERROR6                      -286
#define OCSM_UDP_ERROR7                      -287
#define OCSM_UDP_ERROR8                      -288
#define OCSM_UDP_ERROR9                      -289

#define OCSM_OP_STACK_UNDERFLOW              -291
#define OCSM_OP_STACK_OVERFLOW               -292
#define OCSM_RPN_STACK_UNDERFLOW             -293
#define OCSM_RPN_STACK_OVERFLOW              -294
#define OCSM_TOKEN_STACK_UNDERFLOW           -295
#define OCSM_TOKEN_STACK_OVERFLOW            -296
#define OCSM_UNSUPPORTED                     -298
#define OCSM_INTERNAL_ERROR                  -299

#endif  /* _OPENCSM_H_ */
