/*
 ************************************************************************
 *                                                                      *
 * serveCSM.c -- server for driving OpenCSM                             *
 *                                                                      *
 *              Written by John Dannenhoffer @ Syracuse University      *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2012/2015  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <unistd.h>

#ifdef WIN32
    #define WIN32_LEAN_AND_MEAN
    #include <windows.h>
    #include <winsock2.h>
#endif

#include "egads.h"

#define CINT    const int
#define CDOUBLE const double
#define CCHAR   const char

#ifdef WIN32
    #define snprintf _snprintf
#endif

#define STRNCPY(A, B, LEN) strncpy(A, B, LEN); A[LEN-1] = '\0';

#include "common.h"
#include "OpenCSM.h"

#include "udp.h"

#include "wsserver.h"

/***********************************************************************/
/*                                                                     */
/* macros (including those that go along with common.h)                */
/*                                                                     */
/***********************************************************************/

#ifdef DEBUG
   #define DOPEN {if (dbg_fp == NULL) dbg_fp = fopen("serveCSM.dbg", "w");}
   static  FILE *dbg_fp=NULL;
#endif

//static void *realloc_temp=NULL;            /* used by RALLOC macro */

#define  RED(COLOR)      (float)(COLOR / 0x10000        ) / (float)(255)
#define  GREEN(COLOR)    (float)(COLOR / 0x00100 % 0x100) / (float)(255)
#define  BLUE(COLOR)     (float)(COLOR           % 0x100) / (float)(255)

#define MAX_EXPR_LEN     512
#define MAX_STR_LEN    32767

/***********************************************************************/
/*                                                                     */
/* structures                                                          */
/*                                                                     */
/***********************************************************************/

/* blue-white-red spectrum */
static float color_map[256*3] =
{ 0.0000, 0.0000, 1.0000,    0.0078, 0.0078, 1.0000,   0.0156, 0.0156, 1.0000,    0.0234, 0.0234, 1.0000,
  0.0312, 0.0312, 1.0000,    0.0391, 0.0391, 1.0000,   0.0469, 0.0469, 1.0000,    0.0547, 0.0547, 1.0000,
  0.0625, 0.0625, 1.0000,    0.0703, 0.0703, 1.0000,   0.0781, 0.0781, 1.0000,    0.0859, 0.0859, 1.0000,
  0.0938, 0.0938, 1.0000,    0.1016, 0.1016, 1.0000,   0.1094, 0.1094, 1.0000,    0.1172, 0.1172, 1.0000,
  0.1250, 0.1250, 1.0000,    0.1328, 0.1328, 1.0000,   0.1406, 0.1406, 1.0000,    0.1484, 0.1484, 1.0000,
  0.1562, 0.1562, 1.0000,    0.1641, 0.1641, 1.0000,   0.1719, 0.1719, 1.0000,    0.1797, 0.1797, 1.0000,
  0.1875, 0.1875, 1.0000,    0.1953, 0.1953, 1.0000,   0.2031, 0.2031, 1.0000,    0.2109, 0.2109, 1.0000,
  0.2188, 0.2188, 1.0000,    0.2266, 0.2266, 1.0000,   0.2344, 0.2344, 1.0000,    0.2422, 0.2422, 1.0000,
  0.2500, 0.2500, 1.0000,    0.2578, 0.2578, 1.0000,   0.2656, 0.2656, 1.0000,    0.2734, 0.2734, 1.0000,
  0.2812, 0.2812, 1.0000,    0.2891, 0.2891, 1.0000,   0.2969, 0.2969, 1.0000,    0.3047, 0.3047, 1.0000,
  0.3125, 0.3125, 1.0000,    0.3203, 0.3203, 1.0000,   0.3281, 0.3281, 1.0000,    0.3359, 0.3359, 1.0000,
  0.3438, 0.3438, 1.0000,    0.3516, 0.3516, 1.0000,   0.3594, 0.3594, 1.0000,    0.3672, 0.3672, 1.0000,
  0.3750, 0.3750, 1.0000,    0.3828, 0.3828, 1.0000,   0.3906, 0.3906, 1.0000,    0.3984, 0.3984, 1.0000,
  0.4062, 0.4062, 1.0000,    0.4141, 0.4141, 1.0000,   0.4219, 0.4219, 1.0000,    0.4297, 0.4297, 1.0000,
  0.4375, 0.4375, 1.0000,    0.4453, 0.4453, 1.0000,   0.4531, 0.4531, 1.0000,    0.4609, 0.4609, 1.0000,
  0.4688, 0.4688, 1.0000,    0.4766, 0.4766, 1.0000,   0.4844, 0.4844, 1.0000,    0.4922, 0.4922, 1.0000,
  0.5000, 0.5000, 1.0000,    0.5078, 0.5078, 1.0000,   0.5156, 0.5156, 1.0000,    0.5234, 0.5234, 1.0000,
  0.5312, 0.5312, 1.0000,    0.5391, 0.5391, 1.0000,   0.5469, 0.5469, 1.0000,    0.5547, 0.5547, 1.0000,
  0.5625, 0.5625, 1.0000,    0.5703, 0.5703, 1.0000,   0.5781, 0.5781, 1.0000,    0.5859, 0.5859, 1.0000,
  0.5938, 0.5938, 1.0000,    0.6016, 0.6016, 1.0000,   0.6094, 0.6094, 1.0000,    0.6172, 0.6172, 1.0000,
  0.6250, 0.6250, 1.0000,    0.6328, 0.6328, 1.0000,   0.6406, 0.6406, 1.0000,    0.6484, 0.6484, 1.0000,
  0.6562, 0.6562, 1.0000,    0.6641, 0.6641, 1.0000,   0.6719, 0.6719, 1.0000,    0.6797, 0.6797, 1.0000,
  0.6875, 0.6875, 1.0000,    0.6953, 0.6953, 1.0000,   0.7031, 0.7031, 1.0000,    0.7109, 0.7109, 1.0000,
  0.7188, 0.7188, 1.0000,    0.7266, 0.7266, 1.0000,   0.7344, 0.7344, 1.0000,    0.7422, 0.7422, 1.0000,
  0.7500, 0.7500, 1.0000,    0.7578, 0.7578, 1.0000,   0.7656, 0.7656, 1.0000,    0.7734, 0.7734, 1.0000,
  0.7812, 0.7812, 1.0000,    0.7891, 0.7891, 1.0000,   0.7969, 0.7969, 1.0000,    0.8047, 0.8047, 1.0000,
  0.8125, 0.8125, 1.0000,    0.8203, 0.8203, 1.0000,   0.8281, 0.8281, 1.0000,    0.8359, 0.8359, 1.0000,
  0.8438, 0.8438, 1.0000,    0.8516, 0.8516, 1.0000,   0.8594, 0.8594, 1.0000,    0.8672, 0.8672, 1.0000,
  0.8750, 0.8750, 1.0000,    0.8828, 0.8828, 1.0000,   0.8906, 0.8906, 1.0000,    0.8984, 0.8984, 1.0000,
  0.9062, 0.9062, 1.0000,    0.9141, 0.9141, 1.0000,   0.9219, 0.9219, 1.0000,    0.9297, 0.9297, 1.0000,
  0.9375, 0.9375, 1.0000,    0.9453, 0.9453, 1.0000,   0.9531, 0.9531, 1.0000,    0.9609, 0.9609, 1.0000,
  0.9688, 0.9688, 1.0000,    0.9766, 0.9766, 1.0000,   0.9844, 0.9844, 1.0000,    0.9922, 0.9922, 1.0000,
  1.0000, 1.0000, 1.0000,    1.0000, 0.9922, 0.9922,   1.0000, 0.9844, 0.9844,    1.0000, 0.9766, 0.9766,
  1.0000, 0.9688, 0.9688,    1.0000, 0.9609, 0.9609,   1.0000, 0.9531, 0.9531,    1.0000, 0.9453, 0.9453,
  1.0000, 0.9375, 0.9375,    1.0000, 0.9297, 0.9297,   1.0000, 0.9219, 0.9219,    1.0000, 0.9141, 0.9141,
  1.0000, 0.9062, 0.9062,    1.0000, 0.8984, 0.8984,   1.0000, 0.8906, 0.8906,    1.0000, 0.8828, 0.8828,
  1.0000, 0.8750, 0.8750,    1.0000, 0.8672, 0.8672,   1.0000, 0.8594, 0.8594,    1.0000, 0.8516, 0.8516,
  1.0000, 0.8438, 0.8438,    1.0000, 0.8359, 0.8359,   1.0000, 0.8281, 0.8281,    1.0000, 0.8203, 0.8203,
  1.0000, 0.8125, 0.8125,    1.0000, 0.8047, 0.8047,   1.0000, 0.7969, 0.7969,    1.0000, 0.7891, 0.7891,
  1.0000, 0.7812, 0.7812,    1.0000, 0.7734, 0.7734,   1.0000, 0.7656, 0.7656,    1.0000, 0.7578, 0.7578,
  1.0000, 0.7500, 0.7500,    1.0000, 0.7422, 0.7422,   1.0000, 0.7344, 0.7344,    1.0000, 0.7266, 0.7266,
  1.0000, 0.7188, 0.7188,    1.0000, 0.7109, 0.7109,   1.0000, 0.7031, 0.7031,    1.0000, 0.6953, 0.6953,
  1.0000, 0.6875, 0.6875,    1.0000, 0.6797, 0.6797,   1.0000, 0.6719, 0.6719,    1.0000, 0.6641, 0.6641,
  1.0000, 0.6562, 0.6562,    1.0000, 0.6484, 0.6484,   1.0000, 0.6406, 0.6406,    1.0000, 0.6328, 0.6328,
  1.0000, 0.6250, 0.6250,    1.0000, 0.6172, 0.6172,   1.0000, 0.6094, 0.6094,    1.0000, 0.6016, 0.6016,
  1.0000, 0.5938, 0.5938,    1.0000, 0.5859, 0.5859,   1.0000, 0.5781, 0.5781,    1.0000, 0.5703, 0.5703,
  1.0000, 0.5625, 0.5625,    1.0000, 0.5547, 0.5547,   1.0000, 0.5469, 0.5469,    1.0000, 0.5391, 0.5391,
  1.0000, 0.5312, 0.5312,    1.0000, 0.5234, 0.5234,   1.0000, 0.5156, 0.5156,    1.0000, 0.5078, 0.5078,
  1.0000, 0.5000, 0.5000,    1.0000, 0.4922, 0.4922,   1.0000, 0.4844, 0.4844,    1.0000, 0.4766, 0.4766,
  1.0000, 0.4688, 0.4688,    1.0000, 0.4609, 0.4609,   1.0000, 0.4531, 0.4531,    1.0000, 0.4453, 0.4453,
  1.0000, 0.4375, 0.4375,    1.0000, 0.4297, 0.4297,   1.0000, 0.4219, 0.4219,    1.0000, 0.4141, 0.4141,
  1.0000, 0.4062, 0.4062,    1.0000, 0.3984, 0.3984,   1.0000, 0.3906, 0.3906,    1.0000, 0.3828, 0.3828,
  1.0000, 0.3750, 0.3750,    1.0000, 0.3672, 0.3672,   1.0000, 0.3594, 0.3594,    1.0000, 0.3516, 0.3516,
  1.0000, 0.3438, 0.3438,    1.0000, 0.3359, 0.3359,   1.0000, 0.3281, 0.3281,    1.0000, 0.3203, 0.3203,
  1.0000, 0.3125, 0.3125,    1.0000, 0.3047, 0.3047,   1.0000, 0.2969, 0.2969,    1.0000, 0.2891, 0.2891,
  1.0000, 0.2812, 0.2812,    1.0000, 0.2734, 0.2734,   1.0000, 0.2656, 0.2656,    1.0000, 0.2578, 0.2578,
  1.0000, 0.2500, 0.2500,    1.0000, 0.2422, 0.2422,   1.0000, 0.2344, 0.2344,    1.0000, 0.2266, 0.2266,
  1.0000, 0.2188, 0.2188,    1.0000, 0.2109, 0.2109,   1.0000, 0.2031, 0.2031,    1.0000, 0.1953, 0.1953,
  1.0000, 0.1875, 0.1875,    1.0000, 0.1797, 0.1797,   1.0000, 0.1719, 0.1719,    1.0000, 0.1641, 0.1641,
  1.0000, 0.1562, 0.1562,    1.0000, 0.1484, 0.1484,   1.0000, 0.1406, 0.1406,    1.0000, 0.1328, 0.1328,
  1.0000, 0.1250, 0.1250,    1.0000, 0.1172, 0.1172,   1.0000, 0.1094, 0.1094,    1.0000, 0.1016, 0.1016,
  1.0000, 0.0938, 0.0938,    1.0000, 0.0859, 0.0859,   1.0000, 0.0781, 0.0781,    1.0000, 0.0703, 0.0703,
  1.0000, 0.0625, 0.0625,    1.0000, 0.0547, 0.0547,   1.0000, 0.0469, 0.0469,    1.0000, 0.0391, 0.0391,
  1.0000, 0.0312, 0.0312,    1.0000, 0.0234, 0.0234,   1.0000, 0.0156, 0.0156,    1.0000, 0.0078, 0.0078 };

//$$$/* blue-green-red spectrum */
//$$$static float color_map[256*3] =
//$$${ 0.0000, 0.0000, 1.0000,    0.0000, 0.0157, 1.0000,   0.0000, 0.0314, 1.0000,    0.0000, 0.0471, 1.0000,
//$$$  0.0000, 0.0627, 1.0000,    0.0000, 0.0784, 1.0000,   0.0000, 0.0941, 1.0000,    0.0000, 0.1098, 1.0000,
//$$$  0.0000, 0.1255, 1.0000,    0.0000, 0.1412, 1.0000,   0.0000, 0.1569, 1.0000,    0.0000, 0.1725, 1.0000,
//$$$  0.0000, 0.1882, 1.0000,    0.0000, 0.2039, 1.0000,   0.0000, 0.2196, 1.0000,    0.0000, 0.2353, 1.0000,
//$$$  0.0000, 0.2510, 1.0000,    0.0000, 0.2667, 1.0000,   0.0000, 0.2824, 1.0000,    0.0000, 0.2980, 1.0000,
//$$$  0.0000, 0.3137, 1.0000,    0.0000, 0.3294, 1.0000,   0.0000, 0.3451, 1.0000,    0.0000, 0.3608, 1.0000,
//$$$  0.0000, 0.3765, 1.0000,    0.0000, 0.3922, 1.0000,   0.0000, 0.4078, 1.0000,    0.0000, 0.4235, 1.0000,
//$$$  0.0000, 0.4392, 1.0000,    0.0000, 0.4549, 1.0000,   0.0000, 0.4706, 1.0000,    0.0000, 0.4863, 1.0000,
//$$$  0.0000, 0.5020, 1.0000,    0.0000, 0.5176, 1.0000,   0.0000, 0.5333, 1.0000,    0.0000, 0.5490, 1.0000,
//$$$  0.0000, 0.5647, 1.0000,    0.0000, 0.5804, 1.0000,   0.0000, 0.5961, 1.0000,    0.0000, 0.6118, 1.0000,
//$$$  0.0000, 0.6275, 1.0000,    0.0000, 0.6431, 1.0000,   0.0000, 0.6588, 1.0000,    0.0000, 0.6745, 1.0000,
//$$$  0.0000, 0.6902, 1.0000,    0.0000, 0.7059, 1.0000,   0.0000, 0.7216, 1.0000,    0.0000, 0.7373, 1.0000,
//$$$  0.0000, 0.7529, 1.0000,    0.0000, 0.7686, 1.0000,   0.0000, 0.7843, 1.0000,    0.0000, 0.8000, 1.0000,
//$$$  0.0000, 0.8157, 1.0000,    0.0000, 0.8314, 1.0000,   0.0000, 0.8471, 1.0000,    0.0000, 0.8627, 1.0000,
//$$$  0.0000, 0.8784, 1.0000,    0.0000, 0.8941, 1.0000,   0.0000, 0.9098, 1.0000,    0.0000, 0.9255, 1.0000,
//$$$  0.0000, 0.9412, 1.0000,    0.0000, 0.9569, 1.0000,   0.0000, 0.9725, 1.0000,    0.0000, 0.9882, 1.0000,
//$$$  0.0000, 1.0000, 0.9961,    0.0000, 1.0000, 0.9804,   0.0000, 1.0000, 0.9647,    0.0000, 1.0000, 0.9490,
//$$$  0.0000, 1.0000, 0.9333,    0.0000, 1.0000, 0.9176,   0.0000, 1.0000, 0.9020,    0.0000, 1.0000, 0.8863,
//$$$  0.0000, 1.0000, 0.8706,    0.0000, 1.0000, 0.8549,   0.0000, 1.0000, 0.8392,    0.0000, 1.0000, 0.8235,
//$$$  0.0000, 1.0000, 0.8078,    0.0000, 1.0000, 0.7922,   0.0000, 1.0000, 0.7765,    0.0000, 1.0000, 0.7608,
//$$$  0.0000, 1.0000, 0.7451,    0.0000, 1.0000, 0.7294,   0.0000, 1.0000, 0.7137,    0.0000, 1.0000, 0.6980,
//$$$  0.0000, 1.0000, 0.6824,    0.0000, 1.0000, 0.6667,   0.0000, 1.0000, 0.6510,    0.0000, 1.0000, 0.6353,
//$$$  0.0000, 1.0000, 0.6196,    0.0000, 1.0000, 0.6039,   0.0000, 1.0000, 0.5882,    0.0000, 1.0000, 0.5725,
//$$$  0.0000, 1.0000, 0.5569,    0.0000, 1.0000, 0.5412,   0.0000, 1.0000, 0.5255,    0.0000, 1.0000, 0.5098,
//$$$  0.0000, 1.0000, 0.4941,    0.0000, 1.0000, 0.4784,   0.0000, 1.0000, 0.4627,    0.0000, 1.0000, 0.4471,
//$$$  0.0000, 1.0000, 0.4314,    0.0000, 1.0000, 0.4157,   0.0000, 1.0000, 0.4000,    0.0000, 1.0000, 0.3843,
//$$$  0.0000, 1.0000, 0.3686,    0.0000, 1.0000, 0.3529,   0.0000, 1.0000, 0.3373,    0.0000, 1.0000, 0.3216,
//$$$  0.0000, 1.0000, 0.3059,    0.0000, 1.0000, 0.2902,   0.0000, 1.0000, 0.2745,    0.0000, 1.0000, 0.2588,
//$$$  0.0000, 1.0000, 0.2431,    0.0000, 1.0000, 0.2275,   0.0000, 1.0000, 0.2118,    0.0000, 1.0000, 0.1961,
//$$$  0.0000, 1.0000, 0.1804,    0.0000, 1.0000, 0.1647,   0.0000, 1.0000, 0.1490,    0.0000, 1.0000, 0.1333,
//$$$  0.0000, 1.0000, 0.1176,    0.0000, 1.0000, 0.1020,   0.0000, 1.0000, 0.0863,    0.0000, 1.0000, 0.0706,
//$$$  0.0000, 1.0000, 0.0549,    0.0000, 1.0000, 0.0392,   0.0000, 1.0000, 0.0235,    0.0000, 1.0000, 0.0078,
//$$$  0.0078, 1.0000, 0.0000,    0.0235, 1.0000, 0.0000,   0.0392, 1.0000, 0.0000,    0.0549, 1.0000, 0.0000,
//$$$  0.0706, 1.0000, 0.0000,    0.0863, 1.0000, 0.0000,   0.1020, 1.0000, 0.0000,    0.1176, 1.0000, 0.0000,
//$$$  0.1333, 1.0000, 0.0000,    0.1490, 1.0000, 0.0000,   0.1647, 1.0000, 0.0000,    0.1804, 1.0000, 0.0000,
//$$$  0.1961, 1.0000, 0.0000,    0.2118, 1.0000, 0.0000,   0.2275, 1.0000, 0.0000,    0.2431, 1.0000, 0.0000,
//$$$  0.2588, 1.0000, 0.0000,    0.2745, 1.0000, 0.0000,   0.2902, 1.0000, 0.0000,    0.3059, 1.0000, 0.0000,
//$$$  0.3216, 1.0000, 0.0000,    0.3373, 1.0000, 0.0000,   0.3529, 1.0000, 0.0000,    0.3686, 1.0000, 0.0000,
//$$$  0.3843, 1.0000, 0.0000,    0.4000, 1.0000, 0.0000,   0.4157, 1.0000, 0.0000,    0.4314, 1.0000, 0.0000,
//$$$  0.4471, 1.0000, 0.0000,    0.4627, 1.0000, 0.0000,   0.4784, 1.0000, 0.0000,    0.4941, 1.0000, 0.0000,
//$$$  0.5098, 1.0000, 0.0000,    0.5255, 1.0000, 0.0000,   0.5412, 1.0000, 0.0000,    0.5569, 1.0000, 0.0000,
//$$$  0.5725, 1.0000, 0.0000,    0.5882, 1.0000, 0.0000,   0.6039, 1.0000, 0.0000,    0.6196, 1.0000, 0.0000,
//$$$  0.6353, 1.0000, 0.0000,    0.6510, 1.0000, 0.0000,   0.6667, 1.0000, 0.0000,    0.6824, 1.0000, 0.0000,
//$$$  0.6980, 1.0000, 0.0000,    0.7137, 1.0000, 0.0000,   0.7294, 1.0000, 0.0000,    0.7451, 1.0000, 0.0000,
//$$$  0.7608, 1.0000, 0.0000,    0.7765, 1.0000, 0.0000,   0.7922, 1.0000, 0.0000,    0.8078, 1.0000, 0.0000,
//$$$  0.8235, 1.0000, 0.0000,    0.8392, 1.0000, 0.0000,   0.8549, 1.0000, 0.0000,    0.8706, 1.0000, 0.0000,
//$$$  0.8863, 1.0000, 0.0000,    0.9020, 1.0000, 0.0000,   0.9176, 1.0000, 0.0000,    0.9333, 1.0000, 0.0000,
//$$$  0.9490, 1.0000, 0.0000,    0.9647, 1.0000, 0.0000,   0.9804, 1.0000, 0.0000,    0.9961, 1.0000, 0.0000,
//$$$  1.0000, 0.9882, 0.0000,    1.0000, 0.9725, 0.0000,   1.0000, 0.9569, 0.0000,    1.0000, 0.9412, 0.0000,
//$$$  1.0000, 0.9255, 0.0000,    1.0000, 0.9098, 0.0000,   1.0000, 0.8941, 0.0000,    1.0000, 0.8784, 0.0000,
//$$$  1.0000, 0.8627, 0.0000,    1.0000, 0.8471, 0.0000,   1.0000, 0.8314, 0.0000,    1.0000, 0.8157, 0.0000,
//$$$  1.0000, 0.8000, 0.0000,    1.0000, 0.7843, 0.0000,   1.0000, 0.7686, 0.0000,    1.0000, 0.7529, 0.0000,
//$$$  1.0000, 0.7373, 0.0000,    1.0000, 0.7216, 0.0000,   1.0000, 0.7059, 0.0000,    1.0000, 0.6902, 0.0000,
//$$$  1.0000, 0.6745, 0.0000,    1.0000, 0.6588, 0.0000,   1.0000, 0.6431, 0.0000,    1.0000, 0.6275, 0.0000,
//$$$  1.0000, 0.6118, 0.0000,    1.0000, 0.5961, 0.0000,   1.0000, 0.5804, 0.0000,    1.0000, 0.5647, 0.0000,
//$$$  1.0000, 0.5490, 0.0000,    1.0000, 0.5333, 0.0000,   1.0000, 0.5176, 0.0000,    1.0000, 0.5020, 0.0000,
//$$$  1.0000, 0.4863, 0.0000,    1.0000, 0.4706, 0.0000,   1.0000, 0.4549, 0.0000,    1.0000, 0.4392, 0.0000,
//$$$  1.0000, 0.4235, 0.0000,    1.0000, 0.4078, 0.0000,   1.0000, 0.3922, 0.0000,    1.0000, 0.3765, 0.0000,
//$$$  1.0000, 0.3608, 0.0000,    1.0000, 0.3451, 0.0000,   1.0000, 0.3294, 0.0000,    1.0000, 0.3137, 0.0000,
//$$$  1.0000, 0.2980, 0.0000,    1.0000, 0.2824, 0.0000,   1.0000, 0.2667, 0.0000,    1.0000, 0.2510, 0.0000,
//$$$  1.0000, 0.2353, 0.0000,    1.0000, 0.2196, 0.0000,   1.0000, 0.2039, 0.0000,    1.0000, 0.1882, 0.0000,
//$$$  1.0000, 0.1725, 0.0000,    1.0000, 0.1569, 0.0000,   1.0000, 0.1412, 0.0000,    1.0000, 0.1255, 0.0000,
//$$$  1.0000, 0.1098, 0.0000,    1.0000, 0.0941, 0.0000,   1.0000, 0.0784, 0.0000,    1.0000, 0.0627, 0.0000,
//$$$  1.0000, 0.0471, 0.0000,    1.0000, 0.0314, 0.0000,   1.0000, 0.0157, 0.0000,    1.0000, 0.0000, 0.0000 };

/***********************************************************************/
/*                                                                     */
/* global variables                                                    */
/*                                                                     */
/***********************************************************************/

/* global variable holding a MODL */
static void       *modl;               /* pointer to MODL */
static int        outLevel = 1;        /* default output level */
static double     tessFact = 1;        /* tessellation parameter factor */
static int        printBreps = 0;      /* =1 to print BReps on stack */
static int        batch = 0;           /* =0 to enable visualization */

/* global variables associated with graphical user interface (gui) */
static wvContext  *cntxt;              /* context for the WebViewer */
static int         port   = 7681;      /* port number */

/* global variables associated with undo */
#define MAX_UNDOS  100
static int         nundo  = 0;         /* number of undos */
static void        **undo_modl = NULL;
static char        **undo_text = NULL;

/* global variables associated with scene graph meta-data */
#define MAX_METADATA_LEN 32000
static char        *sgMetaData  = NULL;
static char        *sgTempData  = NULL;
static char        *sgFocusData = NULL;

/* global variables associated with bodyList */
#define MAX_BODYS        999
static int        nbody     = 0;       /* number of built Bodys */
static int        *bodyList = NULL;

/* global variables associated with pending errors */
static int        pendingError = 0;

/* global variables associated with sensitivities */
static float      lims[2]  = {-1.0, +1.0};
static int        haveDots = 0, sensPost = 0;
static double     sensLo = 0.0, sensHi = 0.0;
static char       *dotName=NULL;

/* global variables associated with the response buffer */
static int        max_resp_len = 0;
static int        response_len = 0;
static char       *response = NULL;

/* global variables associated with journals */
static FILE       *jrnl_out = NULL;    /* output journal file */
static char       *filename = NULL;

/***********************************************************************/
/*                                                                     */
/* declarations                                                        */
/*                                                                     */
/***********************************************************************/

/* declarations for high-level routines defined below */

/* declarations for routines defined below */
static void       addToResponse(char *text);
       void       browserMessage(void *wsi, char *text, /*@unused@*/ int lena);
static int        buildBodys(int buildTo, int *builtTo, int *buildStatus);
static int        buildSceneGraph();
static int        getToken(char *text, int nskip, char sep, char *token);
static void       processMessage(char *text);
static void       spec_col(float scalar, float lims[], float out[]);
static int        storeUndo(char *cmd, char *arg);

/* external functions not declared in an .h file */


/***********************************************************************/
/*                                                                     */
/*   main program                                                      */
/*                                                                     */
/***********************************************************************/

int
main(int       argc,                    /* (in)  number of arguments */
     char      *argv[])                 /* (in)  array of arguments */
{

    int       imajor, iminor, status, i, bias, showUsage=0;
    int       builtTo, buildStatus;
    float     fov, zNear, zFar;
    float     eye[3]    = {0.0, 0.0, 7.0};
    float     center[3] = {0.0, 0.0, 0.0};
    float     up[3]     = {0.0, 1.0, 0.0};
    char      *casename=NULL, *jrnlname=NULL, *tempname=NULL;
    char      *text=NULL, *esp_start;
    void      *temp;
    FILE      *jrnl_in=NULL;
    modl_T    *MODL;

    int       jbody, ibody;

    clock_t   old_time, new_time;

    ROUTINE(MAIN);

    /* --------------------------------------------------------------- */

    /* dynamically allocated array so that everything is on heap (not stack) */
    MALLOC(dotName,     char, 129             );
    MALLOC(sgMetaData,  char, MAX_METADATA_LEN);
    MALLOC(sgTempData,  char, MAX_METADATA_LEN);
    MALLOC(sgFocusData, char, 256             );
    MALLOC(bodyList,    int,  MAX_BODYS       );
    MALLOC(filename,    char, 257             );
    MALLOC(casename,    char, 257             );
    MALLOC(jrnlname,    char, 257             );
    MALLOC(tempname,    char, 257             );
    MALLOC(text,        char, MAX_STR_LEN     );

    MALLOC(undo_modl, void*, MAX_UNDOS+1);
    MALLOC(undo_text, char*, MAX_UNDOS+1);
    for (i = 0; i <= MAX_UNDOS; i++) {
        undo_text[i] = NULL;
    }
    for (i = 0; i <= MAX_UNDOS; i++) {
        MALLOC(undo_text[i], char, 32);
    }

    DPRINT0("starting serveCSM");

    /* get the flags and casename(s) from the command line */
    casename[0] = '\0';
    jrnlname[0] = '\0';

    for (i = 1; i < argc; i++) {
        if        (strcmp(argv[i], "-port") == 0) {
            if (i < argc-1) {
                sscanf(argv[++i], "%d", &port);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-jrnl") == 0) {
            if (i < argc-1) {
                STRNCPY(jrnlname, argv[++i], 257);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-outLevel") == 0) {
            if (i < argc-1) {
                sscanf(argv[++i], "%d", &outLevel);
                if (outLevel < 0) outLevel = 0;
                if (outLevel > 3) outLevel = 3;
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-tessFact") == 0) {
            if (i < argc-1) {
                sscanf(argv[++i], "%lf", &tessFact);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-printBreps") == 0) {
            printBreps = 1;
        } else if (strcmp(argv[i], "-batch") == 0) {
            batch = 1;
        } else if (strlen(casename) == 0) {
            STRNCPY(casename, argv[i], 257);
        } else {
            SPRINT0(0, "two casenames given");
            showUsage = 1;
            break;
        }
    }

    if (showUsage) {
        SPRINT0(0, "proper usage: 'serveCSM [-port X] [-jrnl jrnlname] [-outLevel X] [-tessFact factor] [-batch] [-printBreps] [casename[.csm]]'");
        SPRINT0(0, "STOPPING...\a");
        exit(EXIT_FAILURE);
    }

    /* welcome banner */
    (void) ocsmVersion(&imajor, &iminor);

    SPRINT0(1, "**********************************************************");
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "*                    Program serveCSM                    *");
    SPRINT2(1, "*                     version %2d.%02d                      *", imajor, iminor);
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "*        written by John Dannenhoffer, 2010/2015         *");
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "**********************************************************");

    /* set OCSMs output level */
    (void) ocsmSetOutLevel(outLevel);

    /* allocate nominal response buffer */
    max_resp_len = 4096;
    response = (char*) malloc(max_resp_len*sizeof(char));
    if (response == NULL) {
        SPRINT0(0, "ERROR:: error mallocing response");
        exit(EXIT_FAILURE);
    }

    /* strip off .csm (which is assumed to be at the end) if present */
    if (strlen(casename) > 0) {
        STRNCPY(filename, casename, 257);
        if (strstr(casename, ".csm") == NULL) {
            strcat(filename, ".csm");
        }
    } else {
        casename[0] = '\0';
        filename[0] = '\0';
    }

    /* read the .csm file and create the MODL */
    old_time = clock();
    status   = ocsmLoad(filename, &modl);
    new_time = clock();
    SPRINT3(1, "--> ocsmLoad(%s) -> status=%d (%s)", filename, status, ocsmGetText(status));
    SPRINT1(1, "==> ocsmLoad CPUtime=%9.3f sec",
            (double)(new_time-old_time) / (double)(CLOCKS_PER_SEC));

    if (status < SUCCESS) {
        SPRINT0(0, "ERROR:: problem in ocsmLoad");
        exit(EXIT_FAILURE);
    } else {
        MODL = (modl_T*)modl;
    }


    /* check that Branches are properly ordered */
    old_time = clock();
    status   = ocsmCheck(modl);
    new_time = clock();
    SPRINT2(1, "--> ocsmCheck() -> status=%d (%s)",
            status, ocsmGetText(status));
    SPRINT1(1, "==> ocsmCheck CPUtime=%10.3f sec",
            (double)(new_time-old_time) / (double)(CLOCKS_PER_SEC));

    if (status < SUCCESS) {
        SPRINT0(0, "ERROR:: problem in ocsmCheck");
        exit(EXIT_FAILURE);
    }

    /* print out the global Attributes, Parameters, and Branches */
    if (batch == 1) {
        SPRINT0(1, "External Parameter(s):");
        if (outLevel > 0) {
            status = ocsmPrintPmtrs(modl, stdout);
            if (status != SUCCESS) {
                SPRINT1(0, "ocsmPrintPmtrs -> status=%d", status);
            }
        }

        SPRINT0(1, "Branch(es):");
        if (outLevel > 0) {
            status = ocsmPrintBrchs(modl, stdout);
            if (status != SUCCESS) {
                SPRINT1(0, "ocsmPrintBrchs -> status=%d", status);
            }
        }

        SPRINT0(1, "Global Attribute(s):");
        if (outLevel > 0) {
            status = ocsmPrintAttrs(modl, stdout);
            if (status != SUCCESS) {
                SPRINT1(0, "ocsmPrintAttrs -> status=%d", status);
            }
        }
    }

    /* open the output journal file */
    snprintf(tempname, 257, "port%d.jrnl", port);
    jrnl_out = fopen(tempname, "w");

    /* initialize the scene graph meta data */
    if (batch == 0) {
        sgMetaData[ 0] = '\0';
        sgFocusData[0] = '\0';
    }

    /* create the WebViewer context */
    if (batch == 0) {
        bias  =  1;
        fov   = 30.0;
        zNear =  1.0;
        zFar  = 10.0;
        cntxt = wv_createContext(bias, fov, zNear, zFar, eye, center, up);
        if (cntxt == NULL) {
            SPRINT0(0, "ERROR:: failed to create wvContext");
            exit(EXIT_FAILURE);
        }
    }

    /* build the Bodys from the MODL */
    status = buildBodys(0, &builtTo, &buildStatus);

    if (builtTo < 0) {
        SPRINT2(0, "build() detected \"%s\" at %s",
                ocsmGetText(buildStatus), MODL->brch[1-builtTo].name);
        SPRINT0(0, "Configuration only built up to detected error\a");
        pendingError = -builtTo;
    } else if (status != SUCCESS) {
        SPRINT1(0, "ERROR:: build() detected \"%s\"",
                ocsmGetText(buildStatus));
        exit(EXIT_FAILURE);
    }

    /* print out Breps (if requested) */
    if (printBreps == 1) {
        for (ibody = 1; ibody <= MODL->nbody; ibody++) {
            if (MODL->body[ibody].onstack == 1) {
                SPRINT1(1, "Body %d", ibody);
                (void) ocsmPrintBrep(modl, ibody, stdout);
            }
        }
    }

    /* process the input journal file if jrnlname exists */
    if (strlen(jrnlname) > 0) {
        SPRINT1(0, "\n==> Opening input journal file \"%s\"\n", jrnlname);

        jrnl_in = fopen(jrnlname, "r");
        if (jrnl_in == NULL) {
            SPRINT0(0, "ERROR:: Journal file cannot be opened");
            exit(EXIT_FAILURE);
        } else {
            while (1) {
                temp = fgets(text, MAX_STR_LEN, jrnl_in);
                if (temp == NULL) break;

                if (feof(jrnl_in)) break;

                /* if there is a \n, convert it to a \0 */
                text[MAX_STR_LEN-1] = '\0';

                for (i = 0; i < strlen(text); i++) {
                    if (text[i] == '\n') {
                        text[i] =  '\0';
                        break;
                    }
                }

                processMessage(text);
            }

            fclose(jrnl_in);

            SPRINT0(0, "\n==> Closing input journal file\n");
        }
    }

    /* get the command to start the client (if any) */
    if (batch == 0) {
        esp_start = getenv("ESP_START");

        if (esp_start == NULL) {
            esp_start = getenv("WV_START");
        }
    }

    /* start the server */
    if (batch == 0) {
        status = SUCCESS;
        if (wv_startServer(port, NULL, NULL, NULL, 0, cntxt) == 0) {

            /* stay alive a long as we have a client */
            while (wv_statusServer(0)) {
                usleep(500000);

                /* start the browser if the first time through this loop */
                if (status == SUCCESS) {
                    if (esp_start != NULL) {
                        status += system(esp_start);
                    }

                    status++;
                }
            }
        }
    }

    /* cleanup and exit */
    if (jrnl_out != NULL) fclose(jrnl_out);
    jrnl_out = NULL;

    MODL = (modl_T*)modl;

    for (jbody = 0; jbody < nbody; jbody++) {
        ibody = bodyList[jbody];

        if (MODL->body[ibody].etess != NULL) {
            status = EG_deleteObject(MODL->body[ibody].etess);
            SPRINT2(2, "--> EG_deleteObject(etess[%d]) -> status=%d", ibody, status);
        }

        if (MODL->body[ibody].ebody != NULL) {
            status = EG_deleteObject(MODL->body[ibody].ebody);
            SPRINT2(2, "--> EG_deleteObject(ebody[%d]) => status=%d", ibody, status);
        }
    }

    if (MODL->context != NULL) {
        status = EG_setOutLevel(MODL->context, 0);
        if (status < 0) {
            SPRINT1(0, "EG_setOutLevel -> status=%d", status);
        }

        status = EG_close(MODL->context);
        SPRINT1(1, "--> EG_close() -> status=%d", status);

        /* closing the context automatically removed all ebody and etess */
        for (ibody = 1; ibody <= MODL->nbody; ibody++) {
            MODL->body[ibody].ebody = NULL;
            MODL->body[ibody].etess = NULL;
        }
    }

    udp_cleanupAll();

    status = ocsmFree(modl);
    SPRINT2(1, "--> ocsmFree() -> status=%d (%s)", status, ocsmGetText(status));

    status = ocsmFree(NULL);
    SPRINT2(1, "--> ocsmFree(NULL) -> status=%d (%s)", status, ocsmGetText(status));

    free(response);
    response = NULL;

    wv_cleanupServers();

    SPRINT0(0, "==> serveCSM completed successfully");

cleanup:
    if (undo_text != NULL) {
        for (i = 0; i <= MAX_UNDOS; i++) {
            FREE(undo_text[i]);
        }
        FREE(undo_text  );
    }
    FREE(undo_modl  );
    FREE(text       );
    FREE(tempname   );
    FREE(jrnlname   );
    FREE(casename   );
    FREE(filename   );
    FREE(bodyList   );
    FREE(sgFocusData);
    FREE(sgTempData );
    FREE(sgMetaData );
    FREE(dotName    );

    return EXIT_SUCCESS;
}


/***********************************************************************/
/*                                                                     */
/*   addToResponse - add text to response (with buffer length protection) */
/*                                                                     */
/***********************************************************************/

static void
addToResponse(char   *text)             /* (in)  text to add */
{
    int    text_len;

    text_len = strlen(text);

    while (response_len+text_len > max_resp_len-2) {
        max_resp_len +=4096;
        SPRINT1(2, "increasing max_resp_len=%d", max_resp_len);

        response = (char*) realloc(response, max_resp_len*sizeof(char));
        if (response == NULL) {
            SPRINT0(0, "ERROR:: error mallocing response");
            exit(EXIT_FAILURE);
        }
    }

    strcpy(&(response[response_len]), text);

    response_len += text_len;
}


/***********************************************************************/
/*                                                                     */
/*   browserMessage - called when client sends a message to the server */
/*                                                                     */
/***********************************************************************/

void
browserMessage(void    *wsi,
               char    *text,
  /*@unused@*/ int     lena)
{
    int       status;

    char      message2[129];

    modl_T    *MODL;

    ROUTINE(browserMessage);

    /* --------------------------------------------------------------- */

    /* process the Message */
    processMessage(text);

    /* send the response */
    SPRINT1(2, "response-> %s", response);
    wv_sendText(wsi, response);

    /* if the sensitivities were just computed, send a message to
       inform the user about the lowest and highest sensitivity values */
    if (sensPost > 0) {
        snprintf(message2, 128, "Sensitivities are in the range between %f and %f", sensLo, sensHi);
        wv_sendText(wsi, message2);

        sensPost = 0;
    }

    /* send the scene graph meta data if it has not already been sent */
    if (strlen(sgMetaData) > 0) {
        SPRINT1(2, "sgData-> %s", sgMetaData);
        wv_sendText(wsi, sgMetaData);

        /* nullify meta data so that it does not get sent again */
        sgMetaData[0] = '\0';
    }

    if (strlen(sgFocusData) > 0) {
        SPRINT1(2, "sgFocus-> %s", sgFocusData);
        wv_sendText(wsi, sgFocusData);

        /* nullify meta data so that it does not get sent again */
        sgFocusData[0] = '\0';
    }

    /* either open or close the key */
    if        (haveDots >  1) {
        status = wv_setKey(cntxt, 256, color_map, lims[0], lims[1], "d(norm)/d(***)");
        SPRINT0(2, "message-> setWvKey|on|");
        wv_sendText(wsi, "setWvKey|on|");
    } else if (haveDots == 1) {
        status = wv_setKey(cntxt, 256, color_map, lims[0], lims[1], dotName         );
        SPRINT0(2, "message-> setWvKey|on|");
        wv_sendText(wsi, "setWvKey|on|");
    } else {
        status = wv_setKey(cntxt,   0, NULL,      lims[0], lims[1], NULL            );
        SPRINT0(2, "message-> setWvKey|off|");
        wv_sendText(wsi, "setWvKey|off|");
    }
    if (status != SUCCESS) {
        SPRINT1(0, "wv_setKet -> status=%d", status);
    }

    /* send an error message if one is pending */
    if (pendingError > 0) {
        MODL = (modl_T*)modl;

        snprintf(response, max_resp_len, "%s", MODL->errorBuf);
        pendingError = 0;

        SPRINT1(2, "response-> %s", response);
        wv_sendText(wsi, response);

    } else if (pendingError == -1) {
        snprintf(response, max_resp_len, "ERROR:: could not find Design Velocities; shown as zeros");
        pendingError = 0;

        SPRINT1(2, "response-> %s", response);
        wv_sendText(wsi, response);

    }

}


/***********************************************************************/
/*                                                                     */
/*   buildBodys - build Bodys and update scene graph                   */
/*                                                                     */
/***********************************************************************/

static int
buildBodys(int     buildTo,             /* (in)  last Branch to execute */
           int     *builtTo,            /* (out) last Branch successfully executed */
           int     *buildStatus)        /* (out) status returned from build */
{
    int       status = SUCCESS;         /* return status */

    modl_T    *MODL = (modl_T*)modl;

    int       ibody, jbody, status_min, status2;
    double    box[6], size, params[3];
    clock_t   old_time, new_time;

    ego       ebody;

    ROUTINE(buildBodys);

    /* --------------------------------------------------------------- */

    /* if there are no Branches, simply return */
    if (MODL->nbrch <= 0) {
        SPRINT0(1, "--> No Branches, so skipping build");
        *builtTo     = 0;
        *buildStatus = SUCCESS;

    } else {
        /* check that Branches are properly ordered */
        old_time = clock();
        status   = ocsmCheck(modl);
        new_time = clock();
        SPRINT2(0, "--> ocsmCheck() -> status=%d (%s)",
                status, ocsmGetText(status));
        SPRINT1(0, "==> ocsmCheck CPUtime=%10.3f sec",
                (double)(new_time-old_time) / (double)(CLOCKS_PER_SEC));

        if (status < SUCCESS) {
            goto cleanup;
        }

        /* build the Bodys */
        nbody    = MAX_BODYS;
        old_time = clock();
        status   = ocsmBuild(modl, buildTo, builtTo, &nbody, bodyList);
        new_time = clock();
        SPRINT5(0, "--> ocsmBuild(buildTo=%d) -> status=%d (%s), builtTo=%d, nbody=%d",
                buildTo, status, ocsmGetText(status), *builtTo, nbody);
        SPRINT1(0, "==> ocsmBuild CPUtime=%10.3f sec",
                (double)(new_time-old_time) / (double)(CLOCKS_PER_SEC));

        /* print out the Bodys */
        if (outLevel > 0) {
            status2 = ocsmPrintBodys(modl, stdout);
            if (status2 != SUCCESS) printf("ocsmPrintBodys -> status=%d\n", status2);
        }

        if (batch == 1 && status < SUCCESS) {
            SPRINT1(0, "ERROR:: build not completed because error %d was detected", status);
            exit(EXIT_FAILURE);
        }

        *buildStatus = status;

        /* tessellate the Bodys */
        old_time = clock();
        status_min = SUCCESS;
        for (jbody = 0; jbody < nbody; jbody++) {
            ibody = bodyList[jbody];

            /* if we already have an etess, no need to regenerate it here */
            if (MODL->body[ibody].etess != NULL) {
                SPRINT1(0, "--> reusing TessBody(ibody=%d)", ibody);
                continue;
            }

            ebody  = MODL->body[ibody].ebody;
            status = EG_getBoundingBox(ebody, box);
            if (status != SUCCESS) {
                SPRINT1(0, "EG_getBoundingBox -> status=%d", status);
            }

            size = sqrt(SQR(box[3]-box[0]) + SQR(box[4]-box[1]) + SQR(box[5]-box[2]));

            /* vTess parameters */
            params[0] = 0.0250 * size * tessFact;
            params[1] = 0.0010 * size * tessFact;
            params[2] = 15.0;

            ebody = MODL->body[ibody].ebody;

            if (MODL->context != NULL) {
                (void) EG_setOutLevel(MODL->context, 1);
                status = EG_makeTessBody(ebody, params, &(MODL->body[ibody].etess));
                (void) EG_setOutLevel(MODL->context, outLevel);
                SPRINT6(0, "--> EG_makeTessBody(ibody=%4d, params=%10.5f, %10.5f, %10.5f) -> status=%d (%s)",
                        ibody, params[0], params[1], params[2], status, ocsmGetText(status));

                status_min = MIN(status_min, status);
            }
        }

        new_time = clock();
        SPRINT1(0, "==> EG_makeTessBody CPUtime=%10.3f sec",
                (double)(new_time-old_time) / (double)(CLOCKS_PER_SEC));

        status = status_min;
        if (status < SUCCESS) {
            goto cleanup;
        }
    }

    /* build the scene graph */
    if (batch == 0) {
        buildSceneGraph();
    }

cleanup:
    return status;
}


/***********************************************************************/
/*                                                                     */
/*   buildSceneGraph - make a scene graph for wv                       */
/*                                                                     */
/***********************************************************************/

static int
buildSceneGraph()
{
    int       status = SUCCESS;

    int       ibody, jbody, iface, iedge, iattr, nattr, ipmtr, irc;
    int       npnt, ipnt, ntri, itri, igprim, nseg, k, attrs, head;
    int       *segs=NULL, *ivrts=NULL;
    CINT      *ptype, *pindx, *tris, *tric;
    float     focus[4], color[3], *pcolors;
    double    bigbox[6], box[6], size, xyz_dum[3], *vel;
    CDOUBLE   *xyz, *uv, *t;
    char      gpname[33];

    ego       ebody, etess, eface, eedge;
    int       i, nlist, itype;
    CINT      *tempIlist;
    CDOUBLE   *tempRlist;
    CCHAR     *attrName, *tempClist;

    wvData    items[5];

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(buildSceneGraph);

    /* --------------------------------------------------------------- */

    /* remove any graphic primitives that already exist */
    wv_removeAll(cntxt);

    /* find the values needed to adjust the vertices */
    bigbox[0] = bigbox[1] = bigbox[2] = +HUGEQ;
    bigbox[3] = bigbox[4] = bigbox[5] = -HUGEQ;

    for (jbody = 0; jbody < nbody; jbody++) {
        ibody = bodyList[jbody];

        ebody = MODL->body[ibody].ebody;

        status = EG_getBoundingBox(ebody, box);

        if (box[0] < bigbox[0]) bigbox[0] = box[0];
        if (box[1] < bigbox[1]) bigbox[1] = box[1];
        if (box[2] < bigbox[2]) bigbox[2] = box[2];
        if (box[3] > bigbox[3]) bigbox[3] = box[3];
        if (box[4] > bigbox[4]) bigbox[4] = box[4];
        if (box[5] > bigbox[5]) bigbox[5] = box[5];
    }

                                    size = bigbox[3] - bigbox[0];
    if (size < bigbox[4]-bigbox[1]) size = bigbox[4] - bigbox[1];
    if (size < bigbox[5]-bigbox[2]) size = bigbox[5] - bigbox[2];

    focus[0] = (bigbox[0] + bigbox[3]) / 2;
    focus[1] = (bigbox[1] + bigbox[4]) / 2;
    focus[2] = (bigbox[2] + bigbox[5]) / 2;
    focus[3] = size;

    /* generate the scene graph focus data */
    snprintf(sgFocusData, 256, "sgFocus|[%20.12e,%20.12e,%20.12e,%20.12e]", focus[0], focus[1], focus[2], focus[3]);

    /* keep track of minimum and maximum velocities */
    sensLo = +HUGEQ;
    sensHi = -HUGEQ;

    /* initialize the scene graph meta data */
    STRNCPY(sgMetaData, "sgData|{", MAX_METADATA_LEN);

    /* special scene graph if nbody<=0 */
    if (nbody <= 0) {
        snprintf(gpname, 33, "Body 0 Face 0");
        attrs = WV_ON;

        xyz_dum[0] = 0;
        xyz_dum[1] = 0;
        xyz_dum[2] = 0;

        status = wv_setData(WV_REAL64, 1, (void*)xyz_dum, WV_VERTICES, &(items[0]));
        if (status != SUCCESS) {
            SPRINT1(0, "wv_setData -> status=%d", status);
        }

        wv_adjustVerts(&(items[0]), focus);

        color[0] = 0;
        color[1] = 0;
        color[2] = 0;
        status = wv_setData(WV_REAL32, 1, (void*)color, WV_PCOLOR, &(items[1]));
        if (status != SUCCESS) {
            SPRINT1(0, "wv_setData -> status=%d", status);
        }

        igprim = wv_addGPrim(cntxt, gpname, WV_LINE, attrs, 2, items);
        if (igprim < 0) {
            SPRINT1(0, "wv_addGPrim -> igprim=%d", igprim);
        }
    }

    /* loop through the Bodys */
    for (jbody = 0; jbody < nbody; jbody++) {
        ibody = bodyList[jbody];

        etess = MODL->body[ibody].etess;

        /* determine if any of the external Parameters have a velocity */
        haveDots   = 0;
        dotName[0] = '\0';

        for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
            if (MODL->pmtr[ipmtr].type == OCSM_EXTERNAL) {
                for (irc = 0; irc < (MODL->pmtr[ipmtr].nrow)*(MODL->pmtr[ipmtr].ncol); irc++) {
                    if (MODL->pmtr[ipmtr].dot[irc] != 0) {
                        if (fabs(MODL->pmtr[ipmtr].dot[irc]-1) < EPS06) {
                            if (haveDots == 0) {
                                sprintf(dotName, "d(norm)/d(%s)", MODL->pmtr[ipmtr].name);
                            } else {
                                sprintf(dotName, "d(norm)/d(***)");
                            }
                            haveDots++;
                        } else if (MODL->pmtr[ipmtr].dot[irc] != 0) {
                            sprintf(dotName, "d(norm)/d(***)");
                            haveDots++;
                        }
                    }
                }
            }
        }

        /* loop through the Faces within each Body */
        for (iface = 1; iface <= MODL->body[ibody].nface; iface++) {
            status = EG_getTessFace(etess, iface,
                                    &npnt, &xyz, &uv, &ptype, &pindx,
                                    &ntri, &tris, &tric);
            if (status != SUCCESS) {
                SPRINT1(0, "EG_getTessFace -> status=%d", status);
            }

            /* name and attributes */
            snprintf(gpname, 33, "Body %d Face %d", ibody, iface);
            if (haveDots >= 1) {
                attrs = WV_ON | WV_SHADING | WV_ORIENTATION;
            } else {
                attrs = WV_ON              | WV_ORIENTATION;
            }

            /* vertices */
            status = wv_setData(WV_REAL64, npnt, (void*)xyz, WV_VERTICES, &(items[0]));
            if (status != SUCCESS) {
                SPRINT1(0, "wv_setData -> status=%d", status);
            }

            wv_adjustVerts(&(items[0]), focus);

            /* triangles */
            status = wv_setData(WV_INT32, 3*ntri, (void*)tris, WV_INDICES, &(items[1]));
            if (status != SUCCESS) {
                SPRINT1(0, "wv_setData -> status=%d", status);
            }

            /* smooth triangle colors */
            if (haveDots >= 1) {
                sensPost = 1;

                vel = (double *) malloc(npnt * sizeof(double));
                if (vel == NULL) goto cleanup;

                pcolors = (float *) malloc(3*npnt*sizeof(float));
                if (pcolors == NULL) goto cleanup;

                status = ocsmGetVel(MODL, ibody, OCSM_FACE, iface, npnt, NULL, vel);
                if (status != SUCCESS) {
                    pendingError = -1;
                }

                for (ipnt = 0; ipnt < npnt; ipnt++) {
                    spec_col((float)(vel[ipnt]), lims, &(pcolors[3*ipnt]));

                    if (vel[ipnt] < sensLo) sensLo = vel[ipnt];
                    if (vel[ipnt] > sensHi) sensHi = vel[ipnt];
                }
                status = wv_setData(WV_REAL32, npnt, (void*)pcolors, WV_COLORS, &(items[2]));
                if (status != SUCCESS) {
                    SPRINT1(0, "wv_setData -> status=%d", status);
                }

                free(pcolors);
                free(vel);

            /* constant triangle colors */
            } else {
                color[0] = RED(  MODL->body[ibody].face[iface].gratt.color);
                color[1] = GREEN(MODL->body[ibody].face[iface].gratt.color);
                color[2] = BLUE( MODL->body[ibody].face[iface].gratt.color);
                status = wv_setData(WV_REAL32, 1, (void*)color, WV_COLORS, &(items[2]));
                if (status != SUCCESS) {
                    SPRINT1(0, "wv_setData -> status=%d", status);
                }
            }

            /* triangle sides (segments) */
            nseg = 0;
            for (itri = 0; itri < ntri; itri++) {
                for (k = 0; k < 3; k++) {
                    if (tric[3*itri+k] < itri+1) {
                        nseg++;
                    }
                }
            }

            segs = (int*) malloc(2*nseg*sizeof(int));
            if (segs == NULL) goto cleanup;

            nseg = 0;
            for (itri = 0; itri < ntri; itri++) {
                for (k = 0; k < 3; k++) {
                    if (tric[3*itri+k] < itri+1) {
                        segs[2*nseg  ] = tris[3*itri+(k+1)%3];
                        segs[2*nseg+1] = tris[3*itri+(k+2)%3];
                        nseg++;
                    }
                }
            }

            status = wv_setData(WV_INT32, 2*nseg, (void*)segs, WV_LINDICES, &(items[3]));
            if (status != SUCCESS) {
                SPRINT1(0, "wv_setData -> status=%d", status);
            }

            free(segs);

            /* segment colors */
            color[0] = 0.0;
            color[1] = 0.0;
            color[2] = 0.0;
            status = wv_setData(WV_REAL32, 1, (void*)color, WV_LCOLOR, &(items[4]));
            if (status != SUCCESS) {
                SPRINT1(0, "wv_setData -> status=%d", status);
            }

            /* make graphic primitive */
            igprim = wv_addGPrim(cntxt, gpname, WV_TRIANGLE, attrs, 5, items);
            if (igprim >= 0) {

                /* make line width 1 */
                cntxt->gPrims[igprim].lWidth = 1.0;
            }

            eface  = MODL->body[ibody].face[iface].eface;
            status = EG_attributeNum(eface, &nattr);

            /* add Face to meta data (if there is room) */
            if (strlen(sgMetaData) < MAX_METADATA_LEN-1000) {
                if (nattr > 0) {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[",  sgTempData, gpname);
                } else {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[]", sgTempData, gpname);
                }

                for (iattr = 1; iattr <= nattr; iattr++) {
                    status = EG_attributeGet(eface, iattr, &attrName, &itype, &nlist,
                                             &tempIlist, &tempRlist, &tempClist);

                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\",\"", sgTempData, attrName);

                    if        (itype == ATTRINT) {
                        for (i = 0; i < nlist ; i++) {
                            STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                            snprintf(sgMetaData, MAX_METADATA_LEN, "%s %d", sgTempData, tempIlist[i]);
                        }
                    } else if (itype == ATTRREAL) {
                        for (i = 0; i < nlist ; i++) {
                            STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                            snprintf(sgMetaData, MAX_METADATA_LEN, "%s %f", sgTempData, tempRlist[i]);
                        }
                    } else if (itype == ATTRSTRING) {
                        STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                        snprintf(sgMetaData, MAX_METADATA_LEN, "%s %s ", sgTempData, tempClist);
                    }

                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\",", sgTempData);
                }
                sgMetaData[MAX_METADATA_LEN-1] = '\0';
                STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                snprintf(sgMetaData, MAX_METADATA_LEN, "%s],", sgTempData);
            }
        }

        /* loop through the Edges within each Body */
        for (iedge = 1; iedge <= MODL->body[ibody].nedge; iedge++) {
            status = EG_getTessEdge(etess, iedge, &npnt, &xyz, &t);
            if (status != SUCCESS) {
                SPRINT1(0, "EG_getTessEdge -> status=%d", status);
            }

            /* name and attributes */
            snprintf(gpname, 33, "Body %d Edge %d", ibody, iedge);
            attrs = WV_ON;

            /* vertices */
            status = wv_setData(WV_REAL64, npnt, (void*)xyz, WV_VERTICES, &(items[0]));
            if (status != SUCCESS) {
                SPRINT1(0, "wv_setData -> status=%d", status);
            }

            wv_adjustVerts(&(items[0]), focus);

            /* segments */
            ivrts = (int*) malloc(2*(npnt-1)*sizeof(int));
            if (ivrts == NULL) goto cleanup;

            for (ipnt = 0; ipnt < npnt-1; ipnt++) {
                ivrts[2*ipnt  ] = ipnt + 1;
                ivrts[2*ipnt+1] = ipnt + 2;
            }

            status = wv_setData(WV_INT32, 2*(npnt-1), (void*)ivrts, WV_INDICES, &(items[1]));
            if (status != SUCCESS) {
                SPRINT1(0, "wv_setData -> status=%d", status);
            }

            free(ivrts);

            /* line colors */
            color[0] = RED(  MODL->body[ibody].edge[iedge].gratt.color);
            color[1] = GREEN(MODL->body[ibody].edge[iedge].gratt.color);
            color[2] = BLUE( MODL->body[ibody].edge[iedge].gratt.color);
            status = wv_setData(WV_REAL32, 1, (void*)color, WV_COLORS, &(items[2]));
            if (status != SUCCESS) {
                SPRINT1(0, "wv_setData -> status=%d", status);
            }

            /* points */
            ivrts = (int*) malloc(npnt*sizeof(int));
            if (ivrts == NULL) goto cleanup;

            for (ipnt = 0; ipnt < npnt; ipnt++) {
                ivrts[ipnt] = ipnt + 1;
            }

            status = wv_setData(WV_INT32, npnt, (void*)ivrts, WV_PINDICES, &(items[3]));
            if (status != SUCCESS) {
                SPRINT1(0, "wv_setData -> status=%d", status);
            }

            free(ivrts);

            /* point colors */
            color[0] = 0;
            color[1] = 0;
            color[2] = 0;
            status = wv_setData(WV_REAL32, 1, (void*)color, WV_PCOLOR, &(items[4]));
            if (status != SUCCESS) {
                SPRINT1(0, "wv_setData -> status=%d", status);
            }

            /* make graphic primitive */
            igprim = wv_addGPrim(cntxt, gpname, WV_LINE, attrs, 5, items);
            if (igprim >= 0) {

                /* make line width 2 */
                cntxt->gPrims[igprim].lWidth = 2.0;

                /* make point size 5 */
                cntxt->gPrims[igprim].pSize  = 5.0;

                /* add arrow heads */
                head = npnt - 1;
                status = wv_addArrowHeads(cntxt, igprim, 0.05, 1, &head);
                if (status != SUCCESS) {
                    SPRINT1(0, "wv_addArrowHeads -> status=%d", status);
                }
            }

            eedge  = MODL->body[ibody].edge[iedge].eedge;
            status = EG_attributeNum(eedge, &nattr);

            /* add Edge to meta data (if there is room) */
            if (strlen(sgMetaData) < MAX_METADATA_LEN-1000) {
                if (nattr > 0) {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[",  sgTempData, gpname);
                } else {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[]", sgTempData, gpname);
                }

                for (iattr = 1; iattr <= nattr; iattr++) {
                    status = EG_attributeGet(eedge, iattr, &attrName, &itype, &nlist,
                                             &tempIlist, &tempRlist, &tempClist);

                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\",\"", sgTempData, attrName);

                    if        (itype == ATTRINT) {
                        for (i = 0; i < nlist ; i++) {
                            STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                            snprintf(sgMetaData, MAX_METADATA_LEN, "%s %d", sgTempData, tempIlist[i]);
                        }
                    } else if (itype == ATTRREAL) {
                        for (i = 0; i < nlist ; i++) {
                            STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                            snprintf(sgMetaData, MAX_METADATA_LEN, "%s %f", sgTempData, tempRlist[i]);
                        }
                    } else if (itype == ATTRSTRING) {
                        STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                        snprintf(sgMetaData, MAX_METADATA_LEN, "%s %s ", sgTempData, tempClist);
                    }

                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\",", sgTempData);
                }
                sgMetaData[MAX_METADATA_LEN-1] = '\0';
                STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                snprintf(sgMetaData, MAX_METADATA_LEN, "%s],", sgTempData);
            }
        }
    }

    /* finish the scene graph meta data */
    sgMetaData[MAX_METADATA_LEN-1] = '\0';
    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
    snprintf(sgMetaData, MAX_METADATA_LEN, "%s}", sgTempData);

cleanup:
    return status;
}


/***********************************************************************/
/*                                                                     */
/*   getToken - get a token from a string                              */
/*                                                                     */
/***********************************************************************/

static int
getToken(char   *text,                  /* (in)  full text */
         int    nskip,                  /* (in)  tokens to skip */
         char   sep,                    /* (in)  separator character */
         char   *token)                 /* (out) token */
{
    int    lentok, i, count, iskip;

    token[0] = '\0';
    lentok   = 0;

    /* count the number of separators */
    count = 0;
    for (i = 0; i < strlen(text); i++) {
        if (text[i] == sep) {
            count++;
        }
    }

    if (count < nskip+1) return 0;

    /* skip over nskip tokens */
    i = 0;
    for (iskip = 0; iskip < nskip; iskip++) {
        while (text[i] != sep) {
            i++;
        }
        i++;
    }

    /* extract the token we are looking for */
    while (text[i] != sep) {
        token[lentok++] = text[i++];
        token[lentok  ] = '\0';

        if (lentok >= MAX_EXPR_LEN-1) {
            SPRINT0(0, "ERROR:: token exceeds MAX_EXPR_LEN");
            break;
        }
    }

    return strlen(token);
}


/***********************************************************************/
/*                                                                     */
/*   processMessage - process the message from client and create the response */
/*                                                                     */
/***********************************************************************/

static void
processMessage(char    *text)
{
    int       i, status, ibrch, itype, builtTo, buildStatus, ichar;
    int       ipmtr, jpmtr, nrow, ncol, irow, icol, index, iattr, actv, icount;
    char      *pEnd;

    char      *name=NULL,  *type=NULL, *valu=NULL;
    char      *arg1=NULL,  *arg2=NULL, *arg3=NULL;
    char      *arg4=NULL,  *arg5=NULL, *arg6=NULL;
    char      *arg7=NULL,  *arg8=NULL, *arg9=NULL;
    char      *entry=NULL, *temp=NULL, *temp2=NULL;
    char      begs[1024], vars[1024], cons[1024], segs[1024];

    static FILE  *fp=NULL;

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(processMessage);

    /* --------------------------------------------------------------- */

    name  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    type  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    valu  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg1  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg2  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg3  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg4  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg5  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg6  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg7  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg8  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg9  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    entry = (char *) malloc(MAX_STR_LEN*sizeof(char));
    temp  = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    temp2 = (char *) malloc(MAX_EXPR_LEN*sizeof(char));

    if (name  == NULL || type == NULL || valu  == NULL ||
        arg1  == NULL || arg2 == NULL || arg3  == NULL ||
        arg4  == NULL || arg5 == NULL || arg6  == NULL ||
        arg7  == NULL || arg8 == NULL || arg9  == NULL ||
        entry == NULL || temp == NULL || temp2 == NULL   ) {
        SPRINT0(0, "ERROR:: error mallocing arrays in processMessage");
        exit(EXIT_FAILURE);
    }

    SPRINT1(1, "==> processMessage(text=%s)", text);

    /* initialize the response */
    response_len = 0;
    response[0] = '\0';

    /* NO-OP */
    if (strlen(text) == 0) {

    /* "identify|" */
    } else if (strncmp(text, "identify|", 9) == 0) {

        /* build the response */
        snprintf(response, max_resp_len, "identify|serveCSM|");
        response_len = strlen(response);

    /* "getPmtrs|" */
    } else if (strncmp(text, "getPmtrs|", 9) == 0) {

        icount = 0;

        /* build the response in JSON format */
        snprintf(response, max_resp_len, "getPmtrs|[");
        response_len = strlen(response);

        /* external Parameters first */
        for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
            if (MODL->pmtr[ipmtr].type != OCSM_EXTERNAL) continue;

            snprintf(entry, MAX_STR_LEN, "{\"name\":\"%s\",\"type\":%d,\"nrow\":%d,\"ncol\":%d,\"value\":[",
                     MODL->pmtr[ipmtr].name,
                     MODL->pmtr[ipmtr].type,
                     MODL->pmtr[ipmtr].nrow,
                     MODL->pmtr[ipmtr].ncol);
            addToResponse(entry);

            index = 0;
            for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                    if (irow < MODL->pmtr[ipmtr].nrow || icol < MODL->pmtr[ipmtr].ncol) {
                        snprintf(entry, MAX_STR_LEN, "%lg,", MODL->pmtr[ipmtr].value[index++]);
                    } else {
                        snprintf(entry, MAX_STR_LEN, "%lg],\"dot\":[", MODL->pmtr[ipmtr].value[index++]);
                    }
                    addToResponse(entry);
                }
            }

            index = 0;
            for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                    if (irow < MODL->pmtr[ipmtr].nrow || icol < MODL->pmtr[ipmtr].ncol) {
                        snprintf(entry, MAX_STR_LEN, "%lg,", MODL->pmtr[ipmtr].dot[index++]);
                    } else {
                        snprintf(entry, MAX_STR_LEN, "%lg]", MODL->pmtr[ipmtr].dot[index++]);
                    }
                    addToResponse(entry);
                }
            }

            icount++;
            if (icount < MODL->npmtr) {
                addToResponse("},");
            } else {
                addToResponse("}]");
            }
        }

        /* internal Parameters last */
        for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
            if (MODL->pmtr[ipmtr].type != OCSM_INTERNAL) continue;

            snprintf(entry, MAX_STR_LEN, "{\"name\":\"%s\",\"type\":%d,\"nrow\":%d,\"ncol\":%d,\"value\":[",
                     MODL->pmtr[ipmtr].name,
                     MODL->pmtr[ipmtr].type,
                     MODL->pmtr[ipmtr].nrow,
                     MODL->pmtr[ipmtr].ncol);
            addToResponse(entry);

            index = 0;
            for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                    if (irow < MODL->pmtr[ipmtr].nrow || icol < MODL->pmtr[ipmtr].ncol) {
                        snprintf(entry, MAX_STR_LEN, "%lg,", MODL->pmtr[ipmtr].value[index++]);
                    } else {
                        snprintf(entry, MAX_STR_LEN, "%lg],\"dot\":[", MODL->pmtr[ipmtr].value[index++]);
                    }
                    addToResponse(entry);
                }
            }

            for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                    if (irow < MODL->pmtr[ipmtr].nrow || icol < MODL->pmtr[ipmtr].ncol) {
                        snprintf(entry, MAX_STR_LEN, "0,");
                    } else {
                        snprintf(entry, MAX_STR_LEN, "0]");
                    }
                    addToResponse(entry);
                }
            }

            icount++;
            if (icount < MODL->npmtr) {
                addToResponse("},");
            } else {
                addToResponse("}]");
            }
        }

    /* "newPmtr|name|nrow|ncol| value1| ..." */
    } else if (strncmp(text, "newPmtr|", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        nrow = 0;
        ncol = 0;

        if (getToken(text,  1, '|', name) == 0) name[0] = '\0';
        if (getToken(text,  2, '|', arg1)     ) nrow = strtol(arg1, &pEnd, 10);
        if (getToken(text,  3, '|', arg2)     ) ncol = strtol(arg2, &pEnd, 10);

        /* store an undo snapshot */
        status = storeUndo("newPmtr", name);
        if (status != SUCCESS) {
            SPRINT1(0, "storeUndo -> status=%d", status);
        }

        /* build the response */
        status = ocsmNewPmtr(MODL, name, OCSM_EXTERNAL, nrow, ncol);

        if (status == SUCCESS) {
            ipmtr = MODL->npmtr;

            i = 4;
            for (irow = 1; irow <= nrow; irow++) {
                for (icol = 1; icol <= ncol; icol++) {
                    if (getToken(text, i, '|', arg3)) {
                        (void) ocsmSetValu(MODL, ipmtr, irow, icol, arg3);
                    }

                    i++;
                }
            }

            snprintf(response, max_resp_len, "newPmtr|");
        } else {
            snprintf(response, max_resp_len, "ERROR:: newPmtr() detected: %s",
                     ocsmGetText(status));
        }
        response_len = strlen(response);

    /* "setPmtr|pmtrname|irow|icol|value1| " */
    } else if (strncmp(text, "setPmtr|", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        ipmtr = 0;
        irow  = 0;
        icol  = 0;

        getToken(text, 1, '|', arg1);

        for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
            if (strcmp(arg1, MODL->pmtr[jpmtr].name) == 0) {
                ipmtr = jpmtr;
                break;
            }
        }

        if (ipmtr > 0) {
            if (getToken(text, 2, '|', arg2)) irow  = strtol(arg2, &pEnd, 10);
            if (getToken(text, 3, '|', arg3)) icol  = strtol(arg3, &pEnd, 10);

            /* store an undo snapshot */
            status = storeUndo("setPmtr", MODL->pmtr[ipmtr].name);
            if (status != SUCCESS) {
                SPRINT1(0, "storeUndo -> status=%d", status);
            }

            if (getToken(text, 4, '|', arg4)) {
                status = ocsmSetValu(MODL, ipmtr, irow, icol, arg4);
                if (status != SUCCESS) {
                    SPRINT1(0, "ocsmSetValu -> status=%d", status);
                }
            } else {
                status = -999;
            }

            /* build the response */
            if (status == SUCCESS) {
                snprintf(response, max_resp_len, "setPmtr|");
            } else {
                snprintf(response, max_resp_len, "ERROR:: setPmtr(%d, %d, %d) detected: %s",
                         ipmtr, irow, icol, ocsmGetText(status));
            }
        } else {
            snprintf(response, max_resp_len, "ERROR:: setPmtr could not find name \"%s\"", arg1);
        }
        response_len = strlen(response);

    /* "clrVels|" */
    } else if (strncmp(text, "clrVels|", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        status = ocsmSetVelD(MODL, 0, 0, 0, 0.0);
        if (status != SUCCESS) {
            SPRINT1(0, "ocsmSetVelD -> status=%d", status);
        }

        /* store an undo snapshot */
        status = storeUndo("clrVels", "");
        if (status != SUCCESS) {
            SPRINT1(0, "storeUndo -> status=%d", status);
        }

        /* build the response */
        if (status == SUCCESS) {
            snprintf(response, max_resp_len, "clrVels|");
        } else {
            snprintf(response, max_resp_len, "ERROR:: clrVels() detected: %s",
                     ocsmGetText(status));
        }
        response_len = strlen(response);

    /* "setVel|pmtrname|irow|icol|vel| " */
    } else if (strncmp(text, "setVel|", 7) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        ipmtr = 0;
        irow  = 0;
        icol  = 0;

        getToken(text, 1, '|', arg1);

        for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
            if (strcmp(arg1, MODL->pmtr[jpmtr].name) == 0) {
                ipmtr = jpmtr;
                break;
            }
        }

        if (ipmtr > 0) {
            if (getToken(text, 2, '|', arg2)) irow  = strtol(arg2, &pEnd, 10);
            if (getToken(text, 3, '|', arg3)) icol  = strtol(arg3, &pEnd, 10);

            if (getToken(text, 4, '|', arg4)) {
                status = ocsmSetVel(MODL, ipmtr, irow, icol, arg4);
                if (status != SUCCESS) {
                    SPRINT1(0, "ocsmSetVel -> status=%d", status);
                }
            }

            /* store an undo snapshot */
            status = storeUndo("setVel", MODL->pmtr[ipmtr].name);
            if (status != SUCCESS) {
                SPRINT1(0, "storeUndo -> status=%d", status);
            }

            /* build the response */
            if (status == SUCCESS) {
                snprintf(response, max_resp_len, "setVel|");
            } else {
                snprintf(response, max_resp_len, "ERROR:: setVel(%d, %d, %d) detected: %s",
                         ipmtr, irow, icol, ocsmGetText(status));
            }
        } else {
            snprintf(response, max_resp_len, "ERROR:: setVel could not find name \"%s\"", arg1);
        }
        response_len = strlen(response);

    /* "getBrchs|" */
    } else if (strncmp(text, "getBrchs|", 9) == 0) {

        /* build the response in JSON format */
        snprintf(response, max_resp_len, "getBrchs|[");
        response_len = strlen(response);

        for (ibrch = 1; ibrch <= MODL->nbrch; ibrch++) {
            snprintf(entry, MAX_STR_LEN, "{\"name\":\"%s\",\"type\":\"%s\",\"actv\":%d,\"indent\":%d,\"attrs\":[",
                     MODL->brch[ibrch].name,
         ocsmGetText(MODL->brch[ibrch].type),
                     MODL->brch[ibrch].actv,
                     MODL->brch[ibrch].indent);
            addToResponse(entry);

            for (iattr = 0; iattr < MODL->brch[ibrch].nattr; iattr++) {
                if (iattr < MODL->brch[ibrch].nattr-1) {
                    snprintf(entry, MAX_STR_LEN, "[\"%s\",\"%s\"],",
                             MODL->brch[ibrch].attr[iattr].name,
                             MODL->brch[ibrch].attr[iattr].value);
                } else {
                    snprintf(entry, MAX_STR_LEN, "[\"%s\",\"%s\"]",
                             MODL->brch[ibrch].attr[iattr].name,
                             MODL->brch[ibrch].attr[iattr].value);
                }
                addToResponse(entry);
            }

            snprintf(entry, MAX_STR_LEN, "],\"ileft\":%d,\"irite\":%d,\"ichld\":%d,\"args\":[",
                     MODL->brch[ibrch].ileft,
                     MODL->brch[ibrch].irite,
                     MODL->brch[ibrch].ichld);
            addToResponse(entry);

            if (MODL->brch[ibrch].narg >= 1) {
                snprintf(entry, MAX_STR_LEN, "\"%s\"", MODL->brch[ibrch].arg1);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 2) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg2);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 3) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg3);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 4) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg4);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 5) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg5);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 6) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg6);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 7) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg7);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 8) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg8);
                addToResponse(entry);
            }
            if (MODL->brch[ibrch].narg >= 9) {
                snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg9);
                addToResponse(entry);
            }

            if (ibrch < MODL->nbrch) {
                snprintf(entry, MAX_STR_LEN, "]},");
            } else {
                snprintf(entry, MAX_STR_LEN, "]}]");
            }
            addToResponse(entry);
        }

    /* "newBrch|ibrch|type|arg1|arg2|arg3|arg4|arg5|arg6|arg7|arg8|arg9| aname1|avalu1|" */
    } else if (strncmp(text, "newBrch|", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        ibrch = 0;
        itype = 0;

        if (getToken(text,  1, '|', arg1)) ibrch = strtol(arg1, &pEnd, 10);

        if (getToken(text,  2, '|', type) >  0) itype = ocsmGetCode(type);
        if (getToken(text,  3, '|', arg1) == 0) arg1[0] = '\0';
        if (getToken(text,  4, '|', arg2) == 0) arg2[0] = '\0';
        if (getToken(text,  5, '|', arg3) == 0) arg3[0] = '\0';
        if (getToken(text,  6, '|', arg4) == 0) arg4[0] = '\0';
        if (getToken(text,  7, '|', arg5) == 0) arg5[0] = '\0';
        if (getToken(text,  8, '|', arg6) == 0) arg6[0] = '\0';
        if (getToken(text,  9, '|', arg7) == 0) arg7[0] = '\0';
        if (getToken(text, 10, '|', arg8) == 0) arg8[0] = '\0';
        if (getToken(text, 11, '|', arg9) == 0) arg9[0] = '\0';

        /* store an undo snapshot */
        status = storeUndo("newBrch", type);
        if (status != SUCCESS) {
            SPRINT1(0, "storeUndo -> status=%d", status);
        }

        /* build the response */
        status = ocsmNewBrch(MODL, ibrch, itype, arg1, arg2, arg3,
                                                 arg4, arg5, arg6,
                                                 arg7, arg8, arg9);
        if (status == SUCCESS) {
            status = ocsmCheck(MODL);

            if (status == SUCCESS) {
                snprintf(response, max_resp_len, "newBrch|");
            } else {
                snprintf(response, max_resp_len, "ERROR:: newBrch(ibrch=%d) detected: %s",
                         ibrch, ocsmGetText(status));
            }
        } else {
            snprintf(response, max_resp_len, "ERROR:: newBrch(ibrch=%d) detected: %s",
                     ibrch, ocsmGetText(status));
        }
        response_len = strlen(response);


    /* "setBrch|ibrch|name|actv|arg1|arg2|arg3|arg4|arg5|arg6|arg7|arg8|arg9| aname1|avalu1| ..." */
    } else if (strncmp(text, "setBrch|", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        ibrch = 0;
        actv  = 0;

        if (getToken(text, 1, '|', arg1)) ibrch = strtol(arg1, &pEnd, 10);

        /* store an undo snapshot */
        status = storeUndo("setBrch", MODL->brch[ibrch].name);
        if (status != SUCCESS) {
            SPRINT1(0, "storeUndo -> status=%d", status);
        }

        /* build the response */
        if (ibrch >= 1 && ibrch <= MODL->nbrch) {
            if (getToken(text,  2, '|', name)) (void) ocsmSetName(MODL, ibrch, name);

            if (getToken(text,  3, '|', arg1)) {
                if (strcmp(arg1, "suppressed") == 0) {
                    (void) ocsmSetBrch(MODL, ibrch, OCSM_SUPPRESSED);
                    actv = 1;
                } else {
                    (void) ocsmSetBrch(MODL, ibrch, OCSM_ACTIVE);
                    actv = 1;
                }
            }

            if (getToken(text,  4, '|', arg1)) (void) ocsmSetArg(MODL, ibrch, 1, arg1);
            if (getToken(text,  5, '|', arg2)) (void) ocsmSetArg(MODL, ibrch, 2, arg2);
            if (getToken(text,  6, '|', arg3)) (void) ocsmSetArg(MODL, ibrch, 3, arg3);
            if (getToken(text,  7, '|', arg4)) (void) ocsmSetArg(MODL, ibrch, 4, arg4);
            if (getToken(text,  8, '|', arg5)) (void) ocsmSetArg(MODL, ibrch, 5, arg5);
            if (getToken(text,  9, '|', arg6)) (void) ocsmSetArg(MODL, ibrch, 6, arg6);
            if (getToken(text, 10, '|', arg7)) (void) ocsmSetArg(MODL, ibrch, 7, arg7);
            if (getToken(text, 11, '|', arg8)) (void) ocsmSetArg(MODL, ibrch, 8, arg8);
            if (getToken(text, 12, '|', arg9)) (void) ocsmSetArg(MODL, ibrch, 9, arg9);

            i = 13;
            while (1) {
                if (getToken(text, i++, '|', name) == 0) break;
                if (getToken(text, i++, '|', valu) == 0) break;

                (void) ocsmSetAttr(MODL, ibrch, name, valu);
            }

            if (actv > 0) {
                status = ocsmCheck(MODL);

                if (status >= SUCCESS) {
                    snprintf(response, max_resp_len, "setBrch|");
                } else {
                    snprintf(response, max_resp_len, "ERROR:: setBrch(ibrch=%d) detected: %s",
                             ibrch, ocsmGetText(status));
                }
            } else {
                snprintf(response, max_resp_len, "setBrch|");
            }
        } else {
            status = OCSM_ILLEGAL_BRCH_INDEX;
            snprintf(response, max_resp_len, "ERROR: setBrch(%d) detected: %s",
                     ibrch, ocsmGetText(status));
        }
        response_len = strlen(response);


    /* "delBrch|ibrch|" */
    } else if (strncmp(text, "delBrch|", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract argument */
        ibrch = 0;

        if (getToken(text, 1, '|', arg1)) ibrch = strtol(arg1, &pEnd, 10);

        /* store an undo snapshot */
        status = storeUndo("delBrch", MODL->brch[ibrch].name);
        if (status != SUCCESS) {
            SPRINT1(0, "storeUndo -> status=%d", status);
        }

        /* delete the Branch */
        status = ocsmDelBrch(MODL, ibrch);

        /* check that the Branches are properly ordered */
        if (status == SUCCESS) {
            status = ocsmCheck(MODL);
        }

        /* build the response */
        if (status == SUCCESS) {
            snprintf(response, max_resp_len, "delBrch|");
        } else {
            snprintf(response, max_resp_len, "ERROR:: delBrch(%d) detected: %s",
                     ibrch, ocsmGetText(status));
        }
        response_len = strlen(response);

    /* "setAttr|ibrch|aname|avalue|" */
    } else if (strncmp(text, "setAttr|", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        ibrch = 0;

        if (getToken(text, 1, '|', arg1)) ibrch = strtol(arg1, &pEnd, 10);
        getToken(text, 2, '|', arg2);
        getToken(text, 3, '|', arg3);

        /* store an undo snapshot */
        status = storeUndo("setAttr", MODL->brch[ibrch].name);
        if (status != SUCCESS) {
            SPRINT1(0, "storeUndo -> status=%d", status);
        }

        /* set the Attribute */
        status = ocsmSetAttr(MODL, ibrch, arg2, arg3);

        /* build the response */
        if (status == SUCCESS) {
            snprintf(response, max_resp_len, "setAttr|");
        } else {
            snprintf(response, max_resp_len, "ERROR: setAttr(%d, %s, %s) detected: %s",
                     ibrch, arg2, arg3, ocsmGetText(status));
        }
        response_len = strlen(response);

    /* "undo|" */
    } else if (strncmp(text, "undo|", 5) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* build the response */
        if (nundo <= 0) {
            snprintf(response, max_resp_len, "ERROR:: there is nothing to undo");

        } else {
            /* remove the current MODL */
            status = ocsmFree(modl);
            if (status < SUCCESS) {
                snprintf(response, max_resp_len, "ERROR:: undo(%s) detected: %s",
                         arg1, ocsmGetText(status));
            } else {
                /* repoint MODL to the saved modl */
                modl = undo_modl[--nundo];
                snprintf(response, max_resp_len, "undo|%s|", undo_text[nundo]);
            }
        }
        response_len = strlen(response);

    /* "save|filename" */
    } else if (strncmp(text, "save|", 5) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract argument */
        getToken(text, 1, '|', filename);

        /* save the file */
        status = ocsmSave(MODL, filename);

        /* build the response */
        if (status == SUCCESS) {
            snprintf(response, max_resp_len, "save|");
        } else {
            snprintf(response, max_resp_len, "ERROR:: save(%s) detected: %s",
                     arg1, ocsmGetText(status));
        }
        response_len = strlen(response);

    /* "getCsmFile|" */
    } else if (strncmp(text, "getCsmFile|", 11) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* build the response */
        snprintf(response, max_resp_len, "getCsmFile|%s|", filename);
        response_len = strlen(response);

        /* add the input file, line by line */
        if (strlen(filename) > 0) {
            fp = fopen(filename, "r");

            while (1) {
                entry = fgets(entry, MAX_STR_LEN-1, fp);
                if (feof(fp) != 0) break;

                addToResponse(entry);
            }

            fclose(fp);
            fp = NULL;
        }

    /* "setCsmFileBeg|" */
    } else if (strncmp(text, "setCsmFileBeg|", 14) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* open the casefile and overwrite */
        fp = fopen(filename, "w");

        ichar = 14;
        while (text[ichar] != '\0') {
            fprintf(fp, "%c", text[ichar++]);
        }

    /* "setCsmFileMid|" */
    } else if (strncmp(text, "setCsmFileMid|", 14) == 0) {

        ichar = 14;
        while (text[ichar] != '\0') {
            fprintf(fp, "%c", text[ichar++]);
        }

    /* "setCsmFileEnd|" */
    } else if (strncmp(text, "setCsmFileEnd|", 14) == 0) {

        fclose(fp);
        fp = NULL;

        /* load the new case */
        status = ocsmFree(modl);
        if (status != SUCCESS) {
            SPRINT1(0, "ocsmFree -> status=%d", status);
        }

        status = ocsmFree(NULL);
        if (status != SUCCESS) {
            SPRINT1(0, "ocsmFree -> status=%d", status);
        }

        status = ocsmLoad(filename, &modl);

        if (status < 0) {
            snprintf(response, max_resp_len, "%s", MODL->errorBuf);
        } else {
            MODL = (modl_T*)modl;

            status = buildBodys(0, &builtTo, &buildStatus);

            if (builtTo < 0 || buildStatus < SUCCESS) {
                snprintf(response, max_resp_len, "build|%s|", MODL->errorBuf);
            } else if (status == SUCCESS) {
                snprintf(response, max_resp_len, "build|%d|%d|", builtTo, nbody);
            } else {
                snprintf(response, max_resp_len, "%s", MODL->errorBuf);
            }
        }

        response_len = strlen(response);

    /* "build|"  */
    } else if (strncmp(text, "build|", 6) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract argument */
        ibrch = 0;
        if (getToken(text, 1, '|', arg1)) ibrch = strtol(arg1, &pEnd, 10);

        /* build the response */
        status = buildBodys(ibrch, &builtTo, &buildStatus);

        if (builtTo < 0 || buildStatus < SUCCESS) {
            snprintf(response, max_resp_len, "build|%s|", MODL->errorBuf);
        } else if (status == SUCCESS) {
            snprintf(response, max_resp_len, "build|%d|%d|", builtTo, nbody);
        } else {
            snprintf(response, max_resp_len, "build|%s|", MODL->errorBuf);
        }
        response_len = strlen(response);

    /* "loadSketch|" */
    } else if (strncmp(text, "loadSketch|", 11) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract argument */
        ibrch = 0;
        if (getToken(text, 1, '|', arg1)) ibrch = strtol(arg1, &pEnd, 10);

        status = ocsmGetSketch(MODL, ibrch, begs, vars, cons, segs);

        if (status != SUCCESS) {
            snprintf(response, max_resp_len, "loadSketch|%s|", MODL->errorBuf);
            response_len = strlen(response);
        } else {
            snprintf(response, max_resp_len, "loadSketch|%s|%s|%s|%s|", begs, vars, cons, segs);
            response_len = strlen(response);
        }

    /* "solveSketch|" */
    } else if (strncmp(text, "solveSketch|", 12) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        getToken(text, 1, '|', arg1);
        getToken(text, 2, '|', arg2);

        /* solve the sketch (which always returns a response) */
        (void) ocsmSolveSketch(MODL, arg1, arg2, arg3);

        if (strlen(arg3) == 0) {
            snprintf(response, max_resp_len, "solveSketch|%s|", MODL->errorBuf);
        } else {
            snprintf(response, max_resp_len, "solveSketch|%s|", arg3);
        }
        response_len = strlen(response);

    /* "saveSketch|" */
    } else if (strncmp(text, "saveSketch|", 11) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        ibrch = 0;
        if (getToken(text, 1, '|', arg1)) ibrch = strtol(arg1, &pEnd, 10);

        getToken(text, 2, '|', arg2);
        getToken(text, 3, '|', arg3);
        getToken(text, 4, '|', arg4);

        /* save the sketch */
        status = ocsmSaveSketch(MODL, ibrch, arg2, arg3, arg4);

        if (status == SUCCESS) {
            snprintf(response, max_resp_len, "saveSketch|ok|");
            response_len = strlen(response);
        } else {
            snprintf(response, max_resp_len, "saveSketch|error|");
            response_len = strlen(response);
        }

    /* "setLims|lo|hi|" */
    } else if (strncmp(text, "setLims|", 8) == 0) {

        /* extract arguments */
        if (getToken(text, 1, '|', arg1)) lims[0] = strtod(arg1, &pEnd);
        if (getToken(text, 2, '|', arg2)) lims[1] = strtod(arg2, &pEnd);

        /* build the response */
        snprintf(response, max_resp_len, "setLims|");
        response_len = strlen(response);

        /* update the scene graph */
        if (batch == 0) {
            buildSceneGraph();
        }
    }

//cleanup:
    free(temp2);
    free(temp);
    free(entry);
    free(arg9);
    free(arg8);
    free(arg7);
    free(arg6);
    free(arg5);
    free(arg4);
    free(arg3);
    free(arg2);
    free(arg1);
    free(valu);
    free(type);
    free(name);
}


/***********************************************************************/
/*                                                                     */
/*   spec_col - return color for a given scalar value                  */
/*                                                                     */
/***********************************************************************/

static void
spec_col(float  scalar,
         float  lims[],
         float  color[])
{
    int   indx;
    float frac;

    if (lims[0] == lims[1]) {
        color[0] = 0.0;
        color[1] = 1.0;
        color[2] = 0.0;
    } else if (scalar <= lims[0]) {
        color[0] = color_map[0];
        color[1] = color_map[1];
        color[2] = color_map[2];
    } else if (scalar >= lims[1]) {
        color[0] = color_map[3*255  ];
        color[1] = color_map[3*255+1];
        color[2] = color_map[3*255+2];
    } else {
        frac  = 255.0 * (scalar - lims[0]) / (lims[1] - lims[0]);
        if (frac < 0  ) frac = 0;
        if (frac > 255) frac = 255;
        indx  = frac;
        frac -= indx;
        if (indx == 255) {
            indx--;
            frac += 1.0;
        }

        color[0] = frac * color_map[3*(indx+1)  ] + (1.0-frac) * color_map[3*indx  ];
        color[1] = frac * color_map[3*(indx+1)+1] + (1.0-frac) * color_map[3*indx+1];
        color[2] = frac * color_map[3*(indx-1)+2] + (1.0-frac) * color_map[3*indx+2];
    }
}


/***********************************************************************/
/*                                                                     */
/*   storeUndo - store an undo for the current command                 */
/*                                                                     */
/***********************************************************************/

static int
storeUndo(char   *cmd,                  /* (in)  current command */
          char   *arg)                  /* (in)  current argument */
{
    int       status = SUCCESS;         /* return status */

    int       iundo;
//$$$    char      text[MAX_EXPR_LEN];
    char      *text;

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(storeUndo);

    /* --------------------------------------------------------------- */

    MALLOC(text, char, MAX_EXPR_LEN);

    /* if the undos are full, discard the most ancient one */
    if (nundo >= MAX_UNDOS) {
        status = ocsmFree(undo_modl[0]);
        CHECK_STATUS(ocsmFree);

        for (iundo = 0; iundo < nundo; iundo++) {
            undo_modl[iundo] = undo_modl[iundo+1];
            (void) STRNCPY(undo_text[iundo], undo_text[iundo+1], 32);
        }

        nundo--;
    }

    /* store an undo snapshot */
    snprintf(text, MAX_EXPR_LEN, "%s %s", cmd, arg);
    STRNCPY(undo_text[nundo], text, 31);

    status = ocsmCopy(MODL, &undo_modl[nundo]);
    CHECK_STATUS(ocsmCopy);

    nundo++;

    SPRINT2(1, "~~> ocsmCopy() -> status=%d  (nundo=%d)",
            status, nundo);

cleanup:
    FREE(text);

    return status;
}

