/*
 ************************************************************************
 *                                                                      *
 * udpParsec -- udp file to generate a parsec airfoil                   *
 *                                                                      *
 *            Written by John Dannenhoffer @ Syracuse University        *
 *            Patterned after code written by Bob Haimes  @ MIT         *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2011/2015  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#define NUMUDPARGS 4
#include "udpUtilities.h"

/* shorthands for accessing argument values and velocities */
#define YTE(     IUDP)    ((double *) (udps[IUDP].arg[0].val))[0]
#define POLY(    IUDP,I)  ((double *) (udps[IUDP].arg[1].val))[I]
#define PARAM(   IUDP,I)  ((double *) (udps[IUDP].arg[2].val))[I]
#define MEANLINE(IUDP)    ((int    *) (udps[IUDP].arg[3].val))[0]

/* data about possible arguments */
static char  *argNames[NUMUDPARGS] = {"yte",    "poly",   "param",  "meanline", };
static int    argTypes[NUMUDPARGS] = {ATTRREAL, ATTRREAL, ATTRREAL, ATTRINT,    };
static int    argIdefs[NUMUDPARGS] = {0,        0,        0,        0,          };
static double argDdefs[NUMUDPARGS] = {0.,       0.,       0.,       0.,         };

/* get utility routines: udpErrorStr, udpInitialize, udpReset, udpSet,
                         udpGet, udpVel, udpClean, udpMesh */
#include "udpUtilities.c"

/* suppport routine defined below */
static int matsol(double A[], double b[], int n, double x[]);

#define EPS12 1.0e-12


/*
 ************************************************************************
 *                                                                      *
 *   udpExecute - execute the primitive                                 *
 *                                                                      *
 ************************************************************************
 */

int
udpExecute(ego  context,                /* (in)  EGADS context */
           ego  *ebody,                 /* (out) Body pointer */
           int  *nMesh,                 /* (out) number of associated meshes */
           char *string[])              /* (out) error message */
{
    int     status = EGADS_SUCCESS;

    int     npoly, nparam, npts, i, j, ile, ite, ipoly;
    int     sense[3], sizes[2], periodic;
    double  zeta, xx, yy, *pts=NULL, *polyTop=NULL, *polyBot=NULL;
    double  lefac, Amat[25], Rhs[5], poly[5];
    double  rle, xtop, ytop, ctop, ttop, xbot, ybot, cbot, tbot;
    double  data[18], tle, tdata[2], result[3], range[4], eval[18], norm[3];
    double  dxytol = 1.0e-6;
    ego     enodes[4], eedges[3], ecurve, eloop, eface, enew;

    /* default return values */
    *ebody  = NULL;
    *nMesh  = 0;
    *string = NULL;

    /* check arguments */
    npoly  = udps[0].arg[1].size;
    nparam = udps[0].arg[2].size;

    if (udps[0].arg[0].size > 1) {
        printf(" udpExecute: yte should be a scalar\n");
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if (npoly > 1 && nparam > 1) {
        printf(" udpExecute: poly and param cannot both be set\n");
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if (nparam > 1 && nparam != 9) {
        printf(" udpExecute: there should be 9 elements in Param\n");
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if (npoly > 1 && npoly%2 != 0) {
        printf(" udpExecute: there should be an even number of elements in Poly\n");
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if (npoly <= 1 && nparam <= 1) {
        printf(" udpExecute: neither poly nor param was set\n");
        status  = EGADS_NODATA;
        *string = udpErrorStr(status);
        goto cleanup;

    }

    /* cache copy of arguments for future use */
    status = cacheUdp();
    if (status < 0) {
        printf(" udpExecute: problem caching arguments\n");
        *string = udpErrorStr(status);
        goto cleanup;
    }

#ifdef DEBUG
    printf("yte(%d)      = %f\n", numUdp, YTE(     numUdp));
    printf("poly(%d)     = %f\n", numUdp, POLY(    numUdp,0));
    printf("param(%d)    = %f\n", numUdp, PARAM(   numUdp,0));
    printf("meanline(%d) = %d\n", numUdp, MEANLINE(numUdp));
#endif

    /* create the polynomial coefficients if Nparam > 0 */
    if (nparam > 1) {

        npoly = 6;

        /* create top polynomial */
        polyTop = (double*) malloc(npoly*sizeof(double));
        if (polyTop == NULL) {
            status = EGADS_MALLOC;
            goto cleanup;
        }

        rle  = PARAM(0,0);
        xtop = PARAM(0,1);
        ytop = PARAM(0,2);
        ctop = PARAM(0,3);
        ttop = PARAM(0,4);

        if (rle <= 0) {
            printf(" udpExecute: rle = %f <= 0\n", rle);
            status  = EGADS_RANGERR;
            *string = udpErrorStr(status);
            goto cleanup;
        } else if (xtop <= 0) {
            printf(" udpExecute: xtop = %f <= 0\n", xtop);
            *string = udpErrorStr(status);
            status  = EGADS_RANGERR;
            goto cleanup;
        }

        lefac = sqrt(2*rle);

        Amat[ 0] = 1;
        Amat[ 1] = 1;
        Amat[ 2] = 1;
        Amat[ 3] = 1;
        Amat[ 4] = 1;
        Rhs[  0] = YTE(0) - lefac;

        Amat[ 5] =         pow(xtop, 1.5);
        Amat[ 6] =         pow(xtop, 2.5);
        Amat[ 7] =         pow(xtop, 3.5);
        Amat[ 8] =         pow(xtop, 4.5);
        Amat[ 9] =         pow(xtop, 5.5);
        Rhs[  1] = ytop - lefac * sqrt(xtop);

        Amat[10] =  1.5;
        Amat[11] =  2.5;
        Amat[12] =  3.5;
        Amat[13] =  4.5;
        Amat[14] =  5.5;
        Rhs[  2] = tan(ttop*PI/180) - 0.5 * lefac;

        Amat[15] =  1.5  * pow(xtop, 0.5);
        Amat[16] =  2.5  * pow(xtop, 1.5);
        Amat[17] =  3.5  * pow(xtop, 2.5);
        Amat[18] =  4.5  * pow(xtop, 3.5);
        Amat[19] =  5.5  * pow(xtop, 4.5);
        Rhs[  3] =  -.5  * lefac / sqrt(xtop);

        Amat[20] =  0.75 * pow(xtop, -.5);
        Amat[21] =  3.75 * pow(xtop, 0.5);
        Amat[22] =  8.75 * pow(xtop, 1.5);
        Amat[23] = 15.75 * pow(xtop, 2.5);
        Amat[24] = 24.75 * pow(xtop, 3.5);
        Rhs[  4] = ctop + 0.25 * lefac / pow(xtop, 1.5);

        status = matsol(Amat, Rhs, 5, poly);
        if (status != EGADS_SUCCESS) {
            printf(" udpExecute: top matrix is singular\n");
            *string = udpErrorStr(status);
            goto cleanup;
        }

        polyTop[0] = lefac;
        polyTop[1] = poly[0];
        polyTop[2] = poly[1];
        polyTop[3] = poly[2];
        polyTop[4] = poly[3];
        polyTop[5] = poly[4];

        /* create bottom polynomial */
        polyBot = (double*) malloc(6*sizeof(double));
        if (polyBot == NULL) {
            status = EGADS_MALLOC;
            goto cleanup;
        }

        rle  = PARAM(0,0);
        xbot = PARAM(0,5);
        ybot = PARAM(0,6);
        cbot = PARAM(0,7);
        tbot = PARAM(0,8);

        if (rle <= 0) {
            printf(" udpExecute: rle = %f <= 0\n", rle);
            status = EGADS_RANGERR;
            goto cleanup;
        } else if (xbot <= 0) {
            printf(" udpExecute: xbot = %f <= 0\n", xbot);
            status = EGADS_RANGERR;
            goto cleanup;
        }

        lefac = -sqrt(2*rle);

        Amat[ 0] = 1;
        Amat[ 1] = 1;
        Amat[ 2] = 1;
        Amat[ 3] = 1;
        Amat[ 4] = 1;
        Rhs[  0] = YTE(0) - lefac;

        Amat[ 5] =         pow(xbot, 1.5);
        Amat[ 6] =         pow(xbot, 2.5);
        Amat[ 7] =         pow(xbot, 3.5);
        Amat[ 8] =         pow(xbot, 4.5);
        Amat[ 9] =         pow(xbot, 5.5);
        Rhs[  1] = ybot - lefac * sqrt(xbot);

        Amat[10] =  1.5;
        Amat[11] =  2.5;
        Amat[12] =  3.5;
        Amat[13] =  4.5;
        Amat[14] =  5.5;
        Rhs[  2] = tan(tbot*PI/180) - 0.5 * lefac;

        Amat[15] =  1.5  * pow(xbot, 0.5);
        Amat[16] =  2.5  * pow(xbot, 1.5);
        Amat[17] =  3.5  * pow(xbot, 2.5);
        Amat[18] =  4.5  * pow(xbot, 3.5);
        Amat[19] =  5.5  * pow(xbot, 4.5);
        Rhs[  3] =  -.5  * lefac / sqrt(xbot);

        Amat[20] =  0.75 * pow(xbot, -.5);
        Amat[21] =  3.75 * pow(xbot, 0.5);
        Amat[22] =  8.75 * pow(xbot, 1.5);
        Amat[23] = 15.75 * pow(xbot, 2.5);
        Amat[24] = 24.75 * pow(xbot, 3.5);
        Rhs[  4] = cbot + 0.25 * lefac / pow(xbot, 1.5);

        status = matsol(Amat, Rhs, 5, poly);
        if (status != EGADS_SUCCESS) {
            printf(" udpExecute: bot matrix is singular\n");
            *string = udpErrorStr(status);
            goto cleanup;
        }

        polyBot[0] = lefac;
        polyBot[1] = poly[0];
        polyBot[2] = poly[1];
        polyBot[3] = poly[2];
        polyBot[4] = poly[3];
        polyBot[5] = poly[4];

    /* use given polynomial coefficients because Npoly > 0 */
    } else {

        npoly = 1 + npoly / 2;

        polyTop = (double*) malloc((npoly+1)*sizeof(double));
        if (polyTop == NULL) {
            status  = EGADS_MALLOC;
            *string = udpErrorStr(status);
            goto cleanup;
        }

        polyBot = (double*) malloc((npoly+1)*sizeof(double));
        if (polyBot == NULL) {
            status  = EGADS_MALLOC;
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* copy polynomial coefficients and close airfoil */
        polyTop[npoly-1] = YTE(0);
        for (ipoly = 0; ipoly < npoly-1; ipoly++) {
            polyTop[ipoly  ] =  POLY(0,ipoly);
            polyTop[npoly-1] -= POLY(0,ipoly);
        }

        polyBot[npoly-1] = YTE(0);
        for (ipoly = 0; ipoly < npoly-1; ipoly++) {
            polyBot[ipoly  ] =  POLY(0,ipoly+npoly-1);
            polyBot[npoly-1] -= POLY(0,ipoly+npoly-1);
        }
    }

//$$$    Poly[numUdp] = (double*) EG_alloc((Npoly[0]+1)*sizeof(double));
//$$$    if (Poly[numUdp] == NULL) {
//$$$        status = EGADS_MALLOC;
//$$$        goto cleanup;
//$$$    }
//$$$
//$$$    Param[numUdp] = (double*) EG_alloc((nparam+1)*sizeof(double));
//$$$    if (Param[numUdp] == NULL) {
//$$$        status = EGADS_MALLOC;
//$$$        goto cleanup;
//$$$    }
//$$$
//$$$    /* copy in the data */
//$$$    Yte[     numUdp] = Yte[     0];
//$$$    Npoly[   numUdp] = Npoly [  0];
//$$$    Nparam[  numUdp] = Nparam[  0];
//$$$    Meanline[numUdp] = Meanline[0];
//$$$
//$$$    for (ipoly = 0; ipoly < Npoly[0]; ipoly++) {
//$$$        Poly[numUdp][ipoly] = Poly[0][ipoly];
//$$$    }
//$$$
//$$$    for (iparam = 0; iparam < nparam; iparam++) {
//$$$        Param[numUdp][iparam] = Param[0][iparam];
//$$$    }

    /* mallocs required by Windows compiler */
    npts = 101;
    pts = (double*)malloc((3*npts)*sizeof(double));
    if (pts == NULL) {
        status  = EGADS_MALLOC;
        *string = udpErrorStr(status);
        goto cleanup;
    }

    /* points around airfoil (upper and then lower) */
    for (i = 0; i < npts; i++) {
        zeta = TWOPI * i / (npts-1);
        xx   = (1 + cos(zeta)) / 2;

        if (i < npts/2) {
            yy = 0;
            for (ipoly = 0; ipoly < npoly; ipoly++) {
                yy += polyTop[ipoly] * pow(xx, ipoly+0.5);
            }

            pts[3*i  ] = xx;
            pts[3*i+1] = yy;
            pts[3*i+2] = 0;
        } else if (i == npts/2) {
            pts[3*i  ] = 0;
            pts[3*i+1] = 0;
            pts[3*i+2] = 0;
        } else {
            yy = 0;
            for (ipoly = 0; ipoly < npoly; ipoly++) {
                yy += polyBot[ipoly] * pow(xx, ipoly+0.5);
            }

            pts[3*i  ] = xx;
            pts[3*i+1] = yy;
            pts[3*i+2] = 0;
        }
    }

    /* if meanline[0] is zero, then create a FaceBody (the default) */
    if (MEANLINE(0) == 0) {

        /* create Node at upper trailing edge */
        i = 0;
        data[0] = pts[3*i  ];
        data[1] = pts[3*i+1];
        data[2] = pts[3*i+2];
        status = EG_makeTopology(context, NULL, NODE, 0,
                                 data, 0, NULL, NULL, &(enodes[0]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create Node at leading edge */
        i = (npts - 1) / 2;
        data[0] = pts[3*i  ];
        data[1] = pts[3*i+1];
        data[2] = pts[3*i+2];
        status = EG_makeTopology(context, NULL, NODE, 0,
                                 data, 0, NULL, NULL, &(enodes[1]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create Node at lower trailing edge */
        i = npts - 1;
        data[0] = pts[3*i  ];
        data[1] = pts[3*i+1];
        data[2] = pts[3*i+2];
        status = EG_makeTopology(context, NULL, NODE, 0,
                                 data, 0, NULL, NULL, &(enodes[2]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create spline curve from upper TE, to LE, to lower TE */
        sizes[0] = npts;
        sizes[1] = 0;
        status = EG_approximate(context, 0, dxytol, sizes, pts, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* find the t for the leading edge */
        i = (npts - 1) / 2;
        data[0] = pts[3*i  ];
        data[1] = pts[3*i+1];
        data[2] = pts[3*i+2];
        status = EG_invEvaluate(ecurve, data, &tle, result);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make Edge for upper surface */
        tdata[0] = 0;
        tdata[1] = tle;

        status = EG_makeTopology(context, ecurve, EDGE, TWONODE,
                                 tdata, 2, &(enodes[0]), NULL, &(eedges[0]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make Edge for lower surface */
        tdata[0] = tle;
        tdata[1] = 1;

        status = EG_makeTopology(context, ecurve, EDGE, TWONODE,
                                 tdata, 2, &(enodes[1]), NULL, &(eedges[1]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create loop of the two Edges */
        sense[0] = SFORWARD;
        sense[1] = SFORWARD;

        status = EG_makeTopology(context, NULL, LOOP, CLOSED,
                                 NULL, 2, eedges, sense, &eloop);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make Face from the loop */
        status = EG_makeFace(eloop, SFORWARD, NULL, &eface);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* find the direction of the Face normal */
        status = EG_getRange(eface, range, &periodic);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
        goto cleanup;
        }

        range[0] = (range[0] + range[1]) / 2;
        range[1] = (range[2] + range[3]) / 2;

        status = EG_evaluate(eface, range, eval);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        norm[0] = eval[4] * eval[8] - eval[5] * eval[7];
        norm[1] = eval[5] * eval[6] - eval[3] * eval[8];
        norm[2] = eval[3] * eval[7] - eval[4] * eval[6];

        /* if the normal is not positive, flip the Face */
        if (norm[2] < 0) {
            status = EG_flipObject(eface, &enew);
            if (status != EGADS_SUCCESS) {
                *string = udpErrorStr(status);
                goto cleanup;
            }
            eface = enew;
        }

        /* create the FaceBody (which will be returned) */
        status = EG_makeTopology(context, NULL, BODY, FACEBODY,
                                 NULL, 1, &eface, sense, ebody);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

    /* otherwise create a WireBody of the meanline */
    } else {

        /* collapse the upper and lower surfaces into the meanline
              (over-writing  the old lower surface) */
        ile = (npts - 1) / 2;
        ite =  npts - 1;
        for (i = ile+1, j = ile-1; i < ite; i++, j--) {
            pts[3*i  ] = (pts[3*i  ] + pts[3*j  ]) / 2;
            pts[3*i+1] = (pts[3*i+1] + pts[3*j+1]) / 2;
            pts[3*i+2] = (pts[3*i+2] + pts[3*j+2]) / 2;
        }

        /* create Node at leading edge */
        data[0] = pts[3*ile  ];
        data[1] = pts[3*ile+1];
        data[2] = pts[3*ile+2];
        status = EG_makeTopology(context, NULL, NODE, 0,
                                 data, 0, NULL, NULL, &(enodes[0]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create Node at trailing edge */
        data[0] = pts[3*ite  ];
        data[1] = pts[3*ite+1];
        data[2] = pts[3*ite+2];
        status = EG_makeTopology(context, NULL, NODE, 0,
                                 data, 0, NULL, NULL, &(enodes[1]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create spline curve from LE to TE */
        sizes[0] = ile + 1;
        sizes[1] = 0;
        status = EG_approximate(context, 0, dxytol, sizes, &(pts[3*ile]), &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make Edge for camberline */
        data[0] = pts[3*ile  ];
        data[1] = pts[3*ile+1];
        data[2] = pts[3*ile+2];
        status = EG_invEvaluate(ecurve, data, &(tdata[0]), result);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        data[0] = pts[3*ite  ];
        data[1] = pts[3*ite+1];
        data[2] = pts[3*ite+2];
        status = EG_invEvaluate(ecurve, data, &(tdata[1]), result);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_makeTopology(context, ecurve, EDGE, TWONODE,
                                 tdata, 2, &(enodes[0]), NULL, &(eedges[0]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create loop of the Edges */
        sense[0] = SFORWARD;

        status = EG_makeTopology(context, NULL, LOOP, OPEN,
                                 NULL, 1, eedges, sense, &eloop);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create the WireBody (which will be returned) */
        status = EG_makeTopology(context, NULL, BODY, WIREBODY,
                                 NULL, 1, &eloop, NULL, ebody);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }
    }

    /* set the output value(s) */

    /* remember this model (body) */
    udps[numUdp].ebody = *ebody;

cleanup:
    if (pts     != NULL) free(pts    );
    if (polyTop != NULL) free(polyTop);
    if (polyBot != NULL) free(polyBot);

    if (status != EGADS_SUCCESS) {
        *string = udpErrorStr(status);
    }

    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpSensitivity - return sensitivity derivatives for the "real" argument *
 *                                                                      *
 ************************************************************************
 */

int
udpSensitivity(ego    ebody,            /* (in)  Body pointer */
   /*@unused@*/int    npnt,             /* (in)  number of points */
   /*@unused@*/int    entType,          /* (in)  OCSM entity type */
   /*@unused@*/int    entIndex,         /* (in)  OCSM entity index (bias-1) */
   /*@unused@*/double uvs[],            /* (in)  parametric coordinates for evaluation */
   /*@unused@*/double vels[])           /* (out) velocities */
{
    int iudp, judp;

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == udps[judp].ebody) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    /* this routine is not written yet */
    return EGADS_NOLOAD;
}


/*
 ************************************************************************
 *                                                                      *
 *   matsol - Gaussian elimination with partial pivoting                *
 *                                                                      *
 ************************************************************************
 */

static int
matsol(double    A[],                   /* (in)  matrix to be solved (stored rowwise) */
                                        /* (out) upper-triangular form of matrix */
       double    b[],                   /* (in)  right hand side */
                                        /* (out) right-hand side after swapping */
       int       n,                     /* (in)  size of matrix */
       double    x[])                   /* (out) solution of A*x=b */
{
    int       status = EGADS_SUCCESS;   /* (out) return status */

    int       ir, jc, kc, imax;
    double    amax, swap, fact;

    /* --------------------------------------------------------------- */

    /* reduce each column of A */
    for (kc = 0; kc < n; kc++) {

        /* find pivot element */
        imax = kc;
        amax = fabs(A[kc*n+kc]);

        for (ir = kc+1; ir < n; ir++) {
            if (fabs(A[ir*n+kc]) > amax) {
                imax = ir;
                amax = fabs(A[ir*n+kc]);
            }
        }

        /* check for possibly-singular matrix (ie, near-zero pivot) */
        if (amax < EPS12) {
            status  = EGADS_DEGEN;
            goto cleanup;
        }

        /* if diagonal is not pivot, swap rows in A and b */
        if (imax != kc) {
            for (jc = 0; jc < n; jc++) {
                swap         = A[kc  *n+jc];
                A[kc  *n+jc] = A[imax*n+jc];
                A[imax*n+jc] = swap;
            }

            swap    = b[kc  ];
            b[kc  ] = b[imax];
            b[imax] = swap;
        }

        /* row-reduce part of matrix to the bottom of and right of [kc,kc] */
        for (ir = kc+1; ir < n; ir++) {
            fact = A[ir*n+kc] / A[kc*n+kc];

            for (jc = kc+1; jc < n; jc++) {
                A[ir*n+jc] -= fact * A[kc*n+jc];
            }

            b[ir] -= fact * b[kc];

            A[ir*n+kc] = 0;
        }
    }

    /* back-substitution pass */
    x[n-1] = b[n-1] / A[(n-1)*n+(n-1)];

    for (jc = n-2; jc >= 0; jc--) {
        x[jc] = b[jc];
        for (kc = jc+1; kc < n; kc++) {
            x[jc] -= A[jc*n+kc] * x[kc];
        }
        x[jc] /= A[jc*n+jc];
    }

cleanup:
    return status;
}
