/*
 ************************************************************************
 *                                                                      *
 * udpWaffle -- udp file to generate 2D waffle                          *
 *                                                                      *
 *            Written by John Dannenhoffer @ Syracuse University        *
 *            Patterned after code written by Bob Haimes  @ MIT         *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2011/2015  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#define NUMUDPARGS 2
#include "udpUtilities.h"

/* shorthands for accessing argument values and velocities */
#define DEPTH(   IUDP)    ((double *) (udps[IUDP].arg[0].val))[0]
#define SEGMENTS(IUDP,I)  ((double *) (udps[IUDP].arg[1].val))[I]

/* data about possible arguments */
static char  *argNames[NUMUDPARGS] = {"depth",  "segments", };
static int    argTypes[NUMUDPARGS] = {ATTRREAL, ATTRREAL,   };
static int    argIdefs[NUMUDPARGS] = {0,        0,          };
static double argDdefs[NUMUDPARGS] = {1.,       0.,         };

/* get utility routines: udpErrorStr, udpInitialize, udpReset, udpSet,
                         udpGet, udpVel, udpClean, udpMesh */
#include "udpUtilities.c"


/*
 ************************************************************************
 *                                                                      *
 *   udpExecute - execute the primitive                                 *
 *                                                                      *
 ************************************************************************
 */

int
udpExecute(ego  context,                /* (in)  EGADS context */
           ego  *ebody,                 /* (out) Body pointer */
           int  *nMesh,                 /* (out) number of associated meshes */
           char *string[])              /* (out) error message */
{
    int     status = EGADS_SUCCESS;

    int     inode, jnode, nnode, iseg, jseg, nseg, jedge, jface, senses[4];
    int     *ibeg=NULL, *iend=NULL;
    int     ia, ib, ic, id, MAXSIZE=1000;
    double  *X=NULL, *Y=NULL;
    double  xyz[20], data[6], xyz_out[20], D, s, t, xx, yy;

    double  EPS06 = 1.0e-6;
    ego     *enodes=NULL, *eedges=NULL, *efaces=NULL, ecurve, echild[4], eloop, eshell;

#ifdef DEBUG
    printf("udpExecute(context=%lx)\n", (long)context);
    printf("depth(0)    = %f\n", DEPTH(     0));
    printf("segments(0) = %f",   SEGMENTS(0,0));
    for (iseg = 1; iseg < udps[0].arg[1].size; iseg++) {
        printf(" %f", SEGMENTS(0,iseg));
    }
    printf("\n");
#endif

    /* default return values */
    *ebody  = NULL;
    *nMesh  = 0;
    *string = NULL;

    /* check arguments */
    nseg = udps[0].arg[1].size / 4;

    if (udps[0].arg[0].size > 1) {
        printf(" udpExecute: depth should be a scalar\n");
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if (DEPTH(0) <= 0) {
        printf(" udpExecute: depth = %f <= 0\n", DEPTH(0));
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if (udps[0].arg[1].size%4 != 0) {
        printf(" udpExecute: segments must be divisible by 4\n");
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;
    }

    /* cache copy of arguments for future use */
    status = cacheUdp();
    if (status < 0) {
        printf(" udpExecute: problem caching arguments\n");
        *string = udpErrorStr(status);
        goto cleanup;
    }

#ifdef DEBUG
    printf("depth(%d)    = %f\n", numUdp, DEPTH(   numUdp));
    printf("segments(%d) = %f",   numUdp, SEGMENTS(numUdp,0));
    for (iseg = 1; iseg < udps[0].arg[1].size; iseg++) {
        printf(" %f", SEGMENTS(numUdp,iseg));
    }
    printf("\n");
#endif

    /* make an indexed table of the segments */
    X    = (double *) EG_alloc(MAXSIZE*sizeof(double));  // probably too big
    Y    = (double *) EG_alloc(MAXSIZE*sizeof(double));
    ibeg = (int    *) EG_alloc(MAXSIZE*sizeof(int   ));
    iend = (int    *) EG_alloc(MAXSIZE*sizeof(int   ));

    if (X == NULL || Y == NULL || ibeg == NULL || iend == NULL) {
        status  = EGADS_MALLOC;
        *string = udpErrorStr(status);
        goto cleanup;
    }

    nnode = 0;

    for (iseg = 0; iseg < nseg; iseg++) {
        ibeg[iseg] = -1;
        for (jnode = 0; jnode < nnode; jnode++) {
            if (fabs(SEGMENTS(0,4*iseg  )-X[jnode]) < EPS06 &&
                fabs(SEGMENTS(0,4*iseg+1)-Y[jnode]) < EPS06   ) {
                ibeg[iseg] = jnode;
                break;
            }
        }

        if (ibeg[iseg] < 0) {
            ibeg[iseg] = nnode;

            X[nnode] = SEGMENTS(0,4*iseg  );
            Y[nnode] = SEGMENTS(0,4*iseg+1);
            nnode++;
        }

        iend[iseg] = -1;
        for (jnode = 0; jnode < nnode; jnode++) {
            if (fabs(SEGMENTS(0,4*iseg+2)-X[jnode]) < EPS06 &&
                fabs(SEGMENTS(0,4*iseg+3)-Y[jnode]) < EPS06   ) {
                iend[iseg] = jnode;
                break;
            }
        }

        if (iend[iseg] < 0) {
            iend[iseg] = nnode;

            X[nnode] = SEGMENTS(0,4*iseg+2);
            Y[nnode] = SEGMENTS(0,4*iseg+3);
            nnode++;
        }
    }

    /* check for intersections */
    for (jseg = 0; jseg < nseg; jseg++) {
        for (iseg = jseg+1; iseg < nseg; iseg++) {
            ia = ibeg[iseg];
            ib = iend[iseg];
            ic = ibeg[jseg];
            id = iend[jseg];

            D = (X[ib] - X[ia]) * (Y[ic] - Y[id]) - (X[ic] - X[id]) * (Y[ib] - Y[ia]);
            if (fabs(D) > EPS06) {
                s = ((X[ic] - X[ia]) * (Y[ic] - Y[id]) - (X[ic] - X[id]) * (Y[ic] - Y[ia])) / D;
                t = ((X[ib] - X[ia]) * (Y[ic] - Y[ia]) - (X[ic] - X[ia]) * (Y[ib] - Y[ia])) / D;

                if (s > -EPS06 && s < 1+EPS06 &&
                    t > -EPS06 && t < 1+EPS06   ) {
                    xx = (1 - s) * X[ia] + s * X[ib];
                    yy = (1 - s) * Y[ia] + s * Y[ib];

                    inode = -1;
                    for (jnode = 0; jnode < nnode; jnode++) {
                        if (fabs(xx-X[jnode]) < EPS06 &&
                            fabs(yy-Y[jnode]) < EPS06   ) {
                            inode = jnode;
                            break;
                        }
                    }

                    if (inode < 0) {
                        inode = nnode;

                        X[nnode] = xx;
                        Y[nnode] = yy;
                        nnode++;
                    }

                    if (ia != inode && ib != inode) {
                        ibeg[nseg] = inode;
                        iend[nseg] = ib;
                        nseg++;

                        iend[iseg] = inode;
                    }

                    if (ic != inode && id != inode) {
                        ibeg[nseg] = inode;
                        iend[nseg] = id;
                        nseg++;

                        iend[jseg] = inode;
                    }
                }
            }
        }
    }

    /* check for degenerate segments */
    for (jseg = 0; jseg < nseg; jseg++) {
        if (ibeg[jseg] == iend[jseg]) {
            printf(" udpExecute: segment %d is degenerate\n", iseg);
            status = EGADS_DEGEN;
            goto cleanup;
        }
    }

    /* allocate space for the egos */
    enodes = (ego *) EG_alloc((2*nnode       )*sizeof(ego));
    eedges = (ego *) EG_alloc((  nnode+2*nseg)*sizeof(ego));
    efaces = (ego *) EG_alloc((          nseg)*sizeof(ego));

    if (enodes == NULL || eedges == NULL || efaces == NULL) {
        status = EGADS_MALLOC;
        goto cleanup;
    }

    /* set up enodes at Z=0 and at Z=Depth */
    jnode = 0;

    for (inode = 0; inode < nnode; inode++) {
        xyz[0] = X[inode];
        xyz[1] = Y[inode];
        xyz[2] = 0;

        status = EG_makeTopology(context, NULL, NODE, 0,
                                 xyz, 0, NULL, NULL, &(enodes[jnode]));
        if (status != EGADS_SUCCESS) goto cleanup;

        jnode++;
    }

    for (inode = 0; inode < nnode; inode++) {
        xyz[0] = X[inode];
        xyz[1] = Y[inode];
        xyz[2] = DEPTH(0);

        status = EG_makeTopology(context, NULL, NODE, 0, xyz, 0,
                                 NULL, NULL, &(enodes[jnode]));
        if (status != EGADS_SUCCESS) goto cleanup;

        jnode++;
    }

    /* set up the edges on the z=0 plane */
    jedge = 0;

    for (iseg = 0; iseg < nseg; iseg++) {
        xyz[0] = X[ibeg[iseg]];
        xyz[1] = Y[ibeg[iseg]];
        xyz[2] = 0;
        xyz[3] = X[iend[iseg]] - X[ibeg[iseg]];
        xyz[4] = Y[iend[iseg]] - Y[ibeg[iseg]];
        xyz[5] = 0;

        status = EG_makeGeometry(context, CURVE, LINE, NULL,
                                 NULL, xyz, &ecurve);
        if (status != EGADS_SUCCESS) goto cleanup;

        echild[0] = enodes[ibeg[iseg]];
        echild[1] = enodes[iend[iseg]];

        status = EG_invEvaluate(ecurve, xyz, &(data[0]), xyz_out);
        if (status != EGADS_SUCCESS) goto cleanup;

        xyz[0] = X[iend[iseg]];
        xyz[1] = Y[iend[iseg]];
        xyz[2] = 0;

        status = EG_invEvaluate(ecurve, xyz, &(data[1]), xyz_out);
        if (status != EGADS_SUCCESS) goto cleanup;

        status = EG_makeTopology(context, ecurve, EDGE, TWONODE,
                                 data, 2, echild, NULL, &(eedges[jedge]));
        if (status != EGADS_SUCCESS) goto cleanup;

        jedge++;
    }

    /* set up the edges on the z=Depth plane */
    for (iseg = 0; iseg < nseg; iseg++) {
        xyz[0] = X[ibeg[iseg]];
        xyz[1] = Y[ibeg[iseg]];
        xyz[2] = DEPTH(0);
        xyz[3] = X[iend[iseg]] - X[ibeg[iseg]];
        xyz[4] = Y[iend[iseg]] - Y[ibeg[iseg]];
        xyz[5] = 0;

        status = EG_makeGeometry(context, CURVE, LINE, NULL,
                                 NULL, xyz, &ecurve);
        if (status != EGADS_SUCCESS) goto cleanup;

        echild[0] = enodes[ibeg[iseg]+nnode];
        echild[1] = enodes[iend[iseg]+nnode];

        status = EG_invEvaluate(ecurve, xyz, &(data[0]), xyz_out);
        if (status != EGADS_SUCCESS) goto cleanup;

        xyz[0] = X[iend[iseg]];
        xyz[1] = Y[iend[iseg]];
        xyz[2] = DEPTH(0);

        status = EG_invEvaluate(ecurve, xyz, &(data[1]), xyz_out);
        if (status != EGADS_SUCCESS) goto cleanup;

        status = EG_makeTopology(context, ecurve, EDGE, TWONODE,
                                 data, 2, echild, NULL, &(eedges[jedge]));
        if (status != EGADS_SUCCESS) goto cleanup;

        jedge++;
    }

    /* set up the edges between z=0 and Depth */
    for (inode = 0; inode < nnode; inode++) {
        xyz[0] = X[inode];
        xyz[1] = Y[inode];
        xyz[2] = 0;
        xyz[3] = 0;
        xyz[4] = 0;
        xyz[5] = 1;

        status = EG_makeGeometry(context, CURVE, LINE, NULL,
                                 NULL, xyz, &ecurve);
        if (status != EGADS_SUCCESS) goto cleanup;

        echild[0] = enodes[inode      ];
        echild[1] = enodes[inode+nnode];

        status = EG_invEvaluate(ecurve, xyz, &(data[0]), xyz_out);
        if (status != EGADS_SUCCESS) goto cleanup;

        xyz[0] = X[inode];
        xyz[1] = Y[inode];
        xyz[2] = DEPTH(0);

        status = EG_invEvaluate(ecurve, xyz, &(data[1]), xyz_out);
        if (status != EGADS_SUCCESS) goto cleanup;

        status = EG_makeTopology(context, ecurve, EDGE, TWONODE,
                                 data, 2, echild, NULL, &(eedges[jedge]));
        if (status != EGADS_SUCCESS) goto cleanup;

        jedge++;
    }

    /* set up the faces */
    jface = 0;

    for (iseg = 0; iseg < nseg; iseg++) {
        echild[0] = eedges[iseg];
        echild[1] = eedges[2*nseg+iend[iseg]];
        echild[2] = eedges[  nseg+     iseg ];
        echild[3] = eedges[2*nseg+ibeg[iseg]];

        senses[0] = SFORWARD;
        senses[1] = SFORWARD;
        senses[2] = SREVERSE;
        senses[3] = SREVERSE;

        status = EG_makeTopology(context, NULL, LOOP, CLOSED,
                                 NULL, 4, echild, senses, &eloop);
        if (status != EGADS_SUCCESS) goto cleanup;

        status = EG_makeFace(eloop, SFORWARD, NULL, &(efaces[jface]));
        if (status != EGADS_SUCCESS) goto cleanup;

        jseg = iseg + 1;
        status = EG_attributeAdd(efaces[jface], "segment", ATTRINT,
                                 1, &jseg, NULL, NULL);
        if (status != EGADS_SUCCESS) goto cleanup;

        jface++;
    }

    /* make the sheet body */
    status = EG_makeTopology(context, NULL, SHELL, OPEN,
                             NULL, jface, efaces, NULL, &eshell);
    if (status != EGADS_SUCCESS) goto cleanup;

    status = EG_makeTopology(context, NULL, BODY, SHEETBODY,
                             NULL, 1, &eshell, NULL, ebody);
    if (status != EGADS_SUCCESS) goto cleanup;

    /* set the output value(s) */

    /* remember this model (body) */
    udps[numUdp].ebody = *ebody;

cleanup:
    if (efaces != NULL) EG_free(efaces);
    if (eedges != NULL) EG_free(eedges);
    if (enodes != NULL) EG_free(enodes);

    if (X      != NULL) EG_free(X     );
    if (Y      != NULL) EG_free(Y     );
    if (ibeg   != NULL) EG_free(ibeg  );
    if (iend   != NULL) EG_free(iend  );

    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpSensitivity - return sensitivity derivatives for the "real" argument *
 *                                                                      *
 ************************************************************************
 */

int
udpSensitivity(ego    ebody,            /* (in)  Body pointer */
   /*@unused@*/int    npnt,             /* (in)  number of points */
   /*@unused@*/int    entType,          /* (in)  OCSM entity type */
   /*@unused@*/int    entIndex,         /* (in)  OCSM entity index (bias-1) */
   /*@unused@*/double uvs[],            /* (in)  parametric coordinates for evaluation */
   /*@unused@*/double vels[])           /* (out) velocities */
{
    int iudp, judp;

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == udps[judp].ebody) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    /* this routine is not written yet */
    return EGADS_NOLOAD;
}
