// ESP.js implements functions for the Engineering Sketch Pad (ESP)
// written by John Dannenhoffer and Bob Haimes

// Copyright (C) 2010/2015  John F. Dannenhoffer, III (Syracuse University)
//
// This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//    MA  02110-1301  USA

// functions expected by wv
//    wvInitUI()                called by wv-render.js
//    wvUpdateUI()              called by wv-render.js
//    wvServerMessage(text)     called by wv-socket.js
//    wvServerDown()            called by wv-socket.js
//    wvUpdateCanvas(gl)        called by ESP.html

// functions associated with button presses (and associated button presses)
//    activateBuildButton()
//    activateSolveButton()
//    cmdUndo()
//    cmdEdit()
//       editCsmOk()
//       editCsmCancel()
//    cmdQuit()
//    cmdSave()
//    cmdHelp()
//    cmdHome()
//    cmdLeft()
//    cmdRite()
//    cmdBotm()
//    cmdTop()
//    cmdIn()
//    cmdOut()

// functions associated with menu selections (and associated button presses)
//    addPmtr()
//    editPmtr(e)
//       compSens()
//       setVel(e)
//       clrVels()
//       editPmtrOk()
//       editPmtrCancel()
//    addBrch()
//       addBrchOk()
//       addBrchCancel()
//    editBrch(e)
//       delBrch(e)
//       addAttr()
//       enterSketcher()
//       buildTo()
//       editBrchOk()
//       editBrchCancel()
//    editAttr(e)

// functions associated with the mouse in the canvas
//    getMouseDown0(e)
//    getMouseMove0(e)
//    getMouseUp0(e)
//    getMouseRoll0(e)
//    mouseLeftCanvas0(e)

// function associated with the mouse in the Sketcher
//    getMouseDown7(e)
//    getMouseMove7(e)
//    getMouseUp7(e)

//  function associated with key presses in whole document
//    getKeyPress(e)
//    getKeyDown(e)
//    getKeyUp(e)

// functions associated with the mouse in the key window
//    setKeyLimits(e)

// functions associated with toggling display settings
//    toggleViz(e)
//    toggleGrd(e)
//    toggleTrn(e)
//    toggleOri(e)

// functions associated with the Sketcher
//    initializeSketch()
//    loadSketch(begs, vars, cons, segs)
//    sketchKeyPress()
//    solveSketchPre()
//    solveSketchPost(text)
//    quitSketch()
//    saveSketch()
//    undoSketch()
//    saveSketchUdo()
//
//    updateSegData(iseg)
//    drawSketch()
//    drawBezier(context, Xbezier, Ybezier)
//    drawSpline(context, Xspline, Yspline)
//    getClosestSketchPoint()
//    getClosestSketchSegment()
//    getClosestSketchConstraint()

// functions associated with a Tree in the treefrm
//    Tree(doc, treeId) - constructor
//    TreeAddNode(iparent, name, gprim, click, cmenu, prop1, valu1, cbck1, prop2, valu2, cbck2, prop3, valu3, cbck3)
//    TreeBuild()
//    TreeClear()
//    TreeContract(inode)
//    TreeExpand(inode)
//    TreeProp(inode, iprop, onoff)
//    TreeUpdate()

// helper functions
//    resizeFrames()            called byESP.html
//    changeMode(newMode)
//    postMessage(mesg)
//    setupEditBrchForm()
//    setupEditPmtrForm()
//    numberOfPmtrChanges()
//    numberOfBrchChanges()
//    unhighlightColumn1()
//    printObject(obj)
//    sprintf()

"use strict";


//
// callback when the user interface is to be initialized (called by wv-render.js)
//
function wvInitUI()
{
    // alert("wvInitUI()");

    // set up extra storage for matrix-matrix multiplies
    wv.uiMatrix   = new J3DIMatrix4();
    wv.saveMatrix = new J3DIMatrix4(wv.mvMatrix);

                                   // ui cursor variables
    wv.cursorX   = -1;             // current cursor position
    wv.cursorY   = -1;
    wv.keyPress  = -1;             // last key pressed
    wv.keyCode   = -1;
    wv.startX    = -1;             // start of dragging position
    wv.startY    = -1;
    wv.button    = -1;             // button pressed
    wv.modifier  =  0;             // modifier (shift,alt,cntl) bitflag
    wv.flying    =  1;             // flying multiplier (do not set to 0)
    wv.offTop0   =  0;             // offset to upper-left corner of the canvas
    wv.offLeft0  =  0;
    wv.offTop7   =  0;             // offset to upper-left corner of Sketcher
    wv.offLeft7  =  0;
    wv.dragging  =  false;         // true during drag operation
    wv.picking   =  0;             // keycode of command that turned picking on
    wv.locating  =  0;             // keycode of command that turned locating on
    wv.focus     = [0, 0, 0, 1];   // focus data needed in locating
    wv.pmtrStat  =  0;             // -2 latest Parameters are in Tree
                                   // -1 latest Parameters not in Tree (yet)
                                   //  0 need to request Parameters
                                   // >0 waiting for Parameters (request already made)
    wv.brchStat  =  0;             // -2 latest Branches are in Tree
                                   // -1 latest Branches not in Tree (yet)
                                   //  0 need to request Branches
                                   // >0 waiting for Branches (request already made)
    wv.builtTo   = 99999;          // last Branch in previous successful build
    wv.menuEvent = undefined;      // event associated with click in Tree
    wv.server    = undefined;      // string passed back from "identify;"
    wv.loLimit   = -1;             // lower limit in key
    wv.upLimit   = +1;             // upper limit in key
    wv.nchanges  = -1;             // number of changes since last getCsmFile (-1 to initialize)
    wv.filename  = "";             // name of the .csm file
    wv.curFile   = "";             // contents of current .csm file
    wv.curMode   =  0;             //-1 to disable buttons, etc.
                                   // 0 show WebViewer in canvas
                                   // 1 show addBrchForm
                                   // 2 show editBrchForm with addBrchHeader
                                   // 3 show editBrchForm with editBrchHeader
                                   // 4 show editPmtrForm with addPmtrHeader
                                   // 5 show editPmtrForm with editPmtrHeader
                                   // 6 show editCsmForm
                                   // 7 show sketcherForm
    wv.curPmtr   = -1;             // Parameter beibng editted (or -1)
    wv.curBrch   = -1;             // Branch being editted (or -1)
    wv.afterBrch = -1;             // Branch to add after (or -1)
    wv.numArgs   =  0;             // number of arguments in wv.curBrch
    wv.sgData    = {};             // scene graph metadata
    wv.scale     =  1;             // scale factor for axes
//  wv.centerV                     // set to 1 to center view and rotation
//  wv.pick                        // set to 1 to turn picking on
//  wv.locate                      // set to 1 to turn locating on
//  wv.sceneGraph                  // pointer to sceneGraph
//  wv.picked                      // sceneGraph object that was picked
//  wv.located                     // sceneGraph object that was located
//  wv.sceneUpd                    // should be set to 1 to re-render scene
//  wv.sgUpdate                    // =1 if the sceneGraph has been updated
//  wv.canvasID                    // ID of main canvas
//  wv.canvasKY                    // ID of key  canvas
//  wv.drawKey                     // =1 if key window needs to be redrawn
//  wv.socketUt.send(text)         // function to send text to server
//  wv.plotAttrs                   // plot attributes

//  sket.ibrch                     // Branch index that launched Sketcher
//  sket.mode                      // 0 initiaizing
//                                 // 1 drawing
//                                 // 2 setting curvature
//                                 // 3 constraining
//                                 // 4 setting width
//                                 // 5 setting depth
//                                 // 6 solved
//  sket.movingPoint = -1;         // moving point while constraining (or -1 for none)
//  sket.basePoint = undefined;    // base Point for W or D constraint
//  sket.skbegX    = undefined;    // expression for X from SKBEG
//  sket.skbegY    = undefined;    // expression for Y from SKBEG
//  sket.skbegZ    = undefined;    // expression for Z from SKBEG
//  sket.halo      =  5;           // pixels for determining "closeness" while drawing
//  sket.offTop    =  0;           // offset to upper-left corner of the canvas
//  sket.offLeft   =  0;

//  sket.scale     = undefined;    // drawing scale factor
//  sket.xorig     = undefined;    // x-origin for drawing
//  sket.yorig     = undefined;    // y-origin for drawing

//  sket.pnt.x[   i]               // x-coordinate
//  sket.pnt.y[   i]               // y-coordinate
//  sket.pnt.lbl[ i]               // constraint label

//  sket.seg.type[i]               // Segment type (L, C, S, or P)
//  sket.seg.ibeg[i]               // Point at beginning
//  sket.seg.iend[i]               // Point at end
//  sket.seg.xm[  i]               // x-coordinate at middle
//  sket.seg.ym[  i]               // y-coordinate at middle
//  sket.seg.lbl[ i]               // constraint label
//
//  sket.seg.dip[ i]               // dip (>0 for convex, <0 for concave)
//  sket.seg.xc[  i]               // x-coordinate at center
//  sket.seg.yc[  i]               // y-coordinate at center
//  sket.seg.rad[ i]               // radius (screen coordinates)
//  sket.seg.tbeg[i]               // theta at beginning (radians)
//  sket.seg.tend[i]               // theta at end       (radians)

//  sket.var.name[ i]              // variable name
//  sket.var.value[i]              // variable value

//  sket.con.type[  i]             // constraint type (X, Y, P, T, H, V, I, L, R, or S)
//  sket.con.index1[i]             // primary Point or Segment index
//  sket.con.index2[i]             // secondary Point or Segment index
//  sket.con.value[ i]             // (optional) constraint value

//  sket.undo.mode                 // undo snapshot (or undefined)
//  sket.undo.pnt
//  sket.undo.seg
//  sket.undo.var
//  sket.undo.con

    document.addEventListener('keypress', getKeyPress,     false);
    document.addEventListener('keydown',  getKeyDown,      false);
    document.addEventListener('keyup',    getKeyUp,        false);

    var canvas = document.getElementById(wv.canvasID);
    canvas.addEventListener('mousedown',  getMouseDown0,    false);
    canvas.addEventListener('mousemove',  getMouseMove0,    false);
    canvas.addEventListener('mouseup',    getMouseUp0,      false);
    canvas.addEventListener("wheel",      getMouseRoll0,    false);
    canvas.addEventListener('mouseout',   mouseLeftCanvas0, false);

    var sketcher = document.getElementById("sketcher");
    sketcher.addEventListener('mousedown',  getMouseDown7,  false);
    sketcher.addEventListener('mousemove',  getMouseMove7,  false);
    sketcher.addEventListener('mouseup',    getMouseUp7,    false);

    var keycan = document.getElementById(wv.canvasKY);
    keycan.addEventListener('mouseup',    setKeyLimits,    false);

    var quitButton = document.getElementById("quitButton");
    quitButton.hidden = true;
}


//
// callback when the user interface should be updated (called by wv-render.js)
//
function wvUpdateUI()
{
    // alert("wvUpdateUI()");

    // special code for delayed-picking mode
    if (wv.picking > 0) {

        // if something is picked, post a message
        if (wv.picked !== undefined) {

            // second part of '^' operation
            if (wv.picking == 94) {
                var mesg = "Picked: "+wv.picked.gprim;

                try {
                    var attrs = wv.sgData[wv.picked.gprim];
                    for (var i = 0; i < attrs.length; i+=2) {
                        mesg = mesg + "\n        "+attrs[i]+"= "+attrs[i+1];
                    }
                } catch (x) {
                }
                postMessage(mesg);
            }

            wv.picked  = undefined;
            wv.picking = 0;
            wv.pick    = 0;

        // abort picking on mouse motion
        } else if (wv.dragging) {
            postMessage("Picking aborted");

            wv.picking = 0;
            wv.pick    = 0;
        }

        wv.keyPress = -1;
        wv.dragging = false;
    }

    // special code for delayed-locating mode
    if (wv.locating > 0) {

        // if something is located, post a message
        if (wv.located !== undefined) {

            // second part of '@' operation
            if (wv.locating == 64) {
                var xloc = wv.focus[3] * wv.located[0] + wv.focus[0];
                var yloc = wv.focus[3] * wv.located[1] + wv.focus[1];
                var zloc = wv.focus[3] * wv.located[2] + wv.focus[2];

                postMessage("Located: x="+xloc.toFixed(4)+", y="+yloc.toFixed(4)+
                                   ", z="+zloc.toFixed(4));
            }

            wv.located  = undefined;
            wv.locating = 0;
            wv.locate   = 0;

        // abort locating on mouse motion
        } else if (wv.dragging) {
            postMessage("Locating aborted");

            wv.locating = 0;
            wv.locate   = 0;
        }

        wv.keyPress = -1;
        wv.dragging = false;
    }

    // if the Parameters are scheduled to be updated, send a message to
    //    get the Parameters now
    if (wv.pmtrStat > 0) {
        wv.pmtrStat--;
    } else if (wv.pmtrStat == 0) {
        try {
            wv.socketUt.send("identify|");
            wv.socketUt.send("getPmtrs|");
            wv.pmtrStat = 6000;
        } catch (e) {
            // could not send command
        }
    }

    // if the Branches are scheduled to be updated, send a message to
    //    get the Branches now
    if (wv.brchStat > 0) {
        wv.brchStat--;
    } else if (wv.brchStat == 0) {
        try {
            wv.socketUt.send("getBrchs|");
            wv.brchStat = 6000;
        } catch (e) {
            // could not send command
        }
    }

    // if the CsmFile has changed, send a message to
    //    get the CsmFile now
    if (wv.nchanges < 0) {
        try {
            wv.socketUt.send("getCsmFile|");
            wv.nchanges = 0;
        } catch (e) {
            // could not send message
        }
    }

    // if the scene graph and Parameters have been updated, (re-)build the Tree
    if ((wv.sgUpdate == 1 && wv.pmtrStat <= -1 && wv.brchStat <= -1) ||
        (                    wv.pmtrStat == -1 && wv.brchStat == -2) ||
        (                    wv.pmtrStat == -2 && wv.brchStat == -1)   ) {

        if (wv.sceneGraph === undefined) {
            alert("wv.sceneGraph is undefined --- but we need it");
        }

        // if there was a previous Tree, keep track of whether or not
        //    the Parameters, Branches, and Display was open
        var pmtr1Open = 0;
        var pmtr2Open = 0;
        var brchsOpen = 0;

        if (myTree.opened.length > 4) {
            pmtr1Open = myTree.opened[1];
            pmtr2Open = myTree.opened[2];
            brchsOpen = myTree.opened[3];
        }

        // clear previous Nodes from the Tree
        myTree.clear();

        // put the group headers into the Tree
        myTree.addNode(0, "Design Parameters", "", addPmtr, "*");
        myTree.addNode(0, "Local Variables",   "", null,    "" );
        myTree.addNode(0, "Branches",          "", addBrch, "*");
        myTree.addNode(0, "Display",           "", null,    "" );

        // put the Design Parameters into the Tree
        for (var ipmtr = 0; ipmtr < pmtr.length; ipmtr++) {
            var name  = "\u00a0\u00a0"+pmtr[ipmtr].name;
            var type  =                pmtr[ipmtr].type;
            var nrow  =                pmtr[ipmtr].nrow;
            var ncol  =                pmtr[ipmtr].ncol;
            var value =                pmtr[ipmtr].value[0];

            if (nrow > 1 || ncol > 1) {
                value = "["+nrow+"x"+ncol+"]";
            }

            if (type == 500 && pmtr[ipmtr].dot[0] != 0) {
                myTree.addNode(1, "^"+name, "", editPmtr, "*", ""+value, "", "");
            } else if (type == 500) {
                myTree.addNode(1,     name, "", editPmtr, "*", ""+value, "", "");
            }
        }

        wv.pmtrStat = -2;

        // open the Design Parameters (if they were open before the Tree was rebuilt)
        if (pmtr1Open == 1) {
            myTree.opened[1] = 1;
        }

        // put the Local Variables into the Tree
        for (ipmtr = 0; ipmtr < pmtr.length; ipmtr++) {
            var name  = "\u00a0\u00a0"+pmtr[ipmtr].name;
            var type  =                pmtr[ipmtr].type;
            var nrow  =                pmtr[ipmtr].nrow;
            var ncol  =                pmtr[ipmtr].ncol;
            var value =                pmtr[ipmtr].value[0];

            if (nrow > 1 || ncol > 1) {
                value = "["+nrow+"x"+ncol+"]";
            }

            if (type !== 500) {
                myTree.addNode(2,     name, "", null,     "",  ""+value, "", "");
                var inode = myTree.name.length - 1;

                var indx = 0;
                if (nrow > 1 || ncol > 1) {
                    for (var irow = 0; irow < nrow; irow++) {
                        for (var icol = 0; icol < ncol; icol++) {
                            name  = "\u00a0\u00a0\u00a0\u00a0["+(irow+1)+","+(icol+1)+"]";
                            value = pmtr[ipmtr].value[indx++];

                            myTree.addNode(inode, name, "", null, "",
                                           "\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0"+value, "", "");
                        }
                    }
                }


            }
        }

        wv.pmtrStat = -2;

        // open the Local Variables (if they were open before the Tree was rebuilt)
        if (pmtr2Open == 1) {
            myTree.opened[2] = 1;
        }

        // put the Branches into the Tree
        for (var ibrch = 0; ibrch < brch.length; ibrch++) {
            var name  = "\u00a0\u00a0";
            for (var indent = 0; indent < brch[ibrch].indent; indent++) {
                name = name+">";
            }
            name = name+brch[ibrch].name;

            var type  =      brch[ibrch].type;
            var actv;
            if (ibrch >= wv.builtTo) {
                actv = "skipped";
            } else if (brch[ibrch].actv == 301) {
                actv = "suppressed";
            } else if (brch[ibrch].actv == 302) {
                actv = "inactive";
            } else if (brch[ibrch].actv == 303) {
                actv = "deferred";
            } else {
                actv = "";
            }

            myTree.addNode(3, name, "", editBrch, "*", type, "", "", actv, "", "");

            // add the Branch's attributes to the Tree
            var inode = myTree.name.length - 1;
            for (var iattr = 0; iattr < brch[ibrch].attrs.length; iattr++) {
                var aname  = brch[ibrch].attrs[iattr][0];
                var avalue = brch[ibrch].attrs[iattr][1];
                myTree.addNode(inode, "\u00a0\u00a0\u00a0\u00a0"+aname, "", editAttr, "*", avalue, "", "");
            }
        }

        wv.brchStat = -2;

        // open the Branches (if they were open before the Tree was rebuilt)
        if (brchsOpen == 1) {
            myTree.opened[3] = 1;
        }

        // put the Display attributes into the Tree
        for (var gprim in wv.sceneGraph) {

            // parse the name
            var matches = gprim.split(" ");

            var ibody = Number(matches[1]);
            if        (matches[2] == "Face") {
                var iface = matches[3];
            } else if (matches[2] == "Edge") {
                var iedge = matches[3];
            } else {
                alert("unknown type: "+matches[2]);
                continue;
            }

            // determine if Body does not exists
            var kbody = -1;
            for (var jnode = 1; jnode < myTree.name.length; jnode++) {
                if (myTree.name[jnode] == "\u00a0\u00a0Body "+ibody) {
                    kbody = jnode;
                }
            }

            // if Body does not exist, create it and its Face and Edge
            //    subnodes now
            var kface, kedge;
            if (kbody < 0) {
                myTree.addNode(4, "\u00a0\u00a0Body "+ibody, "", null, "",
                               "Viz", "on", toggleViz, "Grd", "off", toggleGrd);
                kbody = myTree.name.length - 1;

                myTree.addNode(kbody, "\u00a0\u00a0\u00a0\u00a0Faces", "", null, "",
                               "Viz", "on", toggleViz, "Grd", "off", toggleGrd, "Trn", "off", toggleTrn);
                kface = myTree.name.length - 1;

                myTree.addNode(kbody, "\u00a0\u00a0\u00a0\u00a0Edges", "", null, "",
                               "Viz", "on", toggleViz, "Grd", "off", toggleGrd, "Ori", "off", toggleOri);
                kedge = myTree.name.length - 1;

            // otherwise, get pointers to the face-group and edge-group Nodes
            } else {
                kface = myTree.child[kbody];
                kedge = kface + 1;
            }

            // make the Tree Node
            if        (matches[2] == "Face") {
                myTree.addNode(kface, "\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0face "+iface, gprim, null, "",
                               "Viz", "on", toggleViz, "Grd", "off", toggleGrd, "Trn", "off", toggleTrn);
            } else if (matches[2] == "Edge") {
                myTree.addNode(kedge, "\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0edge "+iedge, gprim, null, "",
                               "Viz", "on", toggleViz, "Grd", "off", toggleGrd, "Ori", "off", toggleOri);
            }
        }

        // open the Display (by default)
        myTree.opened[4] = 1;

        // mark that we have (re-)built the Tree
        wv.sgUpdate = 0;

        // convert the abstract Tree Nodes into an HTML table
        myTree.build();
    }

    // deal with key presses
    if (wv.keyPress != -1 && wv.curMode == 0) {

        var myKeyPress = String.fromCharCode(wv.keyPress);

        // '?' -- help
        if (myKeyPress == "?") {
            postMessage("........................... Viewer Cursor options ...........................\n" +
                        "ctrl-h <Home> - initial view             ctrl-f        - front view          \n" +
                        "ctrl-l        - leftside view            ctrl-r        - riteside view       \n" +
                        "ctrl-t        - top view                 ctrl-b        - bottom view         \n" +
                        "ctrl-i <PgUp> - zoom in                  ctrl-o <PgDn> - zoom out            \n" +
                        "<Left>        - rotate or xlate left     <Rite>        - rotate or xlate rite\n" +
                        "<Up>          - rotate or xlate up       <Down>        - rotate or xlate down\n" +
                        ">             - save view                <             - recall view         \n" +
                        "^             - query object at cursor   @             - get coords. @ cursor\n" +
                        "!             - toggle flying mode       *             - center view @ cursor\n" +
                        ".............................. ? - get help .................................");

        // '^' -- query at cursor
        } else if (myKeyPress == '^') {
            wv.picking  = 94;
            wv.pick     = 1;
            wv.sceneUpd = 1;

        // '@' -- locate at cursor
        } else if (myKeyPress == '@') {
            wv.locating = 64;
            wv.locate   = 1;

        // '*' -- center view
        } else if (myKeyPress == '*') {
            wv.centerV = 1;
            postMessage("View being centered");

        // '!' -- toggle flying mode
        } else if (myKeyPress == "!") {
            if (wv.flying <= 1) {
                postMessage("Turning flying mode ON");
                wv.flying = 10;
            } else {
                postMessage("Turning flying mode OFF");
                wv.flying = 1;
            }

        // '>' -- save view
        } else if (myKeyPress == ">") {
            postMessage("Saving current view");
            wv.saveMatrix.load(wv.mvMatrix);
            wv.sceneUpd = 1;

        // '<' -- recall view
        } else if (myKeyPress == "<") {
            postMessage("Restoring saved view");
            wv.mvMatrix.load(wv.saveMatrix);
            wv.sceneUpd = 1;

        // '<esc>' -- not used
//        } else if (wv.keyPress == 0 && wv.keyCode == 27) {
//            postMessage("<Esc> is not supported.  Use '?' for help");

        // '<Home>' -- initial view
        } else if (wv.keyPress == 0 && wv.keyCode == 36) {
            wv.mvMatrix.makeIdentity();
            wv.scale    = 1;
            wv.sceneUpd = 1;

        // '<end>' -- not used
        } else if (wv.keyPress == 0 && wv.keyCode == 35) {
            postMessage("<End> is not supported.  Use '?' for help");

        // '<PgUp>' -- zoom in
        } else if (wv.keyPress == 0 && wv.keyCode == 33) {
            if (wv.modifier == 0) {
                wv.mvMatrix.scale(2.0, 2.0, 2.0);
                wv.scale *= 2.0;
            } else {
                wv.mvMatrix.scale(1.25, 1.25, 1.25);
                wv.scale *= 1.25;
            }
            wv.sceneUpd = 1;

        // '<PgDn>' -- zoom out
        } else if (wv.keyPress == 0 && wv.keyCode == 34) {
            if (wv.modifier == 0) {
                wv.mvMatrix.scale(0.5, 0.5, 0.5);
                wv.scale *= 0.5;
            } else {
                wv.mvMatrix.scale(0.8, 0.8, 0.8);
                wv.scale *= 0.8;
            }
            wv.sceneUpd = 1;

        // '<Delete>' -- not used
        } else if (wv.keyPress == 0 && wv.keyCode == 46) {
            postMessage("<Delete> is not supported.  Use '?' for help");

        // '<Left>' -- rotate or translate object left
        } else if (wv.keyPress == 0 && wv.keyCode == 37) {
            if (wv.flying == 1) {
                if (wv.modifier == 0) {
                    wv.mvMatrix.rotate(-30, 0,1,0);
                } else {
                    wv.mvMatrix.rotate( -5, 0,1,0);
                }
            } else {
                if (wv.modifier == 0) {
                    wv.mvMatrix.translate(-0.5, 0.0, 0.0);
                } else {
                    wv.mvMatrix.translate(-0.1, 0.0, 0.0);
                }
            }
            wv.sceneUpd = 1;

        // '<Right>' -- rotate or translate object right
        } else if (wv.keyPress == 0 && wv.keyCode == 39) {
            if (wv.flying == 1) {
                if (wv.modifier == 0) {
                    wv.mvMatrix.rotate(+30, 0,1,0);
                } else {
                    wv.mvMatrix.rotate( +5, 0,1,0);
                }
            } else {
                if (wv.modifier == 0) {
                    wv.mvMatrix.translate(+0.5, 0.0, 0.0);
                } else {
                    wv.mvMatrix.translate(+0.1, 0.0, 0.0);
                }
            }
            wv.sceneUpd = 1;

        // '<Up>' -- rotate or translate object up
        } else if (wv.keyPress == 0 && wv.keyCode == 38) {
            if (wv.flying == 1) {
                if (wv.modifier == 0) {
                    wv.mvMatrix.rotate(-30, 1,0,0);
                } else {
                    wv.mvMatrix.rotate( -5, 1,0,0);
                }
            } else {
                if (wv.modifier == 0) {
                    wv.mvMatrix.translate(0.0, +0.5, 0.0);
                } else {
                    wv.mvMatrix.translate(0.0, +0.1, 0.0);
                }
            }
            wv.sceneUpd = 1;

        // '<Down>' -- rotate or translate object down
        } else if (wv.keyPress == 0 && wv.keyCode == 40) {
            if (wv.flying == 1) {
                if (wv.modifier == 0) {
                    wv.mvMatrix.rotate(+30, 1,0,0);
                } else {
                    wv.mvMatrix.rotate( +5, 1,0,0);
                }
            } else {
                if (wv.modifier == 0) {
                    wv.mvMatrix.translate(0.0, -0.5, 0.0);
                } else {
                    wv.mvMatrix.translate(0.0, -0.1, 0.0);
                }
            }
            wv.sceneUpd = 1;

        // 'ctrl-h' - initial view (same as <Home>)
        } else if ((wv.keyPress == 104 && wv.modifier == 4 && wv.keyCode ==  0) ||
                   (wv.keyPress ==   8 && wv.modifier == 4 && wv.keyCode ==  8)   ) {
            cmdHome();

        // 'ctrl-i' - zoom in (same as <PgUp> without shift)
        } else if ((wv.keyPress == 105 && wv.modifier == 4 && wv.keyCode ==  0) ||
                   (wv.keyPress ==   9 && wv.modifier == 4 && wv.keyCode ==  9)   ) {
            cmdIn();

        // 'ctrl-o' - zoom out (same as <PgDn> without shift)
        } else if ((wv.keyPress == 111 && wv.modifier == 4 && wv.keyCode ==  0) ||
                   (wv.keyPress ==  15 && wv.modifier == 4 && wv.keyCode == 15)   ) {
            cmdOut();

        // 'ctrl-f' - front view (same as <Home>)
        } else if ((wv.keyPress == 102 && wv.modifier == 4 && wv.keyCode ==  0) ||
                   (wv.keyPress ==   6 && wv.modifier == 4 && wv.keyCode ==  6)   ) {
            cmdHome();

        // 'ctrl-r' - riteside view
        } else if ((wv.keyPress == 114 && wv.modifier == 4 && wv.keyCode ==  0) ||
                   (wv.keyPress ==  18 && wv.modifier == 4 && wv.keyCode == 18)   ) {
            cmdRite();

        // 'ctrl-l' - leftside view
        } else if ((wv.keyPress == 108 && wv.modifier == 4 && wv.keyCode ==  0) ||
                   (wv.keyPress ==  12 && wv.modifier == 4 && wv.keyCode == 12)   ) {
            cmdLeft();

        // 'ctrl-t' - top view
        } else if ((wv.keyPress == 116 && wv.modifier == 4 && wv.keyCode ==  0) ||
                   (wv.keyPress ==  20 && wv.modifier == 4 && wv.keyCode == 20)   ) {
            cmdTop();

        // 'ctrl-b' - bottom view
        } else if ((wv.keyPress ==  98 && wv.modifier == 4 && wv.keyCode ==  0) ||
                   (wv.keyPress ==   2 && wv.modifier == 4 && wv.keyCode ==  2)   ) {
            cmdBotm();

        // unknown command
        } else {
            postMessage("Unknown command (keyPress="+wv.keyPress
                        +", modifier="+wv.modifier
                        +", keyCode="+wv.keyCode+").  Use '?' for help");
        }

        wv.keyPress = -1;
    }

    // UI is in screen coordinates (not object)
    wv.uiMatrix.load(wv.mvMatrix);
    wv.mvMatrix.makeIdentity();

    // deal with mouse movement
    if (wv.dragging) {

        // cntrl is down (rotate)
        if (wv.modifier == 4) {
            var angleX =  (wv.startY - wv.cursorY) / 4.0 / wv.flying;
            var angleY = -(wv.startX - wv.cursorX) / 4.0 / wv.flying;
            if ((angleX != 0.0) || (angleY != 0.0)) {
                wv.mvMatrix.rotate(angleX, 1,0,0);
                wv.mvMatrix.rotate(angleY, 0,1,0);
                wv.sceneUpd = 1;
            }

        // alt-shift is down (rotate)
        } else if (wv.modifier == 3) {
            var angleX =  (wv.startY - wv.cursorY) / 4.0 / wv.flying;
            var angleY = -(wv.startX - wv.cursorX) / 4.0 / wv.flying;
            if ((angleX != 0.0) || (angleY != 0.0)) {
                wv.mvMatrix.rotate(angleX, 1,0,0);
                wv.mvMatrix.rotate(angleY, 0,1,0);
                wv.sceneUpd = 1;
            }

        // alt is down (spin)
        } else if (wv.modifier == 2) {
            var xf = wv.startX - wv.width  / 2;
            var yf = wv.startY - wv.height / 2;

            if ((xf != 0.0) || (yf != 0.0)) {
                var theta1 = Math.atan2(yf, xf);
                xf = wv.cursorX - wv.width  / 2;
                yf = wv.cursorY - wv.height / 2;

                if ((xf != 0.0) || (yf != 0.0)) {
                    var dtheta = Math.atan2(yf, xf) - theta1;
                    if (Math.abs(dtheta) < 1.5708) {
                        var angleZ = 128*(dtheta) / 3.1415926 / wv.flying;
                        wv.mvMatrix.rotate(angleZ, 0,0,1);
                        wv.sceneUpd = 1;
                    }
                }
            }

        // shift is down (zoom)
        } else if (wv.modifier == 1) {
            if (wv.cursorY != wv.startY) {
                var scale = Math.exp((wv.cursorY - wv.startY) / 512.0 / wv.flying);
                wv.mvMatrix.scale(scale, scale, scale);
                wv.scale   *= scale;
                wv.sceneUpd = 1;
            }

        // no modifier (translate)
        } else {
            var transX = (wv.cursorX - wv.startX) / 256.0 / wv.flying;
            var transY = (wv.cursorY - wv.startY) / 256.0 / wv.flying;
            if ((transX != 0.0) || (transY != 0.0)) {
                wv.mvMatrix.translate(transX, transY, 0.0);
                wv.sceneUpd = 1;
            }
        }

        // if not flying, then update the start coordinates
        if (wv.flying <= 1) {
            wv.startX = wv.cursorX;
            wv.startY = wv.cursorY;
        }
    }
}


//
// callback when a (text) message is received from the server (called by wv-socket.js)
//
function wvServerMessage(text)
{
    // alert("wvServerMessage(text="+text+")");

    // if it is a null message, do nothing
    if (text.length <= 1) {

    // if it starts with "identify|" post a message */
    } else if (text.substring(0,9) == "identify|") {
        if (wv.server === undefined) {
            wv.server = text.substring(9,text.length-2);
            postMessage("ESP has been initialized and is attached to '"+wv.server+"'");
        }

    // if it starts with "sgData|" store the auxiliary scene graph data
    } else if (text.substring(0,7) == "sgData|") {
        if (text.length > 9) {

            // strip of ",]" and ",}"
            var newtext = text.substring(7,text.length-1);
            newtext = newtext.replace(/,]/g, "]").replace(/,}/g, "}");
            wv.sgData = JSON.parse(newtext);
        } else {
            wv.sgData = {};
        }

    // if it starts with "sgFocus|" store the scene graph focus data
    } else if (text.substring(0,8) == "sgFocus|") {
        if (text.length > 10) {
            wv.focus = JSON.parse(text.substring(8,text.length-1));
        } else {
            wv.focus = [];
        }

    // if it starts with "getPmtrs|" build the (global) pmtr array
    } else if (text.substring(0,9) == "getPmtrs|") {
        if (text.length > 11) {
            pmtr = JSON.parse(text.substring(9,text.length-1));
        } else {
            pmtr = new Array;
        }
        wv.pmtrStat = -1;

    // if it starts with "newPmtr|" do nothing
    } else if (text.substring(0,8) == "newPmtr|") {

    // if it starts with "clrVels|" do nothing
    } else if (text.substring(0,8) == "clrVels|") {

    // if it starts with "setPmtr|" do nothing
    } else if (text.substring(0,8) == "setPmtr|") {

    // if it starts with "setVel|" do nothing
    } else if (text.substring(0,7) == "setVel|") {

    // if it starts with "getBrchs|" build the (global) brch array
    } else if (text.substring(0,9) == "getBrchs|") {
        if (text.length > 11) {
            brch = JSON.parse(text.substring(9,text.length-1));
        } else {
            brch = new Array;
        }
        wv.brchStat = -1;

    // if it starts with "newBrch|" do nothing
    } else if (text.substring(0,8) == "newBrch|") {

    // if it starts with "setBrch|" do nothing
    } else if (text.substring(0,8) == "setBrch|") {

    // if it starts with "delBrch|" do nothing
    } else if (text.substring(0,8) == "delBrch|") {

    // if it starts with "setAttr|" do nothing
    } else if (text.substring(0,8) == "setAttr|") {

    // if it starts with "undo|" get the latest Parameters and
    //                           Branches and update tree (scene)
    } else if (text.substring(0,5) == "undo|") {
        var cmd = text.substring(5,text.length-2);

        wv.sgUpdate = 1;
        wv.brchStat = 0;
        wv.pmtrStat = 0;

        postMessage("Undoing '"+cmd+"' ====> Re-build is needed <====");

    // if it starts with "save|" do nothing
    } else if (text.substring(0,5) == "save|") {

    // if it starts with "build|" reset the build button
    } else if (text.substring(0,6) == "build|") {
        var button = document.getElementById("buildButton");
        button.onclick = null;
        button["innerHTML"] = "Up to date";
        button.style.backgroundColor = null;

        var textList = text.split("|");

        if (textList[1].substring(0,7) == "ERROR::") {
            postMessage(textList[1]);
        } else {
            var ibrch    = Number(textList[1]);
            var nbody    = Number(textList[2]);

            wv.builtTo = ibrch;

            if (ibrch == brch.length) {
                postMessage("Entire build complete, which generated "+nbody+
                            " Body(s)");
            } else if (ibrch >= brch.length) {
                postMessage("Build complete through ibrch="+ibrch+
                            ", which generated "+nbody+" Body(s)");
            } else {
                postMessage("Partial build (through "+brch[ibrch-1].name+
                            ") complete, which generated "+nbody+" Body(s)");
            }
        }

        changeMode(0);

    // if it starts with "loadSketch|" initialize the Sketcher
    } else if (text.substring(0,11) == "loadSketch|") {

        var textList = text.split("|");

        if (textList[1].substring(0,7) == "ERROR::") {
            postMessage(textList[1]);
            alert("Sketch cannot be loaded");
        } else {
            // open the Sketcher
            changeMode(7);
            initializeSketch();

            loadSketch(textList[1], textList[2], textList[3], textList[4]);
        }

    // if it starts with "solveSketch|" update the Sketcher
    } else if (text.substring(0,12) == "solveSketch|") {

        solveSketchPost(text);

    // if it starts with "saveSketch|" either return an error or update the mode
    } else if (text.substring(0,11) == "saveSketch|") {

        if (text.substring(11,14) == "ok|") {
            activateBuildButton();
            changeMode(0);
        } else {
            alert("The sketch could not be saved");
        }

    // if it starts with "setLims|" do nothing
    } else if (text.substring(0,8) == "setLims|") {

    // if it starts with "getCsmFile|" store the results in wv.filename
    //    and wv.curFile
    } else if (text.substring(0,11) == "getCsmFile|") {
        var ichar = 11;

        wv.filename = "";
        while (text[ichar] != "|") {
            wv.filename += text[ichar++];
        }
        ichar++;

        wv.curFile = text.substring(ichar);

    // if it starts with "setCsmFileBeg|" do nothing
    } else if (text.substring(0,14) == "setCsmFileBeg|") {

    // if it starts with "setCsmFileMid|" do nothing
    } else if (text.substring(0,14) == "setCsmFileMid|") {

    // if it starts with "setCsmFileEnd|" do nothing
    } else if (text.substring(0,14) == "setCsmFileEnd|") {

    // if it starts with "setWvKey|" turn key or logo on
    } else if (text.substring(0,9) == "setWvKey|") {
        if (wv.curMode == 0) {
            if (text.substring(9,11) == "on") {
                document.getElementById("WVkey"  ).hidden = false;
                document.getElementById("ESPlogo").hidden = true;
            } else {
                document.getElementById("WVkey"  ).hidden = true;
                document.getElementById("ESPlogo").hidden = false;
            }
        }

    // if it starts with "ERROR:: there is nothing to undo" post the error
    } else if (text.substring(0,32) == "ERROR:: there is nothing to undo") {
        postMessage(text);
        alert("There is nothing to undo");

    // if it starts with "ERROR::" post the error
    } else if (text.substring(0,7) == "ERROR::") {
        postMessage(text);
        alert("Error encountered (see messages)");

        var button = document.getElementById("buildButton");

        button["innerHTML"] = "Fix before re-build";
        button.style.backgroundColor = "#FF3F3F";

        changeMode(0);

        button.onclick = function () {
        };

    // default is to post the message
    } else {
        postMessage(text);
    }
}


//
// callback when the server goes down either on abort or normal closure (called by wv-socket.js)
//
function wvServerDown()
{
    // deactivate the buttons
    wv.curMode = -1;

    postMessage("The server has terminated or network connection has been lost.\n"
                +"ESP will be working off-line");

    // turn the background of the message window pink
    var botm = document.getElementById("brframe");
    botm.style.backgroundColor = "#FFAFAF";
}


//
// callback used to put axes on the canvas (called by ESP.html)
//
function wvUpdateCanvas(gl)
{
    // Construct the identity as projection matrix and pass it in
    wv.mvpMatrix.load(wv.perspectiveMatrix);
    wv.mvpMatrix.setUniform(gl, wv.u_modelViewProjMatrixLoc, false);

    var mv   = new J3DIMatrix4();
    var mVal = wv.mvMatrix.getAsArray();

    mVal[ 3] = 0.0;
    mVal[ 7] = 0.0;
    mVal[11] = 0.0;
    mv.load(mVal);
    mv.scale(1.0/wv.scale, 1.0/wv.scale, 1.0/wv.scale);
    mv.invert();
    mv.transpose();

    // define location of axes in space
    var x    = -1.5 * wv.width / wv.height;
    var y    = -1.5;
    var z    =  0.9;
    var pdel =  0.15;     // length of axes in positive direction
    var mdel =  0.0;      // length of axes in negative direction

    // set up coordinates for axes
    mVal = mv.getAsArray();

    var vertices = new Float32Array(18);
    vertices[ 0] = x - mdel*mVal[ 0];   vertices[ 1] = y - mdel*mVal[ 1];   vertices[ 2] = z - mdel*mVal[ 2];
    vertices[ 3] = x + pdel*mVal[ 0];   vertices[ 4] = y + pdel*mVal[ 1];   vertices[ 5] = z + pdel*mVal[ 2];
    vertices[ 6] = x - mdel*mVal[ 4];   vertices[ 7] = y - mdel*mVal[ 5];   vertices[ 8] = z - mdel*mVal[ 6];
    vertices[ 9] = x + pdel*mVal[ 4];   vertices[10] = y + pdel*mVal[ 5];   vertices[11] = z + pdel*mVal[ 6];
    vertices[12] = x - mdel*mVal[ 8];   vertices[13] = y - mdel*mVal[ 9];   vertices[14] = z - mdel*mVal[10];
    vertices[15] = x + pdel*mVal[ 8];   vertices[16] = y + pdel*mVal[ 9];   vertices[17] = z + pdel*mVal[10];

    // set up colors for the axes
    var colors = new Uint8Array(18);
    colors[ 0] = 255;   colors[ 1] =   0;   colors[ 2] =   0;
    colors[ 3] = 255;   colors[ 4] =   0;   colors[ 5] =   0;
    colors[ 6] =   0;   colors[ 7] = 255;   colors[ 8] =   0;
    colors[ 9] =   0;   colors[10] = 255;   colors[11] =   0;
    colors[12] =   0;   colors[13] =   0;   colors[14] = 255;
    colors[15] =   0;   colors[16] =   0;   colors[17] = 255;

    // draw the axes
    if (gl.lineWidth) {
        gl.lineWidth(3);
    }
    gl.disableVertexAttribArray(2);
    gl.uniform1f(wv.u_wLightLoc, 0.0);

    var buffer = gl.createBuffer();
    gl.bindBuffer(gl.ARRAY_BUFFER, buffer);
    gl.bufferData(gl.ARRAY_BUFFER, vertices, gl.STATIC_DRAW);
    gl.vertexAttribPointer(0, 3, gl.FLOAT, false, 0, 0);
    gl.enableVertexAttribArray(0);

    var cbuf = gl.createBuffer();
    gl.bindBuffer(gl.ARRAY_BUFFER, cbuf);
    gl.bufferData(gl.ARRAY_BUFFER, colors, gl.STATIC_DRAW);
    gl.vertexAttribPointer(1, 3, gl.UNSIGNED_BYTE, false, 0, 0);
    gl.enableVertexAttribArray(1);

    gl.drawArrays(gl.LINES, 0, 6);
    gl.deleteBuffer(buffer);
    gl.deleteBuffer(cbuf);
    gl.uniform1f(wv.u_wLightLoc, 1.0);
}


//
// activate the buildButton
//
function activateBuildButton() {
    // alert("in activateBuildButton()");

    var button = document.getElementById("buildButton");

    button["innerHTML"] = "Press to Re-build";
    button.style.backgroundColor = "#3FFF3F";

    button.onclick = function () {
        if (wv.curMode != 0) {
            alert("Command disabled,  Press 'Cancel' or 'OK' first");
            return;
        } else {
            postMessage("Re-building...");
        }

        // build first so that parameters are updated
        wv.socketUt.send("build|0|");
        wv.brchStat = 6000;

        wv.socketUt.send("getPmtrs|");
        wv.pmtrStat = 6000;

        wv.socketUt.send("getBrchs|");

        var button = document.getElementById("buildButton");
        button.onclick = null;
        button["innerHTML"] = "Re-building...";
        button.style.backgroundColor = "#FFFF3F";

        //inactivate buttons until build is done
        changeMode(-1);
    };
}


//
// activate the solveButton
//
function activateSolveButton() {
    // alert("in activateSolveButton()");

    var button = document.getElementById("buildButton");

    button["innerHTML"] = "Press to Solve";
    button.style.backgroundColor = "#3FFF3F";

    button.onclick = function () {
        if (wv.curMode != 7) {
            alert("Command disabled,  Press 'Cancel' or 'OK' first");
            return;
        }

        // send solve request to the server
        solveSketchPre();
    };
}


//
// callback when "undoButton" is pressed (called by ESP.html)
//
function cmdUndo() {
    // alert("in cmdUndo()");

    if (wv.curMode == 7) {
        undoSketch();
        return;
    } else if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first");
        return;
    } else if (wv.server != "serveCSM") {
        alert("cmdUndo is not implemented for "+wv.server);
        return;
    }

    wv.socketUt.send("undo|");

    // update the UI
    activateBuildButton();
}


//
// callback when "editButton" is pressed (called by ESP.html)
//
function cmdEdit() {
    // alert("in cmdEdit()");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first");
        return;
    } else if (wv.server != "serveCSM") {
        alert("cmdEdit is not implemented for "+wv.server);
        return;
    }

    // if there have been any changes, tell user to save first
    if (wv.nchanges > 0) {
        alert("Changes have been made to the 'Design Parameters'\n" +
              "and/or the 'Branches' via the user-interface.\n\n" +
              "Save these changes before you can edit the .csm file");
        return;
    }

    // fill in the name of the .csm file
    var csmFilename = document.getElementById("editCsmFilename");

    csmFilename["innerHTML"] = "Contents of: "+wv.filename;

    // fill the textarea with the current .csm file
    var csmTextArea = document.getElementById("editCsmTextArea");

    csmTextArea.cols  = 84;
    csmTextArea.rows  = 25;
    csmTextArea.value = wv.curFile;

    // post the editCsmForm
    changeMode(6);
}


//
// callback when "OK" button is pressed in editCsmForm (called by ESP.html)
//
function editCsmOk() {
    // alert("in editCsmOk()");

    var editCsmTextArea = document.getElementById("editCsmTextArea");

    // tell user if no changes were made
    if (wv.curFile == editCsmTextArea.value) {
        alert("No changes were made");
        changeMode(0);
        return;
    } else {
        if (confirm("This will overwrite your input file.  Continue?") !== true) {
            return;
        }
    }

    // because of an apparent limit on the size of text
    //    messages that can be sent from the browser to the
    //    server, we need to send the new file back in
    //    pieces and then reassemble on the server
    var maxMessageSize = 1000;

    var ichar = 0;
    var part  = editCsmTextArea.value.substring(ichar, ichar+maxMessageSize);
    wv.socketUt.send("setCsmFileBeg|"+part);
    ichar += maxMessageSize;

    while (ichar < editCsmTextArea.value.length) {
        part = editCsmTextArea.value.substring(ichar, ichar+maxMessageSize);
        wv.socketUt.send("setCsmFileMid|"+part);
        ichar += maxMessageSize;
    }

    wv.socketUt.send("setCsmFileEnd|");

    // remember the edited .csm file
    wv.curFile = editCsmTextArea.value;

    postMessage("'"+wv.filename+"' file has been changed.  Re-building...");

    // get an updated version of the Parameters and Branches
    wv.pmtrStat = 0;
    wv.brchStat = 0;

    // reset the number of changes
    wv.nchanges = 0;

    // inform the user that a rebuild is in process
    var button = document.getElementById("buildButton");
    button.onclick = null;
    button["innerHTML"] = "Re-building...";
    button.style.backgroundColor = "#FFFF3F";

    // inactivate buttons until build is done
    changeMode( 0);
    changeMode(-1);
}


//
// callback when "Cancel" is pressed in editCsmForm (called by ESP.html)
//
function editCsmCancel() {
    // alert("in editCsmCancel()");

    // return to the WebViewer
    changeMode(0);
}


//
// callback when "quitButton" is pressed (called by ESP.html)
//
function cmdQuit() {
    // alert("in cmdQuit()");

    if (wv.curMode != 7) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first");
        return;
    } else if (wv.server != "serveCSM") {
        alert("cmdQuit is not implemented for "+wv.server);
        return;
    }

    quitSketch();
    activateBuildButton();
}


//
// callback when "saveButton" is pressed (called by ESP.html)
//
function cmdSave() {
    // alert("in cmdSave()");

    if (wv.server != "serveCSM") {
        alert("cmdSave is not implemented for "+wv.server);
        return;

    } else if (wv.curMode == 0) {
        var filename = prompt("Enter filename:", wv.filename);
        if (filename !== null) {
            if (filename.search(".csm") == -1) {
                filename = filename + ".csm";
            }

            postMessage("Saving model to '"+filename+"'");
            wv.socketUt.send("save|"+filename+"|");

            wv.socketUt.send("getFile|");

            wv.nchanges = -1;
        } else {
            postMessage("NOT saving since no filename specified");
        }

    } else if (wv.curMode == 7) {
        saveSketch();
        wv.brchStat = 0;

    } else {
         alert("Command disabled.  Press 'Cancel' or 'OK' first");
        return;

    }
}


//
// callback when "helpButton" is pressed (called by ESP.html)
//
function cmdHelp() {

    // open help in another tab
    window.open("ESP-help.html");
}


//
// callback when "homeButton" is pressed (called by ESP.html)
//
function cmdHome()
{
    if (wv.curMode == 0) {
        wv.mvMatrix.makeIdentity();
        wv.uiMatrix.load(wv.mvMatrix);
        wv.mvMatrix.makeIdentity();
        wv.scale    = 1;
        wv.sceneUpd = 1;
    } else if (wv.curMode == 7) {
        var npnt = sket.pnt.x.length;
        var nseg = sket.seg.type.length;

        if (npnt <= 1) {
            alert("Not enough data to rescale");
            return;
        } else if (sket.scale === undefined || isNaN(sket.scale)) {
            alert("At least one length or radius must be set before rescaling  (sket.scale="+sket.scale+")");
            return;
        } else if (sket.xorig === undefined || isNaN(sket.xorig)) {
            alert("At lest one \"X\" must be set before rescaling  (sket.xorig="+sket.xorig+")");
            return;
        } else if (sket.yorig === undefined || isNaN(sket.yorig)) {
            alert("At lest one \"Y\" must be set before rescaling  (sket.yorig="+sket.yorig+")");
            return;
        }

        // get the extrema of the data
        var xmin = +1.0e+10;
        var xmax = -1.0e+10;
        var ymin = +1.0e+10;
        var ymax = -1.0e+10;

        for (var ipnt = 0; ipnt < npnt; ipnt++) {
            xmin = Math.min(xmin, sket.scale * (sket.pnt.x[ipnt] - sket.xorig));
            xmax = Math.max(xmax, sket.scale * (sket.pnt.x[ipnt] - sket.xorig));
            ymin = Math.min(ymin, sket.scale * (sket.yorig - sket.pnt.y[ipnt]));
            ymax = Math.max(ymax, sket.scale * (sket.yorig - sket.pnt.y[ipnt]));
        }

        // get the size of the canvas
        var canvas = document.getElementById("sketcher");
        var width  = canvas.clientWidth;
        var height = canvas.clientHeight;

        var test1 = (xmax - xmin) / width;
        var test2 = (ymax - ymin) / height;

        // set up sizing so that Sketch fills middle 50 percent of the window
        var newScale = 2 * Math.max(test1, test2);
        var newXorig = (width  - (xmin + xmax) / newScale) / 2;
        var newYorig = (height + (ymin + ymax) / newScale) / 2;

        // convert the Points by the new scale factors
        for (ipnt = 0; ipnt < npnt; ipnt++) {
            var xx = sket.scale * (sket.pnt.x[ipnt] - sket.xorig);
            var yy = sket.scale * (sket.yorig - sket.pnt.y[ipnt]);

            sket.pnt.x[ipnt] = Math.floor(newXorig + xx / newScale);
            sket.pnt.y[ipnt] = Math.floor(newYorig - yy / newScale);
        }

        for (var iseg = 0; iseg < nseg; iseg++) {
            sket.seg.dip[iseg] = sket.seg.dip[iseg] * sket.scale / newScale;

            updateSegData(iseg);
        }

        sket.scale = newScale;
        sket.xorig = newXorig;
        sket.yorig = newYorig;

        drawSketch();
    }
}


//
// callback when "leftButton" is pressed (called by ESP.html)
//
function cmdLeft()
{
    if (wv.curMode == 0) {
        wv.mvMatrix.makeIdentity();
        wv.mvMatrix.rotate(+90, 0,1,0);
        wv.uiMatrix.load(wv.mvMatrix);
        wv.mvMatrix.makeIdentity();
        wv.scale    = 1;
        wv.sceneUpd = 1;
    } else if (wv.curMode == 7) {
        var npnt = sket.pnt.x.length;
        var nseg = sket.seg.type.length;

        if (npnt <= 1) {
            alert("Not enough data to rescale");
            return;
        } else if (sket.scale === undefined || isNaN(sket.scale)) {
            alert("At least one length or radius must be set before rescaling  (sket.scale="+sket.scale+")");
            return;
        } else if (sket.xorig === undefined || isNaN(sket.xorig)) {
            alert("At lest one \"X\" must be set before rescaling  (sket.xorig="+sket.xorig+")");
            return;
        } else if (sket.yorig === undefined || isNaN(sket.yorig)) {
            alert("At lest one \"Y\" must be set before rescaling  (sket.yorig="+sket.yorig+")");
            return;
        }

        var newXorig = sket.xorig - 100;

        for (var ipnt = 0; ipnt < npnt; ipnt++) {
            var xx = sket.scale * (sket.pnt.x[ipnt] - sket.xorig);
            sket.pnt.x[ipnt] = Math.floor(newXorig + xx / sket.scale);
        }

        for (var iseg = 0; iseg < nseg; iseg++) {
            updateSegData(iseg);
        }

        sket.xorig = newXorig;

        drawSketch();
    }
}


//
// callback when "riteButton" is pressed (called by ESP.html)
//
function cmdRite()
{
    if (wv.curMode == 0) {
        wv.mvMatrix.makeIdentity();
        wv.mvMatrix.rotate(-90, 0,1,0);
        wv.uiMatrix.load(wv.mvMatrix);
        wv.mvMatrix.makeIdentity();
        wv.scale    = 1;
        wv.sceneUpd = 1;
    } else if (wv.curMode == 7) {
        var npnt = sket.pnt.x.length;
        var nseg = sket.seg.type.length;

        if (npnt <= 1) {
            alert("Not enough data to rescale");
            return;
        } else if (sket.scale === undefined || isNaN(sket.scale)) {
            alert("At least one length or radius must be set before rescaling  (sket.scale="+sket.scale+")");
            return;
        } else if (sket.xorig === undefined || isNaN(sket.xorig)) {
            alert("At lest one \"X\" must be set before rescaling  (sket.xorig="+sket.xorig+")");
            return;
        } else if (sket.yorig === undefined || isNaN(sket.yorig)) {
            alert("At lest one \"Y\" must be set before rescaling  (sket.yorig="+sket.yorig+")");
            return;
        }

        var newXorig = sket.xorig + 100;

        for (var ipnt = 0; ipnt < npnt; ipnt++) {
            var xx = sket.scale * (sket.pnt.x[ipnt] - sket.xorig);
            sket.pnt.x[ipnt] = Math.floor(newXorig + xx / sket.scale);
        }

        for (var iseg = 0; iseg < nseg; iseg++) {
            updateSegData(iseg);
        }

        sket.xorig = newXorig;

        drawSketch();
    }
}


//
// callback when "botmButton" is pressed (called by ESP.html)
//
function cmdBotm()
{
    if (wv.curMode == 0) {
        wv.mvMatrix.makeIdentity();
        wv.mvMatrix.rotate(-90, 1,0,0);
        wv.uiMatrix.load(wv.mvMatrix);
        wv.mvMatrix.makeIdentity();
        wv.scale    = 1;
        wv.sceneUpd = 1;
    } else if (wv.curMode == 7) {
        var npnt = sket.pnt.x.length;
        var nseg = sket.seg.type.length;

        if (npnt <= 1) {
            alert("Not enough data to rescale");
            return;
        } else if (sket.scale === undefined || isNaN(sket.scale)) {
            alert("At least one length or radius must be set before rescaling  (sket.scale="+sket.scale+")");
            return;
        } else if (sket.xorig === undefined || isNaN(sket.xorig)) {
            alert("At lest one \"X\" must be set before rescaling  (sket.xorig="+sket.xorig+")");
            return;
        } else if (sket.yorig === undefined || isNaN(sket.yorig)) {
            alert("At lest one \"Y\" must be set before rescaling  (sket.yorig="+sket.yorig+")");
            return;
        }

        var newYorig = sket.yorig + 100;

        for (var ipnt = 0; ipnt < npnt; ipnt++) {
            var yy = sket.scale * (sket.yorig - sket.pnt.y[ipnt]);
            sket.pnt.y[ipnt] = Math.floor(newYorig - yy / sket.scale);
        }

        for (var iseg = 0; iseg < nseg; iseg++) {
            updateSegData(iseg);
        }

        sket.yorig = newYorig;

        drawSketch();
    }
}


//
// callback when "topButton" is pressed (called by ESP.html)
//
function cmdTop()
{
    if (wv.curMode == 0) {
        wv.mvMatrix.makeIdentity();
        wv.mvMatrix.rotate(+90, 1,0,0);
        wv.uiMatrix.load(wv.mvMatrix);
        wv.mvMatrix.makeIdentity();
        wv.scale    = 1;
        wv.sceneUpd = 1;
    } else if (wv.curMode == 7) {
        var npnt = sket.pnt.x.length;
        var nseg = sket.seg.type.length;

        if (npnt <= 1) {
            alert("Not enough data to rescale");
            return;
        } else if (sket.scale === undefined || isNaN(sket.scale)) {
            alert("At least one length or radius must be set before rescaling  (sket.scale="+sket.scale+")");
            return;
        } else if (sket.xorig === undefined || isNaN(sket.xorig)) {
            alert("At lest one \"X\" must be set before rescaling  (sket.xorig="+sket.xorig+")");
            return;
        } else if (sket.yorig === undefined || isNaN(sket.yorig)) {
            alert("At lest one \"Y\" must be set before rescaling  (sket.yorig="+sket.yorig+")");
            return;
        }

        var newYorig = sket.yorig - 100;

        for (var ipnt = 0; ipnt < npnt; ipnt++) {
            var yy = sket.scale * (sket.yorig - sket.pnt.y[ipnt]);
            sket.pnt.y[ipnt] = Math.floor(newYorig - yy / sket.scale);
        }

        for (var iseg = 0; iseg < nseg; iseg++) {
            updateSegData(iseg);
        }

        sket.yorig = newYorig;

        drawSketch();
    }
}


//
// callback when "inButton" is pressed (called by ESP.html)
//
function cmdIn()
{
    if (wv.curMode == 0) {
        wv.mvMatrix.load(wv.uiMatrix);
        wv.mvMatrix.scale(2.0, 2.0, 2.0);
        wv.uiMatrix.load(wv.mvMatrix);
        wv.mvMatrix.makeIdentity();
        wv.scale   *= 2.0;
        wv.sceneUpd = 1;
    } else if (wv.curMode == 7) {
        var npnt = sket.pnt.x.length;
        var nseg = sket.seg.type.length;

        if (npnt <= 1) {
            alert("Not enough data to rescale");
            return;
        } else if (sket.scale === undefined || isNaN(sket.scale)) {
            alert("At least one length or radius must be set before rescaling  (sket.scale="+sket.scale+")");
            return;
        } else if (sket.xorig === undefined || isNaN(sket.xorig)) {
            alert("At lest one \"X\" must be set before rescaling  (sket.xorig="+sket.xorig+")");
            return;
        } else if (sket.yorig === undefined || isNaN(sket.yorig)) {
            alert("At lest one \"Y\" must be set before rescaling  (sket.yorig="+sket.yorig+")");
            return;
        }

        var newScale = 0.5 * sket.scale;

        for (var ipnt = 0; ipnt < npnt; ipnt++) {
            var xx = sket.scale * (sket.pnt.x[ipnt] - sket.xorig);
            var yy = sket.scale * (sket.yorig - sket.pnt.y[ipnt]);

            sket.pnt.x[ipnt] = Math.floor(sket.xorig + xx / newScale);
            sket.pnt.y[ipnt] = Math.floor(sket.yorig - yy / newScale);
        }

        for (var iseg = 0; iseg < nseg; iseg++) {
            sket.seg.dip[iseg] = sket.seg.dip[iseg] * sket.scale / newScale;

            updateSegData(iseg);
        }

        sket.scale = newScale;

        drawSketch();
    }
}


//
// callback when "outButton" is pressed (called by ESP.html)
//
function cmdOut()
{
    if (wv.curMode == 0) {
        wv.mvMatrix.load(wv.uiMatrix);
        wv.mvMatrix.scale(0.5, 0.5, 0.5);
        wv.uiMatrix.load(wv.mvMatrix);
        wv.mvMatrix.makeIdentity();
        wv.scale   *= 0.5;
        wv.sceneUpd = 1;
    } else if (wv.curMode == 7) {
        var npnt = sket.pnt.x.length;
        var nseg = sket.seg.type.length;

        if (npnt <= 1) {
            alert("Not enough data to rescale");
            return;
        } else if (sket.scale === undefined || isNaN(sket.scale)) {
            alert("At least one length or radius must be set before rescaling  (sket.scale="+sket.scale+")");
            return;
        } else if (sket.xorig === undefined || isNaN(sket.xorig)) {
            alert("At lest one \"X\" must be set before rescaling  (sket.xorig="+sket.xorig+")");
            return;
        } else if (sket.yorig === undefined || isNaN(sket.yorig)) {
            alert("At lest one \"Y\" must be set before rescaling  (sket.yorig="+sket.yorig+")");
            return;
        }

        var newScale = 2.0 * sket.scale;

        for (var ipnt = 0; ipnt < npnt; ipnt++) {
            var xx = sket.scale * (sket.pnt.x[ipnt] - sket.xorig);
            var yy = sket.scale * (sket.yorig - sket.pnt.y[ipnt]);

            sket.pnt.x[ipnt] = Math.floor(sket.xorig + xx / newScale);
            sket.pnt.y[ipnt] = Math.floor(sket.yorig - yy / newScale);
        }

        for (var iseg = 0; iseg < nseg; iseg++) {
            sket.seg.dip[iseg] = sket.seg.dip[iseg] * sket.scale / newScale;

            updateSegData(iseg);
        }

        sket.scale = newScale;

        drawSketch();
    }
}


//
// callback when "Design Parameters" is pressed in Tree
//
function addPmtr() {
    // alert("addPmtr()");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first");
        return;
    } else if (wv.server != "serveCSM") {
        alert("addPmtr is not implemented for "+wv.server);
        return;
    }

    // get the new name
    var name = prompt("Enter new Parameter name");
    if (name === null) {
        return;
    } else if (name.length <= 0) {
        return;
    }

    // check that name is valid
    if (name.match(/^[a-zA-Z]\w*$/) === null) {
        alert("'"+name+"' is not a valid name");
        return;
    }

    // check that the name does not exist already
    for (var ipmtr = 0; ipmtr < pmtr.length; ipmtr++) {
        if (name == pmtr[ipmtr].name) {
            alert("'"+name+"' already exists");
            return;
        }
    }

    // get the number of rows
    var nrow = prompt("Enter number of rows", "1");
    if (nrow !== null) {
        if (isNaN(nrow)) {
            alert("Number of rows must be positive integer");
            return;
        } else if (Number(nrow) < 1) {
            alert("Number of rows must be positive integer");
            return;
        } else if (Math.round(Number(nrow)) != Number(nrow)) {
            alert("Number of rows must be positive integer");
            return;
        }
    } else {
        return;
    }

    // get the number of columns
    var ncol = prompt("Enter number of columns", "1");
    if (ncol !== null) {
        if (isNaN(ncol)) {
            alert("Number of columns must be positive integer");
            return;
        } else if (Number(ncol) < 1) {
            alert("Number of columns must be positive integer");
            return;
        } else if (Math.round(Number(ncol)) != Number(ncol)) {
            alert("Number of columns must be positive integer");
            return;
        }
    } else {
        return;
    }

    // store the values locally
    var newPmtr = pmtr.length;

    pmtr[newPmtr] = new Array();

    pmtr[newPmtr].name = name;
    pmtr[newPmtr].type = 500;
    pmtr[newPmtr].nrow = nrow;
    pmtr[newPmtr].ncol = ncol;
    pmtr[newPmtr].value = new Array(nrow*ncol);

    for (var i = 0; i < nrow*ncol; i++) {
        pmtr[newPmtr].value[i] = "";
    }

    // remember info for Parameter
    wv.curPmtr = newPmtr;

    // set up editPmtrForm
    if (setupEditPmtrForm() > 0) {
        return;
    }

    // post the editPmtr form (with the addPmtr header)
    changeMode(4);
}


//
// callback when Design Parameter name is pressed in Tree
//
function editPmtr(e) {
    // alert("in editPmtr(e="+e+")");

    if (wv.curMode == 5) {
        // currently editting another Parameter, so cancel (throwing away changes)
        editPmtrCancel();
    } else if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first");
        return;
    } else if (wv.server != "serveCSM") {
        alert("editPmtr is not implemented for "+wv.server);
        return;
    }

    wv.menuEvent = e;

    // get the Tree Node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // highlight this Parameter in the Tree
    var myElem = document.getElementById("node"+inode+"col1");
    myElem.className = "currentTD";

    // get the Parameter name
    var name = myTree.name[inode].replace(/\u00a0/g, "");
    name = name.replace(/\^/g, "");

    // get the Parameter index
    var ipmtr = -1;      // 0-bias
    var jpmtr;           // 1-bias (and temp)
    for (jpmtr = 0; jpmtr < pmtr.length; jpmtr++) {
        if (pmtr[jpmtr].name.replace(/\^/g, "") == name) {
            ipmtr = jpmtr;
            break;
        }
    }
    if (ipmtr < 0) {
        alert(name+" not found");
        return;
    } else {
        jpmtr = ipmtr + 1;
    }

    // remember info for the current Parameter
    wv.curPmtr = ipmtr;

    // set up editPmtrForm
    setupEditPmtrForm();

    // post the editPmtr form (with the editPmtr header)
    changeMode(5);
}


//
// callback when "Compute Sensitivity" is pressed in editPmtrForm (called by ESP.html)
//
function compSens() {
    // alert("in compSens()");

    // disable this command if there were any changes to the Parameter
    if (numberOfPmtrChanges() > 0) {
        alert("Changes were made.  Press 'Cancel' or 'OK' first");
        return;
    }

    // get the Tree Node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // get the Parameter name
    var name = myTree.name[inode].replace(/\u00a0/g, "");
    name = name.replace(/\^/g, "");

    // get the Parameter index
    var ipmtr = -1;      // 0-bias
    var jpmtr;           // 1-bias (and temp)
    for (jpmtr = 0; jpmtr < pmtr.length; jpmtr++) {
        if (pmtr[jpmtr].name.replace(/\^/g, "") == name) {
            ipmtr = jpmtr;
            break;
        }
    }
    if (ipmtr < 0) {
        alert(name+" not found");
        return;
    } else {
        jpmtr = ipmtr + 1;
    }

    // unhighlight the first column of the Tree
    unhighlightColumn1();

    // can only compute sensitivity for a scalar
    if (pmtr[ipmtr].nrow > 1 || pmtr[ipmtr].ncol > 1) {
        alert("Sensitivity can only be computed for a scalar");
        return;
    }

    // clear any previous velocities
    wv.socketUt.send("clrVels|");
    for (jpmtr = 0; jpmtr < pmtr.length; jpmtr++) {
        pmtr[jpmtr].dot[0] = 0;
    }

    // set velocity for ipmtr
    pmtr[ipmtr].dot[0] = 1.0;
    wv.socketUt.send("setVel|"+pmtr[ipmtr].name+"|1|1|1|");
    postMessage("Computing sensitivity with respect to "+pmtr[ipmtr].name);

    // rebuild
    wv.socketUt.send("build|0|");
    wv.brchStat = 6000;

    wv.socketUt.send("getPmtrs|");
    wv.pmtrStat = 6000;

    wv.socketUt.send("getBrchs|");

    var button = document.getElementById("buildButton");
    button.onclick = null;
    button["innerHTML"] = "Re-building...";
    button.style.backgroundColor = "#FFFF3F";

    // inactivate buttons until build is done
    changeMode( 0);
    changeMode(-1);
}


//
// callback when "Set Design Velocity" is pressed in editPmtrForm (called by ESP.html)
//
function setVel() {
    // alert("in setVel()");

    // disable this command if there were any changes to the Parameter
    if (numberOfPmtrChanges() > 0) {
        alert("Changes were made.  Press 'Cancel' or 'OK' first");
        return;
    }

    // get the Tree Node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // get the Parameter name
    var name = myTree.name[inode].replace(/\u00a0/g, "");
    name = name.replace(/\^/g, "");

    // get the Parameter index
    var ipmtr = -1;      // 0-bias
    var jpmtr;           // 1-bias (and temp)
    for (jpmtr = 0; jpmtr < pmtr.length; jpmtr++) {
        if (pmtr[jpmtr].name.replace(/\^/g, "") == name) {
            ipmtr = jpmtr;
            break;
        }
    }
    if (ipmtr < 0) {
        alert(name+" not found");
        return;
    } else {
        jpmtr = ipmtr + 1;
    }

    // unhighlight the first column of the Tree
    unhighlightColumn1();

    // get each of the values
    var index  = -1;
    var nchange = 0;
    for (var irow = 1; irow <= pmtr[ipmtr].nrow; irow++) {
        for (var icol = 1; icol <= pmtr[ipmtr].ncol; icol++) {
            index++;

            // get the new value
            var newVel;
            if (pmtr[ipmtr].nrow == 1 && pmtr[ipmtr].ncol == 1) {
                newVel = prompt("Enter new Design Velocity for "+name,
                                pmtr[ipmtr].dot[index]);
            } else {
                newVel = prompt("Enter new Design Velocity for "+name+
                                "["+irow+","+icol+"]",
                                pmtr[ipmtr].dot[index]);
            }

            // make sure a valid number was entered
            if (newVel === null) {
                continue;
            } else if (isNaN(newVel)) {
                alert("Illegal number format, so Design Velocity not being changed");
                continue;
            } else if (newVel == pmtr[ipmtr].dot[index]) {
                continue;
            } else if (pmtr[ipmtr].nrow == 1 && pmtr[ipmtr].ncol == 1) {
                postMessage("Parameter '"+name+"' has new Design Velocity "+
                            newVel+" ====> Re-build is needed <====");
                nchange++;
            } else {
                postMessage("Parameter '"+name+"["+irow+","+icol+
                            "]' has new Design Velocity "+newVel+
                            " ====> Re-build is needed <====");
                nchange++;
            }

            // store the value locally
            pmtr[ipmtr].dot[index] = Number(newVel);

            // send the new Design Velocity to the server
            wv.socketUt.send("setVel|"+pmtr[ipmtr].name+"|"+irow+"|"+icol+"|"+newVel+"|");

        }
    }

    // update the UI
    if (nchange > 0) {
        wv.nchanges += nchange;

        var myElem = document.getElementById(id);
        myElem.className = "fakelinkoff";

        activateBuildButton();
    }

    // return to the WebViewer
    changeMode(0);
}


//
// callback when "Clear Design Velocities" is pressed in editPmtrForm (called by ESP.html)
//
function clrVels() {
    // alert("clrVels()");

    // disable this command if there were any changes to the Parameter
    if (numberOfPmtrChanges() > 0) {
        alert("Changes were made.  Press 'Cancel' or 'OK' first");
        return;
    } else if (wv.server != "serveCSM") {
        alert("clrVels is not implemented for "+wv.server);
        return;
    }

    // unhighlight the first column of the Tree
    unhighlightColumn1();

    // get an updated Parameter list (so that added Parameter is listed)
    wv.socketUt.send("clrVels|");

    // update the UI
    postMessage("Design Velocities have been cleared ====> Re-build is needed <====");
    activateBuildButton();

    changeMode(0);
}


//
// callback when "OK" is pressed in editPmtrForm (called by ESP.html)
//
function editPmtrOk() {
    // alert("in editPmtrOk()");

    var editPmtrForm = document.getElementById("editPmtrForm");

    var ipmtr = wv.curPmtr;
    var name  = pmtr[ipmtr].name;
    var nrow  = pmtr[ipmtr].nrow;
    var ncol  = pmtr[ipmtr].ncol;
    var irow;
    var icol;

    // make sure that all entries have valid values
    var nchange = 0;
    var index   = -1;
    for (irow = 1; irow <= pmtr[ipmtr].nrow; irow++) {
        for (icol = 1; icol <= pmtr[ipmtr].ncol; icol++) {
            index++;

            // get the new value
            var myInput = editPmtrForm["row"+irow+"col"+icol];
            var value   = myInput.value.replace(/\s/g, "");

            if (value.length <= 0) {
                alert("Entry at (row "+irow+", col "+icol+") is blank");
                return;
            } else if (isNaN(value)) {
                alert("Illegal number format at (row "+irow+", col "+icol+")");
                return;
            }
        }
    }

    // send the new Parameter to the server if in add Brch mode
    if (wv.curMode == 4) {
        var mesg = "newPmtr|"+name+"|"+nrow+"|"+ncol+"|";

        index = -1;
        for (irow = 1; irow <= nrow; irow++) {
            for (icol = 1; icol <= ncol; icol++) {
                index++;
                mesg = mesg+"|";
            }
        }

        wv.socketUt.send(mesg);
    }

    // get each of the values
    index = -1;
    for (irow = 1; irow <= pmtr[ipmtr].nrow; irow++) {
        for (icol = 1; icol <= pmtr[ipmtr].ncol; icol++) {
            index++;

            // get the new value
            var myInput = editPmtrForm["row"+irow+"col"+icol];
            var value = myInput.value.replace(/\s/g, "");

            if (value != pmtr[ipmtr].value[index]) {
                postMessage("Parameter '"+pmtr[ipmtr].name+"["+irow+","+icol+
                            "]' has been changed to "+value+
                            " ====> Re-build is needed <====");
                nchange++;

                // store the value locally
                pmtr[ipmtr].value[index] = Number(value);

                // send the new value to the server
                wv.socketUt.send("setPmtr|"+pmtr[ipmtr].name+"|"+irow+"|"+icol+"|"+value+"|");
            }
        }
    }

    // update the UI
    if (nchange > 0) {
        wv.nchanges += nchange;

        if (wv.curMode != 4) {
            var id     = wv.menuEvent["target"].id;
            var myElem = document.getElementById(id);
            myElem.className = "fakelinkoff";

        // get an updated Parameter list (so that added Pmtr is listed)
        } else {
            wv.socketUt.send("getPmtrs|");
        }

        activateBuildButton();
    }

    // unhighlight the first column of the Tree
    unhighlightColumn1();

    // return to the WebViewer
    changeMode(0);
}


//
// callback when "Cancel" is pressed in editPmtrForm (called by ESP.html)
//
function editPmtrCancel() {
    // alert("in editPmtrCancel()");

    // if we are in process of adding a Parameter, remove it now
    if (wv.curMode == 4) {
        pmtr.splice(pmtr.length-1, 1);
    }

    // unhighlight the first column of the Tree
    unhighlightColumn1();

    // return to the WebViewer
    changeMode(0);
}


//
// callback when "Branch" is pressed in Tree
// callback when "Add new Branch after this Brnch" is pressed in editBrchForm (called by ESP.html)
//
function addBrch() {
    // alert("in addBrch()");

    // this check allows one to select AddBranchAfterThisBranch
    if (wv.menuEvent !== undefined && wv.curMode == 3) {

    } else if (wv.curMode != 0) {
        alert("Changes were made.  Press 'Cancel' or 'OK' first");
        return;
    } else if (wv.server != "serveCSM") {
        alert("addBrch is not implemented for "+wv.server);
        return;
    }

    // unhighlight the first column of the Tree
    unhighlightColumn1();

    // remember
    if (wv.curBrch >= 0) {
        wv.afterBrch = wv.curBrch;
        wv.curBrch   = -1;
    } else {
        wv.afterBrch = brch.length - 1;
        wv.curBrch   = -1;
    }

    // post the addBrch form
    changeMode(1);
}


//
// callback when "OK" is pressed in addBrchForm (called by ESP.html)
//
function addBrchOk() {
    // alert("in addBrchOk()");

    // get the elements on the form
    var elements = document.getElementById("addBrchForm").elements;

    var jelem       = -1;
    for (var ielem = 0; ielem < elements.length; ielem++) {
        if (elements[ielem].checked) {
            jelem = ielem;
            break;
        }
    }

    // if nothing is selected, return to the WebViewer
    if (jelem < 0) {
        alert("Select a Branch type or press 'Cancel'");
        return;
    }

    // if something was picked, initialize the new Branch
    var newBrch = brch.length;

    brch[newBrch] = new Array();

    brch[newBrch].name = "**name_automatically_assigned**";
    brch[newBrch].type = elements[jelem].value;
    brch[newBrch].actv = 300;
    brch[newBrch].args = new Array();

    // remember info for Branch to add after
    wv.curBrch = newBrch;

    // set up editBrchForm (and abort if problem is detected)
    if (setupEditBrchForm() > 0) {
        return;
    }

    // increment the number of changes
    wv.nchanges++;

    // post the editBrch form (with the addBrch header)
    changeMode(2);
}


//
// callback when "Cancel" is pressed in addBrchForm (called by ESP.html)
//
function addBrchCancel() {
    // alert("in addBrchCancel()");

    // return to the WebViewer
    changeMode(0);
}


//
// callback when Branch name is pressed in Tree
//
function editBrch(e) {
    // alert("in editBrch(e="+e+")");

    if (wv.curMode == 3) {
        // currently editting another Branch, so cancel (throwing away changes)
        editBrchCancel();
    } else if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first");
        return;
    } else if (wv.server != "serveCSM") {
        alert("editBrch is not implemented for "+wv.server);
        return;
    }

    changeMode(0);

    wv.menuEvent = e;

    // get the Tree node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // get the Branch name
    var name = myTree.name[inode].replace(/\u00a0/g, "").replace(/>/g, "");

    // get the Branch index
    var ibrch = -1;           // 0-bias
    var jbrch;                // 1-bias (and temp)
    for (jbrch = 0; jbrch < brch.length; jbrch++) {
        if (brch[jbrch].name == name) {
            ibrch = jbrch;
            break;
        }
    }
    if (ibrch < 0) {
        alert(name+" not found");
        return;
    } else {
        jbrch = ibrch + 1;
    }

    // highlight this Branch in the Tree
    var myElem = document.getElementById("node"+inode+"col1");
    myElem.className = "currentTD";

    // highlight the parents and child of this Branch in the Tree */
    var ileft = brch[ibrch].ileft;
    var irite = brch[ibrch].irite;
    var ichld = brch[ibrch].ichld;

    for (var jnode = 0; jnode < myTree.name.length; jnode++) {
        if (ileft > 0) {
            if (myTree.name[jnode].replace(/\u00a0/g, "").replace(/>/g, "") == brch[ileft-1].name) {
                myElem = document.getElementById("node"+jnode+"col1");
                myElem.className = "parentTD";
            }
        }
        if (irite > 0) {
            if (myTree.name[jnode].replace(/\u00a0/g, "").replace(/>/g, "") == brch[irite-1].name) {
                myElem = document.getElementById("node"+jnode+"col1");
                myElem.className = "parentTD";
            }
        }
        if (ichld > 0) {
            if (myTree.name[jnode].replace(/\u00a0/g, "").replace(/>/g, "") == brch[ichld-1].name) {
                myElem = document.getElementById("node"+jnode+"col1");
                myElem.className = "childTD";
            }
        }
    }

    // remember info for the current Branch
    wv.curBrch   = ibrch;

    // set up editBrchForm
    if (setupEditBrchForm() > 0) {
        alert("This Branch cannot be editted");
        unhighlightColumn1();
        changeMode(0);
        return;
    }

    // post the editBrch form (with the editBrch header)
    changeMode(3);
}


//
// callback when "Delete this Branch" is pressed in editBrchForm (called by ESP.html)
//
function delBrch() {
    // alert("in delBrch()");

    // disable this command if there were any chnges to the Branch
    if (numberOfBrchChanges() > 0) {
        alert("Changes were made.  Press 'Cancel' or 'OK' first");
        return;
    }

    var ibrch = -1;
    var name  = "";

    // get the Tree node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // unhighlight the first column of the Tree
    unhighlightColumn1();

    // if this was called by pressing "Delete last Branch", adjust inode
    //    to point to last visible node (as if "Delete this Branch"
    //    was pressed on the last Branch)
    if (inode == 2) {
        inode = myTree.child[inode];

        while (myTree.next[inode] >= 0) {
            inode = myTree.next[inode];
        }
    }

    // get the Branch name
    name = myTree.name[inode].replace(/\u00a0/g, "").replace(/>/g, "");

    // get the Branch index
    for (var jbrch = 0; jbrch < brch.length; jbrch++) {
        if (brch[jbrch].name == name) {
            ibrch = jbrch + 1;
        }
    }
    if (ibrch <= 0) {
        alert(name+" not found");
        return;
    }

    // send the message to the server
    wv.socketUt.send("delBrch|"+ibrch+"|");

    // hide the Tree node that was just updated
    var element = myTree.document.getElementById("node"+inode);
    element.style.display = "none";

    // get an updated version of the Branches
    wv.brchStat = 0;

    // update the UI
    postMessage("Deleting Branch "+name+" ====> Re-build is needed <====");
    activateBuildButton();

    // return to the WebViewer
    changeMode(0);
}


//
// callback when "Add Attribute" is pressed in editBrchForm  (called by ESP.html)
//
function addAttr() {
    // alert("in addAttr()");

    // disable this command if there were any chnges to the Branch
    if (numberOfBrchChanges() > 0) {
        alert("Changes were made.  Press 'Cancel' or 'OK' first");
        return;
    }

    // get the Tree node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // get the Branch name
    var name = myTree.name[inode].replace(/\u00a0/g, "").replace(/>/g, "");

    // get the Branch index
    var ibrch = -1;           // 0-bias
    var jbrch;                // 1-bias (and temp)
    for (jbrch = 0; jbrch < brch.length; jbrch++) {
        if (brch[jbrch].name == name) {
            ibrch = jbrch;
            break;
        }
    }
    if (ibrch < 0) {
        alert(name+" not found");
        return;
    } else {
        jbrch = ibrch + 1;
    }

    // unhighlight the first column of the Tree
    unhighlightColumn1();

    // get the Attribute name and value
    var aname  = prompt("Enter Attribute name");
    if (aname === null) {
        return;
    } else if (aname.length <= 0) {
        return;
    }
    var avalue = prompt("Enter Attribute value");
    if (avalue === null) {
        return;
    } else if (avalue.length <= 0) {
        return;
    }

    // send the new value to the server
    var mesg = "setAttr|"+jbrch+"|"+aname+"|"+avalue+"|";

    wv.socketUt.send(mesg);

    // update the UI
    postMessage("Adding attribute '"+aname+"' (with value "+avalue+") to "+
                name+" ====> Re-build is needed <====");
    activateBuildButton();

    // get an updated version of the Branches
    wv.brchStat = 0;

    // return to  the WebViewer
    changeMode(0);
}


//
// callback when "Enter Sketcher" is pressed in editBrchForm (called by ESP.html)
//
function enterSketcher() {
    // alert("enterSketcher()");

    sket.ibrch = -1;
    var name   = "";

    // get the Tree node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // get the Branch name
    name = myTree.name[inode].replace(/\u00a0/g, "").replace(/>/g, "");

    // get the Branch index
    for (var jbrch = 0; jbrch < brch.length; jbrch++) {
        if (brch[jbrch].name == name) {
            sket.ibrch = jbrch + 1;
        }
    }
    if (sket.ibrch <= 0) {
        alert(name+" not found");
        return;
    }

    // unhighlight the first column of the Tree
    unhighlightColumn1();

    // send the message to the server
    wv.socketUt.send("loadSketch|"+sket.ibrch+"|");

    // inactivate buttons until build is done
    changeMode(-1);

    // initialize the Sketcher
    initializeSketch();
}


//
// callback when "Build to the Branch" is pressed in editBrchForm (called by ESP.html)
//
function buildTo() {
    // alert("buildTo()");

    // disable this command if there were any chnges to the Branch
    if (numberOfBrchChanges() > 0) {
        alert("Changes were made.  Press 'Cancel' or 'OK' first");
        return;
    }

    var ibrch = -1;
    var name  = "";

    // get the Tree node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // get the Branch name
    name = myTree.name[inode].replace(/\u00a0/g, "").replace(/>/g, "");

    // get the Branch index
    for (var jbrch = 0; jbrch < brch.length; jbrch++) {
        if (brch[jbrch].name == name) {
            ibrch = jbrch + 1;
        }
    }
    if (ibrch <= 0) {
        alert(name+" not found");
        return;
    }

    // unhighlight the first column of the Tree
    unhighlightColumn1();

    // send the message to the server
    wv.socketUt.send("build|"+ibrch+"|");

    // update the UI
    postMessage("Re-building only to "+name+"...");

    var button = document.getElementById("buildButton");
    button.onclick = null;
    button["innerHTML"] = "Re-building...";
    button.style.backgroundColor = "#FFFF3F";

    // inactivate buttons until build is done
    changeMode(-1);
}


//
// callback when "OK" button is pressed in editBrchForm (called by ESP.html)
//
function editBrchOk() {
    // alert("in editBrchOk()");

    var editBrchForm = document.getElementById("editBrchForm");

    var ibrch = wv.curBrch;
    var jbrch = ibrch + 1;

    var brchName = editBrchForm.brchName.value.replace(/\s/g, "");
    if (brchName.length <= 0) {
        alert("Name cannot be blank");
        return;
    }

    var mesg;
    var output;
    var nchange = 0;

    if (wv.curMode == 2) {
        mesg = "newBrch|"+(wv.afterBrch+1)+"|"+brch[ibrch].type+"|";
    } else {
        mesg = "setBrch|"+jbrch+"|";
    }

    // make sure that name does not contain a space or ">"
    var newBrchName = editBrchForm.brchName.value;

    if (newBrchName.indexOf(" ") >= 0) {
        alert("Changed name '"+newBrchName+"' cannot contain a space");
        return;
    } else if (newBrchName.indexOf(">") >= 0) {
        alert("Changed name '"+newBrchName+"' cannot contain '>'");
        return;
    }

    if (newBrchName != brch[ibrch].name) {
        // make sure that name does not start with "Brch_"
        if (newBrchName.substring(0,5) == "Brch_") {
            alert("Changed name '"+newBrchName+
                  "' cannot begin with 'Brch_'");
            return;
        }

        // make sure that name does not already exist
        for (var kbrch = 0; kbrch < brch.length; kbrch++) {
            if (brch[kbrch].name == newBrchName) {
                alert("Name '"+newBrchName+"' already exists");
                return;
            }
        }

        // store the value locally
        postMessage("Changing name '"+brch[ibrch].name+"' to '"+
                    newBrchName+"' ====> Re-build is needed <====");
        brch[ibrch].name = newBrchName;

        nchange++;
    }

    if (wv.curMode == 3) {
        mesg = mesg + newBrchName + "|";
    }

    // update the activity
    var newActivity = editBrchForm.activity.value;
    if (brch[ibrch].actv != "none"     &&
        brch[ibrch].actv != newActivity  ) {

        // store the value locally
        brch[ibrch].actv = newActivity;
        if (brch[ibrch].actv == 300) {
            postMessage("Activating Branch "+newBrchName+
                        " ====> Re-build is needed <====");
        } else {
            postMessage("Suppressing Branch "+newBrchName+
                        " ====> Re-build is needed <====");
        }

        // get an updated version of the Branches
        wv.brchStat = 0;

        nchange++;
    }

    if (wv.curMode == 2) {

    } else if (newActivity == 300) {
        mesg = mesg + "active|";
    } else {
        mesg = mesg + "suppressed|";
    }

    // update any necessary arguments
    var prev_value = "dum";
    for (var iarg = 0; iarg < wv.numArgs; iarg++) {
        if        (iarg == 0) {
            var name  = document.getElementById("argName1").firstChild["data"];
            var value = editBrchForm.            argValu1.value.replace(/\s/g, "");
        } else if (iarg == 1) {
            var name  = document.getElementById("argName2").firstChild["data"];
            var value = editBrchForm.            argValu2.value.replace(/\s/g, "");
        } else if (iarg == 2) {
            var name  = document.getElementById("argName3").firstChild["data"];
            var value = editBrchForm.            argValu3.value.replace(/\s/g, "");
        } else if (iarg == 3) {
            var name  = document.getElementById("argName4").firstChild["data"];
            var value = editBrchForm.            argValu4.value.replace(/\s/g, "");
        } else if (iarg == 4) {
            var name  = document.getElementById("argName5").firstChild["data"];
            var value = editBrchForm.            argValu5.value.replace(/\s/g, "");
        } else if (iarg == 5) {
            var name  = document.getElementById("argName6").firstChild["data"];
            var value = editBrchForm.            argValu6.value.replace(/\s/g, "");
        } else if (iarg == 6) {
            var name  = document.getElementById("argName7").firstChild["data"];
            var value = editBrchForm.            argValu7.value.replace(/\s/g, "");
        } else if (iarg == 7) {
            var name  = document.getElementById("argName8").firstChild["data"];
            var value = editBrchForm.            argValu8.value.replace(/\s/g, "");
        } else if (iarg == 8) {
            var name  = document.getElementById("argName9").firstChild["data"];
            var value = editBrchForm.            argValu9.value.replace(/\s/g, "");
        }

        // check for blanks (allowed for some Branch types)
        if (value.length <= 0) {
            if (brch[ibrch].type == "fillet"  ||
                brch[ibrch].type == "chamfer" ||
                brch[ibrch].type == "udparg"  ||
                brch[ibrch].type == "udprim"  ||
                brch[ibrch].type == "select"    ) {
                if (confirm(name+" is blank.  Continue?") !== true) {
                    return;
                }
            } else {
                alert(name+" should not be blank");
                return;
            }
        }

        // make sure non-blank does not follow blank
        if (value.length > 0 && prev_value.length == 0) {
            alert(name+" should not be blank (follows blank)");
            return;
        } else {
            prev_value = value;
        }

        if (name.charAt(0) != "$" || value.length <= 0) {
            output =       value;
        } else {
            output = "$" + value;
        }

        // check for bars
        if (value.indexOf("|") >= 0) {
            alert(name+" should not contain a bar");
            return;
        }

        // check if there were any changes
        if (output != brch[ibrch].args[iarg]) {
            if (wv.curMode == 3) {
                postMessage("Changing "+name+" from '"+brch[ibrch].args[iarg]+
                            "' to '"+output+
                            "' ====> Re-build is needed <====");
            }

            brch[ibrch].args[iarg] = output;
            nchange++;
        }
        mesg = mesg + output + "|";
    }

    // put rest of semicolon at end of mesg
    for (iarg = wv.numArgs; iarg < 9; iarg++) {
        mesg = mesg + "|";
    }

    if (wv.curMode == 3) {
        mesg = mesg + "||";
    }

    if (nchange > 0 || wv.numArgs == 0) {
        wv.nchanges += nchange;

        // send the new values to the server
        wv.socketUt.send(mesg);

        // get an updated Branch list (so that added Branch is listed)
        if (wv.curMode == 2) {
            wv.socketUt.send("getBrchs|");
        }

        // update the UI
        if (wv.curMode == 3) {
            var id     = wv.menuEvent["target"].id;
            var myElem = document.getElementById(id);
            myElem.className = "fakelinkoff";
        }

        postMessage("Branch (type="+brch[ibrch].type+") has been added"+
                "  ====> Re-build is needed <====");

        activateBuildButton();
    } else {
        alert("no changes were made");
    }

    // unhighlight the first column of the Tree
    unhighlightColumn1();

    // return to the WebViewer
    changeMode(0);
}


//
// callback when "Cancel" is pressed in editBrchForm (called by ESP.html)
//
function editBrchCancel() {
    // alert("in editBrchCancel()");

    // if we are in process of adding a Branch, remove it now
    if (wv.curMode == 2) {
        brch.splice(brch.length-1, 1);
    }

    // unhighlight the first column of the Tree
    unhighlightColumn1();

    // return to the WebViewer
    changeMode(0);
}


//
// callback when Attribute name is pressed in Tree
//
function editAttr(e) {
    // alert("in editAttr(e="+e+")");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first");
        return;
    } else if (wv.server != "serveCSM") {
        alert("editBrch is not implemented for "+wv.server);
        return;
    }

    wv.menuEvent = e;

    // get the Tree node
    var id    = wv.menuEvent["target"].id;
    var inode = Number(id.substring(4,id.length-4));

    // get the Attribute name and value
    var aname  = myTree.name[inode].replace(/\u00a0/g, "");
    var avalue = myTree.prop1[inode];

    // get the Branch name
    inode = myTree.parent[inode];

    var name = myTree.name[inode].replace(/\u00a0/g, "").replace(/>/g, "");

    // get the Branch index
    var ibrch = -1;           // 0-bias
    var jbrch;                // 1-bias (and temp)
    for (jbrch = 0; jbrch < brch.length; jbrch++) {
        if (brch[jbrch].name == name) {
            ibrch = jbrch;
            break;
        }
    }
    if (ibrch < 0) {
        alert(name+" not found");
        return;
    } else {
        jbrch = ibrch + 1;
    }

    // get the new Attribute value
    var anew = prompt("Enter new value for '"+aname+"'", avalue);
    if (anew === null) {
        return;
    } else if (anew.length <= 0) {
        return;
    } else if (anew == avalue) {
        return;
    }

    // send the new value to the server
    var mesg = "setAttr|"+jbrch+"|"+aname+"|"+anew+"|";

    wv.socketUt.send(mesg);

    // update the UI
    postMessage("Changing attribute '"+aname+"' to "+
                anew+" ====> Re-build is needed <====");
    activateBuildButton();

    // get an updated version of the Branches
    wv.brchStat = 0;
}


//
// callback when any mouse is pressed in canvas (when wv.curMode==0)
//
function getMouseDown0(e)
{
    if (!e) var e = event;

    wv.startX   =  e.clientX - wv.offLeft             - 1;
    wv.startY   = -e.clientY + wv.offTop  + wv.height + 1;

    wv.dragging = true;
    wv.button   = e.button;

                    wv.modifier  = 0;
    if (e.shiftKey) wv.modifier |= 1;
    if (e.altKey  ) wv.modifier |= 2;
    if (e.ctrlKey ) wv.modifier |= 4;
}


//
// callback when the mouse moves in canvas (when wv.curMode==0)
//
function getMouseMove0(e)
{
    if (!e) var e = event;

    wv.cursorX  =  e.clientX - wv.offLeft             - 1;
    wv.cursorY  = -e.clientY + wv.offTop  + wv.height + 1;

                    wv.modifier  = 0;
    if (e.shiftKey) wv.modifier |= 1;
    if (e.altKey  ) wv.modifier |= 2;
    if (e.ctrlKey ) wv.modifier |= 4;
}


//
// callback when the mouse is released in canvas (when wv.curMode==0)
//
function getMouseUp0(e)
{
    wv.dragging = false;
}


//
// callback when the mouse wheel is rolled in canvas (when wv.curMode==0)
//
function getMouseRoll0(e)
{
    if (e) {

        // zoom in
        if        (e.deltaY > 0) {
            wv.mvMatrix.scale(1.1, 1.1, 1.1);
            wv.scale *= 1.1;
            wv.sceneUpd = 1;

        // zoom out
        } else if (e.deltaY < 0) {
            wv.mvMatrix.scale(0.9, 0.9, 0.9);
            wv.scale *= 0.9;
            wv.sceneUpd = 1;
        }
    }
}


//
// callback when the mouse leaves the canvas (when wv.curMode==0)
//
function mouseLeftCanvas0(e)
{
    if (wv.dragging) {
        wv.dragging = false;
    }
}


//
// callback when any mouse is pressed in Sketcher (when wv.curMode==7)
//
function getMouseDown7(e)
{
    if (!e) var e = event;

    wv.startX   = e.clientX - wv.offLeft7 - 1;
    wv.startY   = e.clientY - wv.offTop7  - 1;

    wv.button   = e.button;

                    wv.modifier  = 0;
    if (e.shiftKey) wv.modifier |= 1;
    if (e.altKey  ) wv.modifier |= 2;
    if (e.ctrlKey ) wv.modifier |= 4;

    // set default key presses that are bound to mouse clicks in Sketcher
    if        (sket.mode == 1) {
        var evnt = new Object();
        evnt.keyCode  = 0;
        evnt.charCode = 108;   // "l"

        getKeyPress(evnt);
    } else if (sket.mode == 2) {
        var evnt = new Object();
        evnt.keyCode  = 0;
        evnt.charCode = 108;   // "l"

        getKeyPress(evnt);
    } else if (sket.mode == 3) {
        wv.cursorX = wv.startX;
        wv.cursorY = wv.startY;

        sket.movingPoint = getClosestSketchPoint();
    } else if (sket.mode == 4) {
        var evnt = new Object();
        evnt.keyCode  = 0;
        evnt.charCode = 124;   // "w"

        getKeyPress(evnt);
    } else if (sket.mode == 5) {
        var evnt = new Object();
        evnt.keyCode  = 0;
        evnt.charCode = 100;   // "d"

        getKeyPress(evnt);
    }
}


//
// callback when the mouse moves in Sketcher (when wv.curMode==7)
//
function getMouseMove7(e)
{
    if (!e) var e = event;

    wv.cursorX  = e.clientX - wv.offLeft7 - 1;
    wv.cursorY  = e.clientY - wv.offTop7  - 1;

                    wv.modifier  = 0;
    if (e.shiftKey) wv.modifier |= 1;
    if (e.altKey  ) wv.modifier |= 2;
    if (e.ctrlKey ) wv.modifier |= 4;

    // if sket.mode==1, then the "proposed" new Segment
    //    is taken care of in drawSketch
    if (sket.mode == 1) {
        drawSketch();

    // if sket.mode==3, then move the point at (wv.startX, wv.startY)
    //    to the current location
    } else if (sket.mode == 3) {
        if (sket.movingPoint >= 0) {
            sket.pnt.x[sket.movingPoint] = wv.cursorX;
            sket.pnt.y[sket.movingPoint] = wv.cursorY;

            for (var iseg = 0; iseg < sket.seg.type.length; iseg++) {
                updateSegData(iseg);
            }

            drawSketch();
        }

    // if sket.mode==4 or 5, then the "proposed" new W or D constraint
    //    is taken care of in drawSketch
    } else if (sket.mode == 4 || sket.mode == 5) {
        drawSketch();

    // if sket.mode==2, then we need to use the cursor location to update
    //    the dip for the last Segment
    } else if (sket.mode == 2) {
        var iseg = sket.seg.type.length - 1;

        var xa = sket.pnt.x[sket.seg.ibeg[iseg]];
        var ya = sket.pnt.y[sket.seg.ibeg[iseg]];
        var xb = sket.pnt.x[sket.seg.iend[iseg]];
        var yb = sket.pnt.y[sket.seg.iend[iseg]];

        // put the cursor on the arc
        var xe  = wv.cursorX;
        var ye  = wv.cursorY;

        var D = (ya - ye) * (xb - xe) - (yb - ye) * (xa - xe);
        if (Math.abs(D) > 1e-6) {
            var s = ((xb - xa) * (xb - xe) - (ya - yb) * (yb - ye)) / D;

            var xc = (xa + xe + s * (ya - ye)) / 2;
            var yc = (ya + ye + s * (xe - xa)) / 2;

            var R = Math.sqrt((xc-xa) * (xc-xa) + (yc-ya) * (yc-ya));
            var L = Math.sqrt((xb-xa) * (xb-xa) + (yb-ya) * (yb-ya));

            if ((xe-xa)*(yb-ya) > (ye-ya)*(xb-xa)) {
                if ((xb-xa)*(yc-ya) > (yb-ya)*(xc-xa)) {
                    sket.seg.dip[iseg] = -R + Math.sqrt(R*R - L*L/4);
                } else {
                    sket.seg.dip[iseg] = -R - Math.sqrt(R*R - L*L/4);
                }
            } else {
                if ((xb-xa)*(yc-ya) > (yb-ya)*(xc-xa)) {
                    sket.seg.dip[iseg] = +R + Math.sqrt(R*R - L*L/4);
                } else {
                    sket.seg.dip[iseg] = +R - Math.sqrt(R*R - L*L/4);
                }
            }

            updateSegData(iseg);
        } else {
            sket.seg.dip[ iseg] = 0;
            sket.seg.xm[  iseg] = (xa + xb) / 2;
            sket.seg.ym[  iseg] = (ya + yb) / 2;
            sket.seg.xc[  iseg] = 0;
            sket.seg.yc[  iseg] = 0;
            sket.seg.rad[ iseg] = 0;
            sket.seg.tbeg[iseg] = 0;
            sket.seg.tend[iseg] = 0;
        }

        drawSketch();
    }
}


//
// callback when the mouse is released in Sketcher (when wv.curMode==7)
//
function getMouseUp7(e)
{
    sket.movingPoint = -1;
}


//
// callback when a key is pressed
//
function getKeyPress(e)
{
//    if (!e) var e = event;

    // if <esc> was pressed, return to base mode
    if (e.charCode == 0 && e.keyCode == 27) {
        activateBuildButton();
        changeMode(0);
    }

    // only process keyPress if using canvas or Sketcher
    if (wv.curMode != 0 && wv.curMode != 7) {
        return;
    }

    wv.keyPress = e.charCode;
    wv.keyCode  = e.keyCode;

                    wv.modifier  = 0;
    if (e.shiftKey) wv.modifier |= 1;
    if (e.altKey  ) wv.modifier |= 2;
    if (e.ctrlKey ) wv.modifier |= 4;

    // make sure that Firefox does not stop websockets if <esc> is pressed
    if (e.charCode == 0) {
        e.preventDefault();
    }

    // if in the Sketcher, process the key press now
    if (wv.curMode == 7) {
        sketchKeyPress();
    }
}


//
// callback when an arrow... or shift is pressed (needed for Chrome)
//
function getKeyDown(e) {
    if (e.charCode == 0) {
        if (e.keyCode == 33 ||          // PgUp
            e.keyCode == 34 ||          // PgDn
            e.keyCode == 36 ||          // Home
            e.keyCode == 37 ||          // Left
            e.keyCode == 38 ||          // Up
            e.keyCode == 39 ||          // Right
            e.keyCode == 40   ) {       // Down
            wv.keyCode  = e.keyCode;
            wv.keyPress = 0;
        } else if (e.keyCode == 16) {   // Shift
            wv.modifier = 1;
        }
    }
}


//
// callback when a shift is released (needed for Chrome)
//
function getKeyUp(e) {

    if (e.charCode == 0 && e.keyCode == 16) {
        wv.modifier = 0;
    }
}


//
// callback when the mouse is pressed in key window
//
function setKeyLimits(e)
{

    // get new limits
    var templo = prompt("Enter new lower limit", wv.loLimit);
    if (isNaN(templo)) {
        alert("Lower limit must be a number");
        return;
    }

    var tempup = prompt("Enter new upper limit", wv.upLimit);
    if (isNaN(tempup)) {
        alert("Upper limit must be a number");
        return;
    }

    if (templo != wv.loLimit || tempup != wv.uplimit) {
        wv.loLimit = templo;
        wv.upLimit = tempup;

        // send the limits back to the server
        wv.socketUt.send("setLims|"+wv.loLimit+"|"+wv.upLimit+"|");
    }
}


//
// callback to toggle Viz property
//
function toggleViz(e) {
    // alert("in toggleViz(e="+e+")");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first");
        return;
    }

    // get the Tree Node
    var inode = e["target"].id.substring(4);
    inode     = inode.substring(0,inode.length-4);
    inode     = Number(inode);

    // toggle the Viz property
    if        (myTree.valu1[inode] == "off") {
        myTree.prop(inode, 1, "on");
    } else if (myTree.valu1[inode] == "on") {
        myTree.prop(inode, 1, "off");
    } else {
        alert("Illegal Viz property:"+myTree.valu1[inode]);
        return;
    }
}


//
// callback to toggle Grd property
//
function toggleGrd(e) {
    // alert("in toggleGrd(e="+e+")");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first");
        return;
    }

    // get the Tree Node
    var inode = e["target"].id.substring(4);
    inode     = inode.substring(0,inode.length-4);
    inode     = Number(inode);

    // toggle the Grd property
    if        (myTree.valu2[inode] == "off") {
        myTree.prop(inode, 2, "on");
    } else if (myTree.valu2[inode] == "on") {
        myTree.prop(inode, 2, "off");
    } else {
        alert("Illegal Grd property:"+myTree.valu2[inode]);
        return;
    }
}


//
// callback to toggle Trn property
//
function toggleTrn(e) {
    // alert("in toggleTrn(e="+e+")");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first");
        return;
    }

    // get the Tree Node
    var inode = e["target"].id.substring(4);
    inode     = inode.substring(0,inode.length-4);
    inode     = Number(inode);

    // toggle the Trn property
    if        (myTree.valu3[inode] == "off") {
        myTree.prop(inode, 3, "on");
    } else if (myTree.valu3[inode] == "on") {
        myTree.prop(inode, 3, "off");
    } else {
        alert("Illegal Trn property:"+myTree.valu3[inode]);
        return;
    }
}


//
// callback to toggle Ori property
//
function toggleOri(e) {
    // alert("in toggleOri(e="+e+")");

    if (wv.curMode != 0) {
        alert("Command disabled.  Press 'Cancel' or 'OK' first");
        return;
    }

    // get the Tree Node
    var inode = e["target"].id.substring(4);
    inode     = inode.substring(0,inode.length-4);
    inode     = Number(inode);

    // toggle the Ori property
    if        (myTree.valu3[inode] == "off") {
        myTree.prop(inode, 3, "on");
    } else if (myTree.valu3[inode] == "on") {
        myTree.prop(inode, 3, "off");
    } else {
        alert("Illegal Ori property:"+myTree.valu3[inode]);
        return;
    }
}


//
// Initialize a Sketch
//
function initializeSketch()
{
//    alert("initializeSketch()");

    if (sket) {
        sket.basePoint = undefined;
        sket.skbegX    = undefined;
        sket.skbegY    = undefined;
        sket.skbegZ    = undefined;
        sket.halo      =  5;      // pixels for determining "closeness" while drawing
        wv.keyPress    = -1;      // last key pressed
        wv.keyCode     = -1;
        wv.button      = -1;      // button pressed
        wv.modifier    =  0;      // modifier (shift,alt,ctrl) bitflag

        // set up Sketcher offsets
        var sketcher = document.getElementById("sketcher");
        wv.offTop7   = sketcher.offsetTop;
        wv.offLeft7  = sketcher.offsetLeft;

        // initialize the arrays
        sket.pnt        = {};
        sket.pnt.x      = [];
        sket.pnt.y      = [];
        sket.pnt.lbl    = [];

        sket.seg        = {};
        sket.seg.type   = [];
        sket.seg.ibeg   = [];
        sket.seg.iend   = [];
        sket.seg.xm     = [];
        sket.seg.ym     = [];
        sket.seg.lbl    = [];

        sket.seg.dip    = [];
        sket.seg.xc     = [];
        sket.seg.yc     = [];
        sket.seg.rad    = [];
        sket.seg.tbeg   = [];
        sket.seg.tend   = [];

        sket.var        = {};
        sket.var.name   = [];
        sket.var.value  = [];

        sket.con        = {};
        sket.con.type   = [];
        sket.con.index1 = [];
        sket.con.index2 = [];
        sket.con.value  = [];

        // mapping between "physical" and "screen" coordinates
        //   x_physical = sket.scale * (x_screen - sket.xorig)
        //   y_physical = sket.scale * (sket.yorig - y_screen)

        // commands that set sket.xorig: "x"
        // commands that set sket.yorig: "y"
        // commands that set sket.scale: "l", "r", "w", or "d"
        //                               "x" if another "x" exists
        //                               "y" if another "y" exists

        // reset scaling info
        sket.scale = undefined;
        sket.xorig = undefined;
        sket.yorig = undefined;

        // initialize undo
        sket.undo = undefined;

        // set mode to initializing
        sket.mode = 0;
    } else {
        alert("ERROR:: sket does not exist");
    }

    // draw the current Sketch
    drawSketch();
}


//
// load a Sketch from a JSON string sent from serveCSM
//
function loadSketch(begs, vars, cons, segs)
{
//    alert("loadSketch()");
//    alert("   begs="+begs);
//    alert("   vars="+vars);
//    alert("   cons="+cons);
//    alert("   segs="+segs);

    // break up the inputs
    var begsList = begs.split(";");
    var varsList = vars.split(";");
    var consList = cons.split(";");
    var segsList = segs.split(";");

    // get the number of points, segments, variables, and constraints
    var npnt = (varsList.length - 1) / 3;
    var ncon = (consList.length - 1) / 4;
    var nseg = (segsList.length - 1) / 3;

    // store the base expressions from the SKBEG statement
    sket.skbegX = begsList[0];
    sket.skbegY = begsList[2];
    sket.skbegZ = begsList[4];

    // find the extrema of the coordinates
    var xmin = Number(begsList[1]);
    var xmax = Number(begsList[1]);
    var ymin = Number(begsList[3]);
    var ymax = Number(begsList[3]);

    for (var ipnt = 1; ipnt < npnt; ipnt++) {
        xmin = Math.min(xmin, Number(varsList[3*ipnt  ]));
        xmax = Math.max(xmax, Number(varsList[3*ipnt  ]));
        ymin = Math.min(ymin, Number(varsList[3*ipnt+1]));
        ymax = Math.max(ymax, Number(varsList[3*ipnt+1]));
    }

    // get the size of the canvas
    var canvas = document.getElementById("sketcher");
    var width  = canvas.clientWidth;
    var height = canvas.clientHeight;

    if (npnt > 1) {
        var test1 = (xmax - xmin) / width;
        var test2 = (ymax - ymin) / height;

        // set up sizing so that Sketch fills middle 50 percent of the window
        sket.scale = 2 * Math.max(test1, test2);
        sket.xorig = (width  - (xmin + xmax) / sket.scale) / 2;
        sket.yorig = (height + (ymin + ymax) / sket.scale) / 2;

    // not enough data to set scales, so guess something base upon the SKBEG statement
    } else {
        sket.scale = 0.01;
        sket.xorig = width  / 2 - begsList[1] / sket.scale;
        sket.yorig = height / 2 + begsList[3] / sket.scale;
    }

    // create the points
    sket.pnt.x   = [];
    sket.pnt.y   = [];
    sket.pnt.lbl = [];

    for (ipnt = 0; ipnt < npnt; ipnt++) {
        sket.pnt.x[  ipnt] = Math.floor(sket.xorig + Number(varsList[3*ipnt  ]) / sket.scale);
        sket.pnt.y[  ipnt] = Math.floor(sket.yorig - Number(varsList[3*ipnt+1]) / sket.scale);
        sket.pnt.lbl[ipnt] = "";
    }

    // overwrite the first point by the begsList
    if (npnt > 0) {
        sket.pnt.x[0] = Math.floor(sket.xorig + Number(begsList[1]) / sket.scale);
        sket.pnt.y[0] = Math.floor(sket.yorig - Number(begsList[3]) / sket.scale);
    }

    // create the segments
    sket.seg.type = [];
    sket.seg.ibeg = [];
    sket.seg.iend = [];
    sket.seg.xm   = [];
    sket.seg.ym   = [];
    sket.seg.lbl  = [];

    sket.seg.dip  = [];
    sket.seg.xc   = [];
    sket.seg.yc   = [];
    sket.seg.rad  = [];
    sket.seg.tbeg = [];
    sket.seg.tend = [];

    for (var iseg = 0; iseg < nseg; iseg++) {
        sket.seg.type[iseg] =        segsList[3*iseg  ];
        sket.seg.ibeg[iseg] = Number(segsList[3*iseg+1]) - 1;
        sket.seg.iend[iseg] = Number(segsList[3*iseg+2]) - 1;
        sket.seg.lbl[ iseg] = "";

        sket.seg.dip[ iseg] = 0;
        sket.seg.xc[  iseg] = 0;
        sket.seg.yc[  iseg] = 0;
        sket.seg.rad[ iseg] = 0;
        sket.seg.tbeg[iseg] = 0;
        sket.seg.tend[iseg] = 0;

        // update the segment if an arc
        if (sket.seg.type[iseg] == "C") {
            if (iseg < npnt-1) {
                sket.seg.dip[iseg] = Number(varsList[3*iseg+5]) / sket.scale;
            } else {
                sket.seg.dip[iseg] = Number(varsList[       2]) / sket.scale;
            }
        }

        updateSegData(iseg);
    }

    // create the variables
    sket.var.name  = [];
    sket.var.value = [];

    var nvar = 0;
    for (var ivar = 0; ivar < npnt; ivar++) {
        sket.var.name[ nvar] = sprintf("::x[%]", ivar+1);
        sket.var.value[nvar] = varsList[3*ivar  ];
        nvar++;

        sket.var.name[ nvar] = sprintf("::y[%]", ivar+1);
        sket.var.value[nvar] = varsList[3*ivar+1];
        nvar++;

        if (varsList.length > 0 && Number(varsList[3*ivar+2]) != 0) {
            sket.var.name[ nvar] = sprintf("::d[%]", ivar+1);
            sket.var.value[nvar] = varsList[3*ivar+2];
            nvar++;
        }
    }

    // create the constraints
    sket.con.type   = [];
    sket.con.index1 = [];
    sket.con.index2 = [];
    sket.con.value  = [];

    for (var icon = 0; icon < ncon; icon++) {
        sket.con.type[  icon] =        consList[4*icon  ];
        sket.con.index1[icon] = Number(consList[4*icon+1]) - 1;
        sket.con.index2[icon] = Number(consList[4*icon+2]) - 1;
        sket.con.value[ icon] =        consList[4*icon+3];

        if (sket.con.type[icon] == "W" || sket.con.type[icon] == "D") {
            // not associated with Point or Segment
        } else if (sket.con.index2[icon] < 0) {
            sket.con.index2[icon] += 1;
            sket.pnt.lbl[sket.con.index1[icon]] += sket.con.type[icon];
        } else {
            sket.seg.lbl[sket.con.index1[icon]] += sket.con.type[icon];
        }
    }

    // if there were no constraints, create X and Y constraints at first Point
    if (ncon == 0) {
        sket.pnt.x[  0] = Math.floor(sket.xorig + Number(begsList[1]) / sket.scale);
        sket.pnt.y[  0] = Math.floor(sket.yorig - Number(begsList[3]) / sket.scale);
        sket.pnt.lbl[0] = "XY";

        sket.var.name[ 0] = "::x[1]";
        sket.var.value[0] = sket.pnt.x[0];

        sket.var.name[ 1] = "::y[1]";
        sket.var.value[1] = sket.pnt.y[0];

        sket.con.type[  0] = "X";
        sket.con.index1[0] = 0;
        sket.con.index2[0] = -1;
        sket.con.value[ 0] = sket.skbegX;

        sket.con.type[  1] = "Y";
        sket.con.index1[1] = 0;
        sket.con.index2[1] = -1;
        sket.con.value[ 1] = sket.skbegY;
    }

    // set the mode depending on status of Sketch
    if (sket.seg.type.length == 0) {
        sket.mode = 1;
    } else if (sket.seg.iend[sket.seg.iend.length-1] != 0) {
        sket.mode = 1;
    } else {
        sket.mode = 3;
    }

    // draw the current Sketch
    drawSketch();
}


//
// process a key press in the Sketcher
//
function sketchKeyPress()
{
//    alert("sketchKeyPress");

    var canvas  = document.getElementById("sketcher");
    var context = canvas.getContext("2d");

    var npnt = sket.pnt.x.length;
    var nseg = sket.seg.type.length;
    var nvar = sket.var.name.length;
    var ncon = sket.con.type.length;

    var myKeyPress = String.fromCharCode(wv.keyPress);

    // 'ctrl-h' --- home
    if ((wv.keyPress == 104 && wv.modifier == 4 && wv.keyCode == 0) ||
        (wv.keyPress ==   8 && wv.modifier == 4 && wv.keyCode == 8)   ) {
        cmdHome();
        return;

    // 'ctrl-i' - zoom in (same as <PgUp> without shift)
    } else if ((wv.keyPress == 105 && wv.modifier == 4 && wv.keyCode ==  0) ||
               (wv.keyPress ==   9 && wv.modifier == 4 && wv.keyCode ==  9)   ) {
        cmdIn();
        return;

    // 'ctrl-o' - zoom out (same as <PgDn> without shift)
    } else if ((wv.keyPress == 111 && wv.modifier == 4 && wv.keyCode ==  0) ||
               (wv.keyPress ==  15 && wv.modifier == 4 && wv.keyCode == 15)   ) {
        cmdOut();
        return;

    // 'ctrl-r' - riteside view
    } else if ((wv.keyPress == 114 && wv.modifier == 4 && wv.keyCode ==  0) ||
               (wv.keyPress ==  18 && wv.modifier == 4 && wv.keyCode == 18)   ) {
        cmdRite();
        return;

    // 'ctrl-l' - leftside view
    } else if ((wv.keyPress == 108 && wv.modifier == 4 && wv.keyCode ==  0) ||
               (wv.keyPress ==  12 && wv.modifier == 4 && wv.keyCode == 12)   ) {
        cmdLeft();
        return;

    // 'ctrl-t' - top view
    } else if ((wv.keyPress == 116 && wv.modifier == 4 && wv.keyCode ==  0) ||
               (wv.keyPress ==  20 && wv.modifier == 4 && wv.keyCode == 20)   ) {
        cmdTop();
        return;

    // 'ctrl-b' - bottom view
    } else if ((wv.keyPress ==  98 && wv.modifier == 4 && wv.keyCode ==  0) ||
               (wv.keyPress ==   2 && wv.modifier == 4 && wv.keyCode ==  2)   ) {
        cmdBotm();
        return;

    // sket.mode 1 options ('l', 'c', 's', 'l', or 'o' to leave sketch open)
    } else if (sket.mode == 1) {
        saveSketchUndo();

        // determine the coordinates (and constraints) for this placement
        var xnew = wv.cursorX;
        var ynew = wv.cursorY;
        var cnew = "";

        for (var ipnt = npnt-1; ipnt >= 0; ipnt--) {
            if (Math.abs(wv.cursorX-sket.pnt.x[ipnt]) < sket.halo) {
                xnew = sket.pnt.x[ipnt];
                if (ipnt == npnt-1) {
                    cnew = "V";
                }
                break;
            }
        }
        for (ipnt = npnt-1; ipnt >= 0; ipnt--) {
            if (Math.abs(wv.cursorY-sket.pnt.y[ipnt]) < sket.halo) {
                ynew = sket.pnt.y[ipnt];
                if (ipnt == npnt-1) {
                    cnew = "H";
                }
                break;
            }
        }

        // "l" --- add a line Segment
        if (myKeyPress == "l" || myKeyPress == "L") {
            var ibeg = npnt - 1;
            var iend = 0;

            // if this Point does not close the Sketch, add a new Point and its variables
            if (Math.abs(sket.pnt.x[0]-xnew) > sket.halo ||
                Math.abs(sket.pnt.y[0]-ynew) > sket.halo   ) {
                iend = npnt;
                sket.pnt.x[  iend] = xnew;
                sket.pnt.y[  iend] = ynew;
                sket.pnt.lbl[iend] = "";

                sket.var.name[ nvar  ] = sprintf("::x[%]", npnt+1);
                sket.var.value[nvar  ] = sprintf("%",    xnew  );

                sket.var.name[ nvar+1] = sprintf("::y[%]", npnt+1);
                sket.var.value[nvar+1] = sprintf("%",    ynew  );
            }

            // add the new Segment
            sket.seg.type[nseg] = "L";
            sket.seg.ibeg[nseg] = ibeg;
            sket.seg.iend[nseg] = iend;
            sket.seg.xm[  nseg] = (sket.pnt.x[ibeg] + sket.pnt.x[iend]) / 2;
            sket.seg.ym[  nseg] = (sket.pnt.y[ibeg] + sket.pnt.y[iend]) / 2;
            sket.seg.lbl[ nseg] = "";

            sket.seg.dip[ nseg] = 0;
            sket.seg.xc[  nseg] = 0;
            sket.seg.yc[  nseg] = 0;
            sket.seg.rad[ nseg] = 0;
            sket.seg.tbeg[nseg] = 0;
            sket.seg.tend[nseg] = 0;

            // add a new "horizontal" or "vertical" constraint
            if        (cnew == "H") {
                sket.con.type[  ncon] = "H";
                sket.con.index1[ncon] = ibeg;
                sket.con.index2[ncon] = iend;
                sket.con.value[ ncon] = 0;

                sket.seg.lbl[nseg] += "H";
            } else if (cnew == "V") {
                sket.con.type[  ncon] = "V";
                sket.con.index1[ncon] = ibeg;
                sket.con.index2[ncon] = iend;
                sket.con.value[ ncon] = 0;

                sket.seg.lbl[nseg] += "V";
            }

            // if Sketch is closed, post a message and swith mode
            if (iend == 0) {
                sket.mode = 3;
            }

        // "c" --- add a circular arc
        } else if (myKeyPress == "c" || myKeyPress == "C") {
            var ibeg = npnt - 1;
            var iend = 0;

            // if this Point does not close the Sketch, add a new Point and its variables
            if (Math.abs(sket.pnt.x[0]-xnew) > sket.halo ||
                Math.abs(sket.pnt.y[0]-ynew) > sket.halo) {
                iend          = npnt;
                sket.pnt.x[  iend] = xnew;
                sket.pnt.y[  iend] = ynew;
                sket.pnt.lbl[iend] = "";

                sket.var.name[ nvar  ] = sprintf("::x[%]", npnt+1);
                sket.var.value[nvar  ] = sprintf("%",    xnew  );

                sket.var.name[ nvar+1] = sprintf("::y[%]", npnt+1);
                sket.var.value[nvar+1] = sprintf("%",    ynew  );
            }

            // add the new Segment
            sket.seg.type[nseg] = "C";
            sket.seg.ibeg[nseg] = ibeg;
            sket.seg.iend[nseg] = iend;
            sket.seg.xm[  nseg] = (sket.pnt.x[ibeg] + sket.pnt.x[iend]) / 2;
            sket.seg.ym[  nseg] = (sket.pnt.y[ibeg] + sket.pnt.y[iend]) / 2;
            sket.seg.lbl[ nseg] = "";

            sket.seg.dip[ nseg] = 0;
            sket.seg.xc[  nseg] = 0;
            sket.seg.yc[  nseg] = 0;
            sket.seg.rad[ nseg] = 0;
            sket.seg.tbeg[nseg] = 0;
            sket.seg.tend[nseg] = 0;

            nvar = sket.var.name.length;

            sket.var.name[ nvar] = sprintf("::d[%]", sket.seg.type.length);
            sket.var.value[nvar] = "0";

            sket.mode = 2;

        // "s" --- add spline Point
        } else if (myKeyPress == "s" || myKeyPress == "S") {
            var ibeg = npnt - 1;
            var iend = 0;

            // if this Point does not close the Sketch, add a new Point and its variables
            if (Math.abs(sket.pnt.x[0]-xnew) > sket.halo ||
                Math.abs(sket.pnt.y[0]-ynew) > sket.halo   ) {
                iend = npnt;
                sket.pnt.x[  iend] = xnew;
                sket.pnt.y[  iend] = ynew;
                sket.pnt.lbl[iend] = "";

                sket.var.name[ nvar  ] = sprintf("::x[%]", npnt+1);
                sket.var.value[nvar  ] = sprintf("%",    xnew  );

                sket.var.name[ nvar+1] = sprintf("::y[%]", npnt+1);
                sket.var.value[nvar+1] = sprintf("%",    ynew  );
            }

            // add the new Segment
            sket.seg.type[nseg] = "S";
            sket.seg.ibeg[nseg] = ibeg;
            sket.seg.iend[nseg] = iend;
            sket.seg.xm[  nseg] = (sket.pnt.x[ibeg] + sket.pnt.x[iend]) / 2;
            sket.seg.ym[  nseg] = (sket.pnt.y[ibeg] + sket.pnt.y[iend]) / 2;
            sket.seg.lbl[ nseg] = "";

            sket.seg.dip[ nseg] = 0;
            sket.seg.xc[  nseg] = 0;
            sket.seg.yc[  nseg] = 0;
            sket.seg.rad[ nseg] = 0;
            sket.seg.tbeg[nseg] = 0;
            sket.seg.tend[nseg] = 0;

            // if Sketch is closed, post a message and swith mode
            if (iend == 0) {
                sket.mode = 3;
            }

        // "b" --- add Bezier Point
        } else if (myKeyPress == "b" || myKeyPress == "B") {
            var ibeg = npnt - 1;
            var iend = 0;

            // if this Point does not close the Sketch, add a new Point and its variables
            if (Math.abs(sket.pnt.x[0]-xnew) > sket.halo ||
                Math.abs(sket.pnt.y[0]-ynew) > sket.halo   ) {
                iend = npnt;
                sket.pnt.x[  iend] = xnew;
                sket.pnt.y[  iend] = ynew;
                sket.pnt.lbl[iend] = "";

                sket.var.name[ nvar  ] = sprintf("::x[%]", npnt+1);
                sket.var.value[nvar  ] = sprintf("%",    xnew  );

                sket.var.name[ nvar+1] = sprintf("::y[%]", npnt+1);
                sket.var.value[nvar+1] = sprintf("%",    ynew  );
            }

            // add the new Segment
            sket.seg.type[nseg] = "B";
            sket.seg.ibeg[nseg] = ibeg;
            sket.seg.iend[nseg] = iend;
            sket.seg.xm[  nseg] = (sket.pnt.x[ibeg] + sket.pnt.x[iend]) / 2;
            sket.seg.ym[  nseg] = (sket.pnt.y[ibeg] + sket.pnt.y[iend]) / 2;
            sket.seg.lbl[ nseg] = "";

            sket.seg.dip[ nseg] = 0;
            sket.seg.xc[  nseg] = 0;
            sket.seg.yc[  nseg] = 0;
            sket.seg.rad[ nseg] = 0;
            sket.seg.tbeg[nseg] = 0;
            sket.seg.tend[nseg] = 0;

            // add a new "horizontal" or "vertical" constraint
            if        (cnew == "H") {
                sket.con.type[  ncon] = "H";
                sket.con.index1[ncon] = ibeg;
                sket.con.index2[ncon] = iend;
                sket.con.value[ ncon] = 0;

                sket.seg.lbl[nseg] += "H";
            } else if (cnew == "V") {
                sket.con.type[  ncon] = "V";
                sket.con.index1[ncon] = ibeg;
                sket.con.index2[ncon] = iend;
                sket.con.value[ ncon] = 0;

                sket.seg.lbl[nseg] += "V";
            }

            // if Sketch is closed, post a message and swith mode
            if (iend == 0) {
                sket.mode = 3;
            }

        // "o"  --- end (open) Sketch
        } else if (myKeyPress == "o") {
            sket.mode = 3;

        } else {
            postMessage("Valid options are 'l', 'c', 's', 'b', and 'o'");
        }

    // sket.mode 2 option (any keyPress or mouseDown)
    } else if (sket.mode == 2) {

        // <any> --- set curvature
        if (sket.seg.iend[sket.seg.iend.length-1] == 0) {
            sket.mode = 3;
        } else {
            sket.mode = 1;
        }

    // sket.mode 3 options
    // if at a Point:          'x', 'y', 'p', 't', 'a', 'w', 'd' or '<' to delete
    // if at s cirarc Segment: 'r', 's', 'x', 'y',               or '<' to delete
    // if at          Segment: 'h', 'v', 'i', 'l',               or '<' to delete
    } else if (sket.mode == 3 || sket.mode == 6) {
        saveSketchUndo();

        // "x" --- set X coordinate of nearest Point
        if (myKeyPress == "x" || myKeyPress == "X") {
            sket.mode = 3;

            var ibest = getClosestSketchPoint();
            var xvalue;
            if (ibest < 0) {
                alert("No Point found");
            } else if (sket.pnt.lbl[ibest].indexOf("X") >= 0) {
                for (var icon = 0; icon < ncon; icon++) {
                    if (sket.con.type[icon] == "X" && sket.con.index1[icon] == ibest) {
                        xvalue = prompt("Enter x value: ", sket.con.value[icon]);
                        if (xvalue !== null) {
                            sket.con.value[icon] = xvalue;
                        }
                        break;
                    }
                }
            } else {
                if (sket.scale === undefined && sket.xorig === undefined) {
                    xvalue = prompt("Enter x value: ");
                    if (xvalue !== null) {
                        for (var icon = 0; icon < ncon; icon++) {
                            if (sket.con.type[icon] == "X" && sket.con.index1[icon] != ibest) {
                                sket.scale = (xvalue - sket.con.value[icon]) / (sket.pnt.x[ibest] - sket.pnt.x[sket.con.index1[icon]]);
                                sket.xorig = sket.pnt.x[sket.con.index1[icon]] - sket.con.value[icon] / sket.scale;
                                break;
                            }
                        }
                    }
                } else if (sket.xorig === undefined) {
                    xvalue = prompt("Enter x value: ");
                    if (xvalue !== null) {
                        sket.xorig = sket.pnt.x[ibest] + xvalue / sket.scale;
                    }
                } else if (sket.scale === undefined) {
                    xvalue = prompt("Enter x value: ");
                    if (xvalue !== null) {
                        for (var icon = 0; icon < ncon; icon++) {
                            if (sket.con.type[icon] == "X" && sket.con.index1[icon] != ibest) {
                                sket.scale = (xvalue - sket.con.value[icon]) / (sket.pnt.x[ibest] - sket.xorig);
                                break;
                            }
                        }
                    }
                } else {
                    xvalue = prompt("Enter x value: ", sket.scale*(sket.pnt.x[ibest]-sket.xorig));
                }

                if (xvalue !== null) {
                    sket.con.type[  ncon] = "X";
                    sket.con.index1[ncon] = ibest;
                    sket.con.index2[ncon] = -1;
                    sket.con.value[ ncon] = xvalue;

                    sket.pnt.lbl[ibest] += "X";
                }
            }

        // "y" --- set Y coordinate of nearest Point
        } else if (myKeyPress == "y" || myKeyPress == "Y") {
            sket.mode = 3;

            var ibest = getClosestSketchPoint();
            var yvalue;
            if (ibest < 0) {
                alert("No Point found");
            } else if (sket.pnt.lbl[ibest].indexOf("Y") >= 0) {
                for (var icon = 0; icon < ncon; icon++) {
                    if (sket.con.type[icon] == "Y" && sket.con.index1[icon] == ibest) {
                        yvalue = prompt("Enter y value: ", sket.con.value[icon]);
                        if (yvalue !== null) {
                            sket.con.value[icon] = yvalue;
                        }
                        break;
                    }
                }
            } else {
                if (sket.scale === undefined && sket.yorig === undefined) {
                    yvalue = prompt("Enter y value: ");
                    if (yvalue !== null) {
                        for (var icon = 0; icon < ncon; icon++) {
                            if (sket.con.type[icon] == "Y" && sket.con.index1[icon] != ibest) {
                                sket.scale = (sket.con.value[icon] - yvalue) / (sket.pnt.y[ibest] - sket.pnt.y[sket.con.index1[icon]]);
                                sket.yorig = sket.pnt.y[sket.con.index1[icon]] + sket.con.value[icon] / sket.scale;
                                break;
                            }
                        }
                    }
                } else if (sket.yorig === undefined) {
                    yvalue = prompt("Enter y value: ");
                    if (yvalue !== null) {
                        sket.yorig = sket.pnt.y[ibest] - yvalue / sket.scale;
                    }
                } else if (sket.scale === undefined) {
                    yvalue = prompt("Enter y value: ");
                    if (yvalue !== null) {
                        for (var icon = 0; icon < ncon; icon++) {
                            if (sket.con.type[icon] == "Y" && sket.con.index1[icon] != ibest) {
                                sket.scale = (sket.con.value[icon] - yvalue) / (sket.pnt.y[ibest] - sket.yorig);
                                break;
                            }
                        }
                    }
                } else {
                    yvalue = prompt("Enter y value: ", sket.scale*(sket.yorig-sket.pnt.y[ibest]));
                }

                if (yvalue !== null) {
                    sket.con.type[  ncon] = "Y";
                    sket.con.index1[ncon] = ibest;
                    sket.con.index2[ncon] = -1;
                    sket.con.value[ ncon] = yvalue;

                    sket.pnt.lbl[ibest] += "Y";
                }
            }

        // "p" --- perpendicularity
        } else if (myKeyPress == "p" || myKeyPress == "P") {
            sket.mode = 3;

            var ibest = getClosestSketchPoint();
            if (ibest < 0) {
                alert("No Point found");
            } else if (sket.pnt.lbl[ibest].indexOf("P") >= 0) {
                alert("'P' constraint already at this Point");
            } else if (sket.pnt.lbl[ibest].indexOf("T") >= 0) {
                alert("'T' constraint already at this Point");
            } else if (sket.pnt.lbl[ibest].indexOf("A") >= 0) {
                alert("'A' constraint already at this Point");
            } else {
                sket.con.type[  ncon] = "P";
                sket.con.index1[ncon] = ibest;
                sket.con.index2[ncon] = -1;
                sket.con.value[ ncon] = 0;

                sket.pnt.lbl[ibest] += "P";
            }

        // "t" --- tangency
        } else if (myKeyPress == "t" || myKeyPress == "T") {
            sket.mode = 3;

            var ibest = getClosestSketchPoint();
            if (ibest < 0) {
                alert("No Point found");
            } else if (sket.pnt.lbl[ibest].indexOf("P") >= 0) {
                alert("'P' constraint already at this Point");
            } else if (sket.pnt.lbl[ibest].indexOf("T") >= 0) {
                alert("'T' constraint already at this Point");
            } else if (sket.pnt.lbl[ibest].indexOf("A") >= 0) {
                alert("'A' constraint already at this Point");
            } else {
                sket.con.type[  ncon] = "T";
                sket.con.index1[ncon] = ibest;
                sket.con.index2[ncon] = -1;
                sket.con.value[ ncon] = 0;

                sket.pnt.lbl[ibest] += "T";
            }

        // "a" --- angle (positive to left)
        } else if (myKeyPress == 'a' || myKeyPress == "A") {
            sket.mode = 3;

            var ibest = getClosestSketchPoint();
            var avalue;
            if (ibest < 0) {
                alert("No Point found");
            } else if (sket.pnt.lbl[ibest].indexOf("P") >= 0) {
                alert("'P' constraint already at this Point");
            } else if (sket.pnt.lbl[ibest].indexOf("T") >= 0) {
                alert("'T' constraint already at this Point");
            } else if (sket.pnt.lbl[ibest].indexOf("A") >= 0) {
                for (var icon = 0; icon < ncon; icon++) {
                    if (sket.con.type[icon] == "A" && sket.con.index1[icon] == ibest) {
                        avalue = prompt("Enter angle (deg): ", sket.con.value[icon]);
                        if (avalue !== null) {
                            sket.con.value[icon] = avalue;
                        }
                        break;
                    }
                }
            } else {
                var ibeg = -1;
                var iend = -1;
                for (var iseg = 0; iseg < sket.seg.type.length; iseg++) {
                    if (sket.seg.iend[iseg] == ibest) {
                        ibeg = sket.seg.ibeg[iseg];
                    }
                    if (sket.seg.ibeg[iseg] == ibest) {
                        iend = sket.seg.iend[iseg];
                    }
                }

                if (ibeg < 0 || iend < 0) {
                    alert("Angle cannot be specified at endpoints");
                    return;                        
                }

                var ang1 = Math.atan2(sket.pnt.y[ibest]-sket.pnt.y[ibeg ],
                                      sket.pnt.x[ibest]-sket.pnt.x[ibeg ]) * 180 / Math.PI;
                var ang2 = Math.atan2(sket.pnt.y[iend ]-sket.pnt.y[ibest],
                                      sket.pnt.x[iend ]-sket.pnt.x[ibest]) * 180 / Math.PI;
                avalue = ang1 - ang2;
                while (avalue < -180) {
                    avalue += 360;
                }
                while (avalue > 180) {
                    avalue -= 360;
                }

                avalue = prompt("Enter angle (deg): ", avalue);

                if (avalue !== null) {
                    sket.con.type[  ncon] = "A";
                    sket.con.index1[ncon] = ibest;
                    sket.con.index2[ncon] = -1;
                    sket.con.value[ ncon] = avalue;

                    sket.pnt.lbl[ibest] += "A";
                }
            }

        // "h" --- Segment is horizontal
        } else if (myKeyPress == "h" || myKeyPress == "H") {
            sket.mode = 3;

            var ibest = getClosestSketchSegment();
            if (ibest < 0) {
                alert("No Segment found");
            } else if (sket.seg.lbl[ibest].indexOf("H") >= 0) {
                alert("'H' constraint already on this Segment");
            } else if (sket.seg.lbl[ibest].indexOf("V") >= 0) {
                alert("'V' constraint already on this Segment");
            } else if (sket.seg.lbl[ibest].indexOf("I") >= 0) {
                alert("'I' constraint already on this Segment");
            } else if (Math.abs(sket.seg.dip[ibest]) > 1e-3) {
                alert("'H' constraint cannot be set on a Cirarc");
            } else {
                var ibeg  = sket.seg.ibeg[ibest];
                var iend  = sket.seg.iend[ibest];

                sket.con.type[  ncon] = "H";
                sket.con.index1[ncon] = ibeg;
                sket.con.index2[ncon] = iend;
                sket.con.value[ ncon] = 0;

                sket.seg.lbl[ibest] += "H";
            }

        // "v" --- Segment is vertical
        } else if (myKeyPress == "v" || myKeyPress == "V") {
            sket.mode = 3;

            var ibest = getClosestSketchSegment();
            if (ibest < 0) {
                alert("No Segment found");
            } else if (sket.seg.lbl[ibest].indexOf("V") >= 0) {
                alert("'V' constraint already on this Segment");
            } else if (sket.seg.lbl[ibest].indexOf("H") >= 0) {
                alert("'H' constraint already on this Segment");
            } else if (sket.seg.lbl[ibest].indexOf("I") >= 0) {
                alert("'I' constraint already on this Segment");
            } else if (Math.abs(sket.seg.dip[ibest]) > 1e-3) {
                alert("'V' constraint cannot be set on a Cirarc");
            } else {
                var ibeg  = sket.seg.ibeg[ibest];
                var iend  = sket.seg.iend[ibest];

                sket.con.type[  ncon] = "V";
                sket.con.index1[ncon] = ibeg;
                sket.con.index2[ncon] = iend;
                sket.con.value[ ncon] = 0;

                sket.seg.lbl[ibest] += "V";
            }

        // "i" --- inclination of Segment
        } else if (myKeyPress == "i" || myKeyPress == "I") {
            sket.mode = 3;

            var ibest = getClosestSketchSegment();
            var ivalue;
            if (ibest < 0) {
                alert("No Segment found");
            } else if (sket.seg.lbl[ibest].indexOf("H") >= 0) {
                alert("'H' constraint already on this Segment");
            } else if (sket.seg.lbl[ibest].indexOf("V") >= 0) {
                alert("'V' constraint already on this Segment");
            } else if (Math.abs(sket.seg.dip[ibest]) > 1e-3) {
                alert("'I' constraint cannot be set on a Cirarc");
            } else if (sket.seg.lbl[ibest].indexOf("I") >= 0) {
                for (var icon = 0; icon < ncon; icon++) {
                    if (sket.con.type[ icon] == "I" && sket.con.index1[icon] == ibest) {
                        ivalue = prompt("Enter inclination angle (deg): ", sket.con.value[icon]);
                        if (ivalue !== null) {
                            sket.con.value[icon] = ivalue;
                        }
                        break;
                    }
                }
            } else {
                var ibeg  = sket.seg.ibeg[ibest];
                var iend  = sket.seg.iend[ibest];

                var iprompt = Math.atan2(sket.pnt.y[ibeg]-sket.pnt.y[iend], sket.pnt.x[iend]-sket.pnt.x[ibeg]) * 180/Math.PI;

                ivalue = prompt("Enter inclination angle (deg): ", iprompt);
                if (ivalue !== null) {
                    sket.con.type[  ncon] = "I";
                    sket.con.index1[ncon] = ibeg;
                    sket.con.index2[ncon] = iend;
                    sket.con.value[ ncon] = ivalue;

                    sket.seg.lbl[ibest] += "I";
                }
            }

        // "l" --- length of Segment
        } else if (myKeyPress == "l" || myKeyPress == "L") {
            sket.mode = 3;

            var ibest = getClosestSketchSegment();
            var lvalue;
            if (ibest < 0) {
                alert("No Segment found");
            } else if (Math.abs(sket.seg.dip[ibest]) > 1e-3) {
                alert("'L' constraint cannot be set on a Cirarc");
            } else if (sket.seg.lbl[ibest].indexOf("L") >= 0) {
                for (var icon = 0; icon < ncon; icon++) {
                    if (sket.con.type[ icon] == "L" && sket.con.index1[icon] == ibest) {
                        lvalue = prompt("Enter length: ", sket.con.value[icon]);
                        if (lvalue !== null) {
                            sket.con.value[icon] = lvalue;
                        }
                        break;
                    }
                }
            } else {
                var ibeg  = sket.seg.ibeg[ibest];
                var iend  = sket.seg.iend[ibest];

                var xbeg = sket.pnt.x[ibeg];
                var ybeg = sket.pnt.y[ibeg];
                var xend = sket.pnt.x[iend];
                var yend = sket.pnt.y[iend];
                var len  = Math.sqrt(Math.pow(xend-xbeg, 2) + Math.pow(yend-ybeg, 2));

                if (sket.scale === undefined) {
                    lvalue = prompt("Enter length: ");
                    if (lvalue !== null) {
                        sket.scale = lvalue / len;
                    }
                } else {
                    lvalue = prompt("Enter length: ", sket.scale*len);
                }

                if (lvalue !== null) {
                    sket.con.type[  ncon] = "L";
                    sket.con.index1[ncon] = ibeg;
                    sket.con.index2[ncon] = iend;
                    sket.con.value[ ncon] = lvalue;

                    sket.seg.lbl[ibest] += "L";
                }
            }

        // "r" --- radius of Segment
        } else if (myKeyPress == "r" || myKeyPress == "R") {
            sket.mode = 3;

            var ibest = getClosestSketchSegment();
            var rvalue;
            if (ibest < 0) {
                alert("No Cirarc found");
            } else if (Math.abs(sket.seg.dip[ibest]) < 1e-3) {
                alert("'R' constraint cannot be set for a Linseg");
            } else if (sket.seg.lbl[ibest].indexOf("R") >= 0) {
                for (var icon = 0; icon < ncon; icon++) {
                    if (sket.con.type[ icon] == "R" && sket.con.index1[icon] == ibest) {
                        rvalue = prompt("Enter radius: ", sket.con.value[icon]);
                        if (rvalue !== null) {
                            sket.con.value[icon] = rvalue;
                        }
                        break;
                    }
                }
            } else {
                var ibeg = sket.seg.ibeg[ibest];
                var iend = sket.seg.iend[ibest];

                var xa = sket.pnt.x[ibeg];
                var ya = sket.pnt.y[ibeg];
                var xb = sket.pnt.x[iend];
                var yb = sket.pnt.y[iend];

                var d  = sket.seg.dip[ibest];

                var L   = Math.sqrt((xb-xa) * (xb-xa) + (yb-ya) * (yb-ya));
                var rad = (L*L + 4*d*d) / (8*d);

                if (sket.scale === undefined) {
                    rvalue = prompt("Enter radius: ");
                    if (rvalue !== null) {
                        sket.scale = rvalue / rad;
                    }
                } else {
                    rvalue = prompt("Enter radius: ", sket.scale*rad);
                }

                if (rvalue !== null) {
                    sket.con.type[  ncon] = "R";
                    sket.con.index1[ncon] = ibeg;
                    sket.con.index2[ncon] = iend;
                    sket.con.value[ ncon] = rvalue;

                    sket.seg.lbl[ibest] += "R";
                }
            }

        // "s" --- sweep angle of Segment
        } else if (myKeyPress == "s" || myKeyPress == "S") {
            sket.mode = 3;

            var ibest = getClosestSketchSegment();
            var avalue;
            if (ibest < 0) {
                alert("No Cirarc found");
            } else if (Math.abs(sket.seg.dip[ibest]) < 1e-3) {
                alert("'S' constraint cannot be set for a Linseg");
            } else if (sket.seg.lbl[ibest].indexOf("S") >= 0) {
                for (var icon = 0; icon < ncon; icon++) {
                    if (sket.con.type[icon] == "S" && sket.con.index1[icon] == ibest) {
                        avalue = prompt("Enter sweep angle (deg): ", sket.con.value[icon]);
                        if (avalue !== null) {
                            sket.con.value[icon] = avalue;
                        }
                        break;
                    }
                }
            } else {
                var ibeg = sket.seg.ibeg[ibest];
                var iend = sket.seg.iend[ibest];

                avalue = prompt("Enter sweep angle (deg): ");

                if (avalue !== null) {
                    sket.con.type[  ncon] = "S";
                    sket.con.index1[ncon] = ibeg;
                    sket.con.index2[ncon] = iend;
                    sket.con.value[ ncon] = avalue;

                    sket.seg.lbl[ibest] += "S";
                }
            }

        // "w" --- width (dx) between Points
        } else if (myKeyPress == "w" || myKeyPress == "W") {
            var bestPnt = getClosestSketchPoint();
            var bestCon = getClosestSketchConstraint();

            if (bestCon !== undefined) {
                var xmid = (  sket.pnt.x[sket.con.index1[bestCon]]
                            + sket.pnt.x[sket.con.index2[bestCon]]) / 2;
                var ymid = (2*sket.pnt.y[sket.con.index1[bestCon]]
                            + sket.pnt.y[sket.con.index2[bestCon]]) / 3;

                var dpnt = Math.abs(sket.pnt.x[bestPnt] - wv.cursorX)
                         + Math.abs(sket.pnt.y[bestPnt] - wv.cursorY);
                var dcon = Math.abs(xmid                - wv.cursorX)
                         + Math.abs(ymid                - wv.cursorY);

                if (dcon < dpnt) {
                    var wvalue = prompt("Enter width: ", sket.con.value[bestCon]);
                    if (wvalue !== null) {
                        sket.con.value[bestCon] = wvalue;
                    }

                    sket.bestPoint = undefined;
                    sket.mode      = 3;
                } else {
                    sket.basePoint = bestPnt;
                    sket.mode      = 4;
                }
            } else {
                sket.basePoint = bestPnt;
                sket.mode      = 4;
            }

        // "d" --- depth (dy) between Points
        } else if (myKeyPress == "d" || myKeyPress == "D") {
            var bestPnt = getClosestSketchPoint();
            var bestCon = getClosestSketchConstraint();

            if (bestCon !== undefined) {
                var xmid = (2*sket.pnt.x[sket.con.index1[bestCon]]
                            + sket.pnt.x[sket.con.index2[bestCon]]) / 3;
                var ymid = (  sket.pnt.y[sket.con.index1[bestCon]]
                            + sket.pnt.y[sket.con.index2[bestCon]]) / 2;

                var dpnt = Math.abs(sket.pnt.x[bestPnt] - wv.cursorX)
                         + Math.abs(sket.pnt.y[bestPnt] - wv.cursorY);
                var dcon = Math.abs(xmid                - wv.cursorX)
                         + Math.abs(ymid                - wv.cursorY);

                if (dcon < dpnt) {
                    var dvalue = prompt("Enter depth: ", sket.con.value[bestCon]);
                    if (dvalue !== null) {
                        sket.con.value[bestCon] = dvalue;
                    }

                    sket.bestPoint = undefined;
                    sket.mode      = 3;
                } else {
                    sket.basePoint = bestPnt;
                    sket.mode      = 5;
                }
            } else {
                sket.basePoint = bestPnt;
                sket.mode      = 5;
            }

        // "<" --- remove all constraints from nearset Point or Segment
        } else if (myKeyPress == "<") {
            sket.mode = 3;

            var ibest = getClosestSketchPoint();
            var jbest = getClosestSketchSegment();
            var kbest = getClosestSketchConstraint();

            var dibest = 999999;
            if (ibest >= 0) {
                dibest = Math.abs(wv.cursorX - sket.pnt.x[ibest])
                       + Math.abs(wv.cursorY - sket.pnt.y[ibest]);
            }
            var djbest = 999999;
            if (jbest >= 0) {
                djbest = Math.abs(wv.cursorX - sket.seg.xm[jbest])
                       + Math.abs(wv.cursorY - sket.seg.ym[jbest]);
            }
            var dkbest = 999999;
            if (kbest !== undefined) {
                if        (sket.con.type[kbest] == "W") {
                    var xmid = (  sket.pnt.x[sket.con.index1[kbest]]
                                + sket.pnt.x[sket.con.index2[kbest]]) / 2;
                    var ymid = (2*sket.pnt.y[sket.con.index1[kbest]]
                                + sket.pnt.y[sket.con.index2[kbest]]) / 3;

                    dkbest = Math.abs(wv.cursorX - xmid)
                           + Math.abs(wv.cursorY - ymid);
                } else if (sket.con.type[kbest] == "D") {
                    var xmid = (2*sket.pnt.x[sket.con.index1[kbest]]
                                + sket.pnt.x[sket.con.index2[kbest]]) / 3;
                    var ymid = (  sket.pnt.y[sket.con.index1[kbest]]
                                + sket.pnt.y[sket.con.index2[kbest]]) / 2;

                    dkbest = Math.abs(wv.cursorX - xmid)
                           + Math.abs(wv.cursorY - ymid);
                }
            }

            // remove constraints at Point ibest
            if (ibest >= 0 && dibest < djbest && dibest < dkbest) {
                postMessage("removing constraints at Point "+ibest);
                sket.pnt.lbl[ibest] = "";

                for (var icon = sket.con.type.length-1; icon >= 0; icon--) {
                    if (sket.con.index1[icon] == ibest &&
                        sket.con.index2[icon] <  0       ) {
                        sket.con.type.splice(  icon, 1);
                        sket.con.index1.splice(icon, 1);
                        sket.con.index2.splice(icon, 1);
                        sket.con.value.splice( icon, 1);
                    }
                }

            // remove constraints at Segment jbest
            } else if (jbest >= 0 && djbest < dkbest) {
                postMessage("removing constraints at Segment "+jbest);
                sket.seg.lbl[jbest] = "";

                for (var icon = sket.con.type.length-1; icon >= 0; icon--) {
                    if (sket.con.index1[icon] == jbest &&
                        sket.con.index2[icon] >= 0       ) {
                        sket.con.type.splice(  icon, 1);
                        sket.con.index1.splice(icon, 1);
                        sket.con.index2.splice(icon, 1);
                        sket.con.value.splice( icon, 1);
                    }
                }

            // remove H or W constraint
            } else {
                postMessage("removing constraint "+kbest);

                sket.con.type.splice(  kbest, 1);
                sket.con.index1.splice(kbest, 1);
                sket.con.index2.splice(kbest, 1);
                sket.con.value.splice( kbest, 1);
            }

        // "?" --- examine Point or Segment
        } else if (myKeyPress == "?") {
            var ibest = getClosestSketchPoint();
            var jbest = getClosestSketchSegment();

            var dibest = 999999;
            if (ibest >= 0) {
                dibest = Math.abs(wv.cursorX - sket.pnt.x[ibest])
                       + Math.abs(wv.cursorY - sket.pnt.y[ibest]);
            }
            var djbest = 999999;
            if (jbest >= 0) {
                djbest = Math.abs(wv.cursorX - sket.seg.xm[jbest])
                       + Math.abs(wv.cursorY - sket.seg.ym[jbest]);
            }

            // examine Point ibest
            if (ibest >= 0 && dibest < djbest) {
                if (sket.pnt.lbl[ibest].length == 0) {
                    postMessage("Point "+(ibest+1)+" has no constraints");
                } else {
                    postMessage("Point "+(ibest+1)+" has constraints: "+sket.pnt.lbl[ibest]);
                }

            // examine Segment jbest
            } else if (jbest >= 0) {
                if        (sket.seg.type[jbest] == "L") {
                    var segtype = "linseg";
                } else if (sket.seg.type[jbest] == "C") {
                    var segtype = "cirarc";
                } else if (sket.seg.type[jbest] == "S") {
                    var segtype = "spline";
                } else if (sket.seg.type[jbest] == "B") {
                    var segtype = "bezier";
                } else {
                    var segtype = "**unknown type**";
                }
                if (sket.seg.lbl[jbest].length == 0) {
                    postMessage("Segment "+(jbest+1)+" ("+segtype+") has no constraints");
                } else {
                    postMessage("Segment "+(jbest+1)+" ("+segtype+") has constraints: "+sket.seg.lbl[jbest]);
                }
            }

        // "@" --- dump sket structure
        } else if (myKeyPress == "@") {
            postMessage("sket.pnt->"+JSON.stringify(sket.pnt));
            postMessage("sket.seg->"+JSON.stringify(sket.seg));
            postMessage("sket.var->"+JSON.stringify(sket.var));
            postMessage("sket.con->"+JSON.stringify(sket.con));

        // invalid option
        } else {
            var ibest = getClosestSketchPoint();
            var jbest = getClosestSketchSegment();

            var dibest = 999999;
            if (ibest >= 0) {
                dibest = Math.abs(wv.cursorX - sket.pnt.x[ibest])
                       + Math.abs(wv.cursorY - sket.pnt.y[ibest]);
            }
            var djbest = 999999;
            if (jbest >= 0) {
                djbest = Math.abs(wv.cursorX - sket.seg.xm[jbest])
                       + Math.abs(wv.cursorY - sket.seg.ym[jbest]);
            }

            // remove constraints at Point ibest
            if (ibest >= 0 && dibest < djbest) {
                postMessage("Valid options are 'x', 'y', 'p', 't', 'w', 'd', and '<'");
            } else {
                postMessage("Valid options are 'h', 'v', 'i', 'l', 'r', 's', and '<'");
            }
        }

        // if .scale has been set, try to update .xorig or .yorig
        //    if a suitable constraint exists
        if (sket.scale !== undefined) {
            if (sket.xorig === undefined) {
                for (var icon = 0; icon < ncon; icon++) {
                    if (sket.con.type[icon] == "X") {
                        sket.xorig = sket.pnt.x[sket.con.index1[icon]] - sket.con.value[icon] / sket.scale;
                        break;
                    }
                }
            }
            if (sket.yorig === undefined) {
                for (var icon = 0; icon < ncon; icon++) {
                    if (sket.con.type[icon] == "Y") {
                        sket.yorig = sket.pnt.y[sket.con.index1[icon]] + sket.con.value[icon] / sket.scale;
                        break;
                    }
                }
            }
        }

    // sket.mode 4 options: w
    } else if (sket.mode == 4) {
        sket.mode = 3;

        var target = getClosestSketchPoint();

        if (target == sket.basePoint) {
            alert("Width not set since base and target Point are the same");
        } else {
            var wvalue = prompt("Enter width: ", wvalue);
            if (wvalue !== null) {
                if (isNaN(wvalue) !== true) {
                    if (sket.pnt.x[target] > sket.pnt.x[sket.basePoint]) {
                        wvalue = +Math.abs(wvalue);
                    } else {
                        wvalue = -Math.abs(wvalue);
                    }
                }

                sket.con.type[  ncon] = "W";
                sket.con.index1[ncon] = sket.basePoint;
                sket.con.index2[ncon] = target;
                sket.con.value[ ncon] = wvalue;
            }
        }

        sket.basePoint = undefined;

    // sket.mode 5 options: d
    } else if (sket.mode == 5) {
        sket.mode = 3;

        var target = getClosestSketchPoint();

        if (target == sket.basePoint) {
            alert("Depth not set since base and target Point are the same");
        } else {
            var dvalue = prompt("Enter depth: ", dvalue);
            if (dvalue !== null) {
                if (isNaN(dvalue) !== true) {
                    if (sket.pnt.y[target] > sket.pnt.y[sket.basePoint]) {
                        dvalue = +Math.abs(dvalue);
                    } else {
                        dvalue = -Math.abs(dvalue);
                    }
                }

                sket.con.type[  ncon] = "D";
                sket.con.index1[ncon] = sket.basePoint;
                sket.con.index2[ncon] = target;
                sket.con.value[ ncon] = dvalue;
            }
        }

        sket.basePoint = undefined;
    }

    drawSketch();
}


//
// called when "solveButton" is pressed (defined in ESP-sketch.html)
//
function solveSketchPre()
{
//    alert("in solveSketchPre()");

    var npnt = sket.pnt.x.length;
    var nvar = sket.var.name.length;
    var ncon = sket.con.type.length;
    var nseg = sket.seg.type.length;

    if (nvar > ncon) {
        alert("Sketch is under-constrained since nvar="+nvar+" but ncon="+ncon);
        return;
    } else if (nvar < ncon) {
        alert("Sketch is over-constrained since nvar="+nvar+" but ncon="+ncon);
        return;
    }

    if (sket.xorig === undefined || isNaN(sket.xorig) ||
        sket.yorig === undefined || isNaN(sket.yorig) ||
        sket.scale === undefined || isNaN(sket.scale)   ) {

        // get the size of the canvas
        var canvas = document.getElementById("sketcher");
        var width  = canvas.clientWidth;
        var height = canvas.clientHeight;

        // guess that scale of body is "5" and that it is centered at origin
        sket.scale = 5 / width;
        sket.xorig = width  / 2;
        sket.yorig = height / 2;
    }

    // create a message to send to the server
    var message = "solveSketch|";

    for (var ipnt = 0; ipnt < npnt; ipnt++) {
        if (ipnt == 0) {
            var im1 = npnt - 1;
        } else {
            var im1 = ipnt - 1;
        }

        message += Number(sket.scale * (sket.pnt.x[ipnt] - sket.xorig)).toFixed(6) + ";";
        message += Number(sket.scale * (sket.yorig - sket.pnt.y[ipnt])).toFixed(6) + ";";

        if (ipnt > 0) {
            message += Number(sket.scale *  sket.seg.dip[ipnt-1]      ).toFixed(6) + ";";
        } else if (sket.seg.iend[nseg-1] == 0) {
            message += Number(sket.scale *  sket.seg.dip[npnt-1]      ).toFixed(6) + ";";
        } else {
            message += "0;";
        }
    }

    message += "|";

    for (var icon = 0; icon < ncon; icon++) {
        message += sket.con.type[icon] + ";" + (sket.con.index1[icon]+1) + ";";
        if (sket.con.index2[icon] >= 0) {
            message += (sket.con.index2[icon] + 1) + ";" + sket.con.value[icon] + ";";
        } else {
            message +=  sket.con.index2[icon]      + ";" + sket.con.value[icon] + ";";
        }
    }

    message += "|";

    wv.socketUt.send(message);
}


//
// update the variables associated with a Sketch
//
function solveSketchPost(text)
{
//    alert("solveSketchPost(text="+text+")");

    // break up based upon the "|"
    var textList = text.split("|");

    if (textList[1].substring(0,7) == "ERROR::") {
        postMessage(textList[1]);
        alert("Sketch was not successfully solved");
        return;
    }

    var npnt = 0;
    for (var i = 0; i < textList[1].length; i++) {
        if (textList[1].charAt(i) == ';') {
            npnt++;
        }
    }
    npnt /= 3;

    // update the values (which are separated by semicolons)
    var values = textList[1].split(";");

    for (var ipnt = 0; ipnt < npnt; ipnt++) {
        if (ipnt == 0) {
            var im1 = npnt - 1;
        } else {
            var im1 = ipnt - 1;
        }

        sket.pnt.x[  ipnt] = Math.floor(sket.xorig + Number(values[3*ipnt  ]) / sket.scale);
        sket.pnt.y[  ipnt] = Math.floor(sket.yorig - Number(values[3*ipnt+1]) / sket.scale);
        sket.seg.dip[im1 ] = Math.floor(             Number(values[3*ipnt+2]) / sket.scale);
    }

    // update data associated with segments
    for (var iseg = 0; iseg < sket.seg.type.length; iseg++) {
        updateSegData(iseg);
    }

    // set the mode
    sket.mode = 6;

    // redraw the Sketch with the updated values
    drawSketch();
}


//
// called when "quit" is pressed (defined in ESP-sketch.html)
//
function quitSketch()
{
//    alert("in quitSketch()");

    changeMode(0);
}


//
// called when "save" is pressed (defined in ESP-sketch.html)
//
function saveSketch()
{
//    alert("in saveSketch()");

    // if the scaling is not set, make sure that at least one length is given
    //    (or else NaN will be sent back to .csm)
    if        (sket.scale === undefined || isNaN(sket.scale)) {
        alert("At lest one length or radius must be set before saving");
        return;
    } else if (sket.xorig === undefined || isNaN(sket.xorig)) {
        alert("At least one \"X\" must be set before saving");
        return;
    } else if (sket.yorig === undefined || isNaN(sket.yorig)) {
        alert("At least one \"Y\" must be set before saving");
        return;
    }

    // alert user that an unsolved Sketch may cause .csm to not solve
    if (sket.mode != 6) {
        if (confirm("Sketch is not solved and may break build.  Continue?") !== true) {
            return;
        }
    }

    var npnt = sket.pnt.x.length;
    var nseg = sket.seg.type.length;
    var ncon = sket.con.type.length;

    // build a string of the variables
    var vars = "";
    for (var ipnt = 0; ipnt < npnt; ipnt++) {
        vars += Number(sket.scale * (sket.pnt.x[ipnt] - sket.xorig)).toFixed(6) + ";";
        vars += Number(sket.scale * (sket.yorig - sket.pnt.y[ipnt])).toFixed(6) + ";";

        if (ipnt > 0) {
            vars += Number(sket.scale *  sket.seg.dip[ipnt-1]      ).toFixed(6) + ";";
        } else if (sket.seg.iend[nseg-1] == 0) {
            vars += Number(sket.scale *  sket.seg.dip[npnt-1]      ).toFixed(6) + ";";
        } else {
            vars += "0;";
        }
    }

    // build a string of the constraints
    var cons = "";
    for (var icon = 0; icon < ncon; icon++) {
        cons += sket.con.type[icon] + ";" + (sket.con.index1[icon]+1) + ";";
        if (sket.con.index2[icon] >= 0) {
            cons += (sket.con.index2[icon] + 1) + ";" + sket.con.value[icon] + ";";
        } else {
            cons +=  sket.con.index2[icon]      + ";" + sket.con.value[icon] + ";";
        }
    }

    // build a string of the segments
    var segs = "";
    for (var iseg = 0; iseg < nseg; iseg++) {
        segs += sket.seg.type[iseg] + ";" + (sket.seg.ibeg[iseg]+1) + ";" +
                                            (sket.seg.iend[iseg]+1) + ";";
    }

    // create the message to be sent to the server
    wv.socketUt.send("saveSketch|"+sket.ibrch+"|"+vars+"|"+cons+"|"+segs+"|");
}


//
// undo last Sketch command
//
function undoSketch()
{
//    alert("undoSketch()");

    // if undo information does not exist, inform the user
    if (sket.undo     === undefined ||
        sket.undo.pnt === undefined ||
        sket.undo.seg === undefined ||
        sket.undo.var === undefined ||
        sket.undo.con === undefined   ) {
        alert("There is nothing to undo");
        return;
    }

    // remove old Sketch info and initialize the arrays
    sket.pnt        = {};
    sket.pnt.x      = [];
    sket.pnt.y      = [];
    sket.pnt.lbl    = [];

    sket.seg        = {};
    sket.seg.type   = [];
    sket.seg.ibeg   = [];
    sket.seg.iend   = [];
    sket.seg.xm     = [];
    sket.seg.ym     = [];
    sket.seg.lbl    = [];

    sket.seg.dip    = [];
    sket.seg.xc     = [];
    sket.seg.yc     = [];
    sket.seg.rad    = [];
    sket.seg.tbeg   = [];
    sket.seg.tend   = [];

    sket.var        = {};
    sket.var.name   = [];
    sket.var.value  = [];

    sket.con        = {};
    sket.con.type   = [];
    sket.con.index1 = [];
    sket.con.index2 = [];
    sket.con.value  = [];

    // copy the undo information into Sketch information
    sket.mode = sket.undo.mode;

    for (var ipnt = 0; ipnt < sket.undo.pnt.x.length; ipnt++) {
        sket.pnt.x[  ipnt] = sket.undo.pnt.x[  ipnt];
        sket.pnt.y[  ipnt] = sket.undo.pnt.y[  ipnt];
        sket.pnt.lbl[ipnt] = sket.undo.pnt.lbl[ipnt];
    }
    for (var iseg = 0; iseg < sket.undo.seg.type.length; iseg++) {
        sket.seg.type[iseg] = sket.undo.seg.type[iseg];
        sket.seg.ibeg[iseg] = sket.undo.seg.ibeg[iseg];
        sket.seg.iend[iseg] = sket.undo.seg.iend[iseg];
        sket.seg.xm[  iseg] = sket.undo.seg.xm[  iseg];
        sket.seg.ym[  iseg] = sket.undo.seg.ym[  iseg];
        sket.seg.lbl[ iseg] = sket.undo.seg.lbl[ iseg];
        sket.seg.dip[ iseg] = sket.undo.seg.dip[ iseg];
        sket.seg.xc[  iseg] = sket.undo.seg.xc[  iseg];
        sket.seg.yc[  iseg] = sket.undo.seg.yc[  iseg];
        sket.seg.rad[ iseg] = sket.undo.seg.rad[ iseg];
        sket.seg.tbeg[iseg] = sket.undo.seg.tbeg[iseg];
        sket.seg.tend[iseg] = sket.undo.seg.tend[iseg];
    }
    for (var ivar = 0; ivar < sket.undo.var.name.length; ivar++) {
        sket.var.name[ ivar] = sket.undo.var.name[ ivar];
        sket.var.value[ivar] = sket.undo.var.value[ivar];
    }
    for (var icon = 0; icon < sket.undo.con.type.length; icon++){
        sket.con.type[  icon] = sket.undo.con.type[  icon];
        sket.con.index1[icon] = sket.undo.con.index1[icon];
        sket.con.index2[icon] = sket.undo.con.index2[icon];
        sket.con.value[ icon] = sket.undo.con.value[ icon];
    }

    // remove the undo information
    sket.undo = {};

    // update the display
    drawSketch();
}


//
// save an undo snapshot
//
function saveSketchUndo()
{
//    alert("saveSketchUndo()");

    // remove old undo info and initialize the arrays
    sket.undo      = {};

    sket.undo.pnt        = {};
    sket.undo.pnt.x      = [];
    sket.undo.pnt.y      = [];
    sket.undo.pnt.lbl    = [];

    sket.undo.seg        = {};
    sket.undo.seg.type   = [];
    sket.undo.seg.ibeg   = [];
    sket.undo.seg.iend   = [];
    sket.undo.seg.xm     = [];
    sket.undo.seg.ym     = [];
    sket.undo.seg.lbl    = [];

    sket.undo.seg.dip    = [];
    sket.undo.seg.xc     = [];
    sket.undo.seg.yc     = [];
    sket.undo.seg.rad    = [];
    sket.undo.seg.tbeg   = [];
    sket.undo.seg.tend   = [];

    sket.undo.var        = {};
    sket.undo.var.name   = [];
    sket.undo.var.value  = [];

    sket.undo.con        = {};
    sket.undo.con.type   = [];
    sket.undo.con.index1 = [];
    sket.undo.con.index2 = [];
    sket.undo.con.value  = [];

    // copy the Sketch information into the undo storage
    sket.undo.mode = sket.mode;

    for (var ipnt = 0; ipnt < sket.pnt.x.length; ipnt++) {
        sket.undo.pnt.x[  ipnt] = sket.pnt.x[  ipnt];
        sket.undo.pnt.y[  ipnt] = sket.pnt.y[  ipnt];
        sket.undo.pnt.lbl[ipnt] = sket.pnt.lbl[ipnt];
    }
    for (var iseg = 0; iseg < sket.seg.type.length; iseg++) {
        sket.undo.seg.type[iseg] = sket.seg.type[iseg];
        sket.undo.seg.ibeg[iseg] = sket.seg.ibeg[iseg];
        sket.undo.seg.iend[iseg] = sket.seg.iend[iseg];
        sket.undo.seg.xm[  iseg] = sket.seg.xm[  iseg];
        sket.undo.seg.ym[  iseg] = sket.seg.ym[  iseg];
        sket.undo.seg.lbl[ iseg] = sket.seg.lbl[ iseg];
        sket.undo.seg.dip[ iseg] = sket.seg.dip[ iseg];
        sket.undo.seg.xc[  iseg] = sket.seg.xc[  iseg];
        sket.undo.seg.yc[  iseg] = sket.seg.yc[  iseg];
        sket.undo.seg.rad[ iseg] = sket.seg.rad[ iseg];
        sket.undo.seg.tbeg[iseg] = sket.seg.tbeg[iseg];
        sket.undo.seg.tend[iseg] = sket.seg.tend[iseg];
    }
    for (var ivar = 0; ivar < sket.var.name.length; ivar++) {
        sket.undo.var.name[ ivar] = sket.var.name[ ivar];
        sket.undo.var.value[ivar] = sket.var.value[ivar];
    }
    for (var icon = 0; icon < sket.con.type.length; icon++){
        sket.undo.con.type[  icon] = sket.con.type[  icon];
        sket.undo.con.index1[icon] = sket.con.index1[icon];
        sket.undo.con.index2[icon] = sket.con.index2[icon];
        sket.undo.con.value[ icon] = sket.con.value[ icon];
    }
}


//
// update the variables associated with a segment
//
function updateSegData(iseg)
{
//    alert("updateSegData(iseg="+iseg+")");

    sket.seg.xm[iseg] = (sket.pnt.x[sket.seg.ibeg[iseg]] + sket.pnt.x[sket.seg.iend[iseg]]) / 2;
    sket.seg.ym[iseg] = (sket.pnt.y[sket.seg.ibeg[iseg]] + sket.pnt.y[sket.seg.iend[iseg]]) / 2;

    if (Math.abs(sket.seg.dip[iseg]) > 1.0e-6) {
        var xa = sket.pnt.x[sket.seg.ibeg[iseg]];
        var ya = sket.pnt.y[sket.seg.ibeg[iseg]];
        var xb = sket.pnt.x[sket.seg.iend[iseg]];
        var yb = sket.pnt.y[sket.seg.iend[iseg]];
        var d  =            sket.seg.dip[ iseg];

        var L  = Math.sqrt((xb-xa) * (xb-xa) + (yb-ya) * (yb-ya));
        var R  = (L*L + 4*d*d) / (8*d);

        sket.seg.xc[iseg] = (xa + xb) / 2 + (R - d) * (yb - ya) / L;
        sket.seg.yc[iseg] = (ya + yb) / 2 - (R - d) * (xb - xa) / L;

        sket.seg.rad[ iseg] = R;
        sket.seg.tbeg[iseg] = Math.atan2(ya-sket.seg.yc[iseg], xa-sket.seg.xc[iseg]);
        sket.seg.tend[iseg] = Math.atan2(yb-sket.seg.yc[iseg], xb-sket.seg.xc[iseg]);

        if (R > 0) {
            sket.seg.tbeg[iseg] -= 2 * Math.PI;
            while (sket.seg.tbeg[iseg] < sket.seg.tend[iseg]) {
                sket.seg.tbeg[iseg] += 2 * Math.PI;
            }
        } else {
            sket.seg.tbeg[iseg] += 2 * Math.PI;
            while (sket.seg.tbeg[iseg] > sket.seg.tend[iseg]) {
                sket.seg.tbeg[iseg] -= 2 * Math.PI;
            }
        }

        var tavg = (sket.seg.tbeg[iseg] + sket.seg.tend[iseg]) / 2;
        sket.seg.xm[iseg] = sket.seg.xc[iseg] + Math.cos(tavg) * Math.abs(R);
        sket.seg.ym[iseg] = sket.seg.yc[iseg] + Math.sin(tavg) * Math.abs(R);
    }
}


//
// Draw the current Sketch
//
function drawSketch()
{
//    alert("drawSketch()");

    // set up context
    var canvas  = document.getElementById("sketcher");
    var context = canvas.getContext("2d");

    // clear the screen
    context.clearRect(0, 0, canvas.width, canvas.height);

    context.lineCap  = "round";
    context.lineJoin = "round";

    // determine number of Points and Segments in Sketch
    var npnt = sket.pnt.x.length;
    var nseg = sket.seg.type.length;
    var ncon = sket.con.type.length;

    // return if no Points
    if (npnt <= 0) {
        var button = document.getElementById("buildButton");
        button.onclick = null;
        button.style.backgroundColor =  "#FFFF3F";
        button["innerHTML"] = "Initializing...";

        return;
    }

    // draw the part of the Sketch that is complete
    if (nseg > 0) {

        // we are editting the Cirarc at the end, so it has to
        //    be drawn separately (since it will be drawn in red)
        if (sket.mode == 2) {

            // the part of the Sketch this done
            context.strokeStyle = "black";
            context.lineWidth   = 3;
            context.beginPath();

            var ibeg = sket.seg.ibeg[0];
            var iend = sket.seg.iend[0];
            context.moveTo(sket.pnt.x[ibeg], sket.pnt.y[ibeg]);

            for (var iseg = 0; iseg < nseg-1; iseg++) {
                iend = sket.seg.iend[iseg];

                if (Math.abs(sket.seg.dip[iseg]) < 1e-6) {
                    context.lineTo(sket.pnt.x[iend], sket.pnt.y[iend]);
                } else if (sket.seg.rad[iseg] > 0) {
                    context.arc(sket.seg.xc[iseg], sket.seg.yc[iseg],
                                +sket.seg.rad[iseg], sket.seg.tbeg[iseg], sket.seg.tend[iseg], 1);
                } else {
                    context.arc(sket.seg.xc[iseg], sket.seg.yc[iseg],
                                -sket.seg.rad[iseg], sket.seg.tbeg[iseg], sket.seg.tend[iseg], 0);
                }
            }

            context.stroke();

            // the Cirarc at the end which is to be rendered in red
            context.strokeStyle = "red";
            context.beginPath();

            if (sket.seg.dip[nseg-1] > 0) {
                context.arc(sket.seg.xc[nseg-1], sket.seg.yc[nseg-1],
                            +sket.seg.rad[nseg-1], sket.seg.tbeg[nseg-1], sket.seg.tend[nseg-1], 1);
            } else {
                context.arc(sket.seg.xc[nseg-1], sket.seg.yc[nseg-1],
                            -sket.seg.rad[nseg-1], sket.seg.tbeg[nseg-1], sket.seg.tend[nseg-1], 0);
            }

            context.stroke();

        // we are NOT editting the Cirarc at the end
        } else {
            context.strokeStyle = "black";
            context.lineWidth   = 3;
            context.beginPath();

            var ibeg = sket.seg.ibeg[0];
            var iend = sket.seg.iend[0];
            context.moveTo(sket.pnt.x[ibeg], sket.pnt.y[ibeg]);

            for (var iseg = 0; iseg < nseg; iseg++) {
                iend = sket.seg.iend[iseg];

                if (sket.seg.type[iseg] == "B") {
                    var Xbezier = [];
                    var Ybezier = [];
                    var nbezier = 2;

                    Xbezier[0] = sket.pnt.x[ibeg];
                    Ybezier[0] = sket.pnt.y[ibeg];
                    Xbezier[1] = sket.pnt.x[iend];
                    Ybezier[1] = sket.pnt.y[iend];

                    while (iseg < nseg-1) {
                        if (sket.seg.type[iseg+1] == "B") {
                            iseg++;
                            iend = sket.seg.iend[iseg];
                            Xbezier[nbezier] = sket.pnt.x[iend];
                            Ybezier[nbezier] = sket.pnt.y[iend];
                            nbezier++;
                        } else {
                            break;
                        }
                    }

                    drawBezier(context, Xbezier, Ybezier);
//                } else if (sket.seg.type[iseg] == "S") {
//                    var Xspline = [];
//                    var Yspline = [];
//                    var nspline = 2;
//
//                    Xspline[0] = sket.pnt.x[ibeg];
//                    Yspline[0] = sket.pnt.y[ibeg];
//                    Xspline[1] = sket.pnt.x[iend];
//                    Yspline[1] = sket.pnt.y[iend];
//
//                    while (iseg < nseg-1) {
//                        if (sket.seg.type[iseg+1] == "S") {
//                            iseg++;
//                            iend = sket.seg.iend[iseg];
//                            Xspline[nspline] = sket.pnt.x[iend];
//                            Yspline[nspline] = sket.pnt.y[iend];
//                            nspline++;
//                        } else {
//                            break;
//                        }
//                    }
//
//                    drawSpline(context, Xspline, Yspline);
                } else if (Math.abs(sket.seg.dip[iseg]) < 1e-6) {
                    context.lineTo(sket.pnt.x[iend], sket.pnt.y[iend]);
                } else if (sket.seg.dip[iseg] > 0) {
                    context.arc(sket.seg.xc[iseg], sket.seg.yc[iseg],
                                +sket.seg.rad[iseg], sket.seg.tbeg[iseg], sket.seg.tend[iseg], 1);
                } else {
                    context.arc(sket.seg.xc[iseg], sket.seg.yc[iseg],
                                -sket.seg.rad[iseg], sket.seg.tbeg[iseg], sket.seg.tend[iseg], 0);
                }

                ibeg = iend;
            }

            context.stroke();

            // if Sketch is closed, fill it
            if (iend == sket.seg.ibeg[0]) {
                if (sket.var.name.length == sket.con.type.length) {
                    context.fillStyle = "#c0f0c0";  // light green
                    context.fill();
                } else {
                    context.fillStyle = "#e0e0e0";  // grey
                    context.fill();
                }
            }

            // if draw straight lines for Segments associated with splines
            //    and beziers
            context.strokeStyle = "cyan";
            context.lineWidth   = 1;
            for (iseg = 0; iseg < nseg; iseg++) {
                if (sket.seg.type[iseg] == "B" ||
                    sket.seg.type[iseg] == "S"   ) {

                    ibeg = sket.seg.ibeg[iseg];
                    iend = sket.seg.iend[iseg];
                    context.beginPath();
                    context.moveTo(sket.pnt.x[ibeg], sket.pnt.y[ibeg]);
                    context.lineTo(sket.pnt.x[iend], sket.pnt.y[iend]);
                    context.stroke();
                }
            }

            // we are creating W constraint
            if (sket.mode == 4) {
                var xmid = (  sket.pnt.x[sket.basePoint] + wv.cursorX) / 2;
                var ymid = (2*sket.pnt.y[sket.basePoint] + wv.cursorY) / 3;

                context.strokeStyle = "red";
                context.lineWidth   = 1;
                context.beginPath();
                context.moveTo(sket.pnt.x[sket.basePoint], sket.pnt.y[sket.basePoint]);
                context.lineTo(sket.pnt.x[sket.basePoint], ymid                      );
                context.lineTo(wv.cursorX,                 ymid                      );
                context.lineTo(wv.cursorX,                 wv.cursorY                );
                context.stroke();
            }

            // we are creating D constraint
            if (sket.mode == 5) {
                var xmid = (2*sket.pnt.x[sket.basePoint] + wv.cursorX) / 3;
                var ymid = (  sket.pnt.y[sket.basePoint] + wv.cursorY) / 2;

                context.strokeStyle = "red";
                context.lineWidth   = 1;
                context.beginPath();
                context.moveTo(sket.pnt.x[sket.basePoint], sket.pnt.y[sket.basePoint]);
                context.lineTo(xmid,                       sket.pnt.y[sket.basePoint]);
                context.lineTo(xmid,                       wv.cursorY                );
                context.lineTo(wv.cursorX,                 wv.cursorY                );
                context.stroke();
            }
        }
    }

    // draw the Points
    context.fillStyle = "black";
    context.fillRect(sket.pnt.x[0]-3, sket.pnt.y[0]-3, 7, 7);

    for (var ipnt = 1; ipnt < npnt; ipnt++) {
        context.fillRect(sket.pnt.x[ipnt]-2, sket.pnt.y[ipnt]-2, 5, 5);
    }

    // display constraint labels associated with the Points
    for (ipnt = 0; ipnt < npnt; ipnt++) {
        if (sket.pnt.lbl[ipnt] != "") {
            context.font         = "10px Verdana";
            context.textAlign    = "center";
            context.textBaseline = "middle";

            var width  = context.measureText(sket.pnt.lbl[ipnt]).width;
            var height = 8;

            context.fillStyle    = "yellow";
            context.fillRect(sket.pnt.x[ipnt]-width/2-2, sket.pnt.y[ipnt]-height/2-2, width+4, height+4);

            context.fillStyle    = "black";
            context.fillText(sket.pnt.lbl[ipnt], sket.pnt.x[ipnt], sket.pnt.y[ipnt]);
        }
    }

    // display constraint labels associated with the Segments
    for (var iseg = 0; iseg < nseg; iseg++) {
        if (sket.seg.lbl[iseg] != "") {
            context.font         = "12px Verdana";
            context.textAlign    = "center";
            context.textBaseline = "middle";

            var width  = context.measureText(sket.seg.lbl[iseg]).width;
            var height = 8;

            context.fillStyle    = "yellow";
            context.fillRect(sket.seg.xm[iseg]-width/2-2, sket.seg.ym[iseg]-height/2-2, width+4, height+4);

            context.fillStyle    = "black";
            context.fillText(sket.seg.lbl[iseg], sket.seg.xm[iseg], sket.seg.ym[iseg]);
        }
    }

    // display width and depth constraints
    for (var icon = 0; icon < ncon; icon++) {
        if (sket.con.type[icon] == "W" || sket.con.type[icon] == "D") {
            var ibeg = sket.con.index1[icon];
            var iend = sket.con.index2[icon];

            context.strokeStyle = "green";
            context.lineWidth   = 1;
            context.beginPath();
            context.moveTo(sket.pnt.x[ibeg], sket.pnt.y[ibeg]);
            if (sket.con.type[icon] == "W") {
                var xmid = (  sket.pnt.x[ibeg] + sket.pnt.x[iend]) / 2;
                var ymid = (2*sket.pnt.y[ibeg] + sket.pnt.y[iend]) / 3;

                context.lineTo(sket.pnt.x[ibeg], ymid);
                context.lineTo(sket.pnt.x[iend], ymid);
            } else {
                var xmid = (2*sket.pnt.x[ibeg] + sket.pnt.x[iend]) / 3;
                var ymid = (  sket.pnt.y[ibeg] + sket.pnt.y[iend]) / 2;

                context.lineTo(xmid, sket.pnt.y[ibeg]);
                context.lineTo(xmid, sket.pnt.y[iend]);
            }
            context.lineTo(sket.pnt.x[iend], sket.pnt.y[iend]);
            context.stroke();

            context.font         = "12px Verdana";
            context.textAlign    = "center";
            context.textBaseline = "middle";

            if (sket.con.type[icon] == "W") {
                var label = "W";
            } else {
                var label = "D";
            }
            var width  = context.measureText(label).width;
            var height = 8;

            context.fillStyle = "yellow";
            context.fillRect(xmid-width/2-2, ymid-height/2-2, width+4, height+4);

            context.fillStyle = "black";
            context.fillText(label, xmid, ymid);
        }
    }

    // if we are in mode 1, draw a line from the last Point to the cursor
    if (sket.mode == 1) {

        // nominal line to cursor is blue
        var color = "blue";

        // if horizontal, make it orange
        if (Math.abs(sket.pnt.y[npnt-1]-wv.cursorY) < sket.halo) {
            color = "orange";

        // otherwise see if it alignes with another point
        } else {
            for (var jpnt = 0; jpnt < npnt-1; jpnt++) {
                if (Math.abs(sket.pnt.y[jpnt]-wv.cursorY) < sket.halo) {
                    context.strokeStyle = "orange";
                    context.lineWidth   = 1;

                    context.beginPath();
                    context.moveTo(sket.pnt.x[jpnt], sket.pnt.y[jpnt]);
                    context.lineTo(wv.cursorX,  wv.cursorY );
                    context.stroke();
                    break;
                }
            }
        }

        // if vertical, make it orange
        if (Math.abs(sket.pnt.x[npnt-1]-wv.cursorX) < sket.halo) {
            color = "orange";

        // otherwise see if it alignes with another point
        } else {
            for (var jpnt = 0; jpnt < npnt-1; jpnt++) {
                if (Math.abs(sket.pnt.x[jpnt]-wv.cursorX) < sket.halo) {
                    context.strokeStyle = "orange";
                    context.lineWidth   = 1;

                    context.beginPath();
                    context.moveTo(sket.pnt.x[jpnt], sket.pnt.y[jpnt]);
                    context.lineTo(wv.cursorX,  wv.cursorY );
                    context.stroke();
                    break;
                }
            }
        }

        // if the cursor is close to the first Point, indicate that
        if (Math.abs(sket.pnt.x[0]-wv.cursorX) < sket.halo &&
            Math.abs(sket.pnt.y[0]-wv.cursorY) < sket.halo   ) {
            context.strokeStyle = "orange";
            context.lineWidth   = 5;
            context.beginPath();
            context.arc(sket.pnt.x[0], sket.pnt.y[0], 17, 0, 2*Math.PI);
            context.stroke();
        }

        // draw the line to the cursor
        context.strokeStyle = color;
        context.lineWidth   = 2;

        context.beginPath();
        context.moveTo(sket.pnt.x[npnt-1], sket.pnt.y[npnt-1]);
        context.lineTo(wv.cursorX,    wv.cursorY   );
        context.stroke();
    }

    // have the Build button show the current Sketcher status
    var button = document.getElementById("buildButton");
    button.onclick = null;
    button.style.backgroundColor =  "#FFFF3F";

    if        (sket.mode == 0) {
        button["innerHTML"] = "Initializing...";
    } else if (sket.mode == 1) {
        button["innerHTML"] = "Drawing...";
    } else if (sket.mode == 2) {
        button["innerHTML"] = "Setting R...";
//  } else if (sket.mode == 3) {
//      handled below
    } else if (sket.mode == 4) {
        button["innerHTML"] = "Setting W...";
    } else if (sket.mode == 5) {
        button["innerHTML"] = "Setting D...";
    } else if (sket.mode == 6) {
        button["innerHTML"] = "Up to date";
        button.style.backgroundColor = null;
    } else if (sket.var.name.length != sket.con.type.length) {
        button["innerHTML"] = "Constraining...";
    } else {
        activateSolveButton();
    }

    // post informtion about current mode in blframe
    var skstat = document.getElementById("SketchStatus");

    var mesg = "ndof=" + sket.var.name.length + "   ncon=" + sket.con.type.length + "\n";
    if        (sket.mode == 1) {
        mesg += "Valid commands are:\n";
        mesg += "  'l'   add linseg\n";
        mesg += "  'c'   add cirarc\n";
        mesg += "  's'   add spline\n";
        mesg += "  'b'   add bezier\n";
        mesg += "  'o'   complete (open) sketch\n";
    } else if (sket.mode == 2) {
        mesg += "Hit any character to set curvature\n";
    } else if (sket.mode == 3) {
        mesg += "Valid constraints at points\n";
        mesg += "  'x' (fix x)     'y' (fix y)\n";
        mesg += "  'p' (perp)      't' (tangent)\n";
        mesg += "  'a' (angle)\n";
        mesg += "  'w' (width)     'd' (depth)\n";
        mesg += "Valid constraints on segments\n";
        mesg += "  'h' (horiz)     'v' (vertical)\n";
        mesg += "  'i' (incline)   'l' (length)\n";
        mesg += "Valid constraints on cirarcs\n";
        mesg += "  'r' (radius)    's' (sweep angle)\n";
        mesg += "Valid commands anywhere\n";
        mesg += "  '<' (delete)    '?' (info)\n";
    } else if (sket.mode == 4) {
        mesg += "Hit any character to set width\n";
    } else if (sket.mode == 5) {
        mesg += "Hit any character to set depth\n";
    }

    var pre  = document.createElement("pre");
    var text = document.createTextNode(mesg);
    pre.appendChild(text);

    skstat.replaceChild(pre, skstat.lastChild);
}


// draw a Bezier curve (with evaluations via Hoerner-like scheme)
//
function drawBezier(context, Xbezier, Ybezier)
{
//    alert("drawBezier()");

    var degree = Xbezier.length - 1;

    for (var j = 1; j <= 4*degree; j++) {
        var t  = j / (4.0 * degree);
        var t1 = 1.0 - t;

        var fact   = 1.0;
        var bicoef = 1;

        var xx = Xbezier[0] * t1;
        var yy = Ybezier[0] * t1;

        for (var i = 1; i < degree; i++) {
            fact    *= t;
            bicoef  *= (degree - i + 1) / i;

            xx = (xx + fact * bicoef * Xbezier[i]) * t1;
            yy = (yy + fact * bicoef * Ybezier[i]) * t1;
        }

        xx += fact * t * Xbezier[degree];
        yy += fact * t * Ybezier[degree];

        context.lineTo(xx, yy);
    }
}


// draw a Spline curve (with evaluations via Hoerner-like scheme)
//
function drawSpline(context, Xspline, Yspline)
{
//    alert("drawSpline()");

    var degree = Xspline.length - 1;

    // convert Spline points into cubic Bezier control points
    var Xbezier = [];
    var Ybezier = [];
    var gam     = [];

    var bet    = 4;
    Xbezier[1] = (6 * Xspline[1] - Xspline[0]) / bet;
    Ybezier[1] = (6 * Yspline[1] - Yspline[0]) / bet;

    for (var i = 1; i < degree-2; i++) {
        gam[i]       = 1 / bet;
        bet          =  4 - gam[i];
        Xbezier[i+1] = (6 * Xspline[i+1] - Xbezier[i]) / bet;
        Ybezier[i+1] = (6 * Yspline[i+1] - Ybezier[i]) / bet;
    }

    gam[degree-2]     = 1 / bet;
    bet               =  4 - gam[degree-2];
    Xbezier[degree-1] = (6 * Xspline[degree-1] - Xspline[degree] - Xbezier[degree-2]) / bet;
    Ybezier[degree-1] = (6 * Yspline[degree-1] - Yspline[degree] - Ybezier[degree-2]) / bet;

    for (i = degree-2; i > 0; i--) {
        Xbezier[i] -= gam[i] * Xbezier[i+1];
        Ybezier[i] -= gam[i] * Ybezier[i+1];
    }

    Xbezier[0] = Xspline[0];
    Ybezier[0] = Yspline[0];

    Xbezier[degree] = Xspline[degree];
    Ybezier[degree] = Yspline[degree];

    // draw the Splines
    for (i = 1; i <= degree; i++) {
        for (var j = 1; j < 5; j++) {
            var t  = 0.25 * j;
            var t1 = 1.0 - t;

            var xx = (t1 * t1 * t1) * (    Xspline[i-1]                 )
                   + (t1 * t1 * t ) * (2 * Xbezier[i-1] +     Xbezier[i])
                   + (t1 * t  * t ) * (    Xbezier[i-1] + 2 * Xbezier[i])
                   + (t  * t  * t ) * (                       Xspline[i]);
            var yy = (t1 * t1 * t1) * (    Yspline[i-1]                 )
                   + (t1 * t1 * t ) * (2 * Ybezier[i-1] +     Ybezier[i])
                   + (t1 * t  * t ) * (    Ybezier[i-1] + 2 * Ybezier[i])
                   + (t  * t  * t ) * (                       Yspline[i]);
            context.lineTo(xx,yy);
        }
    }
}


//
// get Point closest to the cursor
//
function getClosestSketchPoint()
{
//    alert("getClosestSketchPoint()");

    var npnt  = sket.pnt.x.length;
    var ibest = -1;
    var dbest = 999999;

    for (var ipnt = 0; ipnt < npnt; ipnt++) {
        var dtest = Math.abs(wv.cursorX - sket.pnt.x[ipnt])
                  + Math.abs(wv.cursorY - sket.pnt.y[ipnt]);
        if (dtest < dbest) {
            ibest = ipnt;
            dbest = dtest;
        }
    }

    return ibest;
}


//
// get Segment closest to the cursor
//
function getClosestSketchSegment()
{
//    alert("getClosestSketchSegment()");

    var nseg  = sket.seg.type.length;
    var ibest = -1;
    var dbest = 999999;

    for (var iseg = 0; iseg < nseg; iseg++) {
        var dtest = Math.abs(wv.cursorX - sket.seg.xm[iseg])
                  + Math.abs(wv.cursorY - sket.seg.ym[iseg]);
        if (dtest < dbest) {
            ibest = iseg;
            dbest = dtest;
        }
    }

    return ibest;
}


//
// get W or D constrint closest to the cursor
//
function getClosestSketchConstraint()
{
//    alert("getClosestSketchConstraint()");

    var ncon  = sket.con.type.length;
    var ibest = undefined;
    var dbest = 999999;

    for (var icon = 0; icon < ncon; icon++) {
        if        (sket.con.type[icon] == "W") {
            var xmid = (  sket.pnt.x[sket.con.index1[icon]]
                        + sket.pnt.x[sket.con.index2[icon]]) / 2;
            var ymid = (2*sket.pnt.y[sket.con.index1[icon]]
                        + sket.pnt.y[sket.con.index2[icon]]) / 3;
        } else if (sket.con.type[icon] == "D") {
            var xmid = (2*sket.pnt.x[sket.con.index1[icon]]
                        + sket.pnt.x[sket.con.index2[icon]]) / 3;
            var ymid = (  sket.pnt.y[sket.con.index1[icon]]
                        + sket.pnt.y[sket.con.index2[icon]]) / 2;
        } else {
            continue;
        }

        var dtest = Math.abs(wv.cursorX - xmid) + Math.abs(wv.cursorY - ymid);
        if (dtest < dbest) {
            ibest = icon;
            dbest = dtest;
        }
    }

    return ibest;
}


//
// constructor for a Tree
//
function Tree(doc, treeId) {
    // alert("in Tree(doc="+doc+", treeId="+treeId+")");

    // remember the document
    this.document = doc;
    this.treeId   = treeId;

    // arrays to hold the Nodes
    this.name   = new Array();
    this.gprim  = new Array();
    this.click  = new Array();
    this.cmenu  = new Array();
    this.parent = new Array();
    this.child  = new Array();
    this.next   = new Array();
    this.nprop  = new Array();
    this.opened = new Array();

    this.prop1  = new Array();
    this.valu1  = new Array();
    this.cbck1  = new Array();
    this.prop2  = new Array();
    this.valu2  = new Array();
    this.cbck2  = new Array();
    this.prop3  = new Array();
    this.valu3  = new Array();
    this.cbck3  = new Array();

    // initialize Node=0 (the root)
    this.name[  0] = "**root**";
    this.gprim[ 0] = "";
    this.click[ 0] = null;
    this.cmenu[ 0] = "";
    this.parent[0] = -1;
    this.child[ 0] = -1;
    this.next[  0] = -1;
    this.nprop[ 0] =  0;
    this.prop1[ 0] = "";
    this.valu1[ 0] = "";
    this.cbck1[ 0] = null;
    this.prop2[ 0] = "";
    this.valu2[ 0] = "";
    this.cbck2[ 0] = null;
    this.prop3[ 0] = "";
    this.valu3[ 0] = "";
    this.cbck3[ 0] = null;
    this.opened[0] = +1;

    // add methods
    this.addNode  = TreeAddNode;
    this.expand   = TreeExpand;
    this.contract = TreeContract;
    this.prop     = TreeProp;
    this.clear    = TreeClear;
    this.build    = TreeBuild;
    this.update   = TreeUpdate;
}


//
// add a Node to the Tree
//
function TreeAddNode(iparent, name, gprim, click, cmenu, prop1, valu1, cbck1,
                     prop2, valu2, cbck2, prop3, valu3, cbck3)
{
    // alert("in TreeAddNode(iparent="+iparent+", name="+name+", gprim="+gprim+", click="+click+
    //                       ", cmenu="+cmenu+", prop1="+prop1+", valu1="+valu1+",cbck1="+cbck1+
    //                       ", prop2="+prop2+", valu2="+valu2+", cbck2="+cbck2+", prop3="+prop3+
    //                       ", valu3="+valu3+", cbck3="+cbck3+")");

    // validate the input
    if (iparent < 0 || iparent >= this.name.length) {
        alert("iparent="+iparent+" is out of range");
        return;
    }

    // find the next Node index
    var inode = this.name.length;

    // store this Node's values
    this.name[  inode] = name;
    this.gprim[ inode] = gprim;
    this.click[ inode] = click;
    this.cmenu[ inode] = cmenu;
    this.parent[inode] = iparent;
    this.child[ inode] = -1;
    this.next[  inode] = -1;
    this.nprop[ inode] =  0;
    this.opened[inode] =  0;

    // store the properties
    if (prop1 !== undefined) {
        this.nprop[inode] = 1;
        this.prop1[inode] = prop1;
        this.valu1[inode] = valu1;
        this.cbck1[inode] = cbck1;
    }

    if (prop2 !== undefined) {
        this.nprop[inode] = 2;
        this.prop2[inode] = prop2;
        this.valu2[inode] = valu2;
        this.cbck2[inode] = cbck2;
    }

    if (prop3 !== undefined) {
        this.nprop[inode] = 3;
        this.prop3[inode] = prop3;
        this.valu3[inode] = valu3;
        this.cbck3[inode] = cbck3;
    }

    // if the parent does not have a child, link this
    //    new Node to the parent
    if (this.child[iparent] < 0) {
        this.child[iparent] = inode;

    // otherwise link this Node to the last parent's child
    } else {
        var jnode = this.child[iparent];
        while (this.next[jnode] >= 0) {
            jnode = this.next[jnode];
        }

        this.next[jnode] = inode;
    }
}


//
// build the Tree (ie, create the html table from the Nodes)
//
function TreeBuild() {
    // alert("in TreeBuild()");

    var doc = this.document;

    // if the table already exists, delete it and all its children (3 levels)
    var thisTable = doc.getElementById(this.treeId);
    if (thisTable) {
        var child1 = thisTable.lastChild;
        while (child1) {
            var child2 = child1.lastChild;
            while (child2) {
                var child3 = child2.lastChild;
                while (child3) {
                    child2.removeChild(child3);
                    child3 = child2.lastChild;
                }
                child1.removeChild(child2);
                child2 = child1.lastChild;
            }
            thisTable.removeChild(child1);
            child1 = thisTable.lastChild;
        }
        thisTable.parentNode.removeChild(thisTable);
    }

    // build the new table
    var newTable = doc.createElement("table");
    newTable.setAttribute("id", this.treeId);
    doc.getElementById("treefrm").appendChild(newTable);

    // traverse the Nodes using depth-first search
    var inode = 1;
    while (inode > 0) {

        // table row "node"+inode
        var newTR = doc.createElement("TR");
        newTR.setAttribute("id", "node"+inode);
        newTable.appendChild(newTR);

        // table data "node"+inode+"col1"
        var newTDcol1 = doc.createElement("TD");
        newTDcol1.setAttribute("id", "node"+inode+"col1");
        newTDcol1.className = "fakelinkon";
        newTR.appendChild(newTDcol1);

        var newTexta = doc.createTextNode("");
        newTDcol1.appendChild(newTexta);

        // table data "node"+inode+"col2"
        var newTDcol2 = doc.createElement("TD");
        newTDcol2.setAttribute("id", "node"+inode+"col2");
        if (this.cmenu[inode] != "") {
            newTDcol2.className = "fakelinkcmenu";
        }
        newTR.appendChild(newTDcol2);

        var newTextb = doc.createTextNode(this.name[inode]);
        newTDcol2.appendChild(newTextb);

        var name = this.name[inode].replace(/\u00a0/g, "");

        var ibrch = 0;
        for (var jbrch = 0; jbrch < brch.length; jbrch++) {
            if (brch[jbrch].name == name) {
                if (brch[jbrch].ileft == -2) {
                    newTDcol2.className = "errorTD";
                }
                break;
            }
        }

        // table data "node"+inode+"col3"
        if (this.nprop[inode] > 0) {
            var newTDcol3 = doc.createElement("TD");
            newTDcol3.setAttribute("id", "node"+inode+"col3");
            if (this.cbck1[inode] != "") {
                newTDcol3.className = "fakelinkon";
            }
            newTR.appendChild(newTDcol3);

            if (this.nprop[inode] == 1) {
                newTDcol3.setAttribute("colspan", "3");
            }

            var newTextc = doc.createTextNode(this.prop1[inode]);
            newTDcol3.appendChild(newTextc);
        }

        // table data "node:+inode+"col4"
        if (this.nprop[inode] > 1) {
            var newTDcol4 = doc.createElement("TD");
            newTDcol4.setAttribute("id", "node"+inode+"col4");
            if (this.cbck2[inode] != "") {
                newTDcol4.className = "fakelinkon";
            }
            newTR.appendChild(newTDcol4);

            if (this.nprop[inode] == 2) {
                newTDcol4.setAttribute("colspan", "2");
            }

            var newTextd = doc.createTextNode(this.prop2[inode]);
            newTDcol4.appendChild(newTextd);
        }

        // table data "node:+inode+"col5"
        if (this.nprop[inode] > 2) {
            var newTDcol5 = doc.createElement("TD");
            newTDcol5.setAttribute("id", "node"+inode+"col5");
            if (this.cbck3[inode] != "") {
                newTDcol5.className = "fakelinkon";
            }
            newTR.appendChild(newTDcol5);

            var newTextd = doc.createTextNode(this.prop3[inode]);
            newTDcol5.appendChild(newTextd);
        }

        // go to next row
        if        (this.child[inode] >= 0) {
            inode = this.child[inode];
        } else if (this.next[inode] >= 0) {
            inode = this.next[inode];
        } else {
            while (inode > 0) {
                inode = this.parent[inode];
                if (this.parent[inode] == 0) {
                    newTR = doc.createElement("TR");
                    newTR.setAttribute("height", "10px");
                    newTable.appendChild(newTR);
                }
                if (this.next[inode] >= 0) {
                    inode = this.next[inode];
                    break;
                }
            }
        }
    }

    this.update();
}


//
// clear the Tree
//
function TreeClear() {
    // alert("in TreeClear()");

    // remove all but the first Node
    this.name.splice(  1);
    this.gprim.splice( 1);
    this.click.splice( 1);
    this.cmenu.splice( 1);
    this.parent.splice(1);
    this.child.splice( 1);
    this.next.splice(  1);
    this.nprop.splice( 1);
    this.opened.splice(1);

    this.prop1.splice(1);
    this.valu1.splice(1);
    this.cbck1.splice(1);
    this.prop2.splice(1);
    this.valu2.splice(1);
    this.cbck2.splice(1);
    this.prop3.splice(1);
    this.valu3.splice(1);
    this.cbck3.splice(1);

    // reset the root Node
    this.parent[0] = -1;
    this.child[ 0] = -1;
    this.next[  0] = -1;
}


//
// expand a Node in the Tree
//
function TreeContract(inode) {
    // alert("in TreeContract(inode="+inode+")");

    // validate inputs
    if (inode < 0 || inode >= this.opened.length) {
        alert("inode="+inode+" is out of range");
        return;
    }

    // contract inode
    this.opened[inode] = 0;

    // contract all descendents of inode
    for (var jnode = 1; jnode < this.parent.length; jnode++) {
        var iparent = this.parent[jnode];
        while (iparent > 0) {
            if (iparent == inode) {
                this.opened[jnode] = 0;
                break;
            }

            iparent = this.parent[iparent];
        }
    }

    // update the display
    this.update();
}


//
// expand a Node in the Tree
//
function TreeExpand(inode) {
    // alert("in TreeExpand(inode="+inode+")");

    // validate inputs
    if (inode < 0 || inode >= this.opened.length) {
        alert("inode="+inode+" is out of range");
        return;
    }

    // expand inode
    this.opened[inode] = 1;

    // update the display
    this.update();
}


//
// change a property of a Node
//
function TreeProp(inode, iprop, onoff) {
    // alert("in TreeProp(inode="+inode+", iprop="+iprop+", onoff="+onoff+")");

    // validate inputs
    if (inode < 0 || inode >= this.opened.length) {
        alert("inode="+inode+" is out of range");
        return;
    } else if (onoff != "on" && onoff != "off") {
        alert("onoff="+onoff+" is not 'on' or 'off'");
        return;
    }

    // set the property for inode
    if (iprop == 1) {
        this.valu1[inode] = onoff;
    } else if (iprop == 2) {
        this.valu2[inode] = onoff;
    } else if (iprop == 3) {
        this.valu3[inode] = onoff;
    } else {
        alert("iprop="+iprop+" is not 1, 2, or 3");
        return;
    }

    // set property of all descendents of inode
    for (var jnode = 1; jnode < this.parent.length; jnode++) {
        var iparent = this.parent[jnode];
        while (iparent > 0) {
            if (iparent == inode) {
                if        (iprop == 1) {
                    this.valu1[jnode] = onoff;
                } else if (iprop == 2) {
                    this.valu2[jnode] = onoff;
                } else if (iprop == 3) {
                    this.valu3[jnode] = onoff;
                }
                break;
            }

            iparent = this.parent[iparent];
        }
    }

    this.update();
}


//
// update the Tree (after build/expension/contraction/property-set)
//
function TreeUpdate() {
    // alert("in TreeUpdate()");

    var doc = this.document;

    // traverse the Nodes using depth-first search
    for (var inode = 1; inode < this.opened.length; inode++) {
        var element = doc.getElementById("node"+inode);

        // unhide the row
        element.style.display = "table-row";

        // hide the row if one of its parents has .opened=0
        var jnode = this.parent[inode];
        while (jnode != 0) {
            if (this.opened[jnode] == 0) {
                element.style.display = "none";
                break;
            }

            jnode = this.parent[jnode];
        }

        // if the current Node has children, set up appropriate event handler to expand/collapse
        if (this.child[inode] > 0) {
            if (this.opened[inode] == 0) {
                var myElem = doc.getElementById("node"+inode+"col1");
                var This   = this;

                myElem.firstChild.nodeValue = "+";
                myElem.onclick = function () {
                    var thisNode = this.id.substring(4);
                    thisNode     = thisNode.substring(0,thisNode.length-4);
                    This.expand(thisNode);
                };

            } else {
                var myElem = doc.getElementById("node"+inode+"col1");
                var This   = this;

                myElem.firstChild.nodeValue = "-";
                myElem.onclick = function () {
                    var thisNode = this.id.substring(4);
                    thisNode     = thisNode.substring(0,thisNode.length-4);
                    This.contract(thisNode);
                };
            }
        }

        if (this.click[inode] !== null) {
            var myElem = doc.getElementById("node"+inode+"col2");
            myElem.onclick = this.click[inode];
        }

        // set the class of the properties
        if (this.nprop[inode] >= 1) {
            var myElem = doc.getElementById("node"+inode+"col3");
            myElem.onclick = this.cbck1[inode];

            if (this.prop1[inode] == "Viz") {
                if (this.valu1[inode] == "off") {
                    myElem.setAttribute("class", "fakelinkoff");
                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs &= ~wv.plotAttrs.ON;
                        wv.sceneUpd = 1;
                    }
                } else {
                    myElem.setAttribute("class", "fakelinkon");
                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs |=  wv.plotAttrs.ON;
                        wv.sceneUpd = 1;
                    }
                }
            }
        }

        if (this.nprop[inode] >= 2) {
            var myElem = doc.getElementById("node"+inode+"col4");
            myElem.onclick = this.cbck2[inode];

            if (this.prop2[inode] == "Grd") {
                if (this.valu2[ inode] == "off") {
                    myElem.setAttribute("class", "fakelinkoff");

                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs &= ~wv.plotAttrs.LINES;
                        wv.sceneGraph[this.gprim[inode]].attrs &= ~wv.plotAttrs.POINTS;
                        wv.sceneUpd = 1;
                    }
                } else {
                    myElem.setAttribute("class", "fakelinkon");

                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs |=  wv.plotAttrs.LINES;
                        wv.sceneGraph[this.gprim[inode]].attrs |=  wv.plotAttrs.POINTS;
                        wv.sceneUpd = 1;
                    }
                }
            }
        }

        if (this.nprop[inode] >= 3) {
            var myElem = doc.getElementById("node"+inode+"col5");
            myElem.onclick = this.cbck3[inode];

            if (this.prop3[inode] == "Trn") {
                if (this.valu3[ inode] == "off") {
                    myElem.setAttribute("class", "fakelinkoff");

                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs &= ~wv.plotAttrs.TRANSPARENT;
                        wv.sceneUpd = 1;
                    }
                } else {
                    myElem.setAttribute("class", "fakelinkon");

                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs |=  wv.plotAttrs.TRANSPARENT;
                        wv.sceneUpd = 1;
                    }
                }
            } else if (this.prop3[inode] == "Ori") {
                if (this.valu3[ inode] == "off") {
                    myElem.setAttribute("class", "fakelinkoff");

                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs &= ~wv.plotAttrs.ORIENTATION;
                        wv.sceneUpd = 1;
                    }
                } else {
                    myElem.setAttribute("class", "fakelinkon");

                    if (this.gprim[inode] != "") {
                        wv.sceneGraph[this.gprim[inode]].attrs |=  wv.plotAttrs.ORIENTATION;
                        wv.sceneUpd = 1;
                    }
                }
            }
        }
    }
}


//
// callback when "onresize" event occurs (called by ESP.html)
//
// resize the frames (with special handling to width of tlframe and height of brframe)
//
function resizeFrames() {
    // alert("resizeFrames()");

    var scrollSize = 24;

    // get the size of the client (minus amount to account for scrollbars)
    var body = document.getElementById("mainBody");
    var bodyWidth  = body.clientWidth  - scrollSize;
    var bodyHeight = body.clientHeight - scrollSize;

    // get the elements associated with the frames and the canvas
    var topleft = document.getElementById("tlframe");
    var butnfrm = document.getElementById("butnfrm");
    var treefrm = document.getElementById("treefrm");
    var toprite = document.getElementById("trframe");
    var botleft = document.getElementById("blframe");
    var botrite = document.getElementById("brframe");
    var canvas  = document.getElementById(wv.canvasID);
    var sketch  = document.getElementById("sketcher");

    // compute and set the widths of the frames
    //    (do not make tlframe larger than 250px)
    var leftWidth = Math.round(0.25 * bodyWidth);
    if (leftWidth > 250)   leftWidth = 250;
    var riteWidth = bodyWidth - leftWidth;
    var canvWidth = riteWidth - scrollSize;

    topleft.style.width = leftWidth+"px";
    butnfrm.style.width = leftWidth+"px";
    treefrm.style.width = leftWidth+"px";
    toprite.style.width = riteWidth+"px";
    botleft.style.width = leftWidth+"px";
    botrite.style.width = riteWidth+"px";
    canvas.style.width  = canvWidth+"px";
    canvas.width        = canvWidth;
    sketch.style.width  = canvWidth+"px";
    sketch.width        = canvWidth;

    // compute and set the heights of the frames
    //    (do not make brframe larger than 200px)
    var botmHeight = Math.round(0.20 * bodyHeight);
    if (botmHeight > 200)   botmHeight = 200;
    var  topHeight = bodyHeight - botmHeight;
    var canvHeight =  topHeight - scrollSize - 5;
    var keyHeight  = botmHeight - 25;

    topleft.style.height =  topHeight+"px";
    treefrm.style.height = (topHeight-90)+"px";
    toprite.style.height =  topHeight+"px";
    botleft.style.height = botmHeight+"px";
    botrite.style.height = botmHeight+"px";
    canvas.style.height  = canvHeight+"px";
    canvas.height        = canvHeight;
    sketch.style.height  = canvHeight+"px";
    sketch.height        = canvHeight;

    // set up canvas associated with key
    if (wv.canvasKY !== undefined) {
        var canf = document.getElementById(wv.canvasKY);

        var keyfWidth     = leftWidth - 20;
        canf.style.width  = keyfWidth + "px";
        canf.width        = keyfWidth;

        var keyfHeight    = botmHeight - 25;
        canf.style.height = keyfHeight + "px";
        canf.height       = keyfHeight;

        // force a key redraw
        wv.drawKey = 1;
    }
}


//
// change mode for trframe
//
function changeMode(newMode) {
    // alert("in changeMode(newMode="+newMode+")");

    var webViewer      = document.getElementById("WebViewer");
    var addBrchForm    = document.getElementById("addBrchForm");
    var editBrchForm   = document.getElementById("editBrchForm");
    var addBrchHeader  = document.getElementById("addBrchHeader");
    var editBrchHeader = document.getElementById("editBrchHeader");
    var editPmtrForm   = document.getElementById("editPmtrForm");
    var addPmtrHeader  = document.getElementById("addPmtrHeader");
    var editPmtrHeader = document.getElementById("editPmtrHeader");
    var editCsmForm    = document.getElementById("editCsmForm");
    var sketcherForm   = document.getElementById("sketcherForm");

    var wvKey          = document.getElementById("WVkey");
    var sketchStatus   = document.getElementById("SketchStatus");
    var ESPlogo        = document.getElementById("ESPlogo");

    if (newMode == wv.curMode) {
        return;
    } else if (newMode < 0) {
        // used to cause buttons such as "cmdSave" not to be active
        wv.curMode = newMode;
        return;
    } else if (newMode == 0) {
        webViewer.hidden    = false;
        addBrchForm.hidden  = true;
        editBrchForm.hidden = true;
        editPmtrForm.hidden = true;
        editCsmForm.hidden  = true;
        sketcherForm.hidden = true;
        wvKey.hidden        = true;
        sketchStatus.hidden = true;
        ESPlogo.hidden      = false;

        var saveButton = document.getElementById("saveButton");
        saveButton["innerHTML"] = "Save File";

        wv.curMode   = 0;
        wv.curPmtr   = -1;
        wv.curBrch   = -1;
        wv.afterBrch = -1;
        wv.menuEvent = undefined;
    } else if (newMode == 1) {
        webViewer.hidden    = true;
        addBrchForm.hidden  = false;
        editBrchForm.hidden = true;
        editPmtrForm.hidden = true;
        editCsmForm.hidden  = true;
        sketcherForm.hidden = true;
        wvKey.hidden        = true;
        sketchStatus.hidden = true;
        ESPlogo.hidden      = false;

        // unselect all items
        var elements = document.getElementById("addBrchForm").elements;
        for (var ielem = 0; ielem < elements.length; ielem++) {
            elements[ielem].checked = false;
        }

        wv.curMode = 1;
    } else if (newMode == 2) {
        webViewer.hidden    = true;
        addBrchForm.hidden  = true;
        editBrchForm.hidden = false;
        editPmtrForm.hidden = true;
        editCsmForm.hidden  = true;
        sketcherForm.hidden = true;
        wvKey.hidden        = true;
        sketchStatus.hidden = true;
        ESPlogo.hidden      = false;

        addBrchHeader.hidden  = false;
        editBrchHeader.hidden = true;

        wv.curMode = 2;
    } else if (newMode == 3) {
        webViewer.hidden    = true;
        addBrchForm.hidden  = true;
        editBrchForm.hidden = false;
        editPmtrForm.hidden = true;
        editCsmForm.hidden  = true;
        sketcherForm.hidden = true;
        wvKey.hidden        = true;
        sketchStatus.hidden = true;
        ESPlogo.hidden      = false;

        addBrchHeader.hidden  = true;
        editBrchHeader.hidden = false;

        wv.curMode = 3;
    } else if (newMode == 4) {
        webViewer.hidden    = true;
        addBrchForm.hidden  = true;
        editBrchForm.hidden = true;
        editPmtrForm.hidden = false;
        editCsmForm.hidden  = true;
        sketcherForm.hidden = true;
        wvKey.hidden        = true;
        sketchStatus.hidden = true;
        ESPlogo.hidden      = false;

        addPmtrHeader.hidden  = false;
        editPmtrHeader.hidden = true;

        wv.curMode = 4;
    } else if (newMode == 5) {
        webViewer.hidden    = true;
        addBrchForm.hidden  = true;
        editBrchForm.hidden = true;
        editPmtrForm.hidden = false;
        editCsmForm.hidden  = true;
        sketcherForm.hidden = true;
        wvKey.hidden        = true;
        sketchStatus.hidden = true;
        ESPlogo.hidden      = false;

        addPmtrHeader.hidden  = true;
        editPmtrHeader.hidden = false;

        wv.curMode = 5;
    } else if (newMode == 6) {
        webViewer.hidden    = true;
        addBrchForm.hidden  = true;
        editBrchForm.hidden = true;
        editPmtrForm.hidden = true;
        editCsmForm.hidden  = false;
        sketcherForm.hidden = true;
        wvKey.hidden        = true;
        sketchStatus.hidden = true;
        ESPlogo.hidden      = false;

        wv.curMode = 6;
    } else if (newMode == 7) {
        webViewer.hidden    = true;
        addBrchForm.hidden  = true;
        editBrchForm.hidden = true;
        editPmtrForm.hidden = true;
        editCsmForm.hidden  = true;
        sketcherForm.hidden = false;
        wvKey.hidden        = true;
        sketchStatus.hidden = false;
        ESPlogo.hidden      = true;

        var saveButton = document.getElementById("saveButton");
        saveButton["innerHTML"] = "Save Sketch";

        wv.curMode = 7;
    } else {
        alert("Bad new mode = "+newMode);
    }

    if (wv.curMode == 0) {
        var editButton = document.getElementById("editButton");
        var quitButton = document.getElementById("quitButton");

        editButton.hidden = false;
        quitButton.hidden = true;
    } else if (wv.curMode == 7) {
        var editButton = document.getElementById("editButton");
        var quitButton = document.getElementById("quitButton");

        editButton.hidden = true;
        quitButton.hidden = false;
    }
}


//
// post a message into the brframe
//
function postMessage(mesg) {
    // alert("in postMessage(mesg="+mesg+")");

    var botm = document.getElementById("brframe");

    var pre  = document.createElement("pre");
    var text = document.createTextNode(mesg);
    pre.appendChild(text);
    botm.insertBefore(pre, botm.lastChild);

    pre.scrollIntoView();
}


//
// load info into editBrchForm
//
function setupEditBrchForm() {
    // alert("setupEditBrchForm()");

    var ibrch = wv.curBrch;
    var name  = brch[ibrch].name;
    var type  = brch[ibrch].type;

    var editBrchForm = document.getElementById("editBrchForm");

    // turn all arguments off (by default)
    document.getElementById("argName1").parentNode.style.display = "none";
    document.getElementById("argName2").parentNode.style.display = "none";
    document.getElementById("argName3").parentNode.style.display = "none";
    document.getElementById("argName4").parentNode.style.display = "none";
    document.getElementById("argName5").parentNode.style.display = "none";
    document.getElementById("argName6").parentNode.style.display = "none";
    document.getElementById("argName7").parentNode.style.display = "none";
    document.getElementById("argName8").parentNode.style.display = "none";
    document.getElementById("argName9").parentNode.style.display = "none";

    // fill in the Branch type and name
    document.getElementById("brchType").firstChild["data"] = type;
    editBrchForm.            brchName.value                = name;

    // by default, the "Enter Sketcher" button is turned off
    document.getElementById("EnterSketcher").style.display = "none";

    // set up arguments based upon the type
    var argList;
    var defValue;
    var suppress = 0;   // set to 1 if type can be suppressed

    if        (type == "arc") {
        argList  = ["xend", "yend", "zend", "dist", "$plane"];
        defValue = ["",     "",     "",     "",     "xy"    ];
        document.getElementById("EnterSketcher").style.display = "inline";
    } else if (type == "bezier") {
        argList  = ["x", "y", "z"];
        defValue = ["",  "",  "" ];
        document.getElementById("EnterSketcher").style.display = "inline";
    } else if (type == "blend") {
        argList  = ["noseList", "tailList", "reorder"];
        defValue = ["0",        "0",        "0"      ];
        suppress = 1;
    } else if (type == "box") {
        argList  = ["xmin", "ymin", "zmin", "dx", "dy", "dz"];
        defValue = ["",     "",     "",     "",   "",   ""  ];
        suppress = 1;
    } else if (type == "chamfer") {
        argList  = ["radius", "edgeList"];
        defValue = ["",       "0"       ];
        suppress = 1;
    } else if (type == "cirarc") {
        argList  = ["xon", "yon", "zon", "xend", "yend", "zend"];
        defValue = ["",    "",    "",    "",     "",     ""    ];
    } else if (type == "combine") {
        argList  = [];
        defValue = [];
    } else if (type == "cone") {
        argList  = ["xvrtx", "yvrtx", "zvrtx", "xbase", "ybase", "zbase", "radius"];
        defValue = ["",      "",      "",      "",      "",      "",      ""      ];
        suppress = 1;
    } else if (type == "connect") {
        argList  = ["faceList1", "faceList2"];
        defValue = ["",          ""         ];
    } else if (type == "cylinder") {
        argList  = ["xbeg", "ybeg", "zbeg", "xend", "yend", "zend", "radius"];
        defValue = ["",     "",     "",     "",     "",     "",     ""      ];
        suppress = 1;
    } else if (type == "dump") {
        argList  = ["$filename", "remove"];
        defValue = ["",          "0"     ];
    } else if (type == "end") {
        argList  = [];
        defValue = [];
    } else if (type == "extract") {
        argList  = ["index"];
        defValue = [""     ];
    } else if (type == "extrude") {
        argList  = ["dx", "dy", "dz"];
        defValue = ["",   "",   ""  ];
        suppress = 1;
    } else if (type == "fillet") {
        argList  = ["radius", "edgeList"];
        defValue = ["",       "0"       ];
        suppress = 1;
    } else if (type == "hollow") {
        argList  = ["thick", "faceList"];
        defValue = ["",      "0"       ];
        suppress = 1;
    } else if (type == "import") {
        argList  = ["$filename", "bodynumber"];
        defValue = ["",          "1"         ];
        suppress = 1;
    } else if (type == "intersect") {
        argList  = ["$order", "index", "maxtol"];
        defValue = ["none",   "1",     "0"     ];
    } else if (type == "join") {
        argList  = ["toler"];
        defValue = ["0"    ];
    } else if (type == "linseg") {
        argList  = ["x", "y", "z"];
        defValue = ["",  "",  "" ];
        document.getElementById("EnterSketcher").style.display = "inline";
    } else if (type == "loft") {
        alert("Consider using 'rule' or 'blend'");
        argList  = ["smooth"];
        defValue = [""      ];
        suppress = 1;
    } else if (type == "macbeg") {
        alert("'macbeg' is deprecated");
        argList  = ["imacro"];
        defValue = [""      ];
    } else if (type == "macend") {
        alert("'macend' is deprecated");
        argList  = [];
        defValue = [];
    } else if (type == "mark") {
        argList  = [];
        defValue = [];
    } else if (type == "mirror") {
        argList  = ["nx", "ny", "nz", "dist"];
        defValue = ["",   "",   "",   "0"   ];
        suppress = 1;
    } else if (type == "patbeg") {
        argList  = ["$pmtrName", "ncopy"];
        defValue = ["",          ""     ];
    } else if (type == "patbreak") {
        argList  = ["expr"];
        defValue = [""    ];
    } else if (type == "patend") {
        argList  = [];
        defValue = [];
    } else if (type == "project") {
        argList  = ["x", "y", "z", "dx", "dy", "dz"];
        defValue = ["",  "",  "",  "",   "",   ""  ];
    } else if (type == "recall") {
        aleft("'recall' is deprecated");
        argList  = ["imicro"];
        defValue = [""      ];
    } else if (type == "reorder") {
        argList  = ["ishift", "iflip"];
        defValue = ["",       "0"    ];
        suppress = 1;
    } else if (type == "restore") {
        argList  = ["$name", "index"];
        defValue = ["",      "0"    ];
    } else if (type == "revolve") {
        argList  = ["xorig", "yorig", "zorig", "dxaxis", "dyaxis", "dzaxis", "angDeg"];
        defValue = ["",      "",      "",      "",       "",       "",       ""      ];
        suppress = 1;
    } else if (type == "rotatex") {
        argList  = ["angDeg", "yaxis", "zaxis"];
        defValue = ["",       "",      ""     ];
        suppress = 1;
    } else if (type == "rotatey") {
        argList  = ["angDeg", "zaxis", "xaxis"];
        defValue = ["",       "",      ""     ];
        suppress = 1;
    } else if (type == "rotatez") {
        argList  = ["angDeg", "xaxis", "yaxis"];
        defValue = ["",       "",      ""     ];
        suppress = 1;
    } else if (type == "rule") {
        argList  = ["reorder"];
        defValue = ["0"      ];
        suppress = 1;
    } else if (type == "scale") {
        argList  = ["fact"];
        defValue = [""    ];
        suppress = 1;
    } else if (type == "select") {
        argList  = ["$type", "arg1", "arg2", "arg3", "arg4", "arg5", "arg6", "arg7", "arg8"];
        defValue = ["",      "",     "",     "",     "",     "",     "",     "",     ""    ];
    } else if (type == "set") {
        argList  = ["$pmtrName", "exprs"];
        defValue = ["",          ""     ];
    } else if (type == "skbeg") {
        argList  = ["x", "y", "z"];
        defValue = ["",  "",  "" ];
        document.getElementById("EnterSketcher").style.display = "inline";
    } else if (type == "skcon") {
        argList  = ["$type", "index1", "index2", "$value"];
        defValue = ["",      "",       "-1",     "0"     ];
        document.getElementById("EnterSketcher").style.display = "inline";
    } else if (type == "skend") {
        argList  = ["wireOnly"];
        defValue = ["0"];
        document.getElementById("EnterSketcher").style.display = "inline";
    } else if (type == "skvar") {
        argList  = ["$type", "valList"];
        defValue = ["",      ""       ];
        document.getElementById("EnterSketcher").style.display = "inline";
    } else if (type == "solbeg") {
        argList  = ["$varList"];
        defValue = [""        ];
    } else if (type == "solcon") {
        argList  = ["$expr"];
        defValue = [""     ];
    } else if (type == "solend") {
        argList  = [];
        defValue = [];
    } else if (type == "sphere") {
        argList  = ["xcent", "ycent", "zcent", "radius"];
        defValue = ["",      "",      "",      ""      ];
        suppress = 1;
    } else if (type == "spline") {
        argList  = ["x", "y", "z"];
        defValue = ["",  "",  "" ];
        document.getElementById("EnterSketcher").style.display = "inline";
    } else if (type == "store") {
        argList  = ["$name", "index", "keep"];
        defValue = ["",      "0" ,    "0"   ];
    } else if (type == "subtract") {
        argList  = ["$order", "index", "maxtol"];
        defValue = ["none",   "1",     "0"     ];
    } else if (type == "sweep") {
        argList  = [];
        defValue = [];
        suppress = 1;
    } else if (type == "torus") {
        argList  = ["xcent", "ycent", "zcent", "dxaxis", "dyaxis", "dzaxis", "majorRad", "minorRad"];
        defValue = ["",      "",      "",      "",       "",       "",       "",         ""        ];
        suppress = 1;
    } else if (type == "translate") {
        argList  = ["dx", "dy", "dz"];
        defValue = ["",   "",   ""  ];
        suppress = 1;
    } else if (type == "udparg") {
        argList  = ["$primtype", "$argName1", "argValue1", "$argName2", "argValue2", "$argName3", "argValue3", "$argName4", "argValue4"];
        defValue = ["",          "",          "",          "",          "",          "",          "",          "",          ""         ];
    } else if (type == "udprim") {
        argList  = ["$primtype", "$argName1", "argValue1", "$argName2", "argValue2", "$argName3", "argValue3", "$argName4", "argValue4"];
        defValue = ["",          "",          "",          "",          "",          "",          "",          "",          ""         ];
        suppress = 1;
    } else if (type == "union") {
        argList  = ["tomark", "trimList", "maxtol"];
        defValue = ["0",      "0",        "0"     ];
    } else {
        return 1;
    }

    // set up the activity
    if (suppress) {
        if        (brch[ibrch].actv == 300) {     // Active
            editBrchForm.activity.value = 300;
        } else if (brch[ibrch].actv == 301) {     // Suppressed
            editBrchForm.activity.value = 301;
        } else {                                  // Inactive or Deferred
            editBrchForm.activity.value = 300;
        }
        editBrchForm.activity.style.display = "table-row";
    } else {
        editBrchForm.activity.style.display = "none";
    }

    // set up the arguments
    wv.numArgs = argList.length;

    if (wv.numArgs >= 1) {
        document.getElementById("argName1").parentNode.style.display = "table-row";
        document.getElementById("argName1").firstChild["data"]       = argList[0];
        if (wv.curMode == 1 || brch[ibrch].args[0] === undefined) {
            editBrchForm.        argValu1.value = defValue[0];
        } else if (argList[0].charAt(0) == "$") {
            editBrchForm.        argValu1.value = brch[ibrch].args[0].substr(1,brch[ibrch].args[0].length);
        } else {
            editBrchForm.        argValu1.value = brch[ibrch].args[0];
        }
    }
    if (wv.numArgs >= 2) {
        document.getElementById("argName2").parentNode.style.display = "table-row";
        document.getElementById("argName2").firstChild["data"]       = argList[1];
        if (wv.curMode == 1 || brch[ibrch].args[1] === undefined) {
            editBrchForm.        argValu2.value = defValue[1];
        } else if (argList[1].charAt(0) == "$") {
            editBrchForm.        argValu2.value = brch[ibrch].args[1].substr(1,brch[ibrch].args[1].length);
        } else {
            editBrchForm.        argValu2.value = brch[ibrch].args[1];
        }
    }
    if (wv.numArgs >= 3) {
        document.getElementById("argName3").parentNode.style.display = "table-row";
        document.getElementById("argName3").firstChild["data"]       = argList[2];
        if (wv.curMode == 1 || brch[ibrch].args[2] === undefined) {
            editBrchForm.        argValu3.value = defValue[2];
        } else if (argList[2].charAt(0) == "$") {
            editBrchForm.        argValu3.value = brch[ibrch].args[2].substr(1,brch[ibrch].args[2].length);
        } else {
            editBrchForm.        argValu3.value = brch[ibrch].args[2];
        }
    }
    if (wv.numArgs >= 4) {
        document.getElementById("argName4").parentNode.style.display = "table-row";
        document.getElementById("argName4").firstChild["data"]       = argList[3];
        if (wv.curMode == 1 || brch[ibrch].args[3] === undefined) {
            editBrchForm.        argValu4.value = defValue[3];
        } else if (argList[3].charAt(0) == "$") {
            editBrchForm.        argValu4.value = brch[ibrch].args[3].substr(1,brch[ibrch].args[3].length);
        } else {
            editBrchForm.        argValu4.value = brch[ibrch].args[3];
        }
    }
    if (wv.numArgs >= 5) {
        document.getElementById("argName5").parentNode.style.display = "table-row";
        document.getElementById("argName5").firstChild["data"]       = argList[4];
        if (wv.curMode == 1 || brch[ibrch].args[4] === undefined) {
            editBrchForm.        argValu5.value = defValue[4];
        } else if (argList[4].charAt(0) == "$") {
            editBrchForm.        argValu5.value = brch[ibrch].args[4].substr(1,brch[ibrch].args[4].length);
        } else {
            editBrchForm.        argValu5.value = brch[ibrch].args[4];
        }
    }
    if (wv.numArgs >= 6) {
        document.getElementById("argName6").parentNode.style.display = "table-row";
        document.getElementById("argName6").firstChild["data"]       = argList[5];
        if (wv.curMode == 1 || brch[ibrch].args[5] === undefined) {
            editBrchForm.        argValu6.value = defValue[5];
        } else if (argList[5].charAt(0) == "$") {
            editBrchForm.        argValu6.value = brch[ibrch].args[5].substr(1,brch[ibrch].args[5].length);
        } else {
            editBrchForm.        argValu6.value = brch[ibrch].args[5];
        }
    }
    if (wv.numArgs >= 7) {
        document.getElementById("argName7").parentNode.style.display = "table-row";
        document.getElementById("argName7").firstChild["data"]       = argList[6];
        if (wv.curMode == 1 || brch[ibrch].args[6] === undefined) {
            editBrchForm.        argValu7.value = defValue[6];
        } else if (argList[6].charAt(0) == "$") {
            editBrchForm.        argValu7.value = brch[ibrch].args[6].substr(1,brch[ibrch].args[6].length);
        } else {
            editBrchForm.        argValu7.value = brch[ibrch].args[6];
        }
    }
    if (wv.numArgs >= 8) {
        document.getElementById("argName8").parentNode.style.display = "table-row";
        document.getElementById("argName8").firstChild["data"]       = argList[7];
        if (wv.curMode == 1 || brch[ibrch].args[7] === undefined) {
            editBrchForm.        argValu8.value = defValue[7];
        } else if (argList[7].charAt(0) == "$") {
            editBrchForm.        argValu8.value = brch[ibrch].args[7].substr(1,brch[ibrch].args[7].length);
        } else {
            editBrchForm.        argValu8.value = brch[ibrch].args[7];
        }
    }
    if (wv.numArgs >= 9) {
        document.getElementById("argName9").parentNode.style.display = "table-row";
        document.getElementById("argName9").firstChild["data"]       = argList[8];
        if (wv.curMode == 1 || brch[ibrch].args[8] === undefined) {
            editBrchForm.        argValu9.value = defValue[8];
        } else if (argList[8].charAt(0) == "$") {
            editBrchForm.        argValu9.value = brch[ibrch].args[8].substr(1,brch[ibrch].args[8].length);
        } else {
            editBrchForm.        argValu9.value = brch[ibrch].args[8];
        }
    }

    return 0;
}


//
// load info into editPmtrForm
//
function setupEditPmtrForm() {
    // alert("setupEditPmtrForm()");

    var ipmtr = wv.curPmtr;
    var name  = pmtr[ipmtr].name;
    var nrow  = pmtr[ipmtr].nrow;
    var ncol  = pmtr[ipmtr].ncol;

    var editPmtrForm = document.getElementById("editPmtrForm");

    // fill in the Parameter name
    document.getElementById("pmtrName").firstChild["data"] = name;

    var pmtrTable = document.getElementById("editPmtrTable");

    // remove old table entries
    if (pmtrTable) {
        var child1 = pmtrTable.lastChild;
        while (child1) {
            var child2 = child1.lastChild;
            while (child2) {

                var child3 = child2.lastChild;
                while (child3) {
                    child2.removeChild(child3);
                    child3 = child2.lastChild;
                }
                child1.removeChild(child2);
                child2 = child1.lastChild;
            }
            pmtrTable.removeChild(child1);
            child1 = pmtrTable.lastChild;
        }
    }

    // build the table that will contain values
    for (var irow = 0; irow <= nrow; irow++) {
        var newTR = document.createElement("TR");
        pmtrTable.appendChild(newTR);

        // fill the row
        if (irow == 0) {
            var newTD = document.createElement("TD");
            newTR.appendChild(newTD);

            var newText = document.createTextNode("");
            newTD.appendChild(newText);

            for (var icol = 1; icol <= ncol; icol++) {
                newTD = document.createElement("TD");
                newTR.appendChild(newTD);

                newText = document.createTextNode("col\u00a0"+icol);
                newTD.appendChild(newText);
            }
        } else{
            var newTD = document.createElement("TD");
            newTR.appendChild(newTD);

            var newText = document.createTextNode("row\u00a0"+irow);
            newTD.appendChild(newText);

            for (var icol = 1; icol <= ncol; icol++) {
                var indx = icol-1 + (irow-1)*pmtr[ipmtr].ncol;

                newTD = document.createElement("TD");
                newTR.appendChild(newTD);

                var newInput = document.createElement("input");
                newInput.type  = "text";
                newInput.name  = "row"+irow+"col"+icol;
                newInput.size  = 12;
                newInput.value = pmtr[ipmtr].value[indx];
                newTD.appendChild(newInput);
            }
        }
    }

    return 0;
}


//
// count number of Parameter value changes
//
function numberOfPmtrChanges() {
    // alert("in numberOfPmtrChanges()");

    var nchange = 0;

    var editPmtrForm = document.getElementById("editPmtrForm");

    var ipmtr = wv.curPmtr;

    // examine each of the values
    var index   = -1;
    for (var irow = 1; irow <= pmtr[ipmtr].nrow; irow++) {
        for (var icol = 1; icol <= pmtr[ipmtr].ncol; icol++) {
            index++;

            // get the new value
            var myInput = editPmtrForm["row"+irow+"col"+icol];
            var value   = myInput.value.replace(/\s/g, "");

            if (value != pmtr[ipmtr].value[index]) {
                nchange++;
            }
        }
    }

    // return the number of changes
    return nchange;
}


//
// count number of Branch changes
//
function numberOfBrchChanges() {
    // alert("in numberOfBrchChanges()");

    var nchange = 0;

    var editBrchForm = document.getElementById("editBrchForm");

    var ibrch = wv.curBrch;

    // check the name
    if (brch[ibrch].name != editBrchForm.brchName.value.replace(/\s/g, "")) {
        nchange++;
    }

    // check the activity
    if (brch[ibrch].actv != editBrchForm.activity.value) {
        nchange++;
    }

    // check the arguments
    for (var iarg = 0; iarg < wv.numArgs; iarg++) {
        if        (iarg == 0) {
            var name  = document.getElementById("argName1").firstChild["data"];
            var value = editBrchForm.            argValu1.value.replace(/\s/g, "");
        } else if (iarg == 1) {
            var name  = document.getElementById("argName2").firstChild["data"];
            var value = editBrchForm.            argValu2.value.replace(/\s/g, "");
        } else if (iarg == 2) {
            var name  = document.getElementById("argName3").firstChild["data"];
            var value = editBrchForm.            argValu3.value.replace(/\s/g, "");
        } else if (iarg == 3) {
            var name  = document.getElementById("argName4").firstChild["data"];
            var value = editBrchForm.            argValu4.value.replace(/\s/g, "");
        } else if (iarg == 4) {
            var name  = document.getElementById("argName5").firstChild["data"];
            var value = editBrchForm.            argValu5.value.replace(/\s/g, "");
        } else if (iarg == 5) {
            var name  = document.getElementById("argName6").firstChild["data"];
            var value = editBrchForm.            argValu6.value.replace(/\s/g, "");
        } else if (iarg == 6) {
            var name  = document.getElementById("argName7").firstChild["data"];
            var value = editBrchForm.            argValu7.value.replace(/\s/g, "");
        } else if (iarg == 7) {
            var name  = document.getElementById("argName8").firstChild["data"];
            var value = editBrchForm.            argValu8.value.replace(/\s/g, "");
        } else if (iarg == 8) {
            var name  = document.getElementById("argName9").firstChild["data"];
            var value = editBrchForm.            argValu9.value.replace(/\s/g, "");
        }

        var output;
        if (name.charAt(0) != "$" || value.length <= 0) {
            output =       value;
        } else {
            output = "$" + value;
        }

        if (output != brch[ibrch].args[iarg]) {
            nchange++;
        }
    }

    // return the number of changes
    return nchange;
}


//
// unhighlight background colors in column 1 of myTree
//
function unhighlightColumn1() {
    // alert("in unhighlightColumn1()");

    for (var jnode = 0; jnode < myTree.name.length; jnode++) {
        var myElem = document.getElementById("node"+jnode+"col1");
        if (myElem) {
            if (myElem.className == "currentTD" || myElem.className == "parentTD" || myElem.className == "childTD") {
                myElem.className = undefined;
            }
        }
    }
}


//
// print an object and its contents
//
function printObject(obj) {
    var out = '';

    for (var p in obj) {
        out += p + ': ' + obj[p] + '\n';
    }

    alert(out);
}


//
// simple text formatter patterned after C's sprintf
//
function sprintf()
{

    // if no arguments, return an empty string
    var narg = arguments.length;
    if (narg == 0) {
        return "";
    }

    // otherwise, build output from input
    var format = arguments[0];
    var answer = "";
    var iarg   = 1;

    while (1) {
        var indx = format.indexOf("%");
        if (indx >= 0) {
            answer += format.substring(0, indx);
            if (iarg < narg) {
                answer += arguments[iarg++];
            } else {
                answer += "%";
            }
            format  = format.substring(indx+1);
        } else {
            answer += format;
            break;
        }
    }

    return answer;
}