/*
 ************************************************************************
 *                                                                      *
 * OpenCSM -- an open-source constructive solid modeler                 *
 *                                                                      *
 *            Written by John Dannenhoffer @ Syracuse University        *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2010/2015  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <float.h>
#include <math.h>

#include "egads.h"
#include "udp.h"

#define CINT    const int
#define CDOUBLE const double
#define CCHAR   const char

#ifdef WIN32
    #define snprintf _snprintf
    #define SLASH '\\'
#else
    #define SLASH '/'
#endif

#define STRNCPY(A, B, LEN) strncpy(A, B, LEN); A[LEN-1] = '\0';

#include "common.h"
#include "OpenCSM.h"

#if   defined(SHOW_SPLINES)
    #include "grafic.h"
#endif

#define MAX_NAME_LEN       32           /* maximum chars in name */
#define MAX_EXPR_LEN      512           /* maximum chars in expression */
#define MAX_LINE_LEN     2048           /* maximum chars in line in input file */
#define MAX_STR_LEN      4096           /* maximum chars in any string */
#define MAX_STACK_SIZE   1024           /* maximum size of stack */
#define MAX_SKETCH_SIZE  1024           /* maximum points in Sketch */
#define MAX_SOLVER_SIZE   256           /* maximum variable in solver */
#define MAX_NUM_SKETCHES  100           /* maximum number of Sketches in rule/loft/blend */
#define MAX_NUM_PATTERNS   10           /* maximum number of nested patterns */
#define MAX_NUM_MACROS    100           /* maximum number of storage locations */

#define OCSM_MAGIC        4433340       /* magic number */

/*
 ************************************************************************
 *                                                                      *
 * Structures                                                           *
 *                                                                      *
 ************************************************************************
 */

/* "Rpn" contains information associated with Rpn (pseudo-code) */
typedef struct {
    int    type;                       /* type (see below) */
    char   text[MAX_NAME_LEN];         /* associated text */
} rpn_T;

/* "Patn" contains information associated with a patbeg/patend or
          a macbeg/macend pair */
typedef struct {
    int    ipatbeg;                    /* Branch number of patbeg or recall */
    int    ipatend;                    /* Branch number of patend or recall */
    int    ncopy;                      /* total number of copies */
    int    icopy;                      /* current instance number (1->ncopy) */
    int    ipmtr;                      /* Parameter index of iterator */
} patn_T;

/* "Sket: is a Sketch */
typedef struct {
    int    type;                       /* 0 for old style, 1 for xy, 2 for yz, 3 for zx */
    int    size;                       /* size of x[], y[], and z[] */
    int    solved;                     /* =1 if Sketch is solved */
    int    ix;                         /* Parameter index associated with ::x[] */
    int    iy;                         /* Parameter index associated with ::y[] */
    int    iz;                         /* Parameter index associated with ::z[] */
    int    id;                         /* Parameter index associated with ::d[] */

    double xbase;                      /* x-coordinate in skbeg statement */
    double ybase;                      /* y-coordinate in skbeg statement */
    double zbase;                      /* z-coordinate in skbeg statement */

    int    nseg;                       /* number of segments */
    int    itype[MAX_SKETCH_SIZE];     /* Branch type */
    int    ibrch[MAX_SKETCH_SIZE];     /* Branch index */
    double x[MAX_SKETCH_SIZE];
    double y[MAX_SKETCH_SIZE];
    double z[MAX_SKETCH_SIZE];

    int    nvar;                       /* number of variables (at least 2*nseg) */
    int    ipmtr[MAX_SKETCH_SIZE];     /* Parameter index */
    int    index[MAX_SKETCH_SIZE];     /* value index */

    int    ncon;                       /* number of constraints */
    char   con[MAX_SKETCH_SIZE][MAX_EXPR_LEN];
} sket_T;

/*
 ************************************************************************
 *                                                                      *
 * Definitions (for structures above)                                   *
 *                                                                      *
 ************************************************************************
 */

#define           PARSE_NOP         0   /* no operation */
#define           PARSE_OP1         1   /* either add "+" or subtract "-" */
#define           PARSE_OP2         2   /* either multiply "*" or divide "/" */
#define           PARSE_OP3         3   /* exponentiation "^" */
#define           PARSE_OPENP       4   /* open  parenthesis "(" */
#define           PARSE_CLOSEP      5   /* close parenthesis ")" */
#define           PARSE_OPENB       6   /* open  bracket "[" */
#define           PARSE_CLOSEB      7   /* close bracket "]" */
#define           PARSE_COMMA       8   /* comma "," */
#define           PARSE_NAME        9   /* variable name */
#define           PARSE_ARRAY      10   /* array name */
#define           PARSE_FUNC       11   /* function name */
#define           PARSE_NUMBER     12   /* number */
#define           PARSE_STRING     13   /* string */
#define           PARSE_END        14   /* end of Rpn-code */

/*
 ************************************************************************
 *                                                                      *
 * Macros (including those that go along with common.h)                 *
 *                                                                      *
 ************************************************************************
 */

#if   defined DEBUG
    #define DOPEN {if (dbg_fp == NULL) dbg_fp = fopen("OpenCSM.dbg", "w");}
    static  FILE *dbg_fp=NULL;
#endif

static void *realloc_temp=NULL;              /* used by RALLOC macro */

/*
 ************************************************************************
 *                                                                      *
 * Declarations for support routines defined below                      *
 *                                                                      *
 ************************************************************************
 */

static int buildApplied(  modl_T *modl, int ibrch, varg_T args[], int *nstack, int stack[],
                          int npatn, patn_T patn[]);
static int buildBoolean(  modl_T *modl, int ibrch, varg_T args[], int *nstack, int stack[],
                          int npatn, patn_T patn[]);
static int buildGrown(    modl_T *modl, int ibrch, varg_T args[], int *nstack, int stack[],
                          int npatn, patn_T patn[]);
static int buildPrimitive(modl_T *modl, int ibrch, varg_T args[], int *nstack, int stack[],
                          int npatn, patn_T patn[]);
static int buildSketch(   modl_T *modl, int ibrch, varg_T args[], int *nstack, int stack[],
                          int npatn, patn_T patn[], sket_T *sket);
static int buildSolver(   modl_T *modl, int ibrch, varg_T args[], int *ncon, int solcons[]);
static int buildTransform(modl_T *modl, int ibrch, varg_T args[], int *nstack, int stack[]);
       int createPerturbation(modl_T *modl);
static int evalRpn(rpn_T *rpn, modl_T *modl, double *val, double *dot);
static int faceContains(ego eface, double xx, double yy, double zz);
static int finishBody(modl_T *modl, int ibody);
static int finiteDifference(modl_T *modl, int ibody, int seltype, int iselect, int npnt, /*@null@*/double uv[], double dxyz[]);
static int getBodyTolerance(ego ebody, double *toler);
static int getToken(char *text, int nskip, char sep, char *token);
static int makeEdge(modl_T *modl, ego ebeg, ego eend, ego *eedge);
static int makeFace(modl_T *modl, ego eedges[], int fillstyle, ego *eface);
static int matsol(double A[], double b[], int n, double x[]);
static int newBody(modl_T *modl, int ibrch, int brtype, int ileft, int irite, varg_T args[], int hasdots, int botype, int *ibody);
static int parseName(modl_T *modl, char name[], int *ipmtr, int *irow, int *icol);
static int printBodyAttributes(modl_T *modl, int ibody);
static void printEgo(ego obj);
static int recycleBody(modl_T *modl, int ibrch, int brtype, varg_T args[], int hasdots);
       int removePerturbation(modl_T *modl);
static int reorderLoops(modl_T *modl, int nloop, ego eloops[], int startFrom);
static int selectBody(ego emodel, char *order, int index);
static int setFaceAttribute(modl_T *modl, int ibody, int iface, int jbody, int jford, int npatn, patn_T *patn);
static int setupAtPmtrs(modl_T *modl);
static void signalError(/*@null@*/void *modl, int status, int ibrch, char format[], ...);
static int solidBoolean(ego ebodyl, ego ebodyr, int type, double maxtol, ego *emodel);
static int solveSketch(modl_T *modl, sket_T *sket);
static int str2rpn(char str[], rpn_T *rpn);
static int str2val(char str[], modl_T *modl, double *val, double *dot);
       int velocityOfEdge(modl_T *modl, int ibody, int iedge, int npnt, /*@null@*/double t[], double dxyz[]);
       int velocityOfFace(modl_T *modl, int ibody, int iface, int npnt, /*@null@*/double uv[], double dxyz[]);
       int velocityOfNode(modl_T *modl, int ibody, int inode, double dxyz[]);
static int xformToOriginal(modl_T *modl, int ibody, int jbody, double mat[]);
static int xformVelocity(modl_T *modl, int jbody, int nxyz, double xyz_pnt[], int ndxyz, double dxyz[]);

extern int EG_sensitTopo(int iface, double r[], double dxyz[]);

/*
 ************************************************************************
 *                                                                      *
 * Global variables                                                     *
 *                                                                      *
 ************************************************************************
 */

static int outLevel = 1;


/*
 ************************************************************************
 *                                                                      *
 *   ocsmVersion - return current version                               *
 *                                                                      *
 ************************************************************************
 */

int
ocsmVersion(int    *imajor,             /* (out) major version number */
            int    *iminor)             /* (out) minor version number */
{
    int       status = SUCCESS;         /* (out) return status */

    ROUTINE(ocsmVersion);
    DPRINT1("%s() {",
            routine);

    /* --------------------------------------------------------------- */

    *imajor = OCSM_MAJOR_VERSION;
    *iminor = OCSM_MINOR_VERSION;

//cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmSetOutLevel - set output level                                 *
 *                                                                      *
 ************************************************************************
 */

int
ocsmSetOutLevel(int    ilevel)      /* (in)  output level: */
                                    /*       =0 warnings and errors only */
                                    /*       =1 nominal (default) */
                                    /*       =2 debug */
{
    int       status = SUCCESS;         /* (out) return status */

    ROUTINE(ocsmSetOutLevel);
    DPRINT2("%s(ilevel=%d) {",
            routine, ilevel);

    /* --------------------------------------------------------------- */

    outLevel = ilevel;

//cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmLoad - create a MODL by reading a .csm file                    *
 *                                                                      *
 ************************************************************************
 */

int
ocsmLoad(char   filename[],             /* (in)  file to be read (with .csm) */
         void   **modl)                 /* (out) pointer to MODL */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL;

    /* Sketch contains linsegs, cirarcs, splines, and beziers */
    int       nskpt;

    /* number of open patterns */
    int       npatn = 0;

    int       i, j, k, inquote, narg, irow, icol, itmp, indx, count;
    int       ibrch, ipmtr, jpmtr, icount, jcount, insolver;
    double    rows, cols, despmtr, value, dot, bound;
    char      templine[MAX_LINE_LEN], nextline[MAX_LINE_LEN], tmpfileline[MAX_LINE_LEN];
    char      command[MAX_EXPR_LEN];
    char      *str1=NULL, *str2=NULL, *str3=NULL;
    char      *str4=NULL, *str5=NULL, *str6=NULL;
    char      *str7=NULL, *str8=NULL, *str9=NULL;
    char      defn[MAX_EXPR_LEN], tmpfilename[MAX_EXPR_LEN], pathname[MAX_EXPR_LEN];
    void      *temp;

    FILE      *csm_file[11], *tmp_file=NULL;

    ROUTINE(ocsmLoad);
    DPRINT2("%s(filename=%s) {",
            routine, filename);

    /* --------------------------------------------------------------- */

    SPRINT1(1, "--> enter ocsmLoad(filename=%s)", filename);

    #ifdef FORCE_FINITE_DIFFS
        SPRINT0(0, "WARNING:: ***************************************");
        SPRINT0(0, "WARNING:: all sensitivities by finite differences");
        SPRINT0(0, "WARNING:: ***************************************");
    #endif

    MALLOC(str1, char, MAX_LINE_LEN);
    MALLOC(str2, char, MAX_LINE_LEN);
    MALLOC(str3, char, MAX_LINE_LEN);
    MALLOC(str4, char, MAX_LINE_LEN);
    MALLOC(str5, char, MAX_LINE_LEN);
    MALLOC(str6, char, MAX_LINE_LEN);
    MALLOC(str7, char, MAX_LINE_LEN);
    MALLOC(str8, char, MAX_LINE_LEN);
    MALLOC(str9, char, MAX_LINE_LEN);

    /* default return value */
    *modl = NULL;

    if (strlen(filename) > 255) {
        signalError(NULL, OCSM_ILLEGAL_ARGUMENT, 0,
                    "filename has more than 255 characters");
        status = OCSM_ILLEGAL_ARGUMENT;
        goto cleanup;
    }

    /* open the .csm file */
    if (strlen(filename) > 0) {
        csm_file[0] = fopen(filename, "r");
        if (csm_file[0] == NULL) {
            DPRINT0("csm_file = NULL");
            status = OCSM_FILE_NOT_FOUND;
            goto cleanup;
        }
    }

    /* make a new MODL and initialize it */
    MALLOC(MODL, modl_T, 1);

    MODL->magic   = OCSM_MAGIC;
    MODL->checked = 0;
    MODL->nextseq = 1;
    MODL->recycle = 0;

    MODL->sel_body = -1;
    MODL->sel_face = -1;
    MODL->sel_edge = -1;
    MODL->sel_node = -1;

    MODL->level = 0;
    STRNCPY(MODL->filename[0], filename, 255);
    for (i = 1; i < 11; i++) {
        MODL->filename[i][0] = '\0';
    }

    MODL->nattr = 0;
    MODL->attr  = NULL;

    MODL->nstor = 0;
    MODL->stor  = NULL;

    MODL->nbrch = 0;
    MODL->mbrch = 0;
    MODL->brch  = NULL;

    MODL->npmtr = 0;
    MODL->mpmtr = 0;
    MODL->pmtr  = NULL;

    MODL->nbody = 0;
    MODL->mbody = 0;
    MODL->body  = NULL;

    MODL->perturb  = NULL;
    MODL->basemodl = NULL;
    MODL->dtime    = 0;

    MODL->context  = NULL;

    MODL->errorBuf = NULL;
    MALLOC(MODL->errorBuf, char, MAX_STR_LEN);
    MODL->errorBuf[0] = '\0';

    for (i = 0; i < 100; i++) {
        MODL->profile[i].ncall = 0;
        MODL->profile[i].time  = 0;
    }

    /* return value */
    *modl = MODL;

    if (strlen(filename) == 0) {
        goto cleanup;
    }

    /* initialize the number of active Sketch points and patterns */
    nskpt = 0;
    npatn = 0;

    insolver = 0;

    /* read commands from .csm file until the end of file */
    while (feof(csm_file[MODL->level]) == 0) {

        /* read the next line */
        temp = fgets(templine, MAX_LINE_LEN, csm_file[MODL->level]);

        /* we have found the end of the current file (without an end statement) */
        if (temp == NULL) {

            /* if at the top level, we are done */
            if (MODL->level == 0) {

                /* do NOT create a new Branch since having one here would
                   mean that any Branches added at the end would not get executed */
                break;

            /* otherwise, close file and revert to previous level */
            } else {
                fclose(csm_file[MODL->level]);

                /* create a new Branch and decrement the level */
                status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_END,
                                     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
                CHECK_STATUS(ocsmNewBrch);

                (MODL->level)--;

                SPRINT0(1, "\n<<< Reverting to previous file\n");
                continue;
            }
        }

        /* ignore spaces at the beginning of the line */
        for (j = 0; j < strlen(templine); j++) {
            if (templine[0] == ' ') {
                for (k = 0; k < strlen(templine); k++) {
                    templine[k] = templine[k+1];
                }
            }
        }

        /* process the next line:
           - ignore anything after a #
           - ignore anything after a \ and concatenate next line
           - remove any spaces between double quotes (") so that a user
             could include spaces within an expression as long as
             it is surrounded by double quotes */
        if (templine[0] != '#') {
            inquote = 0;
            j       = 0;
            for (i = 0; i < strlen(templine); i++) {
                if (templine[i] == '#') {
                    nextline[j++] = '\0';
                    break;
                } else if (templine[i] == '\\') {
                    temp = fgets(templine, MAX_LINE_LEN, csm_file[MODL->level]);
                    if (temp == NULL) break;
                    i = -1;
                } else if (templine[i] == '"') {
                    inquote = 1 - inquote;
                } else if (templine[i] == ' ' || templine[i] == '\t' || templine[i] == '\n') {
                    if (inquote == 0 && j > 0) {
                        nextline[j++] = templine[i];
                    }
                } else {
                    nextline[j++] = templine[i];
                }

                if (j >= MAX_LINE_LEN-2) {
                    signalError(MODL, OCSM_ILLEGAL_STATEMENT, 0,
                                "input line is too long");
                    SET_STATUS(OCSM_ILLEGAL_STATEMENT, line_too_long);
                }
            }
        } else {
            STRNCPY(nextline, templine, MAX_LINE_LEN);
            j = strlen(nextline);
        }
        nextline[j++] = '\0';

        /* strip white spaces from end of nextline */
        for (i = j-2; i >= 0; i--) {
            if (nextline[i] == ' ' || nextline[i] == '\t' || nextline[i] == '\n'   ) {
                nextline[i] = '\0';
            } else {
                break;
            }
        }
        SPRINT3(1, "    nextline [%2d:%4d]: %s", MODL->level, MODL->nbrch+1, nextline);
        if (strlen(nextline) <= 1) continue;

        /* get the command from the next input line */
        sscanf(nextline, "%2047s", command);

        /* input is: "# comment" */
        if (strncmp(command, "#", 1) == 0) {
            /* nothing to do */

        /* input is: "assert arg1 arg2" */
        } else if (strcmp(command, "assert") == 0) {

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          str1, str2);
            if (narg != 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, assert);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_ASSERT,
                                 str1, str2, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "arc xend yend zend dist $plane=xy" */
        } else if (strcmp(command, "arc") == 0) {
            if (nskpt == 0) {
                SET_STATUS(OCSM_SKETCH_IS_NOT_OPEN, arc);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, arc);
            }

            /* extract arguments */
            strcpy(str5, "$");
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s %2047s %2047s\n",
                          str1, str2, str3, str4, &(str5[1]));
            if (narg == 4) {
                STRNCPY(str5, "$xy", MAX_EXPR_LEN);
            } else if (narg == 5) {
                if        (strcmp(str5, "$xy") == 0) {
                    STRNCPY(str5, "$xy", MAX_EXPR_LEN);
                } else if (strcmp(str5, "$yz") == 0) {
                    STRNCPY(str5, "$yz", MAX_EXPR_LEN);
                } else if (strcmp(str5, "$zx") == 0) {
                    STRNCPY(str5, "$zx", MAX_EXPR_LEN);
                } else {
                    SET_STATUS(OCSM_ILLEGAL_VALUE, arc);
                }
            } else {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, arc);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_ARC,
                                 str1, str2, str3, str4, str5, str6, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

            /* increment the number of Sketch points */
            nskpt++; nskpt++;

        /* input is: "attribute $attrName attrValue */
        } else if (strcmp(command, "attribute") == 0) {

            /* previous Branch will be attributed */
            ibrch = MODL->nbrch;

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          str1, str2);
            if (narg != 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, attribute);
            }

            /* set the Branch's attribute */
            if (ibrch == 0) {
                status = ocsmSetAttr(MODL, ibrch, str1, str2);
                CHECK_STATUS(ocsmSetAttr);
            } else if (MODL->brch[ibrch].type == OCSM_ASSERT    ||
                       MODL->brch[ibrch].type == OCSM_END       ||
                       MODL->brch[ibrch].type == OCSM_INTERFACE ||
                       MODL->brch[ibrch].type == OCSM_MACBEG    ||
                       MODL->brch[ibrch].type == OCSM_MACEND    ||
                       MODL->brch[ibrch].type == OCSM_MARK      ||
                       MODL->brch[ibrch].type == OCSM_PATBEG    ||
                       MODL->brch[ibrch].type == OCSM_PROJECT   ||
                       MODL->brch[ibrch].type == OCSM_SET       ||
                       MODL->brch[ibrch].type == OCSM_SKBEG     ||
                       MODL->brch[ibrch].type == OCSM_SKCON     ||
                       MODL->brch[ibrch].type == OCSM_SKVAR     ||
                       MODL->brch[ibrch].type == OCSM_SOLBEG    ||
                       MODL->brch[ibrch].type == OCSM_STORE     ||
                       MODL->brch[ibrch].type == OCSM_UDPARG      ) {
                signalError(MODL, OCSM_ILLEGAL_ATTRIBUTE, 0,
                            "a \"%s\" Branch cannot be attributed", ocsmGetText(MODL->brch[ibrch].type));
                SET_STATUS(OCSM_ILLEGAL_ATTRIBUTE, attribute);
            } else {
                status = ocsmSetAttr(MODL, ibrch, str1, str2);
                CHECK_STATUS(ocsmSetAttr);
            }

        /* input is: "bezier x y z" */
        } else if (strcmp(command, "bezier") == 0) {
            if (nskpt == 0) {
                SET_STATUS(OCSM_SKETCH_IS_NOT_OPEN, bezier);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, bezier);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s\n",
                          str1, str2, str3);
            if (narg != 3) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, bezier);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_BEZIER,
                                 str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

            /* increment the number of Sketch points */
            nskpt++;

        /* input is: "blend noseList=0 tailList=0 reorder=0" */
        } else if (strcmp(command, "blend") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, blend);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, blend);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s\n",
                          str1, str2, str3);
            if   (narg < 1) {
                STRNCPY(str1, "0", MAX_EXPR_LEN);
            }
            if (narg < 2) {
                STRNCPY(str2, "0", MAX_EXPR_LEN);
            }
            if (narg < 3) {
                STRNCPY(str3, "0", MAX_EXPR_LEN);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_BLEND,
                                 str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "box xbase ybase zbase dx dy dz" */
        } else if (strcmp(command, "box") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, box);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, box);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s %2047s %2047s %2047s\n",
                          str1, str2, str3, str4, str5, str6);
            if (narg != 6) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, box);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_BOX,
                                 str1, str2, str3, str4, str5, str6, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "chamfer radius edgeList=0" */
        } else if (strcmp(command, "chamfer") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, chamfer);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, chamfer);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          str1, str2);
            if (narg == 1) {
                STRNCPY(str2, "0", MAX_EXPR_LEN);
            } else if (narg != 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, chamfer);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_CHAMFER,
                                 str1, str2, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "cirarc xon yon zon xend yend zend" */
        } else if (strcmp(command, "cirarc") == 0) {
            if (nskpt == 0) {
                SET_STATUS(OCSM_SKETCH_IS_NOT_OPEN, cirarc);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, cirarc);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s %2047s %2047s %2047s\n",
                          str1, str2, str3, str4, str5, str6);
            if (narg == 4) {
                STRNCPY(str5, "$xy", MAX_EXPR_LEN);
                str6[0] = '\0';
            } else if (narg == 5) {
                if        (strcmp(str5, "xy") == 0) {
                    STRNCPY(str5, "$xy", MAX_EXPR_LEN);
                } else if (strcmp(str5, "yz") == 0) {
                    STRNCPY(str5, "$yz", MAX_EXPR_LEN);
                } else if (strcmp(str5, "zx") == 0) {
                    STRNCPY(str5, "$zx", MAX_EXPR_LEN);
                } else {
                    SET_STATUS(OCSM_ILLEGAL_VALUE, cirarc);
                }
            } else if (narg != 6) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, cirarc);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_CIRARC,
                                 str1, str2, str3, str4, str5, str6, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

            /* increment the number of Sketch points */
            nskpt++; nskpt++;

        /* input is: "combine" */
        } else if (strcmp(command, "combine") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, combine);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, combine);
            }

            /* no arguments */

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_COMBINE,
                                 NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "cone xvrtx yvrtx zvrtx xbase ybase zbase radius" */
        } else if (strcmp(command, "cone") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, cone);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, cone);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s %2047s %2047s %2047s %2047s\n",
                          str1, str2, str3, str4, str5, str6, str7);
            if (narg != 7) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, cone);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_CONE,
                                 str1, str2, str3, str4, str5, str6, str7, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "connect faceList1 faceList2" */
        } else if (strcmp(command, "connect") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, connect);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, connect);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          str1, str2);
            if (narg != 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, connect);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_CONNECT,
                                 str1, str2, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "cylinder xbeg ybeg zbeg xend yend zend radius" */
        } else if (strcmp(command, "cylinder") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, cylinder);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, cylinder);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s %2047s %2047s %2047s %2047s\n",
                          str1, str2, str3, str4, str5, str6, str7);
            if (narg != 7) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, cylinder);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_CYLINDER,
                                 str1, str2, str3, str4, str5, str6, str7, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "despmtr $pmtrName expression" */
        } else if (strcmp(command, "despmtr") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, despmtr);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, despmtr);
            } else if (MODL->level > 0) {
                signalError(MODL, OCSM_ILLEGAL_STATEMENT, 0,
                            "DESPMTR not allowed in .udc file");
                SET_STATUS(OCSM_ILLEGAL_STATEMENT, despmtr);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          str1, str2);
            if (narg != 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, despmtr);
            }

            /* do not allow pmtrName to start with '@' */
            if (str1[0] == '@') {
                SET_STATUS(OCSM_ILLEGAL_PMTR_NAME, despmtr);
            }

            /* determine if a single- or multi-set (dependent on whether str2
               contains a semi-colon) */
            if (strstr(str2, ";") == NULL) {

                /* break str1 into name[irow,icol] */
                status = parseName(MODL, str1, &jpmtr, &irow, &icol);
                CHECK_STATUS(parseName);

                /* convert jpmtr to EXTERNAL (because parseName creates an INTERNAL) */
                MODL->pmtr[jpmtr].type = OCSM_EXTERNAL;

                status = ocsmGetValu(MODL, jpmtr, irow, icol, &value, &dot);
                CHECK_STATUS(ocsmGetValu);

                if (fabs(value+HUGEQ) > 1 && fabs(dot+HUGEQ) > 1) {
                    SPRINT0(0, "WARNING:: value not overwritten");
                } else {
                    status = ocsmSetValu(MODL, jpmtr, irow, icol, str2);
                    CHECK_STATUS(ocsmSetValu);
                }

            /* multi-set mode */
            } else {
                narg = sscanf(nextline, "%*s %2047s %4047s\n",
                          str1, str2);
                if (narg != 2) {
                    SET_STATUS(OCSM_NOT_ENOUGH_ARGS, multi-set);
                }

                /* look for current Parameter */
                jpmtr = 0;
                for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
                    if (strcmp(MODL->pmtr[ipmtr].name, str1) == 0    &&
                               MODL->pmtr[ipmtr].level == MODL->level  ) {
                        jpmtr = ipmtr;
                        break;
                    }
                }

                /* Parameter must have been defined (in a dimension statement) */
                if (jpmtr <= 0) {
                    SET_STATUS(OCSM_NAME_NOT_FOUND, despmtr);
                } else if (MODL->pmtr[ipmtr].type != OCSM_EXTERNAL) {
                    SET_STATUS(OCSM_PMTR_IS_INTERNAL, despmtr);
                }

                /* store the values */
                icount = 0;
                for (irow = 1; irow <= MODL->pmtr[jpmtr].nrow; irow++) {
                    for (icol = 1; icol <= MODL->pmtr[jpmtr].ncol; icol++) {
                        jcount = 0;

                        while (icount < strlen(str2)) {
                            if (str2[icount] == ';') {
                                icount++;
                                break;
                            } else {
                                defn[jcount  ] = str2[icount];
                                defn[jcount+1] = '\0';
                                icount++;
                                jcount++;
                            }
                        }

                        status = ocsmGetValu(MODL, jpmtr, irow, icol, &value, &dot);
                        CHECK_STATUS(ocsmGetValu);

                        if (fabs(value+HUGEQ) > 1 && fabs(dot+HUGEQ) > 1) {
                            SPRINT0(0, "WARNING:: value not overwritten");
                        } else {
                            if (strcmp(defn, "") != 0) {
                                status = ocsmSetValu(MODL, jpmtr, irow, icol, defn);
                                CHECK_STATUS(ocsmSetValu);
                            }
                        }
                    }
                }
            }

        /* input is: "dimension $pmtrName nrow ncol despmtr" */
        } else if (strcmp(command, "dimension") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, dimension);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, dimension);
            } else if (MODL->level > 0) {
                signalError(MODL, OCSM_ILLEGAL_STATEMENT, 0,
                            "DIMENSION not allowed in .udc file");
                SET_STATUS(OCSM_ILLEGAL_STATEMENT, dimension);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s %2047s\n",
                          str1, str2, str3, str4);
            if (narg < 3) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, dimension);
            }

            /* do not allow pmtrName to start with '@' */
            if (str1[0] == '@') {
                SET_STATUS(OCSM_ILLEGAL_PMTR_NAME, dimension);
            }

            status = str2val(str2, MODL, &rows, &dot);
            CHECK_STATUS(str2val:rows);

            status = str2val(str3, MODL, &cols, &dot);
            CHECK_STATUS(str2val:cols);

            if (narg == 4) {
                status = str2val(str4, MODL, &despmtr, &dot);
                CHECK_STATUS(str2val:despmtr);
            } else {
                despmtr = 0;
            }

            if (NINT(despmtr) == 0) {
                status = ocsmNewPmtr(MODL, str1, OCSM_INTERNAL, NINT(rows), NINT(cols));
                CHECK_STATUS(ocsmNewPmtr);
            } else {
                status = ocsmNewPmtr(MODL, str1, OCSM_EXTERNAL, NINT(rows), NINT(cols));
                CHECK_STATUS(ocsmNewPmtr);
            }

        /* input is: "dump $filename remove=0" */
        } else if (strcmp(command, "dump") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, dump);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, dump);
            }

            /* extract argument */
            strcpy(str1, "$");
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          &(str1[1]), str2);
            if (narg < 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, dump);
            } else if (narg == 1) {
                STRNCPY(str2, "0", MAX_EXPR_LEN);
            }

            /* convert filename if it starts with "$/" to relative pathname */
            if (strncmp(str1, "$$/", 3) == 0) {
                STRNCPY(pathname, MODL->filename[MODL->level], MAX_EXPR_LEN);
                i = strlen(pathname) - 1;
                while (pathname[i] != SLASH && i >= 0) {
                    pathname[i] = '\0';
                    i--;
                }
                STRNCPY(tmpfilename, &(str1[3]), MAX_EXPR_LEN);
                snprintf(str1, MAX_EXPR_LEN, "$%s%s", pathname, tmpfilename);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_DUMP,
                                 str1, str2, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "end" */
        } else if (strcmp(command, "end") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, end);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, end);
            }

            /* if at the top level, we are done */
            if (MODL->level == 0) {

                /* do NOT create a new Branch since having one here would
                   mean that any Branches added at the end would not get executed */
                break;

            /* otherwise, close file and revert to previous level */
            } else {
                fclose(csm_file[MODL->level]);

                /* create a new Branch and decrement the level */
                status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_END,
                                     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
                CHECK_STATUS(ocsmNewBrch);

                (MODL->level)--;

                SPRINT0(1, "\n<<< Reverting to previous file\n");
                continue;
            }

        /* input is: "extract index=0" */
        } else if (strcmp(command, "extract") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, extract);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, extract);
            }

            /* extract argument */
            narg = sscanf(nextline, "%*s %2047s\n",
                          str1);
            if (narg != 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, extract);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_EXTRACT,
                                 str1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "extrude dx dy dz" */
        } else if (strcmp(command, "extrude") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, extrude);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, extrude);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s\n",
                          str1, str2, str3);
            if (narg != 3) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, extrude);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_EXTRUDE,
                                 str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "fillet radius edgeList=0" */
        } else if (strcmp(command, "fillet") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, fillet);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, fillet);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          str1, str2);
            if (narg == 1) {
                STRNCPY(str2, "0", MAX_EXPR_LEN);
            } else if (narg != 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, fillet);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_FILLET,
                                 str1, str2, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "hollow thick faceList=0" */
        } else if (strcmp(command, "hollow") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, hollow);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, hollow);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          str1, str2);
            if (narg == 1) {
                STRNCPY(str2, "0", MAX_EXPR_LEN);
            } else if (narg != 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, hollow);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_HOLLOW,
                                 str1, str2, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "import $filename bodynumber=1" */
        } else if (strcmp(command, "import") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, import);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, import);
            }

            /* extract argument */
            strcpy(str1, "$");
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          &(str1[1]), str2);
            if        (narg == 1) {
                STRNCPY(str2, "1", MAX_EXPR_LEN);
            } else if (narg != 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, import);
            }

            /* convert filename if it starts with "$/" or "$$/" to relative pathname */
            if (strncmp(str1, "$$/", 3) == 0) {
                STRNCPY(pathname, MODL->filename[MODL->level], MAX_EXPR_LEN);
                i = strlen(pathname) - 1;
                while (pathname[i] != SLASH && i >= 0) {
                    pathname[i] = '\0';
                    i--;
                }
                STRNCPY(tmpfilename, &(str1[3]), MAX_EXPR_LEN);
                snprintf(str1, MAX_EXPR_LEN, "$%s%s", pathname, tmpfilename);
            } else if (strncmp(str1, "$$$/", 4) == 0) {
                STRNCPY(pathname, MODL->filename[MODL->level], MAX_EXPR_LEN);
                i = strlen(pathname) - 1;
                while (pathname[i] != SLASH && i >= 0) {
                    pathname[i] = '\0';
                    i--;
                }
                STRNCPY(tmpfilename, &(str1[4]), MAX_EXPR_LEN);
                snprintf(str1, MAX_EXPR_LEN, "$%s%s", pathname, tmpfilename);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_IMPORT,
                                 str1, str2, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "interface $argName $argType default" */
        } else if (strcmp(command, "interface") == 0) {
            if (MODL->level <= 0) {
                signalError(MODL, OCSM_ILLEGAL_STATEMENT, 0,
                            "INTERFACE not allowed in .csm file");
                SET_STATUS(OCSM_ILLEGAL_STATEMENT, interface);
            }

            /* extract arguments */
            strcpy(str1, "$");  strcpy(str2, "$");
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s\n",
                          &(str1[1]), &(str2[1]), str3);
            if (narg != 3) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, interface);
            }

            /* ensure that valid argType is given */
            if ((strcmp(str2, "$in")  != 0) &&
                (strcmp(str2, "$out") != 0) &&
                (strcmp(str2, "$dim") != 0)   ) {
                SET_STATUS(OCSM_ILLEGAL_VALUE, interface);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_INTERFACE,
                                 str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "intersect $order=none index=1 maxtol=0" */
        } else if (strcmp(command, "intersect") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, intersect);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, intersect);
            }

            /* extract arguments */
            strcpy(str1, "$");
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s\n",
                          &(str1[1]), str2, str3);
            if (narg < 1) {
                STRNCPY(str1, "$none", MAX_EXPR_LEN);
            }
            if (narg < 2) {
                STRNCPY(str2, "1",     MAX_EXPR_LEN);
            }
            if (narg < 3) {
                STRNCPY(str3, "0", MAX_EXPR_LEN);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_INTERSECT,
                                 str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "join toler=0" */
        } else if (strcmp(command, "join") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, join);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, join);
            }

            /* extract argument */
            narg = sscanf(nextline, "%*s %2047s\n",
                          str1);
            if (narg < 1) {
                STRNCPY(str1, "0", MAX_EXPR_LEN);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_JOIN,
                                 str1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "lbound $pmtrName expression" */
        } else if (strcmp(command, "lbound") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, lbound);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, lbound);
            } else if (MODL->level > 0) {
                signalError(MODL, OCSM_ILLEGAL_STATEMENT, 0,
                            "LBOUND not allowed in .udc file");
                SET_STATUS(OCSM_ILLEGAL_STATEMENT, lbound);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          str1, str2);
            if (narg != 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, lbound);
            }

            /* do not allow pmtrName to start with '@' */
            if (str1[0] == '@') {
                SET_STATUS(OCSM_ILLEGAL_PMTR_NAME, lbound);
            }

            /* break str1 into name[irow,icol] */
            status = parseName(MODL, str1, &jpmtr, &irow, &icol);
            CHECK_STATUS(parseName);

            /* get the lower bound value */
            status = str2val(str2, MODL, &bound, &dot);
            CHECK_STATUS(str2val:bound);

            /* if name did not contain a "[", then set all rows and columns */
            if (strstr(str1, "[") == NULL) {
                indx = 0;
                for (irow = 1; irow <= MODL->pmtr[jpmtr].nrow; irow++) {
                    for (icol = 1; icol <= MODL->pmtr[jpmtr].ncol; icol++) {
                        MODL->pmtr[jpmtr].lbnd[indx] = bound;
                        indx++;
                    }
                }

            /* otherwise just set the requested value */
            } else {
                if (irow < 1 || irow > MODL->pmtr[jpmtr].nrow ||
                    icol < 1 || icol > MODL->pmtr[jpmtr].ncol   ) {
                    SET_STATUS(OCSM_ILLEGAL_PMTR_INDEX, lbound);
                }

                indx = icol-1 + (irow-1) * MODL->pmtr[jpmtr].ncol;
                MODL->pmtr[jpmtr].lbnd[indx] = bound;
            }

        /* input is: "linseg x y z" */
        } else if (strcmp(command, "linseg") == 0) {
            if (nskpt == 0) {
                SET_STATUS(OCSM_SKETCH_IS_NOT_OPEN, linseg);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, linseg);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s\n",
                          str1, str2, str3);
            if (narg != 3) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, linseg);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_LINSEG,
                                 str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

            /* increment the number of Sketch points */
            nskpt++;

        /* input is: "loft smooth" */
        } else if (strcmp(command, "loft") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, loft);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, loft);
            }

            /* extract argument */
            narg = sscanf(nextline, "%*s %2047s\n",
                          str1);
            if (narg != 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, loft);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_LOFT,
                                 str1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "macbeg imacro" */
        } else if (strcmp(command, "macbeg") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, macbeg);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, macbeg);
            }

            /* extract argument */
            narg = sscanf(nextline, "%*s %2047s\n",
                          str1);
            if (narg != 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, macbeg);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_MACBEG,
                                 str1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "macend" */
        } else if (strcmp(command, "macend") == 0) {
            if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, macend);
            }

            /* no arguments */

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_MACEND,
                                 NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "mark" */
        } else if (strcmp(command, "mark") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, mark);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, mark);
            }

            /* no arguments */

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_MARK,
                                 NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "mirror nx ny nz dist=0" */
        } else if (strcmp(command, "mirror") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, mirror);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, mirror);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s %2047s\n",
                          str1, str2, str3, str4);
            if (narg == 3) {
                STRNCPY(str4, "0", MAX_EXPR_LEN);
            } else if (narg != 4) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, mirror);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_MIRROR,
                                 str1, str2, str3, str4, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "name $branchName" */
        } else if (strcmp(command, "name") == 0) {

            /* previous Branch will be named */
            ibrch = MODL->nbrch;
            if (ibrch < 1) {
                SET_STATUS(OCSM_ILLEGAL_BRCH_INDEX, name);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s\n",
                          str1);
            if (narg != 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, name);
            }

            /* set the Branch's name */
            status = ocsmSetName(MODL, ibrch, str1);
            CHECK_STATUS(ocsmSetName);

        /* input is: "patbeg $pmtrName ncopy" */
        } else if (strcmp(command, "patbeg") == 0) {
            if (npatn >= MAX_NUM_PATTERNS) {
                signalError(MODL, OCSM_NESTED_TOO_DEEPLY, 0,
                            "npatn=%d >0 MAX_PATTERNS=%d", npatn, MAX_NUM_PATTERNS);
                SET_STATUS(OCSM_NESTED_TOO_DEEPLY, patbeg);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, patbeg);
            } else {
                npatn++;
            }

            /* extract arguments */
            strcpy(str1, "$");
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          &(str1[1]), str2);
            if (narg != 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, patbeg);
            }

            /* do not allow pmtrName to start with '@' */
            if (str1[1] == '@') {
                SET_STATUS(OCSM_ILLEGAL_PMTR_NAME, patbeg);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_PATBEG,
                                 str1, str2, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "patbreak expr" */
        } else if (strcmp(command, "patbreak") == 0) {

            if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, patend);
            }

            /* find the matching patbeg */
            i = 1;
            for (ibrch = MODL->nbrch; ibrch > 0; ibrch--) {
                if        (MODL->brch[ibrch].type == OCSM_PATEND) {
                    i++;
                } else if (MODL->brch[ibrch].type == OCSM_PATBEG) {
                    i--;
                    if (i == 0) {
                        break;
                    }
                }
            }
            if (ibrch <= 0) {
                SET_STATUS(OCSM_PATEND_WITHOUT_PATBEG, patbreak);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s\n",
                          str1);
            if (narg != 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, patbreak);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_PATBREAK,
                                 str1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "patend" */
        } else if (strcmp(command, "patend") == 0) {
            if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, patend);
            }

            /* find the matching patbeg */
            i = 1;
            for (ibrch = MODL->nbrch; ibrch > 0; ibrch--) {
                if        (MODL->brch[ibrch].type == OCSM_PATEND) {
                    i++;
                } else if (MODL->brch[ibrch].type == OCSM_PATBEG) {
                    i--;
                    if (i == 0) {
                        break;
                    }
                }
            }
            if (ibrch <= 0) {
                SET_STATUS(OCSM_PATEND_WITHOUT_PATBEG, patend);
            }

            /* no arguments */

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_PATEND,
                                 NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

            npatn--;

        /* input is: "project x y z dx dy dz" */
        } else if (strcmp(command, "project") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, recall);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, recall);
            }

            /* extract argument and save*/
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s %2047s %2047s %2047s\n",
                          str1, str2, str3, str4, str5, str6);
            if (narg < 6) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, project);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_PROJECT,
                                 str1, str2, str3, str4, str5, str6, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "recall imacro" */
        } else if (strcmp(command, "recall") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, recall);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, recall);
            }

            /* extract argument and save*/
            narg = sscanf(nextline, "%*s %2047s\n",
                          str1);
            if (narg != 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, recall);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_RECALL,
                                 str1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "reorder ishift iflip=0" */
        } else if (strcmp(command, "reorder") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, reorder);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, reorder);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          str1, str2);
            if (narg == 1) {
                STRNCPY(str2, "0", MAX_EXPR_LEN);
            } else if (narg != 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, reorder);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_REORDER,
                                 str1, str2, str3, str4, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "restore $name index=0" */
        } else if (strcmp(command, "restore") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, restore);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, restore);
            }

            /* extract arguments */
            strcpy(str1, "$");
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          &(str1[1]), str2);
            if (narg == 1) {
                STRNCPY(str2, "0", MAX_EXPR_LEN);
            } else if (narg != 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, patbeg);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_RESTORE,
                                 str1, str2, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "revolve xorig yorig zorig dxaxis dyaxis dzaxis angDeg" */
        } else if (strcmp(command, "revolve") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, revolve);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, revolve);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s %2047s %2047s %2047s %2047s\n",
                          str1, str2, str3, str4, str5, str6, str7);
            if (narg != 7) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, revolve);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_REVOLVE,
                                 str1, str2, str3, str4, str5, str6, str7, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "rotatex angDeg yaxis zaxis" */
        } else if (strcmp(command, "rotatex") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, rotatex);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, rotatex);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s\n",
                          str1, str2, str3);
            if (narg != 3) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, rotatex);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_ROTATEX,
                                 str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "rotatey angDeg zaxis xaxis" */
        } else if (strcmp(command, "rotatey") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, rotatey);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, rotatey);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s\n",
                          str1, str2, str3);
            if (narg != 3) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, rotatey);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_ROTATEY,
                                 str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "rotatez angDeg xaxis yaxis" */
        } else if (strcmp(command, "rotatez") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, rotatez);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, rotatez);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s\n",
                          str1, str2, str3);
            if (narg != 3) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, rotatez);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_ROTATEZ,
                                 str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "rule reorder=0" */
        } else if (strcmp(command, "rule") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, rule);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, rule);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s\n",
                          str1);
            if (narg < 1) {
                STRNCPY(str1, "0", MAX_EXPR_LEN);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_RULE,
                                 str1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "scale fact" */
        } else if (strcmp(command, "scale") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, scale);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, scale);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s\n",
                          str1);
            if (narg != 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, scale);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SCALE,
                                 str1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "select $type arg1 ..." */
        } else if (strcmp(command, "select") == 0) {

            /* extract arguments */
            strcpy(str1, "$");
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s %2047s %2047s %2047s %2047s %2047s\n",
                          &(str1[1]), str2, str3, str4, str5, str6, str7, str8);

            /* create the new Branch */
            if        (narg < 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, select);
            } else if (strcmp(str1, "$body") == 0) {
                if        (narg == 1) {    /* body */
                    status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SELECT,
                                         str1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
                } else if (narg == 2) {    /* body ibody */
                    status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SELECT,
                                         str1, str2, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
                } else {
                    SET_STATUS(OCSM_NOT_ENOUGH_ARGS, select_body);
                }
            } else if (strcmp(str1, "$face") == 0) {
                if        (narg == 2) {    /* face iface */
                    status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SELECT,
                                         str1, str2, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
                } else if (narg == 3) {    /* face ibody1 iford1 */
                    status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SELECT,
                                         str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
                } else if (narg > 3 ) {    /* face ibody1 iford1 iseq */
                    status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SELECT,
                                         str1, str2, str3, str4, NULL, NULL, NULL, NULL, NULL);
                } else {
                    SET_STATUS(OCSM_NOT_ENOUGH_ARGS, select_face);
                }
            } else if (strcmp(str1, "$edge") == 0) {
                if        (narg == 2) {    /* edge iedge */
                    status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SELECT,
                                         str1, str2, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
                } else if (narg == 5) {    /* edge ibody1 iford1 ibody2 iford2 */
                    status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SELECT,
                                         str1, str2, str3, str4, str5, NULL, NULL, NULL, NULL);
                } else if (narg >  5) {    /* edge ibody1 iford1 ibody2 iford2 iseq */
                    status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SELECT,
                                         str1, str2, str3, str4, str5, str6, NULL, NULL, NULL);
                } else {
                    SET_STATUS(OCSM_NOT_ENOUGH_ARGS, select_edge);
                }
            } else if (strcmp(str1, "$node") == 0) {
                if        (narg == 2) {    /* node inode */
                    status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SELECT,
                                         str1, str2, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
//$$$                } else if (narg == 7) {    /* node ibody1 iford1 ibody2 iford2 ibody3 iford3 */
//$$$                    status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SELECT,
//$$$                                         str1, str2, str3, str4, str5, str6, str7, NULL, NULL);
//$$$                } else if (narg >  7) {    /* node ibody1 iford1 ibody2 iford2 ibody3 iford3 iseq */
//$$$                    status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SELECT,
//$$$                                         str1, str2, str3, str4, str5, str6, str7, str8, NULL);
                } else {
                    status = OCSM_NOT_ENOUGH_ARGS;
                    CHECK_STATUS(select_node);
                }
            } else {
                SET_STATUS(OCSM_ILLEGAL_TYPE, select);
            }
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "set $pmtrName exprs" */
        } else if (strcmp(command, "set") == 0) {
            if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, set);
            }

            /* extract arguments */
            strcpy(str1, "$");
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          &(str1[1]), str2);
            if (narg != 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, set);
            }

            /* do not allow pmtrName to start with '@' */
            if (str1[1] == '@') {
                SET_STATUS(OCSM_ILLEGAL_PMTR_NAME, set);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SET,
                                 str1, str2, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "skbeg x y z" */
        } else if (strcmp(command, "skbeg") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, skbeg);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, skbeg);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s\n",
                          str1, str2, str3);
            if (narg != 3) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, skbeg);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SKBEG,
                                 str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

            /* increment the number of Sketch points */
            nskpt++;

        /* input is: "skcon $type index1 index2=-1 $value=0" */
        } else if (strcmp(command, "skcon") == 0) {
            if (nskpt <= 0) {
                SET_STATUS(OCSM_SKETCH_IS_NOT_OPEN, skcon);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, skcon);
            }

            /* make sure that this follows a skvar or skcon statement */
            if (MODL->brch[MODL->nbrch].type != OCSM_SKVAR &&
                MODL->brch[MODL->nbrch].type != OCSM_SKCON   ) {
                signalError(MODL, OCSM_ILLEGAL_STATEMENT, 0,
                            "only SKVAR or SKCON can preceed SKCON statement");
                SET_STATUS(OCSM_ILLEGAL_STATEMENT, skcon);
            }

            /* extract arguments */
            strcpy(str1, "$");
            strcpy(str4, "$");
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s %2047s\n",
                          &(str1[1]), str2, str3, &(str4[1]));
            if (narg < 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, skcon);
            }
            if (narg < 3) {
                STRNCPY(str3, "-1", MAX_EXPR_LEN);
            }
            if (narg < 4) {
                STRNCPY(str4, "$0", MAX_EXPR_LEN);
            }

            if        (strcmp(str1, "$X") == 0) {
            } else if (strcmp(str1, "$Y") == 0) {
            } else if (strcmp(str1, "$P") == 0) {
            } else if (strcmp(str1, "$T") == 0) {
            } else if (strcmp(str1, "$A") == 0) {
            } else if (strcmp(str1, "$W") == 0) {
            } else if (strcmp(str1, "$D") == 0) {
            } else if (strcmp(str1, "$H") == 0) {
            } else if (strcmp(str1, "$V") == 0) {
            } else if (strcmp(str1, "$I") == 0) {
            } else if (strcmp(str1, "$L") == 0) {
            } else if (strcmp(str1, "$R") == 0) {
            } else if (strcmp(str1, "$S") == 0) {
            } else {
                SET_STATUS(OCSM_ILLEGAL_VALUE, skcon);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SKCON,
                                 str1, str2, str3, str4, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "skend wireonly=0" */
        } else if (strcmp(command, "skend") == 0) {
            if (nskpt < 1) {
                SET_STATUS(OCSM_COLINEAR_SKETCH_POINTS, skend);
            } else if (nskpt > MAX_SKETCH_SIZE) {
                SET_STATUS(OCSM_TOO_MANY_SKETCH_POINTS, skend);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, skend);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s\n",
                          str1);
            if (narg <= 0) {
                STRNCPY(str1, "0", MAX_EXPR_LEN);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SKEND,
                                 str1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

            /* reset the number of Sketch points */
            nskpt = 0;

        /* input is: "skvar $type valList" */
        } else if (strcmp(command, "skvar") == 0) {
            if (nskpt <= 0) {
                SET_STATUS(OCSM_SKETCH_IS_NOT_OPEN, skvar);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, skvar);
            }

            /* make sure that this follows a skbeg statement */
            if (MODL->brch[MODL->nbrch].type != OCSM_SKBEG) {
                signalError(MODL, OCSM_ILLEGAL_STATEMENT, 0,
                            "only SKBEG can preceed SKVAR statement");
                SET_STATUS(OCSM_ILLEGAL_STATEMENT, skvar);
            }

            /* extract arguments */
            strcpy(str1, "$");
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          &(str1[1]), str2);
            if (narg < 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, skvar);
            }

            if (strcmp(str1, "$xy") != 0 &&
                strcmp(str1, "$yz") != 0 &&
                strcmp(str1, "$zx") != 0   ) {
                SET_STATUS(OCSM_ILLEGAL_VALUE, skvar);
            }

            /* make sure that str2 contains the correct number of semicolons */
            count = 0;
            for (i = 0; i < strlen(str2); i++) {
                if (str2[i] == ';') {
                    count++;
                }
            }

            if (count == 0 || count%3 != 0) {
                signalError(MODL, OCSM_ILLEGAL_VALUE, 0,
                            "valList must contain triplets");
                SET_STATUS(OCSM_ILLEGAL_VALUE, skvar);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SKVAR,
                                 str1, str2, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "solbeg $varList" */
        } else if (strcmp(command, "solbeg") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, solbeg);
            }

            /* extract arguments */
            strcpy(str1, "$");
            narg = sscanf(nextline, "%*s %2047s\n",
                          &(str1[1]));
            if (narg != 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, solbeg);
            }

            /* remember that we are in a solver block */
            insolver = 1;

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SOLBEG,
                                 str1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "solcon $expr" */
        } else if (strcmp(command, "solcon") == 0) {
            if (insolver != 1) {
                SET_STATUS(OCSM_SOLVER_IS_NOT_OPEN, solcon);
            }

            /* extract arguments */
            strcpy(str1, "$");
            narg = sscanf(nextline, "%*s %2047s\n",
                          &(str1[1]));
            if (narg != 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, solcon);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SOLCON,
                                 str1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "solend" */
        } else if (strcmp(command, "solend") == 0) {
            if (insolver != 1) {
                SET_STATUS(OCSM_SOLVER_IS_NOT_OPEN, solend);
            }

            /* no arguments */

            /* close the solver */
            insolver = 0;

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SOLEND,
                                 NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "sphere xcent ycent zcent radius" */
        } else if (strcmp(command, "sphere") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, sphere);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, sphere);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s %2047s\n",
                          str1, str2, str3, str4);
            if (narg != 4) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, sphere);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SPHERE,
                                 str1, str2, str3, str4, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "spline x y z" */
        } else if (strcmp(command, "spline") == 0) {
            if (nskpt == 0) {
                SET_STATUS(OCSM_SKETCH_IS_NOT_OPEN, spline);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, spline);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s\n",
                          str1, str2, str3);
            if (narg != 3) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, spline);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SPLINE,
                                 str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

            /* increment the number of Sketch points */
            nskpt++;

        /* input is: "store $name index=0 keep=0" */
        } else if (strcmp(command, "store") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, store);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, store);
            }

            /* extract arguments */
            strcpy(str1, "$");
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s\n",
                          &(str1[1]), str2, str3);
            if (narg < 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, patbeg);
            }
            if (narg < 2) {
                STRNCPY(str2, "0", MAX_EXPR_LEN);
            }
            if (narg < 3) {
                STRNCPY(str3, "0", MAX_EXPR_LEN);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_STORE,
                                 str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "subtract $order=none index=1 maxtol=0" */
        } else if (strcmp(command, "subtract") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, subtract);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, subtract);
            }

            /* extract arguments */
            strcpy(str1, "$");
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s\n",
                          &(str1[1]), str2, str3);
            if (narg < 1) {
                STRNCPY(str1, "$none", MAX_EXPR_LEN);
            }
            if (narg < 2) {
                STRNCPY(str2, "1",     MAX_EXPR_LEN);
            }
            if (narg < 3) {
                STRNCPY(str3, "0",     MAX_EXPR_LEN);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SUBTRACT,
                                 str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "sweep" */
        } else if (strcmp(command, "sweep") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, sweep);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, sweep);
            }

            /* no arguments */

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_SWEEP,
                                 NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "torus xcent ycent zcent dxaxis dyaxis dzaxis majorRad minorRad" */
        } else if (strcmp(command, "torus") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, torus);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, torus);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s %2047s %2047s %2047s %2047s %2047s\n",
                          str1, str2, str3, str4, str5, str6, str7, str8);
            if (narg != 8) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, torus);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_TORUS,
                                 str1, str2, str3, str4, str5, str6, str7, str8, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "translate dx dy dz" */
        } else if (strcmp(command, "translate") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, translate);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, translate);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s\n",
                          str1, str2, str3);
            if (narg != 3) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, translate);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_TRANSLATE,
                                 str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "ubound $pmtrName expression" */
        } else if (strcmp(command, "ubound") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, ubound);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, ubound);
            } else if (MODL->level > 0) {
                signalError(MODL, OCSM_ILLEGAL_STATEMENT, 0,
                            "UBOUND not allowed in .udc file");
                SET_STATUS(OCSM_ILLEGAL_STATEMENT, ubound);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s\n",
                          str1, str2);
            if (narg != 2) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, ubound);
            }

            /* do not allow pmtrName to start with '@' */
            if (str1[0] == '@') {
                SET_STATUS(OCSM_ILLEGAL_PMTR_NAME, ubound);
            }

            /* break str1 into name[irow,icol] */
            status = parseName(MODL, str1, &jpmtr, &irow, &icol);
            CHECK_STATUS(parseName);

            /* get the upper bound value */
            status = str2val(str2, MODL, &bound, &dot);
            CHECK_STATUS(str2val:bound);

            /* if name did not contain a "[", then set all rows and columns */
            if (strstr(str1, "[") == NULL) {
                indx = 0;
                for (irow = 1; irow <= MODL->pmtr[jpmtr].nrow; irow++) {
                    for (icol = 1; icol <= MODL->pmtr[jpmtr].ncol; icol++) {
                        MODL->pmtr[jpmtr].ubnd[indx] = bound;
                        indx++;
                    }
                }

            /* otherwise just set the requested value */
            } else {
                if (irow < 1 || irow > MODL->pmtr[jpmtr].nrow ||
                    icol < 1 || icol > MODL->pmtr[jpmtr].ncol   ) {
                    SET_STATUS(OCSM_ILLEGAL_PMTR_INDEX, ubound);
                }

                indx = icol-1 + (irow-1) * MODL->pmtr[jpmtr].ncol;
                MODL->pmtr[jpmtr].ubnd[indx] = bound;
            }

        /* input is: "udparg $primtype $argName1 argValue1 $argName2 argValue2 $argName3 argValue3 $argName4 argValue4" */
        } else if (strcmp(command, "udparg") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, udparg);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, udparg);
            }

            /* extract arguments */
            strcpy(str1, "$");  strcpy(str2, "$");  strcpy(str4, "$");  strcpy(str6, "$");  strcpy(str8, "$");
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s %2047s %2047s %2047s %2047s %2047s %2047s\n",
                          &(str1[1]), &(str2[1]), str3, &(str4[1]), str5,
                                      &(str6[1]), str7, &(str8[1]), str9);
            if (narg < 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, udparg);
            } else if (narg%2 != 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, udparg);
            }

            if (narg < 3) str3[0] = '\0';
            if (narg < 5) str5[0] = '\0';
            if (narg < 7) str7[0] = '\0';
            if (narg < 9) str9[0] = '\0';

            /* create the new Branch */
            if        (narg < 3) {
                status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_UDPARG,
                                     str1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            } else if (narg < 5) {
                status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_UDPARG,
                                     str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            } else if (narg < 7) {
                status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_UDPARG,
                                     str1, str2, str3, str4, str5, NULL, NULL, NULL, NULL);
            } else if (narg < 9) {
                status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_UDPARG,
                                     str1, str2, str3, str4, str5, str6, str7, NULL, NULL);
            } else {
                status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_UDPARG,
                                     str1, str2, str3, str4, str5, str6, str7, str8, str9);
            }
            CHECK_STATUS(ocsmNewBrch);

        /* input is: "udprim $primtype $argName1 argValue1 $argName2 argValue2 $argName3 argValue3 $argName4 argValue4" */
        } else if (strcmp(command, "udprim") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, udprim);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, udprim);
            }

            /* extract arguments */
            strcpy(str1, "$");  strcpy(str2, "$");  strcpy(str4, "$");  strcpy(str6, "$");  strcpy(str8, "$");
            narg = sscanf(nextline, "%*s %2047s %2047s %2047s %2047s %2047s %2047s %2047s %2047s %2047s\n",
                          &(str1[1]), &(str2[1]), str3, &(str4[1]), str5,
                                      &(str6[1]), str7, &(str8[1]), str9);
            if (narg < 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, udprim);
            } else if (narg%2 != 1) {
                SET_STATUS(OCSM_NOT_ENOUGH_ARGS, udprim);
            }

            if (narg < 3) str3[0] = '\0';
            if (narg < 5) str5[0] = '\0';
            if (narg < 7) str7[0] = '\0';
            if (narg < 9) str9[0] = '\0';

            /* extract "in-line" file if str3, str5, str7, or str9 is << */
            if (strncmp(str3, "<<", MAX_EXPR_LEN) == 0 ||
                strncmp(str5, "<<", MAX_EXPR_LEN) == 0 ||
                strncmp(str7, "<<", MAX_EXPR_LEN) == 0 ||
                strncmp(str9, "<<", MAX_EXPR_LEN) == 0   ) {

                /* look for a tmp file that does not exist */
                for (itmp = 0; itmp < 999; itmp++) {
                    sprintf(tmpfilename, "tmp_OpenCSM_%03d", itmp);

                    tmp_file = fopen(tmpfilename, "r");
                    if (tmp_file != NULL) {
                        fclose(tmp_file);
                        continue;
                    }

                    /* we found an unused tmp file, so copy from filename to tmpfilename now */
                    tmp_file = fopen(tmpfilename, "w");
                    if (tmp_file != NULL) {
                        while (1) {
                            temp = fgets(tmpfileline, MAX_LINE_LEN, csm_file[MODL->level]);
                            if (temp == NULL) break;

                            if (strncmp(tmpfileline, ">>", 2) == 0) {
                                break;
                            }

                            itmp = fputs(tmpfileline, tmp_file);
                            if (itmp < 0) break;
                        }

                        fclose(tmp_file);
                        tmp_file = NULL;
                        SPRINT1(1, "--> tmp file \"%s\" created", tmpfilename);

                        if        (strcmp(str3, "<<") == 0) {
                            sprintf(str3, "$%s", tmpfilename);
                        } else if (strcmp(str5, "<<") == 0) {
                            sprintf(str5, "$%s", tmpfilename);
                        } else if (strcmp(str7, "<<") == 0) {
                            sprintf(str7, "$%s", tmpfilename);
                        } else {
                            sprintf(str9, "$%s", tmpfilename);
                        }

                        break;
                    }
                }
            }

            /* convert arguments that start with "$/" to relative pathnames */
            if (strncmp(str3, "$$/", 3) == 0) {
                STRNCPY(pathname, MODL->filename[MODL->level], MAX_EXPR_LEN);
                i = strlen(pathname) - 1;
                while (pathname[i] != SLASH && i >= 0) {
                    pathname[i] = '\0';
                    i--;
                }
                STRNCPY(tmpfilename, &(str3[3]), MAX_EXPR_LEN);
                snprintf(str3, MAX_EXPR_LEN, "$%s%s", pathname, tmpfilename);
            }
            if (strncmp(str5, "$$/", 3) == 0 ) {
                STRNCPY(pathname, MODL->filename[MODL->level], MAX_EXPR_LEN);
                i = strlen(pathname) - 1;
                while (pathname[i] != SLASH && i >= 0) {
                    pathname[i] = '\0';
                    i--;
                }
                STRNCPY(tmpfilename, &(str5[3]), MAX_EXPR_LEN);
                snprintf(str5, MAX_EXPR_LEN, "$%s%s", pathname, tmpfilename);
            }
            if (strncmp(str7, "$$/", 3) == 0 ) {
                STRNCPY(pathname, MODL->filename[MODL->level], MAX_EXPR_LEN);
                i = strlen(pathname) - 1;
                while (pathname[i] != SLASH && i >= 0) {
                    pathname[i] = '\0';
                    i--;
                }
                STRNCPY(tmpfilename, &(str7[3]), MAX_EXPR_LEN);
                snprintf(str7, MAX_EXPR_LEN, "$%s%s", pathname, tmpfilename);
            }
            if (strncmp(str9, "$$/", 3) == 0 ) {
                STRNCPY(pathname, MODL->filename[MODL->level], MAX_EXPR_LEN);
                i = strlen(pathname) - 1;
                while (pathname[i] != SLASH && i >= 0) {
                    pathname[i] = '\0';
                    i--;
                }
                STRNCPY(tmpfilename, &(str9[3]), MAX_EXPR_LEN);
                snprintf(str9, MAX_EXPR_LEN, "$%s%s", pathname, tmpfilename);
            }

            /* create the new Branch */
            if        (narg < 3) {
                status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_UDPRIM,
                                     str1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            } else if (narg < 5) {
                status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_UDPRIM,
                                     str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            } else if (narg < 7) {
                status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_UDPRIM,
                                     str1, str2, str3, str4, str5, NULL, NULL, NULL, NULL);
            } else if (narg < 9) {
                status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_UDPRIM,
                                     str1, str2, str3, str4, str5, str6, str7, NULL, NULL);
            } else {
                status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_UDPRIM,
                                     str1, str2, str3, str4, str5, str6, str7, str8, str9);
            }
            CHECK_STATUS(ocsmNewBrch);

            /* if the primtype begins with a slash, open that file now */
            if (str1[0] == '$' && str1[1] == '/') {
                if (MODL->level < 10) {

                    strcpy(MODL->filename[MODL->level+1], &(str1[2]));
                    strcat(MODL->filename[MODL->level+1], ".udc");

                    SPRINT1(1, "\n>>> Diverting to file %s\n", MODL->filename[MODL->level+1]);

                    (MODL->level)++;
                    csm_file[MODL->level] = fopen(MODL->filename[MODL->level], "r");
                    if (csm_file[MODL->level] == NULL) {
                        status = OCSM_FILE_NOT_FOUND;
                        goto cleanup;
                    }
                } else {
                    signalError(MODL, OCSM_NESTED_TOO_DEEPLY, 0,
                                "maximum file depth reached");
                }

            /* if the primtype begins with a dollar-slash, open that file now */
            } else if (str1[0] == '$' && str1[1] == '$' && str1[2] == '/') {
                if (MODL->level < 10) {

                    STRNCPY(pathname, MODL->filename[MODL->level], MAX_EXPR_LEN);
                    i = strlen(pathname) - 1;
                    while (pathname[i] != SLASH && i >= 0) {
                        pathname[i] = '\0';
                        i--;
                    }
                    snprintf(MODL->filename[MODL->level+1], MAX_EXPR_LEN, "%s%s.udc", pathname, &(str1[3]));

                    SPRINT1(1, "\n>>> Diverting to file %s\n", MODL->filename[MODL->level+1]);

                    (MODL->level)++;
                    csm_file[MODL->level] = fopen(MODL->filename[MODL->level], "r");
                    if (csm_file[MODL->level] == NULL) {
                        status = OCSM_FILE_NOT_FOUND;
                        goto cleanup;
                    }
                } else {
                    signalError(MODL, OCSM_NESTED_TOO_DEEPLY, 0,
                                "maximum file depth reached");
                }
            }

        /* input is: "union tomark=0 trimList=0 maxtol=0" */
        } else if (strcmp(command, "union") == 0) {
            if (nskpt > 0) {
                SET_STATUS(OCSM_SKETCH_IS_OPEN, union);
            } else if (insolver != 0) {
                SET_STATUS(OCSM_SOLVER_IS_OPEN, union);
            }

            /* extract arguments */
            narg = sscanf(nextline, "%*s %2047s %2047s %2027s\n",
                          str1, str2, str3);
            if (narg < 1) {
                STRNCPY(str1, "0", MAX_EXPR_LEN);
            }
            if (narg < 2) {
                STRNCPY(str2, "0", MAX_EXPR_LEN);
            }
            if (narg < 3) {
                STRNCPY(str3, "0", MAX_EXPR_LEN);
            }

            /* create the new Branch */
            status = ocsmNewBrch(MODL, MODL->nbrch, OCSM_UNION,
                                 str1, str2, str3, NULL, NULL, NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);

        /* illegal command */
        } else {
            status = OCSM_ILLEGAL_STATEMENT;
            goto cleanup;
        }
    }

    /* close the .csm file */
    if (csm_file[MODL->level] != NULL) fclose(csm_file[MODL->level]);

cleanup:
    FREE(str9);
    FREE(str8);
    FREE(str7);
    FREE(str6);
    FREE(str5);
    FREE(str4);
    FREE(str3);
    FREE(str2);
    FREE(str1);

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmSave - save a MODL to a file                                   *
 *                                                                      *
 ************************************************************************
 */

int
ocsmSave(void   *modl,                  /* (in)  pointer to MODL */
         char   filename[])             /* (in)  file to be written (with .csm) */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       ipmtr, ibrch, jbrch, iattr, irow, icol, index, indent;
    FILE      *csm_file;

    ROUTINE(ocsmSave);
    DPRINT2("%s(filename=%s) {",
            routine, filename);

    /* --------------------------------------------------------------- */

    SPRINT1(1, "--> enter ocsmSave(filename=%s)", filename);

    /* open file */
    if (strlen(filename) == 0) {
        status = OCSM_FILE_NOT_FOUND;
        goto cleanup;
    }

    csm_file = fopen(filename, "w");
    if (csm_file == NULL) {
        status = OCSM_FILE_NOT_FOUND;
        goto cleanup;
    }

    /* write header */
    fprintf(csm_file, "# %s written by ocsmSave\n", filename);

    /* write the design (external) Parameters */
    fprintf(csm_file, "\n# Design Parameters:\n");
    for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
        if (MODL->pmtr[ipmtr].nrow > 1 ||
            MODL->pmtr[ipmtr].ncol > 1   ) {
            if (MODL->pmtr[ipmtr].type  == OCSM_EXTERNAL &&
                MODL->pmtr[ipmtr].level == MODL->level     ) {
                fprintf(csm_file, "dimension   %s   %d   %d   1\n",
                        MODL->pmtr[ipmtr].name,
                        MODL->pmtr[ipmtr].nrow,
                        MODL->pmtr[ipmtr].ncol);
            } else if (MODL->pmtr[ipmtr].level == MODL->level) {
                fprintf(csm_file, "dimension   %s   %d   %d   0\n",
                        MODL->pmtr[ipmtr].name,
                        MODL->pmtr[ipmtr].nrow,
                        MODL->pmtr[ipmtr].ncol);
            }
        }

        if (MODL->pmtr[ipmtr].type  == OCSM_EXTERNAL &&
            MODL->pmtr[ipmtr].level == MODL->level     ) {
            index = 0;
            for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                    fprintf(csm_file, "despmtr   %s[%d,%d]   %11.5f\n",
                            MODL->pmtr[ipmtr].name, irow, icol,
                            MODL->pmtr[ipmtr].value[index]);
                    index++;
                }
            }
        }
    }

    /* write the Bounds for any (external) Parameters */
    for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
        if (MODL->pmtr[ipmtr].type  != OCSM_EXTERNAL ||
            MODL->pmtr[ipmtr].level != MODL->level     ) continue;

        index = 0;
        for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
            for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                if (MODL->pmtr[ipmtr].lbnd[index] > -HUGEQ) {
                    fprintf(csm_file, "lbound      %s[%d,%d]   %f\n",
                            MODL->pmtr[ipmtr].name,
                            irow, icol,
                            MODL->pmtr[ipmtr].lbnd[index]);
                }
                if (MODL->pmtr[ipmtr].ubnd[index] < +HUGEQ) {
                    fprintf(csm_file, "ubound      %s[%d,%d]   %f\n",
                            MODL->pmtr[ipmtr].name,
                            irow, icol,
                            MODL->pmtr[ipmtr].ubnd[index]);
                }

                index++;
            }
        }
    }

    /* write the global Attributes */
    for (iattr = 0; iattr < MODL->nattr; iattr++) {
        fprintf(csm_file, "gattrib   %s   %s\n",
                MODL->attr[iattr].name,
                MODL->attr[iattr].value);
    }

    /* make sure indentation is up to date */
    status = ocsmCheck(MODL);
    CHECK_STATUS(ocsmCheck);

    /* write the Branches */
    fprintf(csm_file, "\n# Branches:\n");
    for (ibrch = 1; ibrch <= MODL->nbrch; ibrch++) {
        for (indent = 0; indent < MODL->brch[ibrch].indent; indent++) {
            fprintf(csm_file, "   ");
        }

        if        (MODL->brch[ibrch].type == OCSM_ARC) {
            fprintf(csm_file, "arc       %s   %s   %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3,
                    MODL->brch[ibrch].arg4,
                  &(MODL->brch[ibrch].arg5[1]));
        } else if (MODL->brch[ibrch].type == OCSM_ASSERT) {
            fprintf(csm_file, "assert    %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2);
        } else if (MODL->brch[ibrch].type == OCSM_BEZIER) {
            fprintf(csm_file, "bezier    %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3);
        } else if (MODL->brch[ibrch].type == OCSM_BLEND) {
            fprintf(csm_file, "blend     %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3);
        } else if (MODL->brch[ibrch].type == OCSM_BOX) {
            fprintf(csm_file, "box       %s   %s   %s   %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3,
                    MODL->brch[ibrch].arg4,
                    MODL->brch[ibrch].arg5,
                    MODL->brch[ibrch].arg6);
        } else if (MODL->brch[ibrch].type == OCSM_CHAMFER) {
            fprintf(csm_file, "chamfer   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2);
        } else if (MODL->brch[ibrch].type == OCSM_CIRARC) {
            fprintf(csm_file, "cirarc    %s   %s   %s   %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3,
                    MODL->brch[ibrch].arg4,
                    MODL->brch[ibrch].arg5,
                    MODL->brch[ibrch].arg6);
        } else if (MODL->brch[ibrch].type == OCSM_COMBINE) {
            fprintf(csm_file, "combine   %s\n",
                    MODL->brch[ibrch].arg1);
        } else if (MODL->brch[ibrch].type == OCSM_CONE) {
            fprintf(csm_file, "cone      %s   %s   %s   %s   %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3,
                    MODL->brch[ibrch].arg4,
                    MODL->brch[ibrch].arg5,
                    MODL->brch[ibrch].arg6,
                    MODL->brch[ibrch].arg7);
        } else if (MODL->brch[ibrch].type == OCSM_CONNECT) {
            fprintf(csm_file, "connect     %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2);
        } else if (MODL->brch[ibrch].type == OCSM_CYLINDER) {
            fprintf(csm_file, "cylinder  %s   %s   %s   %s   %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3,
                    MODL->brch[ibrch].arg4,
                    MODL->brch[ibrch].arg5,
                    MODL->brch[ibrch].arg6,
                    MODL->brch[ibrch].arg7);
        } else if (MODL->brch[ibrch].type == OCSM_DUMP) {
            fprintf(csm_file, "dump      %s   %s\n",
                  &(MODL->brch[ibrch].arg1[1]),
                    MODL->brch[ibrch].arg2    );
        } else if (MODL->brch[ibrch].type == OCSM_END) {
            fprintf(csm_file, "end\n");
        } else if (MODL->brch[ibrch].type == OCSM_EXTRACT) {
            fprintf(csm_file, "extract   %s\n",
                    MODL->brch[ibrch].arg1);
        } else if (MODL->brch[ibrch].type == OCSM_EXTRUDE) {
            fprintf(csm_file, "extrude   %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3);
        } else if (MODL->brch[ibrch].type == OCSM_FILLET) {
            fprintf(csm_file, "fillet    %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2);
        } else if (MODL->brch[ibrch].type == OCSM_HOLLOW) {
            fprintf(csm_file, "hollow    %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2);
        } else if (MODL->brch[ibrch].type == OCSM_IMPORT) {
            fprintf(csm_file, "import    %s  %s\n",
                  &(MODL->brch[ibrch].arg1[1]),
                    MODL->brch[ibrch].arg2);
        } else if (MODL->brch[ibrch].type == OCSM_INTERSECT) {
            fprintf(csm_file, "intersect %s   %s   %s\n",
                  &(MODL->brch[ibrch].arg1[1]),
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3    );
        } else if (MODL->brch[ibrch].type == OCSM_JOIN) {
            fprintf(csm_file, "join     %s\n",
                    MODL->brch[ibrch].arg1);
        } else if (MODL->brch[ibrch].type == OCSM_LINSEG) {
            fprintf(csm_file, "linseg    %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3);
        } else if (MODL->brch[ibrch].type == OCSM_LOFT) {
            fprintf(csm_file, "loft      %s\n",
                    MODL->brch[ibrch].arg1);
        } else if (MODL->brch[ibrch].type == OCSM_MACBEG) {
            fprintf(csm_file, "macbeg    %s\n",
                    MODL->brch[ibrch].arg1);
        } else if (MODL->brch[ibrch].type == OCSM_MACEND) {
            fprintf(csm_file, "macend\n");
        } else if (MODL->brch[ibrch].type == OCSM_MARK) {
            fprintf(csm_file, "mark\n");
        } else if (MODL->brch[ibrch].type == OCSM_MIRROR) {
            fprintf(csm_file, "mirror    %s   %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3,
                    MODL->brch[ibrch].arg4);
        } else if (MODL->brch[ibrch].type == OCSM_PATBEG) {
            fprintf(csm_file, "patbeg    %s %s\n",
                  &(MODL->brch[ibrch].arg1[1]),
                    MODL->brch[ibrch].arg2    );
        } else if (MODL->brch[ibrch].type == OCSM_PATBREAK) {
            fprintf(csm_file, "patbreak  %s\n",
                    MODL->brch[ibrch].arg1);
        } else if (MODL->brch[ibrch].type == OCSM_PATEND) {
            fprintf(csm_file, "patend\n");
        } else if (MODL->brch[ibrch].type == OCSM_PROJECT) {
            fprintf(csm_file, "project   %s   %s   %s   %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3,
                    MODL->brch[ibrch].arg4,
                    MODL->brch[ibrch].arg5,
                    MODL->brch[ibrch].arg6);
        } else if (MODL->brch[ibrch].type == OCSM_RECALL) {
            fprintf(csm_file, "recall    %s\n",
                    MODL->brch[ibrch].arg1);
        } else if (MODL->brch[ibrch].type == OCSM_REORDER) {
            fprintf(csm_file, "reorder  %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2);
        } else if (MODL->brch[ibrch].type == OCSM_RESTORE) {
            fprintf(csm_file, "restore   %s %s\n",
                  &(MODL->brch[ibrch].arg1[1]),
                    MODL->brch[ibrch].arg2    );
        } else if (MODL->brch[ibrch].type == OCSM_REVOLVE) {
            fprintf(csm_file, "revolve   %s   %s   %s   %s   %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3,
                    MODL->brch[ibrch].arg4,
                    MODL->brch[ibrch].arg5,
                    MODL->brch[ibrch].arg6,
                    MODL->brch[ibrch].arg7);
        } else if (MODL->brch[ibrch].type == OCSM_ROTATEX) {
            fprintf(csm_file, "rotatex   %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3);
        } else if (MODL->brch[ibrch].type == OCSM_ROTATEY) {
            fprintf(csm_file, "rotatey   %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3);
        } else if (MODL->brch[ibrch].type == OCSM_ROTATEZ) {
            fprintf(csm_file, "rotatez   %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3);
        } else if (MODL->brch[ibrch].type == OCSM_RULE) {
            fprintf(csm_file, "rule      %s\n",
                    MODL->brch[ibrch].arg1);
        } else if (MODL->brch[ibrch].type == OCSM_SCALE) {
            fprintf(csm_file, "scale     %s\n",
                    MODL->brch[ibrch].arg1);
        } else if (MODL->brch[ibrch].type == OCSM_SELECT) {
            fprintf(csm_file, "select    %s",
                  &(MODL->brch[ibrch].arg1[1]));
            if (MODL->brch[ibrch].narg >= 3) {
                fprintf(csm_file, "   %s",
                    MODL->brch[ibrch].arg2);
            }
            if (MODL->brch[ibrch].narg >= 4) {
                fprintf(csm_file, "   %s",
                    MODL->brch[ibrch].arg3);
            }
            if (MODL->brch[ibrch].narg >= 5) {
                fprintf(csm_file, "   %s",
                    MODL->brch[ibrch].arg4);
            }
            if (MODL->brch[ibrch].narg >= 6) {
                fprintf(csm_file, "   %s",
                    MODL->brch[ibrch].arg5);
            }
            if (MODL->brch[ibrch].narg >= 7) {
                fprintf(csm_file, "   %s",
                    MODL->brch[ibrch].arg6);
            }
            if (MODL->brch[ibrch].narg >= 8) {
                fprintf(csm_file, "   %s",
                    MODL->brch[ibrch].arg7);
            }
            if (MODL->brch[ibrch].narg >= 9) {
                fprintf(csm_file, "   %s",
                    MODL->brch[ibrch].arg8);
            }
            fprintf(csm_file, "\n");
        } else if (MODL->brch[ibrch].type == OCSM_SET) {
            fprintf(csm_file, "set       %s %s\n",
                  &(MODL->brch[ibrch].arg1[1]),
                    MODL->brch[ibrch].arg2);
        } else if (MODL->brch[ibrch].type == OCSM_SKBEG) {
            fprintf(csm_file, "skbeg     %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3);
        } else if (MODL->brch[ibrch].type == OCSM_SKCON) {
            fprintf(csm_file, "skcon     %s   %s   %s  %s\n",
                  &(MODL->brch[ibrch].arg1[1]),
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3,
                  &(MODL->brch[ibrch].arg4[1]));
        } else if (MODL->brch[ibrch].type == OCSM_SKEND) {
            fprintf(csm_file, "skend     %s\n",
                    MODL->brch[ibrch].arg1);
        } else if (MODL->brch[ibrch].type == OCSM_SKVAR) {
            fprintf(csm_file, "skvar     %s  %s\n",
                  &(MODL->brch[ibrch].arg1[1]),
                    MODL->brch[ibrch].arg2);
        } else if (MODL->brch[ibrch].type == OCSM_SOLBEG) {
            fprintf(csm_file, "solbeg    %s\n",
                  &(MODL->brch[ibrch].arg1[1]));
        } else if (MODL->brch[ibrch].type == OCSM_SOLCON) {
            fprintf(csm_file, "solcon    %s\n",
                  &(MODL->brch[ibrch].arg1[1]));
        } else if (MODL->brch[ibrch].type == OCSM_SOLEND) {
            fprintf(csm_file, "solend\n");
        } else if (MODL->brch[ibrch].type == OCSM_SPHERE) {
            fprintf(csm_file, "sphere    %s   %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3,
                    MODL->brch[ibrch].arg4);
        } else if (MODL->brch[ibrch].type == OCSM_SPLINE) {
            fprintf(csm_file, "spline    %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3);
        } else if (MODL->brch[ibrch].type == OCSM_STORE) {
            fprintf(csm_file, "store     %s %s %s\n",
                  &(MODL->brch[ibrch].arg1[1]),
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3    );
        } else if (MODL->brch[ibrch].type == OCSM_SUBTRACT) {
            fprintf(csm_file, "subtract  %s   %s   %s\n",
                  &(MODL->brch[ibrch].arg1[1]),
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3    );
        } else if (MODL->brch[ibrch].type == OCSM_SWEEP) {
            fprintf(csm_file, "sweep\n");
        } else if (MODL->brch[ibrch].type == OCSM_TORUS) {
            fprintf(csm_file, "torus     %s   %s   %s   %s   %s   %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3,
                    MODL->brch[ibrch].arg4,
                    MODL->brch[ibrch].arg5,
                    MODL->brch[ibrch].arg6,
                    MODL->brch[ibrch].arg7,
                    MODL->brch[ibrch].arg8);
        } else if (MODL->brch[ibrch].type == OCSM_TRANSLATE) {
            fprintf(csm_file, "translate %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3);
        } else if (MODL->brch[ibrch].type == OCSM_UDPARG) {
            fprintf(csm_file, "udparg    %s",
                  &(MODL->brch[ibrch].arg1[1]));
            if (MODL->brch[ibrch].narg >= 3) {
                fprintf(csm_file, "   %s   %s",
                        &(MODL->brch[ibrch].arg2[1]),
                          MODL->brch[ibrch].arg3);
            }
            if (MODL->brch[ibrch].narg >= 5) {
                fprintf(csm_file, "   %s   %s",
                        &(MODL->brch[ibrch].arg4[1]),
                          MODL->brch[ibrch].arg5);
            }
            if (MODL->brch[ibrch].narg >= 7) {
                fprintf(csm_file, "   %s   %s",
                        &(MODL->brch[ibrch].arg6[1]),
                          MODL->brch[ibrch].arg7);
            }
            if (MODL->brch[ibrch].narg >= 9) {
                fprintf(csm_file, "   %s   %s",
                        &(MODL->brch[ibrch].arg8[1]),
                          MODL->brch[ibrch].arg9);
            }
            fprintf(csm_file, "\n");
        } else if (MODL->brch[ibrch].type == OCSM_UDPRIM) {
            fprintf(csm_file, "udprim    %s",
                  &(MODL->brch[ibrch].arg1[1]));
            if (MODL->brch[ibrch].narg >= 3) {
                fprintf(csm_file, "   %s   %s",
                        &(MODL->brch[ibrch].arg2[1]),
                          MODL->brch[ibrch].arg3);
            }
            if (MODL->brch[ibrch].narg >= 5) {
                fprintf(csm_file, "   %s   %s",
                        &(MODL->brch[ibrch].arg4[1]),
                          MODL->brch[ibrch].arg5);
            }
            if (MODL->brch[ibrch].narg >= 7) {
                fprintf(csm_file, "   %s   %s",
                        &(MODL->brch[ibrch].arg6[1]),
                          MODL->brch[ibrch].arg7);
            }
            if (MODL->brch[ibrch].narg >= 9) {
                fprintf(csm_file, "   %s   %s",
                        &(MODL->brch[ibrch].arg8[1]),
                          MODL->brch[ibrch].arg9);
            }
            fprintf(csm_file, "\n");

            /* skip Branches within */
            if ((MODL->brch[ibrch].arg1[1] == '/'                                    ) ||
                (MODL->brch[ibrch].arg1[1] == '$' && MODL->brch[ibrch].arg1[2] == '/')   ) {
                jbrch = ibrch;
                while (MODL->brch[ibrch].type   != OCSM_END                ||
                       MODL->brch[ibrch].indent != MODL->brch[jbrch].indent  ) {
                    ibrch++;
                }
            }
        } else if (MODL->brch[ibrch].type == OCSM_UNION) {
            fprintf(csm_file, "union     %s   %s   %s\n",
                    MODL->brch[ibrch].arg1,
                    MODL->brch[ibrch].arg2,
                    MODL->brch[ibrch].arg3);
        }

        /* write the name of the Branch (if not the default name) */
        if (strncmp(MODL->brch[ibrch].name, "Brch_", 5) != 0) {
            fprintf(csm_file, "name      %s\n",
                    MODL->brch[ibrch].name);
        }

        /* write the Attributes for the Branch */
        for (iattr = 0; iattr < MODL->brch[ibrch].nattr; iattr++) {
            fprintf(csm_file, "attribute %s   %s\n",
                    MODL->brch[ibrch].attr[iattr].name,
                    MODL->brch[ibrch].attr[iattr].value);
        }
    }

    /* close the file */
    fprintf(csm_file, "\nend\n");
    fclose(csm_file);

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmCopy - copy a MODL                                             *
 *                                                                      *
 ************************************************************************
 */

int
ocsmCopy(void   *srcModl,               /* (in)  pointer to source MODL */
         void   **newModl)              /* (out) pointer to new    MODL */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *SRC_MODL = (modl_T*)srcModl;
    modl_T    *NEW_MODL;

    int       ibrch, iattr, istor, ipmtr, irow, icol, index, i;

    ROUTINE(ocsmCopy);
    DPRINT1("%s() {",
            routine);

    /* --------------------------------------------------------------- */

    /* default return value */
    *newModl = NULL;

    /* check magic number */
    if (SRC_MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (SRC_MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* make a new MODL and initialize it */
    MALLOC(NEW_MODL, modl_T, 1);

    NEW_MODL->magic   = OCSM_MAGIC;
    NEW_MODL->checked = 0;
    NEW_MODL->nextseq = 1;
    NEW_MODL->recycle = 0;

    NEW_MODL->sel_body = -1;
    NEW_MODL->sel_face = -1;
    NEW_MODL->sel_edge = -1;
    NEW_MODL->sel_node = -1;

    NEW_MODL->level = SRC_MODL->level;
    for (i = 0; i < 11; i++) {
        STRNCPY(NEW_MODL->filename[i], SRC_MODL->filename[i], 255);
    }

    NEW_MODL->nattr = 0;
    NEW_MODL->attr  = NULL;

    NEW_MODL->nstor = 0;
    NEW_MODL->stor  = NULL;

    NEW_MODL->nbrch = 0;
    NEW_MODL->mbrch = 0;
    NEW_MODL->brch  = NULL;

    NEW_MODL->npmtr = 0;
    NEW_MODL->mpmtr = 0;
    NEW_MODL->pmtr  = NULL;

    NEW_MODL->nbody = 0;
    NEW_MODL->mbody = 0;
    NEW_MODL->body  = NULL;

    NEW_MODL->perturb  = NULL;
    NEW_MODL->basemodl = NULL;
    NEW_MODL->dtime    = 0;

    NEW_MODL->context = NULL;

    NEW_MODL->errorBuf = NULL;
    MALLOC(NEW_MODL->errorBuf, char, MAX_STR_LEN);
    NEW_MODL->errorBuf[0] = '\0';

    for (i = 0; i < 100; i++) {
        NEW_MODL->profile[i].ncall = SRC_MODL->profile[i].ncall;
        NEW_MODL->profile[i].time  = SRC_MODL->profile[i].time;
    }

    /* return value */
    *newModl = NEW_MODL;

    /* copy the global Parameters */
    for (iattr = 0; iattr < SRC_MODL->nattr; iattr++) {
        status = ocsmSetAttr(NEW_MODL, 0,
                             SRC_MODL->attr[iattr].name,
                             SRC_MODL->attr[iattr].value);
    }

    /* copy the Storge locations */
    NEW_MODL->nstor = SRC_MODL->nstor;

    MALLOC(NEW_MODL->stor, stor_T, NEW_MODL->nstor);

    for (istor = 0; istor < NEW_MODL->nstor; istor++) {
        STRNCPY(NEW_MODL->stor[istor].name, SRC_MODL->stor[istor].name, MAX_NAME_LEN);

        NEW_MODL->stor[istor].index = SRC_MODL->stor[istor].index;
        NEW_MODL->stor[istor].ibody = SRC_MODL->stor[istor].ibody;
        NEW_MODL->stor[istor].ebody = SRC_MODL->stor[istor].ebody;
    }

    /* copy the Parameter table (done before Branch table so that
       new Parameters do not get created by a OCSM_SET or OCSM_PATBEG) */
    for (ipmtr = 1; ipmtr <= SRC_MODL->npmtr; ipmtr++) {
        status = ocsmNewPmtr(NEW_MODL,
                             SRC_MODL->pmtr[ipmtr].name,
                             SRC_MODL->pmtr[ipmtr].type,
                             SRC_MODL->pmtr[ipmtr].nrow,
                             SRC_MODL->pmtr[ipmtr].ncol);
        CHECK_STATUS(ocsmNewPmtr);

        index = 0;
        for (irow = 1; irow <= SRC_MODL->pmtr[ipmtr].nrow; irow++) {
            for (icol = 1; icol <= SRC_MODL->pmtr[ipmtr].ncol; icol++) {
                NEW_MODL->pmtr[ipmtr].value[index] = SRC_MODL->pmtr[ipmtr].value[index];
                NEW_MODL->pmtr[ipmtr].dot[  index] = SRC_MODL->pmtr[ipmtr].dot[  index];
                NEW_MODL->pmtr[ipmtr].lbnd[ index] = SRC_MODL->pmtr[ipmtr].lbnd[ index];
                NEW_MODL->pmtr[ipmtr].ubnd[ index] = SRC_MODL->pmtr[ipmtr].ubnd[ index];
                index++;
            }
        }
    }

    /* copy the Branch table */
    for (ibrch = 1; ibrch <= SRC_MODL->nbrch; ibrch++) {
        status = ocsmNewBrch(NEW_MODL,
                             NEW_MODL->nbrch,
                             SRC_MODL->brch[ibrch].type,
                             SRC_MODL->brch[ibrch].arg1,
                             SRC_MODL->brch[ibrch].arg2,
                             SRC_MODL->brch[ibrch].arg3,
                             SRC_MODL->brch[ibrch].arg4,
                             SRC_MODL->brch[ibrch].arg5,
                             SRC_MODL->brch[ibrch].arg6,
                             SRC_MODL->brch[ibrch].arg7,
                             SRC_MODL->brch[ibrch].arg8,
                             SRC_MODL->brch[ibrch].arg9);
        CHECK_STATUS(ocsmNewBrch);

        /* restore the old sequence number (over-writing the sequence
           number incremented in ocsmNewBrch) */
        NEW_MODL->nextseq = SRC_MODL->nextseq;

        /* copy the name and attributes */
        FREE(  NEW_MODL->brch[ibrch].name);
        MALLOC(NEW_MODL->brch[ibrch].name, char, (int)(strlen(SRC_MODL->brch[ibrch].name)+1));
        strcpy(NEW_MODL->brch[ibrch].name,                    SRC_MODL->brch[ibrch].name    );

        for (iattr = 0; iattr < SRC_MODL->brch[ibrch].nattr; iattr++) {
            status = ocsmSetAttr(NEW_MODL, ibrch,
                                 SRC_MODL->brch[ibrch].attr[iattr].name,
                                 SRC_MODL->brch[ibrch].attr[iattr].value);
        }
    }

    /* set the Branch's activity and link to the other Branches */
    for (ibrch = 1; ibrch <= SRC_MODL->nbrch; ibrch++) {
        NEW_MODL->brch[ibrch].actv  = SRC_MODL->brch[ibrch].actv;
        NEW_MODL->brch[ibrch].ileft = SRC_MODL->brch[ibrch].ileft;
        NEW_MODL->brch[ibrch].irite = SRC_MODL->brch[ibrch].irite;
        NEW_MODL->brch[ibrch].ichld = SRC_MODL->brch[ibrch].ichld;
    }

    /* do NOT copy the Body table */

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmFree - free all storage associated with a MODL                 *
 *                                                                      *
 ************************************************************************
 */

int
ocsmFree(
/*@null@*/void   *modl)                 /* (in)  pointer to MODL (or NULL) */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       itmp, ibrch, iattr, ipmtr, ibody, iarg;
    char      tmpfilename[MAX_EXPR_LEN];

    ROUTINE(ocsmFree);
    DPRINT1("%s() {",
            routine);

    /* --------------------------------------------------------------- */

    /* if modl is NULL, then we are cleaning up tmp files */
    if (MODL == NULL) {
        for (itmp = 0; itmp < 999; itmp++) {
            if (sprintf(tmpfilename, "tmp_OpenCSM_%03d", itmp) < 0) {
                break;
            }

            if (remove(tmpfilename) == 0) {
                SPRINT1(1, "--> tmp file \"%s\" removed", tmpfilename);
            }
        }

        goto cleanup;
    }

    /* check magic number */
    if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* if this has a perturbed Body, free it first */
    status = removePerturbation(MODL);
    CHECK_STATUS(removePerturbation);

    /* free up the Body table */
    for (ibody = 1; ibody <= MODL->nbody; ibody++) {
        if (MODL->body[ibody].etess != NULL) {
            status = EG_deleteObject(MODL->body[ibody].etess);
            CHECK_STATUS(EG_deleteObject);

            MODL->body[ibody].etess = NULL;
        }

        if (MODL->body[ibody].ebody != NULL) {
            ibrch = MODL->body[ibody].ibrch;
            if (MODL->brch[ibrch].type == OCSM_UDPRIM) {
                status = udp_clean(MODL->body[ibody].arg[1].str, MODL->body[ibody].ebody);
                if (status == EGADS_NOTFOUND) {
                    /* this means that some other Body probably cleaned up this udp
                     or that this is a user-defined component (udc) */
                } else {
                    CHECK_STATUS(udp_clean);
                }
            }

            status = EG_deleteObject(MODL->body[ibody].ebody);
            if (status == EGADS_EMPTY) status = SUCCESS;
            CHECK_STATUS(EG_deleteObject);

            MODL->body[ibody].ebody = NULL;
        }

        for (iarg = 1; iarg < 10; iarg++) {
            FREE(MODL->body[ibody].arg[iarg].val);
            FREE(MODL->body[ibody].arg[iarg].dot);
        }

        FREE(MODL->body[ibody].node);
        FREE(MODL->body[ibody].edge);
        FREE(MODL->body[ibody].face);
    }

    FREE(MODL->body);

    /* free up the Parameter table */
    for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
        FREE(MODL->pmtr[ipmtr].name );
        FREE(MODL->pmtr[ipmtr].value);
        FREE(MODL->pmtr[ipmtr].dot  );
        FREE(MODL->pmtr[ipmtr].lbnd );
        FREE(MODL->pmtr[ipmtr].ubnd );
    }

    FREE(MODL->pmtr);

    /* free up the Branch table */
    for (ibrch = 1; ibrch <= MODL->nbrch; ibrch++) {
        for (iattr = 0; iattr < MODL->brch[ibrch].nattr; iattr++) {
            FREE(MODL->brch[ibrch].attr[iattr].name );
            FREE(MODL->brch[ibrch].attr[iattr].value);
        }

        FREE(MODL->brch[ibrch].name);
        FREE(MODL->brch[ibrch].attr);
        FREE(MODL->brch[ibrch].arg1);
        FREE(MODL->brch[ibrch].arg2);
        FREE(MODL->brch[ibrch].arg3);
        FREE(MODL->brch[ibrch].arg4);
        FREE(MODL->brch[ibrch].arg5);
        FREE(MODL->brch[ibrch].arg6);
        FREE(MODL->brch[ibrch].arg7);
        FREE(MODL->brch[ibrch].arg8);
        FREE(MODL->brch[ibrch].arg9);
    }

    FREE(MODL->brch);

    /* free up the Storage table */
    FREE(MODL->stor);

    /* free up the message buffer */
    FREE(MODL->errorBuf);

    /* free up the MODL structure */
    FREE(MODL);

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmInfo - get info about a MODL                                   *
 *                                                                      *
 ************************************************************************
 */

int
ocsmInfo(void   *modl,                  /* (in)  pointer to MODL */
         int    *nbrch,                 /* (out) number of Branches */
         int    *npmtr,                 /* (out) number of Parameters */
         int    *nbody)                 /* (out) number of Bodys */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(ocsmInfo);
    DPRINT1("%s() {",
            routine);

    /* --------------------------------------------------------------- */

    /* default return values */
    *nbrch = -1;
    *npmtr = -1;

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* return values */
    *nbrch = MODL->nbrch;
    *npmtr = MODL->npmtr;
    *nbody = MODL->nbody;

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmCheck - check that Branches are properly ordered               *
 *                                                                      *
 ************************************************************************
 */

int
ocsmCheck(void   *modl)                 /* (in)  pointer to MODL */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       ibrch, ibrchl, ibrchr, jbrch, type, class;
    int       ipass, imark, nsketch, nmacro, npatn;
    int       ileft, ichld, indent;

    ROUTINE(ocsmCheck);
    DPRINT1("%s() {",
            routine);

    /* --------------------------------------------------------------- */

    SPRINT0(1, "--> enter ocsmCheck()");

    /* initially there are no errors */
    ipass = 0;
    npatn = 0;

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* initialize all indentation and parent/child relationships */
    for (ibrch = 1; ibrch <= MODL->nbrch; ibrch++) {
        MODL->brch[ibrch].indent = -1;
        MODL->brch[ibrch].ileft  = -1;
        MODL->brch[ibrch].irite  = -1;
        MODL->brch[ibrch].ichld  = -1;
    }

    /* set up the indentation */
    indent = 0;

    for (ibrch = 1; ibrch <= MODL->nbrch; ibrch++) {

        /* decrease indentation at the end of a block */
        if        (MODL->brch[ibrch].type == OCSM_MACEND  ||
                   MODL->brch[ibrch].type == OCSM_PATEND  ||
                   MODL->brch[ibrch].type == OCSM_SKEND   ||
                   MODL->brch[ibrch].type == OCSM_SOLEND  ||
                   MODL->brch[ibrch].type == OCSM_RULE    ||
                   MODL->brch[ibrch].type == OCSM_LOFT    ||
                   MODL->brch[ibrch].type == OCSM_BLEND   ||
                   MODL->brch[ibrch].type == OCSM_COMBINE ||
                   MODL->brch[ibrch].type == OCSM_END       ) {
            indent--;
        } else if (MODL->brch[ibrch].type == OCSM_UNION && strcmp(MODL->brch[ibrch].arg1, "1") == 0) {
            indent--;
        }

        /* set the indentation */
        MODL->brch[ibrch].indent = indent;

        /* increase indentation at the beginning of a block */
        if        (MODL->brch[ibrch].type == OCSM_MACBEG ||
                   MODL->brch[ibrch].type == OCSM_PATBEG ||
                   MODL->brch[ibrch].type == OCSM_SKBEG  ||
                   MODL->brch[ibrch].type == OCSM_SOLBEG ||
                   MODL->brch[ibrch].type == OCSM_MARK     ) {
            indent++;
        } else if (MODL->brch[ibrch].type    == OCSM_UDPRIM &&
                   MODL->brch[ibrch].arg1[1] == '/'           ) {
            indent++;
        } else if (MODL->brch[ibrch].type    == OCSM_UDPRIM &&
                   MODL->brch[ibrch].arg1[1] == '$'         &&
                   MODL->brch[ibrch].arg1[2] == '/'           ) {
            indent++;
        }
    }

    /* warn if improper nesting was found */
    if (indent == 1 && MODL->brch[MODL->nbrch].type != OCSM_END) {
        /* okay */
    } else if (indent != 0) {
        SPRINT1(0, "WARNING:: indent=%d, but should be 0", indent);
    }

    /* set up linkages to other Branches */
    for (ibrch = 1; ibrch <= MODL->nbrch; ibrch++) {
        type  = MODL->brch[ibrch].type;
        class = MODL->brch[ibrch].class;

        /* PRIMITIVEs have no parents */
        if (class == OCSM_PRIMITIVE) {
            MODL->brch[ibrch ].ichld = 0;

        /* RULEs, LOFTs, BLENDs, and COMBINEs must have a previous MARK. left parent
              is first branch (typically a SKEND or UDPRIM) without a child */
        } else if (type == OCSM_RULE  || type == OCSM_LOFT   ||
                   type == OCSM_BLEND || type == OCSM_COMBINE  ) {
            imark   = 0;
            nsketch = 0;
            for (jbrch = ibrch-1; jbrch > 0; jbrch--) {
                if (MODL->brch[jbrch].type == OCSM_MARK) {
                    imark = 1;
                    break;
                } else if (MODL->brch[jbrch].ichld == 0) {
                    nsketch++;

                    MODL->brch[ibrch].ileft = jbrch;
                    MODL->brch[jbrch].ichld = ibrch;

                    MODL->brch[ibrch].ichld = 0;
                }
            }

            if (imark == 0) {
                status = OCSM_MISSING_MARK;
                goto cleanup;
            } else if (nsketch < 1) {
                signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                            "RULE, LOFT, or BLEND expects 2 or more Bodys since MARK");
                status = OCSM_INSUFFICIENT_BODYS_ON_STACK;
                goto cleanup;
            } else if (nsketch >= MAX_NUM_SKETCHES) {
                signalError(MODL, OCSM_TOO_MANY_BODYS_ON_STACK, ibrch,
                            "too many Sketches (%d) in RULE, LOFT, BLEND, or COMBINE", nsketch);
                status = OCSM_TOO_MANY_BODYS_ON_STACK;
                goto cleanup;
            }

        /* GROWNs (except RULE, LOFT, and BLEND) have a left parent (typically a
               SKEND or UDPRIM) */
        } else if (class == OCSM_GROWN) {  /* except OCSM_RULE, OCSM_LOFT, and OCSM_BLEND */
            ibrchl = 0;
            for (jbrch = ibrch-1; jbrch > 0; jbrch--) {
                if (MODL->brch[jbrch].ichld == 0) {
                    ibrchl = jbrch;
                    break;
                }
            }

            if (ibrchl == 0) {
                signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                            "RULE, LOFT, or BLEND expects a Body on the stack");
                status = OCSM_INSUFFICIENT_BODYS_ON_STACK;
                goto cleanup;
            }

            MODL->brch[ibrch ].ileft = ibrchl;
            MODL->brch[ibrchl].ichld = ibrch;

            MODL->brch[ibrch ].ichld = 0;

        /* APPLIEDs have a Body as its left parent */
        } else if (class == OCSM_APPLIED) {
            ibrchl = 0;
            for (jbrch = ibrch-1; jbrch > 0; jbrch--) {
                if (MODL->brch[jbrch].ichld == 0) {
                    if (MODL->brch[jbrch].type != OCSM_SKEND) {
                        ibrchl = jbrch;
                        break;
                    } else {
                        signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                                    "FILLET, CHAMFER, or HOLLOW expects a Body on the stack");
                        status = OCSM_INSUFFICIENT_BODYS_ON_STACK;
                        goto cleanup;
                    }
                }
            }

            if (ibrchl == 0) {
                signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                            "FILLET, CHAMFER, or HOLLOW expects a Body on the stack");
                status = OCSM_INSUFFICIENT_BODYS_ON_STACK;
            }

            MODL->brch[ibrch ].ileft = ibrchl;
            MODL->brch[ibrchl].ichld = ibrch;

            MODL->brch[ibrch ].ichld = 0;

        /* EXTRACTs have one Body as left parent */
        } else if (type == OCSM_EXTRACT) {
            ibrchl = 0;
            for (jbrch = ibrch-1; jbrch > 0; jbrch--) {
                if (MODL->brch[jbrch].ichld == 0) {
                    if (ibrchl == 0) {
                        ibrchl = jbrch;
                        break;
                    }
                }
            }

            if (ibrchl == 0) {
                signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                            "EXTRACT expects a Body on the stack");
                status = OCSM_INSUFFICIENT_BODYS_ON_STACK;
                goto cleanup;
            }

            MODL->brch[ibrch ].ileft = ibrchl;
            MODL->brch[ibrchl].ichld = ibrch;

            MODL->brch[ibrch ].ichld = 0;

        /* BOOLEANs (except EXTRACT) have two Bodys as parents */
        } else if (class == OCSM_BOOLEAN) {
            ibrchl = 0;
            ibrchr = 0;
            for (jbrch = ibrch-1; jbrch > 0; jbrch--) {
                if (MODL->brch[jbrch].ichld == 0) {
                    if        (ibrchr == 0) {
                        ibrchr = jbrch;
                    } else if (ibrchl == 0) {
                        ibrchl = jbrch;
                        break;
                    }
                }
            }

            if (ibrchl == 0 || ibrchr == 0) {
                if (npatn == 0) {
                    signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                                "BOOLEANs expect 2 Bodys on the stack");
                    status = OCSM_INSUFFICIENT_BODYS_ON_STACK;
                    goto cleanup;
                } else {
                    SPRINT0(0, "WARNING:: Boolean in pattern cannot be checked");
                }
            }

            MODL->brch[ibrch ].ileft = ibrchl;
            MODL->brch[ibrchl].ichld = ibrch;

            MODL->brch[ibrch ].irite = ibrchr;
            MODL->brch[ibrchr].ichld = ibrch;

            MODL->brch[ibrch ].ichld = 0;

        /* TRANSFORMs has a Body or SKEND as parent */
        } else if (class == OCSM_TRANSFORM) {
            ibrchl = 0;
            for (jbrch = ibrch-1; jbrch > 0; jbrch--) {
                if (MODL->brch[jbrch].ichld == 0) {
                    if (ibrchl == 0) {
                        ibrchl = jbrch;
                        break;
                    }
                }
            }

            if (ibrchl == 0) {
                signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                            "TRANSFORMs expect a Body on the stack");
                status = OCSM_INSUFFICIENT_BODYS_ON_STACK;
                goto cleanup;
            }

            MODL->brch[ibrch ].ileft = ibrchl;
            MODL->brch[ibrchl].ichld = ibrch;

            MODL->brch[ibrch ].ichld = 0;

        /* SKBEGs has the SKEND as its child */
        } else if (type == OCSM_SKBEG) {
            ichld = 0;
            for (jbrch = ibrch+1; jbrch <= MODL->nbrch; jbrch++) {
                if (MODL->brch[jbrch].type == OCSM_SKEND) {
                    ichld = jbrch;
                    break;
                }
            }

            if (ichld == 0) {
                status = OCSM_SKETCH_DOES_NOT_CLOSE;
                goto cleanup;
            }

            MODL->brch[ibrch ].ichld = ichld;

        /* SKETCHs (except SKBEGs) are linked to SKBEG and SKEND */
        } else if (class == OCSM_SKETCH) {
            ibrchl = 0;
            for (jbrch = ibrch-1; jbrch > 0; jbrch--) {
                if        (MODL->brch[jbrch].type == OCSM_SKEND) {
                    status = OCSM_SKETCH_IS_OPEN;
                    goto cleanup;
                } else if (MODL->brch[jbrch].type == OCSM_SOLEND) {
                    status = OCSM_SOLVER_IS_NOT_OPEN;
                    goto cleanup;
                } else if (MODL->brch[jbrch].type == OCSM_SKBEG) {
                    ibrchl = jbrch;
                    break;
                }
            }

            if (ibrchl == 0) {
                status = OCSM_SKETCH_IS_NOT_OPEN;
                goto cleanup;
            }

            ichld = 0;
            if (MODL->brch[ibrch].type != OCSM_SKEND) {
                for (jbrch = ibrch+1; jbrch <= MODL->nbrch; jbrch++) {
                    if        (MODL->brch[jbrch].type == OCSM_SKBEG) {
                        status = OCSM_SKETCH_IS_OPEN;
                        goto cleanup;
                    } else if (MODL->brch[jbrch].type == OCSM_SOLBEG) {
                        status = OCSM_SKETCH_IS_OPEN;
                        goto cleanup;
                    } else if (MODL->brch[jbrch].type == OCSM_SKEND) {
                        ichld = jbrch;
                        break;
                    }
                }

                if (ichld == 0) {
                    status = OCSM_SKETCH_DOES_NOT_CLOSE;
                    goto cleanup;
                }
            }

            MODL->brch[ibrch ].ileft = ibrchl;
            MODL->brch[ibrch ].ichld = ichld;
            MODL->brch[ibrchl].ichld = ibrch;

        /* SOLBEGs have no parents */
        } else if (type == OCSM_SOLBEG) {
            MODL->brch[ibrch ].ichld = 0;

        /* SOLVERs (except SOLBEGs) have the SOLBEG as its left parent */
        } else if (class == OCSM_SOLVER) {
            ibrchl = 0;
            for (jbrch = ibrch-1; jbrch > 0; jbrch--) {
                if        (MODL->brch[jbrch].type == OCSM_SOLEND) {
                    status = OCSM_SOLVER_IS_OPEN;
                    goto cleanup;
                } else if (MODL->brch[jbrch].type == OCSM_SKEND) {
                    status = OCSM_SKETCH_IS_NOT_OPEN;
                    goto cleanup;
                } else if (MODL->brch[jbrch].type == OCSM_SOLBEG) {
                    ibrchl = jbrch;
                    break;
                }
            }

            if (ibrchl == 0) {
                status = OCSM_SOLVER_IS_NOT_OPEN;
                goto cleanup;
            }

            MODL->brch[ibrch ].ileft = ibrchl;
            MODL->brch[ibrchl].ichld = ibrch;

            MODL->brch[ibrch ].ichld = 0;

        /* MACBEGs have no parent or children */
        } else if (type == OCSM_MACEND) {
            MODL->brch[ibrch].ileft = 0;
            MODL->brch[ibrch].irite = 0;
            MODL->brch[ibrch].ichld = 0;

            nmacro = 1;
            for (jbrch = ibrch-1; jbrch > 0; jbrch--) {
                if        (MODL->brch[jbrch].type == OCSM_MACEND) {
                    nmacro++;
                } else if (MODL->brch[jbrch].type == OCSM_MACBEG) {
                    nmacro--;
                }
                if (MODL->brch[jbrch].ichld == 0) {
                    MODL->brch[jbrch].ichld = ibrch;
                }
                if (nmacro <= 0) {
                    MODL->brch[ibrch].ileft = jbrch;
                    MODL->brch[ibrch].irite = ibrch;
                    MODL->brch[ibrch].ichld = -1;
                    break;
                }
            }

        } else if (type == OCSM_RECALL) {
            MODL->brch[ibrch ].ileft = 0;
            MODL->brch[ibrch ].irite = 0;
            MODL->brch[ibrch ].ichld = 0;

        } else if (type == OCSM_DUMP || type == OCSM_STORE) {
            ibrchl = 0;
            for (jbrch = ibrch-1; jbrch > 0; jbrch--) {
                if (MODL->brch[jbrch].ichld == 0 &&
                    MODL->brch[jbrch].type  != OCSM_MACEND) {
                    if (ibrchl == 0) {
                        ibrchl = jbrch;
                        break;
                    }
                }
            }

            if (ibrchl == 0) {
                signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                            "DUMP and STORE expect a Body on the stack");
                status = OCSM_INSUFFICIENT_BODYS_ON_STACK;
                goto cleanup;
            }

            MODL->brch[ibrch ].ileft = ibrchl;
            MODL->brch[ibrchl].ichld = ibrch;

            MODL->brch[ibrch ].ichld = 0;

        } else if (type == OCSM_PATBEG) {
            npatn++;

        } else if (type == OCSM_PATEND) {
            npatn--;

        } else if (class == OCSM_UTILITY) { /* except OCSM_MACEND, OCSM_RECALL, OCSM_DUMP, and OCSM_STORE */
            /* no linkages */

        } else {
            status = OCSM_ILLEGAL_TYPE;
        }

        /* remember that this Branch passed the test */
        ipass = ibrch;
    }

    /* activate all Branches that are not suppressed */
    for (ibrch = 1; ibrch <= MODL->nbrch; ibrch++) {
        if (MODL->brch[ibrch].actv != OCSM_SUPPRESSED) {
            MODL->brch[ibrch].actv =  OCSM_ACTIVE;
        }
    }

    /* propagate inactivity from suppressed Branches */
    for (ibrch = 1; ibrch <= MODL->nbrch; ibrch++) {
        if (MODL->brch[ibrch].actv == OCSM_SUPPRESSED) {

            /* from a RULE, LOFT, or BLEND, inactivate left parents until next MARK */
            if (MODL->brch[ibrch].type == OCSM_RULE ||
                MODL->brch[ibrch].type == OCSM_LOFT ||
                MODL->brch[ibrch].type == OCSM_BLEND  ) {
                ileft = MODL->brch[ibrch].ileft;
                while (ileft > 0) {
                    if (MODL->brch[ileft].actv == OCSM_ACTIVE) {
                        MODL->brch[ileft].actv =  OCSM_INACTIVE;
                    }

                    if (MODL->brch[ileft].type == OCSM_MARK) {
                        break;
                    } else {
                        ileft = MODL->brch[ileft].ileft;
                    }
                }

            /* from a GROWN (except RULE, LOFT, and BLEND), inactivate left parents
               until next SKBEG */
            } else if (MODL->brch[ibrch].class == OCSM_GROWN) {
                ileft = MODL->brch[ibrch].ileft;
                while (ileft > 0) {
                    if (MODL->brch[ileft].actv == OCSM_ACTIVE) {
                        MODL->brch[ileft].actv =  OCSM_INACTIVE;
                    }

                    if (MODL->brch[ileft].type == OCSM_SKBEG) {
                        break;
                    } else {
                        ileft = MODL->brch[ileft].ileft;
                    }
                }
            }

            /* if a PRIMITIVE, GROWN, or RESTORE, inactivate children until
               next BOOLEAN */
            if (MODL->brch[ibrch].class == OCSM_PRIMITIVE ||
                MODL->brch[ibrch].class == OCSM_GROWN       ) {
                ichld = MODL->brch[ibrch].ichld;
                while (ichld > 0) {
                    if (MODL->brch[ichld].actv == OCSM_ACTIVE) {
                        MODL->brch[ichld].actv =  OCSM_INACTIVE;
                    }

                    if (MODL->brch[ichld].class == OCSM_BOOLEAN) {

                        /* inactivate any APPLIEDs associated with the BOOLEAN */
                        ichld = MODL->brch[ichld].ichld;
                        while (ichld > 0) {
                            if (MODL->brch[ichld].class == OCSM_APPLIED) {
                                if (MODL->brch[ichld].actv == OCSM_ACTIVE) {
                                    MODL->brch[ichld].actv =  OCSM_INACTIVE;
                                }
                                ichld = MODL->brch[ichld].ichld;
                            } else {
                                break;
                            }
                        }

                        break;
                    } else {
                        ichld = MODL->brch[ichld].ichld;
                    }
                }
            }
        }
    }

    /* defer all active Branches within macro definitions */
    nmacro = 0;
    for (ibrch = 1; ibrch <= MODL->nbrch; ibrch++) {
        if (MODL->brch[ibrch].actv == OCSM_ACTIVE && nmacro > 0) {
            MODL->brch[ibrch].actv = OCSM_DEFERRED;
        }
        if        (MODL->brch[ibrch].type == OCSM_MACBEG) {
            nmacro++;
        } else if (MODL->brch[ibrch].type == OCSM_MACEND) {
            nmacro--;
        }
    }

    /* MODL has passed all checks */
    MODL->checked = 1;

cleanup:
    if (MODL->checked == 1) {
        SPRINT0(1, "--> checks passed");
    } else {
        SPRINT0(1, "--> checks failed");

        /* mark the Branch after the last one that passed as the Branch with the error */
        if (ipass < MODL->nbrch) {
            MODL->brch[ipass+1].ileft = -2;
        } else {
            status = OCSM_ILLEGAL_BRCH_INDEX;
        }
    }

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmBuild - build Bodys by executing the MODL up to a given Branch *
 *                                                                      *
 ************************************************************************
 */

int
ocsmBuild(void   *modl,                 /* (in)  pointer to MODL */
          int    buildTo,               /* (in)  last Branch to execute (or 0 for all) */
          int    *builtTo,              /* (out) last Branch executed successfully */
          int    *nbody,                /* (in)  number of entries allocated in body[] */
                                        /* (out) number of Bodys on the stack */
/*@null@*/int    body[])                /* (out) array  of Bodys on the stack (LIFO)
                                                 (at least *nbody long) */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    /* stack contains Body indices, Sketch indices (if negative), or 0 for MARKs */
    int       nstack, nstack_save, *stack=NULL;

    /* Sketch contains linsegs, cirarcs, splines, and beziers */
    sket_T    *sket=NULL;

    /* storage loaction information */
    int        imacro, *macros=NULL;

    /* pattern information */
    int        npatn;
    patn_T     *patn=NULL;
    double     newValue;

    /* solver information */
    int        nsolcon;
    int        *solcons=NULL;

    int        ibrch, jbrch, type, ibrchl, i, j, k, iface, iedge, inode, nbodyMax, ncatch=0, buildStatus=SUCCESS;
    int        iattr, ipmtr, jpmtr, istor, jstor, icount, nmacro;
    int        ibody, jbody, ibodyl, irow, icol, indx, itype, nlist, npmtr_temp, total_call;
    varg_T     args[10];
    double     toler, value, dot;
    char       pmtrName[MAX_EXPR_LEN], thisArg[MAX_EXPR_LEN], temp[MAX_EXPR_LEN];

    CINT       *tempIlist;
    CDOUBLE    *tempRlist;
    CCHAR      *tempClist;

    int        imajor, iminor, iarg, ival, hasdots;
    int        newtess, ibest, npnt, ntri, itri, AT_iface, AT_xcg, AT_ycg, AT_zcg;
    CINT       *ptype, *pindx, *tris, *tric;
    double     params[3], bbox[6], size, tbest, xbest, ybest, zbest;
    double     x0, y0, z0, x1, y1, z1, x2, y2, z2, s0, s1, tt, mat[9], rhs[3], s0s1tt[3];
    CDOUBLE    *xyz, *uv;
    char       dumpfile[MAX_EXPR_LEN];
    CCHAR      *OC_ver;
    ego        ebody, ebodyl, emodel, etemp, enode, eedge, eface;
    clock_t    old_time, new_time, total_time;

    ROUTINE(ocsmBuild);
    DPRINT2("%s(buildTo=%d) {",
            routine, buildTo);

    /* macro to catch bad status and then continue to finalizing */
#define CATCH_STATUS(X)                                                 \
    if (status < SUCCESS) {                                             \
        nstack = nstack_save;                                           \
        ncatch++;                                                       \
        printf( "WARNING:: build terminated early due to BAD STATUS = %d from %s (called from %s:%d)\n", status, #X, routine, __LINE__); \
        goto finalize;                                                  \
    }

    /* --------------------------------------------------------------- */

    SPRINT1(1, "--> enter ocsmBuild(buildTo=%d)", buildTo);

    MODL->errorBuf[0] = '\0';

    MALLOC(stack,   int,    MAX_STACK_SIZE);
    MALLOC(macros,  int,    MAX_NUM_MACROS+1);
    MALLOC(patn,    patn_T, MAX_NUM_PATTERNS);
    MALLOC(sket,    sket_T, 1);
    MALLOC(solcons, int,    MAX_SOLVER_SIZE);

    *builtTo = 0;
    nbodyMax = *nbody;

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* defer all active Branches within macro definitions */
    nmacro = 0;
    for (ibrch = 1; ibrch <= MODL->nbrch; ibrch++) {
        if (MODL->brch[ibrch].actv == OCSM_ACTIVE && nmacro > 0) {
            MODL->brch[ibrch].actv = OCSM_DEFERRED;
        }
        if        (MODL->brch[ibrch].type == OCSM_MACBEG) {
            nmacro++;
        } else if (MODL->brch[ibrch].type == OCSM_MACEND) {
            nmacro--;
        }
    }

    EG_revision(&imajor, &iminor, &OC_ver);
    SPRINT3(1, "\n    EGADS version %2d.%02d (with %s)\n", imajor, iminor, OC_ver);

    /* create a new EGADS context (if not already set) */
    if (MODL->context == NULL) {
        status = EG_open(&(MODL->context));
        CHECK_STATUS(EG_open);
    }

    status = EG_setOutLevel(MODL->context, 1);
    CHECK_STATUS(EG_setOutLevel);

    /* set the global level */
    MODL->level = 0;

    /* remove internal Parameters that may be left over from a failure
          in a Sketch that was being solved */
    while (MODL->npmtr) {
        if (strcmp(MODL->pmtr[MODL->npmtr].name, "::x") == 0 ||
            strcmp(MODL->pmtr[MODL->npmtr].name, "::y") == 0 ||
            strcmp(MODL->pmtr[MODL->npmtr].name, "::z") == 0 ||
            strcmp(MODL->pmtr[MODL->npmtr].name, "::d") == 0   ) {
            FREE(MODL->pmtr[MODL->npmtr].name );
            FREE(MODL->pmtr[MODL->npmtr].value);
            FREE(MODL->pmtr[MODL->npmtr].dot  );
            FREE(MODL->pmtr[MODL->npmtr].lbnd );
            FREE(MODL->pmtr[MODL->npmtr].ubnd );
            MODL->npmtr--;
        } else {
            break;
        }
    }

    /* print warnings if any external Parameters are out of bounds */
    for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
        if (MODL->pmtr[ipmtr].type  != OCSM_EXTERNAL) continue;

        indx = 0;
        for (irow = 0; irow < MODL->pmtr[ipmtr].nrow; irow++) {
            for (icol = 0; icol < MODL->pmtr[ipmtr].ncol; icol++) {
                if (MODL->pmtr[ipmtr].value[indx] < MODL->pmtr[ipmtr].lbnd[indx]) {
                    SPRINT3(1, "WARNING:: value of %s (%f) is lower  than lower bound (%f)",
                            MODL->pmtr[ipmtr].name,
                            MODL->pmtr[ipmtr].value[indx],
                            MODL->pmtr[ipmtr].lbnd[ indx]);
                }

                if (MODL->pmtr[ipmtr].value[indx] > MODL->pmtr[ipmtr].ubnd[indx]) {
                    SPRINT3(1, "WARNING:: value of %s (%f) is higher than upper bound (%f)",
                            MODL->pmtr[ipmtr].name,
                            MODL->pmtr[ipmtr].value[indx],
                            MODL->pmtr[ipmtr].ubnd[ indx]);
                }

                indx++;
            }
        }
    }

    /* initialize the hasdots flag for all Bodys */
    for (ibody = 1; ibody <= MODL->nbody; ibody++) {
        MODL->body[ibody].hasdots = 0;
    }

    /* find the first Body that has to be regenerated, because its
       Branch is dirty */
    MODL->recycle = 0;
    for (ibody = 1; ibody <= MODL->nbody; ibody++) {
        ibrch = MODL->body[ibody].ibrch;
        if (MODL->brch[ibrch].dirty > 0) {

            /* free up all Bodys starting at ibody */
            for (jbody = ibody; jbody <= MODL->nbody; jbody++) {
                for (iarg = 1; iarg < 10; iarg++) {
                    FREE(MODL->body[jbody].arg[iarg].val);
                    FREE(MODL->body[jbody].arg[iarg].dot);
                }

                FREE(MODL->body[jbody].node);
                FREE(MODL->body[jbody].edge);
                FREE(MODL->body[jbody].face);

                if (MODL->body[jbody].etess != NULL) {
                    status = EG_deleteObject(MODL->body[jbody].etess);
                    CHECK_STATUS(EG_deleteObject);

                    MODL->body[jbody].etess = NULL;
                }

                if (MODL->body[jbody].ebody != NULL) {
                    status = EG_deleteObject(MODL->body[jbody].ebody);
                    if (status == EGADS_EMPTY) status = SUCCESS;
                    CHECK_STATUS(EG_deleteObject);

                    MODL->body[jbody].ebody = NULL;
                }
            }

            /* remove any .ichld pointers to any Bodys at
               ibody or beyond */
            for (jbody = 1; jbody <= MODL->nbody; jbody++) {
                if (MODL->body[jbody].ichld >= ibody) {
                    MODL->body[jbody].ichld = 0;
                }
            }

            break;
        }
        MODL->recycle = ibody;
    }
    MODL->nbody = 0;

    /* initialize the stack */
    nstack = 0;

    /* initialize the number of active Sketch points (and constraints) */
    sket->type   = 0;
    sket->size   = 0;
    sket->solved = 1;
    sket->nseg   = 0;
    sket->nvar   = 0;
    sket->ncon   = 0;

    /* initiailze the number of nested patterns */
    npatn = 0;

    /* initialize the number of solver constraints */
    nsolcon = 0;

    /* initialize the storage locations */
    for (imacro = 1; imacro <= MAX_NUM_MACROS; imacro++) {
        macros[imacro] = -1;
    }

    /* initialize the arguments */
    for (iarg = 1; iarg < 10; iarg++) {
        args[iarg].nval = -1;
        args[iarg].nrow =  0;
        args[iarg].ncol =  0;
        args[iarg].val  = NULL;
        args[iarg].dot  = NULL;
    }

    /* if buildTo was set to zero, reset it (locally) to be greater
       then MODL->nbrch */
    if (buildTo == 0) {
        buildTo = MODL->nbrch + 1;
    }

    /* loop through and process all the Branches (up to buildTo) */
    for (ibrch = 1; ibrch <= MODL->nbrch; ibrch++) {
        nstack_save = nstack;

        type = MODL->brch[ibrch].type;

        /* exit if we exceeded buildTo */
        if (ibrch > buildTo) {
            break;
        }

        /* if Branch is deferred, mark it as active but skip it for now */
        if (MODL->brch[ibrch].actv == OCSM_DEFERRED) {
            MODL->brch[ibrch].actv =  OCSM_ACTIVE;
            SPRINT1(1, "    deferring [%4d]:", ibrch);
            continue;
        }

        /* skip this Branch if it is not active */
        if (MODL->brch[ibrch].actv != OCSM_ACTIVE) {
            SPRINT1(1, "    skipping  [%4d]:", ibrch);
            continue;
        }

        /* if this is an executable Sketch statement and we need
           to solve the Sketch, do it before interpreting the
           arguments and velocities */
        if (type == OCSM_LINSEG || type == OCSM_CIRARC ||
            type == OCSM_ARC    || type == OCSM_SPLINE ||
            type == OCSM_BEZIER                          ) {
            status = solveSketch(MODL, sket);

            /* if the solver did not work, remove the Body associated with
               skbeg */
            if (status < SUCCESS) {
                (MODL->nbody)--;
            }

            CATCH_STATUS(solveSketch);
        }

        /* get the values for the arguments and velocities */
        for (iarg = 1; iarg < 10; iarg++) {
            args[iarg].nval = -1;
            args[iarg].nrow =  0;
            args[iarg].ncol =  0;

            if (iarg > MODL->brch[ibrch].narg) continue;

                                  thisArg[0] = '\0';
            if (iarg == 1) strcpy(thisArg, MODL->brch[ibrch].arg1);
            if (iarg == 2) strcpy(thisArg, MODL->brch[ibrch].arg2);
            if (iarg == 3) strcpy(thisArg, MODL->brch[ibrch].arg3);
            if (iarg == 4) strcpy(thisArg, MODL->brch[ibrch].arg4);
            if (iarg == 5) strcpy(thisArg, MODL->brch[ibrch].arg5);
            if (iarg == 6) strcpy(thisArg, MODL->brch[ibrch].arg6);
            if (iarg == 7) strcpy(thisArg, MODL->brch[ibrch].arg7);
            if (iarg == 8) strcpy(thisArg, MODL->brch[ibrch].arg8);
            if (iarg == 9) strcpy(thisArg, MODL->brch[ibrch].arg9);

            /* argument is a string (which starts with $) */
            if (thisArg[0] == '$') {
                args[iarg].nval = 0;
                MALLOC(args[iarg].str, char, (int)strlen(thisArg));

                strcpy(args[iarg].str, &(thisArg[1]));
            } else {
                jpmtr = 0;

                for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
                    if (strcmp(MODL->pmtr[ipmtr].name, thisArg) == 0 &&
                               MODL->pmtr[ipmtr].level == MODL->level  ) {
                        jpmtr = ipmtr;
                        break;
                    }
                }

                /* argument name matches Parameter name */
                if (jpmtr > 0) {
                    args[iarg].nval = (MODL->pmtr[jpmtr].nrow) * (MODL->pmtr[jpmtr].ncol);
                    args[iarg].nrow =  MODL->pmtr[jpmtr].nrow;
                    args[iarg].ncol =  MODL->pmtr[jpmtr].ncol;
                    MALLOC(args[iarg].val, double, args[iarg].nval);
                    MALLOC(args[iarg].dot, double, args[iarg].nval);

                    for (ival = 0; ival < args[iarg].nval; ival++) {
                        args[iarg].val[ival] = MODL->pmtr[jpmtr].value[ival];
                        args[iarg].dot[ival] = MODL->pmtr[jpmtr].dot[  ival];
                    }

                /* argument contains a semicolon-separated list */
                } else if (strstr(thisArg, ";") != NULL) {
                    /* make sure it ends in a semicolon */
                    i = strlen(thisArg) - 1;
                    if (thisArg[i] != ';') {
                        thisArg[i+1] = ';';
                        thisArg[i+2] = '\0';
                    }

                    /* count the number of semicolons and allocate storage */
                    args[iarg].nval = 0;
                    for (i = 0; i < strlen(thisArg); i++) {
                        if (thisArg[i] == ';') {
                            (args[iarg].nval)++;
                        }
                    }
                    args[iarg].nrow = args[iarg].nval;
                    args[iarg].ncol = 1;
                    MALLOC(args[iarg].val, double, args[iarg].nval);
                    MALLOC(args[iarg].dot, double, args[iarg].nval);

                    /* put the semi-colon-separated arguments into the val/dot */
                    i = 0;
                    for (k = 0; k < args[iarg].nval; k++) {
                        j = 0;
                        temp[j] = '\0';
                        while (thisArg[i] != ';') {
                            temp[j  ] = thisArg[i];
                            temp[j+1] = '\0';
                            j++;
                            i++;
                        }

                        if (strlen(temp) > 0) {
                            status = str2val(temp, MODL, &(args[iarg].val[k]), &(args[iarg].dot[k]));
                            CATCH_STATUS(str2val);
                        } else {
                            SET_STATUS(OCSM_ILLEGAL_VALUE, parser);
                        }
                        i++;
                    }

                /* argument contains an expression */
                } else {
                    args[iarg].nval = 1;
                    args[iarg].nrow = 1;
                    args[iarg].ncol = 1;
                    MALLOC(args[iarg].val, double, args[iarg].nval);
                    MALLOC(args[iarg].dot, double, args[iarg].nval);

                    status = str2val(thisArg, MODL, &(args[iarg].val[0]), &(args[iarg].dot[0]));
                    CATCH_STATUS(str2val);
                }
            }
        }

        /* make sure that there is enough room on the stack */
        if (nstack >= MAX_STACK_SIZE-1) {
            status = OCSM_TOO_MANY_BODYS_ON_STACK;
            goto cleanup;
        }

        /* execute Branch ibrch */
        old_time = clock();

        if        (MODL->brch[ibrch].class == OCSM_PRIMITIVE) {
            status = buildPrimitive(MODL, ibrch, args, &nstack, stack, npatn, patn);
            CATCH_STATUS(buildPrimitive);

            status = setupAtPmtrs(MODL);
            CATCH_STATUS(setupAtPmtrs);

        } else if (MODL->brch[ibrch].class == OCSM_GROWN) {
            status = buildGrown(MODL, ibrch, args, &nstack, stack, npatn, patn);
            CATCH_STATUS(buildGrown);

            status = setupAtPmtrs(MODL);
            CATCH_STATUS(setupAtPmtrs);

        } else if (MODL->brch[ibrch].class == OCSM_APPLIED) {
            status = buildApplied(MODL, ibrch, args, &nstack, stack, npatn, patn);
            CATCH_STATUS(buildApplied);

            status = setupAtPmtrs(MODL);
            CATCH_STATUS(setupAtPmtrs);

        } else if (MODL->brch[ibrch].class == OCSM_BOOLEAN) {
            status = buildBoolean(MODL, ibrch, args, &nstack, stack, npatn, patn);
            CATCH_STATUS(buildBoolean);

            status = setupAtPmtrs(MODL);
            CATCH_STATUS(setupAtPmtrs);

        } else if (MODL->brch[ibrch].class == OCSM_TRANSFORM) {
            status = buildTransform(MODL, ibrch, args, &nstack, stack);
            CATCH_STATUS(buildTransform);

            status = setupAtPmtrs(MODL);
            CATCH_STATUS(setupAtPmtrs);

        } else if (MODL->brch[ibrch].class == OCSM_SKETCH) {
            status = buildSketch(MODL, ibrch, args, &nstack, stack, npatn, patn, sket);
            CATCH_STATUS(buildSketch);

        } else if (MODL->brch[ibrch].class == OCSM_SOLVER) {
            status = buildSolver(MODL, ibrch, args, &nsolcon, solcons);
            CATCH_STATUS(buildSolver);

        /* execute: "set $pmtrName exprs" */
        } else if (type == OCSM_SET) {
            SPRINT3(1, "    executing [%4d] set:            %s  %s",
                    ibrch, args[1].str, MODL->brch[ibrch].arg2);

            /* single-set mode */
            if (args[2].nval == 1) {

                /* break arg1 into name[irow,icol] */
                status = parseName(MODL, args[1].str, &jpmtr, &irow, &icol);
                CATCH_STATUS(parseName);

                /* if the Parameter is EXTERNAL, we have a problem */
                if (MODL->pmtr[jpmtr].type == OCSM_EXTERNAL) {
                    status = OCSM_PMTR_IS_EXTERNAL;
                    CATCH_STATUS(set);
                }

                /* store the value (and its velocity) */
                value = args[2].val[0];
                dot   = args[2].dot[0];

                status = ocsmSetValuD(MODL, jpmtr, irow, icol, value);
                CATCH_STATUS(ocsmSetValu);

                status = ocsmSetVelD(MODL, jpmtr, irow, icol, dot);
                CATCH_STATUS(ocsmSetVelD);

                SPRINT3(1, "                          %s = %11.5f %11.5f",
                        args[1].str, value, dot);

            /* multi-set mode */
            } else {

                /* look for current Parameter */
                jpmtr = 0;
                for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
                    if (strcmp(MODL->pmtr[ipmtr].name, args[1].str) == 0 &&
                               MODL->pmtr[ipmtr].level == MODL->level      ) {
                        jpmtr = ipmtr;
                        break;
                    }
                }

                /* Parameter must have been defined (in a dimension statement) */
                if (jpmtr <= 0) {
                    status = OCSM_NAME_NOT_FOUND;
                    CATCH_STATUS(set);
                } else if (MODL->pmtr[ipmtr].type != OCSM_INTERNAL) {
                    status = OCSM_PMTR_IS_EXTERNAL;
                    CATCH_STATUS(set);
                }

                /* store the values */
                icount = 0;
                value  = 0;
                dot    = 0;
                for (irow = 1; irow <= MODL->pmtr[jpmtr].nrow; irow++) {
                    for (icol = 1; icol <= MODL->pmtr[jpmtr].ncol; icol++) {

                        /* next value (if it exists; otherwise it is extended) */
                        if (icount < args[2].nval) {
                            value = args[2].val[icount];
                            dot   = args[2].dot[icount];
                        }

                        status = ocsmSetValuD(MODL, jpmtr, irow, icol, value);
                        CATCH_STATUS(ocsmSetValuD);

                        status = ocsmSetVelD(MODL, jpmtr, irow, icol, dot);
                        CATCH_STATUS(ocsmSetVelD);

                        if (icount < args[2].nval) {
                            SPRINT5(1, "                          %s[%d,%d] = %11.5f %11.5f",
                                    args[1].str, irow, icol, value, dot);
                        } else {
                            SPRINT5(1, "                          %s[%d,%d] = %11.5f %11.5f (extended)",
                                    args[1].str, irow, icol, value, dot);
                        }

                        icount++;
                    }
                }
            }

        /* execute: "udparg $primtype $argName1 argValue1 $argName2 argValue2 $argName3 argValue3 $argName4 argValue4" */
        } else if (type == OCSM_UDPARG) {
            SPRINT2x(1, "    executing [%4d] udparg:     %s",
                    ibrch, args[1].str);

            hasdots = 0;
            for (iarg = 3; iarg <= MODL->brch[ibrch].narg; iarg+=2) {
                if (args[iarg].nval == 0) {
                    SPRINT2x(1, " %s %s",
                             args[iarg-1].str, args[iarg].str);
                } else {
                    SPRINT2x(1, " %s %11.5f",
                             args[iarg-1].str, args[iarg].val[0]);

                    for (ival = 0; ival < args[iarg].nval; ival++) {
                        if (args[iarg].val[ival] != 0) hasdots = 1;
                    }
                }
            }
            SPRINT0(1, " ");

            status = newBody(MODL, ibrch, OCSM_UDPARG, -1, -1,
                             args, hasdots, OCSM_NULL_BODY, &ibody);
            CATCH_STATUS(newBody);

            /* finish the Body */
            status = finishBody(MODL, ibody);
            CATCH_STATUS(finishBody);

            /* do not push the Body onto the stack */

            SPRINT1(1, "                          Body   %4d created  (NULL Body)",
                ibody);

        /* execute: "select $type arg1..." */
        } else if (type == OCSM_SELECT) {
            if        (MODL->brch[ibrch].narg == 1) {
                SPRINT2(1, "    executing [%4d] select:     %s",
                        ibrch, args[1].str);
            } else if (MODL->brch[ibrch].narg == 2) {
                SPRINT3(1, "    executing [%4d] select:     %s  %11.5f",
                        ibrch, args[1].str,
                        args[2].val[0]);
            } else if (MODL->brch[ibrch].narg == 3) {
                SPRINT4(1, "    executing [%4d] select:     %s  %11.5f %11.5f",
                        ibrch, args[1].str,
                        args[2].val[0], args[3].val[0]);
            } else if (MODL->brch[ibrch].narg == 4) {
                SPRINT5(1, "    executing [%4d] select:     %s  %11.5f %11.5f %11.5f",
                        ibrch, args[1].str,
                        args[2].val[0], args[3].val[0], args[4].val[0]);
            } else if (MODL->brch[ibrch].narg == 5) {
                SPRINT6(1, "    executing [%4d] select:     %s  %11.5f %11.5f %11.5f %11.5f",
                        ibrch, args[1].str,
                        args[2].val[0], args[3].val[0], args[4].val[0], args[5].val[0]);
            } else if (MODL->brch[ibrch].narg == 6) {
                SPRINT7(1, "    executing [%4d] select:     %s  %11.5f %11.5f %11.5f %11.5f %11.5f",
                        ibrch, args[1].str,
                        args[2].val[0], args[3].val[0], args[4].val[0], args[5].val[0], args[6].val[0]);
            } else if (MODL->brch[ibrch].narg == 7) {
                SPRINT8(1, "    executing [%4d] select:     %s  %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f",
                        ibrch, args[1].str,
                        args[2].val[0], args[3].val[0], args[4].val[0], args[5].val[0], args[6].val[0], args[7].val[0]);
            } else if (MODL->brch[ibrch].narg == 8) {
                SPRINT9(1, "    executing [%4d] select:     %s  %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f",
                        ibrch, args[1].str,
                        args[2].val[0], args[3].val[0], args[4].val[0], args[5].val[0], args[6].val[0], args[7].val[0], args[8].val[0]);
            }

            ibodyl = stack[nstack-1];

            /* "select body" */
            if        (strcmp(args[1].str, "body") == 0) {
                if        (MODL->brch[ibrch].narg == 1) {    /* "body" */
                    MODL->sel_body = MODL->nbody;
                    MODL->sel_face = -1;
                    MODL->sel_edge = -1;
                    MODL->sel_node = -1;
                } else if (MODL->brch[ibrch].narg == 2) {    /* "body ibody" */
                    if (NINT(args[2].val[0]) >= 1 && NINT(args[2].val[0]) <= MODL->nbody) {
                        MODL->sel_body = NINT(args[2].val[0]);
                        MODL->sel_face = -1;
                        MODL->sel_edge = -1;
                        MODL->sel_node = -1;
                    } else {
                        status = OCSM_BODY_NOT_FOUND;
                        CATCH_STATUS(select_body);
                    }
                }

                /* put attributes for OCSM_BRANCH on the Body */
                ebody = MODL->body[MODL->sel_body].ebody;

                for (iattr = 0; iattr < MODL->brch[ibrch].nattr; iattr++) {
                    if (MODL->brch[ibrch].attr[iattr].value[0] == '$') {
                        status = EG_attributeAdd(ebody, MODL->brch[ibrch].attr[iattr].name,
                                                 ATTRSTRING, 1,
                                                 NULL, NULL, &(MODL->brch[ibrch].attr[iattr].value[1]));
                        CATCH_STATUS(EG_attributeAdd);
                    } else {
                        ipmtr = -1;
                        for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
                            if (strcmp(MODL->brch[ibrch].attr[iattr].value, MODL->pmtr[jpmtr].name) == 0) {
                                ipmtr = jpmtr;
                                break;
                            }
                        }

                        if (ipmtr > 0) {
                            status = EG_attributeAdd(ebody, MODL->brch[ibrch].attr[iattr].name,
                                                     ATTRREAL, (MODL->pmtr[ipmtr].nrow)*(MODL->pmtr[ipmtr].ncol),
                                                     NULL, MODL->pmtr[ipmtr].value, NULL);
                            CATCH_STATUS(EG_attributeAdd);
                        } else {
                            str2val(MODL->brch[ibrch].attr[iattr].value, MODL, &value, &dot);
                            CATCH_STATUS(str2val);

                            status = EG_attributeAdd(ebody, MODL->brch[ibrch].attr[iattr].name,
                                                     ATTRREAL, 1,
                                                     NULL, &value, NULL);
                            CATCH_STATUS(EG_attributeAdd);
                        }
                    }
                }

            /* "select face" */
            } else if (strcmp(args[1].str, "face") == 0) {
                if (MODL->sel_body < 0) {
                    status = OCSM_BODY_NOT_FOUND;
                    CATCH_STATUS(select_face);
                } else if (MODL->brch[ibrch].narg == 2) {    /* "face iface" */
                    if (NINT(args[2].val[0]) >= 1 && NINT(args[2].val[0]) <= MODL->body[MODL->sel_body].nface) {
                        MODL->sel_face = NINT(args[2].val[0]);
                        MODL->sel_edge = -1;
                        MODL->sel_node = -1;
                    } else {
                        status = OCSM_FACE_NOT_FOUND;
                        CATCH_STATUS(select_face);
                    }

                } else if (MODL->brch[ibrch].narg >= 3) {    /* "face ibody1 iford1 iseq=1" */
                    MODL->sel_face = -1;
                    MODL->sel_edge = -1;
                    MODL->sel_node = -1;

                    if (MODL->brch[ibrch].narg == 3) {
                        args[4].val[0] = 1;
                    }

                    for (iface = 1; iface <= MODL->body[MODL->sel_body].nface; iface++) {
                        status = EG_attributeRet(MODL->body[MODL->sel_body].face[iface].eface,
                                                 "faceID", &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CATCH_STATUS(EG_attributeRet);

                        if (NINT(args[2].val[0]) == tempIlist[0] &&
                            NINT(args[3].val[0]) == tempIlist[1] &&
                            NINT(args[4].val[0]) == tempIlist[2]   ) {
                            MODL->sel_face = iface;
                            break;
                        }
                    }

                    if (MODL->sel_face < 0) {
                        status = OCSM_FACE_NOT_FOUND;
                        CATCH_STATUS(select_face);
                    }
                } else {
                    status = OCSM_NOT_ENOUGH_ARGS;
                    CATCH_STATUS(select_face);
                }

                /* put attributes for OCSM_BRANCH on the Face */
                eface = MODL->body[MODL->sel_body].face[MODL->sel_face].eface;

                for (iattr = 0; iattr < MODL->brch[ibrch].nattr; iattr++) {
                    if (MODL->brch[ibrch].attr[iattr].value[0] == '$') {
                        status = EG_attributeAdd(eface, MODL->brch[ibrch].attr[iattr].name,
                                                 ATTRSTRING, 1,
                                                 NULL, NULL, &(MODL->brch[ibrch].attr[iattr].value[1]));
                        CATCH_STATUS(EG_attributeAdd);
                    } else {
                        ipmtr = -1;
                        for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
                            if (strcmp(MODL->brch[ibrch].attr[iattr].value, MODL->pmtr[jpmtr].name) == 0) {
                                ipmtr = jpmtr;
                                break;
                            }
                        }

                        if (ipmtr > 0) {
                            status = EG_attributeAdd(eface, MODL->brch[ibrch].attr[iattr].name,
                                                     ATTRREAL, (MODL->pmtr[ipmtr].nrow)*(MODL->pmtr[ipmtr].ncol),
                                                     NULL, MODL->pmtr[ipmtr].value, NULL);
                            CATCH_STATUS(EG_attributeAdd);
                        } else {
                            str2val(MODL->brch[ibrch].attr[iattr].value, MODL, &value, &dot);
                            CATCH_STATUS(str2val);

                            status = EG_attributeAdd(eface, MODL->brch[ibrch].attr[iattr].name,
                                                     ATTRREAL, 1,
                                                     NULL, &value, NULL);
                            CATCH_STATUS(EG_attributeAdd);
                        }
                    }
                }

            /* "select edge" */
            } else if (strcmp(args[1].str, "edge") == 0) {
                if (MODL->sel_body < 0) {
                    status = OCSM_BODY_NOT_FOUND;
                    CATCH_STATUS(select_edge);
                } else if (MODL->brch[ibrch].narg == 2) {    /* "edge iedge" */
                    if (NINT(args[2].val[0]) >= 1 && NINT(args[2].val[0]) <= MODL->body[MODL->sel_body].nedge) {
                        MODL->sel_face = -1;
                        MODL->sel_edge = NINT(args[2].val[0]);
                        MODL->sel_node = -1;
                    } else {
                        status = OCSM_EDGE_NOT_FOUND;
                        CATCH_STATUS(select_edge);
                    }
                } else if (MODL->brch[ibrch].narg >= 5) {    /* "edge ibody1 iford1 ibody2 iford2 iseq=1" */
                    MODL->sel_face = -1;
                    MODL->sel_edge = -1;
                    MODL->sel_node = -1;

                    if (MODL->brch[ibrch].narg == 5) {
                        args[6].val[0] = 1;
                    }

                    for (iedge = 1; iedge <= MODL->body[MODL->sel_body].nedge; iedge++) {
                        status = EG_attributeRet(MODL->body[MODL->sel_body].edge[iedge].eedge,
                                                 "edgeID", &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CATCH_STATUS(EG_attributeRet);

                        if        (NINT(args[2].val[0]) == tempIlist[0] &&
                                   NINT(args[3].val[0]) == tempIlist[1] &&
                                   NINT(args[4].val[0]) == tempIlist[2] &&
                                   NINT(args[5].val[0]) == tempIlist[3] &&
                                   NINT(args[6].val[0]) == tempIlist[4]   ) {
                            MODL->sel_edge = iedge;
                            break;
                        } else if (NINT(args[2].val[0]) == tempIlist[2] &&
                                   NINT(args[3].val[0]) == tempIlist[3] &&
                                   NINT(args[4].val[0]) == tempIlist[0] &&
                                   NINT(args[5].val[0]) == tempIlist[1] &&
                                   NINT(args[6].val[0]) == tempIlist[4]   ) {
                            MODL->sel_edge = iedge;
                            break;
                        }
                    }

                    if (MODL->sel_edge < 0) {
                        status = OCSM_EDGE_NOT_FOUND;
                        CATCH_STATUS(select_edge);
                    }
                } else {
                    status = OCSM_NOT_ENOUGH_ARGS;
                    CATCH_STATUS(select_edge);
                }

                /* put attributes for OCSM_BRANCH on the Edge */
                eedge = MODL->body[MODL->sel_body].edge[MODL->sel_edge].eedge;

                for (iattr = 0; iattr < MODL->brch[ibrch].nattr; iattr++) {
                    if (MODL->brch[ibrch].attr[iattr].value[0] == '$') {
                        status = EG_attributeAdd(eedge, MODL->brch[ibrch].attr[iattr].name,
                                                 ATTRSTRING, 1,
                                                 NULL, NULL, &(MODL->brch[ibrch].attr[iattr].value[1]));
                        CATCH_STATUS(EG_attributeAdd);
                    } else {
                        ipmtr = -1;
                        for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
                            if (strcmp(MODL->brch[ibrch].attr[iattr].value, MODL->pmtr[jpmtr].name) == 0) {
                                ipmtr = jpmtr;
                                break;
                            }
                        }

                        if (ipmtr > 0) {
                            status = EG_attributeAdd(eedge, MODL->brch[ibrch].attr[iattr].name,
                                                     ATTRREAL, (MODL->pmtr[ipmtr].nrow)*(MODL->pmtr[ipmtr].ncol),
                                                     NULL, MODL->pmtr[ipmtr].value, NULL);
                            CATCH_STATUS(EG_attributeAdd);
                        } else {
                            str2val(MODL->brch[ibrch].attr[iattr].value, MODL, &value, &dot);
                            CATCH_STATUS(str2val);

                            status = EG_attributeAdd(eedge, MODL->brch[ibrch].attr[iattr].name,
                                                     ATTRREAL, 1,
                                                     NULL, &value, NULL);
                            CATCH_STATUS(EG_attributeAdd);
                        }
                    }
                }

            /* "select node" */
            } else if (strcmp(args[1].str, "node") == 0) {
                if (MODL->sel_body < 0) {
                    status = OCSM_BODY_NOT_FOUND;
                    CATCH_STATUS(select_node);
                } else if (MODL->brch[ibrch].narg == 2) {    /* "node inode" */
                    if (NINT(args[2].val[0]) >= 1 && NINT(args[2].val[0]) <= MODL->body[MODL->sel_body].nnode) {
                        MODL->sel_face = -1;
                        MODL->sel_edge = -1;
                        MODL->sel_node = NINT(args[2].val[0]);
                    } else {
                        status = OCSM_NODE_NOT_FOUND;
                        CATCH_STATUS(select_node);
                    }
                } else if (MODL->brch[ibrch].narg >= 7) {    /* "node ibody1 iford1 ibody2 iford2 ibody3 iford3 iseq=1" */
                    MODL->sel_face = -1;
                    MODL->sel_edge = -1;
                    MODL->sel_node = -1;

                    if (MODL->brch[ibrch].narg == 7) {
                        args[8].val[0] = 1;
                    }

                    for (inode = 1; inode <= MODL->body[MODL->sel_body].nnode; inode++) {
                        status = EG_attributeRet(MODL->body[MODL->sel_body].node[inode].enode,
                                                 "nodeID", &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CATCH_STATUS(EG_attributeRet);

                        if (NINT(args[2].val[0]) == tempIlist[0] &&
                            NINT(args[3].val[0]) == tempIlist[1] &&
                            NINT(args[4].val[0]) == tempIlist[2] &&
                            NINT(args[5].val[0]) == tempIlist[3] &&
                            NINT(args[6].val[0]) == tempIlist[4] &&
                            NINT(args[7].val[0]) == tempIlist[5] &&
                            NINT(args[8].val[0]) == tempIlist[6]   ) {
                            MODL->sel_node = inode;
                            break;
                        }
                    }

                    if (MODL->sel_node < 0) {
                        status = OCSM_NODE_NOT_FOUND;
                        CATCH_STATUS(select_node);
                    }
                } else {
                    status = OCSM_NOT_ENOUGH_ARGS;
                    CATCH_STATUS(select_node);
                }

                /* put attributes for OCSM_BRANCH on the Node */
                enode = MODL->body[MODL->sel_body].node[MODL->sel_node].enode;

                for (iattr = 0; iattr < MODL->brch[ibrch].nattr; iattr++) {
                    if (MODL->brch[ibrch].attr[iattr].value[0] == '$') {
                        status = EG_attributeAdd(enode, MODL->brch[ibrch].attr[iattr].name,
                                                 ATTRSTRING, 1,
                                                 NULL, NULL, &(MODL->brch[ibrch].attr[iattr].value[1]));
                        CATCH_STATUS(EG_attributeAdd);
                    } else {
                        ipmtr = -1;
                        for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
                            if (strcmp(MODL->brch[ibrch].attr[iattr].value, MODL->pmtr[jpmtr].name) == 0) {
                                ipmtr = jpmtr;
                                break;
                            }
                        }

                        if (ipmtr > 0) {
                            status = EG_attributeAdd(enode, MODL->brch[ibrch].attr[iattr].name,
                                                     ATTRREAL, (MODL->pmtr[ipmtr].nrow)*(MODL->pmtr[ipmtr].ncol),
                                                     NULL, MODL->pmtr[ipmtr].value, NULL);
                            CATCH_STATUS(EG_attributeAdd);
                        } else {
                            str2val(MODL->brch[ibrch].attr[iattr].value, MODL, &value, &dot);
                            CATCH_STATUS(str2val);

                            status = EG_attributeAdd(enode, MODL->brch[ibrch].attr[iattr].name,
                                                     ATTRREAL, 1,
                                                     NULL, &value, NULL);
                            CATCH_STATUS(EG_attributeAdd);
                        }
                    }
                }

            } else {
                status = OCSM_ILLEGAL_ARGUMENT;
                CATCH_STATUS(select);
            }

            status = setupAtPmtrs(MODL);
            CATCH_STATUS(setupAtPmtrs);

        /* execute: "project x y z dx dy dz" */
        } else if (type == OCSM_PROJECT) {
            SPRINT7(1, "    executing [%4d] project:    %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f",
                    ibrch, args[1].val[0], args[2].val[0], args[3].val[0], args[4].val[0],
                           args[5].val[0], args[6].val[0]);

            /* get the Bodys on the top of the stack */
            if (nstack < 1) {
                signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                            "PROJECT expects a Body on stack");
                status = OCSM_INSUFFICIENT_BODYS_ON_STACK;
                CATCH_STATUS(project);
            } else {
                ibodyl = stack[nstack-1];
            }

            /* make sure that the current Body has a tessellation */
            newtess = 0;
            if (MODL->body[ibodyl].etess == NULL) {
                status = EG_getBoundingBox(MODL->body[ibodyl].ebody, bbox);
                CATCH_STATUS(EG_getBoundingBox);

                size = sqrt(SQR(bbox[3]-bbox[0]) + SQR(bbox[4]-bbox[1]) + SQR(bbox[5]-bbox[2]));

                params[0] = 0.0250 * size;
                params[1] = 0.0010 * size;
                params[2] = 15.0;

                (void) EG_setOutLevel(MODL->context, 1);
                status = EG_makeTessBody(MODL->body[ibodyl].ebody, params,
                                         &(MODL->body[ibodyl].etess));
                (void) EG_setOutLevel(MODL->context, outLevel);
                CATCH_STATUS(EG_makeTessBody);

                newtess = 1;
            }

            /* loop through all Triangles in all Faces to find closest projection */
            ibest = 0;
            tbest = HUGEQ;
            xbest = HUGEQ;
            ybest = HUGEQ;
            zbest = HUGEQ;

            for (iface = 1; iface <= MODL->body[ibodyl].nface; iface++) {
                status = EG_getTessFace(MODL->body[ibodyl].etess, iface,
                                        &npnt, &xyz, &uv, &ptype, &pindx,
                                        &ntri, &tris, &tric);
                CATCH_STATUS(EG_getFaceTess);

                for (itri = 0; itri < ntri; itri++) {
                    x0 = xyz[3*tris[3*itri  ]-3];
                    y0 = xyz[3*tris[3*itri  ]-2];
                    z0 = xyz[3*tris[3*itri  ]-1];
                    x1 = xyz[3*tris[3*itri+1]-3];
                    y1 = xyz[3*tris[3*itri+1]-2];
                    z1 = xyz[3*tris[3*itri+1]-1];
                    x2 = xyz[3*tris[3*itri+2]-3];
                    y2 = xyz[3*tris[3*itri+2]-2];
                    z2 = xyz[3*tris[3*itri+2]-1];

                    mat[0] = x0 - x2;  mat[1] = x1 - x2;  mat[2] = -args[4].val[0];  rhs[0] = args[1].val[0] - x2;
                    mat[3] = y0 - y2;  mat[4] = y1 - y2;  mat[5] = -args[5].val[0];  rhs[1] = args[2].val[0] - y2;
                    mat[6] = z0 - z2;  mat[7] = z1 - z2;  mat[8] = -args[6].val[0];  rhs[2] = args[3].val[0] - z2;

                    status = matsol(mat, rhs, 3, s0s1tt);
                    if (status != SUCCESS) continue;

                    s0 = s0s1tt[0];
                    s1 = s0s1tt[1];
                    tt = s0s1tt[2];

                    if (tt < 0 || s0 < 0 || s1 < 0 || (1-s0-s1) < 0) continue;

                    if (tt < tbest) {
                        ibest = iface;
                        tbest = tt;
                        xbest = args[1].val[0] + tt * args[4].val[0];
                        ybest = args[2].val[0] + tt * args[5].val[0];
                        zbest = args[3].val[0] + tt * args[6].val[0];
                    }
                }
            }

            /* if a new tessellation was made above, delete it now */
            if (newtess == 1) {
                status = EG_deleteObject(MODL->body[ibodyl].etess);
                CATCH_STATUS(EG_deleteObject);

                MODL->body[ibodyl].etess = NULL;
            }

            /* if any hits, store the results of the best in the @-parameters */
            if (ibest > 0) {
                status = ocsmFindPmtr(MODL, "@iface", OCSM_INTERNAL, 1, 1, &AT_iface);
                CATCH_STATUS(ocsmFindPmtr);

                status = ocsmSetValuD(MODL, AT_iface, 1, 1, (double)ibest);
                CATCH_STATUS(ocsmSetValuD);

                status = ocsmFindPmtr(MODL, "@xcg", OCSM_INTERNAL, 1, 1, &AT_xcg);
                CATCH_STATUS(ocsmFindPmtr);

                status = ocsmSetValuD(MODL, AT_xcg, 1, 1, xbest);
                CATCH_STATUS(ocsmSetValuD);

                status = ocsmFindPmtr(MODL, "@ycg", OCSM_INTERNAL, 1, 1, &AT_ycg);
                CATCH_STATUS(ocsmFindPmtr);

                status = ocsmSetValuD(MODL, AT_ycg, 1, 1, ybest);
                CATCH_STATUS(ocsmSetValuD);

                status = ocsmFindPmtr(MODL, "@zcg", OCSM_INTERNAL, 1, 1, &AT_zcg);
                CATCH_STATUS(ocsmFindPmtr);

                status = ocsmSetValuD(MODL, AT_zcg, 1, 1, zbest);
                CATCH_STATUS(ocsmSetValuD);

            /* otherwise, return an error */
            } else {
                status = OCSM_FACE_NOT_FOUND;
                CATCH_STATUS(project);
            }

        /* execute: "macbeg imacro" */
        } else if (type == OCSM_MACBEG) {
            SPRINT2(1, "    executing [%4d] macbeg:     %11.5f",
                    ibrch, args[1].val[0]);

            /* determine the storage location */
            imacro = NINT(args[1].val[0]);

            if (imacro < 1 || imacro > MAX_NUM_MACROS) {
                status = OCSM_ILLEGAL_MACRO_INDEX;
                CATCH_STATUS(macbeg);
            }

            /* an error if storage location is not empty */
            if (macros[imacro] > 0) {
                status = OCSM_STORAGE_ALREADY_USED;
                CATCH_STATUS(macbeg);
            }


            macros[imacro] = ibrch;

            SPRINT2(1, "                          Storing Branch %4d in storage %d", ibrch, imacro);

        /* execute: "macend" */
        } else if (type == OCSM_MACEND) {
            SPRINT1(1, "    executing [%4d] macend:",
                    ibrch);

            npatn--;
            ibrch = patn[npatn].ipatend;

        /* execute: "recall imacro" */
        } else if (type == OCSM_RECALL) {
            SPRINT2(1, "    executing [%4d] recall:     %11.5f",
                    ibrch, args[1].val[0]);

            /* determine the storage location */
            imacro = NINT(args[1].val[0]);

            if (imacro < 1 || imacro > MAX_NUM_MACROS) {
                status = OCSM_ILLEGAL_MACRO_INDEX;
                CATCH_STATUS(recall);
            } else if (macros[imacro] <= 0) {
                status = OCSM_NOTHING_PREVIOUSLY_STORED;
                CATCH_STATUS(recall);
            }

            patn[npatn].ipatbeg = macros[imacro];
            patn[npatn].ipatend = ibrch;
            patn[npatn].ncopy   =  1;
            patn[npatn].icopy   =  1;
            patn[npatn].ipmtr   = -1;
            npatn++;

            /* start executing just after the macbeg */
            ibrch = macros[imacro];

            SPRINT1(1, "                          Entering storage %d", imacro);

        /* execute: "store $name index=0 keep=0" */
        } else if (type ==OCSM_STORE) {
            SPRINT4(1, "    executing [%4d] store:      %s  %11.5f  %11.5f",
                    ibrch, args[1].str, args[2].val[0], args[3].val[0]);

            /* pop a Body from the stack */
            if (nstack < 1) {
                signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                            "STORE expects a Body on the stack");
                status = OCSM_INSUFFICIENT_BODYS_ON_STACK;
                CATCH_STATUS(store);
            } else if (NINT(args[3].val[0]) == 1) {
                ibodyl = stack[nstack-1];
            } else {
                ibodyl = stack[--nstack];
            }

            /* if ibodyl is 0, it is a mark, so return an error */
            if (ibodyl == 0) {
                signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                            "STORE cannot be applied to a MARK");
                status = OCSM_INSUFFICIENT_BODYS_ON_STACK;
                CATCH_STATUS(store);
            }

            /* see if name/index is already used */
            istor = -1;

            for (jstor = 0; jstor < MODL->nstor; jstor++) {
                if (strcmp(args[1].str, MODL->stor[jstor].name) == 0 &&
                    NINT(args[2].val[0]) ==  MODL->stor[jstor].index   ) {
                    istor = jstor;
                    SPRINT2(0, "WARNING:: storage %s (%d) is being overwritten",
                            MODL->stor[jstor].name, MODL->stor[jstor].index);
                    break;
                }
            }

            /* if storage does not exist, create it now */
            if (istor < 0) {
                istor = MODL->nstor;

                (MODL->nstor)++;
                if (MODL->stor == NULL) {
                    MALLOC(MODL->stor, stor_T, MODL->nstor);
                } else {
                    RALLOC(MODL->stor, stor_T, MODL->nstor);
                }

                STRNCPY(MODL->stor[istor].name, args[1].str, MAX_NAME_LEN);
                MODL->stor[istor].index = NINT(args[2].val[0]);
            }

            /* save the Body info */
            MODL->stor[istor].ibody = ibodyl;
            MODL->stor[istor].ebody = MODL->body[ibodyl].ebody;

            SPRINT1(1, "                          Body   %4d stored", ibodyl);

        /* execute: "patbeg $pmtrName ncopy" */
        } else if (type == OCSM_PATBEG) {
            SPRINT3(1, "    executing [%4d] patbeg:         %s  %11.5f",
                    ibrch, args[1].str, args[2].val[0]);

            if (npatn >= MAX_NUM_PATTERNS) {
                signalError(MODL, OCSM_NESTED_TOO_DEEPLY, ibrch,
                            "npatn=%d >0 MAX_PATTERNS=%d", npatn, MAX_NUM_PATTERNS);
                status = OCSM_NESTED_TOO_DEEPLY;
                CATCH_STATUS(patbeg);
            }

            /* store info about this pattern */
            patn[npatn].ipatbeg = ibrch;
            patn[npatn].ipatend = -1;
            patn[npatn].icopy   = 1;
            patn[npatn].ncopy   = NINT(args[2].val[0]);
            patn[npatn].ipmtr   = -1;

            /* find matching patend */
            i = 1;
            for (ibrchl = ibrch+1; ibrchl <= MODL->nbrch; ibrchl++) {
                if        (MODL->brch[ibrchl].type == OCSM_PATBEG) {
                    i++;
                } else if (MODL->brch[ibrchl].type == OCSM_PATEND) {
                    i--;

                    if (i == 0) {
                        patn[npatn].ipatend = ibrchl;
                        break;
                    }
                }
            }

            /* find the pmtrName */
            status = ocsmFindPmtr(MODL, args[1].str, OCSM_INTERNAL, 1, 1,
                                  &(patn[npatn].ipmtr));
            CATCH_STATUS(ocsmFindPmtr);

            /* if no copies are required, jump to Branch after patend */
            if (patn[npatn].ncopy < 1) {
                if (patn[npatn].ipatend > patn[npatn].ipatbeg) {
                    ibrch = patn[npatn].ipatend;
                } else {
                    break;    /* used if no natching patend */
                }

            /* otherwise increment the number of patterns and continue */
            } else {
                irow = icol = 1;

                status = ocsmSetValuD(MODL, patn[npatn].ipmtr, irow, icol, 1.0);
                CATCH_STATUS(ocsmSetValuD);

                SPRINT2(1, "                          pattern counter: %s = %3d",
                        MODL->pmtr[patn[npatn].ipmtr].name, 1);

                npatn++;
            }

        /* execute: "patbreak expr" */
        } else if (type == OCSM_PATBREAK) {
            SPRINT2(1, "    executing [%4d] patbreak:       %11.5f",
                    ibrch, args[1].val[0]);

            /* break out of this pattern if the argument is positive */
            if (args[1].val[0] > 0) {
                npatn--;

                while (ibrch < MODL->nbrch && MODL->brch[ibrch].type != OCSM_PATEND) {
                    ibrch++;
                }
            }

        /* execute: "patend" */
        } else if (type == OCSM_PATEND) {
            SPRINT1(1, "    executing [%4d] patend:",
                    ibrch);

            /* increment the iterator */
            (patn[npatn-1].icopy)++;

            /* go back for another copy */
            if (patn[npatn-1].icopy <= patn[npatn-1].ncopy) {
                newValue = patn[npatn-1].icopy;

                irow = icol = 1;

                status = ocsmSetValuD(MODL, patn[npatn-1].ipmtr, irow, icol, newValue);
                CATCH_STATUS(ocsmSetValuD);

                ibrch = patn[npatn-1].ipatbeg;

                SPRINT2(1, "                          pattern counter: %s = %3d",
                        MODL->pmtr[patn[npatn-1].ipmtr].name, patn[npatn-1].icopy);

            /* otherwise, we are finished with the pattern */
            } else {
                npatn--;
            }

        /* execute: "mark" */
        } else if (type == OCSM_MARK) {
            SPRINT1(1, "    executing [%4d] mark:",
                    ibrch);

            /* push a Mark onto the stack */
            stack[nstack++] = 0;

            SPRINT0(1, "                          Mark        created");

        /* execute: "dump $filename remove=0" */
        } else if (type == OCSM_DUMP) {
            SPRINT3(1, "    executing [%4d] dump:       %s  %11.5f",
                    ibrch, args[1].str, args[2].val[0]);

            /* pop a Body from the stack */
            if (nstack < 1) {
                signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                            "DUMP expects a Body on the stack");
                status = OCSM_INSUFFICIENT_BODYS_ON_STACK;
                CATCH_STATUS(dump);
            } else {
                ibodyl = stack[--nstack];
            }

            /* make a model and dump the Body */
            status = EG_copyObject(MODL->body[ibodyl].ebody, NULL, &etemp);
            CATCH_STATUS(EG_copyObject);

            ebodyl = MODL->body[ibodyl].ebody;

            status = EG_makeTopology(MODL->context, NULL, MODEL, 0,
                                     NULL, 1, &etemp, NULL, &emodel);
            CATCH_STATUS(EG_makeTopology);

            /* strip the dollarsign off the filename */
            STRNCPY(dumpfile, args[1].str, MAX_EXPR_LEN);

            /* if file exists, delete it now */
            status = remove(dumpfile);
            if (status == 0) {
                SPRINT1(0, "WARNING:: file \"%s\" is being overwritten", dumpfile);
            }

            status = EG_saveModel(emodel, dumpfile);
            CATCH_STATUS(EG_saveModel);

            status = EG_deleteObject(emodel);
            CATCH_STATUS(EG_deleteObject);

            status = EG_copyObject(ebodyl, NULL, &ebody);
            CATCH_STATUS(EG_copyObject);

            SPRINT1(1, "                          Body   %4d dumped", ibodyl);

            /* push the Body back onto the stack (if it should not be removed) */
            if (NINT(args[2].val[0]) == 1) {
                SPRINT1(1, "                          Body   %4d removed", ibodyl);
            } else {
                stack[nstack++] = ibodyl;
            }

        /* execute: "end" */
        } else if (type == OCSM_END) {
            SPRINT1(1, "    executing [%4d] end:", ibrch);

            if (MODL->level <= 0) break;

            /* for each of the interface statements with a "out", write the
               @@-parameter */
            for (jbrch = ibrch-1; jbrch > 0; jbrch--) {
                if (MODL->brch[jbrch].indent == MODL->brch[ibrch].indent) {
                    if (MODL->brch[jbrch].type == OCSM_UDPRIM) {
                        break;
                    } else {
                        signalError(MODL, OCSM_INTERNAL, ibrch,
                                    "expecting UDPRIM, but got %s", ocsmGetText(MODL->brch[jbrch].type));
                        status = OCSM_INTERNAL_ERROR;
                        CATCH_STATUS(end);
                    }
                }
            }

            /* jbrch is first Branch after the udprim statement that called this .udc */
            jbrch++;
            while (jbrch < ibrch) {
                if (MODL->brch[jbrch].type != OCSM_INTERFACE) break;

                if (strcmp(MODL->brch[jbrch].arg2, "$out") == 0) {

                    /* get the value of the local variable */
                    STRNCPY(pmtrName, &(MODL->brch[jbrch].arg1[1]), MAX_EXPR_LEN);

                    status = str2val(pmtrName, MODL, &value, &dot);
                    CATCH_STATUS(str2val);

                    /* find or create an @@-parameter up one level and store the value/dot */
                    (MODL->level)--;
                    snprintf(pmtrName, MAX_EXPR_LEN, "@@%s", &(MODL->brch[jbrch].arg1[1]));

                    status = ocsmFindPmtr(MODL, pmtrName, OCSM_INTERNAL, 1, 1, &ipmtr);
                    CATCH_STATUS(ocsmFindPmtr);

                    status = ocsmSetValuD(MODL, ipmtr, 1, 1, value);
                    CATCH_STATUS(ocsmSetValuD);

                    status = ocsmSetVelD(MODL, ipmtr, 1, 1, dot);
                    CATCH_STATUS(ocsmSetVelD);
                    (MODL->level)++;
                }

                jbrch++;
            }

            /* free up all storage associated with Parameters at MODL->nlevel */
            npmtr_temp = MODL->npmtr;
            for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
                if (MODL->pmtr[ipmtr].level >= MODL->level) {

                    FREE(MODL->pmtr[ipmtr].name );
                    FREE(MODL->pmtr[ipmtr].value);
                    FREE(MODL->pmtr[ipmtr].dot  );
                    FREE(MODL->pmtr[ipmtr].lbnd );
                    FREE(MODL->pmtr[ipmtr].ubnd );
                    npmtr_temp--;
                }
            }

            /* remove all the Parameters at MODL->level */
            for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
                if (MODL->pmtr[ipmtr].level >= MODL->level) {
                    for (jpmtr = ipmtr+1; jpmtr <= MODL->npmtr; jpmtr++) {
                        if (MODL->pmtr[jpmtr].level < MODL->level) {

                            /* copy jpmtr over ipmtr */
                            MODL->pmtr[ipmtr].name  = MODL->pmtr[jpmtr].name;
                            MODL->pmtr[ipmtr].type  = MODL->pmtr[jpmtr].type;
                            MODL->pmtr[ipmtr].level = MODL->pmtr[jpmtr].level;
                            MODL->pmtr[ipmtr].nrow  = MODL->pmtr[jpmtr].nrow;
                            MODL->pmtr[ipmtr].ncol  = MODL->pmtr[jpmtr].ncol;
                            MODL->pmtr[ipmtr].value = MODL->pmtr[jpmtr].value;
                            MODL->pmtr[ipmtr].dot   = MODL->pmtr[jpmtr].dot;
                            MODL->pmtr[ipmtr].lbnd  = MODL->pmtr[jpmtr].lbnd;
                            MODL->pmtr[ipmtr].ubnd  = MODL->pmtr[jpmtr].ubnd;

                            MODL->pmtr[jpmtr].level = MODL->level;
                            break;
                        }
                    }
                }
            }
            MODL->npmtr = npmtr_temp;

            (MODL->level)--;

        /* execute: "assert arg1 arg2" */
        } else if (type == OCSM_ASSERT) {
            SPRINT3(1, "    executing [%4d] assert:     %s %s",
                    ibrch, MODL->brch[ibrch].arg1,
                           MODL->brch[ibrch].arg2);

            /* return error if arg1 != arg2 */
            if (fabs(args[1].val[0]-args[2].val[0]) > 1.0e-6) {
                SPRINT1(0, "    arg1 = %20.8f", args[1].val[0]);
                SPRINT1(0, "    arg2 = %20.8f", args[2].val[0]);
                status = OCSM_ASSERT_FAILED;
                CATCH_STATUS(assert);
            }
        }

        /* keep track of profile info */
        new_time = clock();

        if (type >= 100 && type < 200) {
            MODL->profile[type-100].ncall += 1;
            MODL->profile[type-100].time  += (new_time - old_time);
        }

        /* record that we successfully executed this Branch */
        *builtTo = ibrch;
    }

finalize:
    SPRINT0(1, "    finalizing:");

    /* mark all Branches up to builtTo as as not dirty and those after
       builtTo as dirty */
    for (ibrch = 1; ibrch <= MODL->nbrch; ibrch++) {
        if (ibrch <= *builtTo) {
            MODL->brch[ibrch].dirty = 0;
        } else {
            MODL->brch[ibrch].dirty = 1;
        }
    }

    /* special processing if an error was caught */
    if (ncatch > 0) {
        buildStatus = status;

        /* close the Sketch */
        sket->type   = 0;
        sket->size   = 0;
        sket->solved = 1;
        sket->nseg   = 0;
        sket->nvar   = 0;
        sket->ncon   = 0;

        /* return negative of builtTo */
        (*builtTo) *= -1;
    } else {
        buildStatus = SUCCESS;
    }

    /* make sure the Sketch is not open */
    if (sket->nseg > 0) {
        SET_STATUS(OCSM_SKETCH_IS_OPEN, end);
    }

    /* make sure that user-supplied body[] is big enough */
    if (body != NULL) {
        if (nstack > nbodyMax) {
            status = OCSM_TOO_MANY_BODYS_ON_STACK;
            goto cleanup;
        }
    }

    /* pre-mark all the Bodys as not being on the stack (fixed below) */
    for (ibody = 1; ibody <= MODL->nbody; ibody++) {
        MODL->body[ibody].onstack = 0;
    }

    /* report a warning if there is a mark or NODE_BODY on the stack */
    for (i = 0; i < nstack; i++) {
        ibody = stack[i];
        if (ibody == 0) {
            SPRINT0(0, "WARNING:: mark being removed from the stack");

            for (j = i+1; j < nstack; j++) {
                stack[j-1] = stack[j];
            }
            nstack--;

            i--;  /* repeat in case first moved stack entry was also a mark */
        } else if (ibody > 0) {
            if (MODL->body[ibody].botype == OCSM_NODE_BODY) {
                SPRINT0(0, "WARNING:: node body being removed from the stack");

                for (j = i+1; j < nstack; j++) {
                    stack[j-1] = stack[j];
                }
                nstack--;

                i--;  /* repeat in case first moved stack entry was also a mark */
            }
        }

        /* free up the storage associated with the arguments */
        for (iarg = 1; iarg < 10; iarg++) {
            FREE(args[iarg].val);
            FREE(args[iarg].dot);
        }
    }

    /* mark the stack's Bodys */
    for (i = 0; i < nstack; i++) {
        MODL->body[stack[i]].onstack = 1;
    }

    /* return Bodys on the stack (LIFO) */
    if (body != NULL) {
        *nbody = nstack;
    }

    nbodyMax = nstack;
    for (i = 0; i < nbodyMax; i++) {
        if (body != NULL) {
            body[i] = stack[--nstack];

            status = getBodyTolerance(MODL->body[body[i]].ebody, &toler);
            CHECK_STATUS(getBodyTolerance);

            SPRINT5(1, "    Body %5d   nnode=%-6d   nedge=%-6d   nface=%-6d   toler=%11.4e", body[i],
                    MODL->body[body[i]].nnode,
                    MODL->body[body[i]].nedge,
                    MODL->body[body[i]].nface, toler);

            if (outLevel >= 3) {
                status = printBodyAttributes(MODL, body[i]);
                CHECK_STATUS(printBodyAttributes);
            }
        }
    }

    /* print the profile info */
    SPRINT0(1, "==> Profile information");
    total_time = 0;
    total_call = 0;
    for (i = 0; i < 100; i++) {
        total_call += MODL->profile[i].ncall;
        total_time += MODL->profile[i].time;
    }
    SPRINT0(1, "    Branch type           ncall  time (sec)    pct");
    for (i = 0; i < 100; i++) {
        if (MODL->profile[i].ncall > 0) {
            SPRINT4(1, "    %-20s  %5d  %10.3f  %5.1f", ocsmGetText(i+100), MODL->profile[i].ncall,
                   (double)(MODL->profile[i].time) / (double)(CLOCKS_PER_SEC),
                   (double)(MODL->profile[i].time) / (double)(total_time) * 100);
        }
    }
    SPRINT1(1, "   Total                  %5d", total_call);

cleanup:
    for (iarg = 1; iarg < 10; iarg++) {
        FREE(args[iarg].val);
        FREE(args[iarg].dot);
    }

    FREE(sket   );
    FREE(solcons);
    FREE(patn   );
    FREE(macros );
    FREE(stack  );

#undef CATCH_STATUS
    /* if buildStatus is not success, return it */
    if (buildStatus != SUCCESS) {
        status = buildStatus;
    }

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmNewBrch - create a new Branch                                  *
 *                                                                      *
 ************************************************************************
 */

int
ocsmNewBrch(void   *modl,               /* (in)  pointer to MODL */
            int    iafter,              /* (in)  Branch index (0-nbrch) after which to add */
            int    type,                /* (in)  Branch type (see below) */
  /*@null@*/char   arg1[],              /* (in)  Argument 1 (or NULL) */
  /*@null@*/char   arg2[],              /* (in)  Argument 2 (or NULL) */
  /*@null@*/char   arg3[],              /* (in)  Argument 3 (or NULL) */
  /*@null@*/char   arg4[],              /* (in)  Argument 4 (or NULL) */
  /*@null@*/char   arg5[],              /* (in)  Argument 5 (or NULL) */
  /*@null@*/char   arg6[],              /* (in)  Argument 6 (or NULL) */
  /*@null@*/char   arg7[],              /* (in)  Argument 7 (or NULL) */
  /*@null@*/char   arg8[],              /* (in)  Argument 8 (or NULL) */
  /*@null@*/char   arg9[])              /* (in)  Argument 9 (or NULL) */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       class, narg, ibrch, jbrch;

    ROUTINE(ocsmNewBrch);
    DPRINT3("%s(iafter=%d, type=%d) {",
            routine, iafter, type);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* make sure valid ibrch was given */
    if (iafter < 0 || iafter > MODL->nbrch) {
        status = OCSM_ILLEGAL_BRCH_INDEX;
        goto cleanup;
    }

    /* if this has a perturbed Body, free it first */
    status = removePerturbation(MODL);
    CHECK_STATUS(removePerturbation);

    /* check that valid type is given (and determine class) */
    if        (type == OCSM_ARC) {
        class = OCSM_SKETCH;
        narg = 5;
    } else if (type == OCSM_ASSERT) {
        class = OCSM_UTILITY;
        narg  = 2;
    } else if (type == OCSM_BEZIER) {
        class = OCSM_SKETCH;
        narg  = 3;
    } else if (type == OCSM_BLEND) {
        class = OCSM_GROWN;
        narg  = 3;
    } else if (type == OCSM_BOX) {
        class = OCSM_PRIMITIVE;
        narg  = 6;
    } else if (type == OCSM_CHAMFER) {
        class = OCSM_APPLIED;
        narg  = 2;
    } else if (type == OCSM_CIRARC) {
        class = OCSM_SKETCH;
        narg = 6;
    } else if (type == OCSM_COMBINE) {
        class = OCSM_BOOLEAN;
        narg  = 0;
    } else if (type == OCSM_CONE) {
        class = OCSM_PRIMITIVE;
        narg  = 7;
    } else if (type == OCSM_CONNECT) {
        class = OCSM_BOOLEAN;
        narg  = 2;
    } else if (type == OCSM_CYLINDER) {
        class = OCSM_PRIMITIVE;
        narg  = 7;
    } else if (type == OCSM_DUMP) {
        class = OCSM_UTILITY;
        narg  = 2;
    } else if (type == OCSM_END) {
        class = OCSM_UTILITY;
        narg  = 0;
    } else if (type == OCSM_EXTRACT) {
        class = OCSM_BOOLEAN;
        narg  = 1;
    } else if (type == OCSM_EXTRUDE) {
        class = OCSM_GROWN;
        narg  = 3;
    } else if (type == OCSM_FILLET) {
        class = OCSM_APPLIED;
        narg  = 2;
    } else if (type == OCSM_HOLLOW) {
        class = OCSM_APPLIED;
        narg  = 2;
    } else if (type == OCSM_IMPORT) {
        class = OCSM_PRIMITIVE;
        narg  = 2;
    } else if (type == OCSM_INTERFACE) {
        class = OCSM_UTILITY;
        narg  = 3;
    } else if (type == OCSM_INTERSECT) {
        class = OCSM_BOOLEAN;
        narg  = 3;
    } else if (type == OCSM_JOIN) {
        class = OCSM_BOOLEAN;
        narg  = 1;
    } else if (type == OCSM_LINSEG) {
        class = OCSM_SKETCH;
        narg  = 3;
    } else if (type == OCSM_LOFT) {
        class = OCSM_GROWN;
        narg  = 1;
    } else if (type == OCSM_MACBEG) {
        class = OCSM_UTILITY;
        narg  = 1;
    } else if (type == OCSM_MACEND) {
        class = OCSM_UTILITY;
        narg  = 0;
    } else if (type == OCSM_MARK) {
        class = OCSM_UTILITY;
        narg  = 0;
    } else if (type == OCSM_MIRROR) {
        class = OCSM_TRANSFORM;
        narg  = 4;
    } else if (type == OCSM_PATBEG) {
        class = OCSM_UTILITY;
        narg  = 2;
    } else if (type == OCSM_PATBREAK) {
        class = OCSM_UTILITY;
        narg  = 1;
    } else if (type == OCSM_PATEND) {
        class = OCSM_UTILITY;
        narg  = 0;
    } else if (type == OCSM_PROJECT) {
        class = OCSM_UTILITY;
        narg  = 6;
    } else if (type == OCSM_RECALL) {
        class = OCSM_UTILITY;
        narg  = 1;
    } else if (type == OCSM_REORDER) {
        class = OCSM_TRANSFORM;
        narg  = 2;
    } else if (type == OCSM_RESTORE) {
        class = OCSM_PRIMITIVE;
        narg  = 2;
    } else if (type == OCSM_REVOLVE) {
        class = OCSM_GROWN;
        narg  = 7;
    } else if (type == OCSM_ROTATEX) {
        class = OCSM_TRANSFORM;
        narg  = 3;
    } else if (type == OCSM_ROTATEY) {
        class = OCSM_TRANSFORM;
        narg  = 3;
    } else if (type == OCSM_ROTATEZ) {
        class = OCSM_TRANSFORM;
        narg  = 3;
    } else if (type == OCSM_RULE) {
        class = OCSM_GROWN;
        narg  = 1;
    } else if (type == OCSM_SCALE) {
        class = OCSM_TRANSFORM;
        narg  = 1;
    } else if (type == OCSM_SELECT) {
        class = OCSM_UTILITY;
        narg  = 1;
        if (arg2 != NULL && arg2[0] != '\0') narg = 2;
        if (arg3 != NULL && arg3[0] != '\0') narg = 3;
        if (arg4 != NULL && arg4[0] != '\0') narg = 4;
        if (arg5 != NULL && arg5[0] != '\0') narg = 5;
        if (arg6 != NULL && arg6[0] != '\0') narg = 6;
        if (arg7 != NULL && arg7[0] != '\0') narg = 7;
        if (arg8 != NULL && arg8[0] != '\0') narg = 8;
    } else if (type == OCSM_SET) {
        class = OCSM_UTILITY;
        narg  = 2;
    } else if (type == OCSM_SKBEG) {
        class = OCSM_SKETCH;
        narg  = 3;
    } else if (type == OCSM_SKCON) {
        class = OCSM_SKETCH;
        narg  = 4;
    } else if (type == OCSM_SKEND) {
        class = OCSM_SKETCH;
        narg  = 1;
    } else if (type == OCSM_SKVAR) {
        class = OCSM_SKETCH;
        narg  = 2;
    } else if (type == OCSM_SOLBEG) {
        class = OCSM_SOLVER;
        narg  = 1;
    } else if (type == OCSM_SOLCON) {
        class = OCSM_SOLVER;
        narg  = 1;
    } else if (type == OCSM_SOLEND) {
        class = OCSM_SOLVER;
        narg  = 0;
    } else if (type == OCSM_SPHERE) {
        class = OCSM_PRIMITIVE;
        narg  = 4;
    } else if (type == OCSM_SPLINE) {
        class = OCSM_SKETCH;
        narg  = 3;
    } else if (type == OCSM_STORE) {
        class = OCSM_UTILITY;
        narg  = 3;
    } else if (type == OCSM_SUBTRACT) {
        class = OCSM_BOOLEAN;
        narg  = 3;
    } else if (type == OCSM_SWEEP) {
        class = OCSM_GROWN;
        narg  = 0;
    } else if (type == OCSM_TORUS) {
        class = OCSM_PRIMITIVE;
        narg  = 8;
    } else if (type == OCSM_TRANSLATE) {
        class = OCSM_TRANSFORM;
        narg  = 3;
    } else if (type == OCSM_UDPARG) {
        class = OCSM_UTILITY;
        narg  = 1;
        if (arg2 != NULL && arg2[0] != '\0' &&
            arg3 != NULL && arg3[0] != '\0'   ) narg = 3;
        if (arg4 != NULL && arg4[0] != '\0' &&
            arg5 != NULL && arg5[0] != '\0'   ) narg = 5;
        if (arg6 != NULL && arg6[0] != '\0' &&
            arg7 != NULL && arg7[0] != '\0'   ) narg = 7;
        if (arg8 != NULL && arg8[0] != '\0' &&
            arg9 != NULL && arg9[0] != '\0'   ) narg = 9;
    } else if (type == OCSM_UDPRIM) {
        class = OCSM_PRIMITIVE;
        narg  = 1;
        if (arg2 != NULL && arg2[0] != '\0' &&
            arg3 != NULL && arg3[0] != '\0'   ) narg = 3;
        if (arg4 != NULL && arg4[0] != '\0' &&
            arg5 != NULL && arg5[0] != '\0'   ) narg = 5;
        if (arg6 != NULL && arg6[0] != '\0' &&
            arg7 != NULL && arg7[0] != '\0'   ) narg = 7;
        if (arg8 != NULL && arg8[0] != '\0' &&
            arg9 != NULL && arg9[0] != '\0'   ) narg = 9;
    } else if (type == OCSM_UNION) {
        class = OCSM_BOOLEAN;
        narg  = 3;
    } else {
        status = OCSM_ILLEGAL_TYPE;
        goto cleanup;
    }

    /* mark MODL as not being checked */
    MODL->checked = 0;

    /* increment the number of Branches */
    (MODL->nbrch)++;
    ibrch = iafter + 1;

    /* extend the Branch list (if needed) */
    if (MODL->nbrch > MODL->mbrch) {
        MODL->mbrch += 25;
        if (MODL->brch == NULL) {
            MALLOC(MODL->brch, brch_T, MODL->mbrch+1);
        } else {
            RALLOC(MODL->brch, brch_T, MODL->mbrch+1);
        }
    }

    /* copy the Branches up to make room for the new Branch */
#ifndef __clang_analyzer__
    if (ibrch < MODL->nbrch) {
        for (jbrch = MODL->nbrch; jbrch > ibrch; jbrch--) {
            MODL->brch[jbrch].name   = MODL->brch[jbrch-1].name;
            MODL->brch[jbrch].type   = MODL->brch[jbrch-1].type;
            MODL->brch[jbrch].class  = MODL->brch[jbrch-1].class;
            MODL->brch[jbrch].indent = MODL->brch[jbrch-1].indent;
            MODL->brch[jbrch].actv   = MODL->brch[jbrch-1].actv;
            MODL->brch[jbrch].dirty  = 1;
            MODL->brch[jbrch].nattr  = MODL->brch[jbrch-1].nattr;
            MODL->brch[jbrch].attr   = MODL->brch[jbrch-1].attr;
            MODL->brch[jbrch].ileft  = MODL->brch[jbrch-1].ileft;
            MODL->brch[jbrch].irite  = MODL->brch[jbrch-1].irite;
            MODL->brch[jbrch].ichld  = MODL->brch[jbrch-1].ichld;
            MODL->brch[jbrch].narg   = MODL->brch[jbrch-1].narg;
            MODL->brch[jbrch].arg1   = MODL->brch[jbrch-1].arg1;
            MODL->brch[jbrch].arg2   = MODL->brch[jbrch-1].arg2;
            MODL->brch[jbrch].arg3   = MODL->brch[jbrch-1].arg3;
            MODL->brch[jbrch].arg4   = MODL->brch[jbrch-1].arg4;
            MODL->brch[jbrch].arg5   = MODL->brch[jbrch-1].arg5;
            MODL->brch[jbrch].arg6   = MODL->brch[jbrch-1].arg6;
            MODL->brch[jbrch].arg7   = MODL->brch[jbrch-1].arg7;
            MODL->brch[jbrch].arg8   = MODL->brch[jbrch-1].arg8;
            MODL->brch[jbrch].arg9   = MODL->brch[jbrch-1].arg9;
        }
    }
#endif

    /* initialize the new Branch */
    MODL->brch[ibrch].name   = NULL;
    MODL->brch[ibrch].type   = type;
    MODL->brch[ibrch].class  = class;
    MODL->brch[ibrch].indent = 0;
    MODL->brch[ibrch].actv   =  OCSM_ACTIVE;
    MODL->brch[ibrch].dirty  = 1;
    MODL->brch[ibrch].nattr  = 0;
    MODL->brch[ibrch].attr   = NULL;
    MODL->brch[ibrch].ileft  = -1;
    MODL->brch[ibrch].irite  = -1;
    MODL->brch[ibrch].ichld  = -1;
    MODL->brch[ibrch].narg   =  narg;
    MODL->brch[ibrch].arg1   =  NULL;
    MODL->brch[ibrch].arg2   =  NULL;
    MODL->brch[ibrch].arg3   =  NULL;
    MODL->brch[ibrch].arg4   =  NULL;
    MODL->brch[ibrch].arg5   =  NULL;
    MODL->brch[ibrch].arg6   =  NULL;
    MODL->brch[ibrch].arg7   =  NULL;
    MODL->brch[ibrch].arg8   =  NULL;
    MODL->brch[ibrch].arg9   =  NULL;

    /* default name for the Branch */
    MALLOC(  MODL->brch[ibrch].name, char, 12);
    snprintf(MODL->brch[ibrch].name,       12, "Brch_%06d", MODL->nextseq);
    (MODL->nextseq)++;

    /* save the Arguments */
    if (narg >= 1) {
        if (arg1 != NULL && strlen(arg1) > 0) {
            MALLOC(MODL->brch[ibrch].arg1, char, (int)(strlen(arg1)+1));
            strcpy(MODL->brch[ibrch].arg1,                    arg1    );
        } else {
            status = OCSM_ILLEGAL_NARG;
        }
    }

    if (narg >= 2) {
        if (arg2 != NULL && strlen(arg2) > 0) {
            MALLOC(MODL->brch[ibrch].arg2, char, (int)(strlen(arg2)+1));
            strcpy(MODL->brch[ibrch].arg2,                    arg2    );
        } else {
            status = OCSM_ILLEGAL_NARG;
        }
    }

    if (narg >= 3) {
        if (arg3 != NULL && strlen(arg3) > 0) {
            MALLOC(MODL->brch[ibrch].arg3, char, (int)(strlen(arg3)+1));
            strcpy(MODL->brch[ibrch].arg3,                    arg3    );
        } else {
            status = OCSM_ILLEGAL_NARG;
        }
    }

    if (narg >= 4) {
        if (arg4 != NULL && strlen(arg4) > 0) {
            MALLOC(MODL->brch[ibrch].arg4, char, (int)(strlen(arg4)+1));
            strcpy(MODL->brch[ibrch].arg4,                    arg4    );
        } else {
            status = OCSM_ILLEGAL_NARG;
        }
    }

    if (narg >= 5) {
        if (arg5 != NULL && strlen(arg5) > 0) {
            MALLOC(MODL->brch[ibrch].arg5, char, (int)(strlen(arg5)+1));
            strcpy(MODL->brch[ibrch].arg5,                    arg5    );
        } else {
            status = OCSM_ILLEGAL_NARG;
        }
    }

    if (narg >= 6) {
        if (arg6 != NULL && strlen(arg6) > 0) {
            MALLOC(MODL->brch[ibrch].arg6, char, (int)(strlen(arg6)+1));
            strcpy(MODL->brch[ibrch].arg6,                    arg6    );
        } else {
            status = OCSM_ILLEGAL_NARG;
        }
    }

    if (narg >= 7) {
        if (arg7 != NULL && strlen(arg7) > 0) {
            MALLOC(MODL->brch[ibrch].arg7, char, (int)(strlen(arg7)+1));
            strcpy(MODL->brch[ibrch].arg7,                    arg7    );
        } else {
            status = OCSM_ILLEGAL_NARG;
        }
    }

    if (narg >= 8) {
        if (arg8 != NULL && strlen(arg8) > 0) {
            MALLOC(MODL->brch[ibrch].arg8, char, (int)(strlen(arg8)+1));
            strcpy(MODL->brch[ibrch].arg8,                    arg8    );
        } else {
            status = OCSM_ILLEGAL_NARG;
        }
    }

    if (narg >= 9) {
        if (arg9 != NULL && strlen(arg9) > 0) {
            MALLOC(MODL->brch[ibrch].arg9, char, (int)(strlen(arg9)+1));
            strcpy(MODL->brch[ibrch].arg9,                    arg9    );
        } else {
            status = OCSM_ILLEGAL_NARG;
        }
    }

cleanup:
    if (status < SUCCESS) {
        MODL->nbrch--;
    }

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmGetBrch - get info about a Branch                              *
 *                                                                      *
 ************************************************************************
 */

int
ocsmGetBrch(void   *modl,               /* (in)  pointer to MODL */
            int    ibrch,               /* (in)  Branch index (1-nbrch) */
            int    *type,               /* (out) Branch type */
            int    *class,              /* (out) Branch class */
            int    *actv,               /* (out) Branch Activity */
            int    *ichld,              /* (out) ibrch of child (or 0 if root) */
            int    *ileft,              /* (out) ibrch of left parent (or 0) */
            int    *irite,              /* (out) ibrch of rite parent (or 0) */
            int    *narg,               /* (out) number of Arguments */
            int    *nattr)              /* (out) number of Attributes */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(ocsmGetBrch);
    DPRINT2("%s(ibrch=%d) {",
            routine, ibrch);

    /* --------------------------------------------------------------- */

    /* default return values */
    *type  = 0;
    *class = 0;
    *actv  = 0;
    *ichld = 0;
    *ileft = 0;
    *irite = 0;
    *narg  = 0;
    *nattr = 0;

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ibrch is given */
    if (ibrch < 1 || ibrch > MODL->nbrch) {
        status = OCSM_ILLEGAL_BRCH_INDEX;
        goto cleanup;
    }

    /* return pertinent information */
    *type  = MODL->brch[ibrch].type;
    *class = MODL->brch[ibrch].class;
    *actv  = MODL->brch[ibrch].actv;
    *ichld = MODL->brch[ibrch].ichld;
    *ileft = MODL->brch[ibrch].ileft;
    *irite = MODL->brch[ibrch].irite;
    *narg  = MODL->brch[ibrch].narg;
    *nattr = MODL->brch[ibrch].nattr;

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmSetBrch - set activity for a Branch                            *
 *                                                                      *
 ************************************************************************
 */

int
ocsmSetBrch(void   *modl,               /* (in)  pointer to MODL */
            int    ibrch,               /* (in)  Branch index (1-nbrch) */
            int    actv)                /* (in)  Branch activity */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(ocsmSetBrch);
    DPRINT3("%s(ibrch=%d, actv=%d) {",
            routine, ibrch, actv);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ibrch is given */
    if (ibrch < 1 || ibrch > MODL->nbrch) {
        status = OCSM_ILLEGAL_BRCH_INDEX;
        goto cleanup;
    }

    /* check that a valid activity was given */
    if (actv != OCSM_ACTIVE    &&
        actv != OCSM_SUPPRESSED  ) {
        status = OCSM_ILLEGAL_ACTIVITY;
        goto cleanup;
    }

    /* check that branch type can be suppressed */
    if (MODL->brch[ibrch].class != OCSM_PRIMITIVE &&
        MODL->brch[ibrch].class != OCSM_GROWN     &&
        MODL->brch[ibrch].class != OCSM_APPLIED   &&
        MODL->brch[ibrch].class != OCSM_TRANSFORM   ) {
        status = OCSM_CANNOT_BE_SUPPRESSED;
        goto cleanup;
    }

    /* if this has a perturbed Body, free it first */
    status = removePerturbation(MODL);
    CHECK_STATUS(removePerturbation);

    /* save the activity */
    MODL->brch[ibrch].actv = actv;

    /* mark the Branch as dirty */
    MODL->brch[ibrch].dirty = 1;

    /* mark MODL as not being checked */
    MODL->checked = 0;

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmDelBrch - delete the last Branch                               *
 *                                                                      *
 ************************************************************************
 */

int
ocsmDelBrch(void   *modl,               /* (in)  pointer to MODL */
            int    ibrch)               /* (in)  Branch index (1-nbrch) */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       jbrch, iattr;

    ROUTINE(ocsmDelBrch);
    DPRINT2("%s(ibrch=%d) {",
            routine, ibrch);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check for valid Branch index */
    if (ibrch < 1 || ibrch > MODL->nbrch) {
        status = OCSM_ILLEGAL_BRCH_INDEX;
        goto cleanup;
    }

    /* if this has a perturbed Body, free it first */
    status = removePerturbation(MODL);
    CHECK_STATUS(removePerturbation);

    /* mark MODL as not being checked */
    MODL->checked = 0;

    /* free up the storage associated with ibrch */
    for (iattr = 0; iattr < MODL->brch[ibrch].nattr; iattr++) {
        FREE(MODL->brch[ibrch].attr[iattr].name );
        FREE(MODL->brch[ibrch].attr[iattr].value);
    }

    FREE(MODL->brch[ibrch].name);
    FREE(MODL->brch[ibrch].attr);
    FREE(MODL->brch[ibrch].arg1);
    FREE(MODL->brch[ibrch].arg2);
    FREE(MODL->brch[ibrch].arg3);
    FREE(MODL->brch[ibrch].arg4);
    FREE(MODL->brch[ibrch].arg5);
    FREE(MODL->brch[ibrch].arg6);
    FREE(MODL->brch[ibrch].arg7);
    FREE(MODL->brch[ibrch].arg8);
    FREE(MODL->brch[ibrch].arg9);

    /* move all Branches up to write over deleted Branch */
    for (jbrch = ibrch; jbrch < MODL->nbrch; jbrch++) {
        MODL->brch[jbrch].name   = MODL->brch[jbrch+1].name;
        MODL->brch[jbrch].type   = MODL->brch[jbrch+1].type;
        MODL->brch[jbrch].class  = MODL->brch[jbrch+1].class;
        MODL->brch[jbrch].indent = MODL->brch[jbrch+1].indent;
        MODL->brch[jbrch].actv   = MODL->brch[jbrch+1].actv;
        MODL->brch[jbrch].dirty  = 1;
        MODL->brch[jbrch].nattr  = MODL->brch[jbrch+1].nattr;
        MODL->brch[jbrch].attr   = MODL->brch[jbrch+1].attr;
        MODL->brch[jbrch].ileft  = MODL->brch[jbrch+1].ileft;
        MODL->brch[jbrch].irite  = MODL->brch[jbrch+1].irite;
        MODL->brch[jbrch].ichld  = MODL->brch[jbrch+1].ichld;
        MODL->brch[jbrch].narg   = MODL->brch[jbrch+1].narg;
        MODL->brch[jbrch].arg1   = MODL->brch[jbrch+1].arg1;
        MODL->brch[jbrch].arg2   = MODL->brch[jbrch+1].arg2;
        MODL->brch[jbrch].arg3   = MODL->brch[jbrch+1].arg3;
        MODL->brch[jbrch].arg4   = MODL->brch[jbrch+1].arg4;
        MODL->brch[jbrch].arg5   = MODL->brch[jbrch+1].arg5;
        MODL->brch[jbrch].arg6   = MODL->brch[jbrch+1].arg6;
        MODL->brch[jbrch].arg7   = MODL->brch[jbrch+1].arg7;
        MODL->brch[jbrch].arg8   = MODL->brch[jbrch+1].arg8;
        MODL->brch[jbrch].arg9   = MODL->brch[jbrch+1].arg9;
    }

    /* decrement the number of Branches */
    (MODL->nbrch)--;

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmPrintBrchs - print Branches to file                            *
 *                                                                      *
 ************************************************************************
 */

int
ocsmPrintBrchs(void   *modl,            /* (in)  pointer to MODL */
               FILE   *fp)              /* (in)  pointer to FILE */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       ibrch, i, nindent, iattr;

    ROUTINE(ocsmPrintBrchs);
    DPRINT2("%s(fp=%lx) {",
            routine, (long)fp);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* run ocsmCheck to set up necessary linkages amongst Branches */
    status = ocsmCheck(MODL);
    CHECK_STATUS(ocsmCheck);

    /* print header */
    fprintf(fp, "    ibrch                    type             ileft irite ichld args...\n");

    /* print each Branch */
    for (ibrch = 1; ibrch <= MODL->nbrch; ibrch++) {

        fprintf(fp, "    %5d", ibrch);

        if (MODL->brch[ibrch].dirty > 0) {
            if (MODL->brch[ibrch].actv == OCSM_ACTIVE    ) fprintf(fp, " [a] ");
            if (MODL->brch[ibrch].actv == OCSM_SUPPRESSED) fprintf(fp, " [s] ");
            if (MODL->brch[ibrch].actv == OCSM_INACTIVE  ) fprintf(fp, " [i] ");
            if (MODL->brch[ibrch].actv == OCSM_DEFERRED  ) fprintf(fp, " [d] ");
        } else {
            if (MODL->brch[ibrch].actv == OCSM_ACTIVE    ) fprintf(fp, " (a) ");
            if (MODL->brch[ibrch].actv == OCSM_SUPPRESSED) fprintf(fp, " (s) ");
            if (MODL->brch[ibrch].actv == OCSM_INACTIVE  ) fprintf(fp, " (i) ");
            if (MODL->brch[ibrch].actv == OCSM_DEFERRED  ) fprintf(fp, " (d) ");
        }

        fprintf(fp, " %-14s", MODL->brch[ibrch].name);

        nindent = MODL->brch[ibrch].indent;
        for (i = 0; i < nindent; i++) {
            fprintf(fp, ".");
        }

        fprintf(fp, "%-9s", ocsmGetText(MODL->brch[ibrch].type));

        for (i = nindent; i < 8; i++) {
            fprintf(fp, " ");
        }

        fprintf(fp, "%5d %5d %5d", MODL->brch[ibrch].ileft,
                                   MODL->brch[ibrch].irite,
                                   MODL->brch[ibrch].ichld);

        if (MODL->brch[ibrch].narg >= 1) {
            if (MODL->brch[ibrch].arg1[0] != '$') {
                fprintf(fp, " [%s]",   MODL->brch[ibrch].arg1    );
            } else {
                fprintf(fp, " [%s]", &(MODL->brch[ibrch].arg1[1]));
            }
        }
        if (MODL->brch[ibrch].narg >= 2) {
            if (MODL->brch[ibrch].arg2[0] != '$') {
                fprintf(fp, " [%s]",   MODL->brch[ibrch].arg2    );
            } else {
                fprintf(fp, " [%s]", &(MODL->brch[ibrch].arg2[1]));
            }
        }
        if (MODL->brch[ibrch].narg >= 3) {
            if (MODL->brch[ibrch].arg3[0] != '$') {
                fprintf(fp, " [%s]",   MODL->brch[ibrch].arg3    );
            } else {
                fprintf(fp, " [%s]", &(MODL->brch[ibrch].arg3[1]));
            }
        }
        if (MODL->brch[ibrch].narg >= 4) {
            if (MODL->brch[ibrch].arg4[0] != '$') {
                fprintf(fp, " [%s]",   MODL->brch[ibrch].arg4    );
            } else {
                fprintf(fp, " [%s]", &(MODL->brch[ibrch].arg4[1]));
            }
        }
        if (MODL->brch[ibrch].narg >= 5) {
            if (MODL->brch[ibrch].arg5[0] != '$') {
                fprintf(fp, " [%s]",   MODL->brch[ibrch].arg5    );
            } else {
                fprintf(fp, " [%s]", &(MODL->brch[ibrch].arg5[1]));
            }
        }
        if (MODL->brch[ibrch].narg >= 6) {
            if (MODL->brch[ibrch].arg6[0] != '$') {
                fprintf(fp, " [%s]",   MODL->brch[ibrch].arg6    );
            } else {
                fprintf(fp, " [%s]", &(MODL->brch[ibrch].arg6[1]));
            }
        }
        if (MODL->brch[ibrch].narg >= 7) {
            if (MODL->brch[ibrch].arg7[0] != '$') {
                fprintf(fp, " [%s]",   MODL->brch[ibrch].arg7    );
            } else {
                fprintf(fp, " [%s]", &(MODL->brch[ibrch].arg7[1]));
            }
        }
        if (MODL->brch[ibrch].narg >= 8) {
            if (MODL->brch[ibrch].arg8[0] != '$') {
                fprintf(fp, " [%s]",   MODL->brch[ibrch].arg8    );
            } else {
                fprintf(fp, " [%s]", &(MODL->brch[ibrch].arg8[1]));
            }
        }
        if (MODL->brch[ibrch].narg >= 9) {
            if (MODL->brch[ibrch].arg9[0] != '$') {
                fprintf(fp, " [%s]",   MODL->brch[ibrch].arg9    );
            } else {
                fprintf(fp, " [%s]", &(MODL->brch[ibrch].arg9[1]));
            }
        }

        fprintf(fp, "\n");

        for (iattr = 0; iattr < MODL->brch[ibrch].nattr; iattr++) {
            fprintf(fp, "                                                            %-20s %-20s\n",
                    MODL->brch[ibrch].attr[iattr].name,
                    MODL->brch[ibrch].attr[iattr].value);
        }
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmGetArg - get an Argument for a Branch                          *
 *                                                                      *
 ************************************************************************
 */

int
ocsmGetArg(void   *modl,                /* (in)  pointer to MODL */
           int    ibrch,                /* (in)  Branch index (1-nbrch) */
           int    iarg,                 /* (in)  Argument index (1-narg) */
           char   defn[],               /* (out) Argument definition (at least 129 long) */
           double *value,               /* (out) Argument value */
           double *dot)                 /* (out) Argument velocity */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(ocsmGetArg);
    DPRINT3("%s(ibrch=%d, iarg=%d) {",
            routine, ibrch, iarg);

    /* --------------------------------------------------------------- */

    /* default return values */
    defn[0] = '\0';
    *value  = 0;

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ibrch is given */
    if (ibrch < 1 || ibrch > MODL->nbrch) {
        status = OCSM_ILLEGAL_BRCH_INDEX;
        goto cleanup;
    }

    /* check that valid iarg is given */
    if (iarg < 1 || iarg > MODL->brch[ibrch].narg) {
        status = OCSM_ILLEGAL_ARG_INDEX;
        goto cleanup;
    }

    /* return the definition and current value */
    if (iarg == 1) {
        STRNCPY(defn, MODL->brch[ibrch].arg1, 129);
    }
    if (iarg == 2) {
        STRNCPY(defn, MODL->brch[ibrch].arg2, 129);
    }
    if (iarg == 3) {
        STRNCPY(defn, MODL->brch[ibrch].arg3, 129);
    }
    if (iarg == 4) {
        STRNCPY(defn, MODL->brch[ibrch].arg4, 129);
    }
    if (iarg == 5) {
        STRNCPY(defn, MODL->brch[ibrch].arg5, 129);
    }
    if (iarg == 6) {
        STRNCPY(defn, MODL->brch[ibrch].arg6, 129);
    }
    if (iarg == 7) {
        STRNCPY(defn, MODL->brch[ibrch].arg7, 129);
    }
    if (iarg == 8) {
        STRNCPY(defn, MODL->brch[ibrch].arg8, 129);
    }
    if (iarg == 9) {
        STRNCPY(defn, MODL->brch[ibrch].arg9, 129);
    }

    status = str2val(defn, MODL, value, dot);

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmSetArg - set an Argument for a Branch                          *
 *                                                                      *
 ************************************************************************
 */

int
ocsmSetArg(void   *modl,                /* (in)  pointer to MODL */
           int    ibrch,                /* (in)  Branch index (1-nbrch) */
           int    iarg,                 /* (in)  Argument index (1-narg) */
           char   defn[])               /* (in)  Argument definition */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(ocsmSetArg);
    DPRINT4("%s(ibrch=%d, iarg=%d, defn=%s) {",
            routine, ibrch, iarg, defn);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ibrch is given */
    if (ibrch < 1 || ibrch > MODL->nbrch) {
        status = OCSM_ILLEGAL_BRCH_INDEX;
        goto cleanup;
    }

    /* check that valid iarg is given */
    if (iarg < 1 || iarg > MODL->brch[ibrch].narg) {
        status = OCSM_ILLEGAL_ARG_INDEX;
        goto cleanup;
    }

    /* if this has a perturbed Body, free it first */
    status = removePerturbation(MODL);
    CHECK_STATUS(removePerturbation);

    /* save the definition */
    if        (iarg == 1) {
        if (strcmp(defn, MODL->brch[ibrch].arg1) != 0) {
            FREE(  MODL->brch[ibrch].arg1);
            MALLOC(MODL->brch[ibrch].arg1, char, (int)(strlen(defn)+1));
            strcpy(MODL->brch[ibrch].arg1,                    defn    );
        }
    } else if (iarg == 2) {
        if (strcmp(defn, MODL->brch[ibrch].arg2) != 0) {
            FREE(  MODL->brch[ibrch].arg2);
            MALLOC(MODL->brch[ibrch].arg2, char, (int)(strlen(defn)+1));
            strcpy(MODL->brch[ibrch].arg2,                    defn    );
        }
    } else if (iarg == 3) {
        if (strcmp(defn, MODL->brch[ibrch].arg3) != 0) {
            FREE(  MODL->brch[ibrch].arg3);
            MALLOC(MODL->brch[ibrch].arg3, char, (int)(strlen(defn)+1));
            strcpy(MODL->brch[ibrch].arg3,                    defn    );
        }
    } else if (iarg == 4) {
        if (strcmp(defn, MODL->brch[ibrch].arg4) != 0) {
            FREE(  MODL->brch[ibrch].arg4);
            MALLOC(MODL->brch[ibrch].arg4, char, (int)(strlen(defn)+1));
            strcpy(MODL->brch[ibrch].arg4,                    defn    );
        }
    } else if (iarg == 5) {
        if (strcmp(defn, MODL->brch[ibrch].arg5) != 0) {
            FREE(  MODL->brch[ibrch].arg5);
            MALLOC(MODL->brch[ibrch].arg5, char, (int)(strlen(defn)+1));
            strcpy(MODL->brch[ibrch].arg5,                    defn    );
        }
    } else if (iarg == 6) {
        if (strcmp(defn, MODL->brch[ibrch].arg6) != 0) {
            FREE(  MODL->brch[ibrch].arg6);
            MALLOC(MODL->brch[ibrch].arg6, char, (int)(strlen(defn)+1));
            strcpy(MODL->brch[ibrch].arg6,                    defn    );
        }
    } else if (iarg == 7) {
        if (strcmp(defn, MODL->brch[ibrch].arg7) != 0) {
            FREE(  MODL->brch[ibrch].arg7);
            MALLOC(MODL->brch[ibrch].arg7, char, (int)(strlen(defn)+1));
            strcpy(MODL->brch[ibrch].arg7,                    defn    );
        }
    } else if (iarg == 8) {
        if (strcmp(defn, MODL->brch[ibrch].arg8) != 0) {
            FREE(  MODL->brch[ibrch].arg8);
            MALLOC(MODL->brch[ibrch].arg8, char, (int)(strlen(defn)+1));
            strcpy(MODL->brch[ibrch].arg8,                    defn    );
        }
    } else if (iarg == 9) {
        if (strcmp(defn, MODL->brch[ibrch].arg9) != 0) {
            FREE(  MODL->brch[ibrch].arg9);
            MALLOC(MODL->brch[ibrch].arg9, char, (int)(strlen(defn)+1));
            strcpy(MODL->brch[ibrch].arg9,                    defn    );
        }
    }

    /* mark the Branch as dirty */
    MODL->brch[ibrch].dirty = 1;

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmRetAttr - return an Attribute for a Branch by index            *
 *                                                                      *
 ************************************************************************
 */

int
ocsmRetAttr(void   *modl,               /* (in)  pointer to MODL */
            int    ibrch,               /* (in)  Branch index (1-nbrch) */
            int    iattr,               /* (in)  Attribute index (1-nattr) */
            char   aname[],             /* (out) Attribute name  (at least 129 long) */
            char   avalue[])            /* (out) Attribute value (at least 129 long) */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(ocsmRetAttr);
    DPRINT3("%s(ibrch=%d, iattr=%d) {",
            routine, ibrch, iattr);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ibrch is given */
    if (ibrch < 1 || ibrch > MODL->nbrch) {
        status = OCSM_ILLEGAL_BRCH_INDEX;
        goto cleanup;
    }

    /* check that value iattr is given */
    if (iattr < 1 || iattr > MODL->brch[ibrch].nattr) {
        status = OCSM_ILLEGAL_ARG_INDEX;
        goto cleanup;
    }

    /* return the name and value */
    STRNCPY(aname,  MODL->brch[ibrch].attr[iattr-1].name , 129);
    STRNCPY(avalue, MODL->brch[ibrch].attr[iattr-1].value, 129);

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmGetAttr - get an Attribute for a Branch by name                *
 *                                                                      *
 ************************************************************************
 */

int
ocsmGetAttr(void   *modl,               /* (in)  pointer to MODL */
            int    ibrch,               /* (in)  Branch index (1-nbrch) */
            char   aname[],             /* (in)  Attribute name */
            char   avalue[])            /* (out) Attribute value (at least 129 long) */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       iattr, jattr;

    ROUTINE(ocsmGetAttr);
    DPRINT3("%s(ibrch=%d, aname=%s) {",
            routine, ibrch, aname);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ibrch is given */
    if (ibrch < 0 || ibrch > MODL->nbrch) {
        status = OCSM_ILLEGAL_BRCH_INDEX;
        goto cleanup;
    }

    /* global Attribute */
    if (ibrch == 0) {

        /* determine the Attribute index (if any) */
        iattr = -1;
        for (jattr = 0; jattr < MODL->nattr; jattr++) {
            if (strcmp(MODL->attr[jattr].name, aname) == 0) {
                iattr = jattr;
                break;
            }
        }

        /* error if Attribute name is not found */
        if (iattr < 0) {
            status = OCSM_NAME_NOT_FOUND;
            goto cleanup;
        }

        /* return the value */
        STRNCPY(avalue, MODL->attr[iattr].value, 129);

    /* Attribute for Branch */
    } else {

        /* determine the Attribute index (if any) */
        iattr = -1;
        for (jattr = 0; jattr < MODL->brch[ibrch].nattr; jattr++) {
            if (strcmp(MODL->brch[ibrch].attr[jattr].name, aname) == 0) {
                iattr = jattr;
                break;
            }
        }

        /* error if Attribute name is not found */
        if (iattr < 0) {
            status = OCSM_NAME_NOT_FOUND;
            goto cleanup;
        }

        /* return the value */
        STRNCPY(avalue, MODL->brch[ibrch].attr[iattr].value, 129);
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmSetAttr - set an Attribute for a Branch                        *
 *                                                                      *
 ************************************************************************
 */

int
ocsmSetAttr(void   *modl,               /* (in)  pointer to MODL */
            int    ibrch,               /* (in)  Branch index (1-nbrch) */
            char   aname[],             /* (in)  Attribute name */
            char   avalue[])            /* (in)  Attribute value */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       iattr, jattr;

    ROUTINE(ocsmSetAttr);
    DPRINT4("%s(ibrch=%d, aname=%s, avalue=%s) {",
            routine, ibrch, aname, avalue);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ibrch is given */
    if (ibrch < 0 || ibrch > MODL->nbrch) {
        status = OCSM_ILLEGAL_BRCH_INDEX;
        goto cleanup;
    }

    /* global Attribute */
    if (ibrch == 0) {

        /* determine the Attribute index (if any) */
        iattr = -1;
        for (jattr = 0; jattr < MODL->nattr; jattr++) {
            if (strcmp(MODL->attr[jattr].name, aname) == 0) {
                iattr = jattr;
                break;
            }
        }

        /* if attribute is not found already, create a new one */
        if (iattr < 0) {
            MODL->nattr++;

            if (MODL->attr == NULL) {
                MALLOC(MODL->attr, attr_T, MODL->nattr);
            } else {
                RALLOC(MODL->attr, attr_T, MODL->nattr);
            }

            iattr = MODL->nattr - 1;
        }

        /* set the Attribute's name and value */
        MALLOC(MODL->attr[iattr].name,  char, (int)(strlen(aname )+1));
        strcpy(MODL->attr[iattr].name,                     aname     );
        MALLOC(MODL->attr[iattr].value, char, (int)(strlen(avalue)+1));
        strcpy(MODL->attr[iattr].value,                    avalue    );

    /* Attribute for Branch */
    } else {

        /* determine the Attribute index (if any) */
        iattr = -1;
        for (jattr = 0; jattr < MODL->brch[ibrch].nattr; jattr++) {
            if (strcmp(MODL->brch[ibrch].attr[jattr].name, aname) == 0) {
                iattr = jattr;
                break;
            }
        }

        /* if attribute is not found already, create a new one */
        if (iattr < 0) {
            MODL->brch[ibrch].nattr++;

            if (MODL->brch[ibrch].attr == NULL) {
                MALLOC(MODL->brch[ibrch].attr, attr_T, MODL->brch[ibrch].nattr);
            } else {
                RALLOC(MODL->brch[ibrch].attr, attr_T, MODL->brch[ibrch].nattr);
            }

            iattr = MODL->brch[ibrch].nattr - 1;
        }

        /* set the Attribute's name and value */
        MALLOC(MODL->brch[ibrch].attr[iattr].name,  char, (int)(strlen(aname )+1));
        strcpy(MODL->brch[ibrch].attr[iattr].name,                     aname     );
        MALLOC(MODL->brch[ibrch].attr[iattr].value, char, (int)(strlen(avalue)+1));
        strcpy(MODL->brch[ibrch].attr[iattr].value,                    avalue    );

        /* mark the Branch as dirty */
        MODL->brch[ibrch].dirty = 1;
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmPrintAttrs - print global Attributes to file                   *
 *                                                                      *
 ************************************************************************
 */

int
ocsmPrintAttrs(void   *modl,            /* (in)  pointer to MODL */
               FILE   *fp)              /* (in)  pointer to FILE */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       iattr;

    ROUTINE(ocsmPrintAttrs);
    DPRINT2("%s(fp=%lx) {",
            routine, (long)fp);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    for (iattr = 0; iattr < MODL->nattr; iattr++) {
        fprintf(fp, "    %-24s -> %24s\n",
                MODL->attr[iattr].name,
                MODL->attr[iattr].value);
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmGetName - get the name of a Branch                             *
 *                                                                      *
 ************************************************************************
 */
int
ocsmGetName(void   *modl,               /* (in)  pointer to MODL */
            int    ibrch,               /* (in)  Branch index (1-nbrch) */
            char   name[])              /* (out) Branch name (at least 129 long) */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(ocsmGetName);
    DPRINT2("%s(ibrch=%d) {",
            routine, ibrch);

    /* --------------------------------------------------------------- */

    /* default return values */
    name[0] = '\0';

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ibrch is given */
    if (ibrch < 1 || ibrch > MODL->nbrch) {
        status = OCSM_ILLEGAL_BRCH_INDEX;
        goto cleanup;
    }

    /* return the name */
    STRNCPY(name, MODL->brch[ibrch].name, 129);

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmSetName - set the name for a Branch                            *
 *                                                                      *
 ************************************************************************
 */
int
ocsmSetName(void   *modl,               /* (in)  pointer to MODL */
            int    ibrch,               /* (in)  Branch index (1-nbrch) */
            char   name[])              /* (in)  Branch name */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       jbrch;

    ROUTINE(ocsmSetName);
    DPRINT3("%s(ibrch=%d, name=%s) {",
            routine, ibrch, name);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ibrch is given */
    if (ibrch < 1 || ibrch > MODL->nbrch) {
        status = OCSM_ILLEGAL_BRCH_INDEX;
        goto cleanup;
    }

    /* make sure that the new name is unique */
    for (jbrch = 1; jbrch <= MODL->nbrch; jbrch++) {
        if (strcmp(name, MODL->brch[jbrch].name) == 0) {
            status = OCSM_NAME_NOT_UNIQUE;
            goto cleanup;
        } else if (strncmp(name, "Brch_", 5) == 0) {
            status = OCSM_NAME_NOT_UNIQUE;
            goto cleanup;
        }
    }

    /* name cannot contain ">" */
    if (strstr(name, ">") != NULL) {
        status = OCSM_ILLEGAL_VALUE;
        goto cleanup;
    }

    /* set the name */
    FREE(MODL->brch[ibrch].name);

    MALLOC(MODL->brch[ibrch].name, char, (int)(strlen(name)+1));
    strcpy(MODL->brch[ibrch].name,                    name    );

    /* mark the Branch as dirty */
    MODL->brch[ibrch].dirty = 1;

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmGetSketch - get string data associated with a Sketch           *
 *                                                                      *
 ************************************************************************
 */

int
ocsmGetSketch(void   *modl,             /* (in)  pointer to MODL */
              int    ibrch,             /* (in)  Branch indes (1-nbrch) within Sketch */
              char   begs[],            /* (out) string (1024 long) with SKBEG variables */
              char   vars[],            /* (out) string (1024 long) with Sketch variables */
              char   cons[],            /* (out) string (1024 long) with Sketch constraints */
              char   segs[])            /* (out) string (1024 long) with Sketch segments */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       jbrch, iskbeg, iskend, index1, index2, ibeg, iend;
    double    value, xval, yval, zval, dot;
    char      token[1024];

    ROUTINE(ocsmGetSketch);
    DPRINT2("%s(ibrch=%d) {",
            routine, ibrch);

    /* --------------------------------------------------------------- */

    MODL->errorBuf[0] = '\0';

    /* default returns */
    begs[0] = '\0';
    vars[0] = '\0';
    cons[0] = '\0';
    segs[0] = '\0';

    if        (MODL->brch[ibrch].type == OCSM_SKBEG) {
        iskbeg = ibrch;
        iskend = ibrch + 1;
    } else if (MODL->brch[ibrch].type == OCSM_SKEND) {
        iskbeg = ibrch - 1;
        iskend = ibrch;
    } else {
        iskbeg = ibrch;
        iskend = ibrch;
    }

    /* find SKBEG at or before ibrch */
    while (iskbeg > 0 && MODL->brch[iskbeg].type != OCSM_SKBEG) {
        if (MODL->brch[iskbeg].type == OCSM_SKEND) {
            signalError(MODL, OCSM_ILLEGAL_BRCH_INDEX, ibrch,
                        "ibrch=%d is not within a Sketch (SKEND found first)", ibrch);
            SET_STATUS(OCSM_ILLEGAL_BRCH_INDEX, ocsmGetSketch);
        }
        iskbeg--;
        if (iskbeg <= 0) {
            signalError(MODL, OCSM_ILLEGAL_BRCH_INDEX, ibrch,
                        "ibrch=%d is not within a Sketch (at beginning)", ibrch);
            SET_STATUS(OCSM_ILLEGAL_BRCH_INDEX, ocsmGetSketch);
        }
    }

    /* find SKEND at or after ibrch */
    while (iskend <= MODL->nbrch && MODL->brch[iskend].type != OCSM_SKEND) {
        if (MODL->brch[iskend].type == OCSM_SKBEG) {
            signalError(MODL, OCSM_ILLEGAL_BRCH_INDEX, ibrch,
                        "ibrch=%d is not within a Sketch (SKBEG found first)", ibrch);
            SET_STATUS(OCSM_ILLEGAL_BRCH_INDEX, ocsmGetSketch);
        }
        iskend++;
        if (iskend <= 0) {
            signalError(MODL, OCSM_ILLEGAL_BRCH_INDEX, ibrch,
                        "ibrch=%d is not within a Sketch (at end)", ibrch);
            SET_STATUS(OCSM_ILLEGAL_BRCH_INDEX, ocsmGetSketch);
        }
    }

    /* build the begs string */
    status = str2val(MODL->brch[iskbeg].arg1, MODL, &xval, &dot);
    CHECK_STATUS(str2val);

    status = str2val(MODL->brch[iskbeg].arg2, MODL, &yval, &dot);
    CHECK_STATUS(str2val);

    status = str2val(MODL->brch[iskbeg].arg3, MODL, &zval, &dot);
    CHECK_STATUS(str2val);

    snprintf(begs, 1024, "%s;%f;%s;%f;%s;%f;",
             MODL->brch[iskbeg].arg1, xval,
             MODL->brch[iskbeg].arg2, yval,
             MODL->brch[iskbeg].arg3, zval);

    /* if there are no statements within the Sketch, return an empty Sketch */
    if (iskend == iskbeg+1) {
        goto cleanup;
    }

    /* make sure that there there are only valid statements in the Sketch */
    for (jbrch = iskbeg+1; jbrch < iskend; jbrch++) {
        if (MODL->brch[jbrch].type != OCSM_SKVAR  &&
            MODL->brch[jbrch].type != OCSM_SKCON  &&
            MODL->brch[jbrch].type != OCSM_LINSEG &&
            MODL->brch[jbrch].type != OCSM_ARC    &&
            MODL->brch[jbrch].type != OCSM_SPLINE &&
            MODL->brch[jbrch].type != OCSM_BEZIER   ) {
            signalError(MODL, OCSM_ILLEGAL_STATEMENT, jbrch,
                        "wrong type (%s) in Sketch", ocsmGetText(MODL->brch[jbrch].type));
            SET_STATUS(OCSM_ILLEGAL_STATEMENT, ocsmGetSketch);
        }
    }

    /* make sure that a SKVAR follows the SKBEG */
    if (MODL->brch[iskbeg+1].type != OCSM_SKVAR) {
        signalError(MODL, OCSM_ILLEGAL_STATEMENT, ibrch,
                    "SKVAR does not follow SKBEG");
        SET_STATUS(OCSM_ILLEGAL_STATEMENT, ocsmGetSketch);
    }

    /* create the Sketch variables from the SKVAR statement */
    strncpy(vars, MODL->brch[iskbeg+1].arg2, 1024);

    /* create the Sketch constraints from the SKCON statements */
    jbrch = iskbeg + 2;
    while (jbrch < iskend && MODL->brch[jbrch].type == OCSM_SKCON) {
        status = str2val(MODL->brch[jbrch].arg2, MODL, &value, &dot);
        CHECK_STATUS(str2val);
        index1 = NINT(value);

        status = str2val(MODL->brch[jbrch].arg3, MODL, &value, &dot);
        CHECK_STATUS(str2val);
        index2 = NINT(value);

        snprintf(token, 1024, "%s;%d;%d;%s;",
                 &(MODL->brch[jbrch].arg1[1]), index1, index2, &(MODL->brch[jbrch].arg4[1]));

        strncat(cons, token, 1024);

        jbrch++;
    }

    /* create the Sketch segments from the LINSEG, ARC, SPLINE, and BEZIER statements */
    ibeg = 1;
    while (jbrch < iskend) {
        if (strncmp(MODL->brch[jbrch].arg1, "::x[", 4) != 0 ||
            strncmp(MODL->brch[jbrch].arg2, "::y[", 4) != 0 ||
            strncmp(MODL->brch[jbrch].arg3, "::z[", 4) != 0   ) {
            SET_STATUS(OCSM_ILLEGAL_STATEMENT, ocsmGetSketch);
        }

        if (strcmp(MODL->brch[jbrch].arg1, "::x[1]") == 0) {
            iend = 1;
        } else {
            iend = ibeg + 1;
        }

        if        (MODL->brch[jbrch].type == OCSM_LINSEG) {
            snprintf(token, 1024, "L;%d;%d;", ibeg, iend);
        } else if (MODL->brch[jbrch].type == OCSM_ARC) {
            snprintf(token, 1024, "C;%d;%d;", ibeg, iend);
        } else if (MODL->brch[jbrch].type == OCSM_SPLINE) {
            snprintf(token, 1024, "S;%d;%d;", ibeg, iend);
        } else if (MODL->brch[jbrch].type == OCSM_BEZIER) {
            snprintf(token, 1024, "B;%d;%d;", ibeg, iend);
        }

        strncat(segs, token, 1024);

        ibeg = iend;
        jbrch++;
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmSolveSketch - solve for new Sketch variables                   *
 *                                                                      *
 ************************************************************************
 */

int
ocsmSolveSketch(void   *modl,           /* (in)  pointer to MODL */
                char   vars_in[],       /* (in)  string with Sketch variables */
                char   cons[],          /* (in)  string with Sketch constraints */
                char   vars_out[])      /* (out) string (1024 long) with new Sketch variables */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;
    sket_T    *sket=NULL;

    int       i, im1, ip1, icount, lentok, index1, index2, ivar, nvar, icon, jpmtr, jndex;
    int       ii, jj, ibeg, iend;
    double    val;
    char      token[MAX_EXPR_LEN], type[MAX_EXPR_LEN], value[MAX_EXPR_LEN];
    char      prefix[MAX_EXPR_LEN], suffix[MAX_EXPR_LEN];

    ROUTINE(ocsmSolveSketch);
    DPRINT1("%s() {",
            routine);

    /* --------------------------------------------------------------- */

    MALLOC(sket, sket_T, 1);

    /* default (empty) return */
    vars_out[0] = '\0';

    /* initialize the Sketch */
    sket->type   = 0;
    sket->size   = 0;
    sket->solved = 0;
    sket->nseg   = 0;
    sket->nvar   = 0;
    sket->ncon   = 0;

    /* count the semicolons in vars_in to detemine number of variables */
    nvar = 0;
    for (i = 0; i < strlen(vars_in); i++) {
        if (vars_in[i] == ';') nvar++;
    }

    if (nvar%3 == 0) {
        nvar /= 3;
        sket->size = nvar;
    } else {
        signalError(MODL, OCSM_ILLEGAL_VALUE, 0,
                    "vars_in contains %d semicolons (which is not evenly divisible by 3)", nvar);
        status = OCSM_ILLEGAL_VALUE;
        goto cleanup;
    }

    /* make the new Sketch variables */
    status = ocsmNewPmtr(MODL, "::x", OCSM_INTERNAL, 1, sket->size);
    CHECK_STATUS(ocsmNewPmtr);
    sket->ix = MODL->npmtr;

    status = ocsmNewPmtr(MODL, "::y", OCSM_INTERNAL, 1, sket->size);
    CHECK_STATUS(ocsmNewPmtr);
    sket->iy = MODL->npmtr;

    status = ocsmNewPmtr(MODL, "::z", OCSM_INTERNAL, 1, sket->size);
    CHECK_STATUS(ocsmNewPmtr);
    sket->iz = MODL->npmtr;

    status = ocsmNewPmtr(MODL, "::d", OCSM_INTERNAL, 1, sket->size);
    CHECK_STATUS(ocsmNewPmtr);
    sket->id = MODL->npmtr;

    /* pull the variables out of vars_in */
    icount = 0;
    lentok = 0;
    for (i = 0; i < strlen(vars_in); i++) {
        if (vars_in[i] != ';') {
            token[lentok++] = vars_in[i];
            token[lentok  ] = '\0';
        } else {
            val = strtod(token, NULL);
            if        (icount%3 == 0) {
                MODL->pmtr[sket->ix].value[icount/3] = val;
                MODL->pmtr[sket->ix].dot[  icount/3] = 0;
            } else if (icount%3 == 1) {
                MODL->pmtr[sket->iy].value[icount/3] = val;
                MODL->pmtr[sket->iy].dot[  icount/3] = 0;
            } else {
                MODL->pmtr[sket->id].value[icount/3] = val;
                MODL->pmtr[sket->id].dot[  icount/3] = 0;
            }

            icount++;
            lentok = 0;
        }
    }

    /* make a list of the Sketch variables */
    sket->nvar = 0;
    for (i = 1; i <= sket->size; i++) {
        sket->ipmtr[sket->nvar] = sket->ix;
        sket->index[sket->nvar] = i-1;
        sket->nvar++;

        sket->ipmtr[sket->nvar] = sket->iy;
        sket->index[sket->nvar] = i-1;
        sket->nvar++;

        if (fabs(MODL->pmtr[sket->id].value[i-1]) > EPS06) {
            sket->ipmtr[sket->nvar] = sket->id;
            sket->index[sket->nvar] = i-1;
            sket->nvar++;
        }
    }

    for (ivar = 0; ivar < sket->nvar; ivar++) {
        jpmtr = sket->ipmtr[ivar];
        jndex = sket->index[ivar];

        SPRINT3(1, "    -> creating  %s[%d] = %10.5f", MODL->pmtr[jpmtr].name, jndex+1,
                MODL->pmtr[jpmtr].value[jndex]);
    }

    /* make the constraints */
    icount = 0;
    lentok = 0;
    index1 = 0;
    index2 = 0;
    for (i = 0; i < strlen(cons); i++) {
        if (cons[i] != ';') {
            token[lentok++] = cons[i];
            token[lentok  ] = '\0';
        } else {
            if        (icount%4 == 0) {
                strcpy(type, token);
            } else if (icount%4 == 1) {
                index1 = strtol(token, NULL, 10);
            } else if (icount%4 == 2) {
                index2 = strtol(token, NULL, 10);
            } else {
                strcpy(value, token);

                /* modify value if it contains "::L[" */
                while (strstr(value, "::L[") != NULL) {
                    prefix[0] = '\0';
                    suffix[0] = '\0';
                    ibeg      =   0 ;
                    for (ii = 0; ii < strlen(value); ii++) {
                        if (value[ii  ] == ':' && value[ii+1] == ':' &&
                            value[ii+2] == 'L' && value[ii+3] == '['   ) {
                            if (ii > 0) {
                                strncpy(prefix, value, ii);
                                prefix[ii] = '\0';
                            }
                            ibeg = strtol(&(value[ii+4]), NULL, 10);
                            break;
                        }
                    }

                    if (ibeg < 1 || ibeg > sket->size) {
                        signalError(MODL, OCSM_ILLEGAL_VALUE, 0,
                                    "::L[%d] is not defined", ibeg);
                        status = OCSM_ILLEGAL_VALUE;
                        goto cleanup;
                    } else if (ibeg < sket->size) {
                        iend = ibeg + 1;
                    } else {
                        iend = 1;
                    }

                    for (jj = ii+3; jj < strlen(value); jj++) {
                        if (value[jj] == ']') {
                            strcpy(suffix, &(value[jj+1]));
                            break;
                        }
                    }

                    snprintf(value, MAX_EXPR_LEN, "%shypot(::y[%d]-::y[%d],::x[%d]-::x[%d])%s",
                             prefix, iend, ibeg, iend, ibeg, suffix);
                }

                if        (strcmp(type, "X") == 0 && index2 <= 0) {
                    snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                             "::x[%d]-(%s)",
                             index1, value);
                } else if (strcmp(type, "X") == 0) {
                    snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                             "Xcent(::x[%d],::y[%d],::d[%d],::x[%d],::y[%d])-(%s)",
                             index1, index1, index2, index2, index2, value);
                } else if (strcmp(type, "Y") == 0 && index2 <= 0) {
                    snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                             "::y[%d]-(%s)",
                             index1, value);
                } else if (strcmp(type, "Y") == 0) {
                    snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                             "Ycent(::x[%d],::y[%d],::d[%d],::x[%d],::y[%d])-(%s)",
                             index1, index1, index2, index2, index2, value);
                } else if (strcmp(type, "P") == 0) {
                    im1 = index1 - 1; if (im1 == 0           ) im1 = sket->size;
                    ip1 = index1 + 1; if (ip1 == sket->size+1) ip1 = 1;
                    snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                             "(abs(turnang(::x[%d],::y[%d],::d[%d],::x[%d],::y[%d],::d[%d],::x[%d],::y[%d]))-90)/57.3",
                             im1, im1, index1, index1, index1, ip1, ip1, ip1);
                } else if (strcmp(type, "T") == 0) {
                    im1 = index1 - 1; if (im1 == 0           ) im1 = sket->size;
                    ip1 = index1 + 1; if (ip1 == sket->size+1) ip1 = 1;
                    snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                             "turnang(::x[%d],::y[%d],::d[%d],::x[%d],::y[%d],::d[%d],::x[%d],::y[%d])/57.3",
                             im1, im1, index1, index1, index1, ip1, ip1, ip1);
                } else if (strcmp(type, "A") == 0) {
                    im1 = index1 - 1; if (im1 == 0           ) im1 = sket->size;
                    ip1 = index1 + 1; if (ip1 == sket->size+1) ip1 = 1;
                    snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                             "(turnang(::x[%d],::y[%d],::d[%d],::x[%d],::y[%d],::d[%d],::x[%d],::y[%d])-(%s))/57.3",
                             im1, im1, index1, index1, index1, ip1, ip1, ip1, value);
                } else if (strcmp(type, "W") == 0) {
                    snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                             "::x[%d]-::x[%d]-(%s)",
                             index2, index1, value);
                } else if (strcmp(type, "D") == 0) {
                    snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                             "::y[%d]-::y[%d]-(%s)",
                             index2, index1, value);
                } else if (strcmp(type, "H") == 0) {
                    snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                             "::y[%d]-::y[%d]",
                             index1, index2);
                } else if (strcmp(type, "V") == 0) {
                    snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                             "::x[%d]-::x[%d]",
                             index1, index2);
                } else if (strcmp(type, "I") == 0) {
                    snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                             "(atan2d(::y[%d]-::y[%d],::x[%d]-::x[%d])-(%s))/57.3",
                             index1, index2, index1, index2, value);
                } else if (strcmp(type, "L") == 0) {
                    snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                             "hypot(::y[%d]-::y[%d],::x[%d]-::x[%d])-(%s)",
                             index1, index2, index1, index2, value);
                } else if (strcmp(type, "R") == 0) {
                    snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                             "radius(::x[%d],::y[%d],::d[%d],::x[%d],::y[%d])-(%s)",
                             index1, index1, index2, index2, index2, value);
                } else if (strcmp(type, "S") == 0) {
                    snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                             "(sweep(::x[%d],::y[%d],::d[%d],::x[%d],::y[%d])-(%s))/57.3",
                             index1, index1, index2, index2, index2, value);
                }
            }

            icount++;
            lentok = 0;
        }
    }

    for (icon = 0; icon < sket->ncon; icon++) {
        SPRINT2(2, "    -> setting con[%3d] = %s", icon, sket->con[icon]);
    }

    /* check that number of constraints matched number of variables */
    if        (sket->nvar > sket->ncon) {
        SPRINT2(0, "WARNING:: under-constrained since nvar=%d but ncon=%d",
                sket->nvar, sket->ncon);
        goto removeTemps;
    } else if (sket->nvar < sket->ncon) {
        SPRINT2(0, "WARNING:: over-constrained since nvar=%d but ncon=%d",
                sket->nvar, sket->ncon);
        goto removeTemps;
    }

    /* solve the sketch */
    status = solveSketch(MODL, sket);

    /* if solver fails, simply return nvar=0 */
    if (status < SUCCESS) {
        signalError(MODL, OCSM_NOT_CONVERGED, 0,
                    "solver did not converge");
        goto removeTemps;
    }

    /* solver succeeded, so write the new variables */
    for (ivar = 0; ivar < nvar; ivar++) {
        snprintf(token, MAX_EXPR_LEN, "%.6lf;%.6lf;%.6lf;",
                 MODL->pmtr[sket->ix].value[ivar],
                 MODL->pmtr[sket->iy].value[ivar],
                 MODL->pmtr[sket->id].value[ivar]);
        strcat(vars_out, token);
    }

removeTemps:
    /* remove ::x[], ::y[], ::z[], and ::d[] if they were generated by a skvar statement */
    if (sket->nvar > 0) {
        if (strcmp(MODL->pmtr[MODL->npmtr  ].name, "::d") != 0 ||
            strcmp(MODL->pmtr[MODL->npmtr-1].name, "::z") != 0 ||
            strcmp(MODL->pmtr[MODL->npmtr-2].name, "::y") != 0 ||
            strcmp(MODL->pmtr[MODL->npmtr-3].name, "::x") != 0   ) {
            signalError(MODL, OCSM_INTERNAL_ERROR, 0,
                        "last 4 Parameters are not ::d, ::z, ::y, and ::x");
            status = OCSM_INTERNAL_ERROR;
            goto cleanup;
        }

        for (i = 0; i < 4; i++) {
            FREE(MODL->pmtr[MODL->npmtr].name );
            FREE(MODL->pmtr[MODL->npmtr].value);
            FREE(MODL->pmtr[MODL->npmtr].dot  );
            FREE(MODL->pmtr[MODL->npmtr].lbnd );
            FREE(MODL->pmtr[MODL->npmtr].ubnd );
            MODL->npmtr--;
        }
    }

cleanup:
    FREE(sket);

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmSaveSketch - overwrite Branches associated with a Sketch       *
 *                                                                      *
 ************************************************************************
 */

int
ocsmSaveSketch(void   *modl,            /* (in)  pointer to MODL */
               int    ibrch,            /* (in)  Branch index (1-nbrch) within Sketch */
               char   vars[],           /* (in)  string with Sketch variables */
               char   cons[],           /* (in)  string with Sketch constraints */
               char   segs[])           /* (in)  string with Sketch segments */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       iskbeg, iskend, jbrch, iafter, icon, ncon, iseg, nseg, i;
    char      *arg1=NULL, *arg2=NULL, *arg3=NULL, *arg4=NULL, *arg5=NULL, *arg6=NULL;

    ROUTINE(ocsmSaveSketch);
    DPRINT2("%s(ibrch=%d) {",
            routine, ibrch);

    /* --------------------------------------------------------------- */

    if (outLevel >= 2) {
        SPRINT0(2, "ocsmSaveSketch: at beginning");
        status = ocsmPrintBrchs(MODL, stdout);
        CHECK_STATUS(ocsmPrintBrchs);
    }

    MALLOC(arg1, char, MAX_EXPR_LEN);
    MALLOC(arg2, char, MAX_EXPR_LEN);
    MALLOC(arg3, char, MAX_EXPR_LEN);
    MALLOC(arg4, char, MAX_EXPR_LEN);
    MALLOC(arg5, char, MAX_EXPR_LEN);
    MALLOC(arg6, char, MAX_EXPR_LEN);

    if        (MODL->brch[ibrch].type == OCSM_SKBEG) {
        iskbeg = ibrch;
        iskend = ibrch + 1;
    } else if (MODL->brch[ibrch].type == OCSM_SKEND) {
        iskbeg = ibrch - 1;
        iskend = ibrch;
    } else {
        iskbeg = ibrch;
        iskend = ibrch;
    }

    /* find SKBEG at or before ibrch */
    while (iskbeg > 0 && MODL->brch[iskbeg].type != OCSM_SKBEG) {
        if (MODL->brch[iskbeg].type == OCSM_SKEND) {
            signalError(MODL, OCSM_ILLEGAL_BRCH_INDEX, ibrch,
                        "ibrch=%d is not within a Sketch (SKEND found first)", ibrch);
            SET_STATUS(OCSM_ILLEGAL_BRCH_INDEX, ocsmGetSketch);
        }
        iskbeg--;
        if (iskbeg <= 0) {
            signalError(MODL, OCSM_ILLEGAL_BRCH_INDEX, ibrch,
                        "ibrch=%d is not within a Sketch (at beginning)", ibrch);
            SET_STATUS(OCSM_ILLEGAL_BRCH_INDEX, ocsmGetSketch);
        }
    }

    /* find SKEND at or after ibrch */
    while (iskend <= MODL->nbrch && MODL->brch[iskend].type != OCSM_SKEND) {
        if (MODL->brch[iskend].type == OCSM_SKBEG) {
            signalError(MODL, OCSM_ILLEGAL_BRCH_INDEX, ibrch,
                        "ibrch=%d is not within a Sketch (SKBEG found first)", ibrch);
            SET_STATUS(OCSM_ILLEGAL_BRCH_INDEX, ocsmGetSketch);
        }
        iskend++;
        if (iskend <= 0) {
            signalError(MODL, OCSM_ILLEGAL_BRCH_INDEX, ibrch,
                        "ibrch=%d is not within a Sketch (at end)", ibrch);
            SET_STATUS(OCSM_ILLEGAL_BRCH_INDEX, ocsmGetSketch);
        }
    }

    iafter = iskbeg;

    /* determine the number of constraints */
    ncon = 0;
    for (i = 0; i < strlen(cons); i++) {
        if (cons[i] == ';') ncon++;
    }
    ncon /= 4;

    /* determine the number of segments */
    nseg = 0;
    for (i = 0; i < strlen(segs); i++) {
        if (segs[i] == ';') nseg++;
    }
    nseg /= 3;

    /* delete the Branches between the SKBEG and the SKEND */
    for (jbrch = iskend-1; jbrch > iskbeg; jbrch--) {
        status = ocsmDelBrch(MODL, jbrch);
        CHECK_STATUS(ocsmDelBrch);
    }

    /* update the arguments in the SKBEG Branch */
//$$$    getToken(vars, 0, ';', arg1);
//$$$    getToken(vars, 1, ';', arg2);
//$$$
//$$$    RALLOC(MODL->brch[iskbeg].arg1, char, strlen(arg1)+1);
//$$$    strcpy( MODL->brch[iskbeg].arg1, arg1);
//$$$
//$$$    RALLOC(MODL->brch[iskbeg].arg2, char, strlen(arg2)+1);
//$$$    strcpy( MODL->brch[iskbeg].arg2, arg1);

    /* add the SKVAR Branch */
    status = ocsmNewBrch(MODL, iafter++, OCSM_SKVAR, "$xy", vars, NULL, NULL, NULL,
                                                            NULL, NULL, NULL, NULL);
    CHECK_STATUS(ocsmNewBrch);

    /* add the SKCON Branches */
    for (icon = 0; icon < ncon; icon++) {
        arg1[0] = '$';
        arg4[0] = '$';
        getToken(cons, 4*icon  , ';', &(arg1[1]));
        getToken(cons, 4*icon+1, ';',  arg2    );
        getToken(cons, 4*icon+2, ';',  arg3    );
        getToken(cons, 4*icon+3, ';', &(arg4[1]));

        status = ocsmNewBrch(MODL, iafter++, OCSM_SKCON, arg1, arg2, arg3, arg4, NULL,
                                                               NULL, NULL, NULL, NULL);
        CHECK_STATUS(ocsmNewBrch);
    }

    /* add the LINSEG, ARC, SPLINE, and BEZIER Branches */
    for (iseg = 0; iseg < nseg; iseg++) {
        getToken(segs, 3*iseg,   ';', arg1);
        getToken(segs, 3*iseg+2, ';', arg6);

        snprintf(arg2, MAX_EXPR_LEN, "::x[%s]", arg6);
        snprintf(arg3, MAX_EXPR_LEN, "::y[%s]", arg6);
        snprintf(arg4, MAX_EXPR_LEN, "::z[%s]", arg6);
        snprintf(arg5, MAX_EXPR_LEN, "::d[%s]", arg6);

        if        (arg1[0] == 'L') {
            status = ocsmNewBrch(MODL, iafter++, OCSM_LINSEG, arg2, arg3, arg4, NULL, NULL,
                                                                    NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);
        } else if (arg1[0] == 'C') {
            strcpy(arg6, "$xy");
            status = ocsmNewBrch(MODL, iafter++, OCSM_ARC,    arg2, arg3, arg4, arg5, arg6,
                                                                    NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);
        } else if (arg1[0] == 'S') {
            status = ocsmNewBrch(MODL, iafter++, OCSM_SPLINE, arg2, arg3, arg4, NULL, NULL,
                                                                    NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);
        } else if (arg1[0] == 'B') {
            status = ocsmNewBrch(MODL, iafter++, OCSM_BEZIER, arg2, arg3, arg4, NULL, NULL,
                                                                    NULL, NULL, NULL, NULL);
            CHECK_STATUS(ocsmNewBrch);
        }
    }

    if (outLevel >= 2) {
        SPRINT0(2, "ocsmSaveSketch: after additions");
        status = ocsmPrintBrchs(MODL, stdout);
        CHECK_STATUS(ocsmPrintBrchs);
    }

cleanup:
    FREE(arg6);
    FREE(arg5);
    FREE(arg4);
    FREE(arg3);
    FREE(arg2);
    FREE(arg1);

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmNewPmtr - create a new Parameter                               *
 *                                                                      *
 ************************************************************************
 */

int
ocsmNewPmtr(void   *modl,               /* (in)  pointer to MODL */
            char   name[],              /* (in)  Parameter name */
            int    type,                /* (in)  Parameter type */
            int    nrow,                /* (in)  number of rows */
            int    ncol)                /* (in)  number of columns */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       ipmtr, i, irow, icol, indx;

    ROUTINE(ocsmNewPmtr);
    DPRINT5("%s(name=%s, type=%d, nrow=%d, ncol=%d) {",
            routine, name, type, nrow, ncol);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that pmtrName is not already defined at this level */
    for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
        if (strcmp(MODL->pmtr[ipmtr].name, name) == 0    &&
                   MODL->pmtr[ipmtr].level == MODL->level  ) {
            status = OCSM_NAME_ALREADY_DEFINED;
            goto cleanup;
        }
    }

    /* check that pmtrName is a valid name */
    if (strlen(name) == 0 || strlen(name) >= MAX_NAME_LEN) {
        status = OCSM_ILLEGAL_PMTR_NAME;
        goto cleanup;
    } else if (name[0] == '@') {
    } else if (isalpha(name[0]) == 0 && name[0] != ':') {
        status = OCSM_ILLEGAL_PMTR_NAME;
        goto cleanup;
    } else {
        for (i = 1; i < strlen(name); i++) {
            if        (isalpha(name[i]) != 0) {
            } else if (isdigit(name[i]) != 0) {
            } else if (name[i] == '_') {
            } else if (name[i] == '@') {
            } else if (name[i] == ':') {
            } else {
                status = OCSM_ILLEGAL_PMTR_NAME;
                goto cleanup;
            }
        }
    }

    /* check for a valid Parameter type */
    if (type != OCSM_EXTERNAL && type != OCSM_INTERNAL) {
        status = OCSM_ILLEGAL_TYPE;
        goto cleanup;
    }

    /* extend the Parameter list (if needed) */
    if (MODL->npmtr >= MODL->mpmtr) {
        MODL->mpmtr += 25;
        if (MODL->pmtr == NULL) {
            MALLOC(MODL->pmtr, pmtr_T, MODL->mpmtr+1);
        } else {
            RALLOC(MODL->pmtr, pmtr_T, MODL->mpmtr+1);
        }
    }

    /* create the new Parameter and store its name and defn */
    MODL->npmtr += 1;

    ipmtr = MODL->npmtr;

    MALLOC( MODL->pmtr[ipmtr].name, char, (int)(strlen(name)+1));
    STRNCPY(MODL->pmtr[ipmtr].name, name,       strlen(name)+1 );

    MALLOC( MODL->pmtr[ipmtr].value, double, nrow*ncol);
    MALLOC( MODL->pmtr[ipmtr].dot,   double, nrow*ncol);
    MALLOC( MODL->pmtr[ipmtr].lbnd,  double, nrow*ncol);
    MALLOC( MODL->pmtr[ipmtr].ubnd,  double, nrow*ncol);

    MODL->pmtr[ipmtr].level = MODL->level;
    MODL->pmtr[ipmtr].type  = type;
    MODL->pmtr[ipmtr].nrow  = nrow;
    MODL->pmtr[ipmtr].ncol  = ncol;

    /* set default values, lower and upper Bounds */
    indx = 0;
    for (irow = 0; irow < nrow; irow++) {
        for (icol = 0; icol < ncol; icol++) {
            MODL->pmtr[ipmtr].value[indx] = -HUGEQ;
            MODL->pmtr[ipmtr].dot[  indx] = 0;
            MODL->pmtr[ipmtr].lbnd[ indx] = -HUGEQ;
            MODL->pmtr[ipmtr].ubnd[ indx] = +HUGEQ;
            indx++;
        }
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmFindPmtr - find (or create) a Parameter                        *
 *                                                                      *
 ************************************************************************
 */

int
ocsmFindPmtr(void   *modl,              /* (in)  pointer to MODL */
             char   name[],             /* (in)  Parameter name */
             int    type,               /* (in)  Parameter type */
             int    nrow,               /* (in)  number of rows */
             int    ncol,               /* (in)  number of columns */
             int    *ipmtr)             /* (out) Parameter index */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       jpmtr;

    ROUTINE(ocsmFindPmtr);
    DPRINT5("%s(name=%s, type=%d, nrow=%d, ncol=%d) {",
            routine, name, type, nrow, ncol);

    /* --------------------------------------------------------------- */

    *ipmtr = 0;

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check if name is already defined at this level */
    for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
        if (strcmp(MODL->pmtr[jpmtr].name, name) == 0    &&
                   MODL->pmtr[jpmtr].level == MODL->level  ) {
            *ipmtr = jpmtr;
            goto cleanup;
        }
    }

    /* if we got here, create a new Parameter */
    status = ocsmNewPmtr(MODL, name, type, nrow, ncol);
    CHECK_STATUS(ocsmNewPmtr);

    *ipmtr = MODL->npmtr;

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmGetPmtr - get info about a Parameter                           *
 *                                                                      *
 ************************************************************************
 */

int
ocsmGetPmtr(void   *modl,               /* (in)  pointer to MODL */
            int    ipmtr,               /* (in)  Parameter index (1-npmtr) */
            int    *type,               /* (out) Parameter type */
            int    *nrow,               /* (out) number of rows */
            int    *ncol,               /* (out) number of columns */
            char   name[])              /* (out) Parameter name (at least 33 long) */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(ocsmGetPmtr);
    DPRINT2("%s(ipmtr=%d) {",
            routine, ipmtr);

    /* --------------------------------------------------------------- */

    /* default return values */
    name[0] = '\0';
    *type   = 0;
    *nrow   = 0;
    *ncol   = 0;

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ipmtr is given */
    if (ipmtr < 1 || ipmtr > MODL->npmtr) {
        status = OCSM_ILLEGAL_PMTR_INDEX;
        goto cleanup;
    } else if (MODL->pmtr[ipmtr].level != MODL->level) {
        status = OCSM_ILLEGAL_PMTR_INDEX;
        goto cleanup;
    }

    /* return the name, type, and size */
    STRNCPY(name, MODL->pmtr[ipmtr].name, 32);
    *type  =      MODL->pmtr[ipmtr].type;
    *nrow  =      MODL->pmtr[ipmtr].nrow;
    *ncol  =      MODL->pmtr[ipmtr].ncol;

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmPrintPmtrs - print external Parameters to file                 *
 *                                                                      *
 ************************************************************************
 */

int
ocsmPrintPmtrs(void   *modl,            /* (in)  pointer to MODL */
               FILE   *fp)              /* (in)  pointer to FILE */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       ipmtr, maxlen, irow, icol, index, count;

    ROUTINE(ocsmPrintPmtrs);
    DPRINT2("%s(fp=%lx) {",
            routine, (long)fp);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* find maximum name length */
    maxlen = 0;
    for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
        if (strlen(MODL->pmtr[ipmtr].name) > maxlen) {
            maxlen = strlen(MODL->pmtr[ipmtr].name);
        }
    }

    /* loop through all external Parameters and print its name, value, and defn */
    count = 0;
    for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
        if (MODL->pmtr[ipmtr].type  == OCSM_EXTERNAL &&
            MODL->pmtr[ipmtr].level == MODL->level     ) {
            index = 0;
            for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                    fprintf(fp, "    %5d [e]  %-*s  [%3d,%3d] %11.5f %11.5f    >>",
                            ipmtr, maxlen, MODL->pmtr[ipmtr].name,
                            irow, icol, MODL->pmtr[ipmtr].value[index],
                                        MODL->pmtr[ipmtr].dot[  index]);

                    if (MODL->pmtr[ipmtr].lbnd[index] <= -HUGEQ) {
                        fprintf(fp, " unbounded  << ");
                    } else {
                        fprintf(fp, "%11.5f << ", MODL->pmtr[ipmtr].lbnd[ index]);
                    }

                    if (MODL->pmtr[ipmtr].ubnd[index] >= +HUGEQ) {
                        fprintf(fp, " unbounded \n");
                    } else {
                        fprintf(fp, "%11.5f\n",   MODL->pmtr[ipmtr].ubnd[ index]);
                    }

                    index++;
                }
            }
            count++;
        }
    }

    /* return message if no Paramaters */
    if (count <= 0) {
        fprintf(fp, "    --none--\n");
        goto cleanup;
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmGetValu - get the Value of a Parameter                         *
 *                                                                      *
 ************************************************************************
 */

int
ocsmGetValu(void   *modl,               /* (in)  pointer to MODL */
            int    ipmtr,               /* (in)  Parameter index (1-npmtr) */
            int    irow,                /* (in)  row    index (1-nrow) */
            int    icol,                /* (in)  column index (1-ncol) */
            double *value,              /* (out) Parameter value */
            double *dot)                /* (out) Parameter velocity */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       index;

    ROUTINE(ocsmGetValu);
    DPRINT4("%s(ipmtr=%d, irow=%d, icol=%d) {",
            routine, ipmtr, irow, icol);

    /* --------------------------------------------------------------- */

    /* default return values */
    *value = 0;

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ipmtr, irow, and icol are given */
    if (ipmtr < 1 || ipmtr > MODL->npmtr) {
        status = OCSM_ILLEGAL_PMTR_INDEX;
        goto cleanup;
    } else if (MODL->pmtr[ipmtr].level != MODL->level) {
        status = OCSM_ILLEGAL_PMTR_INDEX;
        goto cleanup;
    } else if (icol == 0) {
        if (irow < 1 || irow > (MODL->pmtr[ipmtr].nrow)*(MODL->pmtr[ipmtr].ncol)) {
            status = OCSM_ILLEGAL_PMTR_INDEX;
            goto cleanup;
        }

        index = irow - 1;
    } else {
        if (irow < 1 || irow > MODL->pmtr[ipmtr].nrow) {
            status = OCSM_ILLEGAL_PMTR_INDEX;
            goto cleanup;
        } else if (icol < 1 || icol > MODL->pmtr[ipmtr].ncol) {
            status = OCSM_ILLEGAL_PMTR_INDEX;
            goto cleanup;
        }

        index = (icol-1) + (irow-1) * (MODL->pmtr[ipmtr].ncol);
    }

    /* return pertinent information */
    *value = MODL->pmtr[ipmtr].value[index];
    *dot   = MODL->pmtr[ipmtr].dot[  index];

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmSetValu - set a Value for a Parameter                          *
 *                                                                      *
 ************************************************************************
 */

int
ocsmSetValu(void   *modl,               /* (in)  pointer to MODL */
            int    ipmtr,               /* (in)  Parameter index (1-npmtr) */
            int    irow,                /* (in)  row    index (1-nrow) */
            int    icol,                /* (in)  column index (1-nrow) */
            char   defn[])              /* (in)  definition of Value */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       index;
    double    value, dot;

    ROUTINE(ocsmSetValu);
    DPRINT5("%s(ipmtr%d, irow=%d, icol=%d, defn=%s) {",
            routine, ipmtr, irow, icol, defn);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ipmtr, irow, and icol are given */
    if (ipmtr < 1 || ipmtr > MODL->npmtr) {
        status = OCSM_ILLEGAL_PMTR_INDEX;
        goto cleanup;
    } else if (MODL->pmtr[ipmtr].level != MODL->level) {
        status = OCSM_ILLEGAL_PMTR_INDEX;
        goto cleanup;
    } else if (icol == 0) {
        if (irow < 1 || irow > (MODL->pmtr[ipmtr].nrow)*(MODL->pmtr[ipmtr].ncol)) {
            status = OCSM_ILLEGAL_PMTR_INDEX;
            goto cleanup;
        }

        index = irow - 1;
    } else {
        if (irow < 1 || irow > MODL->pmtr[ipmtr].nrow) {
            status = OCSM_ILLEGAL_PMTR_INDEX;
            goto cleanup;
        } else if (icol < 1 || icol > MODL->pmtr[ipmtr].ncol) {
            status = OCSM_ILLEGAL_PMTR_INDEX;
            goto cleanup;
        }

        index = (icol-1) + (irow-1) * (MODL->pmtr[ipmtr].ncol);
    }

    /* if this has a perturbed Body, free it first */
    status = removePerturbation(MODL);
    CHECK_STATUS(removePerturbation);

    /* evaluate the definition */
    if (MODL->pmtr[ipmtr].type == OCSM_EXTERNAL) {
        status = str2val(defn, NULL, &value, &dot);
        CHECK_STATUS(str2val);
    } else {
        status = str2val(defn, MODL, &value, &dot);
        CHECK_STATUS(str2val);
    }

    /* set pertinent information */
    MODL->pmtr[ipmtr].value[index] = value;
    MODL->pmtr[ipmtr].dot[  index] = 0;

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmSetValuD - set a (double) Value for a Parameter                *
 *                                                                      *
 ************************************************************************
 */

int
ocsmSetValuD(void   *modl,              /* (in)  pointer to MODL */
             int    ipmtr,              /* (in)  Parameter index (1-npmtr) */
             int    irow,               /* (in)  row    index (1-nrow) */
             int    icol,               /* (in)  column index (1-nrow) */
             double value)              /* (in)  vlaue to set */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       index;

    ROUTINE(ocsmSetValuD);
    DPRINT5("%s(ipmtr%d, irow=%d, icol=%d, value=%f) {",
            routine, ipmtr, irow, icol, value);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ipmtr, irow, and icol are given */
    if (ipmtr < 1 || ipmtr > MODL->npmtr) {
        status = OCSM_ILLEGAL_PMTR_INDEX;
        goto cleanup;
    } else if (MODL->pmtr[ipmtr].level != MODL->level) {
        status = OCSM_ILLEGAL_PMTR_INDEX;
        goto cleanup;
    } else if (icol == 0) {
        if (irow < 1 || irow > (MODL->pmtr[ipmtr].nrow)*(MODL->pmtr[ipmtr].ncol)) {
            status = OCSM_ILLEGAL_PMTR_INDEX;
            goto cleanup;
        }

        index = irow - 1;
    } else {
        if (irow < 1 || irow > MODL->pmtr[ipmtr].nrow) {
            status = OCSM_ILLEGAL_PMTR_INDEX;
            goto cleanup;
        } else if (icol < 1 || icol > MODL->pmtr[ipmtr].ncol) {
            status = OCSM_ILLEGAL_PMTR_INDEX;
            goto cleanup;
        }

        index = (icol-1) + (irow-1) * (MODL->pmtr[ipmtr].ncol);
    }

    /* if this has a perturbed Body, free it first */
    status = removePerturbation(MODL);
    CHECK_STATUS(removePerturbation);

    /* set pertinent information */
    MODL->pmtr[ipmtr].value[index] = value;
    MODL->pmtr[ipmtr].dot[  index] = 0;

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmGetBnds - get the Bounds of a Parameter                        *
 *                                                                      *
 ************************************************************************
 */

int
ocsmGetBnds(void   *modl,               /* (in)  pointer to MODL */
            int    ipmtr,               /* (in)  Parameter index (1-npmtr) */
            int    irow,                /* (in)  row    index (1-nrow) */
            int    icol,                /* (in)  column index (1-ncol) */
            double *lbound,             /* (out) lower Bound */
            double *ubound)             /* (out) upper Bound */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       index;

    ROUTINE(ocsmGetBnds);
    DPRINT4("%s(ipmtr=%d, irow=%d, icol=%d) {",
            routine, ipmtr, irow, icol);

    /* --------------------------------------------------------------- */

    /* default return values */
    *lbound = -HUGEQ;
    *ubound = +HUGEQ;

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ipmtr, irow, and icol are given */
    if (ipmtr < 1 || ipmtr > MODL->npmtr) {
        status = OCSM_ILLEGAL_PMTR_INDEX;
        goto cleanup;
    } else if (MODL->pmtr[ipmtr].level != MODL->level) {
        status = OCSM_ILLEGAL_PMTR_INDEX;
        goto cleanup;
    } else if (icol == 0) {
        if (irow < 1 || irow > (MODL->pmtr[ipmtr].nrow)*(MODL->pmtr[ipmtr].ncol)) {
            status = OCSM_ILLEGAL_PMTR_INDEX;
            goto cleanup;
        }

        index = irow - 1;
    } else {
        if (irow < 1 || irow > MODL->pmtr[ipmtr].nrow) {
            status = OCSM_ILLEGAL_PMTR_INDEX;
            goto cleanup;
        } else if (icol < 1 || icol > MODL->pmtr[ipmtr].ncol) {
            status = OCSM_ILLEGAL_PMTR_INDEX;
            goto cleanup;
        }

        index = (icol-1) + (irow-1) * (MODL->pmtr[ipmtr].ncol);
    }

    /* return pertinent information */
    *lbound = MODL->pmtr[ipmtr].lbnd[index];
    *ubound = MODL->pmtr[ipmtr].ubnd[index];

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmSetVel - set a velocity for a Parameter                        *
 *                                                                      *
 ************************************************************************
 */

int
ocsmSetVel(void   *modl,                /* (in)  pointer to MODL */
           int    ipmtr,                /* (in)  Parameter index (1-npmtr) or 0 for all */
           int    irow,                 /* (in)  row    index (1-nrow)     or 0 for all */
           int    icol,                 /* (in)  column index (1-nrow)     or 0 for all */
           char   defn[])               /* (in)  definition of Velocity */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       index, jpmtr, jrow, jcol;
    double    dot, dum;

    ROUTINE(ocsmSetVel);
    DPRINT5("%s(ipmtr%d, irow=%d, icol=%d, defn=%s) {",
            routine, ipmtr, irow, icol, defn);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ipmtr, irow, and icol are given */
    if (ipmtr == 0) {
        if (irow != 0 || icol != 0) {
            status = OCSM_ILLEGAL_PMTR_INDEX;
            goto cleanup;
        }
    } else if (ipmtr < 0 || ipmtr > MODL->npmtr) {
        status = OCSM_ILLEGAL_PMTR_INDEX;
        goto cleanup;
    } else if (MODL->pmtr[ipmtr].level != MODL->level) {
        status = OCSM_ILLEGAL_PMTR_INDEX;
        goto cleanup;
    } else if (irow < 0 || irow > MODL->pmtr[ipmtr].nrow) {
        status = OCSM_ILLEGAL_ARGUMENT;
        goto cleanup;
    } else if (icol < 0 || icol > MODL->pmtr[ipmtr].ncol) {
        status = OCSM_ILLEGAL_ARGUMENT;
        goto cleanup;
    }

    /* if this has a perturbed Body, free it first */
    status = removePerturbation(MODL);
    CHECK_STATUS(removePerturbation);

    /* evaluate the definition */
    status = str2val(defn, NULL, &dot, &dum);
    CHECK_STATUS(str2val);

    /* set velocities on the Parameters */
    if (ipmtr == 0) {
        for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
            for (jrow = 1; jrow <= MODL->pmtr[jpmtr].nrow; jrow++) {
                for (jcol = 1; jcol <= MODL->pmtr[jpmtr].ncol; jcol++) {
                    index = (jcol-1) + (jrow-1) * (MODL->pmtr[jpmtr].ncol);
                    MODL->pmtr[jpmtr].dot[index] = dot;
                }
            }
        }
    } else if (irow == 0 && icol == 0) {
        for (jrow = 1; jrow <= MODL->pmtr[ipmtr].nrow; jrow++) {
            for (jcol = 1; jcol <= MODL->pmtr[ipmtr].ncol; jcol++) {
                index = (jcol-1) + (jrow-1) * (MODL->pmtr[ipmtr].ncol);
                MODL->pmtr[ipmtr].dot[index] = dot;
            }
        }
    } else if (irow == 0) {
        for (jrow = 1; jrow <= MODL->pmtr[ipmtr].nrow; jrow++) {
            index = (icol-1) + (jrow-1) * (MODL->pmtr[ipmtr].ncol);
            MODL->pmtr[ipmtr].dot[index] = dot;
        }
    } else if (icol == 0) {
        for (jcol = 1; jcol <= MODL->pmtr[ipmtr].ncol; jcol++) {
            index = (jcol-1) + (irow-1) * (MODL->pmtr[ipmtr].ncol);
            MODL->pmtr[ipmtr].dot[index] = dot;
        }
    } else {
        index = (icol-1) + (irow-1) * (MODL->pmtr[ipmtr].ncol);
        MODL->pmtr[ipmtr].dot[index] = dot;
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmSetVelD - set a double velocity for a Parameter                *
 *                                                                      *
 ************************************************************************
 */

int
ocsmSetVelD(void   *modl,               /* (in)  pointer to MODL */
            int    ipmtr,               /* (in)  Parameter index (1-npmtr) or 0 for all */
            int    irow,                /* (in)  row    index (1-nrow)     or 0 for all */
            int    icol,                /* (in)  column index (1-nrow)     or 0 for all */
            double dot)                 /* (in)  velocity to set */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       index, jpmtr, jrow, jcol;

    ROUTINE(ocsmSetVelD);
    DPRINT5("%s(ipmtr%d, irow=%d, icol=%d, value=%f) {",
            routine, ipmtr, irow, icol, dot);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ipmtr, irow, and icol are given */
    if (ipmtr == 0) {
        if (irow != 0 || icol != 0) {
            status = OCSM_ILLEGAL_PMTR_INDEX;
            goto cleanup;
        }
    } else if (ipmtr < 0 || ipmtr > MODL->npmtr) {
        status = OCSM_ILLEGAL_PMTR_INDEX;
        goto cleanup;
    } else if (MODL->pmtr[ipmtr].level != MODL->level) {
        status = OCSM_ILLEGAL_PMTR_INDEX;
        goto cleanup;
    } else if (irow < 0 || irow > MODL->pmtr[ipmtr].nrow) {
        status = OCSM_ILLEGAL_ARGUMENT;
        goto cleanup;
    } else if (icol < 0 || icol > MODL->pmtr[ipmtr].ncol) {
        status = OCSM_ILLEGAL_ARGUMENT;
        goto cleanup;
    }

    /* if this has a perturbed Body, free it first */
    status = removePerturbation(MODL);
    CHECK_STATUS(removePerturbation);

    /* set velocities on the Parameters */
    if (ipmtr == 0) {
        for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
            for (jrow = 1; jrow <= MODL->pmtr[jpmtr].nrow; jrow++) {
                for (jcol = 1; jcol <= MODL->pmtr[jpmtr].ncol; jcol++) {
                    index = (jcol-1) + (jrow-1) * (MODL->pmtr[jpmtr].ncol);
                    MODL->pmtr[jpmtr].dot[index] = dot;
                }
            }
        }
    } else if (irow == 0 && icol == 0) {
        for (jrow = 1; jrow <= MODL->pmtr[ipmtr].nrow; jrow++) {
            for (jcol = 1; jcol <= MODL->pmtr[ipmtr].ncol; jcol++) {
                index = (jcol-1) + (jrow-1) * (MODL->pmtr[ipmtr].ncol);
                MODL->pmtr[ipmtr].dot[index] = dot;
            }
        }
    } else if (irow == 0) {
        for (jrow = 1; jrow <= MODL->pmtr[ipmtr].nrow; jrow++) {
            index = (icol-1) + (jrow-1) * (MODL->pmtr[ipmtr].ncol);
            MODL->pmtr[ipmtr].dot[index] = dot;
        }
    } else if (icol == 0) {
        for (jcol = 1; jcol <= MODL->pmtr[ipmtr].ncol; jcol++) {
            index = (jcol-1) + (irow-1) * (MODL->pmtr[ipmtr].ncol);
            MODL->pmtr[ipmtr].dot[index] = dot;
        }
    } else {
        index = (icol-1) + (irow-1) * (MODL->pmtr[ipmtr].ncol);
        MODL->pmtr[ipmtr].dot[index] = dot;
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmGetUV - get the parametric coordinates on an Edge or Face      *
 *                                                                      *
 ************************************************************************
 */

int
ocsmGetUV(void   *modl,                 /* (in)  pointer to MODL */
          int    ibody,                 /* (in)  Body index (bias-1) */
          int    seltype,               /* (in)  OCSM_EDGE or OCSM_FACE (bias-1) */
          int    iselect,               /* (in)  Edge of Face index (bias-1) */
          int    npnt,                  /* (in)  number of points */
          double xyz[],                 /* (in)  coordinates (3*npnt in length) */
          double uv[])                  /* (out) para coords (1*npnt or 2*npnt in length) */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       iedge, iface, ipnt;
    double    pos[3], data[2], result[3];

    ROUTINE(ocsmGetUV);
    DPRINT5("%s(ibody=%d, seltype=%d, iselect=%d, npnt=%d) {",
            routine, ibody, seltype, iselect, npnt);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (ibody < 1 || ibody > MODL->nbody) {
        status = OCSM_ILLEGAL_BODY_INDEX;
        goto cleanup;
    }

    /* OCSM_EDGE was prescribed */
    if (seltype == OCSM_EDGE) {
        iedge = iselect;

        if (iedge < 0 || iedge > MODL->body[ibody].nedge) {
            status = OCSM_EDGE_NOT_FOUND;
            goto cleanup;
        }

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            pos[0] = xyz[3*ipnt  ];
            pos[1] = xyz[3*ipnt+1];
            pos[2] = xyz[3*ipnt+2];

            status = EG_invEvaluate(MODL->body[ibody].edge[iedge].eedge,
                                    pos, data, result);
            CHECK_STATUS(EG_invEvaluate);

            uv[2*ipnt  ] = data[0];
            uv[2*ipnt  ] = data[1];
        }

    /* OCSM_FACE was prescribed */
    } else if (seltype == OCSM_FACE) {
        iface = iselect;

        if (iface < 0 || iface > MODL->body[ibody].nface) {
            status = OCSM_FACE_NOT_FOUND;
            goto cleanup;
        }

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            pos[0] = xyz[3*ipnt  ];
            pos[1] = xyz[3*ipnt+1];
            pos[2] = xyz[3*ipnt+2];

            status = EG_invEvaluate(MODL->body[ibody].face[iface].eface,
                                    pos, data, result);
            CHECK_STATUS(EG_invEvaluate);

            uv[2*ipnt  ] = data[0];
            uv[2*ipnt+1] = data[1];
        }

    /* bad value for seltype */
    } else {
        status = OCSM_ILLEGAL_ARGUMENT;
        goto cleanup;
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmGetXYZ - get the coordinates on a Node, Edge, or Face          *
 *                                                                      *
 ************************************************************************
 */

int
ocsmGetXYZ(void   *modl,                /* (in)  pointer to MODL */
           int    ibody,                /* (in)  Body index (bias-1) */
           int    seltype,              /* (in)  OCSM_NODE, OCSM_EDGE, or OCSM_FACE */
           int    iselect,              /* (in)  Node, egde, or Face index (bias-1) */
           int    npnt,                 /* (in)  number of points */
 /*@null@*/double uv[],                 /* (in)  para coords (NULL, 1*npnt, or 2*npnt in length) */
           double xyz[])                /* (out) coordinates (3*npnt in length) */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       inode, iedge, iface, ipnt;
    double    params[2], data[18];

    ROUTINE(ocsmGetXYZ);
    DPRINT5("%s(ibody=%d, seltype=%d, iselect=%d, npnt=%d) {",
            routine, ibody, seltype, iselect, npnt);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (ibody < 1 || ibody > MODL->nbody) {
        status = OCSM_ILLEGAL_BODY_INDEX;
        goto cleanup;
    }

    /* OCSM_NODE was prescribed */
    if (seltype == OCSM_NODE) {
        inode = iselect;

        if (inode < 1 || inode > MODL->body[ibody].nnode) {
            status = OCSM_NODE_NOT_FOUND;
            goto cleanup;
        }

        xyz[0] = MODL->body[ibody].node[inode].x;
        xyz[1] = MODL->body[ibody].node[inode].y;
        xyz[2] = MODL->body[ibody].node[inode].z;

    /* OCSM_EDGE was prescribed */
    } else if (seltype == OCSM_EDGE) {
        iedge = iselect;

        if (iedge < 1 || iedge > MODL->body[ibody].nedge) {
            status = OCSM_EDGE_NOT_FOUND;
            goto cleanup;
        }

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            params[0] = uv[ipnt];

            status = EG_evaluate(MODL->body[ibody].edge[iedge].eedge,
                                 params, data);
            CHECK_STATUS(EG_evaluate);

            xyz[3*ipnt  ] = data[0];
            xyz[3*ipnt+1] = data[1];
            xyz[3*ipnt+2] = data[2];
        }

    /* OCSM_FACE was prescribed */
    } else if (seltype == OCSM_FACE) {
        iface = iselect;

        if (iface < 1 || iface > MODL->body[ibody].nface) {
            status = OCSM_FACE_NOT_FOUND;
            goto cleanup;
        }

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            params[0] = uv[2*ipnt  ];
            params[1] = uv[2*ipnt+1];

            status = EG_evaluate(MODL->body[ibody].face[iface].eface,
                                 params, data);
            CHECK_STATUS(EG_evaluate);

            xyz[3*ipnt  ] = data[0];
            xyz[3*ipnt+1] = data[1];
            xyz[3*ipnt+2] = data[2];
        }

    /* bad value for seltype */
    } else {
        status = OCSM_ILLEGAL_ARGUMENT;
        goto cleanup;
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmGetVel - get the velocities of coordinates on a Node, Edge, or Face *
 *                                                                      *
 ************************************************************************
 */

int
ocsmGetVel(void   *modl,                /* (in)  pointer to MODL */
           int    ibody,                /* (in)  Body index (bias-1) */
           int    seltype,              /* (in)  OCSM_NODE, OCSM_EDGE, or OCSM_FACE */
           int    iselect,              /* (in)  Node, Edge, or Face index (bias-1) */
           int    npnt,                 /* (in)  number of points */
 /*@null@*/double uv[],                 /* (in)  para coords
                                                    NULL           for OCSM_NODE
                                                    NULL or 1*npnt for OCSM_EDGE
                                                    NULL or 2*npnt for OCSM_FACE */
           double vel[])                /* (out) velocities (in pre-allocated array)
                                                    3      for OCSM_NODE
                                                    3*npnt for OCSM_EDGE
                                                      npnt for OCSM_FACE */
{
    int       status = SUCCESS;         /* (out) return status */

    int       needfd, jbody, kbody, inode, iedge, iface, ipnt, mtype;
    int       npnt_tess, ntri_tess;
    CINT      *tris, *tric, *ptype, *pindx;
    double    *dxyz=NULL;
    double    params[2], data[18], norm[4], uvlims[4], TdotT, VdotT;
    CDOUBLE   *xyz_tess, *uv_tess;

    int       oclass, nchild, *senses;
    ego       ref1, prev, next, eref, *echilds;

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(ocsmGetVel);
    DPRINT5("%s(ibody=%d, seltype=%d, iselect=%d, npnt=%d) {",
            routine, ibody, seltype, iselect, npnt);

    /* --------------------------------------------------------------- */

    SPRINT3(2, "ocsmGetVel(ibody=%d, seltype=%s, iselect=%d)", ibody, ocsmGetText(seltype), iselect);

    /* hierarch of routines for finding velocities:

       ocsmFree, ocsmNewBrch, ocsmSetBrch, ocsmDelBrch, ocsmSetArg, ocsmSetValu, ocsmSetValuD, ocsmSetVel, ocsmSetValD
       .   removePerturbation
       ocsmGetVel
       .   createPerturbation
       .   .   ocsmBuild
       .   .   EG_mapBody
       .   .   EG_mapTessBody
       .   velocityOfNode
       .   .   finiteDifference
       .   .   velocityOfEdge
       .   .   .   finiteDifference
       .   .   .   velocityOfFace
       .   .   .   .   finiteDifference
       .   velocityOfEdge
       .   .   finiteDifference
       .   .   velocityOfFace
       .   .   .   finiteDifference
       .   velocityOfFace
       .   .   finiteDifference
    */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (ibody < 1 || ibody > MODL->nbody) {
        status = OCSM_ILLEGAL_BODY_INDEX;
        goto cleanup;
    }

    /* default return */
    vel[0] = 0;
    vel[1] = 0;
    vel[2] = 0;

    /* a tessellation is needed if an Edge or Face and uv is not given */
    if ((seltype == OCSM_EDGE || seltype == OCSM_FACE) &&
        uv == NULL && MODL->body[ibody].etess == NULL    ) {
        status = OCSM_NEED_TESSELLATION;
        goto cleanup;
    }

    /* force finite differences */
    #ifdef FORCE_FINITE_DIFFS
        SPRINT0(2, "    force_finite_diffs selected at compile time");
        status = createPerturbation(MODL);
        CHECK_STATUS(createPerturbation);
    #endif

    /* create a perturbation if any parent is a Branch type
       for which we do NOT know how to analytically compute
       the sensitivities */
    if (MODL->perturb == NULL) {
        needfd = 0;
        for (jbody = 1; jbody <= ibody; jbody++) {
            if (MODL->body[jbody].brtype == OCSM_TORUS   ||
                MODL->body[jbody].brtype == OCSM_IMPORT  ||
                MODL->body[jbody].brtype == OCSM_LOFT    ||
                MODL->body[jbody].brtype == OCSM_BLEND   ||
                MODL->body[jbody].brtype == OCSM_REVOLVE ||
                MODL->body[jbody].brtype == OCSM_SWEEP   ||
                MODL->body[jbody].brtype == OCSM_FILLET  ||
                MODL->body[jbody].brtype == OCSM_CHAMFER ||
                MODL->body[jbody].brtype == OCSM_HOLLOW  ||
                MODL->body[jbody].brtype == OCSM_CONNECT ||
                MODL->body[jbody].brtype == OCSM_SKEND     ) {

                /* make sure jbody leads to kbody by following the
                   children back toward the root */
                if (jbody == ibody) {
                    needfd = 1;
                    break;
                } else {
                    kbody = MODL->body[jbody].ichld;
                    while (kbody > 0) {
                        if (kbody == ibody) {
                            needfd = 1;
                            break;
                        }
                        kbody = MODL->body[kbody].ichld;
                    }
                }
            }
            if (needfd > 0) break;
        }

        if (needfd > 0) {
            SPRINT5(2, "    creating perturbation for ibody=%d (called for seltyp=%s, iselect=%d, beacuse jbody=%d is a %s)",
                    ibody, ocsmGetText(seltype), iselect, jbody, ocsmGetText(MODL->body[jbody].brtype));

            status = createPerturbation(MODL);
            CHECK_STATUS(createPerturbation);
        }
    }

    /* if there was a problem with the previous attempt to create
       a perturbation, then return vel=0 */
    if (MODL->dtime == -2) {
        if        (seltype == OCSM_NODE) {
            vel[0] = 0;
            vel[1] = 0;
            vel[2] = 0;
        } else if (seltype == OCSM_EDGE) {
            for (ipnt = 0; ipnt < npnt; ipnt++) {
                vel[3*ipnt  ] = 0;
                vel[3*ipnt+1] = 0;
                vel[3*ipnt+2] = 0;
            }
        } else {
            for (ipnt = 0; ipnt < npnt; ipnt++) {
                vel[ipnt] = 0;
            }
        }
        goto cleanup;
    }

    /* OCSM_NODE was prescribed */
    if (seltype == OCSM_NODE) {
        inode = iselect;

        if (inode < 1 || inode > MODL->body[ibody].nnode) {
            status = OCSM_NODE_NOT_FOUND;
            goto cleanup;
        }

        if (npnt != 1) {
            status = OCSM_ILLEGAL_ARGUMENT;
            goto cleanup;
        }

        status = velocityOfNode(MODL, ibody, inode, vel);
        CHECK_STATUS(velocityOfNode);

    /* OCSM_EDGE was prescribed */
    } else if (seltype == OCSM_EDGE) {
        iedge = iselect;

        if (iedge < 1 || iedge > MODL->body[ibody].nedge) {
            status = OCSM_EDGE_NOT_FOUND;
            goto cleanup;
        }

        /* check for degenerate Edge */
        status = EG_getInfo(MODL->body[ibody].edge[iedge].eedge,
                            &oclass, &mtype, &ref1, &prev, &next);
        CHECK_STATUS(EG_getInfo);

        if (mtype == DEGENERATE) {
            for (ipnt = 0; ipnt < npnt; ipnt++) {
                vel[3*ipnt  ] = 0;
                vel[3*ipnt+1] = 0;
                vel[3*ipnt+2] = 0;
            }

            goto cleanup;
        }

        MALLOC(dxyz, double, 3*npnt);

        status = velocityOfEdge(MODL, ibody, iedge, npnt, uv, dxyz);
        CHECK_STATUS(velocityOfEdge);

        /* return normal component */
        if (uv != NULL) {
            uv_tess = uv;
        } else {
            if (MODL->body[ibody].etess == NULL) {
                status = OCSM_NEED_TESSELLATION;
                goto cleanup;
            }

            status = EG_getTessEdge(MODL->body[ibody].etess, iedge,
                                    &npnt_tess, &xyz_tess, &uv_tess);
            CHECK_STATUS(EG_getTessEdge);

            if (npnt_tess != npnt) {
                status = OCSM_ILLEGAL_ARGUMENT;
                goto cleanup;
            }
        }

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            params[0] = uv_tess[ipnt];
            status = EG_evaluate(MODL->body[ibody].edge[iedge].eedge,
                                 params, data);
            CHECK_STATUS(EG_evaluate);

            TdotT = data[       3] * data[3] + data[       4] * data[4] + data[       5] * data[5];
            VdotT = dxyz[3*ipnt  ] * data[3] + dxyz[3*ipnt+1] * data[4] + dxyz[3*ipnt+2] * data[5];

            vel[3*ipnt  ] = dxyz[3*ipnt  ] - VdotT * data[3] / TdotT;
            vel[3*ipnt+1] = dxyz[3*ipnt+1] - VdotT * data[4] / TdotT;
            vel[3*ipnt+2] = dxyz[3*ipnt+2] - VdotT * data[5] / TdotT;
        }

    /* OCSM_FACE was prescribed */
    } else if (seltype == OCSM_FACE) {
        iface = iselect;

        if (iface < 1 || iface > MODL->body[ibody].nface) {
            status = OCSM_FACE_NOT_FOUND;
            goto cleanup;
        }

        MALLOC(dxyz, double, 3*npnt);

        status = velocityOfFace(MODL, ibody, iface, npnt, uv, dxyz);
        CHECK_STATUS(velocityOfFace);

        /* return normal component */
        if (uv != NULL) {
            uv_tess = uv;
        } else {
            if (MODL->body[ibody].etess == NULL) {
                status = OCSM_NEED_TESSELLATION;
                goto cleanup;
            }

            status = EG_getTessFace(MODL->body[ibody].etess, iface,
                                    &npnt_tess, &xyz_tess, &uv_tess, &ptype, &pindx,
                                    &ntri_tess, &tris, &tric);
            CHECK_STATUS(EG_getTessFace);

            if (npnt_tess != npnt) {
                status = OCSM_ILLEGAL_ARGUMENT;
                goto cleanup;
            }
        }

        status = EG_getInfo(MODL->body[ibody].face[iface].eface,
                            &oclass, &mtype, &ref1, &prev, &next);
        CHECK_STATUS(EG_getInfo);

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            params[0] = uv_tess[2*ipnt  ];
            params[1] = uv_tess[2*ipnt+1];
            status = EG_evaluate(MODL->body[ibody].face[iface].eface,
                                 params, data);
            CHECK_STATUS(EG_evaluate);

            norm[0] = data[4] * data[8] - data[5] * data[7];
            norm[1] = data[5] * data[6] - data[3] * data[8];
            norm[2] = data[3] * data[7] - data[4] * data[6];
            norm[3] = mtype * sqrt(norm[0]*norm[0] + norm[1]*norm[1] + norm[2]*norm[2]);

            /* if the normal does not exist, find the normal of a "nearby" point by
               perturbing away from the degeneracy */
            if (fabs(norm[3]) < EPS12) {
                if (fabs(data[3]) < EPS06 && fabs(data[4]) < EPS06 && fabs(data[5]) < EPS06) {
                    status = EG_getTopology(MODL->body[ibody].face[iface].eface, &eref,
                                            &oclass, &mtype, uvlims, &nchild, &echilds, &senses);
                    CHECK_STATUS(EG_getTopology);

                    params[0] = (uvlims[0] + uvlims[1]) / 2;
                    if (fabs(params[1]-uvlims[2]) < fabs(params[1]-uvlims[3])) {
                        params[1] += 0.0001 * (uvlims[3] - uvlims[2]);
                    } else {
                        params[1] -= 0.0001 * (uvlims[3] - uvlims[2]);
                    }

                    status = EG_evaluate(MODL->body[ibody].face[iface].eface,
                                         params, data);
                    CHECK_STATUS(EG_evaluate);

                    norm[0] = data[4] * data[8] - data[5] * data[7];
                    norm[1] = data[5] * data[6] - data[3] * data[8];
                    norm[2] = data[3] * data[7] - data[4] * data[6];
                    norm[3] = mtype * sqrt(norm[0]*norm[0] + norm[1]*norm[1] + norm[2]*norm[2]);
                }
                if (fabs(data[6]) < EPS06 && fabs(data[7]) < EPS06 && fabs(data[8]) < EPS06) {
                    status = EG_getTopology(MODL->body[ibody].face[iface].eface, &eref,
                                            &oclass, &mtype, uvlims, &nchild, &echilds, &senses);
                    CHECK_STATUS(EG_getTopology);

                    params[1] = (uvlims[2] + uvlims[3]) / 2;
                    if (fabs(params[0]-uvlims[0]) < fabs(params[0]-uvlims[1])) {
                        params[0] += 0.0001 * (uvlims[1] - uvlims[0]);
                    } else {
                        params[0] -= 0.0001 * (uvlims[1] - uvlims[0]);
                    }

                    status = EG_evaluate(MODL->body[ibody].face[iface].eface,
                                         params, data);
                    CHECK_STATUS(EG_evaluate);

                    norm[0] = data[4] * data[8] - data[5] * data[7];
                    norm[1] = data[5] * data[6] - data[3] * data[8];
                    norm[2] = data[3] * data[7] - data[4] * data[6];
                    norm[3] = mtype * sqrt(norm[0]*norm[0] + norm[1]*norm[1] + norm[2]*norm[2]);
                }
            }

            /* if the normal exists, return the normal component of dxyz
                          otherwise, return the magnitude of dxyz */
            if (fabs(norm[3]) > EPS12) {
                vel[ipnt] = (dxyz[3*ipnt  ]*norm[0]
                           + dxyz[3*ipnt+1]*norm[1]
                           + dxyz[3*ipnt+2]*norm[2]) / norm[3];
            } else {
                SPRINT3(0, "WARNING:: using magnitude of velocity for ibody=%d, iface=%d, ipnt=%d",
                        ibody, iface, ipnt);

                vel[ipnt] = sqrt(dxyz[3*ipnt  ]*dxyz[3*ipnt  ]
                               + dxyz[3*ipnt+1]*dxyz[3*ipnt+1]
                               + dxyz[3*ipnt+2]*dxyz[3*ipnt+2]);
            }
        }

    /* bad value for seltype */
    } else {
        status = OCSM_ILLEGAL_ARGUMENT;
        goto cleanup;
    }

cleanup:
    FREE(dxyz);

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmGetBody - get info about a Body                                *
 *                                                                      *
 ************************************************************************
 */

int
ocsmGetBody(void   *modl,               /* (in)  pointer to MODL */
            int    ibody,               /* (in)  Body index (1-nbody) */
            int    *type,               /* (out) Branch type */
            int    *ichld,              /* (out) ibody of child (or 0 if root) */
            int    *ileft,              /* (out) ibody of left parent (or 0) */
            int    *irite,              /* (out) ibody of rite parent (or 0) */
            double vals[],              /* (out) array  of Arguments (at least 10 long) */
            int    *nnode,              /* (out) number of Nodes */
            int    *nedge,              /* (out) number of Edges */
            int    *nface)              /* (out) number of Faces */
{
    int       status = SUCCESS;         /* (out) return status */

    int       iarg;

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(ocsmGetBody);
    DPRINT2("%s(ibody=%d) {",
            routine, ibody);

    /* --------------------------------------------------------------- */

    /* default return values */
    *type  = 0;
    *ichld = 0;
    *ileft = 0;
    *irite = 0;
    *nnode = 0;
    *nedge = 0;
    *nface = 0;

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ibody is given */
    if (ibody < 1 || ibody > MODL->nbody) {
        status = OCSM_ILLEGAL_BODY_INDEX;
        goto cleanup;
    }

    /* return pertinent information */
    *type   = MODL->body[ibody].brtype;
    *ichld  = MODL->body[ibody].ichld;
    *ileft  = MODL->body[ibody].ileft;
    *irite  = MODL->body[ibody].irite;

    for (iarg = 1; iarg < 10; iarg++) {
        vals[iarg] = MODL->body[ibody].arg[iarg].val[0];
    }

    *nnode  = MODL->body[ibody].nnode;
    *nedge  = MODL->body[ibody].nedge;
    *nface  = MODL->body[ibody].nface;

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmPrintBodys - print all Bodys to file                           *
 *                                                                      *
 ************************************************************************
 */

int
ocsmPrintBodys(void   *modl,            /* (in)  pointer to MODL */
               FILE   *fp)              /* (in)  pointer to FILE */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       ibody, iarg;

    ROUTINE(ocsmPrintBodys);
    DPRINT2("%s(fp=%lx) {",
            routine, (long)fp);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* return message if no Bodys */
    if (MODL->nbody <= 0) {
        fprintf(fp, "    --none--\n");
        goto cleanup;
    }

    /* print Body table */
    fprintf(fp, "    ibody ibrch brchType  ileft irite ichld nnode nedge nface   bodyType     args ...\n");
    for (ibody = 1; ibody <= MODL->nbody; ibody++) {
        fprintf(fp, "    %5d %5d %-9s %5d %5d %5d %5d %5d %5d %10s ", ibody,
                MODL->body[ibody].ibrch,
                ocsmGetText(MODL->body[ibody].brtype),
                MODL->body[ibody].ileft,
                MODL->body[ibody].irite,
                MODL->body[ibody].ichld,
                MODL->body[ibody].nnode,
                MODL->body[ibody].nedge,
                MODL->body[ibody].nface,
                ocsmGetText(MODL->body[ibody].botype));

        for (iarg = 1; iarg < 10; iarg++) {
            if        (MODL->body[ibody].arg[iarg].nval == 0) {
                fprintf(fp, "%8.8s ", MODL->body[ibody].arg[iarg].str);
            } else if (MODL->body[ibody].arg[iarg].nval == 1) {
                fprintf(fp, "%8.3f ", MODL->body[ibody].arg[iarg].val[0]);
            } else if (MODL->body[ibody].arg[iarg].nval >  1) {
                fprintf(fp, " [%3d*1] ", MODL->body[ibody].arg[iarg].nval);
            } else {
                break;
            }
        }

        fprintf(fp, "\n");

    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmPrintBrep - print BRep associated with a Body to file          *
 *                                                                      *
 ************************************************************************
 */

int
ocsmPrintBrep(void   *modl,             /* (in)  pointer to MODL */
              int    ibody,             /* (in)  Body index (1-nbody) */
              FILE   *fp)               /* (in)  pointer to FILE */
{
    int       status = SUCCESS;         /* (out) return status */

    int       inode, iedge, iface;

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(ocsmPrintBrep);
    DPRINT3("%s(ibody=%d, fp=%lx) {",
            routine, ibody, (long)fp);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ibody is given */
    if (ibody < 1 || ibody > MODL->nbody) {
        status = OCSM_ILLEGAL_BODY_INDEX;
        goto cleanup;
    }

    /* loop through all Nodes */
    fprintf(fp, "inode nedge ibody     x             y             z\n");
    for (inode = 1; inode <= MODL->body[ibody].nnode; inode++) {
        fprintf(fp, "%5d %5d %5d %12.5f %12.5f %12.5f\n", inode,
                MODL->body[ibody].node[inode].nedge,
                MODL->body[ibody].node[inode].ibody,
                MODL->body[ibody].node[inode].x,
                MODL->body[ibody].node[inode].y,
                MODL->body[ibody].node[inode].z);
    }

    /* loop through all Edges */
    fprintf(fp, "iedge  ibeg  iend ileft irite nface ibody iford imark\n");
    for (iedge = 1; iedge <= MODL->body[ibody].nedge; iedge++) {
        fprintf(fp, "%5d %5d %5d %5d %5d %5d %5d %5d %5d\n", iedge,
                MODL->body[ibody].edge[iedge].ibeg,
                MODL->body[ibody].edge[iedge].iend,
                MODL->body[ibody].edge[iedge].ileft,
                MODL->body[ibody].edge[iedge].irite,
                MODL->body[ibody].edge[iedge].nface,
                MODL->body[ibody].edge[iedge].ibody,
                MODL->body[ibody].edge[iedge].iford,
                MODL->body[ibody].edge[iedge].imark);
    }

    /* loop through all Faces */
    fprintf(fp, "iface ibody iford imark\n");
    for (iface = 1; iface <= MODL->body[ibody].nface; iface++) {
        fprintf(fp, "%5d %5d %5d %5d\n", iface,
                MODL->body[ibody].face[iface].ibody,
                MODL->body[ibody].face[iface].iford,
                MODL->body[ibody].face[iface].imark);
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmGetText - convert an OCSM code to text                         *
 *                                                                      *
 ************************************************************************
 */

char *
ocsmGetText(int    icode)               /* (in)  code to look up */
{

    static char    unknown[]                          = "UNKNOWN";
    static char    success[]                          = "success";

    static char    ocsm_dimension[]                   = "dimension";
    static char    ocsm_despmtr[]                     = "despmtr";
    static char    ocsm_lbound[]                      = "lbound";
    static char    ocsm_ubound[]                      = "ubound";
    static char    ocsm_name[]                        = "name";
    static char    ocsm_attribute[]                   = "attribute";

    static char    ocsm_box[]                         = "box";
    static char    ocsm_sphere[]                      = "sphere";
    static char    ocsm_cone[]                        = "cone";
    static char    ocsm_cylinder[]                    = "cylinder";
    static char    ocsm_torus[]                       = "torus";
    static char    ocsm_import[]                      = "import";
    static char    ocsm_udprim[]                      = "udprim";
    static char    ocsm_restore[]                     = "restore";

    static char    ocsm_extrude[]                     = "extrude";
    static char    ocsm_rule[]                        = "rule";
    static char    ocsm_loft[]                        = "loft";
    static char    ocsm_blend[]                       = "blend";
    static char    ocsm_revolve[]                     = "revolve";
    static char    ocsm_sweep[]                       = "sweep";

    static char    ocsm_fillet[]                      = "fillet";
    static char    ocsm_chamfer[]                     = "chamfer";
    static char    ocsm_hollow[]                      = "hollow";
    static char    ocsm_connect[]                     = "connect";

    static char    ocsm_intersect[]                   = "intersect";
    static char    ocsm_subtract[]                    = "subtract";
    static char    ocsm_union[]                       = "union";
    static char    ocsm_join[]                        = "join";
    static char    ocsm_extract[]                     = "extract";
    static char    ocsm_combine[]                     = "combine";

    static char    ocsm_translate[]                   = "translate";
    static char    ocsm_rotatex[]                     = "rotatex";
    static char    ocsm_rotatey[]                     = "rotatey";
    static char    ocsm_rotatez[]                     = "rotatez";
    static char    ocsm_scale[]                       = "scale";
    static char    ocsm_mirror[]                      = "mirror";
    static char    ocsm_reorder[]                     = "reorder";

    static char    ocsm_skbeg[]                       = "skbeg";
    static char    ocsm_skcon[]                       = "skcon";
    static char    ocsm_skvar[]                       = "skvar";
    static char    ocsm_linseg[]                      = "linseg";
    static char    ocsm_cirarc[]                      = "cirarc";
    static char    ocsm_arc[]                         = "arc";
    static char    ocsm_spline[]                      = "spline";
    static char    ocsm_bezier[]                      = "bezier";
    static char    ocsm_skend[]                       = "skend";

    static char    ocsm_solbeg[]                      = "solbeg";
    static char    ocsm_solcon[]                      = "solcon";
    static char    ocsm_solend[]                      = "solend";

    static char    ocsm_interface[]                   = "interface";
    static char    ocsm_end[]                         = "end";
    static char    ocsm_set[]                         = "set";
    static char    ocsm_udparg[]                      = "udparg";
    static char    ocsm_select[]                      = "select";
    static char    ocsm_project[]                     = "project";
    static char    ocsm_macbeg[]                      = "macbeg";
    static char    ocsm_macend[]                      = "macend";
    static char    ocsm_recall[]                      = "recall";
    static char    ocsm_store[]                       = "store";
    static char    ocsm_patbeg[]                      = "patbeg";
    static char    ocsm_patbreak[]                    = "patbreak";
    static char    ocsm_patend[]                      = "patend";
    static char    ocsm_mark[]                        = "mark";
    static char    ocsm_dump[]                        = "dump";
    static char    ocsm_assert[]                      = "assert";

    static char    ocsm_primitive[]                   = "primitive";
    static char    ocsm_grown[]                       = "grown";
    static char    ocsm_applied[]                     = "applied";
    static char    ocsm_boolean[]                     = "boolean";
    static char    ocsm_transform[]                   = "transform";
    static char    ocsm_sketch[]                      = "sketch";
    static char    ocsm_solver[]                      = "solver";
    static char    ocsm_utility[]                     = "utility";

    static char    ocsm_active[]                      = "active";
    static char    ocsm_suppressed[]                  = "suppressed";
    static char    ocsm_inactive[]                    = "inactive";
    static char    ocsm_deferred[]                    = "deferred";

    static char    ocsm_solid_body[]                  = "solid_body";
    static char    ocsm_sheet_body[]                  = "sheet_body";
    static char    ocsm_wire_body[]                   = "wire_body";
    static char    ocsm_node_body[]                   = "node_body";
    static char    ocsm_null_body[]                   = "null_body";

    static char    ocsm_external[]                    = "external";
    static char    ocsm_internal[]                    = "internal";

    static char    ocsm_node[]                        = "node";
    static char    ocsm_edge[]                        = "edge";
    static char    ocsm_face[]                        = "face";
    static char    ocsm_volume[]                      = "volume";

    static char    ocsm_file_not_found[]              = "file_not_found";
    static char    ocsm_illegal_statement[]           = "illegal_statement";
    static char    ocsm_not_enough_args[]             = "not_enough_args";
    static char    ocsm_name_already_defined[]        = "name_already_defined";
    static char    ocsm_nested_too_deeply[]           = "nested_too_deeply";
    static char    ocsm_patbeg_without_patend[]       = "patbeg_without_patend";
    static char    ocsm_patend_without_patbeg[]       = "patend_without_patbeg";
    static char    ocsm_not_modl_structure[]          = "not_modl_structure";
    static char    ocsm_problem_creating_perturb[]    = "problem_creating_perturb";

    static char    ocsm_missing_mark[]                = "missing_mark";
    static char    ocsm_insufficient_bodys_on_stack[] = "insufficient_bodys_on_stack";
    static char    ocsm_wrong_types_on_stack[]        = "wrong_types_on_stack";
    static char    ocsm_did_not_create_body[]         = "did_not_create_body";
    static char    ocsm_created_too_many_bodys[]      = "created_too_many_bodys";
    static char    ocsm_too_many_bodys_on_stack[]     = "too_many_bodys_on_stack";
    static char    ocsm_error_in_bodys_on_stack[]     = "errors_in_bodys_on_stack";

    static char    ocsm_modl_not_checked[]            = "modl_not_checked";
    static char    ocsm_need_tessellation[]           = "need_tessellation";

    static char    ocsm_body_not_found[]              = "body_not_found";
    static char    ocsm_face_not_found[]              = "face_not_found";
    static char    ocsm_edge_not_found[]              = "edge_not_found";
    static char    ocsm_node_not_found[]              = "node_not_found";
    static char    ocsm_illegal_value[]               = "illegal_value";
    static char    ocsm_illegal_attribute[]           = "illegal_attribute";

    static char    ocsm_sketch_is_open[]              = "sketch_is_open";
    static char    ocsm_sketch_is_not_open[]          = "sketch_is_not_open";
    static char    ocsm_colinear_sketch_points[]      = "colinear_sketch_points";
    static char    ocsm_non_coplanar_sketch_points[]  = "non_coplanar_sketch_points";
    static char    ocsm_too_many_sketch_points[]      = "too_many_sketch_points";
    static char    ocsm_too_few_spline_points[]       = "too_few_spline_points";
    static char    ocsm_sketch_does_not_close[]       = "sketch_does_not_close";
    static char    ocsm_closed_sketch_not_planar[]    = "closed_sketch_not_planar";
    static char    ocsm_assert_failed[]               = "assert_failed";

    static char    ocsm_illegal_char_in_expr[]        = "illegal_char_in_expr";
    static char    ocsm_close_before_open[]           = "close_before_open";
    static char    ocsm_missing_close[]               = "missing_close";
    static char    ocsm_improper_nesting[]            = "improper_nesting";
    static char    ocsm_illegal_token_sequence[]      = "illegal_token_sequence";
    static char    ocsm_illegal_number[]              = "illegal_number";
    static char    ocsm_illegal_pmtr_name[]           = "illegal_pmtr_name";
    static char    ocsm_illegal_func_name[]           = "illegal_func_name";
    static char    ocsm_illegal_type[]                = "illegal_type";
    static char    ocsm_illegal_narg[]                = "illegal_narg";

    static char    ocsm_name_not_found[]              = "name_not_found";
    static char    ocsm_name_not_unique[]             = "name_not_unique";
    static char    ocsm_pmtr_is_external[]            = "pmtr_is_external";
    static char    ocsm_pmtr_is_internal[]            = "pmtr_is_internal";
    static char    ocsm_func_arg_out_of_bounds[]      = "func_arg_out_of_bounds";
    static char    ocsm_val_stack_underflow[]         = "val_stack_underflow";
    static char    ocsm_val_stack_overflow[]          = "val_stack_overflow";

    static char    ocsm_illegal_brch_index[]          = "illegal_brch_index";
    static char    ocsm_illegal_pmtr_index[]          = "illegal_pmtr_index";
    static char    ocsm_illegal_body_index[]          = "illegal_body_index";
    static char    ocsm_illegal_arg_index[]           = "illegal_arg_index";
    static char    ocsm_illegal_activity[]            = "illegal_activity";
    static char    ocsm_illegal_macro_index[]         = "illegal_macro_index";
    static char    ocsm_illegal_argument[]            = "illegal_argument";
    static char    ocsm_cannot_be_suppressed[]        = "cannot_be_suppressed";
    static char    ocsm_storage_already_used[]        = "storage_already_used";
    static char    ocsm_nothing_previously_stored[]   = "nothing_previously_stored";

    static char    ocsm_solver_is_open[]              = "solver_is_open";
    static char    ocsm_solver_is_not_open[]          = "solver_is_not_open";
    static char    ocsm_too_many_solver_vars[]        = "too_many_solver_vars";
    static char    ocsm_underconstrained[]            = "underconstrained";
    static char    ocsm_overconstrained[]             = "overconstrained";
    static char    ocsm_singular_matrix[]             = "singular_matrix";
    static char    ocsm_not_converged[]               = "not_converged";

    static char    ocsm_udp_error1[]                  = "udp_error1";
    static char    ocsm_udp_error2[]                  = "udp_error2";
    static char    ocsm_udp_error3[]                  = "udp_error3";
    static char    ocsm_udp_error4[]                  = "udp_error4";
    static char    ocsm_udp_error5[]                  = "udp_error5";
    static char    ocsm_udp_error6[]                  = "udp_error6";
    static char    ocsm_udp_error7[]                  = "udp_error7";
    static char    ocsm_udp_error8[]                  = "udp_error8";
    static char    ocsm_udp_error9[]                  = "udp_error9";

    static char    ocsm_op_stack_underflow[]          = "op_stack_underflow";
    static char    ocsm_op_stack_overflow[]           = "op_stack_overflow";
    static char    ocsm_rpn_stack_underflow[]         = "rpn_stack_underflow";
    static char    ocsm_rpn_stack_overflow[]          = "rpn_stack_overflow";
    static char    ocsm_token_stack_underflow[]       = "token_stack_underflow";
    static char    ocsm_token_stack_overflow[]        = "token_stack_overflow";
    static char    ocsm_unsupported[]                 = "unsupported";
    static char    ocsm_internal_error[]              = "internal_error";

    static char    egads_notfound[]                   = "egads_notfound";
    static char    egads_nullobj[]                    = "egads_nullobj";
    static char    egads_notobj[]                     = "egads_notobj";
    static char    egads_malloc[]                     = "egads_malloc";
    static char    egads_indexerr[]                   = "egads_indexerr";
    static char    egads_noname[]                     = "egads_noname";
    static char    egads_nodata[]                     = "egads_nodata";
    static char    egads_mixcntx[]                    = "egads_mixcntx";
    static char    egads_notcntx[]                    = "egads_notcntx";
    static char    egads_notxform[]                   = "egads_notxform";
    static char    egads_referce[]                    = "egads_referce";
    static char    egads_nottopo[]                    = "egads_nottopo";
    static char    egads_empty[]                      = "egads_empty";
    static char    egads_nottess[]                    = "egads_nottess";
    static char    egads_notgeom[]                    = "egads_notgeom";
    static char    egads_rangerr[]                    = "egads_rangerr";
    static char    egads_noload[]                     = "egads_noload";
    static char    egads_notmodel[]                   = "egads_notmodel";
    static char    egads_writerr[]                    = "egads_writerr";
    static char    egads_notbody[]                    = "egads_notbody";
    static char    egads_geomerr[]                    = "egads_geomerr";
    static char    egads_topoerr[]                    = "egads_topoerr";
    static char    egads_consterr[]                   = "egads_consterr";
    static char    egads_degen[]                      = "egads_degen";
    static char    egads_notortho[]                   = "egads_notortho";
    static char    egads_badscale[]                   = "egads_badscale";
    static char    egads_ocsegflt[]                   = "egads_ocsegflt";
    static char    egads_topocnt[]                    = "egads_topocnt";
    static char    egads_attrerr[]                    = "egads_attrerr";
    static char    egads_exists[]                     = "egads_exists";

    /* --------------------------------------------------------------- */

    /* OCSM_SUCCESS */
    if (icode == SUCCESS                         ) return success;

    /* not branches */
    if (icode == OCSM_DIMENSION                  ) return ocsm_dimension;
    if (icode == OCSM_DESPMTR                    ) return ocsm_despmtr;
    if (icode == OCSM_LBOUND                     ) return ocsm_lbound;
    if (icode == OCSM_UBOUND                     ) return ocsm_ubound;
    if (icode == OCSM_NAME                       ) return ocsm_name;
    if (icode == OCSM_ATTRIBUTE                  ) return ocsm_attribute;

    /* OCSM_PRIMITIVE */
    if (icode == OCSM_BOX                        ) return ocsm_box;
    if (icode == OCSM_SPHERE                     ) return ocsm_sphere;
    if (icode == OCSM_CONE                       ) return ocsm_cone;
    if (icode == OCSM_CYLINDER                   ) return ocsm_cylinder;
    if (icode == OCSM_TORUS                      ) return ocsm_torus;
    if (icode == OCSM_IMPORT                     ) return ocsm_import;
    if (icode == OCSM_UDPRIM                     ) return ocsm_udprim;
    if (icode == OCSM_RESTORE                    ) return ocsm_restore;

    /* OCSM_GROWN */
    if (icode == OCSM_EXTRUDE                    ) return ocsm_extrude;
    if (icode == OCSM_RULE                       ) return ocsm_rule;
    if (icode == OCSM_LOFT                       ) return ocsm_loft;
    if (icode == OCSM_BLEND                      ) return ocsm_blend;
    if (icode == OCSM_REVOLVE                    ) return ocsm_revolve;
    if (icode == OCSM_SWEEP                      ) return ocsm_sweep;

    /* OCSM_APPLIED */
    if (icode == OCSM_FILLET                     ) return ocsm_fillet;
    if (icode == OCSM_CHAMFER                    ) return ocsm_chamfer;
    if (icode == OCSM_HOLLOW                     ) return ocsm_hollow;
    if (icode == OCSM_CONNECT                    ) return ocsm_connect;

    /* OCSM_BOOLEAN */
    if (icode == OCSM_INTERSECT                  ) return ocsm_intersect;
    if (icode == OCSM_SUBTRACT                   ) return ocsm_subtract;
    if (icode == OCSM_UNION                      ) return ocsm_union;
    if (icode == OCSM_JOIN                       ) return ocsm_join;
    if (icode == OCSM_EXTRACT                    ) return ocsm_extract;
    if (icode == OCSM_COMBINE                    ) return ocsm_combine;

    /* OCSM_TRANSFORM */
    if (icode == OCSM_TRANSLATE                  ) return ocsm_translate;
    if (icode == OCSM_ROTATEX                    ) return ocsm_rotatex;
    if (icode == OCSM_ROTATEY                    ) return ocsm_rotatey;
    if (icode == OCSM_ROTATEZ                    ) return ocsm_rotatez;
    if (icode == OCSM_SCALE                      ) return ocsm_scale;
    if (icode == OCSM_MIRROR                     ) return ocsm_mirror;
    if (icode == OCSM_REORDER                    ) return ocsm_reorder;

    /* OCSM_SKETCH */
    if (icode == OCSM_SKBEG                      ) return ocsm_skbeg;
    if (icode == OCSM_SKCON                      ) return ocsm_skcon;
    if (icode == OCSM_SKVAR                      ) return ocsm_skvar;
    if (icode == OCSM_LINSEG                     ) return ocsm_linseg;
    if (icode == OCSM_CIRARC                     ) return ocsm_cirarc;
    if (icode == OCSM_ARC                        ) return ocsm_arc;
    if (icode == OCSM_SPLINE                     ) return ocsm_spline;
    if (icode == OCSM_BEZIER                     ) return ocsm_bezier;
    if (icode == OCSM_SKEND                      ) return ocsm_skend;

    /* OCSM_SOLVER */
    if (icode == OCSM_SOLBEG                     ) return ocsm_solbeg;
    if (icode == OCSM_SOLCON                     ) return ocsm_solcon;
    if (icode == OCSM_SOLEND                     ) return ocsm_solend;

    /* OCSM_UTILITY */
    if (icode == OCSM_INTERFACE                  ) return ocsm_interface;
    if (icode == OCSM_END                        ) return ocsm_end;
    if (icode == OCSM_SET                        ) return ocsm_set;
    if (icode == OCSM_UDPARG                     ) return ocsm_udparg;
    if (icode == OCSM_SELECT                     ) return ocsm_select;
    if (icode == OCSM_PROJECT                    ) return ocsm_project;
    if (icode == OCSM_MACBEG                     ) return ocsm_macbeg;
    if (icode == OCSM_MACEND                     ) return ocsm_macend;
    if (icode == OCSM_RECALL                     ) return ocsm_recall;
    if (icode == OCSM_STORE                      ) return ocsm_store;
    if (icode == OCSM_PATBEG                     ) return ocsm_patbeg;
    if (icode == OCSM_PATBREAK                   ) return ocsm_patbreak;
    if (icode == OCSM_PATEND                     ) return ocsm_patend;
    if (icode == OCSM_MARK                       ) return ocsm_mark;
    if (icode == OCSM_DUMP                       ) return ocsm_dump;
    if (icode == OCSM_ASSERT                     ) return ocsm_assert;

    /* Branch classes */
    if (icode == OCSM_PRIMITIVE                  ) return ocsm_primitive;
    if (icode == OCSM_GROWN                      ) return ocsm_grown;
    if (icode == OCSM_APPLIED                    ) return ocsm_applied;
    if (icode == OCSM_BOOLEAN                    ) return ocsm_boolean;
    if (icode == OCSM_TRANSFORM                  ) return ocsm_transform;
    if (icode == OCSM_SKETCH                     ) return ocsm_sketch;
    if (icode == OCSM_SOLVER                     ) return ocsm_solver;
    if (icode == OCSM_UTILITY                    ) return ocsm_utility;

    /* Branch activities */
    if (icode == OCSM_ACTIVE                     ) return ocsm_active;
    if (icode == OCSM_SUPPRESSED                 ) return ocsm_suppressed;
    if (icode == OCSM_INACTIVE                   ) return ocsm_inactive;
    if (icode == OCSM_DEFERRED                   ) return ocsm_deferred;

    /* Body types */
    if (icode == OCSM_SOLID_BODY                 ) return ocsm_solid_body;
    if (icode == OCSM_SHEET_BODY                 ) return ocsm_sheet_body;
    if (icode == OCSM_WIRE_BODY                  ) return ocsm_wire_body;
    if (icode == OCSM_NODE_BODY                  ) return ocsm_node_body;
    if (icode == OCSM_NULL_BODY                  ) return ocsm_null_body;

    /* Parameter types */
    if (icode == OCSM_EXTERNAL                   ) return ocsm_external;
    if (icode == OCSM_INTERNAL                   ) return ocsm_internal;

    /* Selector types */
    if (icode == OCSM_NODE                       ) return ocsm_node;
    if (icode == OCSM_EDGE                       ) return ocsm_edge;
    if (icode == OCSM_FACE                       ) return ocsm_face;
    if (icode == OCSM_VOLUME                     ) return ocsm_volume;

     /* Error codes */
    if (icode == OCSM_FILE_NOT_FOUND             ) return ocsm_file_not_found;
    if (icode == OCSM_ILLEGAL_STATEMENT          ) return ocsm_illegal_statement;
    if (icode == OCSM_NOT_ENOUGH_ARGS            ) return ocsm_not_enough_args;
    if (icode == OCSM_NAME_ALREADY_DEFINED       ) return ocsm_name_already_defined;
    if (icode == OCSM_NESTED_TOO_DEEPLY          ) return ocsm_nested_too_deeply;
    if (icode == OCSM_PATBEG_WITHOUT_PATEND      ) return ocsm_patbeg_without_patend;
    if (icode == OCSM_PATEND_WITHOUT_PATBEG      ) return ocsm_patend_without_patbeg;
    if (icode == OCSM_NOT_MODL_STRUCTURE         ) return ocsm_not_modl_structure;
    if (icode == OCSM_PROBLEM_CREATING_PERTURB   ) return ocsm_problem_creating_perturb;

    if (icode == OCSM_MISSING_MARK               ) return ocsm_missing_mark;
    if (icode == OCSM_INSUFFICIENT_BODYS_ON_STACK) return ocsm_insufficient_bodys_on_stack;
    if (icode == OCSM_WRONG_TYPES_ON_STACK       ) return ocsm_wrong_types_on_stack;
    if (icode == OCSM_DID_NOT_CREATE_BODY        ) return ocsm_did_not_create_body;
    if (icode == OCSM_CREATED_TOO_MANY_BODYS     ) return ocsm_created_too_many_bodys;
    if (icode == OCSM_TOO_MANY_BODYS_ON_STACK    ) return ocsm_too_many_bodys_on_stack;
    if (icode == OCSM_ERROR_IN_BODYS_ON_STACK    ) return ocsm_error_in_bodys_on_stack;
    if (icode == OCSM_MODL_NOT_CHECKED           ) return ocsm_modl_not_checked;
    if (icode == OCSM_NEED_TESSELLATION          ) return ocsm_need_tessellation;

    if (icode == OCSM_BODY_NOT_FOUND             ) return ocsm_body_not_found;
    if (icode == OCSM_FACE_NOT_FOUND             ) return ocsm_face_not_found;
    if (icode == OCSM_EDGE_NOT_FOUND             ) return ocsm_edge_not_found;
    if (icode == OCSM_NODE_NOT_FOUND             ) return ocsm_node_not_found;
    if (icode == OCSM_ILLEGAL_VALUE              ) return ocsm_illegal_value;
    if (icode == OCSM_ILLEGAL_ATTRIBUTE          ) return ocsm_illegal_attribute;

    if (icode == OCSM_SKETCH_IS_OPEN             ) return ocsm_sketch_is_open;
    if (icode == OCSM_SKETCH_IS_NOT_OPEN         ) return ocsm_sketch_is_not_open;
    if (icode == OCSM_COLINEAR_SKETCH_POINTS     ) return ocsm_colinear_sketch_points;
    if (icode == OCSM_NON_COPLANAR_SKETCH_POINTS ) return ocsm_non_coplanar_sketch_points;
    if (icode == OCSM_TOO_MANY_SKETCH_POINTS     ) return ocsm_too_many_sketch_points;
    if (icode == OCSM_TOO_FEW_SPLINE_POINTS      ) return ocsm_too_few_spline_points;
    if (icode == OCSM_SKETCH_DOES_NOT_CLOSE      ) return ocsm_sketch_does_not_close;
    if (icode == OCSM_CLOSED_SKETCH_NOT_PLANAR   ) return ocsm_closed_sketch_not_planar;
    if (icode == OCSM_ASSERT_FAILED              ) return ocsm_assert_failed;

    if (icode == OCSM_ILLEGAL_CHAR_IN_EXPR       ) return ocsm_illegal_char_in_expr;
    if (icode == OCSM_CLOSE_BEFORE_OPEN          ) return ocsm_close_before_open;
    if (icode == OCSM_MISSING_CLOSE              ) return ocsm_missing_close;
    if (icode == OCSM_IMPROPER_NESTING           ) return ocsm_improper_nesting;
    if (icode == OCSM_ILLEGAL_TOKEN_SEQUENCE     ) return ocsm_illegal_token_sequence;
    if (icode == OCSM_ILLEGAL_NUMBER             ) return ocsm_illegal_number;
    if (icode == OCSM_ILLEGAL_PMTR_NAME          ) return ocsm_illegal_pmtr_name;
    if (icode == OCSM_ILLEGAL_FUNC_NAME          ) return ocsm_illegal_func_name;
    if (icode == OCSM_ILLEGAL_TYPE               ) return ocsm_illegal_type;
    if (icode == OCSM_ILLEGAL_NARG               ) return ocsm_illegal_narg;

    if (icode == OCSM_NAME_NOT_FOUND             ) return ocsm_name_not_found;
    if (icode == OCSM_NAME_NOT_UNIQUE            ) return ocsm_name_not_unique;
    if (icode == OCSM_PMTR_IS_EXTERNAL           ) return ocsm_pmtr_is_external;
    if (icode == OCSM_PMTR_IS_INTERNAL           ) return ocsm_pmtr_is_internal;
    if (icode == OCSM_FUNC_ARG_OUT_OF_BOUNDS     ) return ocsm_func_arg_out_of_bounds;
    if (icode == OCSM_VAL_STACK_UNDERFLOW        ) return ocsm_val_stack_underflow;
    if (icode == OCSM_VAL_STACK_OVERFLOW         ) return ocsm_val_stack_overflow;

    if (icode == OCSM_ILLEGAL_BRCH_INDEX         ) return ocsm_illegal_brch_index;
    if (icode == OCSM_ILLEGAL_PMTR_INDEX         ) return ocsm_illegal_pmtr_index;
    if (icode == OCSM_ILLEGAL_BODY_INDEX         ) return ocsm_illegal_body_index;
    if (icode == OCSM_ILLEGAL_ARG_INDEX          ) return ocsm_illegal_arg_index;
    if (icode == OCSM_ILLEGAL_ACTIVITY           ) return ocsm_illegal_activity;
    if (icode == OCSM_ILLEGAL_MACRO_INDEX        ) return ocsm_illegal_macro_index;
    if (icode == OCSM_ILLEGAL_ARGUMENT           ) return ocsm_illegal_argument;
    if (icode == OCSM_CANNOT_BE_SUPPRESSED       ) return ocsm_cannot_be_suppressed;
    if (icode == OCSM_STORAGE_ALREADY_USED       ) return ocsm_storage_already_used;
    if (icode == OCSM_NOTHING_PREVIOUSLY_STORED  ) return ocsm_nothing_previously_stored;

    if (icode == OCSM_SOLVER_IS_OPEN             ) return ocsm_solver_is_open;
    if (icode == OCSM_SOLVER_IS_NOT_OPEN         ) return ocsm_solver_is_not_open;
    if (icode == OCSM_TOO_MANY_SOLVER_VARS       ) return ocsm_too_many_solver_vars;
    if (icode == OCSM_UNDERCONSTRAINED           ) return ocsm_underconstrained;
    if (icode == OCSM_OVERCONSTRAINED            ) return ocsm_overconstrained;
    if (icode == OCSM_SINGULAR_MATRIX            ) return ocsm_singular_matrix;
    if (icode == OCSM_NOT_CONVERGED              ) return ocsm_not_converged;

    if (icode == OCSM_UDP_ERROR1                 ) return ocsm_udp_error1;
    if (icode == OCSM_UDP_ERROR2                 ) return ocsm_udp_error2;
    if (icode == OCSM_UDP_ERROR3                 ) return ocsm_udp_error3;
    if (icode == OCSM_UDP_ERROR4                 ) return ocsm_udp_error4;
    if (icode == OCSM_UDP_ERROR5                 ) return ocsm_udp_error5;
    if (icode == OCSM_UDP_ERROR6                 ) return ocsm_udp_error6;
    if (icode == OCSM_UDP_ERROR7                 ) return ocsm_udp_error7;
    if (icode == OCSM_UDP_ERROR8                 ) return ocsm_udp_error8;
    if (icode == OCSM_UDP_ERROR9                 ) return ocsm_udp_error9;

    if (icode == OCSM_OP_STACK_UNDERFLOW         ) return ocsm_op_stack_underflow;
    if (icode == OCSM_OP_STACK_OVERFLOW          ) return ocsm_op_stack_overflow;
    if (icode == OCSM_RPN_STACK_UNDERFLOW        ) return ocsm_rpn_stack_underflow;
    if (icode == OCSM_RPN_STACK_OVERFLOW         ) return ocsm_rpn_stack_overflow;
    if (icode == OCSM_TOKEN_STACK_UNDERFLOW      ) return ocsm_token_stack_underflow;
    if (icode == OCSM_TOKEN_STACK_OVERFLOW       ) return ocsm_token_stack_overflow;
    if (icode == OCSM_UNSUPPORTED                ) return ocsm_unsupported;
    if (icode == OCSM_INTERNAL_ERROR             ) return ocsm_internal_error;

    /* EGADS errors */
    if (icode == EGADS_NOTFOUND                  ) return egads_notfound;
    if (icode == EGADS_NULLOBJ                   ) return egads_nullobj;
    if (icode == EGADS_NOTOBJ                    ) return egads_notobj;
    if (icode == EGADS_MALLOC                    ) return egads_malloc;
    if (icode == EGADS_INDEXERR                  ) return egads_indexerr;
    if (icode == EGADS_NONAME                    ) return egads_noname;
    if (icode == EGADS_NODATA                    ) return egads_nodata;
    if (icode == EGADS_MIXCNTX                   ) return egads_mixcntx;
    if (icode == EGADS_NOTCNTX                   ) return egads_notcntx;
    if (icode == EGADS_NOTXFORM                  ) return egads_notxform;
    if (icode == EGADS_REFERCE                   ) return egads_referce;
    if (icode == EGADS_NOTTOPO                   ) return egads_nottopo;
    if (icode == EGADS_EMPTY                     ) return egads_empty;
    if (icode == EGADS_NOTTESS                   ) return egads_nottess;
    if (icode == EGADS_NOTGEOM                   ) return egads_notgeom;
    if (icode == EGADS_RANGERR                   ) return egads_rangerr;
    if (icode == EGADS_NOLOAD                    ) return egads_noload;
    if (icode == EGADS_NOTMODEL                  ) return egads_notmodel;
    if (icode == EGADS_WRITERR                   ) return egads_writerr;
    if (icode == EGADS_NOTBODY                   ) return egads_notbody;
    if (icode == EGADS_GEOMERR                   ) return egads_geomerr;
    if (icode == EGADS_TOPOERR                   ) return egads_topoerr;
    if (icode == EGADS_CONSTERR                  ) return egads_consterr;
    if (icode == EGADS_DEGEN                     ) return egads_degen;
    if (icode == EGADS_NOTORTHO                  ) return egads_notortho;
    if (icode == EGADS_BADSCALE                  ) return egads_badscale;
    if (icode == EGADS_OCSEGFLT                  ) return egads_ocsegflt;
    if (icode == EGADS_TOPOCNT                   ) return egads_topocnt;
    if (icode == EGADS_ATTRERR                   ) return egads_attrerr;
    if (icode == EGADS_EXISTS                    ) return egads_exists;

    /* default return */
    return unknown;
}


/*
 ************************************************************************
 *                                                                      *
 *   ocsmGetCode - convert text to an OCSM code                         *
 *                                                                      *
 ************************************************************************
 */

int
ocsmGetCode(char   *text)               /* (in)  text to look up */
{

    /* --------------------------------------------------------------- */

    /* OCSM_SUCCESS */
    if (strcmp(text, "success"   ) == 0) return SUCCESS;

    /* not a branch */
    if (strcmp(text, "dimension" ) == 0) return OCSM_DIMENSION;
    if (strcmp(text, "despmtr"   ) == 0) return OCSM_DESPMTR;
    if (strcmp(text, "lbound"    ) == 0) return OCSM_LBOUND;
    if (strcmp(text, "ubound"    ) == 0) return OCSM_UBOUND;
    if (strcmp(text, "name"      ) == 0) return OCSM_NAME;
    if (strcmp(text, "attribute" ) == 0) return OCSM_ATTRIBUTE;
    if (strcmp(text, "end"       ) == 0) return OCSM_END;

    /* OCSM_PRIMITIVE */
    if (strcmp(text, "box"       ) == 0) return OCSM_BOX;
    if (strcmp(text, "sphere"    ) == 0) return OCSM_SPHERE;
    if (strcmp(text, "cone"      ) == 0) return OCSM_CONE;
    if (strcmp(text, "cylinder"  ) == 0) return OCSM_CYLINDER;
    if (strcmp(text, "torus"     ) == 0) return OCSM_TORUS;
    if (strcmp(text, "import"    ) == 0) return OCSM_IMPORT;
    if (strcmp(text, "udprim"    ) == 0) return OCSM_UDPRIM;
    if (strcmp(text, "restore"   ) == 0) return OCSM_RESTORE;

    /* OCSM_GROWN */
    if (strcmp(text, "extrude"   ) == 0) return OCSM_EXTRUDE;
    if (strcmp(text, "rule"      ) == 0) return OCSM_RULE;
    if (strcmp(text, "loft"      ) == 0) return OCSM_LOFT;
    if (strcmp(text, "blend"     ) == 0) return OCSM_BLEND;
    if (strcmp(text, "revolve"   ) == 0) return OCSM_REVOLVE;
    if (strcmp(text, "sweep"     ) == 0) return OCSM_SWEEP;

    /* OCSM_APPLIED */
    if (strcmp(text, "fillet"    ) == 0) return OCSM_FILLET;
    if (strcmp(text, "chamfer"   ) == 0) return OCSM_CHAMFER;
    if (strcmp(text, "hollow"    ) == 0) return OCSM_HOLLOW;
    if (strcmp(text, "connect"   ) == 0) return OCSM_CONNECT;

    /* OCSM_BOOLEAN */
    if (strcmp(text, "intersect" ) == 0) return OCSM_INTERSECT;
    if (strcmp(text, "subtract"  ) == 0) return OCSM_SUBTRACT;
    if (strcmp(text, "union"     ) == 0) return OCSM_UNION;
    if (strcmp(text, "join"      ) == 0) return OCSM_JOIN;
    if (strcmp(text, "extract"   ) == 0) return OCSM_EXTRACT;
    if (strcmp(text, "combine"   ) == 0) return OCSM_COMBINE;

    /* OCSM_TRANSFORM */
    if (strcmp(text, "translate" ) == 0) return OCSM_TRANSLATE;
    if (strcmp(text, "rotatex"   ) == 0) return OCSM_ROTATEX;
    if (strcmp(text, "rotatey"   ) == 0) return OCSM_ROTATEY;
    if (strcmp(text, "rotatez"   ) == 0) return OCSM_ROTATEZ;
    if (strcmp(text, "scale"     ) == 0) return OCSM_SCALE;
    if (strcmp(text, "mirror"    ) == 0) return OCSM_MIRROR;
    if (strcmp(text, "reorder"   ) == 0) return OCSM_REORDER;

    /* OCSM_SKETCH */
    if (strcmp(text, "skbeg"     ) == 0) return OCSM_SKBEG;
    if (strcmp(text, "skcon"     ) == 0) return OCSM_SKCON;
    if (strcmp(text, "skvar"     ) == 0) return OCSM_SKVAR;
    if (strcmp(text, "linseg"    ) == 0) return OCSM_LINSEG;
    if (strcmp(text, "cirarc"    ) == 0) return OCSM_CIRARC;
    if (strcmp(text, "arc"       ) == 0) return OCSM_ARC;
    if (strcmp(text, "spline"    ) == 0) return OCSM_SPLINE;
    if (strcmp(text, "bezier"    ) == 0) return OCSM_BEZIER;
    if (strcmp(text, "skend"     ) == 0) return OCSM_SKEND;

    /* OCSM_SOLVER */
    if (strcmp(text, "solbeg"    ) == 0) return OCSM_SOLBEG;
    if (strcmp(text, "solcon"    ) == 0) return OCSM_SOLCON;
    if (strcmp(text, "solend"    ) == 0) return OCSM_SOLEND;

    /* OCSM_UTILITY */
    if (strcmp(text, "interface" ) == 0) return OCSM_INTERFACE;
    if (strcmp(text, "end"       ) == 0) return OCSM_END;
    if (strcmp(text, "set"       ) == 0) return OCSM_SET;
    if (strcmp(text, "udparg"    ) == 0) return OCSM_UDPARG;
    if (strcmp(text, "select"    ) == 0) return OCSM_SELECT;
    if (strcmp(text, "project"   ) == 0) return OCSM_PROJECT;
    if (strcmp(text, "macbeg"    ) == 0) return OCSM_MACBEG;
    if (strcmp(text, "macend"    ) == 0) return OCSM_MACEND;
    if (strcmp(text, "recall"    ) == 0) return OCSM_RECALL;
    if (strcmp(text, "store"     ) == 0) return OCSM_STORE;
    if (strcmp(text, "patbeg"    ) == 0) return OCSM_PATBEG;
    if (strcmp(text, "patbreak"  ) == 0) return OCSM_PATBREAK;
    if (strcmp(text, "patend"    ) == 0) return OCSM_PATEND;
    if (strcmp(text, "mark"      ) == 0) return OCSM_MARK;
    if (strcmp(text, "dump"      ) == 0) return OCSM_DUMP;
    if (strcmp(text, "assert"    ) == 0) return OCSM_ASSERT;

    /* Branch classes */
    if (strcmp(text, "primitive" ) == 0) return OCSM_PRIMITIVE;
    if (strcmp(text, "grown"     ) == 0) return OCSM_GROWN;
    if (strcmp(text, "applied"   ) == 0) return OCSM_APPLIED;
    if (strcmp(text, "boolean"   ) == 0) return OCSM_BOOLEAN;
    if (strcmp(text, "transform" ) == 0) return OCSM_TRANSFORM;
    if (strcmp(text, "sketch"    ) == 0) return OCSM_SKETCH;
    if (strcmp(text, "solver"    ) == 0) return OCSM_SOLVER;
    if (strcmp(text, "utility"   ) == 0) return OCSM_UTILITY;

    /* Branch activities */
    if (strcmp(text, "active"    ) == 0) return OCSM_ACTIVE;
    if (strcmp(text, "suppressed") == 0) return OCSM_SUPPRESSED;
    if (strcmp(text, "inactive"  ) == 0) return OCSM_INACTIVE;
    if (strcmp(text, "deferred"  ) == 0) return OCSM_DEFERRED;

    /* Body types */
    if (strcmp(text, "solid_body") == 0) return OCSM_SOLID_BODY;
    if (strcmp(text, "sheet_body") == 0) return OCSM_SHEET_BODY;
    if (strcmp(text, "wire_body" ) == 0) return OCSM_WIRE_BODY;
    if (strcmp(text, "node_body" ) == 0) return OCSM_NODE_BODY;
    if (strcmp(text, "null_body" ) == 0) return OCSM_NULL_BODY;

    /* Parameter types */
    if (strcmp(text, "external"  ) == 0) return OCSM_EXTERNAL;
    if (strcmp(text, "internal"  ) == 0) return OCSM_INTERNAL;

    /* Selector type */
    if (strcmp(text, "node"      ) == 0) return OCSM_NODE;
    if (strcmp(text, "edge"      ) == 0) return OCSM_EDGE;
    if (strcmp(text, "face"      ) == 0) return OCSM_FACE;
    if (strcmp(text, "volume"    ) == 0) return OCSM_VOLUME;

    return 0;
}


/*
 ************************************************************************
 *                                                                      *
 *   buildApplied - implement OCSM_APPLIEDs for ocsmBuild               *
 *                                                                      *
 ************************************************************************
 */

static int
buildApplied(modl_T *modl,
             int    ibrch,
             varg_T args[],
             int    *nstack,
             int    stack[],
             int    npatn,
             patn_T patn[])
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int        nelist, *ielist=NULL, *iford1=NULL, *iford2=NULL;
    int        itype, nlist, type, hasdots, i, j, iedge, iface, nface, nedge;
    int        jford1, jford2, ibody, ibodyl, iprnt, jbrch, ipair, npair;
    CINT       *tempIlist;
    double     toler;
    CDOUBLE    *tempRlist;
    CCHAR      *tempClist;

    int        iford, nremove, nf, nchild, oclass, mtype, *senses;
    int        nfacemap, *facemap1=NULL, *facemap2=NULL, iswap;
    double     data[4];
    ego        ebody, ebodyl, *eedges, *efaces, eref;
    ego        *eelist=NULL, *eflist=NULL, *echilds, eshell;

    ROUTINE(buildApplied);
    DPRINT2("%s(ibrch=%d) {",
            routine, ibrch);

    /* --------------------------------------------------------------- */

    type    = MODL->brch[ibrch].type;
    hasdots = 0;

    /* execute: "fillet radius edgeList=0" */
    if (type == OCSM_FILLET) {
        SPRINT3(1, "    executing [%4d] fillet:     %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0]);
        if (args[1].dot[0] != 0) {
            hasdots = 1;

            SPRINT1(1, "                                 %11.5f",
                       args[1].dot[0]);
        }

        /* check for a positive radius */
        if (args[1].val[0] <= 0) {
            SET_STATUS(OCSM_ILLEGAL_ARGUMENT, fillet);
        }

        /* pop a Body from the stack */
        if ((*nstack) < 1) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "FILLET expects a Body on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, fillet);
        } else {
            ibodyl = stack[--(*nstack)];
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* check that ibodyl is not a Sketch */
        if (MODL->body[ibodyl].botype != OCSM_SOLID_BODY) {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                        "FILLET does not expect a Sketch on the stack");
            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, fillet);
        }

        /* cycle back from ibodyl to find the closest left parent Body that
           was either a PRIMITIVE, GROWN, or BOOLEAN */
        iprnt = ibodyl;
        while (iprnt != 0) {
            jbrch = MODL->body[iprnt].ibrch;
            if (MODL->brch[jbrch].class == OCSM_PRIMITIVE ||
                MODL->brch[jbrch].class == OCSM_GROWN     ||
                MODL->brch[jbrch].class == OCSM_BOOLEAN     ) break;
            iprnt = MODL->body[iprnt].ileft;
        }

        /* initialize the list of Edges to which fillets should be applied */
        nelist = 0;
        MALLOC(ielist, int, MODL->body[ibodyl].nedge);

        /* apply fillet to all Edges if edgeList=0 OR iprnt is a Boolean */
        if ((args[2].nval == 1 && args[2].val[0] == 0) ||
            MODL->body[iprnt].brtype == OCSM_INTERSECT ||
            MODL->body[iprnt].brtype == OCSM_SUBTRACT  ||
            MODL->body[iprnt].brtype == OCSM_UNION     ||
            MODL->body[iprnt].brtype == OCSM_JOIN        ) {

            ebodyl = MODL->body[ibodyl].ebody;

            status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedge, &eedges);
            CHECK_STATUS(EG_getBodyTopos);

            for (iedge = 1; iedge <= nedge; iedge++) {
                status = EG_attributeRet(eedges[iedge-1], "body",
                                         &itype, &nlist,
                                         &tempIlist, &tempRlist, &tempClist);
                CHECK_STATUS(EG_attributeRet);

                if (tempIlist[0] == iprnt) {
                    ielist[nelist++] = iedge;
                }
            }

            EG_free(eedges);

        /* otherwise, process the edgeList (in order) */
        } else {

            if (args[2].nval%2 != 0) {
                signalError(MODL, OCSM_ILLEGAL_PMTR_NAME, ibrch,
                            "not an even number of entries in edgeList");
                SET_STATUS(OCSM_ILLEGAL_PMTR_NAME, fillet);
            }

            npair = args[2].nval / 2;
            MALLOC(iford1, int, npair);
            MALLOC(iford2, int, npair);

            for (ipair = 0; ipair < npair; ipair++) {
                iford1[ipair] = NINT(args[2].val[2*ipair  ]);
                iford2[ipair] = NINT(args[2].val[2*ipair+1]);
            }

            /* make sure that we have at least one pair */
            if (npair <= 0) {
                SET_STATUS(OCSM_ILLEGAL_PMTR_NAME, fillet);
            }

            /* process the pairs */
            for (ipair = 0; ipair < npair; ipair++) {

                /* add all Edges to ielist */
                if (iford1[ipair] == 0 && iford2[ipair] == 0) {
                    ebodyl = MODL->body[ibodyl].ebody;

                    status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedge, &eedges);
                    CHECK_STATUS(EG_getBodyTopos);

                    for (iedge = 1; iedge <= nedge; iedge++) {
                        status = EG_attributeRet(eedges[iedge-1], "body",
                                                 &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CHECK_STATUS(EG_attributeRet);

                        if (tempIlist[0] == iprnt) {
                            ielist[nelist++] = iedge;
                        }
                    }

                    EG_free(eedges);

                /* add Edges adjacent to +iford1[ipair] */
                } else if (iford1[ipair] > 0 && iford2[ipair] == 0) {
                    ebodyl = MODL->body[ibodyl].ebody;

                    status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedge, &eedges);
                    CHECK_STATUS(EG_getBodyTopos);

                    for (iedge = 1; iedge <= nedge; iedge++) {
                        status = EG_attributeRet(eedges[iedge-1], "body",
                                                 &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CHECK_STATUS(EG_attributeRet);

                        jford1 = tempIlist[1] % 100;
                        jford2 = tempIlist[1] / 100;

                        if (tempIlist[0] == iprnt) {
                            if (jford1 == iford1[ipair] || jford2 == iford1[ipair]) {
                                ielist[nelist++] = iedge;
                            }
                        }
                    }

                    EG_free(eedges);

                /* remove all Edges adjacent to -iford1[ipair] */
                } else if (iford1[ipair] < 0 && iford2[ipair] == 0) {
                    ebodyl = MODL->body[ibodyl].ebody;

                    status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedge, &eedges);
                    CHECK_STATUS(EG_getBodyTopos);

                    for (iedge = 1; iedge <= nedge; iedge++) {
                        status = EG_attributeRet(eedges[iedge-1], "body",
                                                 &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CHECK_STATUS(EG_attributeRet);

                        jford1 = tempIlist[1] % 100;
                        jford2 = tempIlist[1] / 100;

                        if (tempIlist[0] == iprnt) {
                            if (jford1 == -iford1[ipair] || jford2 == -iford1[ipair]) {
                                for (j = 0; j < nelist; j++) {
                                    if (ielist[j] == iedge) {
                                        ielist[j] = 0;
                                    }
                                }
                            }
                        }
                    }

                    EG_free(eedges);

                /* add Edges adjacent to +iford1[ipair] and +iford2[ipair] */
                } else if (iford1[ipair] > 0 && iford2[ipair] > 0) {
                    ebodyl = MODL->body[ibodyl].ebody;

                    status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedge, &eedges);
                    CHECK_STATUS(EG_getBodyTopos);

                    for (iedge = 1; iedge <= nedge; iedge++) {
                        status = EG_attributeRet(eedges[iedge-1], "body",
                                                 &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CHECK_STATUS(EG_attributeRet);

                        jford1 = tempIlist[1] % 100;
                        jford2 = tempIlist[1] / 100;

                        if (tempIlist[0] == iprnt) {
                            if        (jford1 == iford1[ipair] && jford2 == iford2[ipair]) {
                                ielist[nelist++] = iedge;
                            } else if (jford1 == iford2[ipair] && jford2 == iford1[ipair]) {
                                ielist[nelist++] = iedge;
                            }
                        }
                    }

                    EG_free(eedges);

                /* remove Edges adjacent to -iford1[ipair] and -iford2[ipair] */
                } else if (iford1[ipair] < 0 && iford2[ipair] < 0) {
                    ebodyl = MODL->body[ibodyl].ebody;

                    status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedge, &eedges);
                    CHECK_STATUS(EG_getBodyTopos);

                    for (iedge = 1; iedge <= nedge; iedge++) {
                        status = EG_attributeRet(eedges[iedge-1], "body",
                                                 &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CHECK_STATUS(EG_attributeRet);

                        jford1 = tempIlist[1] % 100;
                        jford2 = tempIlist[1] / 100;

                        if (tempIlist[0] == iprnt) {
                            if ((jford1 == -iford1[ipair] && jford2 == -iford2[ipair]) ||
                                (jford1 == -iford2[ipair] && jford2 == -iford1[ipair])   ) {
                                for (j = 0; j < nelist; j++) {
                                    if (ielist[j] == iedge) {
                                        ielist[j] = 0;
                                    }
                                }
                            }
                        }
                    }

                    EG_free(eedges);
                }
            }
        }

        /* remove deleted entries from the ielist */
        for (j = 0; j < nelist; j++) {
            if (ielist[j] == 0) {
                ielist[j] = ielist[nelist-1];
                nelist--;
                j--;
            }
        }

        for (i = 0; i < nelist; i++) {
            DPRINT2("ielist[%3d] = %4d", i, ielist[i]);
        }
        DPRINT1("radius      = %10.5f", args[1].val[0]);

        /* create the fillets */
        if (nelist > 0) {

            /* create the new Body */
            status = newBody(MODL, ibrch, OCSM_FILLET, ibodyl, -1,
                             args, hasdots, OCSM_SOLID_BODY, &ibody);
            CHECK_STATUS(newBody);

            ebodyl = MODL->body[ibodyl].ebody;

            status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedge, &eedges);
            CHECK_STATUS(EG_getBodyTopos);

            MALLOC(eelist, ego, nelist);

            for (i = 0; i < nelist; i++) {
                SPRINT1(2, "        fillet with iedge=%d", ielist[i]);
                eelist[i] = eedges[ielist[i]-1];
            }

            EG_free(eedges);

            if (outLevel >= 3) {
                SPRINT0(3, "before EG_filletBody: ebodyl");
                printEgo(ebodyl);
            }
            status = EG_filletBody(ebodyl, nelist, eelist, args[1].val[0], &ebody, NULL);
            CHECK_STATUS(EG_filletBody);

            FREE(ielist);
            FREE(eelist);

            MODL->body[ibody].ebody = ebody;

            /* mark the new Faces with the current Body */
            status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
            CHECK_STATUS(EG_getBodyTopos);

            for (iface = 1; iface <= nface; iface++) {
                status = EG_attributeRet(efaces[iface-1], "body",
                                         &itype, &nlist,
                                         &tempIlist, &tempRlist, &tempClist);
                if (status == EGADS_NOTFOUND) {
                    status = setFaceAttribute(MODL, ibody, iface, 0, iface, npatn, patn);
                    CHECK_STATUS(setFaceAttribute);
                } else {
                    CHECK_STATUS(EG_attributeRet);
                }
            }

            EG_free(efaces);

            /* finish the Body */
            status = finishBody(MODL, ibody);
            CHECK_STATUS(finishBody);
        } else {
            SPRINT0(0, "WARNING:: no edges for fillet");
            ibody = ibodyl;
        }

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "chamfer radius edgeList=0" */
    } else if (type == OCSM_CHAMFER) {
        SPRINT3(1, "    executing [%4d] chamfer:    %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0]);
        if (args[1].dot[0] != 0) {
            hasdots = 1;

            SPRINT1(1, "                                 %11.5f",
                       args[1].dot[0]);
        }

        /* check for a positive radius */
        if (args[1].val[0] <= 0) {
            SET_STATUS(OCSM_ILLEGAL_ARGUMENT, chamfer);
        }

        /* pop a Body from the stack */
        if ((*nstack) < 1) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "CHAMFER expects a Body on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, chamfer);
        } else {
            ibodyl = stack[--(*nstack)];
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* check that ibodyl is not a Sketch */
        if (MODL->body[ibodyl].botype != OCSM_SOLID_BODY) {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                        "CHAMFER does not expect a Sketch on the stack");
            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, chamfer);
        }

        /* cycle back from ibodyl to find the closest left parent Body that
           was either a PRIMITIVE, GROWN, or BOOLEAN */
        iprnt = ibodyl;
        while (iprnt != 0) {
            jbrch = MODL->body[iprnt].ibrch;
            if (MODL->brch[jbrch].class == OCSM_PRIMITIVE ||
                MODL->brch[jbrch].class == OCSM_GROWN     ||
                MODL->brch[jbrch].class == OCSM_BOOLEAN     ) break;
            iprnt = MODL->body[iprnt].ileft;
        }

        /* initialize the list of Edges to which chamfers should be applied */
        nelist = 0;
        MALLOC(ielist, int, MODL->body[ibodyl].nedge);

        /* apply chamfer to all Edges if edgeList=0 OR iprnt is a Boolean */
        if ((args[2].nval == 1 && args[2].val[0] == 0) ||
            MODL->body[iprnt].brtype == OCSM_INTERSECT ||
            MODL->body[iprnt].brtype == OCSM_SUBTRACT  ||
            MODL->body[iprnt].brtype == OCSM_UNION     ||
            MODL->body[iprnt].brtype == OCSM_JOIN        ) {

            ebodyl = MODL->body[ibodyl].ebody;

            status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedge, &eedges);
            CHECK_STATUS(EG_getBodyTopos);

            for (iedge = 1; iedge <= nedge; iedge++) {
                status = EG_attributeRet(eedges[iedge-1], "body",
                                         &itype, &nlist,
                                         &tempIlist, &tempRlist, &tempClist);
                CHECK_STATUS(EG_attributeRet);

                if (tempIlist[0] == iprnt) {
                    ielist[nelist++] = iedge;
                }
            }

            EG_free(eedges);

        /* otherwise, process the edgeList (in order) */
        } else {

            if (args[2].nval%2 != 0) {
                signalError(MODL, OCSM_ILLEGAL_PMTR_NAME, ibrch,
                            "not an even number of entries in edgeList");
                SET_STATUS(OCSM_ILLEGAL_PMTR_NAME, fillet);
            }

            npair = args[2].nval / 2;
            MALLOC(iford1, int, npair);
            MALLOC(iford2, int, npair);

            for (ipair = 0; ipair < npair; ipair++) {
                iford1[ipair] = NINT(args[2].val[2*ipair  ]);
                iford2[ipair] = NINT(args[2].val[2*ipair+1]);
            }

            /* make sure that we have at least one pair */
            if (npair <= 0) {
                SET_STATUS(OCSM_ILLEGAL_PMTR_NAME, fillet);
            }

            /* process the pairs */
            for (ipair = 0; ipair < npair; ipair++) {

                /* add all Edges to ielist */
                if (iford1[ipair] == 0 && iford2[ipair] == 0) {
                    ebodyl = MODL->body[ibodyl].ebody;

                    status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedge, &eedges);
                    CHECK_STATUS(EG_getBodyTopos);

                    for (iedge = 1; iedge <= nedge; iedge++) {
                        status = EG_attributeRet(eedges[iedge-1], "body",
                                                 &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CHECK_STATUS(EG_attributeRet);

                        if (tempIlist[0] == iprnt) {
                            ielist[nelist++] = iedge;
                        }
                    }

                    EG_free(eedges);

                /* add Edges adjacent to +iford1[ipair] */
                } else if (iford1[ipair] > 0 && iford2[ipair] == 0) {
                    ebodyl = MODL->body[ibodyl].ebody;

                    status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedge, &eedges);
                    CHECK_STATUS(EG_getBodyTopos);

                    for (iedge = 1; iedge <= nedge; iedge++) {
                        status = EG_attributeRet(eedges[iedge-1], "body",
                                                 &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CHECK_STATUS(EG_attributeRet);

                        jford1 = tempIlist[1] % 100;
                        jford2 = tempIlist[1] / 100;

                        if (tempIlist[0] == iprnt) {
                            if (jford1 == iford1[ipair] || jford2 == iford1[ipair]) {
                                ielist[nelist++] = iedge;
                            }
                        }
                    }

                    EG_free(eedges);

                /* remove all Edges adjacent to -iford1[ipair] */
                } else if (iford1[ipair] < 0 && iford2[ipair] == 0) {
                    ebodyl = MODL->body[ibodyl].ebody;

                    status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedge, &eedges);
                    CHECK_STATUS(EG_getBodyTopos);

                    for (iedge = 1; iedge <= nedge; iedge++) {
                        status = EG_attributeRet(eedges[iedge-1], "body",
                                                 &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CHECK_STATUS(EG_attributeRet);

                        jford1 = tempIlist[1] % 100;
                        jford2 = tempIlist[1] / 100;

                        if (tempIlist[0] == iprnt) {
                            if (jford1 == -iford1[ipair] || jford2 == -iford1[ipair]) {
                                for (j = 0; j < nelist; j++) {
                                    if (ielist[j] == iedge) {
                                        ielist[j] = 0;
                                    }
                                }
                            }
                        }
                    }

                    EG_free(eedges);

                /* add Edges adjacent to +iford1[ipair] and +iford2[ipair] */
                } else if (iford1[ipair] > 0 && iford2[ipair] > 0) {
                    ebodyl = MODL->body[ibodyl].ebody;

                    status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedge, &eedges);
                    CHECK_STATUS(EG_getBodyTopos);

                    for (iedge = 1; iedge <= nedge; iedge++) {
                        status = EG_attributeRet(eedges[iedge-1], "body",
                                                 &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CHECK_STATUS(EG_attributeRet);

                        jford1 = tempIlist[1] % 100;
                        jford2 = tempIlist[1] / 100;

                        if (tempIlist[0] == iprnt) {
                            if        (jford1 == iford1[ipair] && jford2 == iford2[ipair]) {
                                ielist[nelist++] = iedge;
                            } else if (jford1 == iford2[ipair] && jford2 == iford1[ipair]) {
                                ielist[nelist++] = iedge;
                            }
                        }
                    }

                    EG_free(eedges);

                /* remove Edges adjacent to -iford1[ipair] and -iford2[ipair] */
                } else if (iford1[ipair] < 0 && iford2[ipair] < 0) {
                    ebodyl = MODL->body[ibodyl].ebody;

                    status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedge, &eedges);
                    CHECK_STATUS(EG_getBodyTopos);

                    for (iedge = 1; iedge <= nedge; iedge++) {
                        status = EG_attributeRet(eedges[iedge-1], "body",
                                                 &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CHECK_STATUS(EG_attributeRet);

                        jford1 = tempIlist[1] % 100;
                        jford2 = tempIlist[1] / 100;

                        if (tempIlist[0] == iprnt) {
                            if ((jford1 == -iford1[ipair] && jford2 == -iford2[ipair]) ||
                                (jford1 == -iford2[ipair] && jford2 == -iford1[ipair])   ) {
                                for (j = 0; j < nelist; j++) {
                                    if (ielist[j] == iedge) {
                                        ielist[j] = 0;
                                    }
                                }
                            }
                        }
                    }

                    EG_free(eedges);
                }
            }
        }

        /* remove deleted entries from the ielist */
        for (j = 0; j < nelist; j++) {
            if (ielist[j] == 0) {
                ielist[j] = ielist[nelist-1];
                nelist--;
                j--;
            }
        }

        for (i = 0; i < nelist; i++) {
            DPRINT2("ielist[%3d] = %4d", i, ielist[i]);
        }
        DPRINT1("radius      = %10.5f", args[1].val[0]);

        /* create the chamfers */
        if (nelist > 0) {

            /* create the new Body */
            status = newBody(MODL, ibrch, OCSM_CHAMFER, ibodyl, -1,
                             args, hasdots, OCSM_SOLID_BODY, &ibody);
            CHECK_STATUS(newBody);

            ebodyl = MODL->body[ibodyl].ebody;

            status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedge, &eedges);
            CHECK_STATUS(EG_getBodyTopos);

            MALLOC(eelist, ego, nelist);
            MALLOC(eflist, ego, nelist);

            for (i = 0; i < nelist; i++) {
                SPRINT1(2, "        chamfer with iedge=%d", ielist[i]);
                eelist[i] = eedges[ielist[i]-1];

                status = EG_getBodyTopos(ebodyl, eelist[i], FACE, &nf, &efaces);
                CHECK_STATUS(EG_getBodyTopos);

                eflist[i] = efaces[0];
                EG_free(efaces);
            }

            EG_free(eedges);

            if (outLevel >= 3) {
                SPRINT0(3, "before EG_chamferBody: ebodyl");
                printEgo(ebodyl);
            }
            status = EG_chamferBody(ebodyl, nelist, eelist, eflist, args[1].val[0], args[1].val[0], &ebody, NULL);
            CHECK_STATUS(EG_chamferBody);

            FREE(ielist);
            FREE(eelist);
            FREE(eflist);

            MODL->body[ibody].ebody = ebody;

            /* mark the new Faces with the current Body */
            status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
            CHECK_STATUS(EG_getBodyTopos);

            for (iface = 1; iface <= nface; iface++) {
                status = EG_attributeRet(efaces[iface-1], "body",
                                         &itype, &nlist,
                                         &tempIlist, &tempRlist, &tempClist);
                if (status == EGADS_NOTFOUND) {
                    status = setFaceAttribute(MODL, ibody, iface, 0, iface, npatn, patn);
                    CHECK_STATUS(setFaceAttribute);
                } else {
                    CHECK_STATUS(EG_attributeRet);
                }
            }

            EG_free(efaces);

            /* finish the Body */
            status = finishBody(MODL, ibody);
            CHECK_STATUS(finishBody);
        } else {
            SPRINT0(0, "WARNING:: no edges for chamfer");
            ibody = ibodyl;
        }

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "hollow thick faceList=0" */
    } else if (type == OCSM_HOLLOW) {
        SPRINT3(1, "    executing [%4d] hollow:     %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0]);
        if (args[1].dot[0] != 0) {
            hasdots = 1;

            SPRINT1(1, "                                 %11.5f",
                    args[1].dot[0]);
        }

        /* pop an Body from the stack */
        if ((*nstack) < 1) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "HOLLOW expects a Body on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, hollow);
        } else {
            ibodyl = stack[--(*nstack)];
        }

        /* normal hollow command */
        if (args[1].val[0] != 0) {

            /* recycle old Body if not dirty */
            status = recycleBody(MODL, ibrch, type, args, hasdots);
            CHECK_STATUS(recycleBody);

            if (status == 1) {
                stack[(*nstack)++] = MODL->nbody;
                status = 0;
                goto cleanup;
            }

            /* create the Body */
            status = newBody(MODL, ibrch, OCSM_HOLLOW, ibodyl, -1,
                             args, hasdots, OCSM_SOLID_BODY, &ibody);
            CHECK_STATUS(newBody);

            if (args[2].nval == 1 && args[2].val[0] == 0) {
                nremove = 0;
            } else {
                /* make list of ifords */
                if (args[2].nval%2 != 0) {
                    signalError(MODL, OCSM_ILLEGAL_PMTR_NAME, ibrch,
                                "not an even number of entries in faceList");
                    SET_STATUS(OCSM_ILLEGAL_PMTR_NAME, hollow);
                }

                npair = args[2].nval / 2;
                MALLOC(iford1, int, npair);
                MALLOC(iford2, int, npair);

                for (i = 0; i < npair; i++) {
                    iford1[i] = NINT(args[2].val[2*i  ]);
                    iford2[i] = NINT(args[2].val[2*i+1]);
                }

                nremove = 0;
                MALLOC(eflist, ego, MODL->body[ibodyl].nface);

                ebodyl = MODL->body[ibodyl].ebody;

                status = EG_getBodyTopos(ebodyl, NULL, FACE, &nface, &efaces);
                CHECK_STATUS(EG_getBodyTopos);

                for (iface = 1; iface <= nface; iface++) {
                    status = EG_attributeRet(efaces[iface-1], "body",
                                             &itype, &nlist,
                                             &tempIlist, &tempRlist, &tempClist);
                    CHECK_STATUS(EG_attributeRet);

                    for (ipair = 0; ipair < npair; ipair++) {
                        if (tempIlist[0] == iford1[ipair] &&
                            tempIlist[1] == iford2[ipair]   ) {
                            eflist[nremove++] = efaces[iface-1];
                        }
                    }
                }

                EG_free(efaces);
            }

            for (i = 0; i < nremove; i++) {
                DPRINT2("eflist[%3d] = %lx", i, (long)eflist[i]);
            }
            DPRINT1("radius      = %10.5f", args[1].val[0]);

            /* generate the hollow */
            ebodyl = MODL->body[ibodyl].ebody;

            if (outLevel >= 3) {
                SPRINT0(3, "before EG_hollowBody: ebodyl");
                printEgo(ebodyl);
            }
            status = EG_hollowBody(ebodyl, nremove, eflist, args[1].val[0], 1, &ebody, &facemap1);
            CHECK_STATUS(EG_hollowBody);

            MODL->body[ibody].ebody = ebody;

        /* convert SolidBody to SheetBody */
        } else if (MODL->body[ibodyl].botype == OCSM_SOLID_BODY) {
            status = EG_copyObject(MODL->body[ibodyl].ebody, NULL, &ebodyl);
            CHECK_STATUS(EG_copyObject);

            status = EG_getBodyTopos(ebodyl, NULL, FACE, &nface, &efaces);
            CHECK_STATUS(EG_getBodyTopos);

            /* recycle old Body if not dirty */
            status = recycleBody(MODL, ibrch, type, args, hasdots);
            CHECK_STATUS(recycleBody);

            if (status == 1) {
                stack[(*nstack)++] = MODL->nbody;
                status = 0;
                goto cleanup;
            }

            /* make a Shell */
            status = EG_makeTopology(MODL->context, NULL, SHELL, CLOSED,
                                     NULL, nface, efaces, NULL, &eshell);
            CHECK_STATUS(EG_makeTopology);

            /* make the new SheetBody */
            status = newBody(MODL, ibrch, OCSM_HOLLOW, ibodyl, -1,
                             args, hasdots, OCSM_SHEET_BODY, &ibody);
            CHECK_STATUS(newBody);

            status = EG_makeTopology(MODL->context, NULL, BODY, SHEETBODY,
                                     NULL, 1, &eshell, NULL, &ebody);
            CHECK_STATUS(EG_makeTopology);

            EG_free(efaces);

            MODL->body[ibody].ebody = ebody;

       /* convert SheetBody to WireBody */
        } else if (MODL->body[ibodyl].botype == OCSM_SHEET_BODY) {

            /* recycle old Body if not dirty */
            status = recycleBody(MODL, ibrch, type, args, hasdots);
            CHECK_STATUS(recycleBody);

            if (status == 1) {
                stack[(*nstack)++] = MODL->nbody;
                status = 0;
                goto cleanup;
            }

            status = EG_copyObject(MODL->body[ibodyl].ebody, NULL, &ebodyl);
            CHECK_STATUS(EG_copyObject);

            /* get FACE from SheetBody */
            status = EG_getTopology(ebodyl, &eref, &oclass, &mtype,
                                    data, &nface, &efaces, &senses);
            CHECK_STATUS(EG_getTopology);

            if (nface != 1) {
                signalError(MODL, OCSM_INTERNAL_ERROR, ibrch,
                            "SHEET has more than one FACE");
                SET_STATUS(OCSM_INTERNAL_ERROR, hollow);
            }

            /* get LOOP from FACE */
            status = EG_getTopology(efaces[0], &eref, &oclass, &mtype,
                                    data, &nchild, &echilds, &senses);
            CHECK_STATUS(EG_getTopology);

            if (nchild != 1) {
                signalError(MODL, OCSM_INTERNAL_ERROR, ibrch,
                            "SHEET has more than one LOOP");
                SET_STATUS(OCSM_INTERNAL_ERROR, hollow);
            }

            /* make the new WireBody */
            status = newBody(MODL, ibrch, OCSM_HOLLOW, ibodyl, -1,
                             args, hasdots, OCSM_WIRE_BODY, &ibody);
            CHECK_STATUS(newBody);

            status = EG_makeTopology(MODL->context, NULL, BODY, WIREBODY,
                                     NULL, 1, echilds, NULL, &ebody);
            CHECK_STATUS(EG_makeTopology);

            MODL->body[ibody].ebody = ebody;

        /* otherwise an error */
        } else {
            SET_STATUS(OCSM_ILLEGAL_VALUE, hollow);
        }

        /* mark the new Faces with the current Branch */
        status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        if (facemap1 == NULL) {
            iford = 0;
            for (iface = 1; iface <= nface; iface++) {
                status = EG_attributeRet(efaces[iface-1], "body",
                                         &itype, &nlist,
                                         &tempIlist, &tempRlist, &tempClist);

                if (status == EGADS_NOTFOUND) {
                    iford++;
                    status = setFaceAttribute(MODL, ibody, iface, 0, iford, npatn, patn);
                    CHECK_STATUS(setFaceAttribute);
                } else {
                    CHECK_STATUS(EG_attributeRet);
                }
            }
        } else {
            MALLOC(facemap2, int, 4*nface);

            /* make a list of the faceID of the Face from which new Faces are offset */
            nfacemap = 0;
            for (iface = 1; iface <= nface; iface++) {
                status = EG_attributeRet(efaces[iface-1], "body",
                                         &itype, &nlist,
                                         &tempIlist, &tempRlist, &tempClist);

                if (status == EGADS_NOTFOUND) {
                    if (facemap1[2*iface-2] != FACEOFF) {
                        signalError(MODL, OCSM_INTERNAL_ERROR, ibrch,
                                    "facemap1[iface=%d]=%d %d", iface, facemap1[2*iface-2], facemap1[2*iface-1]);
                        status = OCSM_INTERNAL_ERROR;
                        goto cleanup;
                    } else {
                        i = facemap1[2*iface-1];

                        status = EG_attributeRet(MODL->body[ibodyl].face[i].eface, "faceID",
                                                 &itype, &nlist,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CHECK_STATUS(EG_attributeRet);

                        facemap2[4*nfacemap  ] = iface;
                        facemap2[4*nfacemap+1] = tempIlist[0];
                        facemap2[4*nfacemap+2] = tempIlist[1];
                        facemap2[4*nfacemap+3] = tempIlist[2];
                        nfacemap++;
                    }
                }
            }

            /* sort the facemap2 in faceID order */
            for (i = 0; i < nfacemap; i++) {
                for (j = i+1; j < nfacemap; j++) {
                    iswap = 0;

                    if (facemap2[4*j+1] < facemap2[4*i+1]) {
                        iswap = 1;
                    } else if (facemap2[4*j+1] == facemap2[4*i+1]) {
                        if (facemap2[4*j+2] < facemap2[4*i+2]) {
                            iswap = 1;
                        } else if (facemap2[4*j+2] == facemap2[4*i+2]) {
                            if (facemap2[4*j+3] < facemap2[4*i+3]) {
                                iswap = 1;
                            }
                        }
                    }

                    if (iswap > 0) {
                        iswap           = facemap2[4*i  ];
                        facemap2[4*i  ] = facemap2[4*j  ];
                        facemap2[4*j  ] = iswap;

                        iswap           = facemap2[4*i+1];
                        facemap2[4*i+1] = facemap2[4*j+1];
                        facemap2[4*j+1] = iswap;

                        iswap           = facemap2[4*i+2];
                        facemap2[4*i+2] = facemap2[4*j+2];
                        facemap2[4*j+2] = iswap;

                        iswap           = facemap2[4*i+3];
                        facemap2[4*i+3] = facemap2[4*j+3];
                        facemap2[4*j+3] = iswap;
                    }
                }
            }

            /* apply the faceID to the Faces in facemap2 */
            for (i = 0; i < nfacemap; i++) {
                iface = facemap2[4*i  ];

                status = setFaceAttribute(MODL, ibody, iface, 0, i+1, npatn, patn);
                CHECK_STATUS(setFaceAttribute);
            }
            FREE(   facemap2);
            EG_free(facemap1);
            facemap1 = NULL;
        }

        EG_free(efaces);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    }

cleanup:
    FREE(eflist);
    FREE(facemap2);
    if (facemap1 != NULL) EG_free(facemap1);

    FREE(iford1);
    FREE(iford2);
    FREE(ielist);

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   buildBoolean - implement OCSM_BOOLEANs for ocsmBuild               *
 *                                                                      *
 ************************************************************************
 */

static int
buildBoolean(modl_T *modl,
             int    ibrch,
             varg_T args[],
             int    *nstack,
             int    stack[],
             int    npatn,
             patn_T patn[])
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int        type, hasdots, tomark;
    int        ibody, ibodyl, ibodyr, index;
    int        i, j, nchild, nchange, numBodys, bodyList[999];
    double     toler, maxtol;
    char       order[MAX_EXPR_LEN];

    int         oclass, mtype, *senses, nface, nshell;
    int         iedge, iface, senses1[1], nedges, bodyType;
    int         nmatch, *matches=NULL, nfacel, nfacer, nfacedg;
    int         inbodyl, inbodyr, ichild, ibest, iswap;
    int         *sense, imatch, iloop, ifacel, nloopl, nedgel, ifacer, nloopr, nedger;
    int         nsize, fillstyle=0, newFaces=0, itype, nlist, nedge;
    int         *indx=NULL, nindx, ntemp, ii, jj, newEdges=0, ibeg, iend, iokay;
    double      data[20], xyz[3], dirn[4], matrix[12], area;
    double      datal[18], datar[18], bboxl[6], bboxr[6], dtest, dbest;
    ego         ebody, ebodyl, ebodyr, emodel, eref, *ebodys, *echilds=NULL, etemp;
    ego         *eshells=NULL, *efaces=NULL, *eedges=NULL, *elist=NULL;
    ego         eedge, eface, eloop, *efacedgs;
    ego         esurf, *efacesl, *eloopsl, *eedgesl, *efacesr, *eloopsr, *eedgesr;

    int         ntopo, k, nattr, atype, len;
    CINT        *pints, *tempIlist;
    CDOUBLE     *preals, *tempRlist;
    CCHAR       *string, *aname, *tempClist;
    ego         *ptopos, *pchildren, etemp1, etemp2, exform, eflip;
    ego         *enewEdges=NULL, *enewFaces=NULL, eedges_in[4];

    ROUTINE(buildBoolean);
    DPRINT2("%s(ibrch=%d) {",
            routine, ibrch);

    /* --------------------------------------------------------------- */

    type    = MODL->brch[ibrch].type;
    hasdots = 0;

    /* execute: "intersect $order=none index=1 maxtol=0" */
    if (type == OCSM_INTERSECT) {
        SPRINT4(1, "    executing [%4d] intersect:  %s %11.5f %11.4e",
                ibrch, args[1].str, args[2].val[0], args[3].val[0]);

        /* pop 2 Bodys from the stack */
        if ((*nstack) < 2) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "INTERSECT expects 2 Bodys on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, intersect);
        } else {
            ibodyr = stack[--(*nstack)];
            ibodyl = stack[--(*nstack)];
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* check that ibodyl and ibodyr are not Sketches */
        if        (MODL->body[ibodyl].botype != OCSM_SOLID_BODY &&
                   MODL->body[ibodyl].botype != OCSM_SHEET_BODY   ) {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                        "INTERSECT expects left Body to be SOLID or SHEET Body");
            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, intersect);
        } else if (MODL->body[ibodyr].botype != OCSM_SOLID_BODY) {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                        "INTERSECT expects rite Body to be SOLID Body");
            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, intersect);
        }

        /* extract the arguments */
        STRNCPY(order, &(MODL->brch[ibrch].arg1[1]), MAX_EXPR_LEN);
        index = NINT(args[2].val[0]);

        /* create the new Body */
        if (MODL->body[ibodyl].botype == OCSM_SOLID_BODY) {
            status = newBody(MODL, ibrch, OCSM_INTERSECT, ibodyl, ibodyr,
                             args, hasdots, OCSM_SOLID_BODY, &ibody);
            CHECK_STATUS(newBody);
        } else {
            status = newBody(MODL, ibrch, OCSM_INTERSECT, ibodyl, ibodyr,
                             args, hasdots, OCSM_SHEET_BODY, &ibody);
            CHECK_STATUS(newBody);
        }

        if (MODL->body[ibodyl].botype == OCSM_SOLID_BODY) {
            ebodyl = MODL->body[ibodyl].ebody;
            ebodyr = MODL->body[ibodyr].ebody;
        } else {
            status = EG_copyObject(MODL->body[ibodyl].ebody, NULL, &ebody);
            CHECK_STATUS(EG_copyObject);

            status = EG_makeTopology(MODL->context, NULL, MODEL, 0,
                                     NULL, 1, &ebody, NULL, &ebodyl);
            CHECK_STATUS(EG_makeTopology);

            ebodyr = MODL->body[ibodyr].ebody;
        }

        status = solidBoolean(ebodyl, ebodyr, INTERSECTION, args[3].val[0], &emodel);
        CHECK_STATUS(solidBoolean);

        status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                                data, &nchild, &ebodys, &senses);
        CHECK_STATUS(EG_getTopology);

        if (nchild < 1) {
            SET_STATUS(OCSM_DID_NOT_CREATE_BODY, intersect);
        } else if (index > nchild) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "INTERSECT only created %d Bodys but index=%d", nchild, index);
            SET_STATUS(OCSM_ILLEGAL_VALUE, intersect);
        } else {
            i = selectBody(emodel, order, index);

            status = EG_copyObject(ebodys[i], NULL, &ebody);
            CHECK_STATUS(EG_copyObject);

            MODL->body[ibody].ebody = ebody;

            if (nchild > 1) {
                SPRINT1(0, "WARNING:: %d Bodys are being lost", nchild-1);
            }
        }

        status = EG_deleteObject(emodel);
        CHECK_STATUS(EG_deleteObject);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "subtract $order=none index=1" */
    } else if (type == OCSM_SUBTRACT) {
        SPRINT4(1, "    executing [%4d] subtract:   %s %11.5f %11.4e",
                ibrch, args[1].str, args[2].val[0], args[3].val[0]);

        /* pop 2 Bodys from the stack */
        if ((*nstack) < 2) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "SUBTRACT expects 2 Bodys on stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, subtract);
        } else {
            ibodyr = stack[--(*nstack)];
            ibodyl = stack[--(*nstack)];
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* extract the arguments */
        STRNCPY(order, &(MODL->brch[ibrch].arg1[1]), MAX_EXPR_LEN);
        index = NINT(args[2].val[0]);

        /* if both ibodyl and ibodyr are SolidBodys, perform Boolean
           opertion */
        if (MODL->body[ibodyl].botype == OCSM_SOLID_BODY &&
            MODL->body[ibodyr].botype == OCSM_SOLID_BODY   ) {

            /* create the new Body */
            status = newBody(MODL, ibrch, OCSM_SUBTRACT, ibodyl, ibodyr,
                             args, hasdots, OCSM_SOLID_BODY, &ibody);
            CHECK_STATUS(newBody);

            ebodyl = MODL->body[ibodyl].ebody;
            ebodyr = MODL->body[ibodyr].ebody;

            status = solidBoolean(ebodyl, ebodyr, SUBTRACTION, args[3].val[0], &emodel);
            CHECK_STATUS(solidBoolean);

            status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                                    data, &nchild, &ebodys, &senses);
            CHECK_STATUS(EG_getTopology);

            if (nchild < 1) {
                SET_STATUS(OCSM_DID_NOT_CREATE_BODY, subtract);
            } else if (index > nchild) {
                signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                            "SUBTRACT only created %d Bodys but index=%d", nchild, index);
                SET_STATUS(OCSM_ILLEGAL_VALUE, subtract);
            } else {
                i = selectBody(emodel, order, index);

                status = EG_copyObject(ebodys[i], NULL, &ebody);
                CHECK_STATUS(EG_copyObject);

                MODL->body[ibody].ebody = ebody;

                if (nchild > 1) {
                    SPRINT1(0, "WARNING:: %d Bodys are being lost", nchild-1);
                }
            }

            status = EG_deleteObject(emodel);
            CHECK_STATUS(EG_deleteObject);

        /* ibodyl is a SolidBody and ibodyr is a SheetBody, so scribe ibodyl */
        } else if (MODL->body[ibodyl].botype == OCSM_SOLID_BODY &&
                   MODL->body[ibodyr].botype == OCSM_SHEET_BODY   ) {

            ebodyl = MODL->body[ibodyl].ebody;
            ebodyr = MODL->body[ibodyr].ebody;

            if (outLevel >= 3) {
                SPRINT0(3, "before EG_intersection: ebodyl");
                printEgo(ebodyl);
                SPRINT0(3, "before EG_intersection: ebodyr");
                printEgo(ebodyr);
            }
            status = EG_intersection(ebodyl, ebodyr, &nfacedg, &efacedgs, &emodel);
            CHECK_STATUS(EG_intersection);

            if (outLevel >= 3) {
                SPRINT0(3, "before EG_imprintBody: ebodyl");
                printEgo(ebodyl);
            }
            status = EG_imprintBody(ebodyl, nfacedg, efacedgs, &ebody);
            CHECK_STATUS(EG_imprintBody);

            status = EG_deleteObject(emodel);
            CHECK_STATUS(EG_deleteObject);

            EG_free(efacedgs);

            /* create the new Body */
            status = newBody(MODL, ibrch, OCSM_SUBTRACT, ibodyl, ibodyr,
                             args, hasdots, OCSM_SOLID_BODY, &ibody);
            CHECK_STATUS(newBody);

            MODL->body[ibody].ebody = ebody;

        /* ibodyl is a SheetBody and ibodyr is a SheetBody, so scribe ibodyl */
        } else if (MODL->body[ibodyl].botype == OCSM_SHEET_BODY &&
                   MODL->body[ibodyr].botype == OCSM_SHEET_BODY   ) {

            ebodyl = MODL->body[ibodyl].ebody;
            ebodyr = MODL->body[ibodyr].ebody;

            if (outLevel >= 3) {
                SPRINT0(3, "before EG_intersection: ebodyl");
                printEgo(ebodyl);
                SPRINT0(3, "before EG_intersection: ebodyr");
                printEgo(ebodyr);
            }
            status = EG_intersection(ebodyl, ebodyr, &nfacedg, &efacedgs, &emodel);
            CHECK_STATUS(EG_intersection);

            if (outLevel >= 3) {
                SPRINT0(3, "before EG_imprintBody: ebodyl");
                printEgo(ebodyl);
            }
            status = EG_imprintBody(ebodyl, nfacedg, efacedgs, &ebody);
            CHECK_STATUS(EG_imprintBody);

            status = EG_deleteObject(emodel);
            CHECK_STATUS(EG_deleteObject);

            EG_free(efacedgs);

            /* create the new Body */
            status = newBody(MODL, ibrch, OCSM_SUBTRACT, ibodyl, ibodyr,
                             args, hasdots, OCSM_SHEET_BODY, &ibody);
            CHECK_STATUS(newBody);

            MODL->body[ibody].ebody = ebody;

        /* unsupported combination of Body types */
        } else {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                        "unsupported combination of Body types to SUBTRACT");
            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, subtract);
        }

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "union tomark=0 trimList=0 maxtol=0" */
    } else if (type == OCSM_UNION) {
        SPRINT4(1, "    executing [%4d] union:      %11.5f %11.5f %11.4e",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0]);

        /* extract the argument */
        tomark = NINT(args[1].val[0]);
        ibody  = 0;
        maxtol = args[3].val[0];

        /* union top two Bodys on the stack */
        if (tomark == 0) {
            /* pop 2 Bodys from the stack */
            if ((*nstack) < 2) {
                signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                            "UNION expects 2 Bodys on stack");
                SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, union);
            } else {
                ibodyr = stack[--(*nstack)];
                ibodyl = stack[--(*nstack)];
            }

            /* recycle old Body if not dirty */
            status = recycleBody(MODL, ibrch, type, args, hasdots);
            CHECK_STATUS(recycleBody);

            if (status == 1) {
                stack[(*nstack)++] = MODL->nbody;
                status = 0;
                goto cleanup;
            }

            if (MODL->body[ibodyl].botype == OCSM_SOLID_BODY &&
                MODL->body[ibodyr].botype == OCSM_SOLID_BODY   ) {

                /* create the new Body */
                status = newBody(MODL, ibrch, OCSM_UNION, ibodyl, ibodyr,
                                 args, hasdots, OCSM_SOLID_BODY, &ibody);
                CHECK_STATUS(newBody);

                ebodyl = MODL->body[ibodyl].ebody;
                ebodyr = MODL->body[ibodyr].ebody;

                /* normal (untrimmed) union of 2 SOLID Bodys */
                if (args[2].nval < 6) {
                    status = solidBoolean(ebodyl, ebodyr, FUSION, maxtol, &emodel);
                    CHECK_STATUS(solidBoolean);

                    status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                                            data, &nchild, &ebodys, &senses);
                    CHECK_STATUS(EG_getTopology);

                    if (nchild == 1) {
                        status = EG_copyObject(ebodys[0], NULL, &ebody);
                        CHECK_STATUS(EG_copyObject);

                        MODL->body[ibody].ebody = ebody;
                    } else {
                        SET_STATUS(OCSM_DID_NOT_CREATE_BODY, union);
                    }

                    status = EG_deleteObject(emodel);
                    CHECK_STATUS(EG_deleteObject);

                /* trimmed union of 2 SOLID Bodys */
                } else {

                    /* determine if given point is in ebodyl or ebodyr */
                    xyz[0] = args[2].val[0];
                    xyz[1] = args[2].val[1];
                    xyz[2] = args[2].val[2];

                    inbodyl = EG_inTopology(ebodyl, xyz);
                    inbodyr = EG_inTopology(ebodyr, xyz);

                    if        (inbodyl == 0 && inbodyr == 1) {
                    } else if (inbodyl == 1 && inbodyr == 0) {
                        etemp1 = ebodyl;
                        ebodyl = ebodyr;
                        ebodyr = etemp1;
                    } else if (inbodyl == 0 && inbodyl == 0) {
                        signalError(MODL, OCSM_DID_NOT_CREATE_BODY, ibrch,
                                    "trim point is inside both Bodys");
                        SET_STATUS(OCSM_DID_NOT_CREATE_BODY, union);
                    } else {
                        signalError(MODL, OCSM_DID_NOT_CREATE_BODY, ibrch,
                                    "trim point is inside neither Body");
                        SET_STATUS(OCSM_DID_NOT_CREATE_BODY, union);
                    }

                    /* move ebodyl a little oppopste the direction given */
                    dirn[0] = args[2].val[3];
                    dirn[1] = args[2].val[4];
                    dirn[2] = args[2].val[5];

                    dirn[3] = sqrt(dirn[0]*dirn[0] + dirn[1]*dirn[1] + dirn[2]*dirn[2]);

                    toler = 1e-5;

                    matrix[ 0] = 1; matrix[ 1] = 0; matrix[ 2] = 0; matrix[ 3] = -dirn[0] / dirn[3] * toler;
                    matrix[ 4] = 0; matrix[ 5] = 1; matrix[ 6] = 0; matrix[ 7] = -dirn[1] / dirn[3] * toler;
                    matrix[ 8] = 0; matrix[ 9] = 0; matrix[10] = 1; matrix[11] = -dirn[2] / dirn[3] * toler;

                    status = EG_makeTransform(MODL->context, matrix, &exform);
                    CHECK_STATUS(EG_makeTransform);

                    status = EG_copyObject(ebodyl, exform, &etemp1);
                    CHECK_STATUS(EG_copyObject);

                    status = EG_deleteObject(exform);
                    CHECK_STATUS(EG_deleteObject);

                    /* subtract ebodyr from etemp1 (the moved copy of ebodyl) */
                    status = solidBoolean(etemp1, ebodyr, SUBTRACTION, maxtol, &emodel);
                    CHECK_STATUS(solidBoolean);

                    status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                                            data, &nchild, &ebodys, &senses);
                    CHECK_STATUS(EG_getTopology);

                    /* if more than one Body was created, keep the body that contains the trim point */
                    if (nchild > 1) {
                        etemp2 = NULL;
                        for (ichild = 0; ichild < nchild; ichild++) {
                            inbodyl = EG_inTopology(ebodys[ichild], xyz);
                            if (inbodyl == 0) {

                                /* move the kept Body back to its original position */
                                matrix[ 3] *= -1;
                                matrix[ 7] *= -1;
                                matrix[11] *= -1;

                                status = EG_makeTransform(MODL->context, matrix, &exform);
                                CHECK_STATUS(EG_makeTransform);

                                status = EG_copyObject(ebodys[ichild], exform, &etemp2);
                                CHECK_STATUS(EG_copyObject);

                                status = EG_deleteObject(exform);
                                CHECK_STATUS(EG_deleteObject);
                            }
                        }

                        status = EG_deleteObject(emodel);
                        CHECK_STATUS(EG_deleteObject);

                        if (etemp2 == NULL) {
                            signalError(MODL, OCSM_INTERNAL_ERROR, ibrch,
                                        "none of the Bodys from the subtraction contain the trim point");
                            SET_STATUS(OCSM_INTERNAL_ERROR, union);
                        }

                        /* union the kept Body and ibodyr */
                        status = solidBoolean(etemp2, ebodyr, FUSION, maxtol, &emodel);
                        CHECK_STATUS(solidBoolean);

                        status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                                                data, &nchild, &ebodys, &senses);
                        CHECK_STATUS(EG_getTopology);

                        if (nchild == 1) {
                            status = EG_copyObject(ebodys[0], NULL, &ebody);
                            CHECK_STATUS(EG_copyObject);

                            MODL->body[ibody].ebody = ebody;
                        } else {
                            SET_STATUS(OCSM_DID_NOT_CREATE_BODY, union);
                        }

                        status = EG_deleteObject(emodel);
                        CHECK_STATUS(EG_deleteObject);

                        status = EG_deleteObject(etemp1);
                        CHECK_STATUS(EG_deleteObject);

                        status = EG_deleteObject(etemp2);
                        CHECK_STATUS(EG_deleteObject);

                    /* since subtraction only created one Body, do a simple
                       union of the two original Bodys (after deleting the emodel) */
                    } else {
                        status = EG_deleteObject(emodel);
                        CHECK_STATUS(EG_deleteObject);

                        status = solidBoolean(ebodyl, ebodyr, FUSION, maxtol, &emodel);
                        CHECK_STATUS(solidBoolean);

                        status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                                                data, &nchild, &ebodys, &senses);
                        CHECK_STATUS(EG_getTopology);

                        if (nchild == 1) {
                            status = EG_copyObject(ebodys[0], NULL, &ebody);
                            CHECK_STATUS(EG_copyObject);

                            MODL->body[ibody].ebody = ebody;
                        } else {
                            SET_STATUS(OCSM_DID_NOT_CREATE_BODY, union);
                        }

                        status = EG_deleteObject(emodel);
                        CHECK_STATUS(EG_deleteObject);
                    }
                }

                /* finish the Body */
                status = finishBody(MODL, ibody);
                CHECK_STATUS(finishBody);

            } else if (MODL->body[ibodyl].botype == OCSM_SHEET_BODY &&
                       MODL->body[ibodyr].botype == OCSM_SHEET_BODY    ) {

                /* create the new Body */
                status = newBody(MODL, ibrch, OCSM_UNION, ibodyl, ibodyr,
                                 args, hasdots, OCSM_SHEET_BODY, &ibody);
                CHECK_STATUS(newBody);

                status = EG_copyObject(MODL->body[ibodyl].ebody, NULL, &etemp);
                CHECK_STATUS(EG_copyObject);

                status = EG_makeTopology(MODL->context, NULL, MODEL, 0,
                                         NULL, 1, &etemp, NULL, &ebodyl);
                CHECK_STATUS(EG_makeTopology);

                ebodyr = MODL->body[ibodyr].ebody;

                /* if ebodyr is a SHEETBODY that contains only 1 Face, set
                   ebodyr to point to that Face instead */
                status = EG_getTopology(ebodyr, &eref, &oclass, &mtype, data, &nshell, &eshells, &senses);
                CHECK_STATUS(EG_getTopology);

                if (oclass == BODY && mtype == SHEETBODY) {
                    if (nshell == 1) {
                        status = EG_getTopology(eshells[0], &eref, &oclass, &mtype, data, &nface, &efaces, &senses);
                        CHECK_STATUS(EG_getTopology);

                        if (oclass == SHELL && nface == 1) {
                            ebodyr = efaces[0];
                        } else {
                            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                                        "UNION expects rite Body to be single face");
                            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, union);
                        }
                    } else {
                        signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                                    "UNION expects rite Body to be a single shell");
                        SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, union);
                    }
                }

                SPRINT0(2, "ebodyl");
                if (outLevel >= 2) printEgo(ebodyl);

                status = EG_getTopology(ebodyl, &eref, &oclass, &mtype, data, &nchild, &pchildren, &senses);
                CHECK_STATUS(EG_getTopology);

                status = EG_getBodyTopos(pchildren[0], NULL, FACE, &ntopo, &ptopos);
                CHECK_STATUS(EG_getBodyTopos);

                for (i = 0; i < ntopo; i++) {
                    SPRINT1(2, "face %d", i);
                    EG_attributeNum(ptopos[i], &nattr);

                    for (j = 0; j < nattr; j++) {
                        EG_attributeGet(ptopos[i], j+1, &aname, &atype, &len, &pints, &preals, &string);
                        SPRINT1x(2, "%20s ", aname);
                        if (atype == ATTRINT) {
                            for (k = 0; k < len; k++) {
                                SPRINT1x(2, "%5d ", pints[k]);
                            }
                        } else if (atype == ATTRREAL) {
                            for (k = 0; k < len; k++) {
                                SPRINT1x(2, "%f ", preals[k]);
                            }
                        } else {
                            SPRINT1x(2, "atype=%d", atype);
                        }
                        SPRINT0(2, " ");
                    }
                }

                EG_free(ptopos);

                SPRINT0(2, "ebodyr");
                if (outLevel >= 2) printEgo(ebodyr);

                status = EG_getBodyTopos(ebodyr, NULL, FACE, &ntopo, &ptopos);
                CHECK_STATUS(EG_getBodyTopos);

                for (i = 0; i < ntopo; i++) {
                    SPRINT1(2, "face %d", i);
                    EG_attributeNum(ptopos[i], &nattr);

                    for (j = 0; j < nattr; j++) {
                        EG_attributeGet(ptopos[i], j+1, &aname, &atype, &len, &pints, &preals, &string);
                        SPRINT1x(2, "%20s ", aname);
                        if (atype == ATTRINT) {
                            for (k = 0; k < len; k++) {
                                SPRINT1x(2, "%5d ", pints[k]);
                            }
                        } else if (atype == ATTRREAL) {
                            for (k = 0; k < len; k++) {
                                SPRINT1x(2, "%f ", preals[k]);
                            }
                        } else {
                            SPRINT1x(2, "atype=%d", atype);
                        }
                        SPRINT0(2, " ");
                    }
                }

                EG_free(ptopos);

                status = solidBoolean(ebodyl, ebodyr, FUSION, maxtol, &emodel);
                CHECK_STATUS(solidBoolean);

                status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                                        data, &nchild, &ebodys, &senses);
                CHECK_STATUS(EG_getTopology);

                if (nchild == 1) {
                    status = EG_copyObject(ebodys[0], NULL, &ebody);
                    CHECK_STATUS(EG_copyObject);

                    MODL->body[ibody].ebody = ebody;
                } else {
                    SET_STATUS(OCSM_DID_NOT_CREATE_BODY, union);
                }

                status = EG_getBodyTopos(ebody, NULL, FACE, &ntopo, &ptopos);
                CHECK_STATUS(EG_getBodyTopos);

                for (i = 0; i < ntopo; i++) {
                    SPRINT1(2, "face %d", i);
                    EG_attributeNum(ptopos[i], &nattr);

                    for (j = 0; j < nattr; j++) {
                        EG_attributeGet(ptopos[i], j+1, &aname, &atype, &len, &pints, &preals, &string);
                        SPRINT1x(2, "%20s ", aname);
                        if (atype == ATTRINT) {
                            for (k = 0; k < len; k++) {
                                SPRINT1x(2, "%5d ", pints[k]);
                            }
                        } else if (atype == ATTRREAL) {
                            for (k = 0; k < len; k++) {
                                SPRINT1x(2, "%f ", preals[k]);
                            }
                        } else {
                            SPRINT1x(2, "atype=%d", atype);
                        }
                        SPRINT0(2, " ");
                    }
                }

                EG_free(ptopos);

                status = EG_deleteObject(emodel);
                CHECK_STATUS(EG_deleteObject);

                /* finish the Body */
                status = finishBody(MODL, ibody);
                CHECK_STATUS(finishBody);
            } else {
                ibody = 0;
            }

        /* union all SolidBodys back to the mark */
        } else if (tomark == 1) {

            /* make a list of the SolidBodys back to mark */
            numBodys = 0;
            while ((*nstack) > 0) {
                ibodyl = stack[--(*nstack)];

                if (ibodyl == 0) {           /* mark is found */
                    break;
                } else if (MODL->body[ibodyl].botype != OCSM_SOLID_BODY) {
                    signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                                "UNION 1 expects SOLID Body on stack");
                    SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, union_need_SOLID_body);
                } else {
                    bodyList[numBodys++] = ibodyl;
                }
            }

            if (numBodys < 2) {
                signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                            "UNION 1 expects 2 or more Bodys since mark");
                SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, union);
            }

            /* recycle old Body if not dirty (note: special treatment to skip
               to last Body that matches this Branch) */
            status = recycleBody(MODL, ibrch, type, args, hasdots);
            CHECK_STATUS(recycleBody);

            if (status == 1) {
                stack[(*nstack)++] = MODL->nbody;
                status = 0;
                goto cleanup;
            }

            /* loop through all pairs of Bodys until finished */
            nchange = 1;
            while (nchange > 0) {
                nchange = 0;

                for (i = numBodys-1; i >= 0; i--) {
                    if (bodyList[i] <= 0) continue;

                    for (j = numBodys-1; j >= 0; j--) {
                        if (bodyList[j] <= 0) continue;
                        if (j           == i) continue;

                        ibodyl = bodyList[i];
                        ibodyr = bodyList[j];

                        ebodyl = MODL->body[ibodyl].ebody;
                        ebodyr = MODL->body[ibodyr].ebody;

                        status = solidBoolean(ebodyl, ebodyr, FUSION, maxtol, &emodel);
                        CHECK_STATUS(solidBoolean);

                        status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                                                data, &nchild, &ebodys, &senses);
                        CHECK_STATUS(EG_getTopology);

                        if (nchild == 1) {
                            status = newBody(MODL, ibrch, OCSM_UNION, ibodyl, ibodyr,
                                             args, hasdots, OCSM_SOLID_BODY, &ibody);
                            CHECK_STATUS(newBody);

                            SPRINT3(1, "                          ... unioning Bodys %4d and %4d to create Body %4d",
                                    ibodyl, ibodyr, ibody);

                            status = EG_copyObject(ebodys[0], NULL, &ebody);
                            CHECK_STATUS(EG_copyObject);

                            MODL->body[ibody].ebody = ebody;

                            status = finishBody(MODL, ibody);
                            CHECK_STATUS(finishBody);

                            bodyList[i] = ibody;
                            bodyList[j] = 0;
                            nchange++;
                        }

                        status = EG_deleteObject(emodel);
                        CHECK_STATUS(EG_deleteObject);
                    }
                }
            }

            nchange = 0;
            for (i = 0; i < numBodys; i++) {
                if (bodyList[i] > 0) {
                    nchange++;
                }
            }

            if (nchange > 1) {
                SET_STATUS(OCSM_DID_NOT_CREATE_BODY, union);
            }

        /* union all FaceBodys back to mark (into a SolidBody) */
        } else if (tomark == 2) {
            signalError(MODL, OCSM_UNSUPPORTED, ibrch,
                        "UNION 2 no longer supported. use COMBINE instead.");
            SET_STATUS(OCSM_UNSUPPORTED, union_2);

        /* bad value for tomark */
        } else {
            SET_STATUS(OCSM_ILLEGAL_VALUE, union);
        }

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "join toler=0" */
    } else if (type == OCSM_JOIN) {
        SPRINT2(1, "    executing [%4d] join:       %11.5f",
                ibrch, args[1].val[0]);

        /* extract the argument */
        toler = args[1].val[0];

        /* pop 2 Bodys from the stack */
        if ((*nstack) < 2) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "JOIN expects 2 Bodys on stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, join);
        } else {
            ibodyr = stack[--(*nstack)];
            ibodyl = stack[--(*nstack)];
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        if (MODL->body[ibodyl].botype != OCSM_SOLID_BODY ||
            MODL->body[ibodyr].botype != OCSM_SOLID_BODY   ) {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                        "JOIN expects 2 SOLID Bodys on stack");
            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, join);
        }

        ebodyl = MODL->body[ibodyl].ebody;
        ebodyr = MODL->body[ibodyr].ebody;

        /* determine the pairs of matched Faces */
        if (toler == 0) {
            status = EG_matchBodyFaces(ebodyl, ebodyr, &nmatch, &matches);
            CHECK_STATUS(EG_matchBodyFaces);
        } else {
            status = EG_getBodyTopos(ebodyl, NULL, FACE, &nfacel, &efacesl);
            CHECK_STATUS(EG_getBodyTopos);

            status = EG_getBodyTopos(ebodyr, NULL, FACE, &nfacer, &efacesr);
            CHECK_STATUS(EG_getBodyTopos);

            MALLOC(matches, int, (nfacel+nfacer));
            nmatch = 0;

            for (ifacel = 0; ifacel < nfacel; ifacel++) {
                status = EG_getTopology(efacesl[ifacel], &esurf, &oclass, &mtype,
                                        datal, &nloopl, &eloopsl, &sense);
                CHECK_STATUS(EG_getTopology);

                for (ifacer = 0; ifacer < nfacer; ifacer++) {
                    status = EG_getTopology(efacesr[ifacer], &esurf, &oclass, &mtype,
                                            datar, &nloopr, &eloopsr, &sense);
                    CHECK_STATUS(EG_getTopology);

                    SPRINT2(3, "checking ifacel=%d, ifacer=%d", ifacel, ifacer);

                    /* make sure that Faces have the same number of Loops */
                    if (nloopl != nloopr) {
                        SPRINT2(3, "    nloopl=%d != nloopr=%d", nloopl, nloopr);
                        continue;
                    }

                    /* make sure that each Loop has the same number of Edges */
                    imatch = 1;
                    for (iloop = 0; iloop < nloopl; iloop++) {
                        status = EG_getTopology(eloopsl[iloop], &esurf, &oclass, &mtype,
                                                datal, &nedgel, &eedgesl, &sense);
                        CHECK_STATUS(EG_getTopology);

                        status = EG_getTopology(eloopsr[iloop], &esurf, &oclass, &mtype,
                                                datar, &nedger, &eedgesr, &sense);
                        CHECK_STATUS(EG_getTopology);

                        if (nedgel != nedger) {
                            imatch = 0;
                            SPRINT3(3, "    nedgel=%d != nedger=%d  (iloop=%d)", nedgel, nedger, iloop);
                            break;
                        }
                    }
                    if (imatch == 0) continue;

                    /* make sure that the bounding boxes of the Faces match */
                    status = EG_getBoundingBox(efacesl[ifacel], bboxl);
                    CHECK_STATUS(EG_getBoundingBox);

                    status = EG_getBoundingBox(efacesr[ifacer], bboxr);
                    CHECK_STATUS(EG_getBoundingBox);

                    if (fabs(bboxl[0]-bboxr[0]) > toler) {
                        SPRINT2(3, "    bboxl[0]=%f != bboxr[0]=%f", bboxl[0], bboxr[0]);
                        continue;
                    }
                    if (fabs(bboxl[1]-bboxr[1]) > toler) {
                        SPRINT2(3, "    bboxl[1]=%f != bboxr[1]=%f", bboxl[1], bboxr[1]);
                        continue;
                    }
                    if (fabs(bboxl[2]-bboxr[2]) > toler) {
                        SPRINT2(3, "    bboxl[2]=%f != bboxr[2]=%f", bboxl[2], bboxr[2]);
                        continue;
                    }
                    if (fabs(bboxl[3]-bboxr[3]) > toler) {
                        SPRINT2(3, "    bboxl[3]=%f != bboxr[3]=%f", bboxl[3], bboxr[3]);
                        continue;
                    }
                    if (fabs(bboxl[4]-bboxr[4]) > toler) {
                        SPRINT2(3, "    bboxl[4]=%f != bboxr[4]=%f", bboxl[4], bboxr[4]);
                        continue;
                    }
                    if (fabs(bboxl[5]-bboxr[5]) > toler) {
                        SPRINT2(3, "    bboxl[5]=%f != bboxr[5]=%f", bboxl[5], bboxr[5]);
                        continue;
                    }

                    /* make sure that the areas of the Faces match */
                    status = EG_getMassProperties(efacesl[ifacel], datal);
                    CHECK_STATUS(EG_getMassProperties);

                    status = EG_getMassProperties(efacesr[ifacer], datar);
                    CHECK_STATUS(EG_getMassProperties);

                    if (fabs(datal[1]-datar[1]) > toler) {
                        SPRINT2(3, "    datal[1]=%f != datar[1]=%f", datal[1], datar[1]);
                        continue;
                    }

                    /* make sure that the centroids of the Faces match */
                    if (fabs(datal[2]-datar[2]) > toler) {
                        SPRINT2(3, "    datal[2]=%f != datar[2]=%f", datal[2], datar[2]);
                        continue;
                    }
                    if (fabs(datal[3]-datar[3]) > toler) {
                        SPRINT2(3, "    datal[3]=%f != datar[3]=%f", datal[3], datar[3]);
                        continue;
                    }
                    if (fabs(datal[4]-datar[4]) > toler) {
                        SPRINT2(3, "    datal[4]=%f != datar[4]=%f", datal[4], datar[4]);
                        continue;
                    }

                    /* getting here means that the Faces match, so add to the match list */
                    matches[2*nmatch  ] = ifacel + 1;
                    matches[2*nmatch+1] = ifacer + 1;
                    nmatch++;
                }
            }

            EG_free(efacesl);
            EG_free(efacesr);
        }

        if (nmatch < 1) {
            signalError(MODL, OCSM_FACE_NOT_FOUND, ibrch,
                        "expecting nmatch>=1 but got %d", nmatch);
            FREE(matches);
            SET_STATUS(OCSM_FACE_NOT_FOUND, join);
        }

        /* make a list of the Faces to sew.  remove any from list that
           are in matches, which was generated above */
        nfacel = MODL->body[ibodyl].nface;
        nfacer = MODL->body[ibodyr].nface;

        MALLOC(efaces, ego, nfacel+nfacer);

        j = 0;
        for (i = 1; i <= nfacel; i++) {
            efaces[j++] = MODL->body[ibodyl].face[i].eface;
            SPRINT3(3, "adding   Face %d (%lx) to   efaces for Body %d", i, (long)efaces[j-1], ibodyl);
            for (k = 0; k < nmatch; k++) {
                if (i == matches[2*k]) {
                    SPRINT3(3, "removing Face %d (%lx) from efaces for Body %d", i, (long)efaces[j-1], ibodyl);
                    j--;
                    break;
                }
            }
        }
        for (i = 1; i <= nfacer; i++) {
            efaces[j++] = MODL->body[ibodyr].face[i].eface;
            SPRINT3(3, "adding   Face %d (%lx) to   efaces for Body %d", i, (long)efaces[j-1], ibodyr);
            for (k = 0; k < nmatch; k++) {
                if (i == matches[2*k+1]) {
                    SPRINT3(3, "removing Face %d (%lx) from efaces for Body %d", i, (long)efaces[j-1], ibodyl);
                    j--;
                    break;
                }
            }
        }
        SPRINT1(3, "there are %d faces in efaces", j);

        if (toler == 0) {
            EG_free(matches);
            matches = NULL;
        } else {
            FREE(matches);
        }

        status = EG_sewFaces(j, efaces, toler, 0, &emodel);
        CHECK_STATUS(EG_sewFaces);

        FREE(efaces);

        /* if the Model contains a single SolidBody, extract it */
        status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                                data, &nchild, &ebodys, &senses);
        CHECK_STATUS(EG_getTopology);

        if (nchild < 1) {
            SET_STATUS(OCSM_DID_NOT_CREATE_BODY, join);
        } else if (nchild > 1) {
            SET_STATUS(OCSM_CREATED_TOO_MANY_BODYS, join);
        } else {
            status = EG_getTopology(ebodys[0], &eref, &oclass, &mtype,
                                    data, &nface, &efaces, &senses);
            CHECK_STATUS(EG_getTopology);

            if (oclass != BODY || mtype != SOLIDBODY) {
                signalError(MODL, OCSM_DID_NOT_CREATE_BODY, ibrch,
                            "JOIN expected a SOLILD Body to be produced");
                SET_STATUS(OCSM_DID_NOT_CREATE_BODY, join);
            }

            /* create the new Body */
            status = newBody(MODL, ibrch, OCSM_JOIN, ibodyl, ibodyr,
                             args, hasdots, OCSM_SOLID_BODY, &ibody);
            CHECK_STATUS(newBody);

            status = EG_copyObject(ebodys[0], NULL, &ebody);
            CHECK_STATUS(EG_copyObject);

            MODL->body[ibody].ebody = ebody;
        }

        status = EG_deleteObject(emodel);
        CHECK_STATUS(EG_deleteObject);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "extract index=0" */
    } else if (type == OCSM_EXTRACT) {
        SPRINT2(1, "    executing [%4d] extract:    %11.5f",
                ibrch, args[1].val[0]);

        /* pop a Body from the stack */
        if ((*nstack) < 1) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "EXTRACT expects a Body on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, extract);
        } else {
            ibodyl = stack[--(*nstack)];
        }

        /* we want to extract a Face */
        if (NINT(args[1].val[0]) > 0) {
            iface = NINT(args[1].val[0]);

            if (MODL->body[ibodyl].botype != OCSM_SOLID_BODY &&
                MODL->body[ibodyl].botype != OCSM_SHEET_BODY   ) {
                signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                            "EXTRACT expects a SOLID or SHEET Body on stack");
                SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, extract);
            }

            /* recycle old Body if not dirty */
            status = recycleBody(MODL, ibrch, type, args, hasdots);
            CHECK_STATUS(recycleBody);

            if (status == 1) {
                stack[(*nstack)++] = MODL->nbody;
                status = 0;
                goto cleanup;
            }

            /* create the Body */
            status = newBody(MODL, ibrch, OCSM_EXTRACT, -1, -1,
                             args, hasdots, OCSM_SHEET_BODY, &ibody);
            CHECK_STATUS(newBody);

            eface = MODL->body[ibodyl].face[iface].eface;

            status = EG_makeTopology(MODL->context, NULL, BODY, FACEBODY,
                                     NULL, 1, &eface, NULL, &ebody);
            CHECK_STATUS(EG_makeTopology);

            MODL->body[ibody].ebody = ebody;

        /* extracting an Edge from ibodyl */
        } else if (NINT(args[1].val[0]) < 0) {
            iedge = NINT(-args[1].val[0]);

            if (MODL->body[ibodyl].botype != OCSM_SOLID_BODY &&
                MODL->body[ibodyl].botype != OCSM_SHEET_BODY   ) {
                signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                            "EXTRACT expects a SOLID or SHEET Body on stack");
                SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, extract);
            }

            /* create the Body */
            status = newBody(MODL, ibrch, OCSM_EXTRACT, -1, -1,
                             args, hasdots, OCSM_WIRE_BODY, &ibody);
            CHECK_STATUS(newBody);

            eedge = MODL->body[ibodyl].edge[iedge].eedge;

            senses1[0] = SFORWARD;
            status = EG_makeTopology(MODL->context, NULL, LOOP, OPEN,
                                     NULL, 1, &eedge, senses1, &eloop);
            CHECK_STATUS(EG_makeTopology);

            status = EG_makeTopology(MODL->context, NULL, BODY, WIREBODY,
                                     NULL, 1, &eloop, NULL, &ebody);
            CHECK_STATUS(EG_makeTopology);

            MODL->body[ibody].ebody = ebody;

        /* extracting the shell from ibodyl */
        } else {
            if (MODL->body[ibodyl].botype != OCSM_SOLID_BODY) {
                signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                            "EXTRACT expects a SOLID Body on the stack");
                SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, extract);
            }

            /* create the Body */
            status = newBody(MODL, ibrch, OCSM_EXTRACT, -1, -1,
                             args, hasdots, OCSM_SHEET_BODY, &ibody);
            CHECK_STATUS(newBody);

            ebodyl = MODL->body[ibodyl].ebody;

            status = EG_getBodyTopos(ebodyl, NULL, SHELL, &nshell, &eshells);
            CHECK_STATUS(EG_getBodyTopos);

            status = EG_makeTopology(MODL->context, NULL, BODY, SHEETBODY,
                                     NULL, 1, eshells, NULL, &ebody);
            CHECK_STATUS(EG_makeTopology);

            EG_free(eshells);

            MODL->body[ibody].ebody = ebody;
        }

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "combine" */
    } else if (type == OCSM_COMBINE) {
        SPRINT1(1, "    executing [%4d] combine:",
                ibrch);

        /* make a list to the Bodys back to mark */
        numBodys = 0;
        nedges   = 0;
        while ((*nstack) > 0) {
            ibodyl = stack[--(*nstack)];

            if (ibodyl == 0) {               /* mark is found */
                break;
            } else if (MODL->body[ibodyl].botype == OCSM_WIRE_BODY ||
                       MODL->body[ibodyl].botype == OCSM_SHEET_BODY  ) {
                bodyList[numBodys++] = ibodyl;
                nedges += MODL->body[ibodyl].nedge;
            } else {
                signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                            "COMBINE expects WIRE or SHEET Body on stack");
                SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, combine);
            }
        }

        if (numBodys < 2) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "COMBINE expects two or more Bodys since mark");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, combine);
        }

        bodyType = MODL->body[bodyList[0]].botype;
        for (i = 1; i < numBodys; i++) {
            if (MODL->body[bodyList[i]].botype != bodyType) {
                signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                            "COMBINE found mixed Body types on stack");
                SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, combine);
            }
        }

        ibodyl = bodyList[numBodys-1];
        ibodyr = bodyList[0         ];

        /* recycle old Body if not dirty (note: special treatment to skip
           to last Body that matches this Branch) */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* combine WireBodys into a Facebody */
        if (bodyType == OCSM_WIRE_BODY) {
            /* make list of Edges in the WireBodys */
            MALLOC(eedges, ego, nedges);

            k = 0;
            for (i = numBodys-1; i >= 0; i--) {
                ebody = MODL->body[bodyList[i]].ebody;

                status = EG_getTopology(ebody, &eref, &oclass, &mtype, data, &nchild, &echilds, &senses);
                CHECK_STATUS(EG_getTopology);

                eloop = echilds[0];

                status = EG_getTopology(eloop, &eref, &oclass, &mtype, data, &nchild, &echilds, &senses);
                CHECK_STATUS(EG_getTopology);

                for (j = 0; j < nchild; j++) {
                    status = EG_copyObject(echilds[j], NULL, &(eedges[k++]));
                    CHECK_STATUS(EG_copyObject);

                    if (outLevel >= 2) {
                        SPRINT1(2, "before makeLoop: eedges[%d]", k-1);
                        printEgo(eedges[k-1]);
                    }
                }
            }

            /* make the Loop, Face, and then the FaceBody */
            status = EG_makeLoop(nedges, eedges, NULL, 0, &eloop);
            CHECK_STATUS(EG_makeLoop);

            if (outLevel >= 2) {
                SPRINT0(2, "eloop");
                printEgo(eloop);
            }

            if (status > 0) {
                EG_free(eloop);
                signalError(MODL, OCSM_DID_NOT_CREATE_BODY, ibrch,
                            "Edges could not combined into a single Loop");
                SET_STATUS(OCSM_DID_NOT_CREATE_BODY, combine);
            }

            /* make sure that Loop starts at first Edge */

            status = EG_getArea(eloop, NULL, &area);
            CHECK_STATUS(EG_getArea);

            /* if the area<0, then we need to flip the Loop now so that when
               we make the Face, we can use SREVERSE to flip the loop back.  if
               this is not done, the Loop will be flipped in the makeTopology(BODY),
               resulting in a FACEBODY that cannot be used in rule or blend */
            if (area < 0) {
                SPRINT1(0, "WARNING:: backwards Loop has area=%f that will be reversed", area);

                eflip = eloop;

                status = EG_flipObject(eflip, &eloop);
                CHECK_STATUS(EG_flipObject);
            }

            if (area > 0) {
                status = EG_makeFace(eloop, SFORWARD, NULL, &eface);
                CHECK_STATUS(EG_makeFace);
            } else {
                status = EG_makeFace(eloop, SREVERSE, NULL, &eface);
                CHECK_STATUS(EG_makeFace);
            }

            status = EG_makeTopology(MODL->context, NULL, BODY, FACEBODY,
                                     NULL, 1, &eface, NULL, &ebody);
            CHECK_STATUS(EG_makeTopology);

            /* create the Body */
            status = newBody(MODL, ibrch, OCSM_COMBINE, ibodyl, ibodyr,
                             args, hasdots, OCSM_SHEET_BODY, &ibody);
            CHECK_STATUS(newBody);

            MODL->body[ibody].ebody = ebody;

            /* attach the Body to its children */
            for (i = 0; i < numBodys; i++) {
                MODL->body[bodyList[i]].ichld = ibody;
            }

            /* mark the new Faces with the current Branch */
            status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
            CHECK_STATUS(EG_getBodyTopos);

            for (iface = 1; iface <= nface; iface++) {
                status = setFaceAttribute(MODL, ibody, iface, 0, iface, npatn, patn);
                CHECK_STATUS(setFaceAttribute);
            }

            EG_free(efaces);

            /* finish the Body */
            status = finishBody(MODL, ibody);
            CHECK_STATUS(finishBody);

            /* push the Body onto the stack */
            stack[(*nstack)++] = ibody;

            status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
            CHECK_STATUS(getBodyTolerance);

            SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                    ibody, toler);

        /* combine SheetBodys into a SolidBody */
        } else if (bodyType == OCSM_SHEET_BODY) {

            /* make a list of the FaceBodys */
            MALLOC(elist, ego, numBodys);

            for (i = 0; i < numBodys; i++) {
                ebodyl = MODL->body[bodyList[i]].ebody;

                status = EG_getTopology(ebodyl, &eref, &oclass, &mtype,
                                        data, &nface, &efaces, &senses);
                CHECK_STATUS(EG_getTopology);

                if (nface != 1) {
                    signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                                "SHEET has more than one FACE");
                    SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, combine);
                }

                elist[i] = efaces[0];
            }

            /* sew them together into a Model using default tolerance */
            status = EG_sewFaces(numBodys, elist, 0, 0, &emodel);
            CHECK_STATUS(EG_sewfaces);

            /* if the Model contains a single SolidBody, extract it */
            status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                                    data, &nchild, &ebodys, &senses);
            CHECK_STATUS(EG_getTopology);

            if (nchild < 1) {
                SET_STATUS(OCSM_DID_NOT_CREATE_BODY, combine);
            } else {
                status = EG_getTopology(ebodys[0], &eref, &oclass, &mtype,
                                        data, &nface, &efaces, &senses);
                CHECK_STATUS(EG_getTopology);

                if (oclass != BODY || mtype != SOLIDBODY) {
                    signalError(MODL, OCSM_DID_NOT_CREATE_BODY, ibrch,
                                "COMBINE was expecting that a SOLID Body would be produced");
                    SET_STATUS(OCSM_DID_NOT_CREATE_BODY, combine);
                }

                status = EG_copyObject(ebodys[0], NULL, &ebody);
                CHECK_STATUS(EG_copyObject);

                if (nchild > 1) {
                    SPRINT1(0, "WARNING:: %d Bodys are being lost", nchild-1);
                }
            }

            status = EG_deleteObject(emodel);
            CHECK_STATUS(EG_deleteObject);

            /* create the Body */
            status = newBody(MODL, ibrch, OCSM_COMBINE, ibodyl, ibodyr,
                             args, hasdots, OCSM_SOLID_BODY, &ibody);
            CHECK_STATUS(newBody);

            MODL->body[ibody].ebody = ebody;

            /* attach the Body to its children */
            for (i = 0; i < numBodys; i++) {
                MODL->body[bodyList[i]].ichld = ibody;
            }

            /* finish the Body */
            status = finishBody(MODL, ibody);
            CHECK_STATUS(finishBody);

            /* push the Body onto the stack */
            stack[(*nstack)++] = ibody;

            status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
            CHECK_STATUS(getBodyTolerance);

            SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                    ibody, toler);
        }

    /* execute: "connect faceList1 faceList2" */
    } else if (type == OCSM_CONNECT) {
        SPRINT3(1, "    executing [%4d] connect:    %11.5f  %11.5f",
                ibrch, args[1].val[0], args[2].val[0]);

        /* pop two Bodys from the stack */
        if ((*nstack) < 2) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "CONNECT expects to Bodys on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, connect);
        } else {
            ibodyr = stack[--(*nstack)];
            ibodyl = stack[--(*nstack)];
        }

        /* make sure that the number of values in args[1] and args[2] are the same */
        if (args[1].nval == args[2].nval) {
            nsize = args[1].nval;
        } else {
            signalError(MODL, OCSM_ILLEGAL_ARGUMENT, ibrch,
                        "faceList1 and faceList2 must be the same size");
            SET_STATUS(OCSM_ILLEGAL_ARGUMENT, connect);
        }

        /* make sure all Faces in faceList1 and faceList2 have 4 Edges */
        for (ii = 0; ii < nsize; ii++) {
            iface = NINT(args[1].val[ii]);
            nedge = 0;
            for (iedge = 1; iedge <= MODL->body[ibodyl].nedge; iedge++) {
                if (MODL->body[ibodyl].edge[iedge].ileft == iface ||
                    MODL->body[ibodyl].edge[iedge].irite == iface   ) {
                    nedge++;
                }
            }
            if (nedge != 4) {
                signalError(MODL, OCSM_ILLEGAL_ARGUMENT, ibrch,
                            "Face %d:%d has %d Edges (not 4)", ibodyl, iface, nedge);
                SET_STATUS(OCSM_ILLEGAL_ARGUMENT, connect);
            }

            iface = NINT(args[2].val[ii]);
            nedge = 0;
            for (iedge = 1; iedge <= MODL->body[ibodyr].nedge; iedge++) {
                if (MODL->body[ibodyr].edge[iedge].ileft == iface ||
                    MODL->body[ibodyr].edge[iedge].irite == iface   ) {
                    nedge++;
                }
            }
            if (nedge != 4) {
                signalError(MODL, OCSM_ILLEGAL_ARGUMENT, ibrch,
                            "Face %d:%d has %d Edges (not 4)", ibodyr, iface, nedge);
                SET_STATUS(OCSM_ILLEGAL_ARGUMENT, connect);
            }
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_CONNECT, ibodyl, ibodyr,
                         args, hasdots, OCSM_SOLID_BODY, &ibody);
        CHECK_STATUS(newBody);

        /* allocate arrays */
        nindx = 2 * nsize + 1;
        ntemp = 2 * nsize + MODL->body[ibodyl].nface + MODL->body[ibodyr].nface;

        MALLOC(indx,      int, 6*nindx  );
        MALLOC(enewEdges, ego,   nindx+1);
        MALLOC(enewFaces, ego,   ntemp  );

        for (ii = 0; ii < 6*nindx; ii++) {
            indx[ii] = 0;
        }
        for (newEdges = 0; newEdges <= nindx; newEdges++) {
            enewEdges[newEdges] = NULL;
        }
        for (newFaces = 0; newFaces <  ntemp; newFaces++) {
            enewFaces[newFaces] = NULL;
        }


        /* set up the index table (nindx*6) that contains the Node, Edge,
           and Face indices of the entities to be bridged

                      0     1     2     3     4    ...  nindx

           row5     node  edge  node  edge  node  edge  node }
           row4     edge  face  edge  face  edge  face  edge } ibodyr
           row3     node  edge  node  edge  node  edge  node }

           row2     node  edge  node  edge  node  edge  node }
           row1     edge  face  edge  face  edge  face  edge } ibodyl
           row0     node  edge  node  edge  node  edge  node }
        */

        /* set up indx map for a single pair of Faces */
#ifndef __clang_analyzer__
        if (nsize == 1) {

            /* set up the indx map for ibodyl */
            indx[ 4] = NINT(args[1].val[0]);

            for (iedge = 1; iedge <= MODL->body[ibodyl].nedge; iedge++) {
                if        (MODL->body[ibodyl].edge[iedge].ileft == indx[ 4]) {
                    indx[ 1] = iedge;
                    indx[ 0] = MODL->body[ibodyl].edge[iedge].ibeg;
                    indx[ 2] = MODL->body[ibodyl].edge[iedge].iend;
                    break;
                } else if (MODL->body[ibodyl].edge[iedge].irite == indx[ 4]) {
                    indx[ 1] = iedge;
                    indx[ 0] = MODL->body[ibodyl].edge[iedge].iend;
                    indx[ 2] = MODL->body[ibodyl].edge[iedge].ibeg;
                    break;
                }
            }

            for (iedge = 1; iedge <= MODL->body[ibodyl].nedge; iedge++) {
                if        (MODL->body[ibodyl].edge[iedge].ileft == indx[ 4] &&
                           MODL->body[ibodyl].edge[iedge].ibeg  == indx[ 2]   ) {
                    indx[ 5] = iedge;
                    indx[ 8] = MODL->body[ibodyl].edge[iedge].iend;
                    break;
                } else if (MODL->body[ibodyl].edge[iedge].irite == indx[ 4] &&
                           MODL->body[ibodyl].edge[iedge].iend  == indx[ 2]   ) {
                    indx[ 5] = iedge;
                    indx[ 8] = MODL->body[ibodyl].edge[iedge].ibeg;
                    break;
                }
            }

            for (iedge = 1; iedge <= MODL->body[ibodyl].nedge; iedge++) {
                if        (MODL->body[ibodyl].edge[iedge].ileft == indx[ 4] &&
                           MODL->body[ibodyl].edge[iedge].ibeg  == indx[ 8]   ) {
                    indx[ 7] = iedge;
                    indx[ 6] = MODL->body[ibodyl].edge[iedge].iend;
                    break;
                } else if (MODL->body[ibodyl].edge[iedge].irite == indx[ 4] &&
                           MODL->body[ibodyl].edge[iedge].iend  == indx[ 8]   ) {
                    indx[ 7] = iedge;
                    indx[ 6] = MODL->body[ibodyl].edge[iedge].ibeg;
                    break;
                }
            }

            for (iedge = 1; iedge <= MODL->body[ibodyl].nedge; iedge++) {
                if        (MODL->body[ibodyl].edge[iedge].ileft == indx[ 4] &&
                           MODL->body[ibodyl].edge[iedge].ibeg  == indx[ 6] &&
                           MODL->body[ibodyl].edge[iedge].iend  == indx[ 0]   ) {
                    indx[ 3] = iedge;
                    break;
                } else if (MODL->body[ibodyl].edge[iedge].irite == indx[ 4] &&
                           MODL->body[ibodyl].edge[iedge].iend  == indx[ 6] &&
                           MODL->body[ibodyl].edge[iedge].ibeg  == indx[ 0]   ) {
                    indx[ 3] = iedge;
                    break;
                }
            }

            /* set up the indx map for ibodyr */
            indx[13] = NINT(args[2].val[0]);

            for (iedge = 1; iedge <= MODL->body[ibodyr].nedge; iedge++) {
                if        (MODL->body[ibodyr].edge[iedge].irite == indx[13]) {
                    indx[10] = iedge;
                    indx[ 9] = MODL->body[ibodyr].edge[iedge].ibeg;
                    indx[11] = MODL->body[ibodyr].edge[iedge].iend;
                    break;
                } else if (MODL->body[ibodyr].edge[iedge].ileft == indx[13]) {
                    indx[10] = iedge;
                    indx[ 9] = MODL->body[ibodyr].edge[iedge].iend;
                    indx[11] = MODL->body[ibodyr].edge[iedge].ibeg;
                    break;
                }
            }

            for (iedge = 1; iedge <= MODL->body[ibodyr].nedge; iedge++) {
                if        (MODL->body[ibodyr].edge[iedge].irite == indx[13] &&
                           MODL->body[ibodyr].edge[iedge].ibeg  == indx[11]   ) {
                    indx[14] = iedge;
                    indx[17] = MODL->body[ibodyr].edge[iedge].iend;
                    break;
                } else if (MODL->body[ibodyr].edge[iedge].ileft == indx[13] &&
                           MODL->body[ibodyr].edge[iedge].iend  == indx[11]   ) {
                    indx[14] = iedge;
                    indx[17] = MODL->body[ibodyr].edge[iedge].ibeg;
                    break;
                }
            }

            for (iedge = 1; iedge <= MODL->body[ibodyr].nedge; iedge++) {
                if        (MODL->body[ibodyr].edge[iedge].irite == indx[13] &&
                           MODL->body[ibodyr].edge[iedge].ibeg  == indx[17]   ) {
                    indx[16] = iedge;
                    indx[15] = MODL->body[ibodyr].edge[iedge].iend;
                    break;
                } else if (MODL->body[ibodyr].edge[iedge].ileft == indx[13] &&
                           MODL->body[ibodyr].edge[iedge].iend  == indx[17]   ) {
                    indx[16] = iedge;
                    indx[15] = MODL->body[ibodyr].edge[iedge].ibeg;
                    break;
                }
            }

            for (iedge = 1; iedge <= MODL->body[ibodyr].nedge; iedge++) {
                if        (MODL->body[ibodyr].edge[iedge].irite == indx[13] &&
                           MODL->body[ibodyr].edge[iedge].ibeg  == indx[15] &&
                           MODL->body[ibodyr].edge[iedge].iend  == indx[ 9]   ) {
                    indx[12] = iedge;
                    break;
                } else if (MODL->body[ibodyr].edge[iedge].ileft == indx[13] &&
                           MODL->body[ibodyr].edge[iedge].iend  == indx[15] &&
                           MODL->body[ibodyr].edge[iedge].ibeg  == indx[ 9]   ) {
                    indx[12] = iedge;
                    break;
                }
            }

            /* rotate the ibodyr Face until we get the shortest
               Edges between ibodyl and ibodyr */
            ibest = 0;
            dbest = sqrt(SQR(MODL->body[ibodyl].node[indx[ 0]].x-MODL->body[ibodyr].node[indx[ 9]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 0]].y-MODL->body[ibodyr].node[indx[ 9]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 0]].z-MODL->body[ibodyr].node[indx[ 9]].z))
                   +sqrt(SQR(MODL->body[ibodyl].node[indx[ 2]].x-MODL->body[ibodyr].node[indx[11]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 2]].y-MODL->body[ibodyr].node[indx[11]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 2]].z-MODL->body[ibodyr].node[indx[11]].z))
                   +sqrt(SQR(MODL->body[ibodyl].node[indx[ 8]].x-MODL->body[ibodyr].node[indx[17]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 8]].y-MODL->body[ibodyr].node[indx[17]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 8]].z-MODL->body[ibodyr].node[indx[17]].z))
                   +sqrt(SQR(MODL->body[ibodyl].node[indx[ 6]].x-MODL->body[ibodyr].node[indx[15]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 6]].y-MODL->body[ibodyr].node[indx[15]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 6]].z-MODL->body[ibodyr].node[indx[15]].z));


            dtest = sqrt(SQR(MODL->body[ibodyl].node[indx[ 0]].x-MODL->body[ibodyr].node[indx[15]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 0]].y-MODL->body[ibodyr].node[indx[15]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 0]].z-MODL->body[ibodyr].node[indx[15]].z))
                   +sqrt(SQR(MODL->body[ibodyl].node[indx[ 2]].x-MODL->body[ibodyr].node[indx[ 9]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 2]].y-MODL->body[ibodyr].node[indx[ 9]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 2]].z-MODL->body[ibodyr].node[indx[ 9]].z))
                   +sqrt(SQR(MODL->body[ibodyl].node[indx[ 8]].x-MODL->body[ibodyr].node[indx[11]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 8]].y-MODL->body[ibodyr].node[indx[11]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 8]].z-MODL->body[ibodyr].node[indx[11]].z))
                   +sqrt(SQR(MODL->body[ibodyl].node[indx[ 6]].x-MODL->body[ibodyr].node[indx[17]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 6]].y-MODL->body[ibodyr].node[indx[17]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 6]].z-MODL->body[ibodyr].node[indx[17]].z));
            if (dtest < dbest) {
                ibest = 1;
                dbest = dtest;
            }

            dtest = sqrt(SQR(MODL->body[ibodyl].node[indx[ 0]].x-MODL->body[ibodyr].node[indx[17]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 0]].y-MODL->body[ibodyr].node[indx[17]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 0]].z-MODL->body[ibodyr].node[indx[17]].z))
                   +sqrt(SQR(MODL->body[ibodyl].node[indx[ 2]].x-MODL->body[ibodyr].node[indx[15]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 2]].y-MODL->body[ibodyr].node[indx[15]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 2]].z-MODL->body[ibodyr].node[indx[15]].z))
                   +sqrt(SQR(MODL->body[ibodyl].node[indx[ 8]].x-MODL->body[ibodyr].node[indx[ 9]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 8]].y-MODL->body[ibodyr].node[indx[ 9]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 8]].z-MODL->body[ibodyr].node[indx[ 9]].z))
                   +sqrt(SQR(MODL->body[ibodyl].node[indx[ 6]].x-MODL->body[ibodyr].node[indx[11]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 6]].y-MODL->body[ibodyr].node[indx[11]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 6]].z-MODL->body[ibodyr].node[indx[11]].z));
            if (dtest < dbest) {
                ibest = 2;
                dbest = dtest;
            }

            dtest = sqrt(SQR(MODL->body[ibodyl].node[indx[ 0]].x-MODL->body[ibodyr].node[indx[11]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 0]].y-MODL->body[ibodyr].node[indx[11]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 0]].z-MODL->body[ibodyr].node[indx[11]].z))
                   +sqrt(SQR(MODL->body[ibodyl].node[indx[ 2]].x-MODL->body[ibodyr].node[indx[17]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 2]].y-MODL->body[ibodyr].node[indx[17]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 2]].z-MODL->body[ibodyr].node[indx[17]].z))
                   +sqrt(SQR(MODL->body[ibodyl].node[indx[ 8]].x-MODL->body[ibodyr].node[indx[15]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 8]].y-MODL->body[ibodyr].node[indx[15]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 8]].z-MODL->body[ibodyr].node[indx[15]].z))
                   +sqrt(SQR(MODL->body[ibodyl].node[indx[ 6]].x-MODL->body[ibodyr].node[indx[ 9]].x)
                        +SQR(MODL->body[ibodyl].node[indx[ 6]].y-MODL->body[ibodyr].node[indx[ 9]].y)
                        +SQR(MODL->body[ibodyl].node[indx[ 6]].z-MODL->body[ibodyr].node[indx[ 9]].z));
            if (dtest < dbest) {
                ibest = 3;
                dbest = dtest;
            }

            if        (ibest == 1) {
                iswap    = indx[ 9];
                indx[ 9] = indx[15];
                indx[15] = indx[17];
                indx[17] = indx[11];
                indx[11] = iswap;

                iswap    = indx[12];
                indx[12] = indx[16];
                indx[16] = indx[14];
                indx[14] = indx[10];
                indx[10] = iswap;
            } else if (ibest == 2) {
                iswap    = indx[ 9];
                indx[ 9] = indx[17];
                indx[17] = iswap;

                iswap    = indx[10];
                indx[10] = indx[16];
                indx[16] = iswap;

                iswap    = indx[11];
                indx[11] = indx[15];
                indx[15] = iswap;

                iswap    = indx[12];
                indx[12] = indx[14];
                indx[14] = iswap;
            } else if (ibest == 3) {
                iswap    = indx[ 9];
                indx[ 9] = indx[11];
                indx[11] = indx[17];
                indx[17] = indx[15];
                indx[15] = iswap;

                iswap    = indx[10];
                indx[10] = indx[14];
                indx[14] = indx[16];
                indx[16] = indx[12];
                indx[12] = iswap;
            }

        /* set up indx map for multiple pairs of Faces (which must be contiguous) */
        } else {

            /* insert Faces into middle rows */
            for (ii = 0; ii < nsize; ii++) {
                indx[  nindx+2*ii+1] = NINT(args[1].val[ii]);
                indx[4*nindx+2*ii+1] = NINT(args[2].val[ii]);
            }

            /* insert interior Edges and their Nodes */
            for (ii = 1; ii < nsize; ii++) {
                for (iedge = 1; iedge <= MODL->body[ibodyl].nedge; iedge++) {
                    if        (MODL->body[ibodyl].edge[iedge].ileft == indx[  nindx+2*ii-1] &&
                               MODL->body[ibodyl].edge[iedge].irite == indx[  nindx+2*ii+1]   ) {
                        indx[  nindx+2*ii  ] = iedge;
                        indx[        2*ii  ] = MODL->body[ibodyl].edge[iedge].ibeg;
                        indx[2*nindx+2*ii  ] = MODL->body[ibodyl].edge[iedge].iend;
                    } else if (MODL->body[ibodyl].edge[iedge].ileft == indx[  nindx+2*ii+1] &&
                               MODL->body[ibodyl].edge[iedge].irite == indx[  nindx+2*ii-1]   ) {
                        indx[  nindx+2*ii  ] = iedge;
                        indx[        2*ii  ] = MODL->body[ibodyl].edge[iedge].iend;
                        indx[2*nindx+2*ii  ] = MODL->body[ibodyl].edge[iedge].ibeg;
                    }
                }

                for (iedge = 1; iedge <= MODL->body[ibodyr].nedge; iedge++) {
                    if        (MODL->body[ibodyr].edge[iedge].ileft == indx[4*nindx+2*ii-1] &&
                               MODL->body[ibodyr].edge[iedge].irite == indx[4*nindx+2*ii+1]   ) {
                        indx[4*nindx+2*ii  ] = iedge;
                        indx[3*nindx+2*ii  ] = MODL->body[ibodyr].edge[iedge].iend;
                        indx[5*nindx+2*ii  ] = MODL->body[ibodyr].edge[iedge].ibeg;
                    } else if (MODL->body[ibodyr].edge[iedge].ileft == indx[4*nindx+2*ii+1] &&
                               MODL->body[ibodyr].edge[iedge].irite == indx[4*nindx+2*ii-1]   ) {
                        indx[4*nindx+2*ii  ] = iedge;
                        indx[3*nindx+2*ii  ] = MODL->body[ibodyr].edge[iedge].ibeg;
                        indx[5*nindx+2*ii  ] = MODL->body[ibodyr].edge[iedge].iend;
                    }
                }
            }

            /* insert top and bottom Edges and corner Nodes */
            for (ii = 0; ii < nsize; ii++) {

                /* Edges in ibodyl */
                iface = indx[  nindx+2*ii+1];
                for (iedge = 1; iedge <= MODL->body[ibodyl].nedge; iedge++) {
                    if (iedge == indx[  nindx+2*ii  ]) continue;
                    if (iedge == indx[  nindx+2*ii+2]) continue;

                    if (MODL->body[ibodyl].edge[iedge].ileft == iface ||
                        MODL->body[ibodyl].edge[iedge].irite == iface   ) {
                        if        (ii>0       && MODL->body[ibodyl].edge[iedge].ibeg == indx[        2*ii  ]) {
                            indx[        2*ii+1] = iedge;
                            indx[        2*ii+2] = MODL->body[ibodyl].edge[iedge].iend;
                        } else if (ii>0       && MODL->body[ibodyl].edge[iedge].iend == indx[        2*ii  ]) {
                            indx[        2*ii+1] = iedge;
                            indx[        2*ii+2] = MODL->body[ibodyl].edge[iedge].ibeg;
                        } else if (ii<nsize-1 && MODL->body[ibodyl].edge[iedge].ibeg == indx[        2*ii+2]) {
                            indx[        2*ii+1] = iedge;
                            indx[        2*ii  ] = MODL->body[ibodyl].edge[iedge].iend;
                        } else if (ii<nsize-1 && MODL->body[ibodyl].edge[iedge].iend == indx[        2*ii+2]) {
                            indx[        2*ii+1] = iedge;
                            indx[        2*ii  ] = MODL->body[ibodyl].edge[iedge].ibeg;
                        }

                        if        (ii>0       && MODL->body[ibodyl].edge[iedge].ibeg == indx[2*nindx+2*ii  ]) {
                            indx[2*nindx+2*ii+1] = iedge;
                            indx[2*nindx+2*ii+2] = MODL->body[ibodyl].edge[iedge].iend;
                        } else if (ii>0       && MODL->body[ibodyl].edge[iedge].iend == indx[2*nindx+2*ii  ]) {
                            indx[2*nindx+2*ii+1] = iedge;
                            indx[2*nindx+2*ii+2] = MODL->body[ibodyl].edge[iedge].ibeg;
                        } else if (ii<nsize-1 && MODL->body[ibodyl].edge[iedge].ibeg == indx[2*nindx+2*ii+2]) {
                            indx[2*nindx+2*ii+1] = iedge;
                            indx[2*nindx+2*ii  ] = MODL->body[ibodyl].edge[iedge].iend;
                        } else if (ii<nsize-1 && MODL->body[ibodyl].edge[iedge].iend == indx[2*nindx+2*ii+2]) {
                            indx[2*nindx+2*ii+1] = iedge;
                            indx[2*nindx+2*ii  ] = MODL->body[ibodyl].edge[iedge].ibeg;
                        }
                    }
                }

                /* Edges in ibodyr */
                iface = indx[4*nindx+2*ii+1];
                for (iedge = 1; iedge <= MODL->body[ibodyr].nedge; iedge++) {
                    if (iedge == indx[4*nindx+2*ii  ]) continue;
                    if (iedge == indx[4*nindx+2*ii+2]) continue;

                    if (MODL->body[ibodyr].edge[iedge].ileft == iface ||
                        MODL->body[ibodyr].edge[iedge].irite == iface   ) {
                        if        (ii>0       && MODL->body[ibodyr].edge[iedge].ibeg == indx[3*nindx+2*ii  ]) {
                            indx[3*nindx+2*ii+1] = iedge;
                            indx[3*nindx+2*ii+2] = MODL->body[ibodyr].edge[iedge].iend;
                        } else if (ii>0       && MODL->body[ibodyr].edge[iedge].iend == indx[3*nindx+2*ii  ]) {
                            indx[3*nindx+2*ii+1] = iedge;
                            indx[3*nindx+2*ii+2] = MODL->body[ibodyr].edge[iedge].ibeg;
                        } else if (ii<nsize-1 && MODL->body[ibodyr].edge[iedge].ibeg == indx[3*nindx+2*ii+2]) {
                            indx[3*nindx+2*ii+1] = iedge;
                            indx[3*nindx+2*ii  ] = MODL->body[ibodyr].edge[iedge].iend;
                        } else if (ii<nsize-1 && MODL->body[ibodyr].edge[iedge].iend == indx[3*nindx+2*ii+2]) {
                            indx[3*nindx+2*ii+1] = iedge;
                            indx[3*nindx+2*ii  ] = MODL->body[ibodyr].edge[iedge].ibeg;
                        }

                        if        (ii>0       && MODL->body[ibodyr].edge[iedge].ibeg == indx[5*nindx+2*ii  ]) {
                            indx[5*nindx+2*ii+1] = iedge;
                            indx[5*nindx+2*ii+2] = MODL->body[ibodyr].edge[iedge].iend;
                        } else if (ii>0       && MODL->body[ibodyr].edge[iedge].iend == indx[5*nindx+2*ii  ]) {
                            indx[5*nindx+2*ii+1] = iedge;
                            indx[5*nindx+2*ii+2] = MODL->body[ibodyr].edge[iedge].ibeg;
                        } else if (ii<nsize-1 && MODL->body[ibodyr].edge[iedge].ibeg == indx[5*nindx+2*ii+2]) {
                            indx[5*nindx+2*ii+1] = iedge;
                            indx[5*nindx+2*ii  ] = MODL->body[ibodyr].edge[iedge].iend;
                        } else if (ii<nsize-1 && MODL->body[ibodyr].edge[iedge].iend == indx[5*nindx+2*ii+2]) {
                            indx[5*nindx+2*ii+1] = iedge;
                            indx[5*nindx+2*ii  ] = MODL->body[ibodyr].edge[iedge].ibeg;
                        }
                    }
                }
            }

            /* insert Edges at beginning */
            iface = indx[  nindx+1];
            for (iedge = 1; iedge <= MODL->body[ibodyl].nedge; iedge++) {
                if (MODL->body[ibodyl].edge[iedge].ileft == iface ||
                    MODL->body[ibodyl].edge[iedge].irite == iface   ) {
                    if        (MODL->body[ibodyl].edge[iedge].ibeg == indx[2*nindx] &&
                               MODL->body[ibodyl].edge[iedge].iend == indx[      0]   ) {
                        indx[  nindx] = iedge;
                        break;
                    } else if (MODL->body[ibodyl].edge[iedge].ibeg == indx[      0] &&
                               MODL->body[ibodyl].edge[iedge].iend == indx[2*nindx]   ) {
                        indx[  nindx] = iedge;
                        break;
                    }
                }
            }

            iface = indx[4*nindx+1];
            for (iedge = 1; iedge <= MODL->body[ibodyr].nedge; iedge++) {
                if (MODL->body[ibodyr].edge[iedge].ileft == iface ||
                    MODL->body[ibodyr].edge[iedge].irite == iface   ) {
                    if        (MODL->body[ibodyr].edge[iedge].ibeg == indx[5*nindx] &&
                               MODL->body[ibodyr].edge[iedge].iend == indx[3*nindx]   ) {
                        indx[4*nindx] = iedge;
                        break;
                    } else if (MODL->body[ibodyr].edge[iedge].ibeg == indx[3*nindx] &&
                               MODL->body[ibodyr].edge[iedge].iend == indx[5*nindx]   ) {
                        indx[4*nindx] = iedge;
                        break;
                    }
                }
            }

            /* insert Edges at end */
            iface = indx[2*nindx-2];
            for (iedge = 1; iedge <= MODL->body[ibodyl].nedge; iedge++) {
                if (MODL->body[ibodyl].edge[iedge].ileft == iface ||
                    MODL->body[ibodyl].edge[iedge].irite == iface   ) {
                    if        (MODL->body[ibodyl].edge[iedge].ibeg == indx[  nindx-1] &&
                               MODL->body[ibodyl].edge[iedge].iend == indx[3*nindx-1]   ) {
                        indx[2*nindx-1] = iedge;
                        break;
                    } else if (MODL->body[ibodyl].edge[iedge].ibeg == indx[3*nindx-1] &&
                               MODL->body[ibodyl].edge[iedge].iend == indx[  nindx-1]   ) {
                        indx[2*nindx-1] = iedge;
                        break;
                    }
                }
            }

            iface = indx[5*nindx-2];
            for (iedge = 1; iedge <= MODL->body[ibodyr].nedge; iedge++) {
                if (MODL->body[ibodyr].edge[iedge].ileft == iface ||
                    MODL->body[ibodyr].edge[iedge].irite == iface   ) {
                    if        (MODL->body[ibodyr].edge[iedge].ibeg == indx[4*nindx-1] &&
                               MODL->body[ibodyr].edge[iedge].iend == indx[6*nindx-1]   ) {
                        indx[5*nindx-1] = iedge;
                        break;
                    } else if (MODL->body[ibodyr].edge[iedge].ibeg == indx[6*nindx-1] &&
                               MODL->body[ibodyr].edge[iedge].iend == indx[4*nindx-1]   ) {
                        indx[5*nindx-1] = iedge;
                        break;
                    }
                }
            }
        }

        /* if any entries in the indx map are zero,  then something went wrong */
        for (ii = 0; ii < 6*nindx; ii++) {
            if (indx[ii] <= 0) {
                signalError(MODL, OCSM_INTERNAL_ERROR, ibrch,
                            "problem finding indx map.  nindx=%d, indx[%d]=%d", nindx, ii, indx[ii]);

                for (jj = 0; jj < 6; jj++) {
                    SPRINT0x(0, "        ");
                    for (ii = 0; ii < nindx; ii++) {
                        SPRINT1x(0, "%5d", indx[jj*nindx+ii]);
                    }
                    SPRINT0(0, " ");
                }

                SPRINT1(0, "        ibodyl=%d", ibodyl);
                for (iedge = 1; iedge <= MODL->body[ibodyl].nedge; iedge++) {
                    SPRINT5(0, "           iedge=%-5d ileft=%-5d irite=%-5d ibeg=%-5d iend=%-5d", iedge,
                            MODL->body[ibodyl].edge[iedge].ileft,
                            MODL->body[ibodyl].edge[iedge].irite,
                            MODL->body[ibodyl].edge[iedge].ibeg,
                            MODL->body[ibodyl].edge[iedge].iend);
                }

                SPRINT1(0, "        ibodyr=%d", ibodyr);
                for (iedge = 1; iedge <= MODL->body[ibodyr].nedge; iedge++) {
                    SPRINT5(0, "           iedge=%-5d ileft=%-5d irite=%-5d ibeg=%-5d iend=%-5d", iedge,
                            MODL->body[ibodyr].edge[iedge].ileft,
                            MODL->body[ibodyr].edge[iedge].irite,
                            MODL->body[ibodyr].edge[iedge].ibeg,
                            MODL->body[ibodyr].edge[iedge].iend);
                }

                SET_STATUS(OCSM_INTERNAL_ERROR, connect);
            }
        }

        /* make the bridging Edges */
        newEdges = 0;
        for (ii = 0; ii <= nindx; ii+=2) {
            /* bottom */
            ibeg = indx[        ii];
            iend = indx[3*nindx+ii];
            status = makeEdge(MODL,
                              MODL->body[ibodyr].node[iend].enode,
                              MODL->body[ibodyl].node[ibeg].enode, &(enewEdges[newEdges++]));
            CHECK_STATUS(makeEdge);

            /* top */
            ibeg = indx[2*nindx+ii];
            iend = indx[5*nindx+ii];
            status = makeEdge(MODL,
                              MODL->body[ibodyr].node[iend].enode,
                              MODL->body[ibodyl].node[ibeg].enode, &(enewEdges[newEdges++]));
            CHECK_STATUS(makeEdge);
        }
        /* make the bridging Faces */
        newFaces = 0;
        for (ii = 0; ii < nindx-1; ii+=2) {
            /* bottom */
            ibeg = indx[       +ii+1];
            iend = indx[3*nindx+ii+1];
            eedges_in[0] = enewEdges[ii  ];
            eedges_in[1] = MODL->body[ibodyl].edge[ibeg].eedge;
            eedges_in[2] = enewEdges[ii+2];
            eedges_in[3] = MODL->body[ibodyr].edge[iend].eedge;
            status = makeFace(MODL, eedges_in, fillstyle, &(enewFaces[newFaces++]));
            CHECK_STATUS(makeFace);

            /* top */
            ibeg = indx[2*nindx+ii+1];
            iend = indx[5*nindx+ii+1];
            eedges_in[0] = enewEdges[ii+1];
            eedges_in[1] = MODL->body[ibodyl].edge[ibeg].eedge;
            eedges_in[2] = enewEdges[ii+3];
            eedges_in[3] = MODL->body[ibodyr].edge[iend].eedge;
            status = makeFace(MODL, eedges_in, fillstyle, &(enewFaces[newFaces++]));
            CHECK_STATUS(makeFace);
        }

        /* beg cap */
        ibeg = indx[  nindx  ];
        iend = indx[4*nindx  ];
        eedges_in[0] = enewEdges[0];
        eedges_in[1] = MODL->body[ibodyl].edge[ibeg].eedge;
        eedges_in[2] = enewEdges[1];
        eedges_in[3] = MODL->body[ibodyr].edge[iend].eedge;
        status = makeFace(MODL, eedges_in, fillstyle, &(enewFaces[newFaces++]));
        CHECK_STATUS(makeFace);

        /* end cap */
        ibeg = indx[2*nindx-1];
        iend = indx[5*nindx-1];
        eedges_in[0] = enewEdges[newEdges-2];
        eedges_in[1] = MODL->body[ibodyl].edge[ibeg].eedge;
        eedges_in[2] = enewEdges[newEdges-1];
        eedges_in[3] = MODL->body[ibodyr].edge[iend].eedge;
        status = makeFace(MODL, eedges_in, fillstyle, &(enewFaces[newFaces++]));
        CHECK_STATUS(makeFace);
#endif

        /* add the Faces that are not in the faceLists into enewFaces */
        status = EG_getBodyTopos(MODL->body[ibodyl].ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        for (iface = 1; iface <= nface; iface++) {
            iokay = 1;

            for (ii = 0; ii < nsize; ii++) {
                if (iface == NINT(args[1].val[ii])) {
                    iokay = 0;
                    break;
                }
            }

            if (iokay == 1) {
                enewFaces[newFaces++] = efaces[iface-1];
            }
        }

        EG_free(efaces);

        status = EG_getBodyTopos(MODL->body[ibodyr].ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        for (iface = 1; iface <= nface; iface++) {
            iokay = 1;

            for (ii = 0; ii < nsize; ii++) {
                if (iface == NINT(args[2].val[ii])) {
                    iokay = 0;
                    break;
                }
            }

            if (iokay == 1) {
                enewFaces[newFaces++] = efaces[iface-1];
            }
        }

        EG_free(efaces);

        /* sew the Faces together (with 0 tolerance, since everything should
           exactly fit tigether) */
        status = EG_sewFaces(newFaces, enewFaces, 0.0, 0, &emodel);
        CHECK_STATUS(EG_sewFaces);

        /* if the Model contains a single SolidBody, extract it */
        status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                                data, &nchild, &ebodys, &senses);
        CHECK_STATUS(EG_getTopology);

        if (nchild < 1) {
            SET_STATUS(OCSM_DID_NOT_CREATE_BODY, connect);
        } else if (nchild > 1) {
            SET_STATUS(OCSM_CREATED_TOO_MANY_BODYS, connect);
        } else {
            status = EG_getTopology(ebodys[0], &eref, &oclass, &mtype,
                                    data, &nface, &efaces, &senses);
            CHECK_STATUS(EG_getTopology);

            if (oclass != BODY || mtype != SOLIDBODY) {
                signalError(MODL, OCSM_DID_NOT_CREATE_BODY, ibrch,
                            "CONNECT expected a SOLILD Body to be produced");
                SET_STATUS(OCSM_DID_NOT_CREATE_BODY, connect);
            }

            status = EG_copyObject(ebodys[0], NULL, &ebody);
            CHECK_STATUS(EG_copyObject);

            MODL->body[ibody].ebody = ebody;
        }

        status = EG_deleteObject(emodel);
        CHECK_STATUS(EG_deleteObject);

        /* mark the new Faces with the current Body */
        status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        for (iface = 1; iface <= nface; iface++) {
            status = EG_attributeRet(efaces[iface-1], "body",
                                     &itype, &nlist,
                                     &tempIlist, &tempRlist, &tempClist);
            if (status == EGADS_NOTFOUND) {
                status = setFaceAttribute(MODL, ibody, iface, 0, iface, npatn, patn);
                CHECK_STATUS(setFaceAttribute);
            } else {
                CHECK_STATUS(EG_attributeRet);
            }
        }

        EG_free(efaces);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);
    }

cleanup:
    FREE(elist    );
    FREE(eedges   );
    FREE(matches  );
    FREE(enewFaces);
    FREE(enewEdges);
    FREE(indx     );

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   buildGrown - implement OCSM_GROWNs for ocsmBuild                   *
 *                                                                      *
 ************************************************************************
 */

static int
buildGrown(modl_T *modl,
           int    ibrch,
           varg_T args[],
           int    *nstack,
           int    stack[],
           int    npatn,
           patn_T patn[])
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       type, hasdots, iface, nface, nloops, loftOpts, itypebeg, itypeend, i, ival;
    int       numSketches, numEdges, ibody, ibodyl, ibodyr, isketches[MAX_NUM_SKETCHES];
    double    toler, dirn[3], alen;

    int        oclass, mtype, *senses, iford1, nedges, nfaces, nchild;
    int        j, nosePmtr, tailPmtr, mode, nsew, *pinfo=NULL;
    double     data[4], Rnose[8], Rtail[8], matrix[12], len;
    double     dx, dy, dz, cosx, sinx, cosy, siny, cosz, sinz;
    double     *prv=NULL, bbox[6], vals[7];
    ego        ebody, ebodyl, ebodyr, *efaces, *echildren, esketches[MAX_NUM_SKETCHES];
    ego        eref, exform, etemp, *echilds, *esew, emodel;

    ROUTINE(buildGrown);
    DPRINT2("%s(ibrch=%d) {",
            routine, ibrch);

    /* --------------------------------------------------------------- */

    type    = MODL->brch[ibrch].type;
    hasdots = 0;

    /* execute: "extrude dx dy dz" */
    if (type == OCSM_EXTRUDE) {
        SPRINT4(1, "    executing [%4d] extrude:    %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0) {
            hasdots = 1;

            SPRINT3(1, "                                 %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0]);
        }

        /* check to be sure that the extrude distance is not zero */
        if (fabs(args[1].val[0]) < EPS06 &&
            fabs(args[2].val[0]) < EPS06 &&
            fabs(args[3].val[0]) < EPS06   ) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "extrude distance must be non-zero");
            status = OCSM_ILLEGAL_VALUE;
            goto cleanup;
        }

        /* pop a Sketch from the stack */
        if ((*nstack) < 1) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "EXTRUDE expects a Body on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, extrude);
        } else {
                ibodyl = stack[--(*nstack)];
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        alen = sqrt(args[1].val[0]*args[1].val[0] + args[2].val[0]*args[2].val[0] + args[3].val[0]*args[3].val[0]);
        dirn[0] = args[1].val[0] / alen;
        dirn[1] = args[2].val[0] / alen;
        dirn[2] = args[3].val[0] / alen;

        /* ibodyl is a SheetBody, so create a SolidBody */
        if (MODL->body[ibodyl].botype == OCSM_SHEET_BODY) {

            /* create the Body */
            status = newBody(MODL, ibrch, OCSM_EXTRUDE, ibodyl, -1,
                             args, hasdots, OCSM_SOLID_BODY, &ibody);
            CHECK_STATUS(newBody);

            ebodyl = MODL->body[ibodyl].ebody;

            if (outLevel >= 3) {
                SPRINT0(3, "before EG_extrude: ebodyl");
                printEgo(ebodyl);
            }
            status = EG_extrude(ebodyl, alen, dirn, &ebody);
            CHECK_STATUS(EG_extrude);

            MODL->body[ibody].ebody = ebody;

        /* ibodyl is a WireBody, so create a SheetBody */
        } else if (MODL->body[ibodyl].botype == OCSM_WIRE_BODY) {

            /* create the Body */
            status = newBody(MODL, ibrch, OCSM_EXTRUDE, ibodyl, -1,
                             args, hasdots, OCSM_SHEET_BODY, &ibody);
            CHECK_STATUS(newBody);

            ebodyl = MODL->body[ibodyl].ebody;

            if (outLevel >= 3) {
                SPRINT0(3, "before EG_extrude: ebodyl");
                printEgo(ebodyl);
            }
            status = EG_extrude(ebodyl, alen, dirn, &ebody);
            CHECK_STATUS(EG_extrude);

            MODL->body[ibody].ebody = ebody;
        } else {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                        "EXTRUDE expects a SHEET or WIRE Body");
            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, extrude);
        }

        /* mark the new Faces with the current Branch */
        status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        for (iface = 1; iface <= nface; iface++) {
            if        (iface == nface-1) {
                iford1 = 1;
            } else if (iface == nface  ) {
                iford1 = 2;
            } else {
                iford1 = iface + 2;
            }
            status = setFaceAttribute(MODL, ibody, iface, 0, iford1, npatn, patn);
            CHECK_STATUS(setFaceAttribute);
        }

        EG_free(efaces);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "rule reorder=0" */
    } else if (type == OCSM_RULE) {
        SPRINT2(1, "    executing [%4d] rule:       %11.5f",
                ibrch, args[1].val[0]);

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            while ((*nstack) > 0) {
                ibodyl = stack[--(*nstack)];
                if (ibodyl == 0) break;      /* mark is found */
            }

            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* pop Sketches from the stack (until the Mark) and put them
           into isketches from first to last */
        numSketches =  0;
        numEdges    = -1;
        itypebeg    = -1;
        itypeend    = -1;
        ibodyl      = -1;
        while ((*nstack) > 0) {
            ibodyl = stack[--(*nstack)];

            if (ibodyl == 0) break;          /* mark is found */

            if (MODL->body[ibodyl].botype == OCSM_NODE_BODY) {
                if (numSketches == 0) {
                    itypeend = MODL->body[ibodyl].botype;
                }

                itypebeg = MODL->body[ibodyl].botype;

                for (j = numSketches; j > 0; j--) {
                    isketches[j] = isketches[j-1];
                    esketches[j] = esketches[j-1];
                }
                isketches[0] = ibodyl;
                esketches[0] = MODL->body[ibodyl].ebody;
                numSketches++;

            } else if (MODL->body[ibodyl].botype == OCSM_WIRE_BODY ||
                       MODL->body[ibodyl].botype == OCSM_SHEET_BODY  ) {
                ebodyl = MODL->body[ibodyl].ebody;

                status = EG_getBodyTopos(ebodyl, NULL, LOOP, &nloops, &echildren);
                CHECK_STATUS(EG_getBodyTopos);

                EG_free(echildren);

                if (nloops != 1) {
                    signalError(MODL, OCSM_ERROR_IN_BODYS_ON_STACK, ibrch,
                                "expecting 1 Loop but found %d", nloops);
                    SET_STATUS(OCSM_ERROR_IN_BODYS_ON_STACK, rule);
                }

                status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedges, &echildren);
                CHECK_STATUS(EG_getBodyTopos);

                EG_free(echildren);

                if (numSketches == 0) {
                    itypeend = MODL->body[ibodyl].botype;
                }

                itypebeg = MODL->body[ibodyl].botype;

                if (numEdges < 0) {
                    numEdges = nedges;
                } else if (numEdges != nedges) {
                    signalError(MODL, OCSM_ERROR_IN_BODYS_ON_STACK, ibrch,
                                "expecting %d Edges but found %d", numEdges, nedges);
                    SET_STATUS(OCSM_ERROR_IN_BODYS_ON_STACK, rule);
                }

                for (j = numSketches; j > 0; j--) {
                    isketches[j] = isketches[j-1];
                    esketches[j] = esketches[j-1];
                }
                isketches[0] = ibodyl;
                esketches[0] = ebodyl;
                numSketches++;
            } else {
                signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                            "Body %d is neither NODE, WIRE, nor SHEET", ibodyl);
                SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, rule);
            }
        }

        if (numSketches < 2) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "RULE expects 2 or more Bodys since mark");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, rule);
        } else {
            SPRINT2(1, "                          ruling  %d Sketches with %d Edges each...",
                    numSketches, numEdges);
        }

        /* create the Body (assume for now it is a SolidBody --- to be fixed below) */
        status = newBody(MODL, ibrch, OCSM_RULE, ibodyl, ibodyl,
                         args, hasdots, OCSM_SOLID_BODY, &ibody);
        CHECK_STATUS(newBody);

        /* link this new Body with its children */
        for (i = 0; i < numSketches; i++) {
            if (isketches[i] != 0) {
                MODL->body[isketches[i]].ichld = ibody;
            }
        }

        MODL->body[ibody].ileft = isketches[0];
        MODL->body[ibody].irite = isketches[numSketches-1];

        /* first Sketch should be a NODE, WIREBODY, or FACE */
        if (itypebeg == OCSM_SHEET_BODY) {
            status = EG_getBodyTopos(esketches[0], NULL, FACE, &nfaces, &echildren);
            CHECK_STATUS(EG_getBodyTopos);

            esketches[0] = echildren[0];

            EG_free(echildren);
        }

        /* last Sketch should be a NODE, WIREBODY, or FACE */
        if (itypeend == OCSM_SHEET_BODY) {
            status = EG_getBodyTopos(esketches[numSketches-1], NULL, FACE, &nfaces, &echildren);
            CHECK_STATUS(EG_getBodyTopos);

            esketches[numSketches-1] = echildren[0];

            EG_free(echildren);
        }

        /* convert all the intermediate Sketches to LOOPs */
        for (i = 1; i < numSketches-1; i++) {
            status = EG_getBodyTopos(esketches[i], NULL, LOOP, &nloops, &echildren);
            CHECK_STATUS(EG_getBodyTopos);

            esketches[i] = echildren[0];

            EG_free(echildren);
        }

        /* reorder Loops to improve results */
        status = reorderLoops(MODL, numSketches, esketches, NINT(args[1].val[0]));
        CHECK_STATUS(reorderLoops);

        /* create the Body */
        if (outLevel >= 3) {
            int idum;
            for (idum = 0; idum < numSketches; idum++) {
                SPRINT1(3, "before EG_ruled: esketches[%d]", idum);
                printEgo(esketches[idum]);
            }
        }
        status = EG_ruled(numSketches, esketches, &ebody);
        CHECK_STATUS(EG_ruled);

        MODL->body[ibody].ebody = ebody;

        /* update the Body type if a SheetBody was produced */
        status = EG_getTopology(ebody, &eref, &oclass, &mtype,
                                data, &nchild, &echilds, &senses);
        CHECK_STATUS(EG_getTopology);

        if (oclass == BODY) {
            if (mtype == SHEETBODY || mtype == FACEBODY) {
                SPRINT0(0, "WARNING:: rule produced a SheetBody");
                MODL->body[ibody].botype = OCSM_SHEET_BODY;
            } else if (mtype != SOLIDBODY) {
                signalError(MODL, OCSM_INTERNAL_ERROR, ibrch,
                            "Body produced is neither SHEET or SOLID");
                SET_STATUS(OCSM_INTERNAL_ERROR, rule);
            }
        } else {
            signalError(MODL, OCSM_INTERNAL_ERROR, ibrch,
                        "EGO produced is not a Body");
            SET_STATUS(OCSM_INTERNAL_ERROR, rule);
        }

        /* mark the new Faces with the current Branch */
        status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        for (iface = 1; iface <= nface; iface++) {
            if        (iface == nface-1) {
                iford1 = 1;
            } else if (iface == nface  ) {
                iford1 = 2;
            } else {
                iford1 = iface + 2;
            }
            status = setFaceAttribute(MODL, ibody, iface, 0, iford1, npatn, patn);
            CHECK_STATUS(setFaceAttribute);
        }

        EG_free(efaces);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "loft smooth" */
    } else if (type == OCSM_LOFT) {
        SPRINT2(1, "    executing [%4d] loft:       %11.5f",
                ibrch, args[1].val[0]);

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            while ((*nstack) > 0) {
                ibodyl = stack[--(*nstack)];
                if (ibodyl == 0) break;      /* mark is found */
            }

            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* pop Sketches from the stack (until the Mark) and put them
           into isketches from last to first */
        numSketches = 0;
        loftOpts    = -1;
        ibodyl      = -1;
        ibodyr      = -1;
        while ((*nstack) > 0) {
            ibodyl = stack[--(*nstack)];

            if (ibodyl == 0) break;          /* mark is found */

            if (numSketches == 0) {
                if (MODL->body[ibodyl].botype == OCSM_WIRE_BODY) {
                    loftOpts = 0;
                } else {
                    loftOpts = 1;
                }
            }

            ebodyl = MODL->body[ibodyl].ebody;

            if (MODL->body[ibodyl].botype == OCSM_NODE_BODY) {
                isketches[numSketches] = ibodyl;
                esketches[numSketches] = ebodyl;
                numSketches++;
            } else {
                status = EG_getBodyTopos(ebodyl, NULL, LOOP, &nloops, &echildren);
                CHECK_STATUS(EG_getBodyTopos);

                isketches[numSketches] = ibodyl;
                esketches[numSketches] = echildren[0];
                numSketches++;

                EG_free(echildren);
            }

            ibodyr = ibodyl;
        }

        if (numSketches < 2 || ibodyl < 0) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "UNION expects 2 or more Bodys since mark");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, loft);
        } else {
            SPRINT1(1, "                          lofting %d Sketches...", numSketches);
        }
        ibodyl = isketches[0];

        /* determine if SHEET or SOLID should be created */
        if (MODL->body[ibodyr].botype == OCSM_WIRE_BODY) {
            loftOpts = 0;
        }

        /* create the new Body */
        if (loftOpts == 0) {
            status = newBody(MODL, ibrch, OCSM_LOFT, ibodyr, ibodyl,
                             args, hasdots, OCSM_SHEET_BODY, &ibody);
            CHECK_STATUS(newBody);
        } else {
            status = newBody(MODL, ibrch, OCSM_LOFT, ibodyr, ibodyl,
                             args, hasdots, OCSM_SOLID_BODY, &ibody);
            CHECK_STATUS(newBody);
        }

        /* all Bodys in Loft should point to new Body */
        for (i = 0; i < numSketches; i++) {
            if (isketches[i] != 0) {
                MODL->body[isketches[i]].ichld = ibody;
            }
        }

        if (NINT(args[1].val[0]) != 1) {
            loftOpts += 2;
        }

        if (outLevel >= 3) {
            int idum;
            for (idum = 0; idum < numSketches; idum++) {
                SPRINT1(3, "before EG_loft: esketches[%d]", idum);
                printEgo(esketches[idum]);
            }
        }
        status = EG_loft(numSketches, esketches, loftOpts, &ebody);
        CHECK_STATUS(EG_loft);

        /* if a SHEET Body and ibodyl is SHEET, sew them together */
        if (MODL->body[ibody ].botype == OCSM_SHEET_BODY &&
            MODL->body[ibodyl].botype == OCSM_SHEET_BODY   ) {

            status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
            CHECK_STATUS(EG_getBodyTopos);

            nsew = nface + 1;
            MALLOC(esew, ego, nsew);
            for (i = 0; i < nsew-1; i++) {
                esew[i] = efaces[i];
            }

            EG_free(efaces);

            status = EG_getBodyTopos(MODL->body[ibodyl].ebody, NULL, FACE, &nface, &efaces);
            CHECK_STATUS(EG_getBodyTopos);

            esew[nsew-1] = efaces[0];

            EG_free(efaces);

            status = EG_sewFaces(nsew, esew, 0.0, 1, &emodel);
            CHECK_STATUS(EG_sewFaces);

            FREE(esew);

            status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                                    data, &nchild, &echilds, &senses);
            CHECK_STATUS(EG_getTopology);

            ebody = echilds[0];

        /* if a SHEET Body and ibodyr is SHEET, sew them together */
        } else if (MODL->body[ibody ].botype == OCSM_SHEET_BODY &&
                   MODL->body[ibodyr].botype == OCSM_SHEET_BODY   ) {

            status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
            CHECK_STATUS(EG_getBodyTopos);

            nsew = nface + 1;
            MALLOC(esew, ego, nsew);
            for (i = 0; i < nsew-1; i++) {
                esew[i] = efaces[i];
            }

            EG_free(efaces);

            status = EG_getBodyTopos(MODL->body[ibodyr].ebody, NULL, FACE, &nface, &efaces);
            CHECK_STATUS(EG_getBodyTopos);

            esew[nsew-1] = efaces[0];

            EG_free(efaces);

            status = EG_sewFaces(nsew, esew, 0.0, 1, &emodel);
            CHECK_STATUS(EG_sewFaces);

            FREE(esew);

            status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                                    data, &nchild, &echilds, &senses);
            CHECK_STATUS(EG_getTopology);

            ebody = echilds[0];
        }

        MODL->body[ibody].ebody = ebody;

        /* mark the new Faces with the current Branch */
        status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        for (iface = 1; iface <= nface; iface++) {
            if        (iface == nface-1) {
                iford1 = 1;
            } else if (iface == nface  ) {
                iford1 = 2;
            } else {
                iford1 = iface + 2;
            }
            status = setFaceAttribute(MODL, ibody, iface, 0, iford1, npatn, patn);
            CHECK_STATUS(setFaceAttribute);
        }

        EG_free(efaces);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "blend noseList=0 tailList=0 reorder=0" */
    } else if (type == OCSM_BLEND) {
        SPRINT4(1, "    executing [%4d] blend:      %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0]);

        for (ival = 0; ival < args[1].nval; ival++) {
            if (args[1].dot[ival] != 0) hasdots = 1;
        }

        for (ival = 0; ival < args[2].nval; ival++) {
            if (args[2].dot[ival] != 0) hasdots = 1;
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            while ((*nstack) > 0) {
                ibodyl = stack[--(*nstack)];
                if (ibodyl == 0) break;      /* mark is found */
            }

            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* pop Sketches from the stack (until the Mark) and put them
           into isketches from first to last */
        numSketches =  0;
        numEdges    = -1;
        itypebeg    = -1;
        itypeend    = -1;
        ibodyl      = -1;
        while ((*nstack) > 0) {
            ibodyl = stack[--(*nstack)];

            if (ibodyl == 0) break;          /* mark is found */

            if (MODL->body[ibodyl].botype == OCSM_NODE_BODY) {
                if (numSketches == 0) {
                    itypeend = MODL->body[ibodyl].botype;
                }

                itypebeg = MODL->body[ibodyl].botype;

                for (j = numSketches; j > 0; j--) {
                    isketches[j] = isketches[j-1];
                    esketches[j] = esketches[j-1];
                }
                isketches[0] = ibodyl;
                esketches[0] = MODL->body[ibodyl].ebody;
                numSketches++;

            } else if (MODL->body[ibodyl].botype == OCSM_WIRE_BODY ||
                       MODL->body[ibodyl].botype == OCSM_SHEET_BODY  ) {
                ebodyl = MODL->body[ibodyl].ebody;

                status = EG_getBodyTopos(ebodyl, NULL, LOOP, &nloops, &echildren);
                CHECK_STATUS(EG_getBodyTopos);

                EG_free(echildren);

                if (nloops != 1) {
                    signalError(MODL, OCSM_ERROR_IN_BODYS_ON_STACK, ibrch,
                                "expecting 1 Loop but found %d", nloops);
                    SET_STATUS(OCSM_ERROR_IN_BODYS_ON_STACK, blend);
                }

                status = EG_getBodyTopos(ebodyl, NULL, EDGE, &nedges, &echildren);
                CHECK_STATUS(EG_getBodyTopos);

                EG_free(echildren);

                if (numSketches == 0) {
                    itypeend = MODL->body[ibodyl].botype;
                }

                itypebeg = MODL->body[ibodyl].botype;

                if (numEdges < 0) {
                    numEdges = nedges;
                } else if (numEdges != nedges) {
                    signalError(MODL, OCSM_ERROR_IN_BODYS_ON_STACK, ibrch,
                                "expecting %d Edges but found %d", numEdges, nedges);
                    SET_STATUS(OCSM_ERROR_IN_BODYS_ON_STACK, blend);
                }

                for (j = numSketches; j > 0; j--) {
                    isketches[j] = isketches[j-1];
                    esketches[j] = esketches[j-1];
                }
                isketches[0] = ibodyl;
                esketches[0] = ebodyl;
                numSketches++;
            } else {
                signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                            "Body %d is neither NODE, WIRE, nor SHEET", ibodyl);
                SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, blend);
            }
        }

        if (numSketches < 2) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "BLEND expects 2 or more Bodys since mark");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, blend);
        } else {
            SPRINT2(1, "                          blending  %d Sketches with %d Edges each...",
                    numSketches, numEdges);
        }

        /* create the Body (assume for now it is a SolidBody --- to be fixed below) */
        status = newBody(MODL, ibrch, OCSM_BLEND, ibodyl, ibodyl,
                         args, hasdots, OCSM_SOLID_BODY, &ibody);
        CHECK_STATUS(newBody);

        /* link this new Body with its children */
        for (i = 0; i < numSketches; i++) {
            if (isketches[i] != 0) {
                MODL->body[isketches[i]].ichld = ibody;
            }
        }

        MODL->body[ibody].ileft = isketches[0];
        MODL->body[ibody].irite = isketches[numSketches-1];

        /* first Sketch should be a NODE, WIREBODY, or FACE */
        if (itypebeg == OCSM_SHEET_BODY) {
            status = EG_getBodyTopos(esketches[0], NULL, FACE, &nfaces, &echildren);
            CHECK_STATUS(EG_getBodyTopos);

            esketches[0] = echildren[0];

            EG_free(echildren);
        }

        /* last Sketch should be a NODE, WIREBODY, or FACE */
        if (itypeend == OCSM_SHEET_BODY) {
            status = EG_getBodyTopos(esketches[numSketches-1], NULL, FACE, &nfaces, &echildren);
            CHECK_STATUS(EG_getBodyTopos);

            esketches[numSketches-1] = echildren[0];

            EG_free(echildren);
        }

        /* convert all the intermediate Sketches to LOOPs */
        for (i = 1; i < numSketches-1; i++) {
            status = EG_getBodyTopos(esketches[i], NULL, LOOP, &nloops, &echildren);
            CHECK_STATUS(EG_getBodyTopos);

            esketches[i] = echildren[0];

            EG_free(echildren);
        }

        /* store info from noseList into Rnose */
        if (args[1].val[0] == 0) {
            nosePmtr = 0;
        } else {
            nosePmtr = 1;
            if (MODL->body[isketches[0]].botype == OCSM_NODE_BODY) {
                if (fabs(args[1].val[0]) < EPS06) {
                    nosePmtr = 0;
                } else if (args[1].nval >= 8) {
                    for (j = 0; j < 8; j++) {
                        Rnose[j] = args[1].val[j];
                    }

                    /* normalize */
                    len = sqrt(Rnose[1]*Rnose[1] + Rnose[2]*Rnose[2] + Rnose[3]*Rnose[3]);
                    Rnose[1] /= len;
                    Rnose[2] /= len;
                    Rnose[3] /= len;

                    len = sqrt(Rnose[5]*Rnose[5] + Rnose[6]*Rnose[6] + Rnose[7]*Rnose[7]);
                    Rnose[5] /= len;
                    Rnose[6] /= len;
                    Rnose[7] /= len;
                } else {
                    SPRINT1(0, "WARNING:: noseList only has %d values", args[1].nval);
                    SPRINT0(0, "          reverting to sharp nose");
                    nosePmtr = 0;
                }
            } else {
                if (fabs(args[1].val[0]) < EPS06) {
                    nosePmtr = 0;
                } else if (args[1].nval >= 4) {
                    for (j = 0; j < 4; j++) {
                        Rnose[j] = args[1].val[j];
                    }
                    for (j = 4; j < 8; j++) {
                        Rnose[j] = 0;
                    }

                    /* normalize */
                    len = sqrt(Rnose[1]*Rnose[1] + Rnose[2]*Rnose[2] + Rnose[3]*Rnose[3]);
                    Rnose[1] /= len;
                    Rnose[2] /= len;
                    Rnose[3] /= len;
                } else {
                    SPRINT1(0, "WARNING:: noseList only has %d values", args[1].nval);
                    SPRINT0(0, "          slope not being set");
                    nosePmtr = 0;
                }
            }
        }

        /* store info from tailList into Rtail */
        if (args[2].val[0] == 0) {
            tailPmtr = 0;
        } else {
            tailPmtr = 1;
            if (MODL->body[isketches[numSketches-1]].botype == OCSM_NODE_BODY) {
                if (fabs(args[2].val[0]) < EPS06) {
                    tailPmtr = 0;
                } else if (args[2].nval >= 8) {
                    for (j = 0; j < 8; j++) {
                        Rtail[j] = args[2].val[j];
                    }

                    /* normalize */
                    len = sqrt(Rtail[1]*Rtail[1] + Rtail[2]*Rtail[2] + Rtail[3]*Rtail[3]);
                    Rtail[1] /= len;
                    Rtail[2] /= len;
                    Rtail[3] /= len;

                    len = sqrt(Rtail[5]*Rtail[5] + Rtail[6]*Rtail[6] + Rtail[7]*Rtail[7]);
                    Rtail[5] /= len;
                    Rtail[6] /= len;
                    Rtail[7] /= len;
                } else {
                    SPRINT1(0, "WARNING:: tailList only has %d values", args[2].nval);
                    SPRINT0(0, "          reverting to sharp tail");
                    tailPmtr = 0;
                }
            } else {
                if (fabs(args[2].val[0]) < EPS06) {
                    tailPmtr = 0;
                } else if (args[2].nval >= 4) {
                    for (j = 0; j < 4; j++) {
                        Rtail[j] = args[2].val[j];
                    }
                    for (j = 4; j < 8; j++) {
                        Rtail[j] = 0;
                    }

                    /* normalize */
                    len = sqrt(Rtail[1]*Rtail[1] + Rtail[2]*Rtail[2] + Rtail[3]*Rtail[3]);
                    Rtail[1] /= len;
                    Rtail[2] /= len;
                    Rtail[3] /= len;
                } else {
                    SPRINT1(0, "WARNING:: tailList only has %d values", args[2].nval);
                    SPRINT0(0, "          slope not being set");
                    tailPmtr = 0;
                }
            }
        }

        /* reorder Loops to improve results */
        status = reorderLoops(MODL, numSketches, esketches, NINT(args[3].val[0]));
        CHECK_STATUS(reorderLoops);

        /* create the Body */
        if (nosePmtr == 0 && tailPmtr == 0) {
            SPRINT1(2, "call EG_blend(numSketches=%d, NULL, NULL)", numSketches);
            if (outLevel >= 3) {
                int idum;
                for (idum = 0; idum < numSketches; idum++) {
                    SPRINT1(3, "before EG_blend: esketches[%d]", idum);
                    printEgo(esketches[idum]);
                }
            }
            status = EG_blend(numSketches, esketches, NULL,  NULL,  &ebody);
            CHECK_STATUS(EG_blend);
        } else if (nosePmtr == 0) {
            SPRINT8(2, "Rtail=%f %f %f %f\n      %f %f %f %f",
                    Rtail[0], Rtail[1], Rtail[2], Rtail[3],
                    Rtail[4], Rtail[5], Rtail[6], Rtail[7]);

            SPRINT1(2, "call EG_blend(numSketches=%d, NULL, Rtail)", numSketches);
            if (outLevel >= 3) {
                int idum;
                for (idum = 0; idum < numSketches; idum++) {
                    SPRINT1(3, "before EG_blend: esketches[%d]", idum);
                    printEgo(esketches[idum]);
                }
            }
            status = EG_blend(numSketches, esketches, NULL,  Rtail, &ebody);
            CHECK_STATUS(EG_blend);
        } else if (tailPmtr == 0) {
            SPRINT8(2, "Rnose=%f %f %f %f\n      %f %f %f %f",
                    Rnose[0], Rnose[1], Rnose[2], Rnose[3],
                    Rnose[4], Rnose[5], Rnose[6], Rnose[7]);

            SPRINT1(2, "call EG_blend(numSketches=%d, Rnose, NULL)", numSketches);
            if (outLevel >= 3) {
                int idum;
                for (idum = 0; idum < numSketches; idum++) {
                    SPRINT1(3, "before EG_blend: esketches[%d]", idum);
                    printEgo(esketches[idum]);
                }
            }
            status = EG_blend(numSketches, esketches, Rnose, NULL,  &ebody);
            CHECK_STATUS(EG_blend);
        } else {
            SPRINT8(2, "Rnose=%f %f %f %f\n      %f %f %f %f",
                    Rnose[0], Rnose[1], Rnose[2], Rnose[3],
                    Rnose[4], Rnose[5], Rnose[6], Rnose[7]);

            SPRINT8(2, "Rtail=%f %f %f %f\n      %f %f %f %f",
                    Rtail[0], Rtail[1], Rtail[2], Rtail[3],
                    Rtail[4], Rtail[5], Rtail[6], Rtail[7]);

            SPRINT1(2, "call EG_blend(numSketches=%d, Rnose, Rtail)", numSketches);
            if (outLevel >= 3) {
                int idum;
                for (idum = 0; idum < numSketches; idum++) {
                    SPRINT1(3, "before EG_blend: esketches[%d]", idum);
                    printEgo(esketches[idum]);
                }
            }
            status = EG_blend(numSketches, esketches, Rnose, Rtail, &ebody);
            CHECK_STATUS(EG_blend);
        }

        MODL->body[ibody].ebody = ebody;

        /* update the Body type if a SheetBody was produced */
        status = EG_getTopology(ebody, &eref, &oclass, &mtype,
                                data, &nchild, &echilds, &senses);
        CHECK_STATUS(EG_getTopology);

        if (oclass == BODY) {
            if (mtype == SHEETBODY || mtype == FACEBODY) {
                SPRINT0(0, "WARNING:: blend produced a SheetBody");
                MODL->body[ibody].botype = OCSM_SHEET_BODY;
            } else if (mtype != SOLIDBODY) {
                signalError(MODL, OCSM_INTERNAL_ERROR, ibrch,
                            "Body produced is neither SHEET or SOLID");
                SET_STATUS(OCSM_INTERNAL_ERROR, blend);
            }
        } else {
            signalError(MODL, OCSM_INTERNAL_ERROR, ibrch,
                        "EGO produced is not a Body");
            SET_STATUS(OCSM_INTERNAL_ERROR, blend);
        }

        /* mark the new Faces with the current Branch */
        status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        for (iface = 1; iface <= nface; iface++) {
            if        (iface == nface-1) {
                iford1 = 1;
            } else if (iface == nface  ) {
                iford1 = 2;
            } else {
                iford1 = iface + 2;
            }
            status = setFaceAttribute(MODL, ibody, iface, 0, iford1, npatn, patn);
            CHECK_STATUS(setFaceAttribute);
        }

        EG_free(efaces);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "revolve xorig yorig zorig dxaxis dyaxis dzaxis angDeg" */
    } else if (type == OCSM_REVOLVE) {
        SPRINT8(1, "    executing [%4d] revolve:    %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0], args[4].val[0],
                       args[5].val[0], args[6].val[0], args[7].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0 || args[4].dot[0] != 0 ||
            args[5].dot[0] != 0 || args[6].dot[0] != 0 || args[7].dot[0] != 0                   ) {
            hasdots = 1;

            SPRINT7(1, "                                 %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0], args[4].dot[0],
                       args[5].dot[0], args[6].dot[0], args[7].dot[0]);
        }

        /* check to be sure that the axis is not zero */
        if (fabs(args[4].val[0]) < EPS06 &&
            fabs(args[5].val[0]) < EPS06 &&
            fabs(args[6].val[0]) < EPS06   ) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "axis length must be non-zero");
            status = OCSM_ILLEGAL_VALUE;
            goto cleanup;
        }

        /* check to be sure that the angDeg is non-zero */
        if        (fabs(args[7].val[0]) < EPS06) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "angDeg=%f cannot be zero", args[7].val[0]);
            status = OCSM_ILLEGAL_VALUE;
            goto cleanup;
        }

        /* pop a Sketch from the stack */
        if ((*nstack) < 1) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "REVOLVE expects a Body on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, revolve);
        } else {
            ibodyl = stack[--(*nstack)];
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* check that ibodyl is a SHEET Body */
        if (MODL->body[ibodyl].botype != OCSM_SHEET_BODY) {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                        "REVOLVE is expecting a SHEET Body");
            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, revolve);
        }

        /* check that a valid angle is given */
        if (args[7].val[0] < -360.0 || args[7].val[0] > 360) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "angle (%f) is not between -360 and +360", args[7].val[0]);
            SET_STATUS(OCSM_ILLEGAL_VALUE, revolve);
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_REVOLVE, ibodyl, -1,
                         args, hasdots, OCSM_SOLID_BODY, &ibody);
        CHECK_STATUS(newBody);

        status = EG_getTopology(MODL->body[ibodyl].ebody, &eref, &oclass, &mtype,
                                data, &nchild, &echilds, &senses);
        CHECK_STATUS(EG_getTopology);

        if        (oclass == BODY && mtype == FACEBODY) {
            status = EG_getTopology(MODL->body[ibodyl].ebody, &eref, &oclass, &mtype,
                                    data, &nchild, &echilds, &senses);
            CHECK_STATUS(EG_getTopology);

            ebodyl = echilds[0];

        } else if (oclass == BODY && mtype == SHEETBODY) {
            if (nchild != 1) {
                signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                            "found %d Shells in SHEETBODY but was expecting 1", nchild);
                SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, revolve);
            }

            etemp = echilds[0];
            status = EG_getTopology(etemp, &eref, &oclass, &mtype,
                                    data, &nchild, &echilds, &senses);
            CHECK_STATUS(EG_getTopology);

            if (nchild != 1) {
                signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                            "found %d Faces in SHEETBODY but was expecting 1", nchild);
                SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, revolve);
            }

            ebodyl = echilds[0];
        } else {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                        "ebody is neither FaceBody nor SheetBody\n");
            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, revolve);
        }

        /* check if projection of axis of revolution does intersect the Face */
        status = EG_getTopology(ebodyl, &eref, &oclass, &mtype,
                                data, &nchild, &echilds, &senses);
        CHECK_STATUS(EG_getTopology);

        status = EG_getGeometry(eref, &oclass, &mtype, &etemp, &pinfo, &prv);
        CHECK_STATUS(EG_getGeometry);

        if (oclass != SURFACE || mtype != PLANE) {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                        "expecting PLANE, but found oclass=%d, mtype=%d", oclass, mtype);
            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, revolve);
        } else if (fabs(prv[3]) < EPS06 && fabs(prv[6]) < EPS06) {
            status = EG_getBoundingBox(ebodyl, bbox);
            CHECK_STATUS(EG_getBoundingBox);

            if        (args[2].val[0]-bbox[1] <= +EPS06 && args[5].val[0] <= 0) {
            } else if (args[2].val[0]-bbox[4] >= -EPS06 && args[5].val[0] >= 0) {
            } else if (args[3].val[0]-bbox[2] <= +EPS06 && args[6].val[0] <= 0) {
            } else if (args[3].val[0]-bbox[5] >= -EPS06 && args[6].val[0] >= 0) {
            } else {
                SPRINT0(0, "WARNING:: axis may intersect Face and result may be invalid");
                SPRINT2(0, "          bbox: ymin=%12.5f   ymax=%12.5f", bbox[1],        bbox[4]       );
                SPRINT2(0, "          bbox: zmin=%12.5f   zmax=%12.5f", bbox[2],        bbox[5]       );
                SPRINT2(0, "          base: ybeg=%12.5f   zbeg=%12.5f", args[2].val[0], args[3].val[0]);
                SPRINT2(0, "          dirn: dy  =%12.5f   dz  =%12.5f", args[5].val[0], args[6].val[0]);
            }
        } else if (fabs(prv[4]) < EPS06 && fabs(prv[7]) < EPS06) {
            status = EG_getBoundingBox(ebodyl, bbox);
            CHECK_STATUS(EG_getBoundingBox);

            if        (args[3].val[0]-bbox[2] <= +EPS06 && args[6].val[0] <= 0) {
            } else if (args[3].val[0]-bbox[5] >= -EPS06 && args[6].val[0] >= 0) {
            } else if (args[1].val[0]-bbox[0] <= +EPS06 && args[4].val[0] <= 0) {
            } else if (args[1].val[0]-bbox[3] >= -EPS06 && args[4].val[0] >= 0) {
            } else {
                SPRINT0(0, "WARNING:: axis may intersect Face and result may be invalid");
                SPRINT2(0, "          bbox: zmin=%12.5f   zmax=%12.5f", bbox[2],        bbox[5]       );
                SPRINT2(0, "          bbox: xmin=%12.5f   xmax=%12.5f", bbox[0],        bbox[3]       );
                SPRINT2(0, "          base: zbeg=%12.5f   ybeg=%12.5f", args[3].val[0], args[1].val[0]);
                SPRINT2(0, "          dirn: dz  =%12.5f   dy  =%12.5f", args[6].val[0], args[4].val[0]);
            }
        } else if (fabs(prv[5]) < EPS06 && fabs(prv[8]) < EPS06) {
            status = EG_getBoundingBox(ebodyl, bbox);
            CHECK_STATUS(EG_getBoundingBox);

            if        (args[1].val[0]-bbox[0] <= +EPS06 && args[4].val[0] <= 0) {
            } else if (args[1].val[0]-bbox[3] >= -EPS06 && args[4].val[0] >= 0) {
            } else if (args[2].val[0]-bbox[1] <= +EPS06 && args[5].val[0] <= 0) {
            } else if (args[2].val[0]-bbox[4] >= -EPS06 && args[5].val[0] >= 0) {
            } else {
                SPRINT0(0, "WARNING:: axis may intersect Face and result may be invalid");
                SPRINT2(0, "          bbox: xmin=%12.5f   xmax=%12.5f", bbox[0],        bbox[3]       );
                SPRINT2(0, "          bbox: ymin=%12.5f   ymax=%12.5f", bbox[1],        bbox[4]       );
                SPRINT2(0, "          base: xbeg=%12.5f   ybeg=%12.5f", args[1].val[0], args[2].val[0]);
                SPRINT2(0, "          dirn: dx  =%12.5f   dy  =%12.5f", args[4].val[0], args[5].val[0]);
            }
        } else {
            SPRINT0(0, "WARNING:: Face is not aligned with coordinate axes and thus cannot be checked");
            SPRINT3(0, "Point: %12.5f %12.5f %12.5f", prv[0], prv[1], prv[2]);
            SPRINT3(0, "Xaxis: %12.5f %12.5f %12.5f", prv[3], prv[4], prv[5]);
            SPRINT3(0, "Yaxis: %12.5f %12.5f %12.5f", prv[6], prv[7], prv[8]);
        }

        if (args[7].val[0] > 0) {
            if (outLevel >= 3) {
                SPRINT0(3, "before EG_rotate: ebodyl");
                printEgo(ebodyl);
            }

            vals[0] = args[1].val[0];
            vals[1] = args[2].val[0];
            vals[2] = args[3].val[0];
            vals[3] = args[4].val[0];
            vals[4] = args[5].val[0];
            vals[5] = args[6].val[0];

            status = EG_rotate(ebodyl, args[7].val[0], vals, &ebody);
            CHECK_STATUS(EG_rotate);
        } else {
            if (outLevel >= 3) {
                SPRINT0(3, "before EG_rotate: ebodyl");
                printEgo(ebodyl);
            }

            vals[0] = args[1].val[0];
            vals[1] = args[2].val[0];
            vals[2] = args[3].val[0];
            vals[3] = args[4].val[0];
            vals[4] = args[5].val[0];
            vals[5] = args[6].val[0];

            status = EG_rotate(ebodyl, fabs(args[7].val[0]), vals, &etemp);
            CHECK_STATUS(EG_rotate);

            if        (fabs(args[5].val[0]) < EPS06 && fabs(args[6].val[0]) < EPS06) {
                cosx = cos(args[7].val[0] * PIo180);
                sinx = sin(args[7].val[0] * PIo180);
                dy   =     args[2].val[0];
                dz   =     args[3].val[0];

                matrix[ 0] = 1; matrix[ 1] = 0;    matrix[ 2] = 0;    matrix[ 3] = 0;
                matrix[ 4] = 0; matrix[ 5] = cosx; matrix[ 6] =-sinx; matrix[ 7] = dy - dy * cosx + dz * sinx;
                matrix[ 8] = 0; matrix[ 9] = sinx; matrix[10] = cosx; matrix[11] = dz - dy * sinx - dz * cosx;
            } else if (fabs(args[4].val[0]) < EPS06 && fabs(args[6].val[0]) < EPS06) {
                cosy = cos(args[7].val[0] * PIo180);
                siny = sin(args[7].val[0] * PIo180);
                dz   =     args[3].val[0];
                dx   =     args[1].val[0];

                matrix[ 0] = cosy; matrix[ 1] = 0; matrix[ 2] = siny; matrix[ 3] = dx - dz * siny - dx * cosy;
                matrix[ 4] = 0;    matrix[ 5] = 1; matrix[ 6] = 0;    matrix[ 7] = 0;
                matrix[ 8] =-siny; matrix[ 9] = 0; matrix[10] = cosy; matrix[11] = dz - dz * cosy + dx * siny;
            } else if (fabs(args[4].val[0]) < EPS06 && fabs(args[5].val[0]) < EPS06) {
                cosz = cos(args[7].val[0] * PIo180);
                sinz = sin(args[7].val[0] * PIo180);
                dx   =     args[1].val[0];
                dy   =     args[2].val[0];

                matrix[ 0] = cosz; matrix[ 1] =-sinz; matrix[ 2] = 0; matrix[ 3] = dx - dx * cosz + dy * sinz;
                matrix[ 4] = sinz; matrix[ 5] = cosz; matrix[ 6] = 0; matrix[ 7] = dy - dx * sinz - dy * cosz;
                matrix[ 8] = 0;    matrix[ 9] = 0;    matrix[10] = 1; matrix[11] = 0;
            } else {
                signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                            "not aligned with axis (%f, %f, %f)", args[4].val[0], args[5].val[0], args[6].val[0]);
                SET_STATUS(OCSM_ILLEGAL_VALUE, revolve);
            }

            status = EG_makeTransform(MODL->context, matrix, &exform);
            CHECK_STATUS(EG_makeTransform);

            status = EG_copyObject(etemp, exform, &ebody);
            CHECK_STATUS(EG_copyObject);

            status = EG_deleteObject(exform);
            CHECK_STATUS(EG_deleteObject);
        }

        MODL->body[ibody].ebody = ebody;

        /* mark the new Faces with the current Branch */
        status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        for (iface = 1; iface <= nface; iface++) {
            if (fabs(args[7].val[0]-360) < EPS06) {
                iford1 = iface + 2;
            } else if (iface == nface-1) {
                iford1 = 1;
            } else if (iface == nface  ) {
                iford1 = 2;
            } else {
                iford1 = iface + 2;
            }
            status = setFaceAttribute(MODL, ibody, iface, 0, iford1, npatn, patn);
            CHECK_STATUS(setFaceAttribute);
        }

        EG_free(efaces);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "sweep" */
    } else if (type == OCSM_SWEEP) {
        SPRINT1(1, "    executing [%4d] sweep:",
                ibrch);

        /* pop two Sketches from the stack */
        if ((*nstack) < 2) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "SWEEP expects a Body on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, sweep);
        } else {
            ibodyr = stack[--(*nstack)];
            ibodyl = stack[--(*nstack)];
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* check that ibodyl is a Sketch */
        if (MODL->body[ibodyl].botype != OCSM_SHEET_BODY &&
            MODL->body[ibodyl].botype != OCSM_WIRE_BODY    ) {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                        "SWEEP expects a SHEET or WIRE Body on the stack");
            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, sweep);
        }

        /* check that ibodyr is a Sketch */
        if (MODL->body[ibodyr].botype != OCSM_WIRE_BODY) {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                        "SWEEP expetcs a WIRE Body on the stack");
            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, sweep);
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_SWEEP, ibodyl, -1,
                         args, hasdots, OCSM_SOLID_BODY, &ibody);
        CHECK_STATUS(newBody);

        ebodyl = MODL->body[ibodyl].ebody;
        ebodyr = MODL->body[ibodyr].ebody;

        /* arbitrarily set the mode to 0 */
        mode = 0;

        if (outLevel >= 3) {
            SPRINT0(3, "before EG_sweep: ebodyl");
            printEgo(ebodyl);
            SPRINT0(3, "before EG_sweep: ebodyr");
            printEgo(ebodyr);
        }
        status = EG_sweep(ebodyl, ebodyr, mode, &ebody);
        CHECK_STATUS(EG_sweep);

        MODL->body[ibody].ebody = ebody;

        /* mark the new Faces with the current Branch */
        status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        for (iface = 1; iface <= nface; iface++) {
            if        (iface == 1    ) {
                iford1 = 1;
            } else if (iface == nface) {
                iford1 = 2;
            } else {
                iford1 = iface + 2;
            }
            status = setFaceAttribute(MODL, ibody, iface, 0, iford1, npatn, patn);
            CHECK_STATUS(setFaceAttribute);
        }

        EG_free(efaces);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   buildPrimitive - implement OCSM_PRIMITIVEs for ocsmBuild           *
 *                                                                      *
 ************************************************************************
 */

static int
buildPrimitive(modl_T *modl,
               int    ibrch,
               varg_T args[],
               int    *nstack,
               int    stack[],
               int    npatn,
               patn_T patn[])
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int        type, hasdots, iface, nface, ibody, jbody, kbody, jbrch, irc;
    int        istor, jstor, ibodyl;
    double     toler, value, dot;

    int         oclass, mtype, nchild, *sense, iford1, iarg, ival;
    int         udp_num, *udp_types, *udp_idef, udp_nmesh;
    int         valInt;
    int         ipmtr, ij, old_outLevel, periodic, *pinfo=NULL;
    double      data[18], box[6], range[4], *prv=NULL, valDouble, vals[8];
    double      *udp_ddef;
    char        **udp_names, *udp_errStr, primtype[MAX_EXPR_LEN];
    char        dumpfile[MAX_EXPR_LEN];
    ego         ebody, *efaces, topRef, prev, next, esurf, *echild, eref;
    char        atPmtr[128];

    ROUTINE(buildPrimitive);
    DPRINT2("%s(ibrch=%d) {",
            routine, ibrch);

    /* --------------------------------------------------------------- */

    type    = MODL->brch[ibrch].type;
    hasdots = 0;

    /* execute: "box xbase ybase zbase dx dy dz" */
    if (type == OCSM_BOX) {
        SPRINT7(1, "    executing [%4d] box:        %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0], args[4].val[0],
                       args[5].val[0], args[6].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0 || args[4].dot[0] != 0 ||
            args[5].dot[0] != 0 || args[6].dot[0] != 0                                   ) {
            hasdots = 1;
            SPRINT6(1, "                                 %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0], args[4].dot[0],
                       args[5].dot[0], args[6].dot[0]);
        }

        /* check to be sure that dx, dy, and dz are non-zero */
        if        (fabs(args[4].val[0]) < EPS06) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "dx=%f cannot be zero", args[4].val[0]);
            status = OCSM_ILLEGAL_VALUE;
            goto cleanup;
        } else if (fabs(args[5].val[0]) < EPS06) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "dy=%f cannot be zero", args[5].val[0]);
            status = OCSM_ILLEGAL_VALUE;
            goto cleanup;
        } else if (fabs(args[6].val[0]) < EPS06) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "dz=%f cannot be zero", args[6].val[0]);
            status = OCSM_ILLEGAL_VALUE;
            goto cleanup;
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_BOX, -1, -1,
                         args, hasdots, OCSM_SOLID_BODY, &ibody);
        CHECK_STATUS(newBody);

        vals[0] = args[1].val[0];
        vals[1] = args[2].val[0];
        vals[2] = args[3].val[0];
        vals[3] = args[4].val[0];
        vals[4] = args[5].val[0];
        vals[5] = args[6].val[0];

        status = EG_makeSolidBody(MODL->context, BOX, vals, &ebody);
        CHECK_STATUS(EG_makeSolidBody);

        MODL->body[ibody].ebody = ebody;

        /* mark the new Faces with the current Branch */
        status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        iford1 = 0;
        for (iface = 1; iface <= nface; iface++) {
            if        (faceContains(efaces[iface-1], args[1].val[0],
                                                     args[2].val[0]+args[5].val[0]/2,
                                                     args[3].val[0]+args[6].val[0]/2)) {
                iford1 = 1;
            } else if (faceContains(efaces[iface-1], args[1].val[0]+args[4].val[0],
                                                     args[2].val[0]+args[5].val[0]/2,
                                                     args[3].val[0]+args[6].val[0]/2)) {
                iford1 = 2;
            } else if (faceContains(efaces[iface-1], args[1].val[0]+args[4].val[0]/2,
                                                     args[2].val[0],
                                                     args[3].val[0]+args[6].val[0]/2)) {
                iford1 = 3;
            } else if (faceContains(efaces[iface-1], args[1].val[0]+args[4].val[0]/2,
                                                     args[2].val[0]+args[5].val[0],
                                                     args[3].val[0]+args[6].val[0]/2)) {
                iford1 = 4;
            } else if (faceContains(efaces[iface-1], args[1].val[0]+args[4].val[0]/2,
                                                     args[2].val[0]+args[5].val[0]/2,
                                                     args[3].val[0])          ) {
                iford1 = 5;
            } else if (faceContains(efaces[iface-1], args[1].val[0]+args[4].val[0]/2,
                                                     args[2].val[0]+args[5].val[0]/2,
                                                     args[3].val[0]+args[6].val[0])  ) {
                iford1 = 6;
            }

            status = setFaceAttribute(MODL, ibody, iface, 0, iford1, npatn, patn);
            CHECK_STATUS(setFaceAttribute);
        }

        EG_free(efaces);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "sphere xcent ycent zcent radius" */
    } else if (type == OCSM_SPHERE) {
        SPRINT5(1, "    executing [%4d] sphere:     %11.5f %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0], args[4].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0 || args[4].dot[0] != 0) {
            hasdots = 1;

            SPRINT4(1, "                                 %11.5f %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0], args[4].dot[0]);
        }

        /* check to be sure that the radius is non-zero */
        if        (fabs(args[4].val[0]) < EPS06) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "radius=%f cannot be zero", args[4].val[0]);
            status = OCSM_ILLEGAL_VALUE;
            goto cleanup;
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_SPHERE, -1, -1,
                         args, hasdots, OCSM_SOLID_BODY, &ibody);
        CHECK_STATUS(newBody);

        vals[0] = args[1].val[0];
        vals[1] = args[2].val[0];
        vals[2] = args[3].val[0];
        vals[3] = args[4].val[0];

        status = EG_makeSolidBody(MODL->context, SPHERE, vals, &ebody);
        CHECK_STATUS(EG_makeSolidBody);

        MODL->body[ibody].ebody = ebody;

        /* mark the new Faces with the current Branch */
        status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        iford1 = 0;
        for (iface = 1; iface <= nface; iface++) {
            if        (faceContains(efaces[iface-1], args[1].val[0],
                                                     args[2].val[0]-args[4].val[0]/2,
                                                     args[3].val[0]          )) {
                iford1 = 1;
            } else if (faceContains(efaces[iface-1], args[1].val[0],
                                                     args[2].val[0]+args[4].val[0]/2,
                                                     args[3].val[0]          )) {
                iford1 = 2;
            }

            status = setFaceAttribute(MODL, ibody, iface, 0, iford1, npatn, patn);
            CHECK_STATUS(setFaceAttribute);
        }

        EG_free(efaces);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "cone xvrtx yvrtx zvrtx xbase ybase zbase radius" */
    } else if (type == OCSM_CONE) {
        SPRINT8(1, "    executing [%4d] cone:       %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0], args[4].val[0],
                       args[5].val[0], args[6].val[0], args[7].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0 || args[4].dot[0] != 0 ||
            args[5].dot[0] != 0 || args[6].dot[0] != 0 || args[7].dot[0] != 0                   ) {
            hasdots = 1;

            SPRINT7(1, "                                 %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0], args[4].dot[0],
                       args[5].dot[0], args[6].dot[0], args[7].dot[0]);
        }

        /* check to be sure that vertex and base are not the same */
        if (fabs(args[4].val[0]-args[1].val[0]) < EPS06 &&
            fabs(args[5].val[0]-args[2].val[0]) < EPS06 &&
            fabs(args[6].val[0]-args[3].val[0]) < EPS06   ) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "vertex and base must be different");
            status = OCSM_ILLEGAL_VALUE;
            goto cleanup;
        }

        /* check to be sure that the radius is non-zero */
        if        (fabs(args[7].val[0]) < EPS06) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "radius=%f cannot be zero", args[7].val[0]);
            status = OCSM_ILLEGAL_VALUE;
            goto cleanup;
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_CONE, -1, -1,
                         args, hasdots, OCSM_SOLID_BODY, &ibody);
        CHECK_STATUS(newBody);

        vals[0] = args[1].val[0];
        vals[1] = args[2].val[0];
        vals[2] = args[3].val[0];
        vals[3] = args[4].val[0];
        vals[4] = args[5].val[0];
        vals[5] = args[6].val[0];
        vals[6] = args[7].val[0];

        status = EG_makeSolidBody(MODL->context, CONE, vals, &ebody);
        CHECK_STATUS(EG_makeSolidBody);

        MODL->body[ibody].ebody = ebody;

        /* mark the new Faces with the current Branch */
        status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        for (iface = 1; iface <= nface; iface++) {
            iford1 = 0;

            status = EG_getTopology(efaces[iface-1], &esurf, &oclass, &mtype,
                                    data, &nchild, &echild, &sense);
            CHECK_STATUS(EG_getTopology);

            status = EG_getGeometry(esurf, &oclass, &mtype, &eref, &pinfo, &prv);
            CHECK_STATUS(EG_getGeometry);

            if (pinfo != NULL) {
                EG_free(pinfo);
                pinfo = NULL;
            }
            if (prv   != NULL) {
                EG_free(prv  );
                prv   = NULL;
            }

            if (mtype == PLANE) {
                status = EG_getBoundingBox(efaces[iface-1], box);
                CHECK_STATUS(EG_getBoundingBox);

                if        (fabs((box[0]+box[3])/2-args[4].val[0]) < EPS06 &&
                           fabs((box[1]+box[4])/2-args[5].val[0]) < EPS06 &&
                           fabs((box[2]+box[5])/2-args[6].val[0]) < EPS06   ) {
                    iford1 = 2;
                }
            } else {
                status = EG_getRange(efaces[iface-1], range, &periodic);
                CHECK_STATUS(EG_getRange);

                if        (fabs(range[0]      ) < EPS06 &&
                           fabs(range[1]-PI   ) < EPS06   ) {
                    iford1 = 3;
                } else if (fabs(range[0]-PI   ) < EPS06 &&
                           fabs(range[1]-TWOPI) < EPS06   ) {
                    iford1 = 4;
                }
            }

            if (iford1 > 0) {
                status = setFaceAttribute(MODL, ibody, iface, 0, iford1, npatn, patn);
                CHECK_STATUS(setFaceAttribute);
            } else {
                signalError(MODL, OCSM_INTERNAL_ERROR, ibrch,
                            "cannot assign iford to iafce=%d", iface);
                SET_STATUS(OCSM_INTERNAL_ERROR, cone);
            }
        }

        EG_free(efaces);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "cylinder xbeg ybeg zbeg xend yend zend radius" */
    } else if (type == OCSM_CYLINDER) {
        SPRINT8(1, "    executing [%4d] cylinder:   %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0], args[4].val[0],
                       args[5].val[0], args[6].val[0], args[7].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0 || args[4].dot[0] != 0 ||
            args[5].dot[0] != 0 || args[6].dot[0] != 0 || args[7].dot[0] != 0                   ) {
            hasdots = 1;

            SPRINT7(1, "                                 %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0], args[4].dot[0],
                       args[5].dot[0], args[6].dot[0], args[7].dot[0]);
        }

        /* check to be sure that vertex and base are not the same */
        if (fabs(args[4].val[0]-args[1].val[0]) < EPS06 &&
            fabs(args[5].val[0]-args[2].val[0]) < EPS06 &&
            fabs(args[6].val[0]-args[3].val[0]) < EPS06   ) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "vertex and base must be different");
            status = OCSM_ILLEGAL_VALUE;
            goto cleanup;
        }

        /* check to be sure that the radius is non-zero */
        if        (fabs(args[7].val[0]) < EPS06) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "radius=%f cannot be zero", args[7].val[0]);
            status = OCSM_ILLEGAL_VALUE;
            goto cleanup;
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_CYLINDER, -1, -1,
                         args, hasdots, OCSM_SOLID_BODY, &ibody);
        CHECK_STATUS(newBody);

        vals[0] = args[1].val[0];
        vals[1] = args[2].val[0];
        vals[2] = args[3].val[0];
        vals[3] = args[4].val[0];
        vals[4] = args[5].val[0];
        vals[5] = args[6].val[0];
        vals[6] = args[7].val[0];

        status = EG_makeSolidBody(MODL->context, CYLINDER, vals, &ebody);
        CHECK_STATUS(EG_makeSolidBody);

        MODL->body[ibody].ebody = ebody;

        /* mark the new Faces with the current Branch */
        status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        for (iface = 1; iface <= nface; iface++) {
            iford1 = 0;

            status = EG_getTopology(efaces[iface-1], &esurf, &oclass, &mtype,
                                    data, &nchild, &echild, &sense);
            CHECK_STATUS(EG_getTopology);

            status = EG_getGeometry(esurf, &oclass, &mtype, &eref, &pinfo, &prv);
            CHECK_STATUS(EG_getGeometry);

            if (pinfo != NULL) {
                EG_free(pinfo);
                pinfo = NULL;
            }
            if (prv   != NULL) {
                EG_free(prv  );
                prv   = NULL;
            }

            if (mtype == PLANE) {
                status = EG_getBoundingBox(efaces[iface-1], box);
                CHECK_STATUS(EG_getBoundingBox);

                if        (fabs((box[0]+box[3])/2-args[1].val[0]) < EPS06 &&
                           fabs((box[1]+box[4])/2-args[2].val[0]) < EPS06 &&
                           fabs((box[2]+box[5])/2-args[3].val[0]) < EPS06   ) {
                    iford1 = 1;
                } else if (fabs((box[0]+box[3])/2-args[4].val[0]) < EPS06 &&
                           fabs((box[1]+box[4])/2-args[5].val[0]) < EPS06 &&
                           fabs((box[2]+box[5])/2-args[6].val[0]) < EPS06   ) {
                    iford1 = 2;
                }
            } else {
                status = EG_getRange(efaces[iface-1], range, &periodic);
                CHECK_STATUS(EG_getRange);

                if        (fabs(range[0]      ) < EPS06 &&
                           fabs(range[1]-PI   ) < EPS06   ) {
                    iford1 = 3;
                } else if (fabs(range[0]-PI   ) < EPS06 &&
                           fabs(range[1]-TWOPI) < EPS06   ) {
                    iford1 = 4;
                }
            }

            if (iford1 > 0) {
                status = setFaceAttribute(MODL, ibody, iface, 0, iford1, npatn, patn);
                CHECK_STATUS(setFaceAttribute);
            } else {
                signalError(MODL, OCSM_INTERNAL_ERROR, ibrch,
                            "cannot assign iford for iface=%d", iface);
                SET_STATUS(OCSM_INTERNAL_ERROR, cylinder);
            }
        }

        EG_free(efaces);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "torus xcent ycent zcent dxaxis dyaxis dzaxis majorRad minorRad" */
    } else if (type == OCSM_TORUS) {
        SPRINT9(1, "    executing [%4d] torus:      %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0], args[4].val[0],
                       args[5].val[0], args[6].val[0], args[7].val[0], args[8].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0 || args[4].dot[0] != 0 ||
            args[5].dot[0] != 0 || args[6].dot[0] != 0 || args[7].dot[0] != 0 || args[8].dot[0] != 0   ) {
            hasdots = 1;

            SPRINT8(1, "                                 %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0], args[4].dot[0],
                       args[5].dot[0], args[6].dot[0], args[7].dot[0], args[8].dot[0]);
        }

        /* check to be sure that the axis is not zero */
        if (fabs(args[4].val[0]) < EPS06 &&
            fabs(args[5].val[0]) < EPS06 &&
            fabs(args[6].val[0]) < EPS06   ) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "axis length must be non-zero");
            status = OCSM_ILLEGAL_VALUE;
            goto cleanup;
        }

        /* check to be sure that the radii are non-zero */
        if        (fabs(args[7].val[0]) < EPS06) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "majorRad=%f cannot be zero", args[7].val[0]);
            status = OCSM_ILLEGAL_VALUE;
            goto cleanup;
        } else if (fabs(args[8].val[0]) < EPS06) {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "minorRad=%f cannot be zero", args[8].val[0]);
            status = OCSM_ILLEGAL_VALUE;
            goto cleanup;
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_TORUS, -1, -1,
                         args, hasdots, OCSM_SOLID_BODY, &ibody);
        CHECK_STATUS(newBody);

        vals[0] = args[1].val[0];
        vals[1] = args[2].val[0];
        vals[2] = args[3].val[0];
        vals[3] = args[4].val[0];
        vals[4] = args[5].val[0];
        vals[5] = args[6].val[0];
        vals[6] = args[7].val[0];
        vals[7] = args[8].val[0];

        status = EG_makeSolidBody(MODL->context, TORUS, vals, &ebody);
        CHECK_STATUS(EG_makeSolidBody);

        MODL->body[ibody].ebody = ebody;

        /* mark the new Faces with the current Branch */
        status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        for (iface = 1; iface <= nface; iface++) {
            status = setFaceAttribute(MODL, ibody, iface, 0, iface, npatn, patn);
            CHECK_STATUS(setFaceAttribute);
        }

        EG_free(efaces);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "import $filename bodynumber=1" */
    } else if (type == OCSM_IMPORT) {
        SPRINT3(1, "    executing [%4d] import:     %s  %11.5f",
                ibrch, args[1].str, args[2].val[0]);

        /* IMPORTs cannot be recycled, so free up all Bodys starting at ibody */
        ibody = MODL->nbody + 1;
        for (jbody = ibody; jbody <= MODL->recycle; jbody++) {
            for (iarg = 1; iarg < 10; iarg++) {
                FREE(MODL->body[jbody].arg[iarg].val);
                FREE(MODL->body[jbody].arg[iarg].dot);
            }

            FREE(MODL->body[jbody].node);
            FREE(MODL->body[jbody].edge);
            FREE(MODL->body[jbody].face);

            if (MODL->body[jbody].etess != NULL) {
                status = EG_deleteObject(MODL->body[jbody].etess);
                CHECK_STATUS(EG_deleteObject);

                MODL->body[jbody].etess = NULL;
            }

            if (MODL->body[jbody].ebody != NULL) {
                status = EG_deleteObject(MODL->body[jbody].ebody);
                CHECK_STATUS(EG_deleteObject);

                MODL->body[jbody].ebody = NULL;
            }
        }

        MODL->recycle = ibody;

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_IMPORT, -1, -1,
                         args, hasdots, OCSM_SOLID_BODY, &ibody);
        CHECK_STATUS(newBody);

        /* load and execute the user-defined primitive */
        STRNCPY(primtype, "import", MAX_EXPR_LEN);

        status = udp_initialize(primtype, &udp_num, &udp_names, &udp_types, &udp_idef, &udp_ddef);
        if (status == EGADS_NOLOAD) {
            status = EGADS_SUCCESS;
        }
        CHECK_STATUS(udp_initialize);

        status = udp_clrArguments(primtype);
        CHECK_STATUS(udp_clrArguments);

        /* strip the dollarsign off the filename */
        STRNCPY(dumpfile, args[1].str, MAX_EXPR_LEN);

        status = udp_setArgument(primtype, "filename", dumpfile, (int)(strlen(dumpfile)+1));
        CHECK_STATUS(udp_setArgument);

        /* set the Body number (in the import) */
        status = udp_setArgument(primtype, "bodynumber", &(args[2].val[0]), 1);
        CHECK_STATUS(udp_setArgument);

        /* execute udpImport */
        status = udp_executePrim(primtype, MODL->context, &ebody, &udp_nmesh, &udp_errStr);
        if (status < EGADS_SUCCESS && udp_errStr != NULL) {
            SPRINT1(0, "WARNING:: upd returned message: \"%s\"", udp_errStr);
            EG_free(udp_errStr);
        }
        CHECK_STATUS(udp_executePrim);

        /* check the type of ebody */
        status = EG_getInfo(ebody, &oclass, &mtype, &topRef, &prev, &next);
        CHECK_STATUS(EG_getInfo);

        if        (oclass == BODY && mtype == SOLIDBODY) {
//                                     OCSM_SOLID_BODY
        } else if (oclass == BODY && mtype == FACEBODY) {
            MODL->body[ibody].botype = OCSM_SHEET_BODY;
        } else if (oclass == BODY && mtype == SHEETBODY) {
            MODL->body[ibody].botype = OCSM_SHEET_BODY;
        } else if (oclass == BODY && mtype == WIREBODY) {
            MODL->body[ibody].botype = OCSM_WIRE_BODY;
        } else {
            signalError(MODL, OCSM_DID_NOT_CREATE_BODY, ibrch,
                        "IMPORT did not produce a SOLID, SHEET, or WIRE Body");
            SET_STATUS(OCSM_DID_NOT_CREATE_BODY, udp_executePrim);
        }

        /* check the tolerance and issue warning if a problem */
        status = getBodyTolerance(ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        if (toler > 2.0e-7) {
            SPRINT1(0, "WARNING:: toler = %12.4e for import", toler);
        }

        /* set @-parameters for any outputs from the udp */
        for (ij = 0; ij < udp_num; ij++) {
            if        (udp_types[ij] == -ATTRINT) {
                status = udp_getOutput(primtype, ebody, udp_names[ij], (void*)&valInt);
                CHECK_STATUS(udp_getOutput);

                snprintf(atPmtr, 128, "@@%s", udp_names[ij]);

                status = ocsmFindPmtr(MODL, atPmtr, OCSM_INTERNAL, 1, 1, &ipmtr);
                CHECK_STATUS(ocsmFindPmtr);

                valDouble = valInt;
                status = ocsmSetValuD(MODL, ipmtr, 1, 1, valDouble);
                CHECK_STATUS(ocsmSetValuD);
            } else if (udp_types[ij] == -ATTRREAL) {
                status = udp_getOutput(primtype, ebody, udp_names[ij], (void*)&valDouble);
                CHECK_STATUS(udp_getOutput);

                snprintf(atPmtr, 128, "@@%s", udp_names[ij]);

                status = ocsmFindPmtr(MODL, atPmtr, OCSM_INTERNAL, 1, 1, &ipmtr);
                CHECK_STATUS(ocsmFindPmtr);

                status = ocsmSetValuD(MODL, ipmtr, 1, 1, valDouble);
                CHECK_STATUS(ocsmSetValuD);
            }
        }

        MODL->body[ibody].ebody = ebody;

        /* mark the new Faces with the current Branch */
        status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        for (iface = 1; iface <= nface; iface++) {
            old_outLevel = EG_setOutLevel(MODL->context, 0);
            status = EG_attributeDel(efaces[iface-1], "body");
            (void) EG_setOutLevel(MODL->context, old_outLevel);
            if (status != EGADS_NOTFOUND) {
                CHECK_STATUS(EG_attributeDel);
            }

            old_outLevel = EG_setOutLevel(MODL->context, 0);
            status = EG_attributeDel(efaces[iface-1], "brch");
            (void) EG_setOutLevel(MODL->context, old_outLevel);
            if (status != EGADS_NOTFOUND) {
                CHECK_STATUS(EG_attributeDel);
            }

            status = setFaceAttribute(MODL, ibody, iface, 0, iface, npatn, patn);
            CHECK_STATUS(setFaceAttribute);
        }

        EG_free(efaces);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body back onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "udprim $primtype $argName1 argValue1 $argName2 argValue2 $argName3 argValue3 $argName4 argValue4" */
    } else if (type == OCSM_UDPRIM) {

        /* primtype starts with a letter, so use a compiled udprim named 'primtype.so' */
        if ((args[1].str[0] >= 'a' && args[1].str[0] <= 'z') ||
            (args[1].str[0] >= 'A' && args[1].str[0] <= 'Z')   ) {
            SPRINT2x(1, "    executing [%4d] udprim:     %s",
                    ibrch, args[1].str);

            hasdots = 0;
            for (iarg = 3; iarg <= MODL->brch[ibrch].narg; iarg+=2) {
                if (args[iarg].nval == 0) {
                    SPRINT2x(1, " %s %s",
                             args[iarg-1].str, args[iarg].str);
                } else {
                    SPRINT2x(1, " %s %11.5f",
                             args[iarg-1].str, args[iarg].val[0]);
                    for (ival = 0; ival < args[iarg].nval; ival++) {
                        if (args[iarg].val[ival] != 0) hasdots = 1;
                    }
                }
            }
            SPRINT0(1, " (compiled code)");

            /* go back to first associated udparg statement whose primtype matches the
               udprim's primtype */
            jbody = MODL->nbody;
            while (jbody > 0 && MODL->body[jbody].brtype == OCSM_UDPARG &&
                   strcmp(MODL->body[jbody].arg[1].str, args[1].str) == 0 ) {

                if (MODL->body[jbody].hasdots != 0) hasdots = MODL->body[jbody].hasdots;

                jbody--;
            }
            jbody++;

            /* recycle old Body if not dirty */
            status = recycleBody(MODL, ibrch, type, args, hasdots);
            CHECK_STATUS(recycleBody);

            if (status == 1) {
                stack[(*nstack)++] = MODL->nbody;
                status = 0;
                goto cleanup;
            }

            /* create the Body */
            status = newBody(MODL, ibrch, OCSM_UDPRIM, -1, -1,
                             args, hasdots, OCSM_SOLID_BODY, &ibody);
            CHECK_STATUS(newBody);

            /* load and execute the user-defined primitive */
            STRNCPY(primtype, args[1].str, MAX_EXPR_LEN);

            status = udp_initialize(primtype, &udp_num, &udp_names, &udp_types, &udp_idef, &udp_ddef);
            if (status == EGADS_NOLOAD) {
                status = EGADS_SUCCESS;
            }
            CHECK_STATUS(udp_initialize);

            status = udp_clrArguments(primtype);
            CHECK_STATUS(udp_clrArguments);

            /* go back to first associated udparg statement whose primtype matches the
               udprim's primtype */
            jbody = MODL->nbody - 1;
            while (jbody > 0 && MODL->body[jbody].brtype == OCSM_UDPARG &&
                   strcmp(MODL->body[jbody].arg[1].str, args[1].str) == 0 ) {

                jbody--;
            }
            jbody++;

            while (jbody <= MODL->nbody) {
                for (iarg = 3; iarg < 10; iarg+=2) {
                    if (MODL->body[jbody].arg[iarg].nval < 0) break;

                    if (MODL->body[jbody].arg[iarg].nval == 0) {
                        status = udp_setArgument(MODL->body[jbody].arg[1].str,
                                                 MODL->body[jbody].arg[iarg-1].str,
                                                 MODL->body[jbody].arg[iarg  ].str,
                                          strlen(MODL->body[jbody].arg[iarg  ].str));
                        CHECK_STATUS(udp_setArgument);
                    } else {
                        status = udp_setArgument(MODL->body[jbody].arg[1].str,
                                                 MODL->body[jbody].arg[iarg-1].str,
                                                 MODL->body[jbody].arg[iarg  ].val,
                                                 MODL->body[jbody].arg[iarg  ].nval);
                        CHECK_STATUS(udp_setArgument);
                    }
                }

                jbody++;
            }

            /* execute the primitive */
            status = udp_executePrim(primtype, MODL->context, &ebody, &udp_nmesh, &udp_errStr);
            if (status < EGADS_SUCCESS && udp_errStr != NULL) {
                SPRINT1(0, "WARNING:: upd returned message: \"%s\"", udp_errStr);
                EG_free(udp_errStr);
            }
            CHECK_STATUS(udp_executePrim);

            /* check the type of ebody */
            status = EG_getInfo(ebody, &oclass, &mtype, &topRef, &prev, &next);
            CHECK_STATUS(EG_getInfo);

            if        (oclass == BODY && mtype == SOLIDBODY) {
//                                         OCSM_SOLID_BODY
            } else if (oclass == BODY && mtype == FACEBODY) {
                MODL->body[ibody].botype = OCSM_SHEET_BODY;
            } else if (oclass == BODY && mtype == SHEETBODY) {
                MODL->body[ibody].botype = OCSM_SHEET_BODY;
            } else if (oclass == BODY && mtype == WIREBODY) {
                MODL->body[ibody].botype = OCSM_WIRE_BODY;
            } else {
                signalError(MODL, OCSM_DID_NOT_CREATE_BODY, ibrch,
                            "UDPRIM did not create a SOLID, SHEET, or WIRE Body");
                SET_STATUS(OCSM_DID_NOT_CREATE_BODY, udp_executePrim);
            }

            /* check the tolerance and issue warning if a problem */
            status = getBodyTolerance(ebody, &toler);
            CHECK_STATUS(getBodyTolerance);

            if (toler > 2.0e-7) {
                SPRINT1(0, "WARNING:: toler = %12.4e for udprim", toler);
            }

            /* set @@-parameters for any outputs from the udp */
            for (ij = 0; ij < udp_num; ij++) {
                if        (udp_types[ij] == -ATTRINT) {
                    status = udp_getOutput(primtype, ebody, udp_names[ij], (void*)&valInt);
                    CHECK_STATUS(udp_getOutput);

                    snprintf(atPmtr, 128, "@@%s", udp_names[ij]);

                    status = ocsmFindPmtr(MODL, atPmtr, OCSM_INTERNAL, 1, 1, &ipmtr);
                    CHECK_STATUS(ocsmFindPmtr);

                    valDouble = valInt;
                    status = ocsmSetValuD(MODL, ipmtr, 1, 1, valDouble);
                    CHECK_STATUS(ocsmSetValuD);
                } else if (udp_types[ij] == -ATTRREAL) {
                    status = udp_getOutput(primtype, ebody, udp_names[ij], (void*)&valDouble);
                    CHECK_STATUS(udp_getOutput);

                    snprintf(atPmtr, 128, "@@%s", udp_names[ij]);

                    status = ocsmFindPmtr(MODL, atPmtr, OCSM_INTERNAL, 1, 1, &ipmtr);
                    CHECK_STATUS(ocsmFindPmtr);

                    status = ocsmSetValuD(MODL, ipmtr, 1, 1, valDouble);
                    CHECK_STATUS(ocsmSetValuD);
                }
            }

            MODL->body[ibody].ebody = ebody;

            /* mark the new Faces with the current Branch */
            status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
            CHECK_STATUS(EG_getBodyTopos);

            for (iface = 1; iface <= nface; iface++) {
                old_outLevel = EG_setOutLevel(MODL->context, 0);
                status = EG_attributeDel(efaces[iface-1], "body");
                (void) EG_setOutLevel(MODL->context, old_outLevel);
                if (status != EGADS_NOTFOUND) {
                    CHECK_STATUS(EG_attributeDel);
                }

                old_outLevel = EG_setOutLevel(MODL->context, 0);
                status = EG_attributeDel(efaces[iface-1], "brch");
                (void) EG_setOutLevel(MODL->context, old_outLevel);
                if (status != EGADS_NOTFOUND) {
                    CHECK_STATUS(EG_attributeDel);
                }

                status = setFaceAttribute(MODL, ibody, iface, 0, iface, npatn, patn);
                CHECK_STATUS(setFaceAttribute);
            }

            EG_free(efaces);

            /* finish the Body */
            status = finishBody(MODL, ibody);
            CHECK_STATUS(finishBody);

            /* push the Body back onto the stack */
            stack[(*nstack)++] = ibody;

            status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
            CHECK_STATUS(getBodyTolerance);

            SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                    ibody, toler);

        /* if primtype begins with / or $/, then we have user-defined code */
        } else if ((args[1].str[0] == '/'                         ) ||
                   (args[1].str[0] == '$' && args[1].str[1] == '/')   ) {
            SPRINT2x(1, "    executing [%4d] udprim:     %s",
                    ibrch, args[1].str);

            for (iarg = 3; iarg <= MODL->brch[ibrch].narg; iarg+=2) {
                if (args[iarg].nval == 0) {
                    SPRINT2x(1, " %s %s",
                             args[iarg-1].str, args[iarg].str);
                } else {
                    SPRINT2x(1, " %s %11.5f",
                             args[iarg-1].str, args[iarg].val[0]);
                }
            }
            SPRINT0(1, " (scripted code)");

            /* increment the level */
            (MODL->level)++;

            /* find the first udparg statement associated with this udprim */
            jbody = MODL->nbody;
            while (jbody > 0 && MODL->body[jbody].brtype == OCSM_UDPARG &&
                   strcmp(MODL->body[jbody].arg[1].str, args[1].str) == 0 ) {
                jbody--;
            }
            kbody = jbody + 1;

            /* set up values for all Parameters mentioned in interface statements */
            jbrch = ibrch + 1;

            while (MODL->brch[jbrch].type == OCSM_INTERFACE) {
                if (strcmp(MODL->brch[jbrch].arg2, "$in" ) == 0 ||
                    strcmp(MODL->brch[jbrch].arg2, "$out") == 0   ) {

                    /* default size is 1 */
                    status = ocsmNewPmtr(MODL, &(MODL->brch[jbrch].arg1[1]), OCSM_INTERNAL, 1, 1);
                    CHECK_STATUS(ocsmNewPmtr);

                    /* set default value */
                    status = str2val(MODL->brch[jbrch].arg3, MODL, &value, &dot);
                    CHECK_STATUS(str2val);

                    status = ocsmSetValuD(MODL, MODL->npmtr, 1, 1, value);
                    CHECK_STATUS(ocsmSetValu);

                    status = ocsmSetVelD(MODL, MODL->npmtr, 1, 1, dot);
                    CHECK_STATUS(ocsmSetVelD);

                    /* transfer any values from a udparg statement */
                    for (jbody = kbody; jbody <= MODL->nbody; jbody++) {
                        for (iarg = 3; iarg < 10; iarg+=2) {
                            if (MODL->body[jbody].arg[iarg].nval < 0) continue;

                            if (strcmp(MODL->body[jbody].arg[iarg-1].str, &(MODL->brch[jbrch].arg1[1])) == 0) {

                                /* make Parameter bigger if multi-valued */
                                if (MODL->body[jbody].arg[iarg].nval > 1) {
                                    FREE(MODL->pmtr[MODL->npmtr].value);
                                    FREE(MODL->pmtr[MODL->npmtr].dot  );
                                    FREE(MODL->pmtr[MODL->npmtr].lbnd );
                                    FREE(MODL->pmtr[MODL->npmtr].ubnd );

                                    MALLOC(MODL->pmtr[MODL->npmtr].value, double, MODL->body[jbody].arg[iarg].nval);
                                    MALLOC(MODL->pmtr[MODL->npmtr].dot,   double, MODL->body[jbody].arg[iarg].nval);
                                    MALLOC(MODL->pmtr[MODL->npmtr].lbnd,  double, MODL->body[jbody].arg[iarg].nval);
                                    MALLOC(MODL->pmtr[MODL->npmtr].ubnd,  double, MODL->body[jbody].arg[iarg].nval);

                                    MODL->pmtr[MODL->npmtr].nrow = MODL->body[jbody].arg[iarg].nrow;
                                    MODL->pmtr[MODL->npmtr].ncol = MODL->body[jbody].arg[iarg].ncol;
                                }

                                for (ival = 0; ival < MODL->body[jbody].arg[iarg].nval; ival++) {
                                    MODL->pmtr[MODL->npmtr].value[ival] = MODL->body[jbody].arg[iarg].val[ival];
                                    MODL->pmtr[MODL->npmtr].dot[  ival] = MODL->body[jbody].arg[iarg].dot[ival];
                                    MODL->pmtr[MODL->npmtr].lbnd[ ival]  = 0;
                                    MODL->pmtr[MODL->npmtr].ubnd[ ival]  = 0;
                                }
                            }
                        }
                    }

                    /* transfer any values from the udprim statement */
                    for (iarg = 3; iarg <= MODL->brch[jbrch].narg; iarg+=2) {
                        if (strcmp(args[iarg-1].str, &(MODL->brch[jbrch].arg1[1])) == 0) {

                            /* make Parameter bigger if multi-valued */
                            if (args[iarg].nval > 1) {
                                FREE(MODL->pmtr[MODL->npmtr].value);
                                FREE(MODL->pmtr[MODL->npmtr].dot  );
                                FREE(MODL->pmtr[MODL->npmtr].lbnd );
                                FREE(MODL->pmtr[MODL->npmtr].ubnd );

                                MALLOC(MODL->pmtr[MODL->npmtr].value, double, args[iarg].nval);
                                MALLOC(MODL->pmtr[MODL->npmtr].dot,   double, args[iarg].nval);
                                MALLOC(MODL->pmtr[MODL->npmtr].lbnd,  double, args[iarg].nval);
                                MALLOC(MODL->pmtr[MODL->npmtr].ubnd,  double, args[iarg].nval);

                                MODL->pmtr[MODL->npmtr].nrow = args[iarg].nrow;
                                MODL->pmtr[MODL->npmtr].ncol = args[iarg].ncol;
                            }

                            for (ival = 0; ival < args[iarg].nval; ival++) {
                                MODL->pmtr[MODL->npmtr].value[ival] = args[iarg].val[ival];
                                MODL->pmtr[MODL->npmtr].dot[  ival] = args[iarg].dot[ival];
                                MODL->pmtr[MODL->npmtr].lbnd[ ival] = 0;
                                MODL->pmtr[MODL->npmtr].ubnd[ ival] = 0;
                            }
                        }
                    }

                } else if (strcmp(MODL->brch[jbrch].arg2, "$dim") == 0) {
                    status = str2val(MODL->brch[jbrch].arg3, MODL, &value, &dot);
                    CHECK_STATUS(str2val);

                    status = ocsmNewPmtr(MODL, &(MODL->brch[jbrch].arg1[1]), OCSM_INTERNAL, 1, NINT(value));
                    CHECK_STATUS(ocsmNewPmtr);

                    for (irc = 1; irc <= NINT(value); irc++) {
                        status = ocsmSetValuD(MODL, MODL->npmtr, 1, irc, 0);
                    }
                }

                jbrch++;
            }

        /* bad primtype */
        } else {
            signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                        "bad primtype");
            SET_STATUS(OCSM_ILLEGAL_VALUE, udprim);
        }

    /* execute" "restore $name index" */
    } else if (type == OCSM_RESTORE) {
        SPRINT3(1, "    executing [%4d] restore:    %s %11.5f",
                ibrch, args[1].str, args[2].val[0]);

        /* find name/index in storage */
        istor = -1;

        for (jstor = 0; jstor < MODL->nstor; jstor++) {
            if (strcmp(args[1].str, MODL->stor[jstor].name) == 0 &&
                NINT(args[2].val[0]) ==  MODL->stor[jstor].index   ) {
                istor = jstor;
                break;
            }
        }

        /* if storage does not exist, return an error */
        if (istor < 0) {
            SET_STATUS(OCSM_NAME_NOT_FOUND, restore);
        }

        ibodyl = MODL->stor[istor].ibody;

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, MODL->body[ibodyl].hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_RESTORE, ibodyl, -1,
                         args, MODL->body[ibodyl].hasdots, MODL->body[ibodyl].botype, &ibody);
        CHECK_STATUS(newBody);

        status = EG_copyObject(MODL->stor[istor].ebody, NULL, &ebody);
        CHECK_STATUS(EG_copyObject);

        MODL->body[ibody].ebody = ebody;

        /* mark the Faces with the current Branch */
        status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
        CHECK_STATUS(EG_getBodyTopos);

        for (iface = 1; iface <= nface; iface++) {
            status = setFaceAttribute(MODL, ibody, iface, ibrch, MODL->stor[istor].ibody, npatn, patn);
            CHECK_STATUS(setFaceAttribute);
        }

        EG_free(efaces);

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d restored (toler=%11.4e)",
                ibody, toler);

    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   buildSketch - implement OCSM_SKETCHs for ocsmBuild                 *
 *                                                                      *
 ************************************************************************
 */

static int
buildSketch(modl_T *modl,
            int    ibrch,
            varg_T args[],
            int    *nstack,
            int    stack[],
            int    npatn,
            patn_T patn[],
            sket_T *sket)
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       type, hasdots, i, j, n, nspln=0, nbezr=0, ibody, iopen, iswap;
    int       ipmtr, index1, index2, im1, ip1, ibeg, iend;
    double    xlast, ylast, zlast, areax, areay, areaz, area;
    double    xmin, xmax, ymin, ymax, zmin, zmax, pts[3*MAX_SKETCH_SIZE];
    double    dx1, dy1, dz1, ds1, dx2, dy2, dz2, ds2, dotp, swap;

    int        iseg, jseg, nseg, sense[MAX_SKETCH_SIZE], header[4];
    int        iface, nface, nedge, periodic, wireonly;
    double     data[20], tdata[2], result[3], dx, dy, dz;
    double     scent, xcent, ycent, zcent;

    int        jpmtr, jndex, ivar;

    char       value[MAX_EXPR_LEN], prefix[MAX_EXPR_LEN], suffix[MAX_EXPR_LEN];

    double     range[4], eval[18], norm[3];
    ego        ebody, *eedges, *efaces, ecurve, eflip, eloop, enew;
    ego        Enodes[MAX_SKETCH_SIZE+1], Eedges[MAX_SKETCH_SIZE], Efaces[1];

    ROUTINE(buildSketch);
    DPRINT2("%s(ibrch=%d) {",
            routine, ibrch);

    /* --------------------------------------------------------------- */

    type    = MODL->brch[ibrch].type;
    hasdots = 0;

    /* execute: "skbeg x y z" */
    if (type == OCSM_SKBEG) {
        SPRINT4(1, "    executing [%4d] skbeg:      %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0) {
            SPRINT3(1, "                                 %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0]);
        }

        /* store the base variables for the Sketch */
        sket->xbase = args[1].val[0];
        sket->ybase = args[2].val[0];
        sket->zbase = args[3].val[0];

        /* add a segment to the Sketch */
        sket->itype[sket->nseg] = OCSM_SKBEG;
        sket->ibrch[sket->nseg] = ibrch;
        sket->x[    sket->nseg] = args[1].val[0];
        sket->y[    sket->nseg] = args[2].val[0];
        sket->z[    sket->nseg] = args[3].val[0];
        sket->nseg++;

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status != 1) {
            status = newBody(MODL, ibrch, OCSM_SKBEG, -1, -1,
                             args, hasdots, OCSM_SKETCH, &ibody);
            CHECK_STATUS(newBody);
        }

    /* execute: "skvar $type valList" */
    } else if (type == OCSM_SKVAR) {
        SPRINT5(1, "    executing [%4d] skvar:   %s %11.5f %11.5f %11.5f ...",
                ibrch, args[1].str, args[2].val[0], args[2].val[1], args[2].val[2]);

        /* check that Sketch variables are not defined already */
        if (sket->nvar > 0) {
            signalError(MODL, OCSM_ILLEGAL_STATEMENT, ibrch,
                        "Sketch varibles already exist");
            SET_STATUS(OCSM_ILLEGAL_STATEMENT, skvar);
        }

        /* check that sketech variables are not defined yet */
        for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
            if        (strcmp(MODL->pmtr[ipmtr].name, "::x") == 0) {
                signalError(MODL, OCSM_NAME_ALREADY_DEFINED, ibrch,
                            "\"::x\" already exists");
                SET_STATUS(OCSM_NAME_ALREADY_DEFINED, skvar);
            } else if (strcmp(MODL->pmtr[ipmtr].name, "::y") == 0) {
                signalError(MODL, OCSM_NAME_ALREADY_DEFINED, ibrch,
                            "\"::y\" already exists");
                SET_STATUS(OCSM_NAME_ALREADY_DEFINED, skvar);
            } else if (strcmp(MODL->pmtr[ipmtr].name, "::z") == 0) {
                signalError(MODL, OCSM_NAME_ALREADY_DEFINED, ibrch,
                            "\"::z\" already exists");
                SET_STATUS(OCSM_NAME_ALREADY_DEFINED, skvar);
            } else if (strcmp(MODL->pmtr[ipmtr].name, "::d") == 0) {
                signalError(MODL, OCSM_NAME_ALREADY_DEFINED, ibrch,
                            "\"::d\" already exists");
                SET_STATUS(OCSM_NAME_ALREADY_DEFINED, skvar);
            }
        }

        /* make the new Sketch variables */
        sket->size = args[2].nval / 3;

        status = ocsmNewPmtr(MODL, "::x", OCSM_INTERNAL, 1, sket->size);
        CHECK_STATUS(ocsmNewPmtr);
        sket->ix = MODL->npmtr;

        status = ocsmNewPmtr(MODL, "::y", OCSM_INTERNAL, 1, sket->size);
        CHECK_STATUS(ocsmNewPmtr);
        sket->iy = MODL->npmtr;

        status = ocsmNewPmtr(MODL, "::z", OCSM_INTERNAL, 1, sket->size);
        CHECK_STATUS(ocsmNewPmtr);
        sket->iz = MODL->npmtr;

        status = ocsmNewPmtr(MODL, "::d", OCSM_INTERNAL, 1, sket->size);
        CHECK_STATUS(ocsmNewPmtr);
        sket->id = MODL->npmtr;

        /* set the values for the Sketch variables */
        if        (strcmp(args[1].str, "xy") == 0) {
            sket->type = 1;

            for (i = 0; i < sket->size; i++) {
                MODL->pmtr[sket->ix].value[i] = args[2].val[3*i  ];
                MODL->pmtr[sket->iy].value[i] = args[2].val[3*i+1];
                MODL->pmtr[sket->iz].value[i] = sket->zbase;
                MODL->pmtr[sket->id].value[i] = args[2].val[3*i+2];

                MODL->pmtr[sket->ix].dot[i] = 0;
                MODL->pmtr[sket->iy].dot[i] = 0;
                MODL->pmtr[sket->iz].dot[i] = 0;
                MODL->pmtr[sket->id].dot[i] = 0;
            }
        } else if (strcmp(args[1].str, "yz") == 0) {
            sket->type = 2;

            for (i = 0; i < sket->size; i++) {
                MODL->pmtr[sket->ix].value[i] = sket->xbase;
                MODL->pmtr[sket->iy].value[i] = args[2].val[3*i  ];
                MODL->pmtr[sket->iz].value[i] = args[2].val[3*i+1];
                MODL->pmtr[sket->id].value[i] = args[2].val[3*i+2];

                MODL->pmtr[sket->ix].dot[i] = 0;
                MODL->pmtr[sket->iy].dot[i] = 0;
                MODL->pmtr[sket->iz].dot[i] = 0;
                MODL->pmtr[sket->id].dot[i] = 0;
            }
        } else if (strcmp(args[1].str, "zx") == 0) {
            sket->type = 3;

            for (i = 0; i < sket->size; i++) {
                MODL->pmtr[sket->ix].value[i] = args[2].val[3*i+1];
                MODL->pmtr[sket->iy].value[i] = sket->ybase;
                MODL->pmtr[sket->iz].value[i] = args[2].val[3*i  ];
                MODL->pmtr[sket->id].value[i] = args[2].val[3*i+2];

                MODL->pmtr[sket->ix].dot[i] = 0;
                MODL->pmtr[sket->iy].dot[i] = 0;
                MODL->pmtr[sket->iz].dot[i] = 0;
                MODL->pmtr[sket->id].dot[i] = 0;
            }
        }

        /* make a list of the Sketch variables */
        for (i = 1; i <= sket->size; i++) {
            if (sket->type == 1 || sket->type == 3) {
                sket->ipmtr[sket->nvar] = sket->ix;
                sket->index[sket->nvar] = i-1;
                sket->nvar++;
            }

            if (sket->type == 2 || sket->type == 1) {
                sket->ipmtr[sket->nvar] = sket->iy;
                sket->index[sket->nvar] = i-1;
                sket->nvar++;
            }

            if (sket->type == 3 || sket->type == 2) {
                sket->ipmtr[sket->nvar] = sket->iz;
                sket->index[sket->nvar] = i-1;
                sket->nvar++;
            }

            if (fabs(MODL->pmtr[sket->id].value[i-1]) > EPS06) {
                sket->ipmtr[sket->nvar] = sket->id;
                sket->index[sket->nvar] = i-1;
                sket->nvar++;
            }
        }

        for (ivar = 0; ivar < sket->nvar; ivar++) {
            jpmtr = sket->ipmtr[ivar];
            jndex = sket->index[ivar];

            SPRINT3(1, "    -> creating  %s[%d] = %10.5f", MODL->pmtr[jpmtr].name, jndex+1,
                    MODL->pmtr[jpmtr].value[jndex]);
        }

        /* set the "constant" Sketch variables */
        for (i = 0; i < sket->size; i++) {
            if        (sket->type == 1) {
                MODL->pmtr[sket->iz].value[i] = sket->zbase;
                MODL->pmtr[sket->iz].dot[  i] = 0;
            } else if (sket->type == 2) {
                MODL->pmtr[sket->ix].value[i] = sket->xbase;
                MODL->pmtr[sket->ix].dot[  i] = 0;
            } else {
                MODL->pmtr[sket->iy].value[i] = sket->ybase;
                MODL->pmtr[sket->iy].dot[  i] = 0;
            }
        }

    /* execute: "skcon $type index1 index2=-1 $value=0" */
    } else if (type == OCSM_SKCON) {
        SPRINT5(1, "    executing [%4d] skcon:   %s  %11.5f %11.5f %s",
                ibrch, args[1].str, args[2].val[0], args[3].val[0], args[4].str);

        index1 = NINT(args[2].val[0]);
        index2 = NINT(args[3].val[0]);

        /* copy args[4].str to value so that it can be modified if needed below */
        strcpy(value, args[4].str);

        /* modify value if it contains "::L[" */
        while (strstr(value, "::L[") != NULL) {
            prefix[0] = '\0';
            suffix[0] = '\0';
            ibeg      =   0 ;
            for (i = 0; i < strlen(value); i++) {
                if (value[i  ] == ':' && value[i+1] == ':' &&
                    value[i+2] == 'L' && value[i+3] == '['   ) {
                    if (i > 0) {
                        strncpy(prefix, value, i);
                        prefix[i] = '\0';
                    }
                    ibeg = strtol(&(value[i+4]), NULL, 10);
                    break;
                }
            }

            if (ibeg < 1 || ibeg > sket->size) {
                signalError(MODL, OCSM_ILLEGAL_VALUE, ibrch,
                            "::L[%d] is not defined", ibeg);
                status = OCSM_ILLEGAL_VALUE;
                goto cleanup;
            } else if (ibeg < sket->size) {
                iend = ibeg + 1;
            } else {
                iend = 1;
            }

            for (j = i+3; j < strlen(value); j++) {
                if (value[j] == ']') {
                    strcpy(suffix, &(value[j+1]));
                    break;
                }
            }

            if        (sket->type == 1) {
                snprintf(value, MAX_EXPR_LEN, "%shypot(::y[%d]-::y[%d],::x[%d]-::x[%d])%s",
                         prefix, iend, ibeg, iend, ibeg, suffix);
            } else if (sket->type == 2) {
                snprintf(value, MAX_EXPR_LEN, "%shypot(::z[%d]-::z[%d],::y[%d]-::y[%d])%s",
                         prefix, iend, ibeg, iend, ibeg, suffix);
            } else {
                snprintf(value, MAX_EXPR_LEN, "%shypot(::x[%d]-::x[%d],::z[%d]-::z[%d])%s",
                         prefix, iend, ibeg, iend, ibeg, suffix);
            }
        }

        /* constraints on XY plane */
        if (sket->type == 1) {
            if        (strcmp(args[1].str, "X") == 0 && index2 <= 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::x[%d]-(%s)",
                         index1, value);
            } else if (strcmp(args[1].str, "X") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "Xcent(::x[%d],::y[%d],::d[%d],::x[%d],::y[%d])-(%s)",
                         index1, index1, index2, index2, index2, value);
            } else if (strcmp(args[1].str, "Y") == 0 && index2 <= 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::y[%d]-(%s)",
                         index1, value);
            } else if (strcmp(args[1].str, "Y") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "Ycent(::x[%d],::y[%d],::d[%d],::x[%d],::y[%d])-(%s)",
                         index1, index1, index2, index2, index2, value);
            } else if (strcmp(args[1].str, "P") == 0) {
                im1 = index1 - 1; if (im1 == 0           ) im1 = sket->size;
                ip1 = index1 + 1; if (ip1 == sket->size+1) ip1 = 1;
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "(abs(turnang(::x[%d],::y[%d],::d[%d],::x[%d],::y[%d],::d[%d],::x[%d],::y[%d]))-90)/57.3",
                         im1, im1, index1, index1, index1, ip1, ip1, ip1);
            } else if (strcmp(args[1].str, "T") == 0) {
                im1 = index1 - 1; if (im1 == 0           ) im1 = sket->size;
                ip1 = index1 + 1; if (ip1 == sket->size+1) ip1 = 1;
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "turnang(::x[%d],::y[%d],::d[%d],::x[%d],::y[%d],::d[%d],::x[%d],::y[%d])/57.3",
                         im1, im1, index1, index1, index1, ip1, ip1, ip1);
            } else if (strcmp(args[1].str, "A") == 0) {
                im1 = index1 - 1; if (im1 == 0           ) im1 = sket->size;
                ip1 = index1 + 1; if (ip1 == sket->size+1) ip1 = 1;
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "(turnang(::x[%d],::y[%d],::d[%d],::x[%d],::y[%d],::d[%d],::x[%d],::y[%d])-(%s))/57.3",
                         im1, im1, index1, index1, index1, ip1, ip1, ip1, value);
            } else if (strcmp(args[1].str, "W") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::x[%d]-::x[%d]-(%s)",
                         index2, index1, value);
            } else if (strcmp(args[1].str, "D") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::y[%d]-::y[%d]-(%s)",
                         index2, index1, value);
            } else if (strcmp(args[1].str, "H") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::y[%d]-::y[%d]",
                         index1, index2);
            } else if (strcmp(args[1].str, "V") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::x[%d]-::x[%d]",
                         index1, index2);
            } else if (strcmp(args[1].str, "I") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "(atan2d(::y[%d]-::y[%d],::x[%d]-::x[%d])-(%s))/57.3",
                         index2, index1, index2, index1, value);
            } else if (strcmp(args[1].str, "L") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "hypot(::y[%d]-::y[%d],::x[%d]-::x[%d])-(%s)",
                         index1, index2, index1, index2, value);
            } else if (strcmp(args[1].str, "R") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "radius(::x[%d],::y[%d],::d[%d],::x[%d],::y[%d])-(%s)",
                         index1, index1, index2, index2, index2, value);
            } else if (strcmp(args[1].str, "S") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "(sweep(::x[%d],::y[%d],::d[%d],::x[%d],::y[%d])-(%s))/57.3",
                         index1, index1, index2, index2, index2, value);
            }

        /* constraints on YZ plane */
        } else if (sket->type == 2) {
            if        (strcmp(args[1].str, "Y") == 0 && index2 <= 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::y[%d]-(%s)",
                         index1, value);
            } else if (strcmp(args[1].str, "Y") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "Xcent(::y[%d],::z[%d],::d[%d],::y[%d],::z[%d])-(%s)",
                         index1, index1, index2, index2, index2, value);
            } else if (strcmp(args[1].str, "Z") == 0 && index2 <= 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::z[%d]-(%s)",
                         index1, value);
            } else if (strcmp(args[1].str, "Z") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "Ycent(::y[%d],::z[%d],::d[%d],::y[%d],::z[%d])-(%s)",
                         index1, index1, index2, index2, index2, value);
            } else if (strcmp(args[1].str, "P") == 0) {
                im1 = index1 - 1; if (im1 == 0           ) im1 = sket->size;
                ip1 = index1 + 1; if (ip1 == sket->size+1) ip1 = 1;
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "(abs(turnang(::y[%d],::z[%d],::d[%d],::y[%d],::z[%d],::d[%d],::y[%d],::z[%d]))-90)/57.3",
                         im1, im1, index1, index1, index1, ip1, ip1, ip1);
            } else if (strcmp(args[1].str, "T") == 0) {
                im1 = index1 - 1; if (im1 == 0           ) im1 = sket->size;
                ip1 = index1 + 1; if (ip1 == sket->size+1) ip1 = 1;
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "turnang(::y[%d],::z[%d],::d[%d],::y[%d],::z[%d],::d[%d],::y[%d],::z[%d])/57.3",
                         im1, im1, index1, index1, index1, ip1, ip1, ip1);
            } else if (strcmp(args[1].str, "A") == 0) {
                im1 = index1 - 1; if (im1 == 0           ) im1 = sket->size;
                ip1 = index1 + 1; if (ip1 == sket->size+1) ip1 = 1;
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "(turnang(::y[%d],::z[%d],::d[%d],::y[%d],::z[%d],::d[%d],::y[%d],::z[%d])-(%s))/57.3",
                         im1, im1, index1, index1, index1, ip1, ip1, ip1, value);
            } else if (strcmp(args[1].str, "W") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::y[%d]-::y[%d]-(%s)",
                         index2, index1, value);
            } else if (strcmp(args[1].str, "D") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::z[%d]-::z[%d]-(%s)",
                         index2, index1, value);
            } else if (strcmp(args[1].str, "H") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::z[%d]-::z[%d]",
                         index1, index2);
            } else if (strcmp(args[1].str, "V") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::y[%d]-::y[%d]",
                         index1, index2);
            } else if (strcmp(args[1].str, "I") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "(atan2d(::z[%d]-::z[%d],::y[%d]-::y[%d])-(%s))/57.3",
                         index2, index1, index2, index1, value);
            } else if (strcmp(args[1].str, "L") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "hypot(::z[%d]-::z[%d],::y[%d]-::y[%d])-(%s)",
                         index1, index2, index1, index2, value);
            } else if (strcmp(args[1].str, "R") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "radius(::y[%d],::z[%d],::d[%d],::y[%d],::z[%d])-(%s)",
                         index1, index1, index2, index2, index2, value);
            } else if (strcmp(args[1].str, "S") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "(sweep(::y[%d],::z[%d],::d[%d],::y[%d],::z[%d])-(%s))/57.3",
                         index1, index1, index2, index2, index2, value);
            }

        /* constraints on ZX plane */
        } else {
            if        (strcmp(args[1].str, "Z") == 0 && index2 <= 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::z[%d]-(%s)",
                         index1, value);
            } else if (strcmp(args[1].str, "Z") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "Xcent(::z[%d],::x[%d],::d[%d],::z[%d],::x[%d])-(%s)",
                         index1, index1, index2, index2, index2, value);
            } else if (strcmp(args[1].str, "X") == 0 && index2 <= 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::x[%d]-(%s)",
                         index1, value);
            } else if (strcmp(args[1].str, "X") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "Ycent(::z[%d],::x[%d],::d[%d],::z[%d],::x[%d])-(%s)",
                         index1, index1, index2, index2, index2, value);
            } else if (strcmp(args[1].str, "P") == 0) {
                im1 = index1 - 1; if (im1 == 0           ) im1 = sket->size;
                ip1 = index1 + 1; if (ip1 == sket->size+1) ip1 = 1;
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "(abs(turnang(::z[%d],::x[%d],::d[%d],::z[%d],::x[%d],::d[%d],::z[%d],::x[%d]))-90)/57.3",
                         im1, im1, index1, index1, index1, ip1, ip1, ip1);
            } else if (strcmp(args[1].str, "T") == 0) {
                im1 = index1 - 1; if (im1 == 0           ) im1 = sket->size;
                ip1 = index1 + 1; if (ip1 == sket->size+1) ip1 = 1;
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "turnang(::z[%d],::x[%d],::d[%d],::z[%d],::x[%d],::d[%d],::z[%d],::x[%d])/57.3",
                         im1, im1, index1, index1, index1, ip1, ip1, ip1);
            } else if (strcmp(args[1].str, "A") == 0) {
                im1 = index1 - 1; if (im1 == 0           ) im1 = sket->size;
                ip1 = index1 + 1; if (ip1 == sket->size+1) ip1 = 1;
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "(turnang(::z[%d],::x[%d],::d[%d],::z[%d],::x[%d],::d[%d],::z[%d],::x[%d])-(%s))/57.3",
                         im1, im1, index1, index1, index1, ip1, ip1, ip1, value);
            } else if (strcmp(args[1].str, "W") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::z[%d]-::z[%d]-(%s)",
                         index2, index1, value);
            } else if (strcmp(args[1].str, "D") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::x[%d]-::x[%d]-(%s)",
                         index2, index1, value);
            } else if (strcmp(args[1].str, "H") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::x[%d]-::x[%d]",
                         index1, index2);
            } else if (strcmp(args[1].str, "V") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "::z[%d]-::z[%d]",
                         index1, index2);
            } else if (strcmp(args[1].str, "I") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "(atan2d(::x[%d]-::x[%d],::z[%d]-::z[%d])-(%s))/57.3",
                         index2, index1, index2, index1, value);
            } else if (strcmp(args[1].str, "L") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "hypot(::x[%d]-::x[%d],::z[%d]-::z[%d])-(%s)",
                         index1, index2, index1, index2, value);
            } else if (strcmp(args[1].str, "R") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "radius(::z[%d],::x[%d],::d[%d],::z[%d],::x[%d])-(%s)",
                         index1, index1, index2, index2, index2, value);
            } else if (strcmp(args[1].str, "S") == 0) {
                snprintf(sket->con[sket->ncon++], MAX_EXPR_LEN,
                         "(sweep(::z[%d],::x[%d],::d[%d],::z[%d],::x[%d])-(%s))/57.3",
                         index1, index1, index2, index2, index2, value);
            }
        }

        /* make a note that the Sketch needs to be solved */
        sket->solved = 0;

    /* execute: "linseg x y z" */
    } else if (type == OCSM_LINSEG) {
        SPRINT4(1, "    executing [%4d] linseg:     %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0) {
            hasdots = 1;

            SPRINT3(1, "                                 %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0]);
        }

        /* add a segment to the Sketch */
        sket->itype[sket->nseg] = OCSM_LINSEG;
        sket->ibrch[sket->nseg] = ibrch;
        sket->x[    sket->nseg] = args[1].val[0];
        sket->y[    sket->nseg] = args[2].val[0];
        sket->z[    sket->nseg] = args[3].val[0];
        sket->nseg++;

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status != 1) {
            status = newBody(MODL, ibrch, OCSM_LINSEG, MODL->nbody, -1,
                             args, hasdots, OCSM_SKETCH, &ibody);
            CHECK_STATUS(newBody);
        }

    /* execute: "cirarc xon yon zon xend yend zend" */
    } else if (type == OCSM_CIRARC) {
        SPRINT7(1, "    executing [%4d] cirarc:     %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0], args[4].val[0],
                       args[5].val[0], args[6].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0 || args[4].dot[0] != 0 ||
            args[5].dot[0] != 0 || args[6].dot[0] != 0                                   ) {
            hasdots = 1;

            SPRINT6(1, "                                 %11.5f %11.5f %11.5f %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0], args[4].dot[0],
                       args[5].dot[0], args[6].dot[0]);
        }

        /* check for the colinearity of the 3 points */
        dx1 = args[1].val[0] - sket->x[sket->nseg-1];
        dy1 = args[2].val[0] - sket->y[sket->nseg-1];
        dz1 = args[3].val[0] - sket->z[sket->nseg-1];
        ds1 = sqrt(dx1*dx1 + dy1*dy1 + dz1*dz1);

        dx2 = args[4].val[0] - args[1].val[0];
        dy2 = args[5].val[0] - args[2].val[0];
        dz2 = args[6].val[0] - args[3].val[0];
        ds2 = sqrt(dx2*dx2 + dy2*dy2 + dz2*dz2);

        dotp = (dx1*dx2 + dy1*dy2 + dz1*dz2) / ds1 / ds2;

        /* if points are co-linear, convert to a linseg */
        if (fabs(dotp) > 0.9999) {
            SPRINT0(0, "WARNING:: converting to linseg since points are colinear");

            sket->itype[sket->nseg] = OCSM_LINSEG;
            sket->ibrch[sket->nseg] = ibrch;
            sket->x[    sket->nseg] = args[4].val[0];
            sket->y[    sket->nseg] = args[5].val[0];
            sket->z[    sket->nseg] = args[6].val[0];
            sket->nseg++;

            /* recycle old Body if not dirty */
            status = recycleBody(MODL, ibrch, type, args, hasdots);
            CHECK_STATUS(recycleBody);

            if (status != 1) {
                args[1].val[0] = args[4].val[0];
                args[2].val[0] = args[5].val[0];
                args[3].val[0] = args[6].val[0];

                args[1].dot[0] = args[4].dot[0];
                args[2].dot[0] = args[5].dot[0];
                args[3].dot[0] = args[6].dot[0];

                status = newBody(MODL, ibrch, OCSM_LINSEG, MODL->nbody, -1,
                                 args, hasdots, OCSM_SKETCH, &ibody);
                CHECK_STATUS(newBody);
            }

        /* add the cirarc to the Sketch (as two segments) */
        } else {
            sket->itype[sket->nseg] = OCSM_CIRARC;
            sket->ibrch[sket->nseg] = ibrch;
            sket->x[    sket->nseg] = args[1].val[0];
            sket->y[    sket->nseg] = args[2].val[0];
            sket->z[    sket->nseg] = args[3].val[0];
            sket->nseg++;

            sket->itype[sket->nseg] = OCSM_CIRARC;
            sket->ibrch[sket->nseg] = ibrch;
            sket->x[    sket->nseg] = args[4].val[0];
            sket->y[    sket->nseg] = args[5].val[0];
            sket->z[    sket->nseg] = args[6].val[0];
            sket->nseg++;

            /* recycle old Body if not dirty */
            status = recycleBody(MODL, ibrch, type, args, hasdots);
            CHECK_STATUS(recycleBody);

            if (status != 1) {
                status = newBody(MODL, ibrch, OCSM_CIRARC, MODL->nbody, -1,
                                 args, hasdots, OCSM_SKETCH, &ibody);
                CHECK_STATUS(newBody);
            }
        }

    /* execute: "arc xend yend zend dist $plane=xy" */
    } else if (type == OCSM_ARC) {
        SPRINT6(1, "    executing [%4d] arc:        %11.5f %11.5f %11.5f %11.5f %s",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0], args[4].val[0],
                       args[5].str);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0 || args[4].dot[0] != 0) {
            hasdots = 1;

            SPRINT4(1, "                                 %11.5f %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0], args[4].dot[0]);
        }

        /* if points are co-linear, convert to a linseg */
        if (fabs(args[4].val[0]) < 0.0001) {
            SPRINT0(0, "WARNING:: converting to linseg since points are colinear");

            sket->itype[sket->nseg] = OCSM_LINSEG;
            sket->ibrch[sket->nseg] = ibrch;
            sket->x[    sket->nseg] = args[1].val[0];
            sket->y[    sket->nseg] = args[2].val[0];
            sket->z[    sket->nseg] = args[3].val[0];
            sket->nseg++;

            /* recycle old Body if not dirty */
            status = recycleBody(MODL, ibrch, type, args, hasdots);
            CHECK_STATUS(recycleBody);

            if (status != 1) {
                status = newBody(MODL, ibrch, OCSM_LINSEG, MODL->nbody, -1,
                                 args, hasdots, OCSM_SKETCH, &ibody);
                CHECK_STATUS(newBody);
            }

        /* add the arc to the Sketch (as two segments) */
        } else {
            xcent = (args[1].val[0] + sket->x[sket->nseg-1]) / 2;
            ycent = (args[2].val[0] + sket->y[sket->nseg-1]) / 2;
            zcent = (args[3].val[0] + sket->z[sket->nseg-1]) / 2;

            dx    = (args[1].val[0] - sket->x[sket->nseg-1]);
            dy    = (args[2].val[0] - sket->y[sket->nseg-1]);
            dz    = (args[3].val[0] - sket->z[sket->nseg-1]);

            if        (strcmp(args[5].str, "xy") == 0 ) {
                sket->itype[sket->nseg] = OCSM_CIRARC;
                sket->ibrch[sket->nseg] = ibrch;
                sket->x[    sket->nseg] = xcent + args[4].val[0] * dy / sqrt(dx*dx + dy*dy);
                sket->y[    sket->nseg] = ycent - args[4].val[0] * dx / sqrt(dx*dx + dy*dy);
                sket->z[    sket->nseg] = zcent;
                sket->nseg++;
            } else if (strcmp(args[5].str, "yz") == 0) {
                sket->itype[sket->nseg] = OCSM_CIRARC;
                sket->ibrch[sket->nseg] = ibrch;
                sket->y[    sket->nseg] = ycent + args[4].val[0] * dz / sqrt(dy*dy + dz*dz);
                sket->z[    sket->nseg] = zcent - args[4].val[0] * dy / sqrt(dy*dy + dz*dz);
                sket->x[    sket->nseg] = xcent;
                sket->nseg++;
            } else {
                sket->itype[sket->nseg] = OCSM_CIRARC;
                sket->ibrch[sket->nseg] = ibrch;
                sket->z[    sket->nseg] = zcent + args[4].val[0] * dx / sqrt(dz*dz + dx*dx);
                sket->x[    sket->nseg] = xcent - args[4].val[0] * dz / sqrt(dz*dz + dx*dx);
                sket->y[    sket->nseg] = ycent;
                sket->nseg++;
            }

            sket->itype[sket->nseg] = OCSM_CIRARC;
            sket->ibrch[sket->nseg] = ibrch;
            sket->x[    sket->nseg] = args[1].val[0];
            sket->y[    sket->nseg] = args[2].val[0];
            sket->z[    sket->nseg] = args[3].val[0];
            sket->nseg++;

            /* recycle old Body if not dirty */
            status = recycleBody(MODL, ibrch, type, args, hasdots);
            CHECK_STATUS(recycleBody);

            if (status != 1) {
                status = newBody(MODL, ibrch, OCSM_ARC, MODL->nbody, -1,
                                 args, hasdots, OCSM_SKETCH, &ibody);
                CHECK_STATUS(newBody);
            }
        }

    /* execute: "spline x y z" */
    } else if (type == OCSM_SPLINE) {
        SPRINT4(1, "    executing [%4d] spline:     %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0) {
            hasdots = 1;

            SPRINT3(1, "                                 %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0]);
        }

        /* add the spline to the Sketch */
        sket->itype[sket->nseg] = OCSM_SPLINE;
        sket->ibrch[sket->nseg] = ibrch;
        sket->x[    sket->nseg] = args[1].val[0];
        sket->y[    sket->nseg] = args[2].val[0];
        sket->z[    sket->nseg] = args[3].val[0];
        sket->nseg++;

        /* warning for repeated points */
        if (sket->nseg > 1) {
            if (fabs(sket->x[sket->nseg-1]-sket->x[sket->nseg-2]) < EPS06 &&
                fabs(sket->y[sket->nseg-1]-sket->y[sket->nseg-2]) < EPS06 &&
                fabs(sket->z[sket->nseg-1]-sket->z[sket->nseg-2]) < EPS06   ) {
                SPRINT0(0, "WARNING:: repeated point in spline being removed");
                sket->nseg--;
            }
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status != 1) {
            status = newBody(MODL, ibrch, OCSM_SPLINE, MODL->nbody, -1,
                             args, hasdots, OCSM_SKETCH, &ibody);
            CHECK_STATUS(newBody);
        }

    /* execute: "bezier x y z" */
    } else if (type == OCSM_BEZIER) {
        SPRINT4(1, "    executing [%4d] bezier:     %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0) {
            hasdots = 1;

            SPRINT3(1, "                                 %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0]);
        }

        /* add the bezier to the Sketch */
        sket->itype[sket->nseg] = OCSM_BEZIER;
        sket->ibrch[sket->nseg] = ibrch;
        sket->x[    sket->nseg] = args[1].val[0];
        sket->y[    sket->nseg] = args[2].val[0];
        sket->z[    sket->nseg] = args[3].val[0];
        sket->nseg++;

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status != 1) {
            status = newBody(MODL, ibrch, OCSM_BEZIER, MODL->nbody, -1,
                             args, hasdots, OCSM_SKETCH, &ibody);
            CHECK_STATUS(newBody);
        }

    /* execute: "skend wireonly=0" */
    } else if (type == OCSM_SKEND) {
        SPRINT2(1, "    executing [%4d] skend:      %11.5f",
                ibrch, args[1].val[0]);

        wireonly = NINT(args[1].val[0]);

        /* add the skend to the Sketch */
        sket->itype[sket->nseg] = OCSM_SKEND;
        sket->ibrch[sket->nseg] = ibrch;
        sket->x[    sket->nseg] = sket->x[sket->nseg-1];   // not used
        sket->y[    sket->nseg] = sket->y[sket->nseg-1];
        sket->z[    sket->nseg] = sket->z[sket->nseg-1];
        sket->nseg++;

        /* if we have only one Sketch segment, then we are creating a OCSM_NODE_BODY */
        if ((sket->nseg) == 2) {
            /* close the Sketch */
            sket->type   = 0;
            sket->size   = 0;
            sket->solved = 1;
            sket->nseg   = 0;
            sket->nvar   = 0;
            sket->ncon   = 0;

            /* recycle old Body if not dirty */
            status = recycleBody(MODL, ibrch, type, args, hasdots);
            CHECK_STATUS(recycleBody);

            if (status == 1) {
                stack[(*nstack)++] = MODL->nbody;
                status = 0;
                goto cleanup;
            }

            /* create the Body */
            status = newBody(MODL, ibrch, OCSM_SKEND, -1, -1,
                             args, hasdots, OCSM_NODE_BODY, &ibody);
            CHECK_STATUS(newBody);

            pts[0] = sket->x[0];
            pts[1] = sket->y[0];
            pts[2] = sket->z[0];

            status = EG_makeTopology(MODL->context, NULL, NODE, 0, pts, 0, NULL, NULL, &ebody);
            CHECK_STATUS(EG_makeTopology);

            MODL->body[ibody].ebody = ebody;

            /* push the Body onto the stack */
            stack[(*nstack)++] = ibody;

            SPRINT1(1, "                          Node   %4d created", ibody);

        /* we have more than one Sketch segment */
        } else {

            /* recycle old Body if not dirty */
            status = recycleBody(MODL, ibrch, type, args, hasdots);
            CHECK_STATUS(recycleBody);

            if (status == 1) {
                stack[(*nstack)++] = MODL->nbody;

                /* remove ::x[], ::y[], ::z[], and ::d[] if they were generated by a skvar statement */
                if (sket->nvar > 0) {
                    if (strcmp(MODL->pmtr[MODL->npmtr  ].name, "::d") != 0 ||
                        strcmp(MODL->pmtr[MODL->npmtr-1].name, "::z") != 0 ||
                        strcmp(MODL->pmtr[MODL->npmtr-2].name, "::y") != 0 ||
                        strcmp(MODL->pmtr[MODL->npmtr-3].name, "::x") != 0   ) {
                        signalError(MODL, OCSM_INTERNAL_ERROR, ibrch,
                                    "last 4 Parameters are not ::d, ::z, ::y, and ::x");
                        SET_STATUS(OCSM_INTERNAL_ERROR, skend);
                    }

                    for (i = 0; i < 4; i++) {
                        FREE(MODL->pmtr[MODL->npmtr].name );
                        FREE(MODL->pmtr[MODL->npmtr].value);
                        FREE(MODL->pmtr[MODL->npmtr].dot  );
                        FREE(MODL->pmtr[MODL->npmtr].lbnd );
                        FREE(MODL->pmtr[MODL->npmtr].ubnd );
                        MODL->npmtr--;
                    }
                }

                /* close the Sketch */
                sket->type   = 0;
                sket->size   = 0;
                sket->solved = 1;
                sket->nseg   = 0;
                sket->nvar   = 0;
                sket->ncon   = 0;

                goto cleanup;
            }

            /* find the extrema of the Sketch points */
            xmin = sket->x[0];
            xmax = sket->x[0];
            ymin = sket->y[0];
            ymax = sket->y[0];
            zmin = sket->z[0];
            zmax = sket->z[0];
            SPRINT5(2, "skpnt[%3d] %3d %10.5f %10.5f %10.5f",
                    0, sket->itype[0], sket->x[0], sket->y[0], sket->z[0]);

            for (iseg = 1; iseg < sket->nseg; iseg++) {
                if (sket->x[iseg] < xmin) xmin = sket->x[iseg];
                if (sket->x[iseg] > xmax) xmax = sket->x[iseg];
                if (sket->y[iseg] < ymin) ymin = sket->y[iseg];
                if (sket->y[iseg] > ymax) ymax = sket->y[iseg];
                if (sket->z[iseg] < zmin) zmin = sket->z[iseg];
                if (sket->z[iseg] > zmax) zmax = sket->z[iseg];

                SPRINT5(2, "skpnt[%3d] %3d %10.5f %10.5f %10.5f",
                        iseg, sket->itype[iseg], sket->x[iseg], sket->y[iseg], sket->z[iseg]);
            }

//$$$                if        (xmin == xmax && ymin == ymax) {
//$$$                    SPRINT4(0, "xmin=%11.5f, xmax=%11.5f, ymin=%11.5f, ymax=%11.5f", xmin, xmax, ymin, ymax);
//$$$                    status = OCSM_COLINEAR_SKETCH_POINTS;
//$$$                    CHECK_STATUS(skend);
//$$$                } else if (ymin == ymax && zmin == zmax) {
//$$$                    SPRINT4(0, "ymin=%11.5f, ymax=%11.5f, zmin=%11.5f, zmax=%11.5f", ymin, ymax, zmin, zmax);
//$$$                    status = OCSM_COLINEAR_SKETCH_POINTS;
//$$$                    CHECK_STATUS(skend);
//$$$                } else if (zmin == zmax && xmin == xmax) {
//$$$                    SPRINT4(0, "zmin=%11.5f, zmax=%11.5f, xmin=%11.5f, xmax=%11.5f", zmin, zmax, xmin, xmax);
//$$$                    status = OCSM_COLINEAR_SKETCH_POINTS;
//$$$                    CHECK_STATUS(skend);
//$$$                }

            /* determine if the Sketch is open or closed */
            if (fabs(sket->x[sket->nseg-2]-sket->x[0]) < EPS06 &&
                fabs(sket->y[sket->nseg-2]-sket->y[0]) < EPS06 &&
                fabs(sket->z[sket->nseg-2]-sket->z[0]) < EPS06   ) {
                iopen = 0;
            } else {
                iopen = 1;
            }
            SPRINT1(2, "iopen=%d", iopen);

            /* find the approximate area of the Sketch to determine
               if it has been generated counter-clockwise or clockwise */
            areax = 0;
            areay = 0;
            areaz = 0;
            for (iseg = 2; iseg < sket->nseg; iseg++) {
                areax += ( (sket->y[iseg-1]-sket->y[0]) * (sket->z[iseg  ]-sket->z[0])
                         - (sket->y[iseg  ]-sket->y[0]) * (sket->z[iseg-1]-sket->z[0]));
                areay += ( (sket->z[iseg-1]-sket->z[0]) * (sket->x[iseg  ]-sket->x[0])
                         - (sket->z[iseg  ]-sket->z[0]) * (sket->x[iseg-1]-sket->x[0]));
                areaz += ( (sket->x[iseg-1]-sket->x[0]) * (sket->y[iseg  ]-sket->y[0])
                         - (sket->x[iseg  ]-sket->x[0]) * (sket->y[iseg-1]-sket->y[0]));
            }

            area = areax / MAX(xmax-xmin, EPS06)
                 + areay / MAX(ymax-ymin, EPS06)
                 + areaz / MAX(zmax-zmin, EPS06);

            if (iopen == 0 && area < 0) {
                SPRINT0(0,"WARNING:: Sketch is defined clockwise.  flipping direction");

                jseg = sket->nseg - 2;
                for (iseg = 1; iseg < sket->nseg; iseg++) {
                    if (iseg >= jseg) break;

                    iswap             = sket->itype[iseg];
                    sket->itype[iseg] = sket->itype[jseg];
                    sket->itype[jseg] = iswap;

                    iswap             = sket->ibrch[iseg];
                    sket->ibrch[iseg] = sket->ibrch[jseg];
                    sket->ibrch[jseg] = iswap;

                    jseg--;
                }

                jseg = sket->nseg - 3;
                for (iseg = 1; iseg < sket->nseg; iseg++) {
                    if (iseg >= jseg) break;

                    swap          = sket->x[iseg];
                    sket->x[iseg] = sket->x[jseg];
                    sket->x[jseg] = swap;

                    swap          = sket->y[iseg];
                    sket->y[iseg] = sket->y[jseg];
                    sket->y[jseg] = swap;

                    swap          = sket->z[iseg];
                    sket->z[iseg] = sket->z[jseg];
                    sket->z[jseg] = swap;

                    jseg--;
                }
            }

            /* create the beginning node */
            nseg = 0;
            pts[0] = sket->x[0];
            pts[1] = sket->y[0];
            pts[2] = sket->z[0];
            status = EG_makeTopology(MODL->context, NULL, NODE, 0,
                                     pts, 0, NULL, NULL, &(Enodes[0]));
            CHECK_STATUS(EG_makeTopology);

            /* save "last" point */
            xlast = sket->x[0];
            ylast = sket->y[0];
            zlast = sket->z[0];

            /* no points in spline or bezier so far */
            nspln = 0;
            nbezr = 0;

            /* add the lines, circular-arcs, splines, and beziers to the Sketch plane */
            for (iseg = 1; iseg < sket->nseg; iseg++) {

                /* if we were defining a spline but the new segment is not
                   a spline, generate the spline now */
                if (nspln > 0 && sket->itype[iseg] != OCSM_SPLINE) {
                    if (nspln < 3) {
                        SET_STATUS(OCSM_TOO_FEW_SPLINE_POINTS, skend);
                    }

                    SPRINT1(2, "spline (w/%d points):", nspln);
                    for (n = 0; n < nspln; n++) {
                        SPRINT4(2, "%5d %11.5f %11.5f %11.5f",   n, pts[3*n], pts[3*n+1], pts[3*n+2]);
                    }

                    if (iseg < sket->nseg-1 || iopen == 1) {
                        data[0] = pts[3*nspln-3];
                        data[1] = pts[3*nspln-2];
                        data[2] = pts[3*nspln-1];
                        status = EG_makeTopology(MODL->context, NULL, NODE, 0,
                                                 data, 0, NULL, NULL, &(Enodes[nseg+1]));
                        CHECK_STATUS(EG_makeTopology);
                    } else {
                        Enodes[nseg+1] = Enodes[0];
                    }

                    header[0] = nspln;
                    header[1] = 0;

                    status = EG_approximate(MODL->context, 0, EPS06, header, pts, &ecurve);
                    CHECK_STATUS(EG_approximate);

                    status = EG_getRange(ecurve, tdata, &periodic);
                    CHECK_STATUS(EG_getRange);

                    #ifdef SHOW_SPLINES
                    {
                        int    io_kbd=5, io_scr=6, indgr=1+4+16+64;
                        int    i, nline=9, ilin[9], isym[9], nper[9];
                        float  xplot[9000], yplot[9000];
                        double uu, xyz3[9];

                        for (i = 0; i < nline; i++) {
                            ilin[i] = +i;
                            isym[i] = -i;
                            nper[i] = 1000;
                        }

                        for (i = 0; i < 1000; i++) {
                            uu = tdata[0] + (double)i/999 * (tdata[1] - tdata[0]);

                            EG_evaluate(ecurve, &uu, xyz3);

                            xplot[      i] = uu;   yplot[      i] = xyz3[0];   //   x
                            xplot[ 1000+i] = uu;   yplot[ 1000+i] = xyz3[1];   //   y
                            xplot[ 2000+i] = uu;   yplot[ 2000+i] = xyz3[2];   //   z
                            xplot[ 3000+i] = uu;   yplot[ 3000+i] = xyz3[3];   //  dx/du
                            xplot[ 4000+i] = uu;   yplot[ 4000+i] = xyz3[4];   //  dy/du
                            xplot[ 5000+i] = uu;   yplot[ 5000+i] = xyz3[5];   //  dz/du
                            xplot[ 6000+i] = uu;   yplot[ 6000+i] = xyz3[6];   // d2x/du2
                            xplot[ 7000+i] = uu;   yplot[ 7000+i] = xyz3[7];   // d2y/du2
                            xplot[ 8000+i] = uu;   yplot[ 8000+i] = xyz3[8];   // d2z/du2
                        }

                        grinit_(&io_kbd, &io_scr, "generation of spline in ocsmBuild",
                                strlen("generation of spline in ocsmBuild"));
                        grline_(ilin, isym, &nline,                "~x~y~spline data",
                                &indgr, xplot, yplot, nper, strlen("~x~y~spline data"));
                    }
                    #endif

                    status = EG_makeTopology(MODL->context, ecurve, EDGE, TWONODE,
                                             tdata, 2, &(Enodes[nseg]), NULL, &(Eedges[nseg]));
                    CHECK_STATUS(EG_makeTopology);

                    nseg++;

                    nspln = 0;

                    xlast = sket->x[iseg-1];
                    ylast = sket->y[iseg-1];
                    zlast = sket->z[iseg-1];
                }

                /* if we were defining a bezier but the new segment is not
                   a bezier, generate the bezier now */
                if (nbezr > 0 && sket->itype[iseg] != OCSM_BEZIER) {
                    if (nbezr < 2) {
                        SET_STATUS(OCSM_TOO_FEW_SPLINE_POINTS, skend);
                    }

                    SPRINT1(2, "bezier (w/%d points):", nbezr);
                    for (n = 0; n < nbezr; n++) {
                        SPRINT4(2, "%5d %11.5f %11.5f %11.5f",   n, pts[3*n], pts[3*n+1], pts[3*n+2]);
                    }

                    if (iseg < sket->nseg-1 || iopen == 1) {
                        data[0] = pts[3*nbezr-3];
                        data[1] = pts[3*nbezr-2];
                        data[2] = pts[3*nbezr-1];
                        status = EG_makeTopology(MODL->context, NULL, NODE, 0,
                                                 data, 0, NULL, NULL, &(Enodes[nseg+1]));
                        CHECK_STATUS(EG_makeTopology);
                    } else {
                        Enodes[nseg+1] = Enodes[0];
                    }

                    header[0] = 0;
                    header[1] = nbezr - 1;
                    header[2] = nbezr;

                    /* make the BEZIER */
                    status = EG_makeGeometry(MODL->context, CURVE, BEZIER, NULL, header, pts, &ecurve);
                    CHECK_STATUS(EG_makeGeometry);

                    status = EG_getRange(ecurve, tdata, &periodic);
                    CHECK_STATUS(EG_getRange);

                    status = EG_makeTopology(MODL->context, ecurve, EDGE, TWONODE,
                                             tdata, 2, &(Enodes[nseg]), NULL, &(Eedges[nseg]));
                    CHECK_STATUS(EG_makeTopology);

                    nseg++;

                    nbezr = 0;

                    xlast = sket->x[iseg-1];
                    ylast = sket->y[iseg-1];
                    zlast = sket->z[iseg-1];
                }

                /* if the new segment is a skend, stop now */
                if (sket->itype[iseg] == OCSM_SKEND) {
                    break;

                /* add a linseg */
                } else if (sket->itype[iseg] == OCSM_LINSEG) {
                    if (fabs(xlast-sket->x[iseg]) < EPS06 &&
                        fabs(ylast-sket->y[iseg]) < EPS06 &&
                        fabs(zlast-sket->z[iseg]) < EPS06   ) {
                        SPRINT0(2, "skipped (zero length linseg)");
                    } else {
                        SPRINT0(2, "linseg:");
                        SPRINT4(2, "%5d %11.5f %11.5f %11.5f", 0,    xlast,         ylast,         zlast        );
                        SPRINT4(2, "%5d %11.5f %11.5f %11.5f", iseg, sket->x[iseg], sket->y[iseg], sket->z[iseg]);

                        if (iseg < sket->nseg-2 || iopen == 1) {
                            pts[0] = sket->x[iseg];
                            pts[1] = sket->y[iseg];
                            pts[2] = sket->z[iseg];
                            status = EG_makeTopology(MODL->context, NULL, NODE, 0,
                                                     pts, 0, NULL, NULL, &(Enodes[nseg+1]));
                            CHECK_STATUS(EG_makeTopology);
                        } else {
                            Enodes[nseg+1] = Enodes[0];
                        }

                        pts[0] = xlast;
                        pts[1] = ylast;
                        pts[2] = zlast;
                        pts[3] = sket->x[iseg] - xlast;
                        pts[4] = sket->y[iseg] - ylast;
                        pts[5] = sket->z[iseg] - zlast;
                        status = EG_makeGeometry(MODL->context, CURVE, LINE, NULL, NULL, pts, &ecurve);
                        CHECK_STATUS(EG_makeGeometry);

                        status = EG_invEvaluate(ecurve, pts, &(tdata[0]), result);
                        CHECK_STATUS(EG_invEvaluate);

                        pts[0] = sket->x[iseg];
                        pts[1] = sket->y[iseg];
                        pts[2] = sket->z[iseg];
                        status = EG_invEvaluate(ecurve, pts, &(tdata[1]), result);
                        CHECK_STATUS(EG_invEvaluate);

                        status = EG_makeTopology(MODL->context, ecurve, EDGE, TWONODE,
                                                 tdata, 2, &(Enodes[nseg]), NULL, &(Eedges[nseg]));
                        CHECK_STATUS(EG_makeTopology);

                        nseg++;
                    }

                    xlast = sket->x[iseg];
                    ylast = sket->y[iseg];
                    zlast = sket->z[iseg];

                /* add a cirarc */
                } else if (sket->itype[iseg] == OCSM_CIRARC) {
                    SPRINT0(2, "cirarc:");
                    SPRINT4(2, "%5d %11.5f %11.5f %11.5f", 0,      xlast,           ylast,           zlast          );
                    SPRINT4(2, "%5d %11.5f %11.5f %11.5f", iseg,   sket->x[iseg  ], sket->y[iseg  ], sket->z[iseg  ]);
                    SPRINT4(2, "%5d %11.5f %11.5f %11.5f", iseg+1, sket->x[iseg+1], sket->y[iseg+1], sket->z[iseg+1]);

                    if (iseg < sket->nseg-3 || iopen == 1) {
                        pts[0] = sket->x[iseg+1];
                        pts[1] = sket->y[iseg+1];
                        pts[2] = sket->z[iseg+1];
                        status = EG_makeTopology(MODL->context, NULL, NODE, 0,
                                                 pts, 0, NULL, NULL, &(Enodes[nseg+1]));
                        CHECK_STATUS(EG_makeTopology);
                    } else {
                        Enodes[nseg+1] = Enodes[0];
                    }

                    if        (fabs(xlast-sket->x[iseg]) < EPS06 && fabs(xlast-sket->x[iseg+1]) < EPS06) {
                        scent = ((sket->y[iseg+1] - ylast        ) * (sket->y[iseg  ] - sket->y[iseg+1])
                                -(sket->z[iseg+1] - zlast        ) * (sket->z[iseg+1] - sket->z[iseg  ]))
                               /((zlast           - sket->z[iseg]) * (sket->y[iseg  ] - sket->y[iseg+1])
                               - (sket->y[iseg]   - ylast        ) * (sket->z[iseg+1] - sket->z[iseg  ]));
                        ycent = (ylast + sket->y[iseg] + scent * (zlast         - sket->z[iseg])) / 2;
                        zcent = (zlast + sket->z[iseg] + scent * (sket->y[iseg] - ylast        )) / 2;

                        data[0] = xlast;
                        data[1] = ycent;
                        data[2] = zcent;
                        data[3] = 0;
                        data[4] = ylast - ycent;
                        data[5] = zlast - zcent;
                        data[6] = 0;
                        data[7] = -data[5];
                        data[8] = +data[4];
                        data[9] = sqrt(SQR(ylast-ycent) + SQR(zlast-zcent));

                    } else if (fabs(ylast-sket->y[iseg]) < EPS06 && fabs(ylast-sket->y[iseg+1]) < EPS06) {
                        scent = ((sket->z[iseg+1] - zlast        ) * (sket->z[iseg  ] - sket->z[iseg+1])
                                -(sket->x[iseg+1] - xlast        ) * (sket->x[iseg+1] - sket->x[iseg  ]))
                               /((xlast           - sket->x[iseg]) * (sket->z[iseg  ] - sket->z[iseg+1])
                               - (sket->z[iseg]   - zlast        ) * (sket->x[iseg+1] - sket->x[iseg  ]));
                        zcent = (zlast + sket->z[iseg] + scent * (xlast         - sket->x[iseg])) / 2;
                        xcent = (xlast + sket->x[iseg] + scent * (sket->z[iseg] - zlast        )) / 2;

                        data[0] = xcent;
                        data[1] = ylast;
                        data[2] = zcent;
                        data[3] = xlast - xcent;
                        data[4] = 0;
                        data[5] = zlast - zcent;
                        data[6] = +data[5];
                        data[7] = 0;
                        data[8] = -data[3];
                        data[9] = sqrt(SQR(zlast-zcent) + SQR(xlast-xcent));

                    } else if (fabs(zlast-sket->z[iseg]) < EPS06 && fabs(zlast-sket->z[iseg+1]) < EPS06) {
                        scent = ((sket->x[iseg+1] - xlast        ) * (sket->x[iseg  ] - sket->x[iseg+1])
                                -(sket->y[iseg+1] - ylast        ) * (sket->y[iseg+1] - sket->y[iseg  ]))
                               /((ylast           - sket->y[iseg]) * (sket->x[iseg  ] - sket->x[iseg+1])
                               - (sket->x[iseg]   - xlast        ) * (sket->y[iseg+1] - sket->y[iseg  ]));
                        xcent = (xlast + sket->x[iseg] + scent * (ylast         - sket->y[iseg])) / 2;
                        ycent = (ylast + sket->y[iseg] + scent * (sket->x[iseg] - xlast        )) / 2;

                        data[0] = xcent;
                        data[1] = ycent;
                        data[2] = zlast;
                        data[3] = xlast - xcent;
                        data[4] = ylast - ycent;
                        data[5] = 0;
                        data[6] = -data[4];
                        data[7] = +data[3];
                        data[8] = 0;
                        data[9] = sqrt(SQR(xlast-xcent) + SQR(ylast-ycent));

                    } else {
                        SET_STATUS(OCSM_NON_COPLANAR_SKETCH_POINTS, skend);
                    }

                    if (scent > 0) {
                        status = EG_makeGeometry(MODL->context, CURVE, CIRCLE, NULL, NULL, data, &ecurve);
                        CHECK_STATUS(EG_makeGeometry);
                    } else {
                        status = EG_makeGeometry(MODL->context, CURVE, CIRCLE, NULL, NULL, data, &eflip);
                        CHECK_STATUS(EG_makeGeometry);

                        status = EG_flipObject(eflip, &ecurve);
                        CHECK_STATUS(EG_flipObject);
                    }

                    data[0] = xlast;
                    data[1] = ylast;
                    data[2] = zlast;
                    status = EG_invEvaluate(ecurve, data, &(tdata[0]), result);
                    CHECK_STATUS(EG_invEvaluate);

                    data[0] = sket->x[iseg+1];
                    data[1] = sket->y[iseg+1];
                    data[2] = sket->z[iseg+1];
                    status = EG_invEvaluate(ecurve, data, &(tdata[1]), result);
                    CHECK_STATUS(EG_invEvaluate);

                    status = EG_makeTopology(MODL->context, ecurve, EDGE, TWONODE,
                                             tdata, 2, &(Enodes[nseg]), NULL, &(Eedges[nseg]));
                    CHECK_STATUS(EG_makeTopology);

                    nseg++;

                    iseg++;   /* this is because cirarc show up in pairs in iskpnt */

                    xlast = sket->x[iseg];
                    ylast = sket->y[iseg];
                    zlast = sket->z[iseg];

                /* initialize or add points to a spline (to be processed above in future
                   trip through this for loop) */
                } else if (sket->itype[iseg] == OCSM_SPLINE) {
                    if (nspln == 0) {
                        pts[0] = xlast;
                        pts[1] = ylast;
                        pts[2] = zlast;

                        nspln = 1;
                    }

                    pts[3*nspln  ] = sket->x[iseg];
                    pts[3*nspln+1] = sket->y[iseg];
                    pts[3*nspln+2] = sket->z[iseg];

                    nspln++;

                /* initialize or add points to a bezier (to be processed above in future
                   trip through this for loop) */
                } else if (sket->itype[iseg] == OCSM_BEZIER) {
                    if (nbezr == 0) {
                        pts[0] = xlast;
                        pts[1] = ylast;
                        pts[2] = zlast;

                        nbezr = 1;
                    }

                    pts[3*nbezr  ] = sket->x[iseg];
                    pts[3*nbezr+1] = sket->y[iseg];
                    pts[3*nbezr+2] = sket->z[iseg];

                    nbezr++;
                }
            }

            /* close the Sketch */
            for (iseg = 0; iseg < nseg; iseg++) {
                sense[iseg] = SFORWARD;
            }

            if (wireonly == 1 || iopen == 1) {
                if (iopen == 1) {
                    status = EG_makeTopology(MODL->context, NULL, LOOP, OPEN,
                                             NULL, nseg, Eedges, sense, &eloop);
                    CHECK_STATUS(EG_makeTopology);
                } else {
                    status = EG_makeTopology(MODL->context, NULL, LOOP, CLOSED,
                                             NULL, nseg, Eedges, sense, &eloop);
                    CHECK_STATUS(EG_makeTopology);
                }

                status = EG_makeTopology(MODL->context, NULL, BODY, WIREBODY,
                                         NULL, 1, &eloop, NULL, &ebody);
                CHECK_STATUS(EG_makeTopology);
            } else {
                status = EG_makeTopology(MODL->context, NULL, LOOP, CLOSED,
                                         NULL, nseg, Eedges, sense, &eloop);
                CHECK_STATUS(EG_makeTopology);

                status = EG_makeFace(eloop, SFORWARD, NULL, Efaces);

                if (status == EGADS_GEOMERR) {
                    /* check that Sketch is planar */
                    if (fabs(xmax-xmin) > EPS06 &&
                        fabs(ymax-ymin) > EPS06 &&
                        fabs(zmax-zmin) > EPS06   ) {
                        SET_STATUS(OCSM_CLOSED_SKETCH_NOT_PLANAR, skend);
                    }
                }

                CHECK_STATUS(EG_makeFace);

                status = EG_getRange(Efaces[0], range, &periodic);
                CHECK_STATUS(EG_getRange);

                range[0] = (range[0] + range[1]) / 2;
                range[1] = (range[2] + range[3]) / 2;

                status = EG_evaluate(Efaces[0], range, eval);
                CHECK_STATUS(EG_evaluate);

                norm[0] = eval[4] * eval[8] - eval[5] * eval[7];
                norm[1] = eval[5] * eval[6] - eval[3] * eval[8];
                norm[2] = eval[3] * eval[7] - eval[4] * eval[6];

                if        (fabs(norm[0]) >= fabs(norm[1]) &&
                           fabs(norm[0]) >= fabs(norm[2])   ) {
                    if (norm[0] < 0) {
                        status = EG_flipObject(Efaces[0], &enew);
                        CHECK_STATUS(EG_flipObject);
                        Efaces[0] = enew;
                    }
                } else if (fabs(norm[1]) >= fabs(norm[2]) &&
                           fabs(norm[1]) >= fabs(norm[0])   ) {
                    if (norm[1] < 0) {
                        status = EG_flipObject(Efaces[0], &enew);
                        CHECK_STATUS(EG_flipObject);
                        Efaces[0] = enew;
                    }
                } else {
                    if (norm[2] < 0) {
                        status = EG_flipObject(Efaces[0], &enew);
                        CHECK_STATUS(EG_flipObject);
                        Efaces[0] = enew;
                    }
                }

                status = EG_makeTopology(MODL->context, NULL, BODY, FACEBODY,
                                         NULL, 1, Efaces, NULL, &ebody);
                CHECK_STATUS(EG_makeTopology);
            }

            #if   defined(SHOW_SPLINES)
            {
                #define NPER          251
                #define MAX_PLOT_PTS 9999

                int    nplot, nline, ilin[MAX_SKETCH_SIZE+1], isym[MAX_SKETCH_SIZE+1], nper[MAX_SKETCH_SIZE+1];
                int    io_kbd=5, io_scr=6, indgr=1+2+4+16+64;
                float  xplot[MAX_PLOT_PTS], yplot[MAX_PLOT_PTS];
                char   pltitl[30], title[] = "Spline  (skbeg=^, linseg=+, cirarc=o, spline=x)";

                int    j, periodic, oclass, mtype, nchild, *senses;
                double trange[4], tt;
                ego    eref, *ebodys;

                if        (xmin == xmax) {
                    snprintf(pltitl, 30, "~y~z~Sketch at x=%10.5f", xmin);
                } else if (ymin == ymax) {
                    snprintf(pltitl, 30, "~z~x~Sketch at y=%10.5f", ymin);
                } else {
                    snprintf(pltitl, 30, "~x~y~Sketch at z=%10.5f", zmin);
                }

                /* put Sketch points onto plot */
                nline = 0;
                nplot = 0;
                for (i = 0; i < sket->nseg; i++) {
                    ilin[nline] = 0;
                    nper[nline] = 1;

                    if        (sket->itype[i] == OCSM_SKBEG) {
                        isym[nline] = GR_TRIANGLE;
                    } else if (sket->itype[i] == OCSM_LINSEG) {
                        isym[nline] = GR_PLUS;
                    } else if (sket->itype[i] == OCSM_CIRARC) {
                        isym[nline] = GR_CIRCLE;
                    } else if (sket->itype[i] == OCSM_SPLINE) {
                        isym[nline] = GR_X;
                    } else if (sket->itype[i] == OCSM_BEZIER) {
                        isym[nline] = GR_X;
                    }
                    nline++;

                    if        (xmin == xmax) {
                        xplot[nplot] = sket->y[i];
                        yplot[nplot] = sket->z[i];
                        nplot++;
                    } else if (ymin == ymax) {
                        xplot[nplot] = sket->z[i];
                        yplot[nplot] = sket->x[i];
                        nplot++;
                    } else {
                        xplot[nplot] = sket->x[i];
                        yplot[nplot] = sket->y[i];
                        nplot++;
                    }
                }

                /* put Sketch evaluations onto plot */
                status = EG_getTopology(eloop, &eref, &oclass, &mtype,
                                        data, &nchild, &ebodys, &senses);
                CHECK_STATUS(EG_getTopology);

                if (nplot+nchild*NPER > MAX_PLOT_PTS) {
                    signalError(MODL, OCSM_INTERNAL_ERROR, ibrch,
                                "nplot=%d, nchild=%d, NPER=%d, MAX_PLOT_PTS=%d", nplot, nchild, NPER, MAX_PLOT_PTS);
                    SET_STATUS(OCSM_INTERNAL_ERROR, ShowSketch);
                }

                for (j = 0; j < nchild; j++) {
                    status = EG_getRange(ebodys[j], trange, &periodic);
                    CHECK_STATUS(EG_getRange);

                    for (i = 0; i < NPER; i++) {
                        tt = trange[0] + (double)(i) / (double)(NPER-1) * (trange[1] - trange[0]);

                        status = EG_evaluate(ebodys[j], &tt, data);
                        CHECK_STATUS(EG_evaluate);

                        if        (xmin == xmax) {
                            xplot[nplot] = data[1];
                            yplot[nplot] = data[2];
                            nplot++;
                        } else if (ymin == ymax) {
                            xplot[nplot] = data[2];
                            yplot[nplot] = data[0];
                            nplot++;
                        } else {
                            xplot[nplot] = data[0];
                            yplot[nplot] = data[1];
                            nplot++;
                        }
                    }

                    ilin[nline] =  +1;
                    isym[nline] =  -1;
                    nper[nline] = NPER;
                    nline++;
                }

                /* plot */
                grinit_(&io_kbd, &io_scr, title, strlen(title));
                grline_(ilin, isym, &nline, pltitl, &indgr, xplot, yplot, nper, strlen(pltitl));
            }
            #endif

            /* remove ::x[], ::y[], ::z[], and ::d[] if they were generated by a skvar statement */
            if (sket->nvar > 0) {
                if (strcmp(MODL->pmtr[MODL->npmtr  ].name, "::d") != 0 ||
                    strcmp(MODL->pmtr[MODL->npmtr-1].name, "::z") != 0 ||
                    strcmp(MODL->pmtr[MODL->npmtr-2].name, "::y") != 0 ||
                    strcmp(MODL->pmtr[MODL->npmtr-3].name, "::x") != 0   ) {
                    signalError(MODL, OCSM_INTERNAL_ERROR, ibrch,
                                "last 4 Parameters are not ::d, ::z, ::y, and ::x");
                    SET_STATUS(OCSM_INTERNAL_ERROR, skend);
                }

                for (i = 0; i < 4; i++) {
                    FREE(MODL->pmtr[MODL->npmtr].name );
                    FREE(MODL->pmtr[MODL->npmtr].value);
                    FREE(MODL->pmtr[MODL->npmtr].dot  );
                    FREE(MODL->pmtr[MODL->npmtr].lbnd );
                    FREE(MODL->pmtr[MODL->npmtr].ubnd );
                    MODL->npmtr--;
                }
            }

            /* close the Sketch */
            sket->type   = 0;
            sket->size   = 0;
            sket->solved = 1;
            sket->nseg   = 0;
            sket->nvar   = 0;
            sket->ncon   = 0;

            /* make a new WireBody */
            if (iopen == 1 || wireonly == 1) {
                status = newBody(MODL, ibrch, OCSM_SKEND, MODL->nbody, -1,
                                 args, hasdots, OCSM_WIRE_BODY, &ibody);
                CHECK_STATUS(newBody);
            } else {
                status = newBody(MODL, ibrch, OCSM_SKEND, MODL->nbody, -1,
                                 args, hasdots, OCSM_SHEET_BODY, &ibody);
                CHECK_STATUS(newBody);
            }

            MODL->body[ibody].ebody = ebody;

            /* mark the new Faces (if any) with the current Branch */
            status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
            CHECK_STATUS(EG_getBodyTopos);

            for (iface = 1; iface <= nface; iface++) {
                status = setFaceAttribute(MODL, ibody, iface, 0, 0, npatn, patn);
                CHECK_STATUS(setFaceAttribute);
            }

            EG_free(efaces);

            status = EG_getBodyTopos(ebody, NULL, EDGE, &nedge, &eedges);
            CHECK_STATUS(EG_getBodyTopos);

            EG_free(eedges);

            /* finish the Body */
            status = finishBody(MODL, ibody);
            CHECK_STATUS(finishBody);

            /* push the Sketch onto the stack */
            stack[(*nstack)++] = ibody;

            SPRINT2(1, "                          Sketch %4d created with %d Edges",
                    ibody, nedge);
        }
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   buildSolver - implement OCSM_SOLVERs for ocsmBuild                 *
 *                                                                      *
 ************************************************************************
 */

static int
buildSolver(modl_T *modl,
            int    ibrch,
/*@unused@*/varg_T args[],
            int    *ncon,
            int    solcons[])
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       jbrch, type, iter, niter, ipmtr, jpmtr, i, j, ivar, icon;
    double    neg_f0[MAX_SOLVER_SIZE], dfdx[MAX_SOLVER_SIZE*MAX_SOLVER_SIZE];
    double    delx[MAX_SOLVER_SIZE], f0max, f0last, save_value, omega;
    double    value, dot;
    char      name[MAX_EXPR_LEN];

    static int nvar;
    static int solvars[MAX_SOLVER_SIZE];

    ROUTINE(buildSolver);
    DPRINT2("%s(ibrch=%d) {",
            routine, ibrch);

    /* --------------------------------------------------------------- */

    type = MODL->brch[ibrch].type;

    /* execute: "solbeg $varList" */
    if (type == OCSM_SOLBEG) {
        SPRINT2(1, "    executing [%4d] solbeg:         %s",
                ibrch, args[1].str);

        /* initialize the number of solver constraints and variables */
        *ncon = 0;
        nvar  = 0;

        /* make a list of the solver variables by parsing arg1 */
        name[j=0] = '\0';
        for (i = 1; i < strlen(MODL->brch[ibrch].arg1); i++) {
            if (MODL->brch[ibrch].arg1[i] != ';') {
                name[j  ] = MODL->brch[ibrch].arg1[i];
                name[j+1] = '\0';
                j++;
            } else {
                ipmtr = 0;
                for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
                    if (strcmp(MODL->pmtr[jpmtr].name, name) == 0            &&
                               MODL->pmtr[jpmtr].level       == MODL->level  &&
                               MODL->pmtr[jpmtr].type        == OCSM_INTERNAL   ) {
                        ipmtr = jpmtr;
                        break;
                    }
                }

                if (ipmtr == 0) {
                    SPRINT1(0, "WARNING:: name \"%s\" not an INTERNAL parameter", name);
                    status = OCSM_NAME_NOT_FOUND;
                    goto cleanup;
                }

                solvars[nvar++] = ipmtr;

                if (nvar > MAX_SOLVER_SIZE) {
                    SET_STATUS(OCSM_TOO_MANY_SOLVER_VARS, solbeg);
                }

                name[j=0] = '\0';
            }
        }

    /* execute: "solcon expr" */
    } else if (type == OCSM_SOLCON) {
        SPRINT2(1, "    executing [%4d] solcon:         %s",
                ibrch, args[1].str);

        solcons[(*ncon)++] = ibrch;

        if (*ncon > MAX_SOLVER_SIZE) {
            SET_STATUS(OCSM_TOO_MANY_SOLVER_VARS, solcon);
        }

    /* execute: "solend" */
    } else if (type == OCSM_SOLEND) {
        SPRINT1(1, "    executing [%4d] solend:",
                ibrch);

        for (ivar = 0; ivar < nvar; ivar++) {
            jpmtr = solvars[ivar];
            SPRINT3(2, "        var[%2d] = %3d [%s]",
                    ivar, jpmtr, MODL->pmtr[jpmtr].name);
        }
        for (icon = 0; icon < *ncon; icon++) {
            jbrch = solcons[icon];
            SPRINT3(2, "        con[%2d] = %3d [%s]",
                    icon, jbrch, &(MODL->brch[jbrch].arg1[1]));
        }

        /* if the number of constraints does not match the number of
           Sketch Parameters, return an error */
        if        (*ncon < nvar) {
            SET_STATUS(OCSM_UNDERCONSTRAINED, solend);
        } else if (*ncon > nvar) {
            SET_STATUS(OCSM_OVERCONSTRAINED, solend);
        }

        /* if there were no constraints, simply return */
        if (*ncon == 0) {
            goto cleanup;
        }

        /* Newton iteration to change the solver variables until
           the constraints are satisfied */
        niter  = 100;
        omega  = 0.50;
        f0last = 0;
        for (iter = 0; iter < niter; iter++) {

            /* evaluate the constraints */
            f0max = 0;
            for (icon = 0; icon < *ncon; icon++) {
                jbrch  = solcons[icon];
                status = str2val(&(MODL->brch[jbrch].arg1[1]), MODL, &value, &dot);
                CHECK_STATUS(str2val);

                neg_f0[icon] = -value;
                SPRINT2(2,"        f0[%4d] = %11.4e", jbrch, value);

                if (fabs(value) > f0max) {
                    f0max = fabs(value);
                }
            }
            SPRINT2(1, "    -> solving Sketch: iter = %3d,   f0max = %12.4e", iter, f0max);

            /* if we have converged, stop the Newton iterations */
            if (f0max < EPS06) {
                DPRINT0("converged");
                break;
            }

            /* f0max < f0last, we are converging, so increase omega */
            if (f0max < f0last) {
                omega = MIN(1.2*omega, 1);
            }
            f0last = f0max;

            /* build up the Jacobian matrix by perturbing the solver variables
               one at a time */
            for (ivar = 0; ivar < nvar; ivar++) {
                jpmtr = solvars[ivar];

                save_value = MODL->pmtr[jpmtr].value[0];
                MODL->pmtr[jpmtr].value[0] += EPS06;

                for (icon = 0; icon < *ncon; icon++) {
                    jbrch = solcons[icon];
                    status = str2val(&(MODL->brch[jbrch].arg1[1]), MODL, &value, &dot);
                    CHECK_STATUS(str2val);

                    dfdx[icon*(*ncon)+ivar] = (value + neg_f0[icon]) / EPS06;
                }

                MODL->pmtr[jpmtr].value[0] = save_value;
            }

            /* print out the Jacobian matrix */
            DPRINT0("Jacobian matrix");
            for (icon = 0; icon < *ncon; icon++) {
                DPRINT1x("%3d: ", icon);

                for (ivar = 0; ivar < nvar; ivar++) {
                    DPRINT1x("%8.4f ", dfdx[icon*(*ncon)+ivar]);
                }
                DPRINT0(" ");
            }

            /* take the Newton step */
            status = matsol(dfdx, neg_f0, *ncon, delx);
            CHECK_STATUS(matsol);

            for (ivar = 0; ivar < nvar; ivar++) {
                jpmtr = solvars[ivar];
                MODL->pmtr[jpmtr].value[0] += omega * delx[ivar];
            }
        }

        /* now that we have run out of iterations, check for convergence */
        if (f0max > EPS06) {
            status = OCSM_NOT_CONVERGED;
            goto cleanup;
        }

        /* initialize variables for next solver */
        nvar  = 0;
        *ncon = 0;
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   buildTransform - implemet OCSM_TRANSFORMs for ocsmBuild            *
 *                                                                      *
 ************************************************************************
 */

static int
buildTransform(modl_T *modl,
               int    ibrch,
               varg_T args[],
               int    *nstack,
               int    stack[])
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int        type, hasdots, ibody, ibodyl, nloop, nface, nedge;
    int        oclass, mtype, *senses, ishift, iswap, i, j;
    double     toler, matrix[12], data[4];
    double     cosx, cosy, cosz, sinx, siny, sinz, fact;
    double     dx, dy, dz, nx, ny, nz, dist;

    ego        ebody, ebodyl, exform, *eloops, eloop, eref, eswap;
    ego        *eedges, eface, esurf, *efaces, eshell;

    ROUTINE(buildTransform);
    DPRINT2("%s(ibrch=%d) {",
            routine, ibrch);

    /* --------------------------------------------------------------- */

    type    = MODL->brch[ibrch].type;
    hasdots = 0;

    /* execute: "translate dx dy dz" */
    if (type == OCSM_TRANSLATE) {
        SPRINT4(1, "    executing [%4d] translate:  %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0) {
            hasdots = 1;

            SPRINT3(1, "                                 %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0]);
        }

        /* pop a Body from the stack */
        if ((*nstack) < 1) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "TRANSLATE expects a Body on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, translate);
        } else {
            ibodyl = stack[--(*nstack)];
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_TRANSLATE, ibodyl, -1,
                         args, hasdots, MODL->body[ibodyl].botype, &ibody);
        CHECK_STATUS(newBody);

        /* apply the new transformation matrix */
        dx = args[1].val[0];
        dy = args[2].val[0];
        dz = args[3].val[0];

        ebodyl = MODL->body[ibodyl].ebody;

        matrix[ 0] = 1; matrix[ 1] = 0; matrix[ 2] = 0; matrix[ 3] = dx;
        matrix[ 4] = 0; matrix[ 5] = 1; matrix[ 6] = 0; matrix[ 7] = dy;
        matrix[ 8] = 0; matrix[ 9] = 0; matrix[10] = 1; matrix[11] = dz;

        status = EG_makeTransform(MODL->context, matrix, &exform);
        CHECK_STATUS(EG_makeTransform);

        status = EG_copyObject(ebodyl, exform, &ebody);
        CHECK_STATUS(EG_copyObject);

        status = EG_deleteObject(exform);
        CHECK_STATUS(EG_deleteObject);

        MODL->body[ibody].ebody = ebody;

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "rotatex angDeg yaxis zaxis" */
    } else if (type == OCSM_ROTATEX) {
        SPRINT4(1, "    executing [%4d] rotatex:    %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0) {
            hasdots = 1;

            SPRINT3(1, "                                 %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0]);
        }

        /* pop a Body from the stack */
        if ((*nstack) < 1) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "ROTATEX expects a Body on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, rotatex);
        } else {
            ibodyl = stack[--(*nstack)];
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_ROTATEX, ibodyl, -1,
                         args, hasdots, MODL->body[ibodyl].botype, &ibody);
        CHECK_STATUS(newBody);

        /* apply the new transformation matrix */
        cosx = cos(args[1].val[0] * PIo180);
        sinx = sin(args[1].val[0] * PIo180);
        dy   =     args[2].val[0];
        dz   =     args[3].val[0];

        ebodyl = MODL->body[ibodyl].ebody;

        matrix[ 0] = 1; matrix[ 1] = 0;    matrix[ 2] = 0;    matrix[ 3] = 0;
        matrix[ 4] = 0; matrix[ 5] = cosx; matrix[ 6] =-sinx; matrix[ 7] = dy - dy * cosx + dz * sinx;
        matrix[ 8] = 0; matrix[ 9] = sinx; matrix[10] = cosx; matrix[11] = dz - dy * sinx - dz * cosx;

        status = EG_makeTransform(MODL->context, matrix, &exform);
        CHECK_STATUS(EG_makeTransform);

        status = EG_copyObject(ebodyl, exform, &ebody);
        CHECK_STATUS(EG_copyObject);

        status = EG_deleteObject(exform);
        CHECK_STATUS(EG_deleteObject);

        MODL->body[ibody].ebody = ebody;

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "rotatey angDeg zaxis xaxis" */
    } else if (type == OCSM_ROTATEY) {
        SPRINT4(1, "    executing [%4d] rotatey:    %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0) {
            hasdots = 1;

            SPRINT3(1, "                                 %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0]);
        }

        /* pop a Body from the stack */
        if ((*nstack) < 1) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "ROTATEY expects a Body on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, rotatey);
        } else {
            ibodyl = stack[--(*nstack)];
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_ROTATEY, ibodyl, -1,
                         args, hasdots, MODL->body[ibodyl].botype, &ibody);
        CHECK_STATUS(newBody);

        /* apply the new transformation matrix */
        cosy = cos(args[1].val[0] * PIo180);
        siny = sin(args[1].val[0] * PIo180);
        dz   =     args[2].val[0];
        dx   =     args[3].val[0];

        ebodyl = MODL->body[ibodyl].ebody;

        matrix[ 0] = cosy; matrix[ 1] = 0; matrix[ 2] = siny; matrix[ 3] = dx - dz * siny - dx * cosy;
        matrix[ 4] = 0;    matrix[ 5] = 1; matrix[ 6] = 0;    matrix[ 7] = 0;
        matrix[ 8] =-siny; matrix[ 9] = 0; matrix[10] = cosy; matrix[11] = dz - dz * cosy + dx * siny;

        status = EG_makeTransform(MODL->context, matrix, &exform);
        CHECK_STATUS(EG_makeTransform);

        status = EG_copyObject(ebodyl, exform, &ebody);
        CHECK_STATUS(EG_copyObject);

        status = EG_deleteObject(exform);
        CHECK_STATUS(EG_deleteObject);

        MODL->body[ibody].ebody = ebody;

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "rotatez angDeg xaxis yaxis" */
    } else if (type == OCSM_ROTATEZ) {
        SPRINT4(1, "    executing [%4d] rotatez:    %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0) {
            hasdots = 1;

            SPRINT3(1, "                                 %11.5f %11.5f %11.5f",
                       args[1].dot[0], args[2].dot[0], args[3].dot[0]);
        }

        /* pop a Body from the stack */
        if ((*nstack) < 1) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "ROTATEZ expects a Body on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, rotatez);
        } else {
            ibodyl = stack[--(*nstack)];
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_ROTATEZ, ibodyl, -1,
                         args, hasdots, MODL->body[ibodyl].botype, &ibody);
        CHECK_STATUS(newBody);

        /* apply the new transformation matrix */
        cosz = cos(args[1].val[0] * PIo180);
        sinz = sin(args[1].val[0] * PIo180);
        dx   =     args[2].val[0];
        dy   =     args[3].val[0];

        ebodyl = MODL->body[ibodyl].ebody;

        matrix[ 0] = cosz; matrix[ 1] =-sinz; matrix[ 2] = 0; matrix[ 3] = dx - dx * cosz + dy * sinz;
        matrix[ 4] = sinz; matrix[ 5] = cosz; matrix[ 6] = 0; matrix[ 7] = dy - dx * sinz - dy * cosz;
        matrix[ 8] = 0;    matrix[ 9] = 0;    matrix[10] = 1; matrix[11] = 0;

        status = EG_makeTransform(MODL->context, matrix, &exform);
        CHECK_STATUS(EG_makeTransform);

        status = EG_copyObject(ebodyl, exform, &ebody);
        CHECK_STATUS(EG_copyObject);

        status = EG_deleteObject(exform);
        CHECK_STATUS(EG_deleteObject);

        MODL->body[ibody].ebody = ebody;

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "scale fact" */
    } else if (type == OCSM_SCALE) {
        SPRINT2(1, "    executing [%4d] scale:      %11.5f",
                ibrch, args[1].val[0]);
        if (args[1].dot[0] != 0) {
            hasdots = 1;

            SPRINT1(1, "                                 %11.5f",
                       args[1].dot[0]);
        }

        /* pop a Body from the stack */
        if ((*nstack) < 1) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "SCALE expects a Body on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, scale);
        } else {
            ibodyl = stack[--(*nstack)];
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_SCALE, ibodyl, -1,
                         args, hasdots, MODL->body[ibodyl].botype, &ibody);
        CHECK_STATUS(newBody);

        /* apply the new transformation matrix */
        fact = args[1].val[0];

        ebodyl = MODL->body[ibodyl].ebody;

        matrix[ 0] = fact; matrix[ 1] = 0;    matrix[ 2] = 0;    matrix[ 3] = 0;
        matrix[ 4] = 0;    matrix[ 5] = fact; matrix[ 6] = 0;    matrix[ 7] = 0;
        matrix[ 8] = 0;    matrix[ 9] = 0;    matrix[10] = fact; matrix[11] = 0;

        status = EG_makeTransform(MODL->context, matrix, &exform);
        CHECK_STATUS(EG_makeTransform);

        status = EG_copyObject(ebodyl, exform, &ebody);
        CHECK_STATUS(EG_copyObject);

        status = EG_deleteObject(exform);
        CHECK_STATUS(EG_deleteObject);

        MODL->body[ibody].ebody = ebody;

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "mirror nx ny nz dist=0" */
    } else if (type == OCSM_MIRROR) {
        SPRINT5(1, "    executing [%4d] mirror:     %11.5f %11.5f %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0], args[3].val[0], args[4].val[0]);
        if (args[1].dot[0] != 0 || args[2].dot[0] != 0 || args[3].dot[0] != 0 || args[4].dot[0] != 0) {
            hasdots = 1;

            SPRINT4(1, "                                 %11.5f %11.5f %11.5f %11.5f",
                    args[1].dot[0], args[2].dot[0], args[3].dot[0], args[4].dot[0]);
        }

        /* pop a Body from the stack */
        if ((*nstack) < 1) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "MIRROR expects a Body on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, mirror);
        } else {
            ibodyl = stack[--(*nstack)];
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_MIRROR, ibodyl, -1,
                         args, hasdots, MODL->body[ibodyl].botype, &ibody);
        CHECK_STATUS(newBody);

        /* apply the new transformation matrix */
        nx   = args[1].val[0];
        ny   = args[2].val[0];
        nz   = args[3].val[0];
        dist = args[4].val[0];

        fact  = sqrt(nx*nx + ny*ny + nz*nz);
        dx    = nx / fact;
        dy    = ny / fact;
        dz    = nz / fact;

        ebodyl = MODL->body[ibodyl].ebody;

        matrix[ 0] = 1-2*dx*dx;    matrix[ 1] =  -2*dy*dx;    matrix[ 2] =  -2*dz*dx;    matrix[ 3] = 2*dist*dx;
        matrix[ 4] =  -2*dx*dy;    matrix[ 5] = 1-2*dy*dy;    matrix[ 6] =  -2*dz*dy;    matrix[ 7] = 2*dist*dy;
        matrix[ 8] =  -2*dx*dz;    matrix[ 9] =  -2*dy*dz;    matrix[10] = 1-2*dz*dz;    matrix[11] = 2*dist*dz;

        status = EG_makeTransform(MODL->context, matrix, &exform);
        CHECK_STATUS(EG_makeTransform);

        status = EG_copyObject(ebodyl, exform, &ebody);
        CHECK_STATUS(EG_copyObject);

        status = EG_deleteObject(exform);
        CHECK_STATUS(EG_deleteObject);

        MODL->body[ibody].ebody = ebody;

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);

    /* execute: "reorder ishift iflip=0" */
    } else if (type == OCSM_REORDER) {
        SPRINT3(1, "    executing [%4d] reorder:    %11.5f %11.5f",
                ibrch, args[1].val[0], args[2].val[0]);

        /* pop a Body from the stack */
        if ((*nstack) < 1) {
            signalError(MODL, OCSM_INSUFFICIENT_BODYS_ON_STACK, ibrch,
                        "REORDER expects a Body on the stack");
            SET_STATUS(OCSM_INSUFFICIENT_BODYS_ON_STACK, reorder);
        } else {
            ibodyl = stack[--(*nstack)];
        }

        /* recycle old Body if not dirty */
        status = recycleBody(MODL, ibrch, type, args, hasdots);
        CHECK_STATUS(recycleBody);

        if (status == 1) {
            stack[(*nstack)++] = MODL->nbody;
            status = 0;
            goto cleanup;
        }

        /* make sure that ibodyl is either a WIRE or SHEET */
        if (MODL->body[ibodyl].botype != OCSM_WIRE_BODY &&
            MODL->body[ibodyl].botype != OCSM_SHEET_BODY  ) {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                        "REORDER needs a WIRE or SHEET");
            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, reorder);
        }

        ebodyl = MODL->body[ibodyl].ebody;

        /* make sure that the Body has a single Loop */
        status = EG_getBodyTopos(ebodyl, NULL, LOOP, &nloop, &eloops);
        CHECK_STATUS(EG_getBodyTopos);

        eloop = eloops[0];
        EG_free(eloops);

        if (nloop != 1) {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                        "REORDER requires one Loop");
            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, reorder);
        }

        /* create the Body */
        status = newBody(MODL, ibrch, OCSM_REORDER, ibodyl, -1,
                         args, hasdots, MODL->body[ibodyl].botype, &ibody);
        CHECK_STATUS(newBody);

        /* get the Edges and senses in the Loop */
        status = EG_getTopology(eloop, &eref, &oclass, &mtype, data,
                                &nedge, &eedges, &senses);
        CHECK_STATUS(EG_getTopology);

        /* shift the Edges in the Loop */
        if (mtype != CLOSED && NINT(args[1].val[0]) != 0) {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, ibrch,
                        "only closed Loops can be shifted");
            SET_STATUS(OCSM_WRONG_TYPES_ON_STACK, reorder);
        }

        if (args[1].val[0] > 0) {
            for (ishift = 0; ishift < NINT(args[1].val[0]); ishift++) {
                eswap = eedges[0];
                iswap = senses[0];
                for (i = 0; i < nedge-1; i++) {
                    eedges[i] = eedges[i+1];
                    senses[i] = senses[i+1];
                }
                eedges[nedge-1] = eswap;
                senses[nedge-1] = iswap;
            }
        } else if (args[1].val[0] < 0) {
            for (ishift = 0; ishift < -NINT(args[1].val[0]); ishift++) {
                eswap = eedges[nedge-1];
                iswap = senses[nedge-1];
                for (i = nedge-1; i > 0; i--) {
                    eedges[i] = eedges[i-1];
                    senses[i] = senses[i-1];
                }
                eedges[0] = eswap;
                senses[0] = iswap;
            }
        }

        /* flip the Loop */
        if (NINT(args[2].val[0]) != 0) {
            for (i = 0, j = nedge-1; i < j; i++, j--) {
                eswap     = eedges[i];
                iswap     = senses[i];
                eedges[i] = eedges[j];
                senses[i] = senses[j];
                eedges[j] = eswap;
                senses[j] = iswap;
            }

            for (i = 0; i < nedge; i++) {
                senses[i] = -senses[i];
            }
        }

        status = EG_makeTopology(MODL->context, NULL, LOOP, mtype,
                                 NULL, nedge, eedges, senses, &eloop);
        CHECK_STATUS(EG_makeTopology);

        /* make either WIRE or SHEET Body */
        if (MODL->body[ibodyl].botype == OCSM_WIRE_BODY) {
            status = EG_makeTopology(MODL->context, NULL, BODY, WIREBODY,
                                     NULL, 1, &eloop, NULL, &ebody);
            CHECK_STATUS(EG_makeTopology);
        } else {
            status = EG_getTopology(ebodyl, &esurf, &oclass, &mtype, data,
                                    &nface, &efaces, &senses);
            CHECK_STATUS(EG_getTopology);

            status = EG_getTopology(efaces[0], &esurf, &oclass, &mtype, data,
                                    &nloop, &eloops, &senses);
            CHECK_STATUS(EG_getTopology);

            if (NINT(args[2].val[0]) != 0) {
                mtype = -mtype;
            }

            status = EG_makeTopology(MODL->context, esurf, FACE, mtype,
                                     NULL, 1, &eloop, senses, &eface);
            CHECK_STATUS(EG_makeTopology);

            status = EG_attributeDup(efaces[0], eface);
            CHECK_STATUS(EG_attributeDup);

            status = EG_makeTopology(MODL->context, NULL, SHELL, OPEN,
                                     NULL, 1, &eface, NULL, &eshell);
            CHECK_STATUS(EG_makeTopology);

            status = EG_makeTopology(MODL->context, NULL, BODY, SHEETBODY,
                                     NULL, 1, &eshell, NULL, &ebody);
            CHECK_STATUS(EG_makeTopology);
        }

        MODL->body[ibody].ebody = ebody;

        /* finish the Body */
        status = finishBody(MODL, ibody);
        CHECK_STATUS(finishBody);

        /* push the Body onto the stack */
        stack[(*nstack)++] = ibody;

        status = getBodyTolerance(MODL->body[ibody].ebody, &toler);
        CHECK_STATUS(getBodyTolerance);

        SPRINT2(1, "                          Body   %4d created  (toler=%11.4e)",
                ibody, toler);
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   createPerturbation - create perturbed body for finite differencing *
 *                                                                      *
 ************************************************************************
 */

int
createPerturbation(modl_T *MODL)        /* (in)  pointer to base MODL */
{
    int       status = 0;               /* (out) return status */

    int       ibody, nerror, itime, ntime=10;
    int       ipmtr, irow, icol, irc, builtTo, nbody_ptrb;
    modl_T    *PTRB = NULL;
    ego       newBody;

    void      *modl_ptrb=NULL;

    ROUTINE(createPurturbation);
    DPRINT1("%s() {",
            routine);

    /* --------------------------------------------------------------- */

    /* if there is no tessellation for any Body on the stack, return error */
    for (ibody = 1; ibody <= MODL->nbody; ibody++) {
        if (MODL->body[ibody].onstack == 0) continue;

        if (MODL->body[ibody].etess == NULL) {
            status = OCSM_NEED_TESSELLATION;
            goto cleanup;
        }
    }

    /* if we found an error in a previous attempt to create a perturbation,
       return immediately (note: set MODL->dtime=0 before call) */
    if (MODL->dtime == -2) {
        goto cleanup;
    }

    /* if perturb exists, we may be able to return immediately */
    if (MODL->perturb != NULL) {
        PTRB = MODL->perturb;

        /* map base ibody's tessellation onto the ptrb Body */
        for (ibody = 1; ibody <= MODL->nbody; ibody++) {
            if (MODL->body[ibody].etess != NULL) {
                if (PTRB->body[ibody].etess == NULL) {
                    status = EG_mapBody(MODL->body[ibody].ebody, PTRB->body[ibody].ebody,
                                        "faceID", &newBody);
                    if (status == SUCCESS && newBody != NULL) {
                        EG_deleteObject(PTRB->body[ibody].ebody);
                        PTRB->body[ibody].ebody = newBody;
                    }

                    status = EG_mapTessBody(MODL->body[ibody].etess, PTRB->body[ibody].ebody,
                                            &(PTRB->body[ibody].etess));
                    CHECK_STATUS(EG_mapTessBody);
                }
            }
        }

        goto cleanup;
    }

    /* default initial time step */
    MODL->dtime = 0.001;

    /* try up to ntime increasingly smaller time steps */
    for (itime = 0; itime < ntime; itime++) {
        SPRINT1(1, "    MODL->dtime=%f\n", MODL->dtime);
        nerror = 0;

        /* make a copy and set its driving parameters to "base + dtime * dot" */
        status = ocsmCopy(MODL, &modl_ptrb);
        CHECK_STATUS(ocsmCopy);

        PTRB = modl_ptrb;

        MODL->perturb  = PTRB;
        PTRB->basemodl = MODL;

        for (ipmtr = 1; ipmtr <= PTRB->npmtr; ipmtr++) {
            if (PTRB->pmtr[ipmtr].type == OCSM_EXTERNAL) {
                irc = 0;
                for (irow = 0; irow < PTRB->pmtr[ipmtr].nrow; irow++) {
                    for (icol = 0; icol < PTRB->pmtr[ipmtr].ncol; icol++) {
                        PTRB->pmtr[ipmtr].value[irc] += MODL->dtime * PTRB->pmtr[ipmtr].dot[irc];
                        PTRB->pmtr[ipmtr].dot[  irc]  = 0;
                        irc++;
                    }
                }
            }
        }

        /* recheck (this is needed to make sure that the indentation is updated) */
        status = ocsmCheck(PTRB);
        CHECK_STATUS(ocsmCheck);

        /* rebuild the perturbed copy */
        nbody_ptrb = 0;
        SPRINT0(2, ">>>\n>>> building perturbation\n>>>");
        status = ocsmBuild(PTRB, 0, &builtTo, &nbody_ptrb, NULL);

        if (status != SUCCESS) {
            SPRINT2(2, "WARNING:: ocsmBuild() -> status=%d, buildTo=%d\n", status, builtTo);
            nerror++;
        }

        /* check that base and ptrb have same number of Bodys */
        if (nerror == 0) {
            if (MODL->nbody != PTRB->nbody) {
                SPRINT0(1, "WARNING:: Base and perturbed models have different nbody");
                SPRINT1(1, "          MODL->nbody = %d", MODL->nbody);
                SPRINT1(1, "          PTRB->nbody = %d", PTRB->nbody);
                nerror++;
            }
        }

        /* check that base and ptrb Bodys match topologically */
        if (nerror == 0) {
            for (ibody = 1; ibody <= MODL->nbody; ibody++) {
                if (MODL->body[ibody].onstack != 1) continue;

//$$$           /* for now, assume that WireBodys map without checking */
                if        (MODL->body[ibody].botype == OCSM_WIRE_BODY &&
                           PTRB->body[ibody].botype == OCSM_WIRE_BODY   ) {
                    status  = SUCCESS;
                    newBody = NULL;

                /* do actual check for SolidBodys */
                } else {
                    status = EG_mapBody(MODL->body[ibody].ebody, PTRB->body[ibody].ebody,
                                        "faceID", &newBody);
                }

                /* if topologes match, we can proceed */
                if (status == SUCCESS) {
                    if (newBody != NULL) {
                        EG_deleteObject(PTRB->body[ibody].ebody);
                        PTRB->body[ibody].ebody = newBody;

                        status = finishBody(PTRB, ibody);
                        CHECK_STATUS(finishBody);
                    }

                /* otherwise we have an error at this dtime */
                } else {
                    SPRINT2(2, "WARNING:: EG_mapBody(ibody=%d) -> status=%d\n", ibody, status);
                    nerror++;
                    break;
                }
            }
        }

        /* if there are no errors above, then we do not need to look at any more times */
        if (nerror == 0) {
            break;

        /* if there are more times to try, flip and decrease time step and try again */
        } else if (itime < ntime-1) {
            status = ocsmFree(PTRB);
            CHECK_STATUS(ocsmFree);

            MODL->perturb = NULL;
            MODL->dtime /= -2;

        /* otherwise we could not find a sufficiently small time step, so return an error  */
        } else {
            signalError(MODL, OCSM_INTERNAL, 0,
                        "dtime=%f is not sufficiently small", MODL->dtime);

            MODL->dtime = -2;      /* signify a problem */
            status = OCSM_INTERNAL_ERROR;
            goto cleanup;
        }
    }

    /* map base ibody's tessellation onto the ptrb Body */
    for (ibody = 1; ibody <= MODL->nbody; ibody++) {
        if (MODL->body[ibody].etess != NULL) {
            status = EG_mapTessBody(MODL->body[ibody].etess, PTRB->body[ibody].ebody,
                                    &(PTRB->body[ibody].etess));
            CHECK_STATUS(EG_mapTessBody);
        }
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   evalRpn - evaluate Rpn-code                                        *
 *                                                                      *
 ************************************************************************
 */

static int
evalRpn(rpn_T     *rpn,                 /* (in)  pointer to Rpn-code */
        modl_T    *modl,                /* (in)  pointer to MODL */
        double    *val,                 /* (out) value      of expression */
        double    *dot)                 /* (out) derivative of expression */
{
    int       status = SUCCESS;         /* (out) return status */

    int       irpn, nvalstack, ival, ipmtr, jpmtr, irow, icol, irc;
    double    valstack[MAX_EXPR_LEN], val1, dot1, nan1, val2, dot2, nan2, val3, nan3, nan4, nan5, nan6, nan7, nan8;
    #ifndef WIN32
        double __attribute__((unused)) dot3;    // needed to eliminate compiler warning
    #else
        double dot3;
    #endif
    double    xa,  xa_d,  ya,  ya_d,  xb, xb_d, yb, yb_d, xc, xc_d, yc, yc_d;
    double    dab, dab_d, dbc, dbc_d, R,  R_d,  L,  L_d;
    double    Xcent, Xcent_d, Ycent, Ycent_d, Xmidl, Xmidl_d, Ymidl, Ymidl_d;
    double    seglen, seglen_d, phi, phi_d, angle, angle_d, sweep, sweep_d;
    double    angab, angab_d, angbc, angbc_d;
    double    rad2deg = 1 / PIo180;
    char      prefix[MAX_NAME_LEN], suffix[MAX_NAME_LEN];

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(evalRpn);
    DPRINT3("%s(rpn=%lx. modl=%lx) {",
            routine, (long)rpn, (long)modl);

#define PUSH_VAL(VAL,DOT,NAN)                   \
    if (nvalstack < MAX_EXPR_LEN-3) {           \
        valstack[nvalstack++] = VAL;            \
        valstack[nvalstack++] = DOT;            \
        valstack[nvalstack++] = NAN;            \
    } else {                                    \
        status = OCSM_VAL_STACK_OVERFLOW;       \
        goto cleanup;                           \
    }

#define POP_VAL(VAL,DOT,NAN)                    \
    if (nvalstack > 2) {                        \
        NAN = valstack[--nvalstack];            \
        DOT = valstack[--nvalstack];            \
        VAL = valstack[--nvalstack];            \
    } else {                                    \
        status = OCSM_VAL_STACK_UNDERFLOW;      \
        goto cleanup;                           \
    }

    /* --------------------------------------------------------------- */

    /* default answer (in case there is an error) */
    *val = 0;
    *dot = 0;

    /* if we begin with PARSE_STRING, simply return default value */
    if (rpn[0].type == PARSE_STRING) {
        goto cleanup;
    }

    /* start with an empty value stack */
    nvalstack = 0;

    /* loop through whole Rpn-code */
    irpn = 0;
    while (rpn[irpn].type != PARSE_END) {

        SPRINT3(3, "executing %3d type=%2d, text=%s", irpn, rpn[irpn].type, rpn[irpn].text);

        /* PARSE_NUM */
        if (rpn[irpn].type == PARSE_NUMBER) {
            PUSH_VAL(strtod(rpn[irpn].text, (char**)NULL), 0.0, 0);

        /* PARSE_NAME */
        } else if (rpn[irpn].type == PARSE_NAME) {
            jpmtr = 0;
            if (modl != NULL) {
                for (ipmtr = 1; ipmtr <= modl->npmtr; ipmtr++) {
                    prefix[0] = '\0';
                    suffix[0] = '\0';
                    for (irc = 0; irc < MAX_NAME_LEN-1; irc++) {
                        if (rpn[irpn].text[irc] == '\0') {
                            break;
                        } else if (rpn[irpn].text[irc] != '.') {
                            prefix[irc  ] = rpn[irpn].text[irc];
                            prefix[irc+1] = '\0';
                        } else {
                            STRNCPY(suffix, &(rpn[irpn].text[irc]), MAX_NAME_LEN);
                            break;
                        }
                    }

                    if (strcmp(modl->pmtr[ipmtr].name, prefix) == 0  &&
                               modl->pmtr[ipmtr].level == modl->level  ) {
                        jpmtr = ipmtr;

                        /* check for dot-suffix */
                        if (strlen(suffix) == 0) {
                            irow = icol = 1;

                            status = ocsmGetValu(modl, jpmtr, irow, icol, &val1, &dot1);
                            CHECK_STATUS(ocsmGetValu);

                            PUSH_VAL(val1, dot1, 0);
                            break;
                        } else if (strcmp(suffix, ".nrow") == 0) {
                            val1 = modl->pmtr[ipmtr].nrow;
                            PUSH_VAL(val1, 0, 0);
                            break;
                        } else if (strcmp(suffix, ".ncol") == 0) {
                            val1 = modl->pmtr[ipmtr].ncol;
                            PUSH_VAL(val1, 0, 0);
                            break;
                        } else if (strcmp(suffix, ".size") == 0) {
                            val1 = (modl->pmtr[ipmtr].nrow) * (modl->pmtr[ipmtr].ncol);
                            PUSH_VAL(val1, 0, 0);
                            break;
                        } else if (strcmp(suffix, ".sum")  == 0) {
                            val1 = 0;
                            dot1 = 0;
                            for (irc = 0; irc < (modl->pmtr[ipmtr].nrow)*(modl->pmtr[ipmtr].ncol); irc++) {
                                val1 += modl->pmtr[ipmtr].value[irc];
                                dot1 += modl->pmtr[ipmtr].dot[  irc];
                            }
                            PUSH_VAL(val1, dot1, 0);
                            break;
                        } else if (strcmp(suffix, ".min")  == 0) {
                            val1 = modl->pmtr[ipmtr].value[0];
                            dot1 = modl->pmtr[ipmtr].dot[  0];
                            for (irc = 0; irc < (modl->pmtr[ipmtr].nrow)*(modl->pmtr[ipmtr].ncol); irc++) {
                                if (modl->pmtr[ipmtr].value[irc] < val1) {
                                    val1 = modl->pmtr[ipmtr].value[irc];
                                    dot1 = modl->pmtr[ipmtr].dot[  irc];
                                }
                            }
                            PUSH_VAL(val1, dot1, 0);
                            break;
                        } else if (strcmp(suffix, ".max")  == 0) {
                            val1 = modl->pmtr[ipmtr].value[0];
                            dot1 = modl->pmtr[ipmtr].dot[  0];
                            for (irc = 0; irc < (modl->pmtr[ipmtr].nrow)*(modl->pmtr[ipmtr].ncol); irc++) {
                                if (modl->pmtr[ipmtr].value[irc] > val1) {
                                    val1 = modl->pmtr[ipmtr].value[irc];
                                    dot1 = modl->pmtr[ipmtr].dot[  irc];
                                }
                            }
                            PUSH_VAL(val1, dot1, 0);
                            break;
                        } else {
                            signalError(MODL, OCSM_ILLEGAL_PMTR_NAME, 0,
                                        "illegal suffix (%s) in name", suffix);
                            status = OCSM_ILLEGAL_PMTR_NAME;
                            goto cleanup;
                        }
                    }
                }
            }
            if (jpmtr == 0) {
                PUSH_VAL(0, 0, 1);
            }

        /* PARSE_ARRAY */
        } else if (rpn[irpn].type == PARSE_ARRAY) {
            jpmtr = 0;
            if (modl != NULL) {
                for (ipmtr = 1; ipmtr <= modl->npmtr; ipmtr++) {
                    if (strcmp(modl->pmtr[ipmtr].name, rpn[irpn].text) == 0 &&
                               modl->pmtr[ipmtr].level == modl->level         ) {
                        jpmtr = ipmtr;

                        POP_VAL(val1, dot1, nan1);
                        POP_VAL(val2, dot2, nan2);

                        irow = NINT(val2);
                        icol = NINT(val1);

                        status = ocsmGetValu(modl, jpmtr, irow, icol, &val1, &dot1);
                        CHECK_STATUS(ocsmGetValu);

                        PUSH_VAL(val1, dot1, 0);
                        break;
                    }
                }
            }
            if (jpmtr == 0) {
                PUSH_VAL(0, 0, 1);
            }

        /* PARSE_OP1 or PARSE_OP2 or PARSE_OP3 */
        } else if (rpn[irpn].type == PARSE_OP1 ||
                   rpn[irpn].type == PARSE_OP2 ||
                   rpn[irpn].type == PARSE_OP3   ) {
            if        (strcmp(rpn[irpn].text, "+") == 0) {
                POP_VAL(val1, dot1, nan1);
                POP_VAL(val2, dot2, nan2);

                if (nan1 != 0 || nan2 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(val2+val1, dot2+dot1, 0);
                }
            } else if (strcmp(rpn[irpn].text, "-") == 0) {
                POP_VAL(val1, dot1, nan1);
                POP_VAL(val2, dot2, nan2);

                if (nan1 != 0 || nan2 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(val2-val1, dot2-dot1, 0);
                }
            } else if (strcmp(rpn[irpn].text, "*") == 0) {
                POP_VAL(val1, dot1, nan1);
                POP_VAL(val2, dot2, nan2);

                if (nan1 != 0 || nan2 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(val2*val1, dot1*val2+dot2*val1, 0);
                }
            } else if (strcmp(rpn[irpn].text, "/") == 0) {
                POP_VAL(val1, dot1, nan1);
                POP_VAL(val2, dot2, nan2);

                if (nan1 != 0 || nan2 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val1 == 0.0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(val2/val1, (dot2*val1-dot1*val2)/val1/val1, 0);
                }
            } else if (strcmp(rpn[irpn].text, "^") == 0) {
                POP_VAL(val1, dot1, nan1);
                POP_VAL(val2, dot2, nan2);
                ival = (int)val1;

                if (nan1 != 0 || nan2 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val1 == (double)ival && dot1 == 0){
                    PUSH_VAL(pow(val2,ival), dot2*ival*pow(val2,ival-1), 0);
                } else if (val2 < 0.0) {
                    status = OCSM_FUNC_ARG_OUT_OF_BOUNDS;
                    goto cleanup;
                } else if (val2 == 0) {
                    PUSH_VAL(0, 0, 0);
                } else {
                    PUSH_VAL(pow(val2,val1), (dot1*log(val2)+dot2*val1/val2)*pow(val2,val1), 0);
                }
            }

        /* PARSE_FUNC */
        } else if (rpn[irpn].type == PARSE_FUNC) {

            /* pi(x) */
            if        (strcmp(rpn[irpn].text, "pi") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(val1*PI, dot1*PI, 0);
                }

            /* min(x,y) */
            } else if (strcmp(rpn[irpn].text, "min") == 0) {
                POP_VAL(val1, dot1, nan1);
                POP_VAL(val2, dot2, nan2);

                if (nan1 != 0 || nan2 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val1 < val2) {
                    PUSH_VAL(val1, dot1, 0);
                } else {
                    PUSH_VAL(val2, dot2, 0);
                }

            /* max(x,y) */
            } else if (strcmp(rpn[irpn].text, "max") == 0) {
                POP_VAL(val1, dot1, nan1);
                POP_VAL(val2, dot2, nan2);

                if (nan1 != 0 || nan2 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val1 > val2) {
                    PUSH_VAL(val1, dot1, 0);
                } else {
                    PUSH_VAL(val2, dot2, 0);
                }

            /* sqrt(x) */
            } else if (strcmp(rpn[irpn].text, "sqrt") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val1 < 0.0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(sqrt(val1), dot1/2/sqrt(val1), 0);
                }

            /* abs(x) */
            } else if (strcmp(rpn[irpn].text, "abs") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val1 >= 0) {
                    PUSH_VAL(+val1, +dot1, 0);
                } else {
                    PUSH_VAL(-val1, -dot1, 0);
                }

            /* int(x) */
            } else if (strcmp(rpn[irpn].text, "int") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    ival = (int) val1;
                    PUSH_VAL((double)ival, 0.0, 0);
                }

            /* nint(x) */
            } else if (strcmp(rpn[irpn].text, "nint") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    if (val1 > 0) {
                        ival = (int) (val1 + 0.50);
                    } else {
                        ival = (int) (val1 - 0.50);
                    }
                    PUSH_VAL((double)ival, 0.0, 0);
                }

            /* ceil(x) */
            } else if (strcmp(rpn[irpn].text, "ceil") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL((double)ceil(val1), 0.0, 0);
                }

            /* floor(x) */
            } else if (strcmp(rpn[irpn].text, "floor") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL((double)floor(val1), 0.0, 0);
                }

            /* sign(x) */
            } else if (strcmp(rpn[irpn].text, "sign") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(val1, dot1, nan1);
                } else if (val1 > 0) {
                    PUSH_VAL(+1, 0, 0);
                } else if (val1 < 0) {
                    PUSH_VAL(-1, 0, 0);
                } else {
                    PUSH_VAL( 0, 0, 0);
                }

            /* exp(x) */
            } else if (strcmp(rpn[irpn].text, "exp") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(exp(val1), dot1*exp(val1), 0);
                }

            /* log(x) */
            } else if (strcmp(rpn[irpn].text, "log") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val1 < 0.0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(log(val1), dot1/val1, 0);
                }

            /* log10(x) */
            } else if (strcmp(rpn[irpn].text, "log10") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val1 < 0.0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(log10(val1), dot1*log10(exp(1.))/val1, 0);
                }

            /* sin(x) */
            } else if (strcmp(rpn[irpn].text, "sin") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(sin(val1), dot1*cos(val1), 0);
                }

            /* sind(x) */
            } else if (strcmp(rpn[irpn].text, "sind") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    val1 /= rad2deg;
                    dot1 /= rad2deg;
                    PUSH_VAL(sin(val1), dot1*cos(val1), 0);
                }

            /* asin(x) */
            } else if (strcmp(rpn[irpn].text, "asin") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val1 < -1.0 || val1 > 1.0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(asin(val1), dot1/sqrt(1-val1*val1), 0);
                }

            /* asind(x) */
            } else if (strcmp(rpn[irpn].text, "asind") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val1 < -1.0 || val1 > 1.0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(asin(val1)*rad2deg, dot1*rad2deg/sqrt(1-val1*val1), 0);
                }

            /* cos(x) */
            } else if (strcmp(rpn[irpn].text, "cos") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(cos(val1), -dot1*sin(val1), 0);
                }

            /* cosd(x) */
            } else if (strcmp(rpn[irpn].text, "cosd") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    val1 /= rad2deg;
                    dot1 /= rad2deg;
                    PUSH_VAL(cos(val1), -dot1*sin(val1), 0);
                }

            /* acos(x) */
            } else if (strcmp(rpn[irpn].text, "acos") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val1 < -1.0 || val1 > 1.0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(acos(val1), -dot1/sqrt(1-val1*val1), 0);
                }

            /* acosd(x) */
            } else if (strcmp(rpn[irpn].text, "acosd") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val1 < -1.0 || val1 > 1.0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(acos(val1)*rad2deg, -dot1*rad2deg/sqrt(1-val1*val1), 0);
                }

            /* tan(x) */
            } else if (strcmp(rpn[irpn].text, "tan") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(tan(val1), dot1/cos(val1)/cos(val1), 0);
                }

            /* tand(x) */
            } else if (strcmp(rpn[irpn].text, "tand") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    val1 /= rad2deg;
                    dot1 /= rad2deg;

                    PUSH_VAL(tan(val1), dot1/cos(val1)/cos(val1), 0);
                }

            /* atan(x) */
            } else if (strcmp(rpn[irpn].text, "atan") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(atan(val1), dot1/(1+val1*val1), 0);
                }

            /* atand(x) */
            } else if (strcmp(rpn[irpn].text, "atand") == 0) {
                POP_VAL(val1, dot1, nan1);

                if (nan1 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(atan(val1)*rad2deg, dot1*rad2deg/(1+val1*val1), 0);
                }

            /* atan2(y,x) */
            } else if (strcmp(rpn[irpn].text, "atan2") == 0) {
                POP_VAL(val1, dot1, nan1);
                POP_VAL(val2, dot2, nan2);

                if (nan1 != 0 || nan2 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val1 == 0.0 && val2 == 0.0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(atan2(val2,val1), (dot2*val1-dot1*val2)/(val1*val1+val2*val2), 0);
                }

            /* atan2d(y,x) */
            } else if (strcmp(rpn[irpn].text, "atan2d") == 0) {
                POP_VAL(val1, dot1, nan1);
                POP_VAL(val2, dot2, nan2);

                if (nan1 != 0 || nan2 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val1 == 0.0 && val2 == 0.0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    PUSH_VAL(atan2(val2,val1)*rad2deg, (dot2*val1-dot1*val2)*rad2deg/(val1*val1+val2*val2), 0);
                }

            /* hypot(x,y) */
            } else if (strcmp(rpn[irpn].text, "hypot") == 0) {
                POP_VAL(val1, dot1, nan1);
                POP_VAL(val2, dot2, nan2);

                if (nan1 != 0 || nan2 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val1 == 0.0 && val2 == 0.0) {
                    PUSH_VAL(0.0, 0.0, 0);
                } else {
                    PUSH_VAL(sqrt(val1*val1+val2*val2), (dot1*val1+dot2*val2)/sqrt(val1*val1+val2*val2), 0);
                }

            /* Xcent(xa,ya,dab,xb,yb) */
            } else if (strcmp(rpn[irpn].text, "Xcent") == 0) {
                POP_VAL(yb,  yb_d,  nan1);
                POP_VAL(xb,  xb_d,  nan2);
                POP_VAL(dab, dab_d, nan3);
                POP_VAL(ya,  ya_d,  nan4);
                POP_VAL(xa,  xa_d,  nan5);

                if (nan1 != 0 || nan2 != 0 || nan3 != 0 || nan4 != 0 || nan5 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (fabs(dab) < EPS06) {
                    Xcent   = (xa   + xb  ) / 2;
                    Xcent_d = (xa_d + xb_d) / 2;

                    PUSH_VAL(Xcent, Xcent_d, 0);
                } else {
                    L   = sqrt((xb-xa)*(xb  -xa  ) + (yb-ya)*(yb  -ya  )) /  2;
                    L_d =     ((xb-xa)*(xb_d-xa_d) + (yb-ya)*(yb_d-ya_d)) / (4 * L);

                    R   = (L*L + dab*dab) / (2 * dab);
                    R_d = (dab * L * L_d - L * L * dab_d) / (2 * dab * dab);

                    Xcent   = (xa   + xb  ) / 2
                            - (R-dab) * (yb-ya) / (2*L);
                    Xcent_d = (xa_d + xb_d) / 2
                            - (R_d-dab_d) * (yb  -ya  ) / (2*L)
                            - (R  -dab  ) * (yb_d-ya_d) / (2*L)
                            + L_d * (R-dab) * (yb-ya) / (2*L*L);

                    PUSH_VAL(Xcent, Xcent_d, 0);
                }

            /* Ycent(xa,ya,dab,xb,yb) */
            } else if (strcmp(rpn[irpn].text, "Ycent") == 0) {
                POP_VAL(yb,  yb_d,  nan1);
                POP_VAL(xb,  xb_d,  nan2);
                POP_VAL(dab, dab_d, nan3);
                POP_VAL(ya,  ya_d,  nan4);
                POP_VAL(xa,  xa_d,  nan5);

                if (nan1 != 0 || nan2 != 0 || nan3 != 0 || nan4 != 0 || nan5 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (fabs(dab) < EPS06) {
                    Ycent   = (ya   + yb  ) / 2;
                    Ycent_d = (ya_d + yb_d) / 2;

                    PUSH_VAL(Ycent, Ycent_d, 0);
                } else {
                    L   = sqrt((xb-xa)*(xb  -xa  ) + (yb-ya)*(yb  -ya  )) /  2;
                    L_d =     ((xb-xa)*(xb_d-xa_d) + (yb-ya)*(yb_d-ya_d)) / (4 * L);

                    R   = (L*L + dab*dab) / (2 * dab);
                    R_d = (dab * L * L_d - L * L * dab_d) / (2 * dab * dab);

                    Ycent   = (ya + yb) / 2
                            + (R-dab) * (xb-xa) / (2*L);
                    Ycent_d = (ya_d + yb_d) / 2
                            + (R_d-dab_d) * (xb  -xa  ) / (2*L)
                            + (R  -dab  ) * (xb_d-xa_d) / (2*L)
                            + L_d * (R-dab) * (xb-xa) / (2*L*L);

                    PUSH_VAL(Ycent, Ycent_d, 0);
                }

            /* Xmidl(xa,ya,dab,xb,yb) */
            } else if (strcmp(rpn[irpn].text, "Xmidl") == 0) {
                POP_VAL(yb,  yb_d,  nan1);
                POP_VAL(xb,  xb_d,  nan2);
                POP_VAL(dab, dab_d, nan3);
                POP_VAL(ya,  ya_d,  nan4);
                POP_VAL(xa,  xa_d,  nan5);

                if (nan1 != 0 || nan2 != 0 || nan3 != 0 || nan4 != 0 || nan5 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (fabs(dab) < EPS06) {
                    Xmidl   = (xa   + xb  ) / 2;
                    Xmidl_d = (xa_d + xb_d) / 2;

                    PUSH_VAL(Xmidl, Xmidl_d, 0);
                } else {
                    L   = sqrt((xb-xa)*(xb  -xa  ) + (yb-ya)*(yb  -ya  )) /  2;
                    L_d =     ((xb-xa)*(xb_d-xa_d) + (yb-ya)*(yb_d-ya_d)) / (2 * L);

                    Xmidl   = (xa + xb) / 2
                            + dab * (yb-ya) / (2*L);
                    Xmidl_d = (xa_d + xb_d) / 2
                            + dab_d * (yb  -ya  ) / (2*L)
                            + dab   * (yb_d-ya_d) / (2*L)
                            - L_d * dab * (yb-ya) / (2*L*L);

                    PUSH_VAL(Xmidl, Xmidl_d, 0);
                }

            /* Ymidl(xa,ya,dab,xb,yb) */
            } else if (strcmp(rpn[irpn].text, "Ymidl") == 0) {
                POP_VAL(yb,  yb_d,  nan1);
                POP_VAL(xb,  xb_d,  nan2);
                POP_VAL(dab, dab_d, nan3);
                POP_VAL(ya,  ya_d,  nan4);
                POP_VAL(xa,  xa_d,  nan5);

                if (nan1 != 0 || nan2 != 0 || nan3 != 0 || nan4 != 0 || nan5 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (fabs(dab) < EPS06) {
                    Ymidl   = (ya   + yb  ) / 2;
                    Ymidl_d = (ya_d + yb_d) / 2;

                    PUSH_VAL(Ymidl, Ymidl_d, 0);
                } else {
                    L   = sqrt((xb-xa)*(xb  -xa  ) + (yb-ya)*(yb  -ya  )) /  2;
                    L_d =     ((xb-xa)*(xb_d-xa_d) + (yb-ya)*(yb_d-ya_d)) / (2 * L);

                    Ymidl   = (ya   + yb  ) / 2
                            - dab * (xb-xa) / (2*L);
                    Ymidl_d = (ya_d + yb_d) / 2
                            - dab_d * (xb  -xa  ) / (2*L)
                            - dab   * (xb_d-xa_d) / (2*L)
                            + L_d * dab * (xb-xa) / (2*L*L);

                    PUSH_VAL(Ymidl, Ymidl_d, 0);
                }

            /* seglen(xa,ya,dab,xb,yb) */
            } else if (strcmp(rpn[irpn].text, "seglen") == 0) {
                POP_VAL(yb,  yb_d,  nan1);
                POP_VAL(xb,  xb_d,  nan2);
                POP_VAL(dab, dab_d, nan3);
                POP_VAL(ya,  ya_d,  nan4);
                POP_VAL(xa,  xa_d,  nan5);

                if (nan1 != 0 || nan2 != 0 || nan3 != 0 || nan4 != 0 || nan5 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (fabs(dab) < EPS06) {
                    seglen   = sqrt((xb  -xa  )*(xb-xa) + (yb  -ya  )*(yb-ya));
                    seglen_d =     ((xb_d-xa_d)*(xb-xa) + (yb_d-ya_d)*(yb-ya)) / seglen;

                    PUSH_VAL(seglen, seglen_d, 0);
                } else {
                    L   = sqrt((xb-xa)*(xb  -xa  ) + (yb-ya)*(yb  -ya  )) /  2;
                    L_d =     ((xb-xa)*(xb_d-xa_d) + (yb-ya)*(yb_d-ya_d)) / (2 * L);

                    R   = (L*L + dab*dab) / (2 * dab);
                    R_d = (dab * L * L_d - L * L * dab_d) / (2 * dab * dab);

                    phi   = atan(L / (R-dab));
                    phi_d = ((R-dab) * L_d - L * (R_d-dab_d)) / (L*L + (R-dab)*(R-dab));

                    seglen   = 2 * R  * phi;
                    seglen_d = 2 * (R_d * phi + R * phi_d);

                    PUSH_VAL(seglen, seglen_d, 0);
                }

            /* incline(xa,ya,dab,xb,yb) */
            } else if (strcmp(rpn[irpn].text, "incline") == 0) {
                POP_VAL(yb,  yb_d,  nan1);
                POP_VAL(xb,  xb_d,  nan2);
                POP_VAL(dab, dab_d, nan3);
                POP_VAL(ya,  ya_d,  nan4);
                POP_VAL(xa,  xa_d,  nan5);

                if (nan1 != 0 || nan2 != 0 || nan3 != 0 || nan4 != 0 || nan5 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    angle   = atan2(yb-ya, xb-xa);
                    angle_d = ((xb-xa) * (yb_d-ya_d) - (yb-ya) * (xb_d-xa_d))
                            / ((xb-xa) * (xb  -xa  ) + (yb-ya) * (yb  -ya  ));

                    PUSH_VAL(angle*rad2deg, angle_d*rad2deg, 0);
                }

            /* radius(xa,ya,dab,xb,yb) */
            } else if (strcmp(rpn[irpn].text, "radius") == 0) {
                POP_VAL(yb,  yb_d,  nan1);
                POP_VAL(xb,  xb_d,  nan2);
                POP_VAL(dab, dab_d, nan3);
                POP_VAL(ya,  ya_d,  nan4);
                POP_VAL(xa,  xa_d,  nan5);

                if (nan1 != 0 || nan2 != 0 || nan3 != 0 || nan4 != 0 || nan5 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (fabs(dab) < EPS06) {
                    PUSH_VAL(0.0, 0.0, 0);
                } else {
                    L   = sqrt((xb-xa)*(xb  -xa  ) + (yb-ya)*(yb  -ya  )) /  2;
                    L_d =     ((xb-xa)*(xb_d-xa_d) + (yb-ya)*(yb_d-ya_d)) / (2 * L);

                    R   = (L*L + dab*dab) / (2 * dab);
                    R_d = (dab * L * L_d - L * L * dab_d) / (2 * dab * dab);

                    PUSH_VAL(R, R_d, 0);
                }

            /* sweep(xa,ya,dab,xb,yb) */
            } else if (strcmp(rpn[irpn].text, "sweep") == 0) {
                POP_VAL(yb,  yb_d,  nan1);
                POP_VAL(xb,  xb_d,  nan2);
                POP_VAL(dab, dab_d, nan3);
                POP_VAL(ya,  ya_d,  nan4);
                POP_VAL(xa,  xa_d,  nan5);

                if (nan1 != 0 || nan2 != 0 || nan3 != 0 || nan4 != 0 || nan5 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (fabs(dab) < EPS06) {
                    PUSH_VAL(0.0, 0.0, 0);
                } else {
                    L   = sqrt((xb-xa)*(xb  -xa  ) + (yb-ya)*(yb  -ya  )) /  2;
                    L_d =     ((xb-xa)*(xb_d-xa_d) + (yb-ya)*(yb_d-ya_d)) / (2 * L);

                    R   = (L*L + dab*dab) / (2 * dab);
                    R_d = (dab * L * L_d - L * L * dab_d) / (2 * dab * dab);

                    phi   = atan(L / (R-dab));
                    phi_d = ((R-dab) * L_d - L * (R_d-dab_d)) / (L*L + (R-dab)*(R-dab));

                    sweep   = 2 * phi;
                    sweep_d = 2 * phi_d;

                    PUSH_VAL(sweep*rad2deg, sweep_d*rad2deg, 0);
                }

            /* turnang(xa,ya,dab,xb,yb) */
            } else if (strcmp(rpn[irpn].text, "turnang") == 0) {
                POP_VAL(yc,  yc_d,  nan1);
                POP_VAL(xc,  xc_d,  nan2);
                POP_VAL(dbc, dbc_d, nan3);
                POP_VAL(yb,  yb_d,  nan4);
                POP_VAL(xb,  xb_d,  nan5);
                POP_VAL(dab, dab_d, nan6);
                POP_VAL(ya,  ya_d,  nan7);
                POP_VAL(xa,  xa_d,  nan8);

                if (nan1 != 0 || nan2 != 0 || nan3 != 0 || nan4 != 0 ||
                    nan5 != 0 || nan6 != 0 || nan7 != 0 || nan8 != 0   ) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    angab   = atan2(yb-ya, xb-xa);
                    angab_d = ((xb-xa)*(yb_d-ya_d) - (yb-ya)*(xb_d-xa_d))
                            / ((xb-xa)*(xb  -xa  ) + (yb-ya)*(yb  -ya  ));

                    if (fabs(dab) > EPS06) {
                        L   = sqrt((xb-xa)*(xb  -xa  ) + (yb-ya)*(yb  -ya  )) /  2;
                        L_d =     ((xb-xa)*(xb_d-xa_d) + (yb-ya)*(yb_d-ya_d)) / (2 * L);

                        R   = (L*L + dab*dab) / (2 * dab);
                        R_d = (dab * L * L_d - L * L * dab_d) / (2 * dab * dab);

                        phi   = atan(L / (R-dab));
                        phi_d = ((R-dab) * L_d - L * (R_d-dab_d)) / (L*L + (R-dab)*(R-dab));

                        if        (dab >  +L) {   //                              +L < dab
                            angab   += phi + PI;
                            angab_d += phi_d;
                        } else if (dab >   0) {   //                   0 < dab <= +L
                            angab   += phi;
                            angab_d += phi_d;
                        } else if (dab >  -L) {   //        -L < dab < 0
                            angab   += phi;
                            angab_d += phi_d;
                        } else {                  // dab <= -L
                            angab   += phi - PI;
                            angab_d += phi_d;
                        }
                    }

                    angbc = atan2(yc-yb, xc-xb);
                    angbc_d = ((xc-xb)*(yc_d-yb_d) - (yc-yb)*(xc_d-xb_d))
                            / ((xc-xb)*(xc  -xb  ) + (yc-yb)*(yc  -yb  ));

                    if (fabs(dbc) > EPS06) {
                        L   = sqrt((xc-xb)*(xc  -xb  ) + (yc-yb)*(yc  -yb  )) /  2;
                        L_d =     ((xc-xb)*(xc_d-xb_d) + (yc-yb)*(yc_d-yb_d)) / (2 * L);

                        R   = (L*L + dbc*dbc) / (2 * dbc);
                        R_d = (dbc * L * L_d - L * L * dbc_d) / (2 * dbc * dbc);

                        phi   = atan(L / (R-dbc));
                        phi_d = ((R-dbc) * L_d - L * (R_d-dbc_d)) / (L*L + (R-dbc)*(R-dbc));

                        if        (dbc >  +L) {   //                              +L < dbc
                            angbc   -= phi + PI;
                            angbc_d -= phi_d;
                        } else if (dbc >   0) {   //                   0 < dbc <= +L
                            angbc   -= phi;
                            angbc_d -= phi_d;
                        } else if (dbc >  -L) {   //        -L < dbc < 0
                            angbc   -= phi;
                            angbc_d -= phi_d;
                        } else {                  // dbc <= -L
                            angbc   -= phi - PI;
                            angbc_d -= phi_d;
                        }
                    }

                    angle   = angbc   - angab;
                    angle_d = angbc_d - angab_d;
                    while (angle < -PI) {
                        angle += TWOPI;
                    }
                    while (angle > +PI) {
                        angle -= TWOPI;
                    }

                    PUSH_VAL(angle*rad2deg, angle_d*rad2deg, 0);
                }

            /* dip(xa,ya,xb,yb,rad) */
            } else if (strcmp(rpn[irpn].text, "dip") == 0) {
                POP_VAL(R,  R_d,  nan1);
                POP_VAL(yb, yb_d, nan2);
                POP_VAL(xb, xb_d, nan3);
                POP_VAL(ya, ya_d, nan4);
                POP_VAL(xa, xa_d, nan5);

                if (nan1 != 0 || nan2 != 0 || nan3 != 0 ||
                    nan4 != 0 || nan5 != 0                ) {
                    PUSH_VAL(0, 0, 1);
                } else {
                    L   = sqrt((xb-xa)*(xb  -xa  ) + (yb-ya)*(yb  -ya  )) /  2;
                    L_d =     ((xb-xa)*(xb_d-xa_d) + (yb-ya)*(yb_d-ya_d)) / (2 * L);

                    if (L > R) {
                        PUSH_VAL(0, 0, 1);
                    } else {
                        dab   = R   -                        sqrt(R*R - L*L);
                        dab_d = R_d - (R*R_d - L*L_d) / (2 * sqrt(R*R - L*L));
                        PUSH_VAL(dab, dab_d, 0);
                    }
                }

            /* ifzero(test,ifTrue,ifFalse) */
            } else if (strcmp(rpn[irpn].text, "ifzero") == 0) {
                POP_VAL(val1, dot1, nan1);
                POP_VAL(val2, dot2, nan2);
                POP_VAL(val3, dot3, nan3);

                if (nan3 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val3 == 0) {
                    PUSH_VAL(val2, dot2, nan2);
                } else {
                    PUSH_VAL(val1, dot1, nan1);
                }

            /* ifpos(test,ifTrue,ifFalse) */
            } else if (strcmp(rpn[irpn].text, "ifpos") == 0) {
                POP_VAL(val1, dot1, nan1);
                POP_VAL(val2, dot2, nan2);
                POP_VAL(val3, dot3, nan3);

                if (nan3 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val3 > 0) {
                    PUSH_VAL(val2, dot2, nan2);
                } else {
                    PUSH_VAL(val1, dot1, nan1);
                }

            /* ifneg(test,ifTrue,ifFalse) */
            } else if (strcmp(rpn[irpn].text, "ifneg") == 0) {
                POP_VAL(val1, dot1, nan1);
                POP_VAL(val2, dot2, nan2);
                POP_VAL(val3, dot3, nan3);

                if (nan3 != 0) {
                    PUSH_VAL(0, 0, 1);
                } else if (val3 < 0) {
                    PUSH_VAL(val2, dot2, nan2);
                } else {
                    PUSH_VAL(val1, dot1, nan1);
                }

            /* ifnan(test,ifTrue,ifFalse) */
            } else if (strcmp(rpn[irpn].text, "ifnan") == 0) {
                POP_VAL(val1, dot1, nan1);
                POP_VAL(val2, dot2, nan2);
                POP_VAL(val3, dot3, nan3);

                if (nan3 != 0) {
                    PUSH_VAL(val2, dot2, nan2);
                } else {
                    PUSH_VAL(val1, dot1, nan1);
                }
            }
        } else {
            SPRINT2(3, "skipping type %d (%s)", rpn[irpn].type, rpn[irpn].text);
        }

        if (outLevel >= 3) {
            int ifoo;
            for (ifoo = 0; ifoo < nvalstack; ifoo += 3) {
                SPRINT3(3, "    %10.5f %10.5f %10.5f",
                        valstack[ifoo], valstack[ifoo+1], valstack[ifoo+2]);
            }
        }

        irpn++;
    }

    /* if there is only one entry on the valstack, return it */
    POP_VAL(*val, *dot, nan1);

    if (nan1 != 0) {
        status = OCSM_FUNC_ARG_OUT_OF_BOUNDS;
    } else if (nvalstack > 0) {
        status = OCSM_VAL_STACK_OVERFLOW;
    }

#undef PUSH_VAL
#undef POP_VAL

cleanup:
    DPRINT4("%s --> status=%d, val=%f, dot=%f}", routine, status, *val, *dot);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   faceContains - determine if point is in Face's bounding box        *
 *                                                                      *
 ************************************************************************
 */

static int
faceContains(ego    eface,              /* (in)  pointer to Face */
             double xx,                 /* (in)  x-coordinate */
             double yy,                 /* (in)  y-coordinate */
             double zz)                 /* (in)  z-coordinate */
{
    int       status = 0;               /* (out) =0  not in bounding box */
                                        /*       =1      in bounding box */

    double    box[6];

    ROUTINE(faceContains);
    DPRINT4("%s(xx=%f, yy=%f, zz=%f) {",
            routine, xx, yy, zz);

    /* --------------------------------------------------------------- */

    status = EG_getBoundingBox(eface, box);
    CHECK_STATUS(EG_getBoundingBox);

    if (xx < box[0] || yy < box[1] || zz < box[2] ||
        xx > box[3] || yy > box[4] || zz > box[5]   ) {
        status = 0;
    } else {
        status = 1;
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   finishBody - finish the definition of the body                     *
 *                                                                      *
 ************************************************************************
 */

static int
finishBody(modl_T *modl,                /* (in)  pointer to MODL */
           int    ibody)                /* (in)  Body index (1-nbody) */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       nnode, nedge, nface, inode, iedge, iface, ibodyl, ibodyr;
    int       jedge, jface, nodes[2], iokayl, iokayr, ipmtr, jpmtr;
    int       ileft, irite, ibeg, iend, nlist, itype, iattrib[7], iswap;
    CINT      *tempIlist;
    double    data[18], value, dot;
    CDOUBLE   *tempRlist;
    CCHAR     *tempClist;
    body_T    *body=NULL;

    int       nchild, ichild, oclass, mtype, ibrch, nn, old_outLevel, nchange, *senses;
    int       nloop, iloop, ntemp, iattr;
    double    datai[18], dataj[18];
    ego       ebody, *enodes, eedge, *eedges, eface, *efaces, *echildren;
    ego       eref, *echilds, *eloops, topRef, prev, next;

    ROUTINE(finishBody);
    DPRINT2("%s(ibody=%d) {",
            routine, ibody);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ibody is given */
    if (ibody < 1 || ibody > MODL->nbody) {
        status = OCSM_ILLEGAL_BODY_INDEX;
        goto cleanup;
    }

    /* get Body info */
    body = &(MODL->body[ibody]);

    /* print ego information */
    if (outLevel >= 3) {
        SPRINT1(3, "finishing Body %d:", ibody);
        printEgo(body->ebody);
    }

    /* if a NULL Body, set up quantities */
    if (body->botype == OCSM_NULL_BODY) {
        body->nnode = 0;
        body->nedge = 0;
        body->nface = 0;

        goto cleanup;
    }

    /* make sure that all SOlidBodys have a positive volume */
    if (body->botype == OCSM_SOLID_BODY) {
        status = EG_getMassProperties(body->ebody, data);
        CHECK_STATUS(EG_getMassProperties);

        if (data[0] <= 0) {
            SPRINT1(0, "WARNING:: Body might be inside-out (volume=%10.3e)", data[0]);
        }
    }

    /* get the number of Nodes, Edges, and Faces associated with this Body */
    nnode = 0;
    nedge = 0;
    nface = 0;

    ebody  = body->ebody;

    status = EG_getInfo(ebody, &oclass, &mtype, &topRef, &prev, &next);
    CHECK_STATUS(EG_getInfo);

    if (oclass != BODY) {
        goto cleanup;
    }

    status = EG_getBodyTopos(ebody, NULL, NODE, &nn, &enodes);
    CHECK_STATUS(EG_getBodyTopos:NODE);
    nnode += nn;
    EG_free(enodes);

    status = EG_getBodyTopos(ebody, NULL, EDGE, &nn, &eedges);
    CHECK_STATUS(EG_getBodyTopos:EDGE);
    nedge += nn;
    EG_free(eedges);

    status = EG_getBodyTopos(ebody, NULL, FACE, &nn, &efaces);
    CHECK_STATUS(EG_getBodyTopos:FACE);
    nface += nn;
    EG_free(efaces);

    body->nnode = nnode;
    body->nedge = nedge;
    body->nface = nface;

    /* if we are creting a perturbation (that is, is basemodl != NULL),
       make sure that the number of Nodes, Edges, and Faces matches
       the base model */
    if (MODL->basemodl != NULL) {
        if (ibody > MODL->basemodl->nbody) {
            SPRINT2(0, "WARNING:: trying to finish ibody=%d but base model only had %d Bodys",
                    ibody, MODL->basemodl->nbody);
            status = OCSM_PROBLEM_CREATING_PERTURB;
            goto cleanup;
        }

        if (nnode != MODL->basemodl->body[ibody].nnode) {
            SPRINT3(0, "WARNING:: nnode disagree for ibody %d, PRTB=%d, BASE=%d",
                    ibody, nnode, MODL->basemodl->body[ibody].nnode);
            status = OCSM_PROBLEM_CREATING_PERTURB;
            goto cleanup;
        }

        if (nedge != MODL->basemodl->body[ibody].nedge) {
            SPRINT3(0, "WARNING:: nedge disagree for ibody %d, PRTB=%d, BASE=%d",
                    ibody, nedge, MODL->basemodl->body[ibody].nedge);
            status = OCSM_PROBLEM_CREATING_PERTURB;
            goto cleanup;
        }

        if (nface != MODL->basemodl->body[ibody].nface) {
            SPRINT3(0, "WARNING:: nface disagree for ibody %d, PRTB=%d, BASE=%d",
                    ibody, nface, MODL->basemodl->body[ibody].nface);
            status = OCSM_PROBLEM_CREATING_PERTURB;
            goto cleanup;
        }
    }

    /* add any global attributes to this Body */
    ebody = body->ebody;

    for (iattr = 0; iattr < MODL->nattr; iattr++) {
        if (MODL->attr[iattr].value[0] == '$') {
            status = EG_attributeAdd(ebody, MODL->attr[iattr].name,
                                     ATTRSTRING, 1,
                                     NULL, NULL, &(MODL->attr[iattr].value[1]));
            CHECK_STATUS(EG_attributeAdd);
        } else {
            ipmtr = -1;
            for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
                if (strcmp(MODL->attr[iattr].name, MODL->pmtr[jpmtr].name) == 0) {
                    ipmtr = jpmtr;
                    break;
                }
            }

            if (ipmtr > 0) {
                status = EG_attributeAdd(ebody, MODL->attr[iattr].name,
                                         ATTRREAL, (MODL->pmtr[ipmtr].nrow)*(MODL->pmtr[ipmtr].ncol),
                                         NULL, MODL->pmtr[ipmtr].value, NULL);
                CHECK_STATUS(EG_attributeAdd);
            } else {
                str2val(MODL->attr[iattr].value, MODL, &value, &dot);
                CHECK_STATUS(str2val);

                status = EG_attributeAdd(ebody, MODL->attr[iattr].name,
                                         ATTRREAL, 1,
                                         NULL, &value, NULL);
                CHECK_STATUS(EG_attributeAdd);
            }
        }
    }

    /* add any attributes associated with this Branch to this Body */
    ibrch = MODL->body[ibody].ibrch;

    for (iattr = 0; iattr < MODL->brch[ibrch].nattr; iattr++) {
        if (MODL->brch[ibrch].attr[iattr].value[0] == '$') {
            status = EG_attributeAdd(ebody, MODL->brch[ibrch].attr[iattr].name,
                                     ATTRSTRING, 1,
                                     NULL, NULL, &(MODL->brch[ibrch].attr[iattr].value[1]));
            CHECK_STATUS(EG_attributeAdd);
        } else {
            ipmtr = -1;
            for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
                if (strcmp(MODL->brch[ibrch].attr[iattr].value, MODL->pmtr[jpmtr].name) == 0) {
                    ipmtr = jpmtr;
                    break;
                }
            }

            if (ipmtr > 0) {
                status = EG_attributeAdd(ebody, MODL->brch[ibrch].attr[iattr].name,
                                         ATTRREAL, (MODL->pmtr[ipmtr].nrow)*(MODL->pmtr[ipmtr].ncol),
                                         NULL, MODL->pmtr[ipmtr].value, NULL);
                CHECK_STATUS(EG_attributeAdd);
            } else {
                str2val(MODL->brch[ibrch].attr[iattr].value, MODL, &value, &dot);
                CHECK_STATUS(str2val);

                status = EG_attributeAdd(ebody, MODL->brch[ibrch].attr[iattr].name,
                                         ATTRREAL, 1,
                                         NULL, &value, NULL);
                CHECK_STATUS(EG_attributeAdd);
            }
        }
    }

    /* add the current Branch as an attribute for this Body */
    status = EG_attributeAdd(ebody, "body", ATTRINT, 1,
                             &ibody, NULL, NULL);
    CHECK_STATUS(EG_attributeAdd);

    status = EG_attributeAdd(ebody, "brch", ATTRINT, 1,
                             &ibrch, NULL, NULL);
    CHECK_STATUS(EG_attributeAdd);

    /* initialize the Nodes for this Body */
    if (nnode < 1) {
        SPRINT2(0, "WARNING:: nnode=%d for ibody=%d", nnode, ibody);
        status = OCSM_INTERNAL_ERROR;
        goto cleanup;
    }

    MALLOC(body->node, node_T, nnode+1);

    status = EG_getBodyTopos(body->ebody, NULL, NODE, &ntemp, &enodes);
    CHECK_STATUS(EG_getBodyTopos);

    if (ntemp != nnode) {
        signalError(MODL, OCSM_INTERNAL_ERROR, 0,
                    "ntemp=%d does not match nnode=%d", ntemp, nnode);
        SET_STATUS(OCSM_INTERNAL_ERROR, EG_getBodyTopos);
    }

    for (inode = 1; inode <= nnode; inode++) {
        body->node[inode].enode = enodes[inode-1];
        body->node[inode].nedge = 0;
        body->node[inode].ibody = 0;
    }

    EG_free(enodes);

    for (inode = 1; inode <= nnode; inode++) {
        status = EG_getTopology(MODL->body[ibody].node[inode].enode,
                                &eref, &oclass, &mtype, data,
                                &nchild, &echilds, &senses);
        CHECK_STATUS(EG_getTopology);

        body->node[inode].x = data[0];
        body->node[inode].y = data[1];
        body->node[inode].z = data[2];
    }

    body->gratt.object = NULL;
    body->gratt.active = 1;                              /* active */
    body->gratt.color  = 0x00000000;                     /* black */
    body->gratt.ptsize = 5;
    body->gratt.render = 64;                             /* FORWARD */
    body->gratt.dirty  = 1;

    /* initialize the Edges for this Body */
    if (nedge < 1) {
        SPRINT2(0, "WARNING:: nedge=%d for ibody=%d", nedge, ibody);
        status = OCSM_INTERNAL_ERROR;
        goto cleanup;
    }

    MALLOC(body->edge, edge_T, nedge+1);

    status = EG_getBodyTopos(body->ebody, NULL, EDGE, &ntemp, &eedges);
    CHECK_STATUS(EG_getBodyTopos);

    if (ntemp != nedge) {
        signalError(MODL, OCSM_INTERNAL_ERROR, 0,
                    "ntemp=%d does not match nedge=%d", ntemp, nedge);
        SET_STATUS(OCSM_INTERNAL_ERROR, EG_getBodyTopos);
    }

    for (iedge = 1; iedge <= nedge; iedge++) {
        body->edge[iedge].eedge = eedges[iedge-1];

        status = EG_getTopology(eedges[iedge-1], &eref, &oclass, &mtype,
                                data, &nchild, &echilds, &senses);
        CHECK_STATUS(EG_getTopology);

        nodes[0] = status = EG_indexBodyTopo(body->ebody, echilds[0]);
        CHECK_STATUS(EG_indexBodyTopo);

        nodes[1] = status = EG_indexBodyTopo(body->ebody, echilds[1]);
        CHECK_STATUS(EG_indexBodyTopo);

        body->edge[iedge].ibeg         = nodes[0];
        body->edge[iedge].iend         = nodes[1];
        body->edge[iedge].ileft        = -1;
        body->edge[iedge].irite        = -1;
        body->edge[iedge].nface        =  0;
        body->edge[iedge].ibody        = -1;
        body->edge[iedge].iford        = -1;
        body->edge[iedge].imark        = -1;
        body->edge[iedge].gratt.object = NULL;
        body->edge[iedge].gratt.active = 1;              /* active */
        body->edge[iedge].gratt.color  = 0x00ff0000;     /* red */
        body->edge[iedge].gratt.bcolor = 0x00ffffff;     /* white */
        body->edge[iedge].gratt.mcolor = 0x00000000;     /* black */
        body->edge[iedge].gratt.lwidth = 2;
        body->edge[iedge].gratt.ptsize = 3;
        body->edge[iedge].gratt.render = 2 + 64;         /* FOREGROUND | FORWARD */
        body->edge[iedge].gratt.dirty  = 1;
    }

    EG_free(eedges);

    /* initialize the Faces for this Body */
    MALLOC(body->face, face_T, nface+1);

    status = EG_getBodyTopos(body->ebody, NULL, FACE, &ntemp, &efaces);
    CHECK_STATUS(EG_getBodyTopos);

    if (ntemp != nface) {
        signalError(MODL, OCSM_INTERNAL_ERROR, 0,
                    "ntemp=%d does not match nface=%d", ntemp, nface);
        SET_STATUS(OCSM_INTERNAL_ERROR, EG_getBodyTopos);
    }

    for (iface = 1; iface <= nface; iface++) {
        body->face[iface].ibody        = -1;
        body->face[iface].iford        = -1;
        body->face[iface].imark        = -1;
        body->face[iface].gratt.object = NULL;
        body->face[iface].gratt.active = 1;              /* active */
        body->face[iface].gratt.color  = 0x00ffffa3;     /* yellow */
        body->face[iface].gratt.bcolor = 0x003f3f00;     /* dark yellow */
        body->face[iface].gratt.mcolor = 0x00bfbfbf;     /* lt grey */
        body->face[iface].gratt.lwidth = 1;
        body->face[iface].gratt.ptsize = 1;
        body->face[iface].gratt.render = 2 + 4 + 64;     /* FOREGROUND | ORIENTATION | FORWARD */
        body->face[iface].gratt.dirty  = 1;

        body->face[iface].eface = efaces[iface-1];
    }

    EG_free(efaces);

    nedge = body->nedge;
    nface = body->nface;

    /* determine the number of Edges adjacent to each Node */
    for (iedge = 1; iedge <= nedge; iedge++) {
        body->node[body->edge[iedge].ibeg].nedge++;
        body->node[body->edge[iedge].iend].nedge++;
    }

    /* set up ileft and irite on the adjacent Edges */
    for (iface = 1; iface <= nface; iface++) {
        status = EG_getTopology(body->face[iface].eface, &eref,
                                &oclass, &mtype, data, &nloop, &eloops, &senses);
        CHECK_STATUS(EG_getTopology);

        for (iloop = 0; iloop < nloop; iloop++) {
            status = EG_getTopology(eloops[iloop], &eref,
                                    &oclass, &mtype, data, &nchild, &echildren, &senses);
            CHECK_STATUS(EG_getTopology);

            for (ichild = 0; ichild < nchild; ichild++) {
                iedge = status = EG_indexBodyTopo(body->ebody, echildren[ichild]);
                CHECK_STATUS(EG_indexBodyTopo);

                if (iedge < 1 || iedge > body->nedge) {
                    signalError(MODL, OCSM_INTERNAL_ERROR, 0,
                                "echildren[%d] not found in body", ichild);
                    status = OCSM_INTERNAL_ERROR;
                    goto cleanup;
                }

                if (senses[ichild] == SFORWARD) {
                    body->edge[iedge].nface++;
                    body->edge[iedge].ileft = iface;
                } else {
                    body->edge[iedge].nface++;
                    body->edge[iedge].irite = iface;
                }
            }
        }
    }

    /* retrieve the Marks from the Edge and Face attributes */
    for (iedge = 1; iedge <= nedge; iedge++) {
        eedge = body->edge[iedge].eedge;
        status = EG_attributeRet(eedge, "mark", &itype, &nlist,
                                 &tempIlist, &tempRlist, &tempClist);

        if (status == SUCCESS && nlist > 0) {
            body->edge[iedge].imark = tempIlist[0];
        }
    }

    for (iface = 1; iface <= nface; iface++) {
        eface = body->face[iface].eface;
        status = EG_attributeRet(eface, "mark", &itype, &nlist,
                                 &tempIlist, &tempRlist, &tempClist);

        if (status == SUCCESS && nlist > 0) {
            body->face[iface].imark = tempIlist[0];
        }
    }

    /* retrieve the Branch and Face order from the Faces attributes */
    for (iface = 1; iface <= nface; iface++) {
        eface = body->face[iface].eface;
        status = EG_attributeRet(eface, "body", &itype, &nlist,
                                 &tempIlist, &tempRlist, &tempClist);

        if        (status == SUCCESS && nlist == 2) {
            body->face[iface].ibody = tempIlist[0];
            body->face[iface].iford = tempIlist[1];
        } else if (status == SUCCESS && nlist == 1) {
            body->face[iface].ibody = tempIlist[0];
            body->face[iface].iford = -1;
        } else {
            signalError(MODL, OCSM_INTERNAL_ERROR, 0,
                        "\"body\" attribute error for iface=%d (nlist=%d)", iface, nlist);
            SET_STATUS(OCSM_INTERNAL_ERROR, no_attribute);
        }

        status = EG_attributeRet(eface, "mark", &itype, &nlist,
                                 &tempIlist, &tempRlist, &tempClist);

        if (status == SUCCESS && nlist > 0) {
            body->face[iface].imark = tempIlist[0];
        }
    }

    /* set up the Branch and iford for each Edge */
    for (iedge = 1; iedge <= nedge; iedge++) {
        ileft = body->edge[iedge].ileft;
        irite = body->edge[iedge].irite;

        /* mark as error if non-manifold */
        if        (ileft < 1 && irite > 0) {
            body->edge[iedge].ibody = body->face[irite].ibody;
            body->edge[iedge].iford = -3;

        } else if (ileft > 0 && irite < 1) {
            body->edge[iedge].ibody = body->face[ileft].ibody;
            body->edge[iedge].iford = -3;

        } else if (ileft < 1 || irite < 1) {
            body->edge[iedge].ibody = -3;
            body->edge[iedge].iford = -3;

        /* if ileft and irite have same Body, then Edge has that Body too */
        } else if (body->face[ileft].ibody == body->face[irite].ibody) {
            body->edge[iedge].ibody = body->face[ileft].ibody;
            body->edge[iedge].iford = 100 * MIN(body->face[ileft].iford,
                                                body->face[irite].iford)
                                          + MAX(body->face[ileft].iford,
                                                body->face[irite].iford);

        /* search up the tree to find the first common Branch */
        } else {
            body->edge[iedge].ibody = -4;
            body->edge[iedge].iford = -4;

            /* try each Body as a candidate */
            for (ibody = 1; ibody <= MODL->nbody; ibody++) {

                /* see if you can get to ibody from ileft */
                iokayl = 0;
                ibodyl = body->face[ileft].ibody;
                while (ibodyl > 0) {
                    if (ibodyl == ibody) {
                        iokayl++;
                        break;
                    }

                    ibodyl = MODL->body[ibodyl].ichld;
                }

                /* see if you can get to ibody from irite */
                iokayr = 0;
                ibodyr = body->face[irite].ibody;
                while (ibodyr > 0) {
                    if (ibodyr == ibody) {
                        iokayr++;
                        break;
                    }

                    ibodyr = MODL->body[ibodyr].ichld;
                }

                /* success, so mark Edge with ibody and jump out of loop */
                if (iokayl == 1 && iokayr == 1) {
                    body->edge[iedge].ibody = ibody;
                    body->edge[iedge].iford = 0;
                    break;
                }
            }
        }
    }

    /* set up the Body for each Edge */
    for (iedge = 1; iedge <= nedge; iedge++) {
        ileft = body->edge[iedge].ileft;
        irite = body->edge[iedge].irite;

        /* mark as error if non-manifold */
        if        (ileft < 1 && irite > 0) {
            body->edge[iedge].ibody = body->face[irite].ibody;
            body->edge[iedge].iford = -3;

        } else if (ileft > 0 && irite < 1) {
            body->edge[iedge].ibody = body->face[ileft].ibody;
            body->edge[iedge].iford = -3;

        } else if (ileft < 1 || irite < 1) {
            body->edge[iedge].ibody = -3;
            body->edge[iedge].iford = -3;

        /* if ileft and irite have same Body, then Edge has that Body too */
        } else if (body->face[ileft].ibody == body->face[irite].ibody) {
            body->edge[iedge].ibody = body->face[ileft].ibody;

        /* search up the tree to find the first common Body */
        } else {
            body->edge[iedge].ibody = -4;

            /* try each Body as a candidate */
            for (ibody = 1; ibody <= MODL->nbody; ibody++) {

                /* see if you can get to ibody from ileft */
                iokayl = 0;
                ibodyl = body->face[ileft].ibody;
                while (ibodyl > 0) {
                    if (ibodyl == ibody) {
                        iokayl++;
                        break;
                    }

                    ibodyl = MODL->body[ibodyl].ichld;
                }

                /* see if you can get to ibody from irite */
                iokayr = 0;
                ibodyr = body->face[irite].ibody;
                while (ibodyr > 0) {
                    if (ibodyr == ibody) {
                        iokayr++;
                        break;
                    }

                    ibodyr = MODL->body[ibodyr].ichld;
                }

                /* success, so mark Edge with ibody and jump out of loop */
                if (iokayl == 1 && iokayr == 1) {
                    body->edge[iedge].ibody = ibody;
                    break;
                }
            }
        }
    }

    /* classify Edges as interior (green) if iford!=0.  otherwise, the edge is exterior (blue) */
    for (iedge = 1; iedge <= nedge; iedge++) {
        if (body->edge[iedge].iford != 0) {
            body->edge[iedge].gratt.color = 0x0000ff00;          /* green */
        } else {
            body->edge[iedge].gratt.color = 0x000000ff;          /* blue  */
        }
    }

    /* set up the Body for each Node */
#ifndef __clang_analyzer__
    for (iedge = 1; iedge <= nedge; iedge++) {
        ibeg = body->edge[iedge].ibeg;
        iend = body->edge[iedge].iend;

        if (body->edge[iedge].ibody > body->node[ibeg ].ibody) {
            body->node[ibeg ].ibody = body->edge[iedge].ibody;
        }

        if (body->edge[iedge].ibody > body->node[iend ].ibody) {
            body->node[iend ].ibody = body->edge[iedge].ibody;
        }
    }
#endif

    /* store the Body and Face order as attributes for each Edge */
    for (iedge = 1; iedge <= nedge; iedge++) {
        eedge = body->edge[iedge].eedge;
        iattrib[0] = body->edge[iedge].ibody;
        iattrib[1] = body->edge[iedge].iford;

        status = EG_attributeAdd(eedge, "body", ATTRINT, 2,
                                 iattrib, NULL, NULL);
        CHECK_STATUS(EG_attributeAdd);
    }

    /* report un-attributed Edges and Faces */
    for (iedge = 1; iedge <= nedge; iedge++) {
        if        (body->edge[iedge].iford == -3   ) {

        } else if (body->edge[iedge].ibody <=  0 ||
                   body->edge[iedge].iford <   0   ) {
            SPRINT3(0, "WARNING:: Edge %3d has .ibody=%d  .iford=%d",
                    iedge, body->edge[iedge].ibody,
                           body->edge[iedge].iford);
        }
    }

    for (iface = 1; iface <= nface; iface++) {
        if (body->face[iface].ibody <= 0 ||
            body->face[iface].iford <  0   ) {
            SPRINT3(0, "WARNING:: Face %3d has .ibody=%d  .iford=%d",
                    iface, body->face[iface].ibody,
                           body->face[iface].iford);
        }
    }

    /* remove the faceID from all Faces and edgeID from all Edges */
    old_outLevel = EG_setOutLevel(MODL->context, 0);
    for (iface = 1; iface <= nface; iface++) {
        status = EG_attributeDel(body->face[iface].eface, "faceID");
    }
    for (iedge = 1; iedge <= nedge; iedge++) {
        status = EG_attributeDel(body->edge[iedge].eedge, "edgeID");
    }
    (void) EG_setOutLevel(MODL->context, old_outLevel);

    /* add the faceID attribute to all Faces */
    for (iface = 1; iface <= nface; iface++) {
        status = EG_attributeRet(body->face[iface].eface, "body",
                                 &itype, &nlist,
                                 &tempIlist, &tempRlist, &tempClist);
        CHECK_STATUS(EG_attributeRet);

        iattrib[0] = tempIlist[0];
        iattrib[1] = tempIlist[1];
        iattrib[2] = 1;

        for (jface = 1; jface < iface; jface++) {
            status = EG_attributeRet(body->face[jface].eface, "faceID",
                                     &itype, &nlist,
                                     &tempIlist, &tempRlist, &tempClist);
            CHECK_STATUS(EG_attributeRet);

            if (tempIlist[0] == iattrib[0] &&
                tempIlist[1] == iattrib[1]   ) {
                iattrib[2] ++;
            }
        }

        status = EG_attributeAdd(body->face[iface].eface, "faceID",
                                 ATTRINT, 3, iattrib, NULL, NULL);
        CHECK_STATUS(EG_attributeAdd);
    }

    /* reorder the faceIDs based upon cg location when sequence > 1 */
    nchange = 1;
    while (nchange > 0) {
        nchange = 0;

        for (iface = 1; iface <= nface; iface++) {
            status = EG_attributeRet(body->face[iface].eface, "faceID",
                                     &itype, &nlist,
                                     &tempIlist, &tempRlist, &tempClist);
            CHECK_STATUS(EG_attributeRet);

            if (tempIlist[2] > 1) {
                iattrib[0] = tempIlist[0];
                iattrib[1] = tempIlist[1];
                iattrib[2] = tempIlist[2];

                for (jface = 1; jface <= nface; jface++) {
                    status = EG_attributeRet(body->face[jface].eface, "faceID",
                                             &itype, &nlist,
                                             &tempIlist, &tempRlist, &tempClist);
                    CHECK_STATUS(EG_attributeRet);

                    if (tempIlist[0] == iattrib[0]  &&
                        tempIlist[1] == iattrib[1]  &&
                        tempIlist[2] == iattrib[2]-1  ) {
                        status = EG_getMassProperties(body->face[jface].eface, dataj);
                        CHECK_STATUS(EG_getMassProperties);

                        status = EG_getMassProperties(body->face[iface].eface, datai);
                        CHECK_STATUS(EG_getMassProperties);

                        iswap = 0;
                        if        (fabs(datai[2]-dataj[2]) > EPS06) {
                            if (datai[2] < dataj[2]) {
                                iswap = 1;
                            }
                        } else if (fabs(datai[3]-dataj[3]) > EPS06) {
                            if (datai[3] < dataj[3]) {
                                iswap = 1;
                            }
                        } else if (fabs(datai[4]-dataj[4]) > EPS06) {
                            if (datai[4] < dataj[4]) {
                                iswap = 1;
                            }
                        } else if (datai[1] < dataj[1]) {
                            iswap = 1;
                        }

                        if (iswap > 0) {
                            status = EG_attributeAdd(body->face[jface].eface, "faceID",
                                                     ATTRINT, 3, iattrib, NULL, NULL);
                            CHECK_STATUS(EG_attributeAdd);

                            iattrib[2] --;

                            status = EG_attributeAdd(body->face[iface].eface, "faceID",
                                                     ATTRINT, 3, iattrib, NULL, NULL);
                            CHECK_STATUS(EG_attributeAdd);

                            nchange++;
                        }
                    }
                }
            }
        }
    }

    /* add the edgeID attribute to all Edges */
    for (iedge = 1; iedge <= nedge; iedge++) {
        ileft = body->edge[iedge].ileft;
        if (ileft >= 0) {
            status = EG_attributeRet(body->face[ileft].eface, "body",
                                     &itype, &nlist,
                                     &tempIlist, &tempRlist, &tempClist);
            CHECK_STATUS(EG_attributeRet);

            iattrib[0] = tempIlist[0];
            iattrib[1] = tempIlist[1];
        } else {
            iattrib[0] = 0;
            iattrib[1] = 0;
        }

        irite = body->edge[iedge].irite;
        if (irite >= 0) {
            status = EG_attributeRet(body->face[irite].eface, "body",
                                     &itype, &nlist,
                                     &tempIlist, &tempRlist, &tempClist);
            CHECK_STATUS(EG_attributeRet);

            iattrib[2] = tempIlist[0];
            iattrib[3] = tempIlist[1];
        } else {
            iattrib[2] = 0;
            iattrib[3] = 0;
        }

        if (100*iattrib[0]+iattrib[1] > 100*iattrib[2]+iattrib[3]) {
            iswap      = iattrib[0];
            iattrib[0] = iattrib[2];
            iattrib[2] = iswap;

            iswap      = iattrib[1];
            iattrib[1] = iattrib[3];
            iattrib[3] = iswap;
        }

        iattrib[4] = 1;

        for (jedge = 1; jedge < iedge; jedge++) {
            status = EG_attributeRet(body->edge[jedge].eedge, "edgeID",
                                     &itype, &nlist,
                                     &tempIlist, &tempRlist, &tempClist);
            CHECK_STATUS(EG_attributeRet);

            if (tempIlist[0] == iattrib[0] &&
                tempIlist[1] == iattrib[1] &&
                tempIlist[2] == iattrib[2] &&
                tempIlist[3] == iattrib[3]   ) {
                iattrib[4] ++;
            }
        }

        status = EG_attributeAdd(body->edge[iedge].eedge, "edgeID",
                                 ATTRINT, 5, iattrib, NULL, NULL);
        CHECK_STATUS(EG_attributeAdd);
    }

    /* reorder the edgeIDs based upon cg location when sequence > 1 */
    nchange = 1;
    while (nchange > 0) {
        nchange = 0;

        for (iedge = 1; iedge <= nedge; iedge++) {
            status = EG_attributeRet(body->edge[iedge].eedge, "edgeID",
                                     &itype, &nlist,
                                     &tempIlist, &tempRlist, &tempClist);
            CHECK_STATUS(EG_attributeRet);

            if (tempIlist[2] > 1) {
                iattrib[0] = tempIlist[0];
                iattrib[1] = tempIlist[1];
                iattrib[2] = tempIlist[2];
                iattrib[3] = tempIlist[3];
                iattrib[4] = tempIlist[4];

                for (jedge = 1; jedge <= nedge; jedge++) {
                    status = EG_attributeRet(body->edge[jedge].eedge, "edgeID",
                                             &itype, &nlist,
                                             &tempIlist, &tempRlist, &tempClist);
                    CHECK_STATUS(EG_attributeRet);

                    if (tempIlist[0] == iattrib[0]  &&
                        tempIlist[1] == iattrib[1]  &&
                        tempIlist[2] == iattrib[2]  &&
                        tempIlist[3] == iattrib[3]  &&
                        tempIlist[4] == iattrib[4]-1  ) {
                        status = EG_getMassProperties(body->edge[jedge].eedge, dataj);
                        CHECK_STATUS(EG_getMassProperties);

                        status = EG_getMassProperties(body->edge[iedge].eedge, datai);
                        CHECK_STATUS(EG_getMassProperties);

                        iswap = 0;
                        if        (fabs(datai[2]-dataj[2]) > EPS06) {
                            if (datai[2] < dataj[2]) {
                                iswap = 1;
                            }
                        } else if (fabs(datai[3]-dataj[3]) > EPS06) {
                            if (datai[3] < dataj[3]) {
                                iswap = 1;
                            }
                        } else if (fabs(datai[4]-dataj[4]) > EPS06) {
                            if (datai[4] < dataj[4]) {
                                iswap = 1;
                            }
                        } else if (datai[1] < dataj[1]) {
                            iswap = 1;
                        }

                        if (iswap > 0) {
                            status = EG_attributeAdd(body->edge[jedge].eedge, "edgeID",
                                                     ATTRINT, 5, iattrib, NULL, NULL);
                            CHECK_STATUS(EG_attributeAdd);

                            iattrib[4] --;

                            status = EG_attributeAdd(body->edge[iedge].eedge, "edgeID",
                                                     ATTRINT, 5, iattrib, NULL, NULL);
                            CHECK_STATUS(EG_attributeAdd);

                            nchange++;
                        }
                    }
                }
            }
        }
    }

    /* reset selection info */
    MODL->sel_body = MODL->nbody;
    MODL->sel_node = -1;
    MODL->sel_edge = -1;
    MODL->sel_face = -1;

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   finiteDifference - get derivatives by finite differences           *
 *                                                                      *
 ************************************************************************
 */

static int
finiteDifference(modl_T *modl,
                 int    ibody,
                 int    seltype,
                 int    iselect,
                 int    npnt,
       /*@null@*/double uv[],
                 double dxyz[])
{
    int       status = SUCCESS;         /* (out) return status */

    int       inode, iedge, iface;

    modl_T    *MODL = (modl_T*)modl;

    int       npnt_base, npnt_ptrb, ntri_base, ntri_ptrb, ip0, ip1, ip2, itri, ipnt;
    int       oclass, mtype, nchild, *senses, itype, nlist, jnode;
    CINT      *tris_base, *tris_ptrb, *tric_base, *tric_ptrb;
    CINT      *ptype_base, *ptype_ptrb, *pindx_base, *pindx_ptrb;
    CINT      *tempIlist;
    double    uv_face[2], weights[3], data_base[18], data_ptrb[18];
    CDOUBLE   *xyz_base, *xyz_ptrb, *uv_base, *uv_ptrb;
    ego       eref, *echilds;

    ROUTINE(finiteDifference);
    DPRINT5("%s(ibody=%d, seltype=%d, iselect=%d, npnt=%d) {",
            routine, ibody, seltype, iselect, npnt);

    /* --------------------------------------------------------------- */

    if (MODL->dtime == -2) {
        for (ipnt = 0; ipnt < npnt; ipnt++) {
            dxyz[3*ipnt  ] = 0;
            dxyz[3*ipnt+1] = 0;
            dxyz[3*ipnt+2] = 0;
        }
    } else if (seltype == OCSM_FACE) {
        iface = iselect;

        /* get the base and perturbed tessellations */
        status = EG_getTessFace(MODL->body[ibody].etess, iface,
                                &npnt_base, &xyz_base, &uv_base, &ptype_base, &pindx_base,
                                &ntri_base, &tris_base, &tric_base);
        CHECK_STATUS(EG_getTessFace);

        status = EG_getTessFace(MODL->perturb->body[ibody].etess, -iface,
                                &npnt_ptrb, &xyz_ptrb, &uv_ptrb, &ptype_ptrb, &pindx_ptrb,
                                &ntri_ptrb, &tris_ptrb, &tric_ptrb);
        CHECK_STATUS(EG_getTessFace);

        /* if using tessellation points, just take finite difference of locations */
        if (uv == NULL) {
            for (ipnt = 0; ipnt < npnt; ipnt++) {
                dxyz[3*ipnt  ] = (xyz_ptrb[3*ipnt  ] - xyz_base[3*ipnt  ]) / MODL->dtime;
                dxyz[3*ipnt+1] = (xyz_ptrb[3*ipnt+1] - xyz_base[3*ipnt+1]) / MODL->dtime;
                dxyz[3*ipnt+2] = (xyz_ptrb[3*ipnt+2] - xyz_base[3*ipnt+2]) / MODL->dtime;
            }

        /* otherwise, interpolate into the tessellation and then take finite differences */
        } else {
            for (ipnt = 0; ipnt < npnt; ipnt++) {
                uv_face[0] = uv[2*ipnt  ];
                uv_face[1] = uv[2*ipnt+1];

                status = EG_locateTessBody(MODL->body[ibody].etess, 1,
                                           &iface, uv_face, &itri, weights);
                CHECK_STATUS(EG_locateTessBody);

                ip0 = tris_base[3*(itri-1)  ] - 1;
                ip1 = tris_base[3*(itri-1)+1] - 1;
                ip2 = tris_base[3*(itri-1)+2] - 1;

                dxyz[3*ipnt  ] = ((xyz_ptrb[3*ip0  ] * weights[0]
                                  +xyz_ptrb[3*ip1  ] * weights[1]
                                  +xyz_ptrb[3*ip2  ] * weights[2])
                                 -(xyz_base[3*ip0  ] * weights[0]
                                  +xyz_base[3*ip1  ] * weights[1]
                                  +xyz_base[3*ip2  ] * weights[2])) / MODL->dtime;
                dxyz[3*ipnt+1] = ((xyz_ptrb[3*ip0+1] * weights[0]
                                  +xyz_ptrb[3*ip1+1] * weights[1]
                                  +xyz_ptrb[3*ip2+1] * weights[2])
                                 -(xyz_base[3*ip0+1] * weights[0]
                                  +xyz_base[3*ip1+1] * weights[1]
                                  +xyz_base[3*ip2+1] * weights[2])) / MODL->dtime;
                dxyz[3*ipnt+2] = ((xyz_ptrb[3*ip0+2] * weights[0]
                                  +xyz_ptrb[3*ip1+2] * weights[1]
                                  +xyz_ptrb[3*ip2+2] * weights[2])
                                 -(xyz_base[3*ip0+2] * weights[0]
                                  +xyz_base[3*ip1+2] * weights[1]
                                  +xyz_base[3*ip2+2] * weights[2])) / MODL->dtime;
            }
        }
    } else if (seltype == OCSM_EDGE) {
        iedge = iselect;

        /* get the base and perturbed tessellations */
        status = EG_getTessEdge(MODL->body[ibody].etess, iedge,
                                &npnt_base, &xyz_base, &uv_base);
        CHECK_STATUS(EG_getTessEdge);

        status = EG_getTessEdge(MODL->perturb->body[ibody].etess, -iedge,
                                &npnt_ptrb, &xyz_ptrb, &uv_ptrb);
        CHECK_STATUS(EG_getTessEdge);

        /* if using tessellation points, just take finite difference of locations */
        if (uv == NULL) {
            for (ipnt = 0; ipnt < npnt; ipnt++) {
                dxyz[3*ipnt  ] = (xyz_ptrb[3*ipnt  ] - xyz_base[3*ipnt  ]) / MODL->dtime;
                dxyz[3*ipnt+1] = (xyz_ptrb[3*ipnt+1] - xyz_base[3*ipnt+1]) / MODL->dtime;
                dxyz[3*ipnt+2] = (xyz_ptrb[3*ipnt+2] - xyz_base[3*ipnt+2]) / MODL->dtime;
            }

            /* otherwise, interpolate into the tessellation and then take finite differences */
        } else {
            for (ipnt = 0; ipnt < npnt; ipnt++) {
                status = EG_evaluate(MODL->body[ibody].edge[iedge].eedge,
                                     &(uv[ipnt]), data_base);
                CHECK_STATUS(EG_evaluate);

                status = EG_evaluate(MODL->perturb->body[ibody].edge[iedge].eedge,
                                     &(uv[ipnt]), data_ptrb);
                CHECK_STATUS(EG_evaluate);

                dxyz[3*ipnt  ] = (data_ptrb[0] - data_base[0]) / MODL->dtime;
                dxyz[3*ipnt+1] = (data_ptrb[1] - data_base[1]) / MODL->dtime;
                dxyz[3*ipnt+2] = (data_ptrb[2] - data_base[2]) / MODL->dtime;
            }
        }

    } else if (seltype == OCSM_NODE) {
        inode = iselect;

        /* if .nMap exists, use mapped Node location in the perturbed Body */
        status = EG_attributeRet(MODL->perturb->body[ibody].ebody, ".nMap",
                                 &itype, &nlist, &tempIlist, NULL, NULL);
        if (status == SUCCESS) {
            jnode = tempIlist[inode-1];
        } else {
            jnode = inode;
        }

        /* get the two locations */
        status = EG_getTopology(MODL->body[ibody].node[inode].enode, &eref,
                                &oclass, &mtype, data_base, &nchild, &echilds, &senses);
        CHECK_STATUS(EG_getTopology);

        status = EG_getTopology(MODL->perturb->body[ibody].node[jnode].enode, &eref,
                                &oclass, &mtype, data_ptrb, &nchild, &echilds, &senses);
        CHECK_STATUS(EG_getTopology);

        dxyz[0] = (data_ptrb[0] - data_base[0]) / MODL->dtime;
        dxyz[1] = (data_ptrb[1] - data_base[1]) / MODL->dtime;
        dxyz[2] = (data_ptrb[2] - data_base[2]) / MODL->dtime;

        /* bad value for seltype */
    } else {
        status = OCSM_ILLEGAL_ARGUMENT;
        goto cleanup;
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   getBodyTolerance - get the largest tolerance associated with       *
 *                      ebody and its Edges/Faces                       *
 *                                                                      *
 ************************************************************************
 */

static int
getBodyTolerance(ego    ebody,          /* (in)  pointer to Body */
                 double *toler)         /* (out) largest tolerance */
{
    int       status = 0;               /* (out) body that satisfies criterion (bias-0) */

    int       nface, nedge, iface, iedge, oclass, mtype;
    double    tol;

    ego       *efaces, *eedges, eref, eprev, enext;

    ROUTINE(getBodyTolerance);
    DPRINT1("%s() {",
            routine);

    /* --------------------------------------------------------------- */

    *toler = 0;

    /* if ebody is really a MODEL, return a tolerance of -1 */
    status = EG_getInfo(ebody, &oclass, &mtype, &eref, &eprev, &enext);
    CHECK_STATUS(EG_getInfo);

    if (oclass == MODEL) {
        *toler = -1;
        goto cleanup;
    }

    /* get tolerance associated with the Body */
    status = EG_getTolerance(ebody, &tol);
    CHECK_STATUS(EG_getTolerance);

    SPRINT1(3, "    body         toler=%11.4e", tol);

    *toler = MAX(*toler, tol);

    /* for each Face associated with ebody, get its tolerance */
    status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
    CHECK_STATUS(EG_getBodyTopos);

    for (iface = 0; iface < nface; iface++) {
        status = EG_getTolerance(efaces[iface], &tol);
        CHECK_STATUS(EG_getTolerance);

        SPRINT2(3, "    iface=%-5d  toler=%11.4e", iface+1, tol);

        *toler = MAX(*toler, tol);
    }

    EG_free(efaces);

    /* for each Edge associated with ebody, get its tolerance */
    status = EG_getBodyTopos(ebody, NULL, EDGE, &nedge, &eedges);
    CHECK_STATUS(EG_getBodyTopos);

    for (iedge = 0; iedge < nedge; iedge++) {
        status = EG_getTolerance(eedges[iedge], &tol);
        CHECK_STATUS(EG_getTolerance);

        SPRINT2(3, "    iedge=%-5d  toler=%11.4e", iedge+1, tol);

        *toler = MAX(*toler, tol);
    }

    EG_free(eedges);

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/***********************************************************************/
/*                                                                     */
/*   getToken - get a token from a string                              */
/*                                                                     */
/***********************************************************************/

static int
getToken(char   *text,                  /* (in)  full text */
         int    nskip,                  /* (in)  tokens to skip */
         char   sep,                    /* (in)  seperator character */
         char   *token)                 /* (out) token */
{
    int    lentok, i, count, iskip;

    token[0] = '\0';
    lentok   = 0;

    /* count the number of separators */
    count = 0;
    for (i = 0; i < strlen(text); i++) {
        if (text[i] == sep) {
            count++;
        }
    }

    if (count < nskip+1) return 0;

    /* skip over nskip tokens */
    i = 0;
    for (iskip = 0; iskip < nskip; iskip++) {
        while (text[i] != sep) {
            i++;
        }
        i++;
    }

    /* extract the token we are looking for */
    while (text[i] != sep) {
        token[lentok++] = text[i++];
        token[lentok  ] = '\0';

        if (lentok >= MAX_EXPR_LEN-1) {
            signalError(NULL, OCSM_INTERNAL_ERROR, 0,
                        "token exceeds MAX_EXPR_LEN");
            break;
        }
    }

    return strlen(token);
}


/***********************************************************************/
/*                                                                     */
/*   makeEdge - make an Edge between eNodes                            */
/*                                                                     */
/***********************************************************************/

static int
makeEdge(modl_T *modl,                  /* (in)  pointer to MODL */
         ego    ebeg,                   /* (in)  enode at beginning */
         ego    eend,                   /* (in)  enode at end */
         ego    *eedge)                 /* (out) created Edge */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       oclass, mtype, nchild, *senses;
    double    xyz_beg[4], xyz_end[4], xyz_line[6], tdata[2], result[3];
    ego       eref, ecurve, *echilds, enodes[2];

    ROUTINE(makeEdge);
    DPRINT1("%s() {",
            routine);

    /* --------------------------------------------------------------- */

    /* get the coordinates of the two Nodes */
    status = EG_getTopology(ebeg, &eref, &oclass, &mtype,
                            xyz_beg, &nchild, &echilds, &senses);
    CHECK_STATUS(EG_getTopology);

    status = EG_getTopology(eend, &eref, &oclass, &mtype,
                            xyz_end, &nchild, &echilds, &senses);
    CHECK_STATUS(EG_getTopology);

    /* create a line between the Nodes */
    xyz_line[0] = xyz_beg[0];
    xyz_line[1] = xyz_beg[1];
    xyz_line[2] = xyz_beg[2];
    xyz_line[3] = xyz_end[0] - xyz_beg[0];
    xyz_line[4] = xyz_end[1] - xyz_beg[1];
    xyz_line[5] = xyz_end[2] - xyz_beg[2];

    status = EG_makeGeometry(MODL->context, CURVE, LINE, NULL, NULL, xyz_line, &ecurve);
    CHECK_STATUS(EG_makeGeometry);

    /* find the parametric coordinates of the endpoints */
    status = EG_invEvaluate(ecurve, xyz_beg, &(tdata[0]), result);
    CHECK_STATUS(EG_invEvaluate);

    status = EG_invEvaluate(ecurve, xyz_end, &(tdata[1]), result);
    CHECK_STATUS(EG_invEvaluate);

    /* make the Edge */
    enodes[0] = ebeg;
    enodes[1] = eend;
    status = EG_makeTopology(MODL->context, ecurve, EDGE, TWONODE,
                             tdata, 2, enodes, NULL, eedge);
    CHECK_STATUS(EG_makeTopology);

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/***********************************************************************/
/*                                                                     */
/*   makeFace - make a Face from an array of Edges                     */
/*                                                                     */
/***********************************************************************/

static int
makeFace(modl_T *modl,                  /* (in)  pointer to MODL */
         ego    eedges_in[],            /* (in)  array of 4 Edges */
         int    fillstyle,              /* (in)  fillstyle argument to EG_isoCline */
         ego    *eface)                 /* (out) created Face */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       iedge, nedge, senses[8];
    ego       eedges[8], eloop, esurface;

    ROUTINE(makeFace);
    DPRINT2("%s(fillstyle=%d) {",
            routine, fillstyle);

    /* --------------------------------------------------------------- */

    /* copy the eedges_in array into (the local) eedges */
    eedges[0] = eedges_in[0];
    eedges[1] = eedges_in[1];
    eedges[2] = eedges_in[2];
    eedges[3] = eedges_in[3];

    if (eedges[3] == NULL) {
        nedge = 3;
    } else {
        nedge = 4;
    }

    senses[0] = SFORWARD;
    senses[1] = SFORWARD;
    senses[2] = SFORWARD;
    senses[3] = SFORWARD;

    /* create a Loop from the Edges */
    status = EG_makeTopology(MODL->context, NULL, LOOP, CLOSED,
                             NULL, nedge, eedges, senses, &eloop);
    CHECK_STATUS(EG_makeTopology);

    /* create a Surface from the Loop (using the overloaded EG_isoCline) */
    status = EG_isoCline(eloop, fillstyle, 0.0, &esurface);
    CHECK_STATUS(EG_isoCline);

    status = EG_deleteObject(eloop);
    CHECK_STATUS(EG_deleteObject);

    /* update the Loop to include the Surface and PCurves */
    for (iedge = 0; iedge < nedge; iedge++) {
        status = EG_otherCurve(esurface, eedges[iedge], 0.0, &(eedges[iedge+nedge]));
        CHECK_STATUS(EG_otherCurve);
    }

    /* make a new Loop and a Face using this Surface */
    status = EG_makeTopology(MODL->context, esurface, LOOP, CLOSED,
                             NULL, nedge, eedges, senses, &eloop);
    CHECK_STATUS(EG_makeTopology);

    status = EG_makeTopology(MODL->context, esurface, FACE, SFORWARD,
                             NULL, 1, &eloop, senses, eface);
    CHECK_STATUS(EG_makeTopology);

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   matsol - Gaussian elimination with partial pivoting                *
 *                                                                      *
 ************************************************************************
 */

static int
matsol(double    A[],                   /* (in)  matrix to be solved (stored rowwise) */
                                        /* (out) upper-triangular form of matrix */
       double    b[],                   /* (in)  right hand side */
                                        /* (out) right-hand side after swapping */
       int       n,                     /* (in)  size of matrix */
       double    x[])                   /* (out) solution of A*x=b */
{
    int       status = SUCCESS;         /* (out) return status */

    int       ir, jc, kc, imax;
    double    amax, swap, fact;

    ROUTINE(matsol);
    DPRINT2("%s(n=%d) {",
            routine, n);

    /* --------------------------------------------------------------- */

    /* reduce each column of A */
    for (kc = 0; kc < n; kc++) {

        /* find pivot element */
        imax = kc;
        amax = fabs(A[kc*n+kc]);

        for (ir = kc+1; ir < n; ir++) {
            if (fabs(A[ir*n+kc]) > amax) {
                imax = ir;
                amax = fabs(A[ir*n+kc]);
            }
        }

        /* check for possibly-singular matrix (ie, near-zero pivot) */
        if (amax < EPS12) {
            status = OCSM_SINGULAR_MATRIX;
            goto cleanup;
        }

        /* if diagonal is not pivot, swap rows in A and b */
        if (imax != kc) {
            for (jc = 0; jc < n; jc++) {
                swap         = A[kc  *n+jc];
                A[kc  *n+jc] = A[imax*n+jc];
                A[imax*n+jc] = swap;
            }

            swap    = b[kc  ];
            b[kc  ] = b[imax];
            b[imax] = swap;
        }

        /* row-reduce part of matrix to the bottom of and right of [kc,kc] */
        for (ir = kc+1; ir < n; ir++) {
            fact = A[ir*n+kc] / A[kc*n+kc];

            for (jc = kc+1; jc < n; jc++) {
                A[ir*n+jc] -= fact * A[kc*n+jc];
            }

            b[ir] -= fact * b[kc];

            A[ir*n+kc] = 0;
        }
    }

    /* back-substitution pass */
    x[n-1] = b[n-1] / A[(n-1)*n+(n-1)];

    for (jc = n-2; jc >= 0; jc--) {
        x[jc] = b[jc];
        for (kc = jc+1; kc < n; kc++) {
            x[jc] -= A[jc*n+kc] * x[kc];
        }
        x[jc] /= A[jc*n+jc];
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   newBody - create and initialize a new Body                         *
 *                                                                      *
 ************************************************************************
 */

static int
newBody(modl_T *modl,                   /* (in)  pointer to MODL */
        int    ibrch,                   /* (in)  Branch index (1-mbrch) */
        int    brtype,                  /* (in)  Branch type */
        int    ileft,                   /* (in)  left parent Body (or 0) */
        int    irite,                   /* (in)  rite parent Body (or 0) */
        varg_T args[],                  /* (in)  array of arguments */
        int    hasdots,                 /* (in)  =1 if any arguments have non-zero dots */
        int    botype,                  /* (in)  Body type */
        int    *ibody)                  /* (out) new Body index */
{
    int       status = SUCCESS;         /* (out) return status */

    int       iarg, ival;

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(newBody);
    DPRINT6("%s(ibrch=%d, brtype=%d, ileft=%d, irite=%d, botype=%d) {",
            routine, ibrch, brtype, ileft, irite, botype);

    /* --------------------------------------------------------------- */

    /* extend the Body list (if needed) */
    if (MODL->nbody >= MODL->mbody) {
        MODL->mbody += 25;
        if (MODL->body == NULL) {
            MALLOC(MODL->body, body_T, MODL->mbody+1);
        } else {
            RALLOC(MODL->body, body_T, MODL->mbody+1);
        }
    }

    /* create the new Body and initialize it */
    MODL->nbody += 1;

    *ibody = MODL->nbody;
    MODL->body[*ibody].ibrch  = ibrch;
    MODL->body[*ibody].brtype = brtype;
    MODL->body[*ibody].ileft  = ileft;
    MODL->body[*ibody].irite  = irite;
    MODL->body[*ibody].ichld  = 0;

    for (iarg = 1; iarg < 10; iarg++) {
        MODL->body[*ibody].arg[iarg].nval = args[iarg].nval;
        MODL->body[*ibody].arg[iarg].nrow = args[iarg].nrow;
        MODL->body[*ibody].arg[iarg].ncol = args[iarg].ncol;
        MODL->body[*ibody].arg[iarg].val  = NULL;
        MODL->body[*ibody].arg[iarg].dot  = NULL;

        if (args[iarg].nval == 0) {
            MALLOC(MODL->body[*ibody].arg[iarg].str, char, (int)(strlen(args[iarg].str)+1));

            strcpy(MODL->body[*ibody].arg[iarg].str, args[iarg].str);
        } else if (args[iarg].nval > 0) {
            MALLOC(MODL->body[*ibody].arg[iarg].val, double, args[iarg].nval);
            MALLOC(MODL->body[*ibody].arg[iarg].dot, double, args[iarg].nval);

            for (ival = 0; ival < args[iarg].nval; ival++) {
                MODL->body[*ibody].arg[iarg].val[ival] = args[iarg].val[ival];
                MODL->body[*ibody].arg[iarg].dot[ival] = args[iarg].dot[ival];
            }
        }
    }

    MODL->body[*ibody].ebody = NULL;
    MODL->body[*ibody].etess = NULL;

    MODL->body[*ibody].onstack = 0;
    MODL->body[*ibody].hasdots = hasdots;
    MODL->body[*ibody].botype  = botype;
    MODL->body[*ibody].nnode   = 0;
    MODL->body[*ibody].node    = NULL;
    MODL->body[*ibody].nedge   = 0;
    MODL->body[*ibody].edge    = NULL;
    MODL->body[*ibody].nface   = 0;
    MODL->body[*ibody].face    = NULL;

    /* link children */
    if (ileft > 0) {
        MODL->body[ileft].ichld = *ibody;
    }
    if (irite > 0) {
        MODL->body[irite].ichld = *ibody;
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   parseName - parse name into Parameter name, row, and column        *
 *                                                                      *
 ************************************************************************
 */

static int
parseName(modl_T *modl,                 /* (in)  pointer to MODL */
          char   name[],                /* (in)  name to parse */
          int    *ipmtr,                /* (out) Parameter index (1-npmtr) */
          int    *irow,                 /* (out) row index (bias-1) */
          int    *icol)                 /* (int) column index (bias-1) */
{
    int       status = 0;               /* (out) body that satisfies criterion (bias-0) */

    modl_T    *MODL = (modl_T*)modl;

    int       i, icount, jcount, jpmtr;
    double    rownum, colnum, dot;
    char      pmtrName[MAX_EXPR_LEN], row[MAX_EXPR_LEN], col[MAX_EXPR_LEN];


    ROUTINE(parseName);
    DPRINT2("%s(name=%s) {",
            routine, name);

    /* --------------------------------------------------------------- */

    /* peel off each part of name[irow,icol] */
    pmtrName[0] = '\0';
    row[0]      = '\0';
    col[0]      = '\0';
    icount      = 0;
    jcount      = 0;

    for (i = 0; i < strlen(name); i++) {
        if (icount == 0) {
            if (name[i] != '[') {
                pmtrName[jcount  ] = name[i];
                pmtrName[jcount+1] = '\0';
                jcount++;
            } else {
                icount++;
                jcount = 0;
            }
        } else if (icount == 1) {
            if (name[i] != ',') {
                row[jcount  ] = name[i];
                row[jcount+1] = '\0';
                jcount++;
            } else {
                icount++;
                jcount = 0;
            }
        } else {
            if (name[i] != ']') {
                col[jcount  ] = name[i];
                col[jcount+1] = '\0';
                jcount++;
            } else {
                icount++;
                break;
            }
        }
    }

    /* return error if not in specified form */
    if (icount != 0 && icount != 3) {
        status = OCSM_ILLEGAL_PMTR_NAME;
        goto cleanup;
    }

    /* convert row and col to an integer */
    if (strlen(row) == 0) {
        *irow = 1;
    } else {
        status = str2val(row, MODL, &rownum, &dot);
        CHECK_STATUS(str2val:row);

        *irow = NINT(rownum);
    }

    if (strlen(col) == 0) {
        *icol = 1;
    } else {
        status = str2val(col, MODL, &colnum, &dot);
        CHECK_STATUS(str2val:col);

        *icol = NINT(colnum);
    }

    /* look for current Parameter index */
    *ipmtr = 0;
    for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
        if (strcmp(MODL->pmtr[jpmtr].name, pmtrName) == 0 &&
                   MODL->pmtr[jpmtr].level == MODL->level   ) {
            *ipmtr = jpmtr;
            break;
        }
    }

    /* create new Parameter (if needed) */
    if (*ipmtr == 0) {
        status = ocsmNewPmtr(MODL, pmtrName, OCSM_INTERNAL, 1, 1);
        CHECK_STATUS(ocsmNewPmtr);

        *ipmtr = MODL->npmtr;
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   printBodyAttributes - prints attributes associated with an ebody   *
 *                                                                      *
 ************************************************************************
 */

static int
printBodyAttributes(modl_T *modl,       /* (in)  pointer to MODL */
                    int    ibody)       /* (in)  Body index (1-nbody) */
{
    int       status = 0;               /* (out) body that satisfies criterion (bias-0) */

    modl_T    *MODL = (modl_T*)modl;

    int       nlist, nattr, iattr, nface, nedge, iface, iedge;

    int       itype, i;
    CINT      *tempIlist;
    CDOUBLE   *tempRlist;
    CCHAR     *attrName, *tempClist;
    ego       ebody, *efaces, *eedges;

    ROUTINE(printBodyAttributes);
    DPRINT1("%s() {",
            routine);

    /* --------------------------------------------------------------- */

    /* get attributes associated with the Body */
    ebody = MODL->body[ibody].ebody;

    status = EG_attributeNum(ebody, &nattr);
    CHECK_STATUS(EG_attributeNum);

    for (iattr = 1; iattr <= nattr; iattr++) {
        status = EG_attributeGet(ebody, iattr, &attrName, &itype, &nlist,
                                 &tempIlist, &tempRlist, &tempClist);
        CHECK_STATUS(EG_attributeGet);

        SPRINT1x(3, "        %-20s =", attrName);

        if        (itype == ATTRINT) {
            for (i = 0; i < nlist; i++) {
                SPRINT1x(3, "%5d ", tempIlist[i]);
            }
            SPRINT0(3, " ");
        } else if (itype == ATTRREAL) {
            for (i = 0; i < nlist; i++) {
                SPRINT1x(3, "%11.5f ", tempRlist[i]);
            }
            SPRINT0(3, " ");
        } else if (itype == ATTRSTRING) {
            SPRINT1(3, "%s", tempClist);
        }
    }

    /* for each Face associated with ebody, get its attributes and tolerances */
    status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
    CHECK_STATUS(EG_getBodyTopos);

    for (iface = 0; iface < nface; iface++) {
        SPRINT1(3, "    iface  = %d", iface);

        status = EG_attributeNum(efaces[iface], &nattr);
        CHECK_STATUS(EG_attributeNum);

        for (iattr = 1; iattr <= nattr; iattr++) {
            status = EG_attributeGet(efaces[iface], iattr, &attrName, &itype, &nlist,
                                     &tempIlist, &tempRlist, &tempClist);
            CHECK_STATUS(EG_attributeGet);

            SPRINT1x(3, "        %-20s =", attrName);

            if        (itype == ATTRINT) {
                for (i = 0; i < nlist; i++) {
                    SPRINT1x(3, "%5d ", tempIlist[i]);
                }
                SPRINT0(3, " ");
            } else if (itype == ATTRREAL) {
                for (i = 0; i < nlist; i++) {
                    SPRINT1x(3, "%11.5f ", tempRlist[i]);
                }
                SPRINT0(3, " ");
            } else if (itype == ATTRSTRING) {
                SPRINT1(3, "%s", tempClist);
            }
        }
    }

    EG_free(efaces);

    /* for each Edge associated with ebody, get its attributes and tolerances */
    status = EG_getBodyTopos(ebody, NULL, EDGE, &nedge, &eedges);
    CHECK_STATUS(EG_getBodyTopos);

    for (iedge = 0; iedge < nedge; iedge++) {
        SPRINT1(3, "    iedge  = %d", iedge);

        status = EG_attributeNum(eedges[iedge], &nattr);
        CHECK_STATUS(EG_attributeNum);

        for (iattr = 1; iattr <= nattr; iattr++) {
            status = EG_attributeGet(eedges[iedge], iattr, &attrName, &itype, &nlist,
                                     &tempIlist, &tempRlist, &tempClist);
            CHECK_STATUS(EG_attributeGet);

            SPRINT1x(3, "        %-20s =", attrName);

            if        (itype == ATTRINT) {
                for (i = 0; i < nlist; i++) {
                    SPRINT1x(3, "%5d ", tempIlist[i]);
                }
                SPRINT0(3, " ");
            } else if (itype == ATTRREAL) {
                for (i = 0; i < nlist; i++) {
                    SPRINT1x(3, "%11.5f ", tempRlist[i]);
                }
                SPRINT0(3, " ");
            } else if (itype == ATTRSTRING) {
                SPRINT1(3, "%s", tempClist);
            }
        }
    }

    EG_free(eedges);

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   printEgo - print EGADS topology for 5 levels                       *
 *                                                                      *
 ************************************************************************
 */

static void
printEgo(ego    obj)                    /* (in)  ego to start */
{
    ego    eref0,    eref1,    eref2,    eref3,    eref4,   eref5;
    int    oclass0,  oclass1,  oclass2,  oclass3,  oclass4, oclass5;
    int    mtype0,   mtype1,   mtype2,   mtype3,   mtype4,  mtype5;
    int    nchild0,  nchild1,  nchild2,  nchild3,  nchild4, nchild5;
    int    ichild0,  ichild1,  ichild2,  ichild3,  ichild4, i;
    double data0[4], data1[4], data2[4], data3[4], data4[4], data5[4];
    ego    *ebodys0, *ebodys1, *ebodys2, *ebodys3, *ebodys4, *ebodys5;
    int    *senses0, *senses1, *senses2, *senses3, *senses4, *senses5;
    int    status, count=0;
    ego    topref, prev, next, refobj, context, prev1, next1;

    /* names take from egadsTypes.h */
    char *classname[27] = {"contxt",       "transform",  "tessellation",  "nil",
                           "empty",        "reference",  "ERROR 6",       "ERROR 7",
                           "ERROR 8",      "ERROR 9",    "pcurve",        "curve",
                           "surface",      "ERROR 13",   "ERROR 14",      "ERROR 15",
                           "ERROR 16",     "ERROR 17",   "ERROR 18",      "ERROR 19",
                           "node",         "edge",       "loop",          "face",
                           "shell",        "body",       "model"};

    /* pcurves and curves */
    char *mtypename1[10] = {"ERROR 0",     "line",       "circle",        "ellipse",
                            "parabola",    "hyperbola",  "trimmed",       "bezier",
                            "bspline",     "offset"};

    /* surfaces */
    char *mtypename2[12] = {"ERROR 0",     "plane",      "spherical",     "cylindrical",
                            "revolution",  "toroidal",   "ERROR 6",       "ERROR 7",
                            "ERROR 8",     "ERROR 9",    "conical",       "extrusion"};

    /* faces */
    char *mtypename3[ 3] = {"sreverse",    "nomtype",    "sforward"};

    /* other topology */
    char *mtypename4[10] = {"nomtype",     "onenode",    "twonode",       "open",
                            "closed",      "degenerate", "wirebody",      "facebody",
                            "sheetbody",   "solidbody"};

    char mtypename[32];


    /* --------------------------------------------------------------- */

    if (obj == NULL) {
        SPRINT0(0, "NULL");
        return;
    }

    status = EG_getContext(obj, &context);
    if (status < 0) {
        SPRINT1(0, "EG_getContext -> status=%d", status);
    }

    status = EG_getTopology(obj, &eref0, &oclass0, &mtype0,
                            data0, &nchild0, &ebodys0, &senses0);
    if (status != EGADS_SUCCESS) {
        EG_getInfo(obj, &oclass0, &mtype0, &topref, &prev, &next);
        nchild0 = -1;
        eref0   = topref;
    }

    if (oclass0 != NODE) {
        if (oclass0 == PCURVE || oclass0 == CURVE) {
            strcpy(mtypename, mtypename1[mtype0]);
        } else if (oclass0 == SURFACE) {
            strcpy(mtypename, mtypename2[mtype0]);
        } else if (oclass0 == FACE) {
            strcpy(mtypename, mtypename3[mtype0+1]);
        } else if (oclass0 == EDGE  || oclass0 == LOOP ||
                   oclass0 == SHELL || oclass0 == BODY   ) {
            strcpy(mtypename, mtypename4[mtype0]);
        } else {
            strcpy(mtypename, "");
        }

        SPRINT6(0, "oclass0=%3d (%s)  mtype0=%3d (%s)  obj=%lx,  eref0=%lx",
                oclass0, classname[oclass0], mtype0, mtypename, (long)obj, (long)eref0);

        if (oclass0 == LOOP) {
            SPRINT0x(0, "< senses=");
            for (i = 0; i < nchild0; i++) {
                SPRINT1x(0, "%2d ", senses0[i]);
            }
            SPRINT0(0, " ");
        }
    } else {
        SPRINT7(0, "oclass0=%3d (%s)  mtype0=%3d,  obj=%lx,  data0=%20.10e %20.10e %20.10e",
                oclass0, classname[oclass0], mtype0, (long)obj, data0[0], data0[1], data0[2]);
    }

    next = obj->tref;
    prev = NULL;
    while (next != NULL) {
        refobj = next->attrs;
        if (refobj != context) {
            count++;
            EG_getInfo(refobj, &oclass1, &mtype1, &topref, &prev1, &next1);
            SPRINT5(0, "< refcount=%2d, refobj=%lx  (oclass=%2d (%s) mtype=%2d)",
                    count, (long)refobj, oclass1, classname[oclass1], mtype1);
        }
        prev = next;
        next = prev->blind;
    }

    for (ichild0 = 0; ichild0 < nchild0; ichild0++) {
        status = EG_getTopology(ebodys0[ichild0], &eref1, &oclass1, &mtype1,
                                data1, &nchild1, &ebodys1, &senses1);
        if (status != EGADS_SUCCESS) {
            EG_getInfo(obj, &oclass1, &mtype1, &topref, &prev, &next);
            nchild1 = -1;
            eref1   = topref;
        }

        if (oclass1 != NODE) {
            if (oclass1 == PCURVE || oclass1 == CURVE) {
                strcpy(mtypename, mtypename1[mtype1]);
            } else if (oclass1 == SURFACE) {
                strcpy(mtypename, mtypename2[mtype1]);
            } else if (oclass1 == FACE) {
                strcpy(mtypename, mtypename3[mtype1+1]);
            } else if (oclass1 == EDGE  || oclass1 == LOOP ||
                       oclass1 == SHELL || oclass1 == BODY   ) {
                strcpy(mtypename, mtypename4[mtype1]);
            } else {
                strcpy(mtypename, "");
            }

            SPRINT6(0, ". oclass1=%3d (%s)  mtype1=%3d (%s)  obj=%lx,  eref1=%lx",
                    oclass1, classname[oclass1], mtype1, mtypename, (long)(ebodys0[ichild0]), (long)eref1);

            if (oclass1 == LOOP) {
                SPRINT0x(0, ". < senses=");
                for (i = 0; i < nchild1; i++) {
                    SPRINT1x(0, "%2d ", senses1[i]);
                }
                SPRINT0(0, " ");
            }
        } else {
            SPRINT7(0, ". oclass1=%3d (%s)  mtype1=%3d,  obj=%lx,  data1=%20.10e %20.10e %20.10e",
                    oclass1, classname[oclass1], mtype1, (long)(ebodys0[ichild0]), data1[0], data1[1], data1[2]);
        }

        for (ichild1 = 0; ichild1 < nchild1; ichild1++) {
            status = EG_getTopology(ebodys1[ichild1], &eref2, &oclass2, &mtype2,
                                    data2, &nchild2, &ebodys2, &senses2);
            if (status != EGADS_SUCCESS) {
                EG_getInfo(obj, &oclass2, &mtype2, &topref, &prev, &next);
                nchild2 = -1;
                eref2   = topref;
            }

            if (oclass2 != NODE) {
                if (oclass2 == PCURVE || oclass2 == CURVE) {
                    strcpy(mtypename, mtypename1[mtype2]);
                } else if (oclass2 == SURFACE) {
                    strcpy(mtypename, mtypename2[mtype2]);
                } else if (oclass2 == FACE) {
                    strcpy(mtypename, mtypename3[mtype2+1]);
                } else if (oclass2 == EDGE  || oclass2 == LOOP ||
                           oclass2 == SHELL || oclass2 == BODY   ) {
                    strcpy(mtypename, mtypename4[mtype2]);
                } else {
                    strcpy(mtypename, "");
                }

                SPRINT6(0, ". . oclass2=%3d (%s)  mtype2=%3d (%s)  obj=%lx,  eref2=%lx",
                        oclass2, classname[oclass2], mtype2, mtypename, (long)(ebodys1[ichild1]), (long)eref2);

                if (oclass2 == LOOP) {
                    SPRINT0x(0, ". . < senses=");
                    for (i=0; i < nchild2; i++) {
                        SPRINT1x(0, "%2d ", senses2[i]);
                    }
                    SPRINT0(0, " ");
                }
            } else {
                SPRINT7(0, ". . oclass2=%3d (%s)  mtype2=%3d,  obj=%lx,  data2=%20.10e %20.10e %20.10e",
                        oclass2, classname[oclass2], mtype2, (long)(ebodys1[ichild1]), data2[0], data2[1], data2[2]);
            }

            for (ichild2 = 0; ichild2 < nchild2; ichild2++) {
                status = EG_getTopology(ebodys2[ichild2], &eref3, &oclass3, &mtype3,
                               data3, &nchild3, &ebodys3, &senses3);
                if (status != EGADS_SUCCESS) {
                    EG_getInfo(obj, &oclass3, &mtype3, &topref, &prev, &next);
                    nchild3 = -1;
                    eref3   = topref;
                }

                if (oclass3 != NODE) {
                    if (oclass3 == PCURVE || oclass3 == CURVE) {
                        strcpy(mtypename, mtypename1[mtype3]);
                    } else if (oclass3 == SURFACE) {
                        strcpy(mtypename, mtypename2[mtype3]);
                    } else if (oclass3 == FACE) {
                        strcpy(mtypename, mtypename3[mtype3+1]);
                    } else if (oclass3 == EDGE  || oclass3 == LOOP ||
                               oclass3 == SHELL || oclass3 == BODY   ) {
                        strcpy(mtypename, mtypename4[mtype3]);
                    } else {
                        strcpy(mtypename, "");
                    }

                    SPRINT6(0, ". . . oclass3=%3d (%s)  mtype3=%3d (%s)  obj=%lx,  eref3=%lx",
                            oclass3, classname[oclass3], mtype3, mtypename, (long)(ebodys2[ichild2]), (long)eref3);

                    if (oclass3 == LOOP) {
                        SPRINT0x(0, ". . . < senses=");
                        for (i=0; i < nchild3; i++) {
                            SPRINT1x(0, "%2d ", senses3[i]);
                        }
                        SPRINT0(0, " ");
                    }
                } else {
                    SPRINT7(0, ". . . oclass3=%3d (%s)  mtype3=%3d,  obj=%lx,  data3=%20.10e %20.10e %20.10e",
                            oclass3, classname[oclass3], mtype3, (long)(ebodys2[ichild2]), data3[0], data3[1], data3[2]);
                }

                for (ichild3 = 0; ichild3 < nchild3; ichild3++) {
                    status = EG_getTopology(ebodys3[ichild3], &eref4, &oclass4, &mtype4,
                                   data4, &nchild4, &ebodys4, &senses4);
                    if (status != EGADS_SUCCESS) {
                        EG_getInfo(obj, &oclass4, &mtype4, &topref, &prev, &next);
                        nchild4 = -1;
                        eref4   = topref;
                    }

                    if (oclass4 != NODE) {
                        if (oclass4 == PCURVE || oclass4 == CURVE) {
                            strcpy(mtypename, mtypename1[mtype4]);
                        } else if (oclass4 == SURFACE) {
                            strcpy(mtypename, mtypename2[mtype4]);
                        } else if (oclass4 == FACE) {
                            strcpy(mtypename, mtypename3[mtype4+1]);
                        } else if (oclass4 == EDGE  || oclass4 == LOOP ||
                                   oclass4 == SHELL || oclass4 == BODY   ) {
                            strcpy(mtypename, mtypename4[mtype4]);
                        } else {
                            strcpy(mtypename, "");
                        }

                        SPRINT6(0, ". . . . oclass4=%3d (%s)  mtype4=%3d (%s)  obj=%lx,  eref4=%lx",
                                oclass4, classname[oclass4], mtype4, mtypename, (long)(ebodys3[ichild3]), (long)eref4);

                        if (oclass4 == LOOP) {
                            SPRINT0x(0, ". . . . < senses=");
                            for (i=0; i < nchild4; i++) {
                                SPRINT1x(0, "%2d ", senses4[i]);
                            }
                            SPRINT0(0, " ");
                        }
                    } else {
                        SPRINT7(0, ". . . . oclass4=%3d (%s)  mtype4=%3d,  obj=%lx,  data4=%20.10e %20.10e %20.10e",
                                oclass4, classname[oclass4], mtype4, (long)(ebodys3[ichild3]), data4[0], data4[1], data4[2]);
                    }

                    for (ichild4 = 0; ichild4 < nchild4; ichild4++) {
                        status = EG_getTopology(ebodys4[ichild4], &eref5, &oclass5, &mtype5,
                                                data5, &nchild5, &ebodys5, &senses5);
                        if (status != EGADS_SUCCESS) {
                            EG_getInfo(obj, &oclass5, &mtype5, &topref, &prev, &next);
                            nchild5 = -1;
                            eref5   = topref;
                        }

                        if (oclass5 != NODE) {
                            if (oclass5 == PCURVE || oclass5 == CURVE) {
                                strcpy(mtypename, mtypename1[mtype5]);
                            } else if (oclass5 == SURFACE) {
                                strcpy(mtypename, mtypename2[mtype5]);
                            } else if (oclass5 == FACE) {
                                strcpy(mtypename, mtypename3[mtype5+1]);
                            } else if (oclass5 == EDGE  || oclass5 == LOOP ||
                                       oclass5 == SHELL || oclass5 == BODY   ) {
                                strcpy(mtypename, mtypename4[mtype5]);
                            } else {
                                strcpy(mtypename, "");
                            }

                            SPRINT6(0, ". . . . . oclass5=%3d (%s)  mtype5=%3d (%s)  obj=%lx,  eref5=%lx",
                                    oclass5, classname[oclass5], mtype5, mtypename, (long)(ebodys4[ichild4]), (long)eref5);

                            if (oclass5 == LOOP) {
                                SPRINT0x(0, ". . . . . < senses=");
                                for (i=0; i < nchild5; i++) {
                                    SPRINT1x(0, "%2d ", senses5[i]);
                                }
                                SPRINT0(0, " ");
                            }
                        } else {
                            SPRINT7(0, ". . . . . oclass5=%3d (%s)  mtype5=%3d,  obj=%lx,  data5=%20.10e %20.10e %20.10e",
                                    oclass5, classname[oclass5], mtype5, (long)(ebodys4[ichild4]), data5[0], data5[1], data5[2]);
                        }
                    }
                }
            }
        }
    }
}


/*
 ************************************************************************
 *                                                                      *
 *   recycleBody - recycle a Body if its arguments are unchanged        *
 *                                                                      *
 ************************************************************************
 */

static int
recycleBody(modl_T *modl,               /* (in)  pointer to MODL */
            int    ibrch,               /* (in)  Branch index (1-nbrch) */
            int    brtype,              /* (in)  Branch type */
            varg_T args[],              /* (in)  array  of arguments */
            int    hasdots)             /* (in)  =1 if any arguments have non-zero dots */
{
    int       status = SUCCESS;         /* (out) return status */

    int       iarg, ival;

    modl_T    *MODL = (modl_T*)modl;

    int       ibody, jbody, okay;

    ROUTINE(recycleBody);
    DPRINT3("%s(ibrch=%d, brtype=%d) {",
            routine, ibrch, brtype);

    /* --------------------------------------------------------------- */

    /* if we have exhaused the recycling, return status=0 (to signify
       that an old Body was not recycled) */
    if (MODL->nbody+1 > MODL->recycle) {
        status = 0;
        goto cleanup;
    }

    /* by default use the next Body */
    ibody = MODL->nbody + 1;

    /* if a OCSM_UNION and tomark==1, find the last Body that matches ibrch */
    if (brtype == OCSM_UNION && NINT(args[1].val[0]) == 1) {
        while ((ibody+1) <= MODL->recycle) {
            if (MODL->body[ibody+1].ibrch  == ibrch) {
                SPRINT1(0, "                          ... skipping Body %4d", ibody);
                ibody++;
            } else {
                break;
            }
        }
    }

    /* make sure that the next Body's Branch can be recycled.  it
       cannot under the following circumstances: */
    okay = 1;

    /* if the ibrch or brtype is not the expected one (this can happen if
       a previously suppressed Branch has been activated), we need to rebuild */
    if (MODL->body[ibody].ibrch  != ibrch ||
        MODL->body[ibody].brtype != brtype  ) {
        okay = 0;
    }

    /* if the value of any of the arguments has changed, we need to rebuild */
    for (iarg = 1; iarg < 10; iarg++) {
        if (MODL->body[ibody].arg[iarg].nval != args[iarg].nval) {
            okay = 0;
            break;
        }

        for (ival = 0; ival < args[iarg].nval; ival++) {
            if (MODL->body[ibody].arg[iarg].val[ival] != args[iarg].val[ival]) {
                okay = 0;
                break;
            }
        }
    }

    /* free up all Bodys starting at ibody */
    if (okay == 0) {
        for (jbody = ibody; jbody <= MODL->recycle; jbody++) {
            for (iarg = 1; iarg < 10; iarg++) {
                FREE(MODL->body[jbody].arg[iarg].val);
                FREE(MODL->body[jbody].arg[iarg].dot);
            }

            FREE(MODL->body[jbody].node);
            FREE(MODL->body[jbody].edge);
            FREE(MODL->body[jbody].face);

            if (MODL->body[jbody].etess != NULL) {
                status = EG_deleteObject(MODL->body[jbody].etess);
                CHECK_STATUS(EG_deleteObject);

                MODL->body[jbody].etess = NULL;
            }

            if (MODL->body[jbody].ebody != NULL) {
                status = EG_deleteObject(MODL->body[jbody].ebody);
                CHECK_STATUS(EG_deleteObject);

                MODL->body[jbody].ebody = NULL;
            }
        }

        MODL->recycle = ibody;

        /* return status=0 (to signify that an old Body was not recycled) */
        status = 0;
        goto cleanup;
    }

    /* increment number of Bodys */
    MODL->nbody = ibody;
    SPRINT1(1, "                          Body   %4d recycled", MODL->nbody);

    /* store the velocities (since they may have changed even though
       the arguments may not have changed) */
    MODL->body[ibody].hasdots = hasdots;

    for (iarg = 1; iarg < 10; iarg++) {
        for (ival = 0; ival < args[iarg].nval; ival++) {
            MODL->body[ibody].arg[iarg].dot[ival] = args[iarg].dot[ival];
        }
    }

    /* reset the selection */
    MODL->sel_body = ibody;
    MODL->sel_node = -1;
    MODL->sel_edge = -1;
    MODL->sel_face = -1;

    /* return status=1 (to signify that an old Body WAS recycled) */
    status = 1;

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   removePerurbation - remove a perturbation if it exists             *
 *                                                                      *
 ************************************************************************
 */

int
removePerturbation(modl_T *modl)        /* (in)  pointer to MODL */
{
    int       status = 0;               /* (out) body that satisfies criterion (bias-0) */

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(removePerturbation);
    DPRINT1("%s() {",
            routine);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* if this has a perturbed Body, free it and reset data */
    if (MODL->perturb != NULL) {
        status = ocsmFree(MODL->perturb);
        CHECK_STATUS(ocsmFree);

        MODL->perturb = NULL;
        MODL->dtime   = 0;
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   reorderLoops - reorder Loops to improve rules and blends           *
 *                                                                      *
 ************************************************************************
 */

static int
reorderLoops(modl_T *modl,              /* (in)  pointer to MODL */
             int    nloop,              /* (in)  number of Loops */
             ego    eloops[],           /* (both)array  of Loops */
             int    startFrom)          /* (in)  +1 to start at beg, -1 to start at end, 0 to return */

{
    int       status = 0;               /* (out) body that satisfies criterion (bias-0) */

    modl_T    *MODL = (modl_T*)modl;

    int      iloop, jloop, iedge, jedge, itest, ishift, modified, sense;
    int      oclassi, oclassj, oclassk, mtypei, mtypej, mtypek, nedgei, nedgej, nedgek;
    int      *sensesi, *sensesj, *sensesk, nnode, *sensesnew=NULL;
    double   uvlimitsi[4], uvlimitsj[4], uvlimitsk[4], *xyzi=NULL, *xyzj=NULL;
    double   areai[3], areaj[3], dotprod, ltest, lshift;
    ego      erefi, erefj, erefk, *eedgesi, *eedgesj, *eedgesk;
    ego      *enodes, *elist, *eedgesnew=NULL, etemp;

    ROUTINE(reorderLoops);
    DPRINT3("%s(nloop=%d, startFrom=%d) {",
            routine, nloop, startFrom);

    /* --------------------------------------------------------------- */

    if (outLevel >= 3) {
        for (iloop = 0; iloop < nloop; iloop++) {
            SPRINT1(3, "at beg of reorderLoops: eloops[%d]", iloop);
            printEgo(eloops[iloop]);
        }
    }

    /* extract just the Loop for interior Sketches and convert bounding
       WireBodys into Loops */
    for (iloop = 0; iloop < nloop; iloop++) {
        status = EG_getTopology(eloops[iloop], &erefi,
                                &oclassi, &mtypei, uvlimitsi, &nedgei, &eedgesi, &sensesi);
        CHECK_STATUS(EG_getTopology);

        /* if a Loop, leave alone */
        if (oclassi == LOOP) {

        /* if a Node, leave alone */
        } else if (oclassi == NODE) {

        /* if an interior Face, convert to Loop */
        } else if (oclassi == FACE) {
            if (iloop > 0 && iloop < nloop-1) {
                eloops[iloop] = eedgesi[0];
            }

        /* if a WireBody, convert to Loop */
        } else if (oclassi == BODY && mtypei == WIREBODY) {
            eloops[iloop] = eedgesi[0];

        /* not a recognized  type */
        } else {
            signalError(MODL, OCSM_WRONG_TYPES_ON_STACK, 0,
                        "Sketch %d is not FACE, NODE, or WIREBODY", iloop);
            status = OCSM_WRONG_TYPES_ON_STACK;
            goto cleanup;
        }
    }

    /* the base Sketch (which will not be modified) is at either end */
    if (startFrom == 0) {
        goto cleanup;
    } else if (startFrom > 0) {
        iloop = 0;
    } else {
        iloop = nloop - 1;
    }

    /* process each pair of Sketches in turn */
    while (1) {
        if (startFrom > 0) {
            jloop = iloop + 1;
            if (jloop >= nloop) break;
        } else {
            jloop = iloop - 1;
            if (jloop < 0) break;
        }
        modified = 0;

        /* get info on iloop */
        status = EG_getTopology(eloops[iloop], &erefi,
                                &oclassi, &mtypei, uvlimitsi, &nedgei, &eedgesi, &sensesi);
        CHECK_STATUS(EG_getTopology);

        if (oclassi == FACE) {
            etemp  = eedgesi[0];
            status = EG_getTopology(etemp, &erefi,
                                    &oclassi, &mtypei, uvlimitsi, &nedgei, &eedgesi, &sensesi);
            CHECK_STATUS(EG_getTopology);
        }

        /* get info on jloop */
        status = EG_getTopology(eloops[jloop], &erefj,
                                &oclassj, &mtypej, uvlimitsj, &nedgej, &eedgesj, &sensesj);
        CHECK_STATUS(EG_getTopology);

        if (oclassj == FACE) {
            etemp  = eedgesj[0];
            status = EG_getTopology(etemp, &erefj,
                                    &oclassj, &mtypej, uvlimitsj, &nedgej, &eedgesj, &sensesj);
            CHECK_STATUS(EG_getTopology);
        }

        if (oclassi != LOOP   || oclassj != LOOP  ) {
            iloop = jloop;
            continue;
        }
        if (mtypei  != CLOSED || mtypej  != CLOSED) {
            iloop = jloop;
            continue;
        }

        /* make sure that Loops have same number of Edges */
        if (nedgei != nedgej) {
            signalError(MODL, OCSM_INTERNAL_ERROR, 0,
                        "Loop %d has %d Edges, but Loop %d has %d Edges", iloop, nedgei, jloop, nedgej);
            status = OCSM_INTERNAL_ERROR;
            goto cleanup;
        } else if (nedgei < 1) {
            signalError(MODL, OCSM_INTERNAL_ERROR, 0,
                        "Loop %d only has %d Edges", iloop, nedgei);
            status = OCSM_INTERNAL_ERROR;
            goto cleanup;
        }

#ifdef __clang_analyzer__
        nedgej = nedgei;
#endif

        /* set up the coordinates at the Nodes in the two Loops */
        MALLOC(xyzi, double, 3*nedgei);
        MALLOC(xyzj, double, 3*nedgej);

        for (iedge = 0; iedge < nedgei; iedge++) {
            status = EG_getTopology(eedgesi[iedge], &erefk,
                                    &oclassk, &mtypek, uvlimitsk, &nnode, &enodes, &sensesk);
            CHECK_STATUS(EG_getTopology);

            if (sensesi[iedge] > 0) {
                status = EG_getTopology(enodes[0], &erefk,
                                        &oclassk, &mtypek, &(xyzi[3*iedge]), &nnode, &elist, &sensesk);
                CHECK_STATUS(EG_getTopology);
            } else {
                status = EG_getTopology(enodes[1], &erefk,
                                        &oclassk, &mtypek, &(xyzi[3*iedge]), &nnode, &elist, &sensesk);
                CHECK_STATUS(EG_getTopology);
            }
        }

        for (iedge = 0; iedge < nedgej; iedge++) {
            status = EG_getTopology(eedgesj[iedge], &erefk,
                                    &oclassk, &mtypek, uvlimitsk, &nnode, &enodes, &sensesk);
            CHECK_STATUS(EG_getTopology);

            if (sensesj[iedge] > 0) {
                status = EG_getTopology(enodes[0], &erefk,
                                        &oclassk, &mtypek, &(xyzj[3*iedge]), &nnode, &elist, &sensesk);
                CHECK_STATUS(EG_getTopology);
            } else {
                status = EG_getTopology(enodes[1], &erefk,
                                        &oclassk, &mtypek, &(xyzj[3*iedge]), &nnode, &elist, &sensesk);
                CHECK_STATUS(EG_getTopology);
            }
        }

        /* find the area of iloop and jloop */
        areai[0] = areai[1] = areai[2] = 0;
        areaj[0] = areaj[1] = areaj[2] = 0;

        for (iedge = 1; iedge < nedgei-1; iedge++) {
            areai[0] += (xyzi[1] - xyzi[3*nedgei-2]) * (xyzi[3*iedge+2] - xyzi[3*nedgei-1])
                      - (xyzi[2] - xyzi[3*nedgei-1]) * (xyzi[3*iedge+1] - xyzi[3*nedgei-2]);
            areai[1] += (xyzi[2] - xyzi[3*nedgei-1]) * (xyzi[3*iedge  ] - xyzi[3*nedgei-3])
                      - (xyzi[0] - xyzi[3*nedgei-3]) * (xyzi[3*iedge+2] - xyzi[3*nedgei-1]);
            areai[2] += (xyzi[0] - xyzi[3*nedgei-3]) * (xyzi[3*iedge+1] - xyzi[3*nedgei-2])
                      - (xyzi[1] - xyzi[3*nedgei-2]) * (xyzi[3*iedge  ] - xyzi[3*nedgei-3]);

            areaj[0] += (xyzj[1] - xyzj[3*nedgej-2]) * (xyzj[3*iedge+2] - xyzj[3*nedgej-1])
                      - (xyzj[2] - xyzj[3*nedgej-1]) * (xyzj[3*iedge+1] - xyzj[3*nedgej-2]);
            areaj[1] += (xyzj[2] - xyzj[3*nedgej-1]) * (xyzj[3*iedge  ] - xyzj[3*nedgej-3])
                      - (xyzj[0] - xyzj[3*nedgej-3]) * (xyzj[3*iedge+2] - xyzj[3*nedgej-1]);
            areaj[2] += (xyzj[0] - xyzj[3*nedgej-3]) * (xyzj[3*iedge+1] - xyzj[3*nedgej-2])
                      - (xyzj[1] - xyzj[3*nedgej-2]) * (xyzj[3*iedge  ] - xyzj[3*nedgej-3]);
        }

        /* if the dot products of the areas is negative, flip the direction of jloop */
        dotprod = areai[0] * areaj[0] + areai[1] * areaj[1] + areai[2] * areaj[2];
        if (dotprod < 0) {
            SPRINT1(0, "WARNING:: reversing Loop %d", jloop);
            modified++;

            etemp  = eloops[jloop];
            status = EG_flipObject(etemp, &(eloops[jloop]));
            CHECK_STATUS(EG_flipObject);

            status = EG_getTopology(eloops[jloop], &erefj,
                                    &oclassj, &mtypej, uvlimitsj, &nedgej, &eedgesj, &sensesj);
            CHECK_STATUS(EG_getTopology);

            if (oclassj == FACE) {
                etemp  = eedgesj[0];
                status = EG_getTopology(etemp, &erefj,
                                        &oclassj, &mtypej, uvlimitsj, &nedgej, &eedgesj, &sensesj);
                CHECK_STATUS(EG_getTopology);
            }

            for (iedge = 0; iedge < nedgej; iedge++) {
                status = EG_getTopology(eedgesj[iedge], &erefk,
                                        &oclassk, &mtypek, uvlimitsk, &nnode, &enodes, &sensesk);
                CHECK_STATUS(EG_getTopology);

                if (sensesj[iedge] > 0) {
                    status = EG_getTopology(enodes[0], &erefk,
                                            &oclassk, &mtypek, &(xyzj[3*iedge]), &nnode, &elist, &sensesk);
                    CHECK_STATUS(EG_getTopology);
                } else {
                    status = EG_getTopology(enodes[1], &erefk,
                                            &oclassk, &mtypek, &(xyzj[3*iedge]), &nnode, &elist, &sensesk);
                    CHECK_STATUS(EG_getTopology);
                }
            }
        }

        /* find the shift of jloop that minimizes the distance between the Nodes of iloop
           and the Nodes of jloop */
        ishift = 0;
        lshift = HUGEQ;

        /* compute the sum of the distances for each candidate shift */
        for (itest = 0; itest < nedgei; itest++) {

            ltest = 0;
            for (iedge = 0; iedge < nedgei; iedge++) {
                jedge = (iedge + itest) % nedgei;

                ltest += sqrt(pow(xyzi[3*iedge  ]-xyzj[3*jedge  ], 2)
                             +pow(xyzi[3*iedge+1]-xyzj[3*jedge+1], 2)
                             +pow(xyzi[3*iedge+2]-xyzj[3*jedge+2],2));
            }

            if (ltest < lshift) {
                ishift = itest;
                lshift = ltest;
            }
        }

        /* create the new rotated Loop */
        if (ishift > 0) {
            SPRINT2(0, "WARNING:: shifting  Loop %d with ishift=%d", jloop, ishift);
            modified++;

            MALLOC(eedgesnew, ego, 2*nedgej);
            MALLOC(sensesnew, int, 2*nedgej);

            /* shift the Loop by ishift */
            for (iedge = 0; iedge < nedgej; iedge++) {
                jedge = (iedge + ishift) % nedgej;
                eedgesnew[iedge] = eedgesj[jedge];
                sensesnew[iedge] = sensesj[jedge];
            }

            status = EG_getTopology(eloops[jloop], &erefk,
                                    &oclassk, &mtypek, uvlimitsk, &nedgek, &eedgesk, &sensesk);
            CHECK_STATUS(EG_getTopology);

            if (oclassk == FACE) {
                for (iedge = 0; iedge < nedgej; iedge++) {
                    jedge = (iedge + ishift) % nedgej;
                    eedgesnew[iedge+nedgej] = eedgesj[jedge+nedgej];
                    sensesnew[iedge+nedgej] = sensesj[jedge+nedgej];
                }
            }

            /* make new Loop */
            status = EG_makeTopology(MODL->context, erefk, LOOP, CLOSED,
                                     NULL, nedgej, eedgesnew, sensesnew, &etemp);
            CHECK_STATUS(EG_makeTopology);

            FREE(sensesnew);
            FREE(eedgesnew);

            /* either store the new Loop or make a new Face (if the original jloop
               was a Face) */
            if (oclassk != FACE) {
                eloops[jloop] = etemp;
            } else {
                sense = SFORWARD;
                status = EG_makeTopology(MODL->context, erefk, FACE, mtypek,
                                         NULL, 1, &etemp, &sense, &(eloops[jloop]));
                CHECK_STATUS(EG_makeTopology);
            }
        }

        FREE(xyzj);
        FREE(xyzi);

        /* next iloop,jloop pair */
        iloop = jloop;
    }

//$$$    /* print out info about Loops */
//$$$    for (iloop = 0; iloop < nloop; iloop++) {
//$$$        etemp = eloops[iloop];
//$$$        status = EG_getTopology(etemp, &erefi,
//$$$                                &oclassi, &mtypei, uvlimitsi, &nedgei, &eedgesi, &sensesi);
//$$$        CHECK_STATUS(EG_getTopology);
//$$$
//$$$        printf("\niloop=%d,  oclassi=%2d,  mtypei=%2d,  nedgei=%2d\n",
//$$$               iloop, oclassi, mtypei, nedgei);
//$$$
//$$$        if (oclassi == NODE) continue;
//$$$
//$$$        while (oclassi != LOOP) {
//$$$            etemp = eedgesi[0];
//$$$
//$$$            status = EG_getTopology(etemp, &erefi,
//$$$                                    &oclassi, &mtypei, uvlimitsi, &nedgei, &eedgesi, &sensesi);
//$$$            CHECK_STATUS(EG_getTopology);
//$$$
//$$$            printf("iloop=%d,  oclassi=%2d,  mtypei=%2d,  nedgei=%2d\n",
//$$$                   iloop, oclassi, mtypei, nedgei);
//$$$        }
//$$$
//$$$        for (iedge = 0; iedge < nedgei; iedge++) {
//$$$            status = EG_getTopology(eedgesi[iedge], &erefk,
//$$$                                    &oclassk, &mtypek, uvlimitsk, &nedgek, &enodes, &sensesk);
//$$$            CHECK_STATUS(EG_getTopology);
//$$$
//$$$            double xyzbeg[3], xyzend[3];
//$$$
//$$$            status = EG_getTopology(enodes[0], &erefk,
//$$$                                    &oclassj, &mtypej, xyzbeg, &nedgej, &eedgesj, &sensesj);
//$$$            CHECK_STATUS(EG_getTopology);
//$$$            status = EG_getTopology(enodes[1], &erefk,
//$$$                                    &oclassj, &mtypej, xyzend, &nedgej, &eedgesj, &sensesj);
//$$$            CHECK_STATUS(EG_getTopology);
//$$$
//$$$            if (sensesi[iedge] > 0) {
//$$$                printf("iedge=%d,  oclassk=%2d,  mtypek=%2d,  nedgek=%2d,  %10.4f  %10.4f  %10.4f\n",
//$$$                       iedge, oclassk, mtypek, nedgek, xyzend[0]-xyzbeg[0], xyzend[1]-xyzbeg[1], xyzend[2]-xyzbeg[2]);
//$$$            } else {
//$$$                printf("iedge=%d,  oclassk=%2d,  mtypek=%2d,  nedgek=%2d,  %10.4f  %10.4f  %10.4f\n",
//$$$                       iedge, oclassk, mtypek, nedgek, -xyzend[0]+xyzbeg[0], -xyzend[1]+xyzbeg[1], -xyzend[2]+xyzbeg[2]);
//$$$            }
//$$$        }
//$$$    }

cleanup:
    FREE(xyzj);
    FREE(xyzi);

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   selectBody - select a Body from a Model                            *
 *                                                                      *
 ************************************************************************
 */

static int
selectBody(ego    emodel,               /* (in)  pointer to Model */
           char*  order,                /* (in)  order type */
           int    index)                /* (in)  index into list (1->nchild) */
{
    int       status = 0;               /* (out) body that satisfies criterion (bias-0) */

    int       i, j, imin, oclass, mtype, nchild, *senses;
    double    box[14], datamin, datamax, data2[4], *data=NULL;
    ego       eref, *ebodys;

    ROUTINE(selectBody);
    DPRINT3("%s(order=%s, index=%d) {",
            routine, order, index);

    /* --------------------------------------------------------------- */

    status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                            data2, &nchild, &ebodys, &senses);
    CHECK_STATUS(EG_getTopology);

    /* if nchild is just one, return the answer immediately */
    if (nchild == 1) {
        status = 0;
        goto cleanup;
    } else if (nchild < 1) {
        SPRINT2(0, "WARNING:: emodel(=%lx) has %d children", (long)emodel, nchild);
        status = OCSM_INTERNAL_ERROR;
        goto cleanup;
    }

    /* make an array that has data based upon order.  note that we
       will always want to return the (index-1)'th smallest entry so that
       quantities that are to maximized are stored as negative value */

    MALLOC(data, double, nchild);

    if        (strcmp(order, "none") == 0) {
        for (i = 0; i < nchild; i++) {
            data[i] = i;
        }
    } else if (strcmp(order, "xmin") == 0) {
        for (i = 0; i < nchild; i++) {
            status = EG_getBoundingBox(ebodys[i], box);
            CHECK_STATUS(EG_getBoundingBox);

            data[i] = +box[0];
        }
    } else if (strcmp(order, "xmax") == 0) {
        for (i = 0; i < nchild; i++) {
            status = EG_getBoundingBox(ebodys[i], box);
            CHECK_STATUS(EG_getBoundingBox);

            data[i] = -box[3];
        }
    } else if (strcmp(order, "ymin") == 0) {
        for (i = 0; i < nchild; i++) {
            status = EG_getBoundingBox(ebodys[i], box);
            CHECK_STATUS(EG_getBoundingBox);

            data[i] = +box[1];
        }
    } else if (strcmp(order, "ymax") == 0) {
        for (i = 0; i < nchild; i++) {
            status = EG_getBoundingBox(ebodys[i], box);
            CHECK_STATUS(EG_getBoundingBox);

            data[i] = -box[4];
        }
    } else if (strcmp(order, "zmin") == 0) {
        for (i = 0; i < nchild; i++) {
            status = EG_getBoundingBox(ebodys[i], box);
            CHECK_STATUS(EG_getBoundingBox);

            data[i] = +box[2];
        }
    } else if (strcmp(order, "zmax") == 0) {
        for (i = 0; i < nchild; i++) {
            status = EG_getBoundingBox(ebodys[i], box);
            CHECK_STATUS(EG_getBoundingBox);

            data[i] = -box[5];
        }
    } else if (strcmp(order, "amin") == 0) {
        for (i = 0; i < nchild; i++) {
            status = EG_getMassProperties(ebodys[i], box);
            CHECK_STATUS(EG_getMassProperties);

            data[i] = +box[1];
        }
    } else if (strcmp(order, "amax") == 0) {
        for (i = 0; i < nchild; i++) {
            status = EG_getMassProperties(ebodys[i], box);
            CHECK_STATUS(EG_getMassProperties);

            data[i] = -box[1];
        }
    } else if (strcmp(order, "vmin") == 0) {
        for (i = 0; i < nchild; i++) {
            status = EG_getMassProperties(ebodys[i], box);
            CHECK_STATUS(EG_getMassProperties);

            data[i] = +box[0];
        }
    } else if (strcmp(order, "vmax") == 0) {
        for (i = 0; i < nchild; i++) {
            status = EG_getMassProperties(ebodys[i], box);
            CHECK_STATUS(EG_getMassProperties);

            data[i] = -box[0];
        }
    } else {
        status = OCSM_ILLEGAL_TYPE;
        goto cleanup;
    }

    /* find the largest entry in data */
    datamax = data[0];
    for (i = 1; i < nchild; i++) {
        if (data[i] > datamax) datamax = data[i];
    }

    /* move the smallest index-1 entries to the end of the list */
    for (j = 0; j < index-1; j++) {
        datamin = data[0];
        imin    = 0;
        for (i = 1; i < nchild; i++) {
            if (data[i] < datamin) {
                datamin = data[i];
                imin    = i;
            }
        }

        data[imin] = (++datamax);
    }

    /* return the smallest entry in data */
    datamin = data[0];
    imin    = 0;
    for (i = 1; i < nchild; i++) {
        if (data[i] < datamin) {
            datamin = data[i];
            imin    = i;
        }
    }

    status = imin;

cleanup:
    FREE(data);

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   setFaceAttribute - set attribute(s) to Face                        *
 *                                                                      *
 ************************************************************************
 */

static int
setFaceAttribute(modl_T *modl,          /* (in)  pointer to MODL */
                 int    ibody,          /* (in)  Body index (1-nbody) */
                 int    iface,          /* (in)  Face index (1-nface) */
                 int    jbody,          /* (in)  Branch index (=0 for automatic) */
                 int    jford,          /* (in)  Face order */
                 int    npatn,          /* (in)  number of active patterns */
                 patn_T *patn)          /* (in)  array  of active patterns */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int       *iattrib=NULL, nattrib, i, ibrch, iattr, ipmtr, jpmtr;
    double    value, dot;
    body_T    *body=NULL;

    int       nface;
    ego       *efaces;

    ROUTINE(setFaceAttribute);
    DPRINT6("%s(ibody=%d, iface=%d, jbody=%d, jford=%d, npatn=%d) {",
            routine, ibody, iface, jbody, jford, npatn);

    /* --------------------------------------------------------------- */

    /* check magic number */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (MODL->magic != OCSM_MAGIC) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    }

    /* check that valid ibody is given */
    if (ibody == 0) {
        goto cleanup;
    } else if (ibody < 1 || ibody > MODL->nbody) {
        status = OCSM_ILLEGAL_BODY_INDEX;
        goto cleanup;
    }

    /* get Body info */
    body = &(MODL->body[ibody]);
    ibrch =  MODL->body[ibody].ibrch;

    /* allocate storage for new attributes */
    MALLOC(iattrib, int, 2+2*npatn);

    nattrib = 0;

    if (jbody == 0) {
        iattrib[nattrib++] = ibody;
        iattrib[nattrib++] = jford;
    } else {
        iattrib[nattrib++] = jbody;
        iattrib[nattrib++] = jford;
    }

    for (i = npatn-1; i >= 0; i--) {
        iattrib[nattrib++] = patn[i].ipatbeg;
        iattrib[nattrib++] = patn[i].icopy;
    }

    /* save the Body attributes */
    status = EG_getBodyTopos(body->ebody, NULL, FACE, &nface, &efaces);
    CHECK_STATUS(EG_getBodyTopos);

    if (nface > 0) {
        if (MODL->brch[ibrch].type != OCSM_RESTORE) {
            status = EG_attributeAdd(efaces[iface-1], "body", ATTRINT,
                                     2, iattrib, NULL, NULL);
            CHECK_STATUS(EG_attributeAdd);
        }

        iattrib[0] = ibrch;
        status = EG_attributeAdd(efaces[iface-1], "brch", ATTRINT,
                                 nattrib, iattrib, NULL, NULL);
        CHECK_STATUS(EG_attributeAdd);

        for (iattr = 0; iattr < MODL->brch[ibrch].nattr; iattr++) {
            if (MODL->brch[ibrch].attr[iattr].value[0] == '$') {
                status = EG_attributeAdd(efaces[iface-1], MODL->brch[ibrch].attr[iattr].name,
                                         ATTRSTRING, 1,
                                         NULL, NULL, &(MODL->brch[ibrch].attr[iattr].value[1]));
                CHECK_STATUS(EG_attributeAdd);
            } else {
                ipmtr = -1;
                for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
                    if (strcmp(MODL->brch[ibrch].attr[iattr].value, MODL->pmtr[jpmtr].name) == 0) {
                        ipmtr = jpmtr;
                        break;
                    }
                }

                if (ipmtr > 0) {
                    status = EG_attributeAdd(efaces[iface-1], MODL->brch[ibrch].attr[iattr].name,
                                             ATTRREAL, (MODL->pmtr[ipmtr].nrow)*(MODL->pmtr[ipmtr].ncol),
                                             NULL, MODL->pmtr[ipmtr].value, NULL);
                    CHECK_STATUS(EG_attributeAdd);
                } else {
                    str2val(MODL->brch[ibrch].attr[iattr].value, MODL, &value, &dot);
                    CHECK_STATUS(str2val);

                    status = EG_attributeAdd(efaces[iface-1], MODL->brch[ibrch].attr[iattr].name,
                                             ATTRREAL, 1,
                                             NULL, &value, NULL);
                    CHECK_STATUS(EG_attributeAdd);
                }
            }
        }

        EG_free(efaces);
    }

cleanup:
    FREE(iattrib);

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   setupAtPmtrs - sets up Parameters starting with @                  *
 *                                                                      *
 ************************************************************************
 */

static int
setupAtPmtrs(modl_T *modl)              /* (in)  pointer to MODL */
{
    int       status = 0;               /* (out) body that satisfies criterion (bias-0) */

    modl_T    *MODL = (modl_T*)modl;

    int       AT_nbody=0,  AT_nface=0, AT_nedge=0,  AT_nnode=0;
    int       AT_ibody=0,  AT_iface=0, AT_iedge=0,  AT_inode=0, AT_ibody1=0, AT_ibody2=0;
    int       AT_xmin=0,   AT_ymin=0,  AT_zmin=0,   AT_xmax=0,  AT_ymax=0,   AT_zmax=0;
    int       AT_length=0, AT_area=0,  AT_volume=0;
    int       AT_xcg=0,    AT_ycg=0,   AT_zcg=0;
    int       AT_Ixx=0,    AT_Ixy=0,   AT_Ixz=0;
    int       AT_Iyx=0,    AT_Iyy=0,   AT_Iyz=0;
    int       AT_Izx=0,    AT_Izy=0,   AT_Izz=0;
    double    box[6];

    int       itype, nlist, oclass, mtype, nchild, *sense;
    CINT      *tempIlist;
    double    data[14];
    CDOUBLE   *tempRlist;
    CCHAR     *tempClist;
    ego       ebody=NULL, eref, *echild, *etemp;

    ROUTINE(setupAtPmtrs);
    DPRINT1("%s() {",
            routine);

    /* --------------------------------------------------------------- */

    /* if there are no Bodys, then return without setting up any @-Parameters */
    if (MODL->nbody <= 0) {
        goto cleanup;
    }

    /* if current Body is NULL, return without setting up any @-Parameters */
    if (MODL->body[MODL->nbody].botype == OCSM_NULL_BODY) {
        goto cleanup;
    }

    /* get the Parameter indices for each of the variables (at this level) */
    status = ocsmFindPmtr(MODL, "@nbody", OCSM_INTERNAL, 1, 1, &AT_nbody);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@nface", OCSM_INTERNAL, 1, 1, &AT_nface);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@nedge", OCSM_INTERNAL, 1, 1, &AT_nedge);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@nnode", OCSM_INTERNAL, 1, 1, &AT_nnode);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@ibody", OCSM_INTERNAL, 1, 1, &AT_ibody);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@iface", OCSM_INTERNAL, 1, 1, &AT_iface);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@iedge", OCSM_INTERNAL, 1, 1, &AT_iedge);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@inode", OCSM_INTERNAL, 1, 1, &AT_inode);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@ibody1", OCSM_INTERNAL, 1, 1, &AT_ibody1);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@ibody2", OCSM_INTERNAL, 1, 1, &AT_ibody2);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@xmin", OCSM_INTERNAL, 1, 1, &AT_xmin);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@ymin", OCSM_INTERNAL, 1, 1, &AT_ymin);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@zmin", OCSM_INTERNAL, 1, 1, &AT_zmin);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@xmax", OCSM_INTERNAL, 1, 1, &AT_xmax);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@ymax", OCSM_INTERNAL, 1, 1, &AT_ymax);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@zmax", OCSM_INTERNAL, 1, 1, &AT_zmax);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@length", OCSM_INTERNAL, 1, 1, &AT_length);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@area", OCSM_INTERNAL, 1, 1, &AT_area);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@volume", OCSM_INTERNAL, 1, 1, &AT_volume);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@xcg", OCSM_INTERNAL, 1, 1, &AT_xcg);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@ycg", OCSM_INTERNAL, 1, 1, &AT_ycg);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@zcg", OCSM_INTERNAL, 1, 1, &AT_zcg);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@Ixx", OCSM_INTERNAL, 1, 1, &AT_Ixx);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@Ixy", OCSM_INTERNAL, 1, 1, &AT_Ixy);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@Ixz", OCSM_INTERNAL, 1, 1, &AT_Ixz);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@Iyx", OCSM_INTERNAL, 1, 1, &AT_Iyx);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@Iyy", OCSM_INTERNAL, 1, 1, &AT_Iyy);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@Iyz", OCSM_INTERNAL, 1, 1, &AT_Iyz);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@Izx", OCSM_INTERNAL, 1, 1, &AT_Izx);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@Izy", OCSM_INTERNAL, 1, 1, &AT_Izy);
    CHECK_STATUS(ocsmFindPmtr);

    status = ocsmFindPmtr(MODL, "@Izz", OCSM_INTERNAL, 1, 1, &AT_Izz);
    CHECK_STATUS(ocsmFindPmtr);

    /* if no Body is selected, select the last Body */
    if (MODL->sel_body < 0) {
        MODL->sel_body = MODL->nbody;
        SPRINT1(0, "WARNING:: selecting Body %d by default", MODL->sel_body);
    }

    /* find ebody (which actually points to a Body, Face, Edge, or Node) */
    if        (MODL->sel_face >= 0) {
        ebody = MODL->body[MODL->sel_body].face[MODL->sel_face].eface;
    } else if (MODL->sel_edge >= 0) {
        ebody = MODL->body[MODL->sel_body].edge[MODL->sel_edge].eedge;
    } else if (MODL->sel_node >= 0) {
        ebody = MODL->body[MODL->sel_body].node[MODL->sel_node].enode;
    } else {
        ebody = MODL->body[MODL->sel_body].ebody;
    }

    /* set @nbody and @ibody */
    status = ocsmSetValuD(MODL, AT_nbody, 1, 1, (double)(MODL->nbody));
    CHECK_STATUS(ocsmSetValuD);

    status = ocsmSetValuD(MODL, AT_ibody, 1, 1, (double)(MODL->sel_body));
    CHECK_STATUS(ocsmSetValuD);

    /* set @nface, @nedge, and @nnode */
    status = ocsmSetValuD(MODL, AT_nface, 1, 1, (double)(MODL->body[MODL->sel_body].nface));
    CHECK_STATUS(ocsmSetValuD);

    status = ocsmSetValuD(MODL, AT_nedge, 1, 1, (double)(MODL->body[MODL->sel_body].nedge));
    CHECK_STATUS(ocsmSetValuD);

    status = ocsmSetValuD(MODL, AT_nnode, 1, 1, (double)(MODL->body[MODL->sel_body].nnode));
    CHECK_STATUS(ocsmSetValuD);

    /* set @iface, @iedge, and @inode */
    status = ocsmSetValuD(MODL, AT_iface, 1, 1, (double)(MODL->sel_face));
    CHECK_STATUS(ocsmSetValuD);

    status = ocsmSetValuD(MODL, AT_iedge, 1, 1, (double)(MODL->sel_edge));
    CHECK_STATUS(ocsmSetValuD);

    status = ocsmSetValuD(MODL, AT_inode, 1, 1, (double)(MODL->sel_node));
    CHECK_STATUS(ocsmSetValuD);

    /* set @ibody1 and @ibody2 */
    if (MODL->sel_edge > 0 || MODL->sel_face > 0) {
        status = EG_attributeRet(ebody, "body", &itype, &nlist,
                                 &tempIlist, &tempRlist, &tempClist);
        CHECK_STATUS(EG_attributeRet);

        if (itype == ATTRINT && nlist == 2) {
            status = ocsmSetValuD(MODL, AT_ibody1, 1, 1, (double)(tempIlist[0]));
            CHECK_STATUS(ocsmSetValuD);

            status = ocsmSetValuD(MODL, AT_ibody2, 1, 1, (double)(tempIlist[1]));
            CHECK_STATUS(ocsmSetValuD);
        } else {
            status = ocsmSetValuD(MODL, AT_ibody1, 1, 1, -1.0);
            CHECK_STATUS(ocsmSetValuD);

            status = ocsmSetValuD(MODL, AT_ibody2, 1, 1, -1.0);
            CHECK_STATUS(ocsmSetValuD);
        }
    } else {
        status = ocsmSetValuD(MODL, AT_ibody1, 1, 1, -1.0);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_ibody2, 1, 1, -1.0);
        CHECK_STATUS(ocsmSetValuD);
    }

    /* set @xmin, @xmax, @ymin, @ymax, @zmin, and @zmax */
    if (MODL->sel_edge >= 0) {
        status = EG_getTopology(ebody, &eref, &oclass, &mtype,
                                data, &nchild, &echild, &sense);
        CHECK_STATUS(EG_getTopology);

        status = EG_getTopology(echild[0], &eref, &oclass, &mtype,
                                &(box[0]), &nchild, &etemp, &sense);
        CHECK_STATUS(EG_getTopology);

        status = EG_getTopology(echild[1], &eref, &oclass, &mtype,
                                &(box[3]), &nchild, &etemp, &sense);
        CHECK_STATUS(EG_getTopology);
    } else {
        status = EG_getBoundingBox(ebody, box);
        CHECK_STATUS(EG_attributeNum);
    }

    status = ocsmSetValuD(MODL, AT_xmin, 1, 1, box[0]);
    CHECK_STATUS(ocsmSetValuD);

    status = ocsmSetValuD(MODL, AT_ymin, 1, 1, box[1]);
    CHECK_STATUS(ocsmSetValuD);

    status = ocsmSetValuD(MODL, AT_zmin, 1, 1, box[2]);
    CHECK_STATUS(ocsmSetValuD);

    status = ocsmSetValuD(MODL, AT_xmax, 1, 1, box[3]);
    CHECK_STATUS(ocsmSetValuD);

    status = ocsmSetValuD(MODL, AT_ymax, 1, 1, box[4]);
    CHECK_STATUS(ocsmSetValuD);

    status = ocsmSetValuD(MODL, AT_zmax, 1, 1, box[5]);
    CHECK_STATUS(ocsmSetValuD);

    /* set mass properties */
    if (MODL->sel_node < 0) {
        status = EG_getMassProperties(ebody, data);
        CHECK_STATUS(EG_getMassProperties);

        /* set @length, @area, and @volume */
        if        (MODL->sel_face >= 0) {
            status = ocsmSetValuD(MODL, AT_length, 1, 1, 0.0);
            CHECK_STATUS(ocsmSetValuD);

            status = ocsmSetValuD(MODL, AT_area,   1, 1, data[1]);
            CHECK_STATUS(ocsmSetValuD);

            status = ocsmSetValuD(MODL, AT_volume, 1, 1, 0.0);
            CHECK_STATUS(ocsmSetValuD);
        } else if (MODL->sel_edge >= 0) {
            status = ocsmSetValuD(MODL, AT_length, 1, 1, data[1]);
            CHECK_STATUS(ocsmSetValuD);

            status = ocsmSetValuD(MODL, AT_area,   1, 1, 0.0);
            CHECK_STATUS(ocsmSetValuD);

            status = ocsmSetValuD(MODL, AT_volume, 1, 1, 0.0);
            CHECK_STATUS(ocsmSetValuD);
        } else {
            status = ocsmSetValuD(MODL, AT_length, 1, 1, 0.0);
            CHECK_STATUS(ocsmSetValuD);

            status = ocsmSetValuD(MODL, AT_area,   1, 1, data[1]);
            CHECK_STATUS(ocsmSetValuD);

            status = ocsmSetValuD(MODL, AT_volume, 1, 1, data[0]);
            CHECK_STATUS(ocsmSetValuD);
        }

        /* set @xcg, @ycg, and @zcg */
        status = ocsmSetValuD(MODL, AT_xcg, 1, 1, data[2]);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_ycg, 1, 1, data[3]);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_zcg, 1, 1, data[4]);
        CHECK_STATUS(ocsmSetValuD);

        /* set @Ixx, @Ixy, @Ixz, @Iyx, @Iyy, @Iyz, @Izx, @Izy, and @Izz */
        status = ocsmSetValuD(MODL, AT_Ixx, 1, 1, data[5]);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Ixy, 1, 1, data[6]);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Ixz, 1, 1, data[7]);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Iyx, 1, 1, data[8]);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Iyy, 1, 1, data[9]);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Iyz, 1, 1, data[10]);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Izx, 1, 1, data[11]);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Izy, 1, 1, data[12]);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Izz, 1, 1, data[13]);
        CHECK_STATUS(ocsmSetValuD);
    } else {
        status = ocsmSetValuD(MODL, AT_length, 1, 1, 0.0);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_area,   1, 1, 0.0);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_volume, 1, 1, 0.0);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_xcg,    1, 1, MODL->body[MODL->sel_body].node[MODL->sel_node].x);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_ycg,    1, 1, MODL->body[MODL->sel_body].node[MODL->sel_node].y);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_zcg,    1, 1, MODL->body[MODL->sel_body].node[MODL->sel_node].z);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Ixx,    1, 1, 0.0);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Ixy,    1, 1, 0.0);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Ixz,    1, 1, 0.0);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Iyx,    1, 1, 0.0);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Iyy,    1, 1, 0.0);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Iyz,    1, 1, 0.0);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Izx,    1, 1, 0.0);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Izy,    1, 1, 0.0);
        CHECK_STATUS(ocsmSetValuD);

        status = ocsmSetValuD(MODL, AT_Izz,    1, 1, 0.0);
        CHECK_STATUS(ocsmSetValuD);
    }

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   signalError - add to error message buffer                          *
 *                                                                      *
 ************************************************************************
 */

static void
signalError(
  /*@null@*/void   *modl,               /* (in)  pointer to MODL */
            int    status,              /* (in)  status flag */
            int    ibrch,               /* (in)  Branch index (bias-1) */
            char   format[],            /* (in)  format specifier */
            ...)                        /* (in)  variable arguments for format */
{
    va_list   args;

    char      errorBuf[MAX_STR_LEN];

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(signalError);
    DPRINT3("%s(status=%d, ibrch=%d) {",
            routine, status, ibrch);

    /* --------------------------------------------------------------- */

    /* set up the va structure */
    va_start(args, format);

    /* write the error in the console and add to the errorBuf */
    if (ibrch > 0) {
        snprintf(errorBuf, MAX_STR_LEN-1, "ERROR:: (%s) in Branch %s\n        ",
                 ocsmGetText(status), MODL->brch[ibrch].name);
    } else {
        snprintf(errorBuf, MAX_STR_LEN-1, "ERROR:: (%s)\n        ",
                 ocsmGetText(status));
    }
    printf("%s", errorBuf);
    if (MODL != NULL) {
        strncat(MODL->errorBuf, errorBuf, MAX_STR_LEN-1);
    }

    vsnprintf(errorBuf, MAX_STR_LEN-1, format, args);
    printf("%s", errorBuf);
    if (MODL != NULL) {
        strncat(MODL->errorBuf, errorBuf, MAX_STR_LEN-1);
    }

    snprintf(errorBuf, MAX_STR_LEN-1, "\n");
    printf("%s", errorBuf);
    if (MODL != NULL) {
        strncat(MODL->errorBuf, errorBuf, MAX_STR_LEN-1);
    }

    /* clean up the va structure */
    va_end(args);
}


/*
 ************************************************************************
 *                                                                      *
 *   solidBoolean - performs solid boolean and loosens tolers if needed *
 *                                                                      *
 ************************************************************************
 */

static int
solidBoolean(ego    ebodyl,             /* (in)  Body on left */
             ego    ebodyr,             /* (in)  Body on rite */
             int    type,               /* (in)  boolean type */
             double maxtol,             /* (in)  maximum allowable tolerance */
             ego    *emodel)            /* (out) model containing result */
{
    int       status = 0;               /* (out) body that satisfies criterion (bias-0) */

    int       oclass, mtype, nchild, *senses, itry, old_outLevel;
    double    tolerl, tolerr, data[4];
    ego       eref, *echilds, ebodyll=NULL, ebodyrr=NULL, context;

    ROUTINE(solidBoolean);
    DPRINT5("%s(ebodyl=%lx, ebodyr=%lx, type=%d, maxtol=%f) {",
            routine, (long)ebodyl, (long)ebodyr, type, maxtol);

    /* --------------------------------------------------------------- */

    status = EG_getContext(ebodyl, &context);
    CHECK_STATUS(EG_getContext);

    if (outLevel >= 3) {
        SPRINT0(3, "before EG_solidBoolean: ebodyl");
        printEgo(ebodyl);
        SPRINT0(3, "before EG_solidBoolean: ebodyr");
        printEgo(ebodyr);
    }

    /* if ebodyl is a model, just do a simple FUSION */
    status = EG_getTopology(ebodyl, &eref, &oclass, &mtype,
                            data, &nchild, &echilds, &senses);
    CHECK_STATUS(eg_gettopology);

    if (oclass == MODEL) {
        status = EG_solidBoolean(ebodyl, ebodyr, type, emodel);
        goto cleanup;
    }

    /* make copies of Bodys so that any increased tolerance will
       be applied only locally */
    status = EG_copyObject(ebodyl, NULL, &ebodyll);
    CHECK_STATUS(EG_copyObject);

    status = EG_copyObject(ebodyr, NULL, &ebodyrr);
    CHECK_STATUS(EG_copyObject);

    /* find the tolerances of the orginal Bodys */
    status = EG_getTolerance(ebodyll, &tolerl);
    CHECK_STATUS(EG_getTolerance);

    status = EG_getTolerance(ebodyrr, &tolerr);
    CHECK_STATUS(EG_getTolerance);
    SPRINT3(3, "    maxtol=%e, tolerl=%e, tolerr=%e", maxtol, tolerl, tolerr);

    /* loop to keep increasing tolerance until we either get a good
       result or we have reached the specified tolerance */
    for (itry = 0; itry < 10; itry++) {

        /* try the boolean operation */
        old_outLevel = EG_setOutLevel(context, 0);

        status = EG_solidBoolean(ebodyll, ebodyrr, type, emodel);
        SPRINT2(3, "    EG_solidBoolean(itry=%d) -> status=%d", itry, status);

        (void) EG_setOutLevel(context, old_outLevel);

        /* the operation succeeded */
        if (status == EGADS_SUCCESS) {

            /* if on the first try, so simply return */
            if (itry == 0) {
                goto cleanup;

            /* otherwise, generate a warning about new tolerance */
            } else {
                SPRINT1(0, "WARNING:: tolerance increased to %e to get good result", MAX(tolerl,tolerr));
                goto cleanup;
            }
        }

        /* otherwise, remove the bad model and try increasing the tolerance (by a factor of 4)
           up to the user-specified value */
        if (*emodel != NULL) {
            status = EG_deleteObject(*emodel);
            CHECK_STATUS(EG_deleteObject);
        }

        if        (tolerl < tolerr) {
            tolerl = tolerr;
        } else if (tolerr < tolerl) {
            tolerr = tolerl;
        } else if (tolerl < maxtol) {
            tolerl = MIN(4*tolerl, maxtol);
            tolerr = tolerl;
        } else {
            goto cleanup;
        }

        SPRINT1(3, "    increasing tolerl to %e", tolerl);
        status = EG_setTolerance(ebodyll, tolerl);
        CHECK_STATUS(EG_setTolerance);

        SPRINT1(3, "    increasing tolerr to %e", tolerr);
        status = EG_setTolerance(ebodyrr, tolerr);
        CHECK_STATUS(EG_setTolerance);
    }

    signalError(NULL, OCSM_INTERNAL_ERROR, 0,
                "max trys exceeded in solidBoolean");
    status = OCSM_INTERNAL_ERROR;

cleanup:
    if (ebodyll != NULL) EG_deleteObject(ebodyll);
    if (ebodyrr != NULL) EG_deleteObject(ebodyrr);

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   solveSketch - solve a Sketch                                       *
 *                                                                      *
 ************************************************************************
 */

static int
solveSketch(modl_T *modl,
            sket_T *sket)
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    int        jpmtr, jndex, ivar, icon, iter, niter;
    double     value, dot;
    double     *f0neg=NULL, *dfdx=NULL, *delx=NULL, f0max, f0last, save_value, omega, dfdx_max;

    ROUTINE(solveSketch);
    DPRINT1("%s() {",
            routine);

    /* --------------------------------------------------------------- */

    if (sket->solved == 1) {
        goto cleanup;
    }

    /* check for under- or over-constrained Sketch */
    if        (sket->ncon < sket->nvar) {
        SPRINT2(0, "WARNING:: under-constrained Sketch (nvar=%d but ncon=%d) will not be solved",
                sket->nvar, sket->ncon);
    } else if (sket->ncon > sket->nvar) {
        SPRINT2(0, "WARNING:: over-constrained Sketch (nvar=%d but ncon=%d) will not be solved",
                sket->nvar, sket->ncon);
    } else {

        /* set up for function evaluation and Jacobian */
        MALLOC(f0neg, double,               sket->ncon );
        MALLOC(dfdx,  double, (sket->nvar)*(sket->ncon));
        MALLOC(delx,  double,  sket->nvar              );

        /* Newton iteration to change the Sketch variables until
           the constraints are satisfied */
        niter  = 25;
        omega  = 0.25;
        f0last = 0;

        for (iter = 0; iter < niter; iter++) {

            /* evaluate the constraints */
            f0max = 0;
            for (icon = 0; icon < sket->ncon; icon++) {
                status = str2val(sket->con[icon], MODL, &value, &dot);
                CHECK_STATUS(str2val);

                f0neg[icon] = -value;
                SPRINT2(2, "       f0[%4d] = %11.4e", icon, value);

                if (fabs(value) > f0max) {
                    f0max = fabs(value);
                }
            }

            SPRINT2(1, "    -> solving   iter = %3d,   f0max = %12.4e", iter, f0max);

            /* if we have converged, stop the Newton iterations */
            if (f0max < EPS12) {
                DPRINT0("converged");
                break;
            }

            /* f0max < f0last, we are converging, so increase omega */
            if (f0max < f0last) {
                omega = MIN(1.2*omega, 1);
                /* otherwise, decrease omega */
            } else {
                omega = omega / 2.0;
            }
            f0last = f0max;

            /* build up the Jacobian matrix by perturbing the solver variables
               one at a time */
            for (ivar = 0; ivar < sket->nvar; ivar++) {
                jpmtr = sket->ipmtr[ivar];
                jndex = sket->index[ivar];

                save_value = MODL->pmtr[jpmtr].value[jndex];
                MODL->pmtr[jpmtr].value[jndex] += EPS06;

                for (icon = 0; icon < sket->ncon; icon++) {
                    status = str2val(sket->con[icon], MODL, &value, &dot);
                    CHECK_STATUS(str2val);

                    dfdx[icon*(sket->ncon)+ivar] = (value + f0neg[icon]) / EPS06;
                }

                MODL->pmtr[jpmtr].value[jndex] = save_value;
            }

            /* print out the Jacobian matrix */
            SPRINT0(2, "Jacobian matrix");
            for (icon = 0; icon < sket->ncon; icon++) {
                dfdx_max = 0;
                SPRINT1x(2, "%3d: ", icon);

                for (ivar = 0; ivar < sket->nvar; ivar++) {
                    SPRINT1x(2, "%8.4f ", dfdx[icon*(sket->ncon)+ivar]);
                    if (fabs(dfdx[icon*(sket->ncon)+ivar]) > dfdx_max) {
                        dfdx_max = fabs(dfdx[icon*(sket->ncon)+ivar]);
                    }
                }
                SPRINT1(2, " | %12.3e", dfdx_max);
            }

            /* take the Newton step */
            status = matsol(dfdx, f0neg, sket->ncon, delx);
            CHECK_STATUS(matsol);

            for (ivar = 0; ivar < sket->nvar; ivar++) {
                jpmtr = sket->ipmtr[ivar];
                jndex = sket->index[ivar];

                MODL->pmtr[jpmtr].value[jndex] += omega * delx[ivar];

                SPRINT4(2, "       x [%4d] = %11.5f  (%s[%d])",
                        ivar, MODL->pmtr[jpmtr].value[jndex], MODL->pmtr[jpmtr].name, jndex+1);
            }
        }

        /* print final solution */
        for (ivar = 0; ivar < sket->nvar; ivar++) {
            jpmtr = sket->ipmtr[ivar];
            jndex = sket->index[ivar];

            SPRINT3(1, "    -> updating  %s[%d] = %10.5f", MODL->pmtr[jpmtr].name, jndex+1,
                    MODL->pmtr[jpmtr].value[jndex]);
        }

        /* now that we have run out of iterations, check for convergence */
        if (f0max > EPS06) {
            status = OCSM_NOT_CONVERGED;
            goto cleanup;
        }

        sket->solved = 1;
    }

cleanup:
    FREE(delx );
    FREE(dfdx );
    FREE(f0neg);

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   str2rpn - convert expression to Rpn-code                           *
 *                                                                      *
 ************************************************************************
 */

static int
str2rpn(char      str[],                /* (in)  string containing expression */
        rpn_T     *rpn)                 /* (in)  pointer to Rpn-code */
{
    int       status = SUCCESS;         /* (out) return status */

    typedef struct {
        int       type;                 /* type of token */
        char      text[MAX_NAME_LEN];   /* text associated with token */
    } tok_T;

    tok_T     token[  MAX_EXPR_LEN],
              opstack[MAX_EXPR_LEN];
    int       istr, jstr, ntoken, nrpn, nopstack, nparen, nbrakt;
    int       i, j, count, type;
    char      text[MAX_NAME_LEN];

    ROUTINE(str2rpn);
    DPRINT2("%s(str=%s) {",
            routine, str);

#define ADD_TOKEN(TYPE, TEXT)                                \
    if (ntoken < MAX_EXPR_LEN-1) {                           \
        token[ntoken].type = TYPE;                           \
        STRNCPY(token[ntoken].text, TEXT, MAX_NAME_LEN);     \
        ntoken++;                                            \
    } else {                                                 \
        status = OCSM_TOKEN_STACK_OVERFLOW;                  \
        goto cleanup;                                        \
    }

#define ADD_TOKEN_CHAR(TYPE, CHAR)                           \
    if (ntoken < MAX_EXPR_LEN-1) {                           \
        token[ntoken].type = TYPE;                           \
        token[ntoken].text[0] = CHAR;                        \
        token[ntoken].text[1] = '\0';                        \
        ntoken++;                                            \
    } else {                                                 \
        status = OCSM_TOKEN_STACK_OVERFLOW;                  \
        goto cleanup;                                        \
    }

#define PUSH_OP(TYPE, TEXT)                                  \
    if (nopstack < MAX_EXPR_LEN-1) {                         \
        opstack[nopstack].type = TYPE;                       \
        STRNCPY(opstack[nopstack].text, TEXT, MAX_NAME_LEN); \
        nopstack++;                                          \
    } else {                                                 \
        status = OCSM_OP_STACK_OVERFLOW;                     \
        goto cleanup;                                        \
    }

#define POP_OP(TYPE, TEXT)                                    \
    if (nopstack > 0) {                                       \
        nopstack--;                                           \
        TYPE = opstack[nopstack].type;                        \
        STRNCPY(TEXT, opstack[nopstack].text, MAX_NAME_LEN);  \
    } else {                                                  \
        status = OCSM_OP_STACK_UNDERFLOW;                     \
        goto cleanup;                                         \
    }

#define PUSH_RPN(TYPE, TEXT)                                 \
    if (nrpn < MAX_EXPR_LEN-1) {                             \
        rpn[nrpn].type = TYPE;                               \
        STRNCPY(rpn[nrpn].text, TEXT, MAX_NAME_LEN);         \
        nrpn++;                                              \
    } else {                                                 \
        status = OCSM_RPN_STACK_OVERFLOW;                    \
        goto cleanup;                                        \
    }

#define POP_RPN(TYPE, TEXT)                                  \
    if (nrpn > 0) {                                          \
        nrpn--;                                              \
        TYPE = rpn[nrpn].type;                               \
        STRNCPY(TEXT, rpn[nrpk].text, MAX_NAME_LEN);         \
    } else {                                                 \
        status = OCSM_RPN_STACK_UNDERFLOW;                   \
        goto cleanup;                                        \
    }

    /* --------------------------------------------------------------- */

    SPRINT1(3, "str=%s", str);

    /* convert str to tokens
       - surrounded by parentheses
       - remove white space
       - check for valid character
       - convert unary ops to binary ops:
            (+  to  (0+
            (-  to  (0-
            [+  to  [0+
            [-  to  [0-
            ,+  to  ,0+
            ,-  to  ,0-
       - check count and nesting of parentheses
    */
    ntoken = 0;               /* number of tokens */
    nparen = 1;               /* number of unmatched open parens */
    nbrakt = 0;               /* number of unmatched open brackets */

    /* if first character is dollarsign, then str is a string, so return 0  */
    if (str[0] == '$') {
        nrpn = 0;
        PUSH_RPN(PARSE_STRING, " ");
        goto cleanup;
    }

    /* first token is an open paren */
    ADD_TOKEN(PARSE_OPENP, "(");

    /* loop through all characters of the string containing the expression */
    for (istr = 0; istr < strlen(str); istr++) {

        /* ignore exclamation point (for backward compatibility) */
        if        (str[istr] == '!') {

        /* open parenthesis */
        } else if (str[istr] == '(') {

            /* add the new PARSE_OPENP token */
            ADD_TOKEN(PARSE_OPENP, "(");

            nparen++;

        /* close parenthesis */
        } else if (str[istr] == ')') {

            /* add the new PARSE_CLOSEP token */
            ADD_TOKEN(PARSE_CLOSEP, ")");

            nparen--;

            /* make sure there is a matching open for this close */
            if (nparen < 1) {
                SPRINT1(1, ") before ( at character %d", istr);
                status = OCSM_CLOSE_BEFORE_OPEN;
                goto cleanup;
            }

        /* open bracket */
        } else if (str[istr] == '[') {

            /* add the new PARSE_OPENB token */
            ADD_TOKEN(PARSE_OPENB, "[");

            nbrakt++;

        /* close bracket */
        } else if (str[istr] == ']') {

            /* add the new PARSE_CLOSEB token */
            ADD_TOKEN(PARSE_CLOSEB, "]");

            nbrakt--;

            /* make sure there is a matching open for this close */
            if (nbrakt < 0) {
                SPRINT1(1, "] before [ at character %d", istr);
                status = OCSM_CLOSE_BEFORE_OPEN;
                goto cleanup;
            }

        /* comma (which separates the arguments of a function) */
        } else if (str[istr] == ',') {

            /* add the new PARSE_COMMA token */
            ADD_TOKEN(PARSE_COMMA, ",");

        /* digit (0-9) or period (which starts a number) */
        } else if (str[istr] == '.' || isdigit(str[istr]) != 0) {

            /* start a new PARSE_NUMBER token */
            ADD_TOKEN_CHAR(PARSE_NUMBER, str[istr]);

            /* add new characters from str as long as we are within a number */
            i = 1;
            for (jstr = istr+1; jstr < strlen(str); jstr++) {
                if (isdigit(str[jstr]) != 0 || str[jstr] == '.') {
                    token[ntoken-1].text[i  ] = str[jstr];
                    token[ntoken-1].text[i+1] = '\0';
                    i++;
                    istr++;

                /* if we find an 'E' or 'e' while making a number, put it
                   and the character after it (which might be a + or - or
                   digit) into the number */
                } else if (str[jstr] == 'E' || str[jstr] == 'e') {
                    token[ntoken-1].text[i  ] = str[jstr  ];
                    token[ntoken-1].text[i+1] = str[jstr+1];
                    token[ntoken-1].text[i+2] = '\0';
                    i++; i++;
                    istr++; istr++;
                    jstr++;   /* advance jstr since we consumed 2 characters */
                } else {
                    break;
                }
            }

        /* letter (a-z, A-Z, :, and @) (which starts a name or func) */
        } else if (isalpha(str[istr]) != 0 || str[istr] == ':'
                                           || str[istr] == '@') {

            /* start a new PARSE_NAME token */
            ADD_TOKEN_CHAR(PARSE_NAME, str[istr]);

            /* add new characters from str as long as we are within a name */
            i = 1;
            for (jstr = istr+1; jstr < strlen(str); jstr++) {
                if (isalpha(str[jstr]) !=  0  || isdigit(str[jstr]) !=  0 ||
                            str[jstr]  == '_' ||         str[jstr]  == '@'||
                            str[jstr]  == ':' ||         str[jstr]  == '.'  ) {
                    token[ntoken-1].text[i  ] = str[jstr];
                    token[ntoken-1].text[i+1] = '\0';
                    i++;
                    istr++;

                /* ignore exclamation points */
                } else if (str[jstr] == '!') {
                    istr++;

                /* convert to PARSE_FUNC if followed by open parenthesis */
                } else if (str[jstr] == '(') {
                    token[ntoken-1].type = PARSE_FUNC;
                    break;

                /* convert to PARSE_ARRAY if followed by open bracket */
                } else if (str[jstr] == '[') {
                    token[ntoken-1].type = PARSE_ARRAY;
                    break;

                } else {
                    break;
                }
            }

        /* plus or minus */
        } else if (str[istr] == '+' || str[istr] == '-') {

            /* if previous character is comma or open, convert this
               unary +/- into a binary +/- by adding a "0" before it */
            if (token[ntoken-1].type == PARSE_OPENP ||
                token[ntoken-1].type == PARSE_OPENB ||
                token[ntoken-1].type == PARSE_COMMA   ) {
                ADD_TOKEN(PARSE_NUMBER, "0");
            }

            /* add the new PARSE_OP1 token */
            ADD_TOKEN_CHAR(PARSE_OP1, str[istr]);

        /* asterisk or slash */
        } else if (str[istr] == '*' || str[istr] == '/') {

            /* add the new PARSE_OP2 token */
            ADD_TOKEN_CHAR(PARSE_OP2, str[istr]);

        /* caret */
        } else if (str[istr] == '^') {

            /* add the new PARSE_OP3 token */
            ADD_TOKEN_CHAR(PARSE_OP3, str[istr]);

        /* white space (' ' or \t or \n) */
        } else if (str[istr] == ' ' || str[istr] == '\t' || str[istr] == '\n') {

        /* illegal character */
        } else {
            SPRINT2(1, "Illegal character at character %d (%c)", istr, str[istr]);
            status = OCSM_ILLEGAL_CHAR_IN_EXPR;
            goto cleanup;
        }
    }

    /* add a close parenthesis at the end */
    ADD_TOKEN(PARSE_CLOSEP, ")");

    nparen--;

    /* verify that the open and close parentheses are balanced */
    if        (nparen < 0) {
        SPRINT0(1, "More ) than (");
        status = OCSM_CLOSE_BEFORE_OPEN;
        goto cleanup;
    } else if (nparen > 0) {
        SPRINT0(1, "More ( than )");
        status = OCSM_MISSING_CLOSE;
        goto cleanup;
    }

    if        (nbrakt < 0) {
        SPRINT0(1, "More ] than [:");
        status = OCSM_CLOSE_BEFORE_OPEN;
        goto cleanup;
    } else if (nbrakt > 0) {
        SPRINT0(1, "More [ than ]");
        status = OCSM_MISSING_CLOSE;
        goto cleanup;
    }

    /* verify that the open and close parentheses are properly nested */
    for (i = 1; i < ntoken; i++) {
        if (token[i].type == PARSE_CLOSEP ||
            token[i].type == PARSE_CLOSEB   ) {
            count = 0;
            for (j = i-1; j >= 0; j--) {
                if (token[j].type == PARSE_OPENP ||
                    token[j].type == PARSE_OPENB   ) {
                    if (count == 0) {
                        if        (token[i].type == PARSE_CLOSEP &&
                                   token[j].type != PARSE_OPENP    ) {
                            SPRINT0(1, ") found but expected ]");
                            status = OCSM_IMPROPER_NESTING;
                            goto cleanup;
                        } else if (token[i].type == PARSE_CLOSEB &&
                                   token[j].type != PARSE_OPENB    ) {
                            SPRINT0(1, "] found by expected )");
                            status = OCSM_IMPROPER_NESTING;
                            goto cleanup;
                        }
                        break;
                    } else {
                        count--;
                    }
                } else if (token[j].type == PARSE_CLOSEP ||
                           token[j].type == PARSE_CLOSEB   ) {
                    count++;
                }
            }
        }
    }

    /* check for a proper sequencing of tokens */
    for (i = 0; i < ntoken-1; i++) {
        if         (token[i].type == PARSE_OP1  ||
                    token[i].type == PARSE_OP2  ||
                    token[i].type == PARSE_OP3    ) {
            if (token[i+1].type == PARSE_OP1    ||
                token[i+1].type == PARSE_OP2    ||
                token[i+1].type == PARSE_OP3    ||
                token[i+1].type == PARSE_CLOSEP ||
                token[i+1].type == PARSE_OPENB  ||
                token[i+1].type == PARSE_CLOSEB ||
                token[i+1].type == PARSE_COMMA    ) {
                SPRINT2(1, "Illegal token sequence: %s %s", token[i].text, token[i+1].text);
                status = OCSM_ILLEGAL_TOKEN_SEQUENCE;
                goto cleanup;
            }
        } else if (token[i].type == PARSE_OPENP) {
            if (token[i+1].type == PARSE_OP1    ||
                token[i+1].type == PARSE_OP2    ||
                token[i+1].type == PARSE_OP3    ||
                token[i+1].type == PARSE_CLOSEP ||
                token[i+1].type == PARSE_OPENB  ||
                token[i+1].type == PARSE_CLOSEB ||
                token[i+1].type == PARSE_COMMA    ) {
                SPRINT2(1, "Illegal token sequence: %s %s", token[i].text, token[i+1].text);
                status = OCSM_ILLEGAL_TOKEN_SEQUENCE;
                goto cleanup;
            }
        } else if (token[i].type == PARSE_CLOSEP) {
            if (token[i+1].type == PARSE_OPENP  ||
                token[i+1].type == PARSE_OPENB  ||
                token[i+1].type == PARSE_NAME   ||
                token[i+1].type == PARSE_ARRAY  ||
                token[i+1].type == PARSE_FUNC   ||
                token[i+1].type == PARSE_NUMBER   ) {
                SPRINT2(1, "Illegal token sequence: %s %s", token[i].text, token[i+1].text);
                status = OCSM_ILLEGAL_TOKEN_SEQUENCE;
                goto cleanup;
            }
        } else if (token[i].type == PARSE_OPENB) {
            if (token[i+1].type == PARSE_OP1    ||
                token[i+1].type == PARSE_OP2    ||
                token[i+1].type == PARSE_OP3    ||
                token[i+1].type == PARSE_CLOSEP ||
                token[i+1].type == PARSE_OPENB  ||
                token[i+1].type == PARSE_CLOSEB ||
                token[i+1].type == PARSE_COMMA    ) {
                SPRINT2(1, "Illegal token sequence: %s %s", token[i].text, token[i+1].text);
                status = OCSM_ILLEGAL_TOKEN_SEQUENCE;
                goto cleanup;
            }
        } else if (token[i].type == PARSE_CLOSEB) {
            if (token[i+1].type == PARSE_OPENP  ||
                token[i+1].type == PARSE_OPENB  ||
                token[i+1].type == PARSE_NAME   ||
                token[i+1].type == PARSE_ARRAY  ||
                token[i+1].type == PARSE_FUNC   ||
                token[i+1].type == PARSE_NUMBER   ) {
                SPRINT2(1, "Illegal token sequence: %s %s", token[i].text, token[i+1].text);
                status = OCSM_ILLEGAL_TOKEN_SEQUENCE;
                goto cleanup;
            }
        } else if (token[i].type == PARSE_COMMA) {
            if (token[i+1].type == PARSE_OP1    ||
                token[i+1].type == PARSE_OP2    ||
                token[i+1].type == PARSE_OP3    ||
                token[i+1].type == PARSE_CLOSEP ||
                token[i+1].type == PARSE_OPENB  ||
                token[i+1].type == PARSE_CLOSEB ||
                token[i+1].type == PARSE_COMMA    ) {
                SPRINT2(1, "Illegal token sequence: %s %s", token[i].text, token[i+1].text);
                status = OCSM_ILLEGAL_TOKEN_SEQUENCE;
                goto cleanup;
            }
        } else if (token[i].type == PARSE_NAME) {
            if (token[i+1].type == PARSE_OPENP  ||
                token[i+1].type == PARSE_NAME   ||
                token[i+1].type == PARSE_ARRAY  ||
                token[i+1].type == PARSE_FUNC   ||
                token[i+1].type == PARSE_NUMBER   ) {
                SPRINT2(1, "Illegal token sequence: %s %s", token[i].text, token[i+1].text);
                status = OCSM_ILLEGAL_TOKEN_SEQUENCE;
                goto cleanup;
            }
        } else if (token[i].type == PARSE_ARRAY) {
            if (token[i+1].type == PARSE_OPENP  ||
                token[i+1].type == PARSE_NAME   ||
                token[i+1].type == PARSE_ARRAY  ||
                token[i+1].type == PARSE_FUNC   ||
                token[i+1].type == PARSE_NUMBER   ) {
                SPRINT2(1, "Illegal token sequence: %s %s", token[i].text, token[i+1].text);
                status = OCSM_ILLEGAL_TOKEN_SEQUENCE;
                goto cleanup;
            }
        } else if (token[i].type == PARSE_NUMBER) {
            if (token[i+1].type == PARSE_OPENP  ||
                token[i+1].type == PARSE_OPENB  ||
                token[i+1].type == PARSE_NAME   ||
                token[i+1].type == PARSE_ARRAY  ||
                token[i+1].type == PARSE_FUNC   ||
                token[i+1].type == PARSE_NUMBER   ) {
                SPRINT2(1, "Illegal token sequence: %s %s", token[i].text, token[i+1].text);
                status = OCSM_ILLEGAL_TOKEN_SEQUENCE;
                goto cleanup;
            }
        } else if (token[i].type == PARSE_FUNC    ) {
            if (token[i+1].type != PARSE_OPENP    ) {
                SPRINT2(1, "Illegal token sequence: %s %s", token[i].text, token[i+1].text);
                status = OCSM_ILLEGAL_TOKEN_SEQUENCE;
                goto cleanup;
            }
        }
    }

    /* make sure all function names are known */
    for (i = 0; i < ntoken; i++) {
        if (token[i].type == PARSE_FUNC) {
            if (strcmp(token[i].text, "pi"     ) == 0 ||
                strcmp(token[i].text, "min"    ) == 0 ||
                strcmp(token[i].text, "max"    ) == 0 ||
                strcmp(token[i].text, "sqrt"   ) == 0 ||
                strcmp(token[i].text, "abs"    ) == 0 ||
                strcmp(token[i].text, "int"    ) == 0 ||
                strcmp(token[i].text, "nint"   ) == 0 ||
                strcmp(token[i].text, "ceil"   ) == 0 ||
                strcmp(token[i].text, "floor"  ) == 0 ||
                strcmp(token[i].text, "sign"   ) == 0 ||
                strcmp(token[i].text, "exp"    ) == 0 ||
                strcmp(token[i].text, "log"    ) == 0 ||
                strcmp(token[i].text, "log10"  ) == 0 ||
                strcmp(token[i].text, "sin"    ) == 0 ||
                strcmp(token[i].text, "sind"   ) == 0 ||
                strcmp(token[i].text, "asin"   ) == 0 ||
                strcmp(token[i].text, "asind"  ) == 0 ||
                strcmp(token[i].text, "cos"    ) == 0 ||
                strcmp(token[i].text, "cosd"   ) == 0 ||
                strcmp(token[i].text, "acos"   ) == 0 ||
                strcmp(token[i].text, "acosd"  ) == 0 ||
                strcmp(token[i].text, "tan"    ) == 0 ||
                strcmp(token[i].text, "tand"   ) == 0 ||
                strcmp(token[i].text, "atan"   ) == 0 ||
                strcmp(token[i].text, "atand"  ) == 0 ||
                strcmp(token[i].text, "atan2"  ) == 0 ||
                strcmp(token[i].text, "atan2d" ) == 0 ||
                strcmp(token[i].text, "hypot"  ) == 0 ||
                strcmp(token[i].text, "Xcent"  ) == 0 ||
                strcmp(token[i].text, "Ycent"  ) == 0 ||
                strcmp(token[i].text, "Xmidl"  ) == 0 ||
                strcmp(token[i].text, "Ymidl"  ) == 0 ||
                strcmp(token[i].text, "seglen" ) == 0 ||
                strcmp(token[i].text, "incline") == 0 ||
                strcmp(token[i].text, "radius" ) == 0 ||
                strcmp(token[i].text, "sweep"  ) == 0 ||
                strcmp(token[i].text, "turnang") == 0 ||
                strcmp(token[i].text, "dip"    ) == 0 ||
                strcmp(token[i].text, "ifzero" ) == 0 ||
                strcmp(token[i].text, "ifpos"  ) == 0 ||
                strcmp(token[i].text, "ifneg"  ) == 0 ||
                strcmp(token[i].text, "ifnan"  ) == 0   ) {
            } else {
                SPRINT1(1, "Illegel function name: %s", token[i].text);
                status = OCSM_ILLEGAL_FUNC_NAME;
                goto cleanup;
            }
        }
    }

    /* make sure all numbers are properly formed */
    for (i = 0; i < ntoken; i++) {
        if (token[i].type == PARSE_NUMBER) {
            count = 0;
            for (j = 0; j < strlen(token[i].text); j++) {
                if (token[i].text[j] == '.') count++;
            }

            if (count > 1) {
                SPRINT1(3, "Illegal number: %s", token[i].text);
                status = OCSM_ILLEGAL_NUMBER;
                goto cleanup;
            }
        }
    }

    /* print the tokens */
    if (outLevel >= 3) {
        SPRINT0(3, "token list");
        for (i = 0; i < ntoken; i++) {
            SPRINT3(3, "    %3d: type=%2d, text=%s", i, token[i].type, token[i].text);
        }
    }

    /* start with empty Rpn-code and op-stacks */
    nrpn     = 0;
    nopstack = 0;

    /* create the Rpn-code stack by cycling through the tokens */
    for (i = 0; i < ntoken; i++) {

        /* PARSE_NAME and PARSE_NUMBER */
        if (token[i].type == PARSE_NAME  ||
            token[i].type == PARSE_NUMBER  ) {
            PUSH_RPN(token[i].type, token[i].text);

        /* PARSE_OP1 */
        } else if (token[i].type == PARSE_OP1) {
            while (nopstack > 0) {
                POP_OP(type, text);
                if (type == PARSE_OPENP || type == PARSE_OPENB ||
                    type == PARSE_COMMA ||
                    type == PARSE_FUNC  || type == PARSE_ARRAY  ) {
                    PUSH_OP(type, text);
                    break;
                } else {
                    PUSH_RPN(type, text);
                }
            }

            PUSH_OP(token[i].type, token[i].text);

        /* PARSE_OP2 */
        } else if (token[i].type == PARSE_OP2) {
            while (nopstack > 0) {
                POP_OP(type, text);
                if (type == PARSE_OPENP || type == PARSE_OPENB ||
                    type == PARSE_COMMA ||
                    type == PARSE_FUNC  || type == PARSE_ARRAY ||
                    type == PARSE_OP1                            ) {
                    PUSH_OP(type, text);
                    break;
                } else {
                    PUSH_RPN(type, text);
                }
            }

            PUSH_OP(token[i].type, token[i].text);

        /* PARSE_OP3 */
        } else if (token[i].type == PARSE_OP3) {
            while (nopstack > 0) {
                POP_OP(type, text);
                if (type == PARSE_OPENP || type == PARSE_OPENB ||
                    type == PARSE_COMMA ||
                    type == PARSE_FUNC  || type == PARSE_ARRAY ||
                    type == PARSE_OP1   || type == PARSE_OP2    ) {
                    PUSH_OP(type, text);
                    break;
                } else {
                    PUSH_RPN(type, text);
                }
            }

            PUSH_OP(token[i].type, token[i].text);

        /* PARSE_OPENP */
        } else if (token[i].type == PARSE_OPENP) {
            PUSH_OP(token[i].type, token[i].text);

        /* PARSE_CLOSEP */
        } else if (token[i].type == PARSE_CLOSEP) {
            while (nopstack > 0) {
                POP_OP(type, text);
                if  (type == PARSE_OPENP) {
                    if (nopstack <= 0) {

                    } else if (opstack[nopstack-1].type == PARSE_FUNC) {
                        POP_OP(type, text);
                        PUSH_RPN(type, text);
                    }
                    break;
                } else if (type == PARSE_OPENB) {
                    SPRINT0(1, "Found  [ ... )");
                    status = OCSM_ILLEGAL_TOKEN_SEQUENCE;
                    goto cleanup;
                } else {
                    PUSH_RPN(type, text);
                }
            }

        /* PARSE_OPENB */
        } else if (token[i].type == PARSE_OPENB) {
            PUSH_OP(token[i].type, token[i].text);

        /* PARSE_CLOSEB */
        } else if (token[i].type == PARSE_CLOSEB) {
            while (nopstack > 0) {
                POP_OP(type, text);
                if (type == PARSE_OPENP) {
                    SPRINT0(1, "Found ( ... ]");
                    status = OCSM_ILLEGAL_TOKEN_SEQUENCE;
                    goto cleanup;
                } else if (type == PARSE_OPENB || type == PARSE_COMMA) {
                    PUSH_OP(type, text);
                    break;
                } else {
                    PUSH_RPN(type, text);
                }
            }

            POP_OP(type, text);

            /* two subscripts for array */
            if (type == PARSE_COMMA) {
                POP_OP(type, text);
                POP_OP(type, text);
                PUSH_RPN(type, text);

            /* one subscript for array */
            } else if (type == PARSE_OPENB) {
                PUSH_RPN(PARSE_NUMBER, "0");

                POP_OP(type, text);
                PUSH_RPN(type, text);

            } else {
                SPRINT0(1, "Expected array name when ] found");
                status = OCSM_ILLEGAL_TOKEN_SEQUENCE;
                goto cleanup;
            }

        /* PARSE_COMMA */
        } else if (token[i].type == PARSE_COMMA) {
            while (nopstack > 0) {
                POP_OP(type, text);
                if (type == PARSE_OPENP) {
                    PUSH_OP(type, text);
                    break;
                } else if (type == PARSE_OPENB) {
                    PUSH_OP(type, text);
                    PUSH_OP(PARSE_COMMA, ",");
                    break;
                } else {
                    PUSH_RPN(type, text);
                }
            }

        /* PARSE_FUNC */
        } else if (token[i].type == PARSE_FUNC) {
            PUSH_OP(token[i].type, token[i].text);

        /* PARSE_ARRAY */
        } else if (token[i].type == PARSE_ARRAY) {
            PUSH_OP(token[i].type, token[i].text);

        }
    }

    /* add everything left on the op-stack to the Rpn-code */
    while (nopstack > 0) {
        POP_OP(type, text);

        PUSH_RPN(type, text);
    }

    /* add a PARSE_END to the end of the Rpn-code */
    PUSH_RPN(PARSE_END, "");

    /* print the Rpn-code */
    if (outLevel >= 3) {
        SPRINT0(3, "rpn-code list");
        for (i = 0; i < nrpn; i++) {
            SPRINT3(3, "    %3d: type=%2d, text=%s", i, rpn[i].type, rpn[i].text);
        }
    }

#undef PUSH_OP
#undef POP_OP

cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   str2val - convert and evaluate expression (in string)              *
 *                                                                      *
 ************************************************************************
 */

static int
str2val(char      str[],                /* (in)  string containing expression */
        modl_T    *modl,                /* (in)  pointer to MODL */
        double    *val,                 /* (out) value      of expression */
        double    *dot)                 /* (out) derivative of expression */
{
    int       status = SUCCESS;         /* (out) return status */

    modl_T    *MODL = (modl_T*)modl;

    rpn_T     *rpn;                     /* Rpn-code */

    ROUTINE(str2val);
    DPRINT3("%s(str=%s, modl=%lx) {",
            routine, str, (long)modl);

    /* --------------------------------------------------------------- */

    MALLOC(rpn, rpn_T, MAX_STACK_SIZE);

    /* short-cut if expression is a single digit */
    if (strlen(str) == 1) {
        if (str[0] >= '0' && str[0] <= '9') {
            *val = str[0] - '0';
            *dot = 0;
            goto cleanup;
        }
    }

    /* convert the expression to Rpn-code */
    status = str2rpn(str, rpn);
    if (status != SUCCESS) {
        signalError(MODL, OCSM_ILLEGAL_VALUE, 0,
                    "could not parse \"%s\"", str);
    }
    CHECK_STATUS(str2rpn);

    /* evaluate the Rpn-code */
    status = evalRpn(rpn, modl, val, dot);
    if (status != SUCCESS) {
        signalError(MODL, OCSM_ILLEGAL_VALUE, 0,
                    "could not evaluate \"%s\"", str);
    }
    CHECK_STATUS(evalRpn);

cleanup:
    FREE(rpn);

    DPRINT4("%s --> status=%d, val=%f, dot=%f}", routine, status, *val, *dot);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   velocityOfEdge - compute Edge velocities analytically              *
 *                                                                      *
 ************************************************************************
 */

int
velocityOfEdge(modl_T *MODL,            /* (in)  pointer to MODL */
               int    ibody,            /* (in)  Body index (bias-1) */
               int    iedge,            /* (in)  Edge index (bias-1) */
               int    npnt,             /* (in)  number of points */
     /*@null@*/double t[],              /* (in)  para coords ( npnt in length) */
               double dxyz[])           /* (out) velocities (3*npnt in length) */
{
    int       status = 0;               /* (out) return status */

    int       jedge, kedge, ipnt, jbody, kbody, npnt_tess, ileft, irite, iarg, nchange;
    double    tt, uvleft[2], uvrite[2], dxyzleft[3], dxyzrite[3], xyzleft[18], xyzrite[18];
    double    normleft[4], normrite[4], A[9], b[3];
    double    data[18], data2[3], mat[12], ubar, vbar, wbar;
    double    dx, dy, dz, num, numdot, den, dendot;
    double    xbeg, ybeg, zbeg, xend, yend, zend, xbase, ybase, zbase, xvrtx, yvrtx, zvrtx, rad;
    double    tht, thtdot, phi, phidot, len, lendot, psi, psidot;
    double    x1, x1dot, y1, y1dot, z1, z1dot;
    double    x2, x2dot, y2, y2dot, z2, z2dot;
    double    x3, x3dot, y3, y3dot, z3, z3dot;
    double    x4, x4dot, y4, y4dot, z4, z4dot;
    double    x5, x5dot, r5, r5dot, t5, t5dot, s5, s5dot;
    double    *ts=NULL, *xyz_pnt=NULL;
    CDOUBLE   *xyz_tess, *t_tess;

    int         ibrch, udp_num, *udp_types, *udp_idef, needfd;
    int         mtype, oclass;
    double      *udp_ddef;
    char        **udp_names, primtype[MAX_EXPR_LEN];
    ego         eref, eprev, enext;

    ROUTINE(velocityOfEdge);
    DPRINT4("%s(ibody=%d, iedge=%d, npnt=%d) {",
            routine, ibody, iedge, npnt);

    /* --------------------------------------------------------------- */

    SPRINT3(2, "    velocityOfEdge(ibody=%d, iedge=%d, npnt=%d)", ibody, iedge, npnt);

    if        (ibody < 1 || ibody > MODL->nbody) {
        status = OCSM_ILLEGAL_BODY_INDEX;
        goto cleanup;
    } else if (iedge < 1 || iedge > MODL->body[ibody].nedge) {
        status = OCSM_EDGE_NOT_FOUND;
        goto cleanup;
    }

    /* if this is a degenerate Edge, return zero velocities */
    status = EG_getInfo(MODL->body[ibody].edge[iedge].eedge,
                        &oclass, &mtype, &eref, &eprev, &enext);
    CHECK_STATUS(EG_getInfo);

    if (mtype == DEGENERATE) {
        for (ipnt = 0; ipnt < npnt; ipnt++) {
            dxyz[3*ipnt  ] = 0;
            dxyz[3*ipnt+1] = 0;
            dxyz[3*ipnt+2] = 0;
        }

        goto cleanup;
    }

    /* if there is an active perturbation, get velocity by finite
       differences */
    if (MODL->perturb != NULL) {
        status = finiteDifference(MODL, ibody, OCSM_EDGE, iedge, npnt, t, dxyz);
        CHECK_STATUS(finiteDifference);

        goto cleanup;
    }

    /* if the Edge was generated by an OCSM_BOOLEAN or an OCSM_GROWN,
       the velocity can be obtained from Face velocities */
    jbody = MODL->body[ibody].edge[iedge].ibody;
    if (MODL->body[jbody].brtype == OCSM_INTERSECT ||
        MODL->body[jbody].brtype == OCSM_SUBTRACT  ||
        MODL->body[jbody].brtype == OCSM_UNION     ||
        MODL->body[jbody].brtype == OCSM_JOIN      ||
        MODL->body[jbody].brtype == OCSM_EXTRUDE   ||
        MODL->body[jbody].brtype == OCSM_RULE      ||
        MODL->body[jbody].brtype == OCSM_LOFT      ||
        MODL->body[jbody].brtype == OCSM_BLEND     ||
        MODL->body[jbody].brtype == OCSM_REVOLVE   ||
        MODL->body[jbody].brtype == OCSM_SWEEP       ) {

        ileft = MODL->body[ibody].edge[iedge].ileft;
        if (ileft < 1 || ileft > MODL->body[ibody].nface) {
            status = OCSM_FACE_NOT_FOUND;
            goto cleanup;
        }
        irite = MODL->body[ibody].edge[iedge].irite;
        if (irite < 1 || irite > MODL->body[ibody].nface) {
            status = OCSM_FACE_NOT_FOUND;
            goto cleanup;
        }

        /* if t==NULL, then we need to get tessellation info for iedge */
        if (t == NULL) {
            status = EG_getTessEdge(MODL->body[ibody].etess, iedge,
                                    &npnt_tess, &xyz_tess, &t_tess);
            CHECK_STATUS(EG_getTessEdge);

            if (npnt != npnt_tess) {
                status = OCSM_ILLEGAL_ARGUMENT;
                goto cleanup;
            }
        }

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            if (t == NULL) {
                tt = t_tess[ipnt];
            } else {
                tt = t[ipnt];
            }
            status = EG_evaluate(MODL->body[ibody].edge[iedge].eedge,
                                 &tt, data);
            CHECK_STATUS(EG_evaluate);

            /* velocity of left Face */
            status = EG_getEdgeUV(MODL->body[ibody].face[ileft].eface,
                                  MODL->body[ibody].edge[iedge].eedge, 0, tt, uvleft);
            CHECK_STATUS(EG_getEdgeUV);

            status = velocityOfFace(MODL, ibody, ileft, 1, uvleft, dxyzleft);
            CHECK_STATUS(velocityOfFace);

            /* velocity of rite Face */
            status = EG_getEdgeUV(MODL->body[ibody].face[irite].eface,
                                  MODL->body[ibody].edge[iedge].eedge, 0, tt, uvrite);
            CHECK_STATUS(EG_getEdgeUV);

            status = velocityOfFace(MODL, ibody, irite, 1, uvrite, dxyzrite);
            CHECK_STATUS(velocityOfFace);

            /* if all velocities are 0, then returned velocity is 0 too */
            if (fabs(dxyzleft[0]) < EPS06 &&
                fabs(dxyzleft[1]) < EPS06 &&
                fabs(dxyzleft[2]) < EPS06 &&
                fabs(dxyzrite[0]) < EPS06 &&
                fabs(dxyzrite[1]) < EPS06 &&
                fabs(dxyzrite[2]) < EPS06   ) {
                dxyz[3*ipnt  ] = 0;
                dxyz[3*ipnt+1] = 0;
                dxyz[3*ipnt+2] = 0;

            /* otherwise find a combination of the Face velocities */
            } else {

                /* left Face normal */
                status = EG_evaluate(MODL->body[ibody].face[ileft].eface, uvleft, xyzleft);
                CHECK_STATUS(EG_evaluate);

                normleft[0] = xyzleft[4] * xyzleft[8] - xyzleft[5] * xyzleft[7];
                normleft[1] = xyzleft[5] * xyzleft[6] - xyzleft[3] * xyzleft[8];
                normleft[2] = xyzleft[3] * xyzleft[7] - xyzleft[4] * xyzleft[6];
                normleft[3] = sqrt(normleft[0]*normleft[0] + normleft[1]*normleft[1] + normleft[2]*normleft[2]);

                /* rite Face normal */
                status = EG_evaluate(MODL->body[ibody].face[irite].eface, uvrite, xyzrite);
                CHECK_STATUS(EG_evaluate);

                normrite[0] = xyzrite[4] * xyzrite[8] - xyzrite[5] * xyzrite[7];
                normrite[1] = xyzrite[5] * xyzrite[6] - xyzrite[3] * xyzrite[8];
                normrite[2] = xyzrite[3] * xyzrite[7] - xyzrite[4] * xyzrite[6];
                normrite[3] = sqrt(normrite[0]*normrite[0] + normrite[1]*normrite[1] + normrite[2]*normrite[2]);

                /* just use face velocity if normleft and normrite are parallel */
                if (fabs(normleft[1]*normrite[2]-normleft[2]*normrite[1]) < EPS12*normleft[3]*normrite[3] &&
                    fabs(normleft[2]*normrite[0]-normleft[0]*normrite[2]) < EPS12*normleft[3]*normrite[3] &&
                    fabs(normleft[0]*normrite[1]-normleft[1]*normrite[0]) < EPS12*normleft[3]*normrite[3]   ) {
                    dxyz[3*ipnt  ] = dxyzleft[0];
                    dxyz[3*ipnt+1] = dxyzleft[1];
                    dxyz[3*ipnt+2] = dxyzleft[2];

                /* find a common velocity that is consistent with the
                   normal velocities of the left and rite Face */
                } else {
                    A[0] = normleft[0];
                    A[1] = normleft[1];
                    A[2] = normleft[2];
                    b[0] = normleft[0] * dxyzleft[0]
                         + normleft[1] * dxyzleft[1]
                         + normleft[2] * dxyzleft[2];

                    A[3] = normrite[0];
                    A[4] = normrite[1];
                    A[5] = normrite[2];
                    b[1] = normrite[0] * dxyzrite[0]
                         + normrite[1] * dxyzrite[1]
                         + normrite[2] * dxyzrite[2];

                    A[6] = data[3];
                    A[7] = data[4];
                    A[8] = data[5];
                    b[2] = 0;

                    status = matsol(A, b, 3, &(dxyz[3*ipnt]));
                    CHECK_STATUS(matsol);
                }
            }
        }

        goto cleanup;
    }

    /* jbody is the Body in which this Edge first appeared */
    jbody = MODL->body[ibody].edge[iedge].ibody;

    /* find the transformtion matrix which is required
       to track a point from its location in ibody back to
       its corresponding location in jbody */
    status = xformToOriginal(MODL, ibody, jbody, mat);
    CHECK_STATUS(xformToOriginal);

    /* if t==NULL, then find the locations of the tessellation points when
       the Edge was first created */
    if (t == NULL) {
        status = EG_getTessEdge(MODL->body[ibody].etess, iedge,
                                &npnt_tess, &xyz_tess, &t_tess);
        CHECK_STATUS(EG_getTessEdge);

        if (npnt != npnt_tess) {
            status = OCSM_ILLEGAL_ARGUMENT;
            goto cleanup;
        } else {
            MALLOC(xyz_pnt, double, 3*npnt_tess);

            /* xyz_pnt are the tessellation points transformed back
               to the their locations in jbody */
            for (ipnt = 0; ipnt < npnt_tess; ipnt++) {
                xyz_pnt[3*ipnt  ] = xyz_tess[3*ipnt  ] * mat[ 0]
                                  + xyz_tess[3*ipnt+1] * mat[ 1]
                                  + xyz_tess[3*ipnt+2] * mat[ 2] + mat[ 3];
                xyz_pnt[3*ipnt+1] = xyz_tess[3*ipnt  ] * mat[ 4]
                                  + xyz_tess[3*ipnt+1] * mat[ 5]
                                  + xyz_tess[3*ipnt+2] * mat[ 6] + mat[ 7];
                xyz_pnt[3*ipnt+2] = xyz_tess[3*ipnt  ] * mat[ 8]
                                  + xyz_tess[3*ipnt+1] * mat[ 9]
                                  + xyz_tess[3*ipnt+2] * mat[10] + mat[11];
            }
        }

    /* otherwise, find the locations at the given t's when the Edge
       was first created */
    } else {
        npnt_tess = npnt;
        MALLOC(xyz_pnt, double, 3*npnt_tess);

        for (ipnt = 0; ipnt < npnt_tess; ipnt++) {
            status = EG_evaluate(MODL->body[ibody].edge[iedge].eedge, &(t[ipnt]), data);
            CHECK_STATUS(EG_evaluate);

            xyz_pnt[3*ipnt  ] = data[0] * mat[ 0]
                              + data[1] * mat[ 1]
                              + data[2] * mat[ 2] + mat[ 3];
            xyz_pnt[3*ipnt+1] = data[0] * mat[ 4]
                              + data[1] * mat[ 5]
                              + data[2] * mat[ 6] + mat[ 7];
            xyz_pnt[3*ipnt+2] = data[0] * mat[ 8]
                              + data[1] * mat[ 9]
                              + data[2] * mat[10] + mat[11];
        }
    }

    /* find the number of Parameters that have changed */
    nchange = MODL->body[jbody].hasdots;

    /* if there are no Parameter changes, then the velocities are all 0 (and do not
       bother computing the derivatives below) */
    if (nchange == 0) {
        SPRINT0(2, "        -> setting to zero");
        for (ipnt = 0; ipnt < npnt; ipnt++) {
            dxyz[3*ipnt  ] = 0;
            dxyz[3*ipnt+1] = 0;
            dxyz[3*ipnt+2] = 0;
        }

    /* get the velocities for a box */
    } else if (MODL->body[jbody].brtype == OCSM_BOX) {
        SPRINT0(2, "        -> analytical for BOX");

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            data2[0] = xyz_pnt[3*ipnt  ];
            data2[1] = xyz_pnt[3*ipnt+1];
            data2[2] = xyz_pnt[3*ipnt+2];

            ubar = (data2[0] - MODL->body[jbody].arg[1].val[0]) / MODL->body[jbody].arg[4].val[0];
            vbar = (data2[1] - MODL->body[jbody].arg[2].val[0]) / MODL->body[jbody].arg[5].val[0];
            wbar = (data2[2] - MODL->body[jbody].arg[3].val[0]) / MODL->body[jbody].arg[6].val[0];

            dxyz[3*ipnt  ] = MODL->body[jbody].arg[1].dot[0] + MODL->body[jbody].arg[4].dot[0] * ubar;
            dxyz[3*ipnt+1] = MODL->body[jbody].arg[2].dot[0] + MODL->body[jbody].arg[5].dot[0] * vbar;
            dxyz[3*ipnt+2] = MODL->body[jbody].arg[3].dot[0] + MODL->body[jbody].arg[6].dot[0] * wbar;
        }

    /* get the velocities for a sphere */
    } else if (MODL->body[jbody].brtype == OCSM_SPHERE) {
        SPRINT0(2, "        -> analytical for SPHERE");

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            data2[0] = xyz_pnt[3*ipnt  ];
            data2[1] = xyz_pnt[3*ipnt+1];
            data2[2] = xyz_pnt[3*ipnt+2];

            ubar = (data2[0] - MODL->body[jbody].arg[1].val[0]) / MODL->body[jbody].arg[4].val[0];
            vbar = (data2[1] - MODL->body[jbody].arg[2].val[0]) / MODL->body[jbody].arg[4].val[0];
            wbar = (data2[2] - MODL->body[jbody].arg[3].val[0]) / MODL->body[jbody].arg[4].val[0];

            dxyz[3*ipnt  ] = MODL->body[jbody].arg[1].dot[0] + MODL->body[jbody].arg[4].dot[0] * ubar;
            dxyz[3*ipnt+1] = MODL->body[jbody].arg[2].dot[0] + MODL->body[jbody].arg[4].dot[0] * vbar;
            dxyz[3*ipnt+2] = MODL->body[jbody].arg[3].dot[0] + MODL->body[jbody].arg[4].dot[0] * wbar;
        }

    /* get the velocities for a cone */
    } else if (MODL->body[jbody].brtype == OCSM_CONE) {
        SPRINT0(2, "        -> analytical for CONE");

        xvrtx = MODL->body[jbody].arg[1].val[0];
        yvrtx = MODL->body[jbody].arg[2].val[0];
        zvrtx = MODL->body[jbody].arg[3].val[0];

        xbase = MODL->body[jbody].arg[4].val[0];
        ybase = MODL->body[jbody].arg[5].val[0];
        zbase = MODL->body[jbody].arg[6].val[0];

        rad   = MODL->body[jbody].arg[7].val[0];

        dx    = xbase - xvrtx;
        dy    = ybase - yvrtx;
        dz    = zbase - zvrtx;

        tht = atan2(dy, dx);
        phi = atan2(dz, sqrt(dy*dy + dx*dx));
        len = sqrt(dx*dx + dy*dy + dz*dz);
        psi = atan2(rad, len);

        num = dy;
        den = dx;
        if (fabs(num) > EPS06 && fabs(den) > EPS06) {
            numdot = MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0];
            dendot = MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0];
            thtdot = (dendot * num - numdot * den) / (num * num + den * den);
        } else {
            thtdot = 0;
        }

        num = dz;
        den = sqrt(dy*dy + dx*dx);
        if (fabs(num) > EPS06 && fabs(den) > EPS06) {
            numdot = MODL->body[jbody].arg[6].dot[0] - MODL->body[jbody].arg[3].dot[0];
            dendot = ( dx * (MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0])
                     + dy * (MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0])) / den;
            phidot = (dendot * num - numdot * den) / (num * num + den * den);
        } else {
            phidot = 0;
        }

        lendot = ( dx * (MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0])
                 + dy * (MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0])
                 + dz * (MODL->body[jbody].arg[6].dot[0] - MODL->body[jbody].arg[3].dot[0])) / len;

        psidot = (len * MODL->body[jbody].arg[7].dot[0] - lendot * rad) / (len * len * rad * rad);

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            x1 = xyz_pnt[3*ipnt  ];
            y1 = xyz_pnt[3*ipnt+1];
            z1 = xyz_pnt[3*ipnt+2];

            /* move to origin */
            x2 = x1 - xvrtx;
            y2 = y1 - yvrtx;
            z2 = z1 - zvrtx;

            /* rotate around z2 axis so that cone lies in (x3,z3) plane */
            x3 = x2 * cos(tht) + y2 * sin(tht);
            y3 = y2 * cos(tht) - x2 * sin(tht);
            z3 = z2;

            /* rotate around y3 axis so that cone lies along x3 axis */
            x4 = x3 * cos(phi) + z3 * sin(phi);
            y4 = y3;
            z4 = z3 * cos(phi) - x3 * sin(phi);

            /* conical coordinates (x, psi, theta) */
            x5 = x4;
            s5 = atan2(sqrt(y4*y4 + z4*z4), x4);   // psi5
            t5 = atan2(z4, y4);                    // theta5

            /* set velocities on the cone in its canonical orientation */
            x5dot = x5 / len * lendot;
            t5dot = 0;
            s5dot = s5 / psi * psidot;

            /* transform velocities back to original orientation */
            x4dot = x5dot;
            y4dot = x5dot * cos(t5)      * tan(s5)
                  + s5dot * cos(t5) * x5 / cos(s5) / cos(s5)
                  - t5dot * sin(t5) * x5 * tan(s5);
            z4dot = x5dot * sin(t5)      * tan(s5)
                  + s5dot * sin(t5) * x5 / cos(s5) / cos(s5)
                  + t5dot * cos(t5) * x5 * tan(s5);

            x3dot = x4dot * cos(phi) - x4 * phidot * sin(phi)
                  - z4dot * sin(phi) - z4 * phidot * cos(phi);
            y3dot = y4dot;
            z3dot = z4dot * cos(phi) - z4 * phidot * sin(phi)
                  + x4dot * sin(phi) + x4 * phidot * cos(phi);

            x2dot = x3dot * cos(tht) - x3 * thtdot * sin(tht)
                  - y3dot * sin(tht) - y3 * thtdot * cos(tht);
            y2dot = y3dot * cos(tht) - y3 * thtdot * sin(tht)
                  + x3dot * sin(tht) + x3 * thtdot * cos(tht);
            z2dot = z3dot;

            x1dot = x2dot + MODL->body[jbody].arg[1].dot[0];
            y1dot = y2dot + MODL->body[jbody].arg[2].dot[0];
            z1dot = z2dot + MODL->body[jbody].arg[3].dot[0];

            dxyz[3*ipnt  ] = x1dot;
            dxyz[3*ipnt+1] = y1dot;
            dxyz[3*ipnt+2] = z1dot;
        }

    /* get the velocities for a cylinder */
    } else if (MODL->body[jbody].brtype == OCSM_CYLINDER) {
        SPRINT0(2, "        -> analytical for CYLINDER");

        xbeg = MODL->body[jbody].arg[1].val[0];
        ybeg = MODL->body[jbody].arg[2].val[0];
        zbeg = MODL->body[jbody].arg[3].val[0];

        xend = MODL->body[jbody].arg[4].val[0];
        yend = MODL->body[jbody].arg[5].val[0];
        zend = MODL->body[jbody].arg[6].val[0];

        rad  = MODL->body[jbody].arg[7].val[0];

        dx   = xend - xbeg;
        dy   = yend - ybeg;
        dz   = zend - zbeg;

        tht = atan2(dy, dx);
        phi = atan2(dz, sqrt(dy*dy + dx*dx));
        len = sqrt(dx*dx + dy*dy + dz*dz);

        num = dy;
        den = dx;
        if (fabs(num) > EPS06 && fabs(den) > EPS06) {
            numdot = MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0];
            dendot = MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0];
            thtdot = (dendot * num - numdot * den) / (num * num + den * den);
        } else {
            thtdot = 0;
        }

        num = dz;
        den = sqrt(dy*dy + dx*dx);
        if (fabs(num) > EPS06 && fabs(den) > EPS06) {
            numdot = MODL->body[jbody].arg[6].dot[0] - MODL->body[jbody].arg[3].dot[0];
            dendot = ( dx * (MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0])
                     + dy * (MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0])) / den;
            phidot = (dendot * num - numdot * den) / (num * num + den * den);
        } else {
            phidot = 0;
        }

        lendot = ( dx * (MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0])
                 + dy * (MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0])
                 + dz * (MODL->body[jbody].arg[6].dot[0] - MODL->body[jbody].arg[3].dot[0])) / len;

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            x1 = xyz_pnt[3*ipnt  ];
            y1 = xyz_pnt[3*ipnt+1];
            z1 = xyz_pnt[3*ipnt+2];

            /* move to origin */
            x2 = x1 - xbeg;
            y2 = y1 - ybeg;
            z2 = z1 - zbeg;

            /* rotate around z2 axis so that cylinder lies in (x3,z3) plane */
            x3 = x2 * cos(tht) + y2 * sin(tht);
            y3 = y2 * cos(tht) - x2 * sin(tht);
            z3 = z2;

            /* rotate around y3 axis so that cylinder lies along x3 axis */
            x4 = x3 * cos(phi) + z3 * sin(phi);
            y4 = y3;
            z4 = z3 * cos(phi) - x3 * sin(phi);

            /* cylindrical coordinates (x, r, theta) */
            x5 = x4;
            r5 = sqrt(y4*y4 + z4*z4);   // rad5
            t5 = atan2(z4, y4);         // theta5

            /* set velocities on the cylinder in its canonical orientation */
            x5dot = x5 / len * lendot;
            r5dot = r5 / rad * MODL->body[jbody].arg[7].dot[0];
            t5dot = 0;

            /* transform velocities back to original orientation */
            x4dot = x5dot;
            y4dot = r5dot * cos(t5) - r5 * t5dot * sin(t5);
            z4dot = r5dot * sin(t5) + r5 * t5dot * cos(t5);

            x3dot = x4dot * cos(phi) - x4 * phidot * sin(phi)
                  - z4dot * sin(phi) - z4 * phidot * cos(phi);
            y3dot = y4dot;
            z3dot = z4dot * cos(phi) - z4 * phidot * sin(phi)
                  + x4dot * sin(phi) + x4 * phidot * cos(phi);

            x2dot = x3dot * cos(tht) - x3 * thtdot * sin(tht)
                  - y3dot * sin(tht) - y3 * thtdot * cos(tht);
            y2dot = y3dot * cos(tht) - y3 * thtdot * sin(tht)
                  + x3dot * sin(tht) + x3 * thtdot * cos(tht);
            z2dot = z3dot;

            x1dot = x2dot + MODL->body[jbody].arg[1].dot[0];
            y1dot = y2dot + MODL->body[jbody].arg[2].dot[0];
            z1dot = z2dot + MODL->body[jbody].arg[3].dot[0];

            dxyz[3*ipnt  ] = x1dot;
            dxyz[3*ipnt+1] = y1dot;
            dxyz[3*ipnt+2] = z1dot;
        }

    /* get the velocities for a udprim */
    } else if (MODL->body[jbody].brtype == OCSM_UDPRIM) {
        SPRINT0(2, "        -> analytical for UDPRIM");

        /* load and execute the user-defined primitive */
        ibrch = MODL->body[jbody].ibrch;

        STRNCPY(primtype, &(MODL->brch[ibrch].arg1[1]), MAX_EXPR_LEN);

        status = udp_initialize(primtype, &udp_num, &udp_names, &udp_types, &udp_idef, &udp_ddef);
        if (status == EGADS_NOLOAD) {
            status = EGADS_SUCCESS;
        }
        CHECK_STATUS(udp_initialize);

        status = udp_clrArguments(primtype);
        CHECK_STATUS(udp_clrArguments);

        needfd = 0;

        /* go back to first associated udparg statement whose primtype matches the
           udprim's primtype */
        kbody = jbody - 1;
        while (kbody > 0 && MODL->body[kbody].brtype == OCSM_UDPARG &&
               strcmp(MODL->body[kbody].arg[1].str, MODL->body[jbody].arg[1].str) == 0 ) {
            kbody--;
        }
        kbody++;

        while (kbody <= jbody) {
            for (iarg = 3; iarg < 10; iarg+=2) {
                if (MODL->body[kbody].arg[iarg].nval < 0) break;

                if (MODL->body[kbody].arg[iarg].nval > 0) {
                    status = udp_setVelocity(MODL->body[kbody].arg[1].str,
                                             MODL->body[jbody].ebody,
                                             MODL->body[kbody].arg[iarg-1].str,
                                             MODL->body[kbody].arg[iarg  ].dot,
                                             MODL->body[kbody].arg[iarg  ].nval);
                    if (status != EGADS_SUCCESS) {
                        /* indexerr indicates that a variable is not available for sensitivity,
                           (for example an integer), so we will just get zeroes */
                        if (status == EGADS_INDEXERR) {
                            SPRINT2(0, "WARNING:: sensitivity cannot be computed for \"%s\" in %s",
                                    MODL->body[kbody].arg[iarg-1].str, MODL->body[kbody].arg[1].str);
                        } else {
                            needfd++;
                        }
                    }
                }
            }

            kbody++;
        }

        /* if any of the velocities could not be set, we need finite differences */
        if (needfd > 0) {
            SPRINT0(2, "        -> analytical derivaitives not available, so reverting to finiteDifferences");

            status = finiteDifference(MODL, ibody, OCSM_EDGE, iedge, npnt, t, dxyz);
            CHECK_STATUS(finiteDifference);
            goto cleanup;
        }

        MALLOC(ts, double, npnt);

        if (t == NULL) {
            for (ipnt = 0 ; ipnt < npnt; ipnt++) {
                ts[ipnt]  = t_tess[ipnt];
            }
        } else {
            for (ipnt = 0 ; ipnt < npnt; ipnt++) {
                ts[ipnt] = t[ipnt];
            }
        }

        /* find the Edge in jbody that maps to Edge iedge in ibody */
        if (jbody == ibody) {
            jedge = iedge;
        } else if (MODL->body[ibody].edge[iedge].iford == -3) {
//$$$                /* this might not be right */
            jedge = iedge;
        } else {
            jedge = -1;
            for (kedge = 1; kedge <= MODL->body[jbody].nedge; kedge++) {
                if (MODL->body[jbody].edge[kedge].iford == MODL->body[ibody].edge[iedge].iford) {
                    jedge = kedge;
                    break;
                }
            }
        }
        SPRINT4(2, "        ibody=%d,iedge=%d maps to jbody=%d,jedge=%d", ibody, iedge, jbody, jedge);

        if (jedge > 0) {
            status = udp_sensitivity(primtype, MODL->body[jbody].ebody,
                                     npnt, OCSM_EDGE, jedge, ts, dxyz);
            CHECK_STATUS(udp_sensitivity);
        } else {
            signalError(MODL, OCSM_INTERNAL_ERROR, 0,
                        "cannot find Edge %d:%d in jbody=%d", ibody, iedge, jbody);
            status = OCSM_INTERNAL_ERROR;
            goto cleanup;
        }

    /* unknown primitive type, so get velocities by finite differences */
    } else {
        SPRINT1(2, "        -> using finiteDifference (brtype=%s)",
                ocsmGetText(MODL->body[jbody].brtype));

        status = finiteDifference(MODL, ibody, OCSM_EDGE, iedge, npnt, t, dxyz);
        CHECK_STATUS(finiteDifference);
        goto cleanup;
    }

    /* follow the children (up to the root) and modify the velocities
       for any OCSM_TRANSFORM that is visited */
    while (jbody != 0) {

        status = xformVelocity(MODL, jbody, npnt_tess, xyz_pnt, npnt, dxyz);
        CHECK_STATUS(xformVelocity);

        jbody = MODL->body[jbody].ichld;
    }

cleanup:
    FREE(ts);
    FREE(xyz_pnt);

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   velocityOfFace - compute Face velocities analytically              *
 *                                                                      *
 ************************************************************************
 */

int
velocityOfFace(modl_T *MODL,            /* (in)  pointer to MODL */
               int    ibody,            /* (in)  Body index (bias-1) */
               int    iface,            /* (in)  Face index (bias-1) */
               int    npnt,             /* (in)  number of points */
     /*@null@*/double uv[],             /* (in)  para coords (2*npnt in length) */
               double dxyz[])           /* (out) velocities  (3*npnt in length) */
{
    int       status = 0;               /* (out) return status */

    int       jface, kface, ipnt, jbody, kbody, lbody, npnt_tess, ntri_tess;
    int       iarg, nchange, nedge, nsketch=0, isketch[MAX_NUM_SKETCHES];
    int       isketchs, isketchn, iedge, i, j, iswap;
    CINT      *tris, *tric, *ptype, *pindx;
    double    params[2], data[18], data2[3], mat[12], ubar, vbar, wbar, trange[4];
    double    dx, dy, dz, num, numdot, den, dendot, frac;
    double    uvlimits[4], ulimits[4], ulimitn[5], *ts=NULL, *tn=NULL, *vels=NULL, *veln=NULL;
    double    xbeg, ybeg, zbeg, xend, yend, zend, xbase, ybase, zbase, xvrtx, yvrtx, zvrtx, rad;
    double    tht, thtdot, phi, phidot, len, lendot, psi, psidot;
    double    x1, x1dot, y1, y1dot, z1, z1dot;
    double    x2, x2dot, y2, y2dot, z2, z2dot;
    double    x3, x3dot, y3, y3dot, z3, z3dot;
    double    x4, x4dot, y4, y4dot, z4, z4dot;
    double    x5, x5dot, r5, r5dot, t5, t5dot, s5, s5dot;
    double    *uvs=NULL, *xyz_pnt=NULL;
    CDOUBLE   *xyz_tess, *uv_tess;

    int         ibrch, udp_num, *udp_types, *udp_idef, needfd, periodic;
    double      *udp_ddef;
    char        **udp_names, primtype[MAX_EXPR_LEN];

    ROUTINE(velocityOfFace);
    DPRINT4("%s(ibody=%d, iface=%d, npnt=%d) {",
            routine, ibody, iface, npnt);

    /* --------------------------------------------------------------- */

    SPRINT3(2, "    velocityOfFace(ibody=%d, iface=%d, npnt=%d)", ibody, iface, npnt);

    if        (ibody < 1 || ibody > MODL->nbody) {
        status = OCSM_ILLEGAL_BODY_INDEX;
        goto cleanup;
    } else if (iface < 1 || iface > MODL->body[ibody].nface) {
        status = OCSM_FACE_NOT_FOUND;
        goto cleanup;
    }

    /* if there is an active perturbation, get velocity by finite
       differences */
    if (MODL->perturb != NULL) {
        status = finiteDifference(MODL, ibody, OCSM_FACE, iface, npnt, uv, dxyz);
        CHECK_STATUS(finiteDifference);

        goto cleanup;
    }

    /* jbody is the Body in which this Face first appeared */
    jbody = MODL->body[ibody].face[iface].ibody;

    /* find the transformtion matrix which is required
       to track a point from its location in ibody back to
       its corresponding location in jbody */
    status = xformToOriginal(MODL, ibody, jbody, mat);
    CHECK_STATUS(xformToOriginal);

    /* if uv==NULL, then find the locations of the tessellation points when
       the Face was first created */
    if (uv == NULL) {
        status = EG_getTessFace(MODL->body[ibody].etess, iface,
                                &npnt_tess, &xyz_tess, &uv_tess, &ptype, &pindx,
                                &ntri_tess, &tris, &tric);
        CHECK_STATUS(EG_getTessFace);

        if (npnt != npnt_tess) {
            status = OCSM_ILLEGAL_ARGUMENT;
            goto cleanup;
        } else {
            MALLOC(xyz_pnt, double, 3*npnt_tess);

            /* xyz_pnt are the tessellation points transformed back
               to the their locations in jbody */
            for (ipnt = 0; ipnt < npnt_tess; ipnt++) {
                xyz_pnt[3*ipnt  ] = xyz_tess[3*ipnt  ] * mat[ 0]
                                  + xyz_tess[3*ipnt+1] * mat[ 1]
                                  + xyz_tess[3*ipnt+2] * mat[ 2] + mat[ 3];
                xyz_pnt[3*ipnt+1] = xyz_tess[3*ipnt  ] * mat[ 4]
                                  + xyz_tess[3*ipnt+1] * mat[ 5]
                                  + xyz_tess[3*ipnt+2] * mat[ 6] + mat[ 7];
                xyz_pnt[3*ipnt+2] = xyz_tess[3*ipnt  ] * mat[ 8]
                                  + xyz_tess[3*ipnt+1] * mat[ 9]
                                  + xyz_tess[3*ipnt+2] * mat[10] + mat[11];
            }
        }

    /* otherwise, find the locations at the given uv's when the Face
       was first created */
    } else {
        npnt_tess = npnt;
        MALLOC(xyz_pnt, double, 3*npnt_tess);

        for (ipnt = 0; ipnt < npnt_tess; ipnt++) {
            params[0] = uv[2*ipnt  ];
            params[1] = uv[2*ipnt+1];

            status = EG_evaluate(MODL->body[ibody].face[iface].eface,
                                 params, data);
            CHECK_STATUS(EG_evaluate);

            xyz_pnt[3*ipnt  ] = data[0] * mat[ 0]
                              + data[1] * mat[ 1]
                              + data[2] * mat[ 2] + mat[ 3];
            xyz_pnt[3*ipnt+1] = data[0] * mat[ 4]
                              + data[1] * mat[ 5]
                              + data[2] * mat[ 6] + mat[ 7];
            xyz_pnt[3*ipnt+2] = data[0] * mat[ 8]
                              + data[1] * mat[ 9]
                              + data[2] * mat[10] + mat[11];
        }
    }

    /* find the number of Parameters that have changed */
    nchange = MODL->body[jbody].hasdots;

    /* if a OCSM_EXTRUDE, count changes associated with the Sketch */
    if (MODL->body[jbody].brtype == OCSM_EXTRUDE) {
        isketch[0] = MODL->body[jbody].ileft;

        lbody = isketch[0];
        nchange += MODL->body[lbody].hasdots;

        nsketch++;
    }

    /* if a OCSM_RULE or OCSM_BLEND, count changes associated with the Sketches */
    if (MODL->body[jbody].brtype == OCSM_RULE ||
        MODL->body[jbody].brtype == OCSM_BLEND  ) {
        for (kbody = MODL->body[jbody].irite; kbody >= MODL->body[jbody].ileft; kbody--) {
            if (MODL->body[kbody].ichld == jbody) {
                isketch[nsketch] = kbody;

                lbody = isketch[nsketch];
                while (1) {
                    nchange += MODL->body[lbody].hasdots;

                    lbody = MODL->body[lbody].ileft;
                    if (lbody <= 0) break;
                }

                nsketch++;
            }
        }

        /* reverse order of Sketches */
        i = 0;
        j = nsketch - 1;
        while (i < j) {
            iswap      = isketch[i];
            isketch[i] = isketch[j];
            isketch[j] = iswap;

            i++;
            j--;
        }
    }

    /* if there are no Parameter changes, then the velocities are all 0 (and do not
       bother computing the derivatives below) */
    if (nchange == 0) {
        SPRINT0(2, "        -> setting to zero");
        for (ipnt = 0; ipnt < npnt; ipnt++) {
            dxyz[3*ipnt  ] = 0;
            dxyz[3*ipnt+1] = 0;
            dxyz[3*ipnt+2] = 0;
        }

    /* get the velocities for a box */
    } else if (MODL->body[jbody].brtype == OCSM_BOX) {
        SPRINT0(2, "        -> analytical for BOX");

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            data2[0] = xyz_pnt[3*ipnt  ];
            data2[1] = xyz_pnt[3*ipnt+1];
            data2[2] = xyz_pnt[3*ipnt+2];

            ubar = (data2[0] - MODL->body[jbody].arg[1].val[0]) / MODL->body[jbody].arg[4].val[0];
            vbar = (data2[1] - MODL->body[jbody].arg[2].val[0]) / MODL->body[jbody].arg[5].val[0];
            wbar = (data2[2] - MODL->body[jbody].arg[3].val[0]) / MODL->body[jbody].arg[6].val[0];

            dxyz[3*ipnt  ] = MODL->body[jbody].arg[1].dot[0] + MODL->body[jbody].arg[4].dot[0] * ubar;
            dxyz[3*ipnt+1] = MODL->body[jbody].arg[2].dot[0] + MODL->body[jbody].arg[5].dot[0] * vbar;
            dxyz[3*ipnt+2] = MODL->body[jbody].arg[3].dot[0] + MODL->body[jbody].arg[6].dot[0] * wbar;
        }

    /* get the velocities for a sphere */
    } else if (MODL->body[jbody].brtype == OCSM_SPHERE) {
        SPRINT0(2, "        -> analytical for SPHERE");

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            data2[0] = xyz_pnt[3*ipnt  ];
            data2[1] = xyz_pnt[3*ipnt+1];
            data2[2] = xyz_pnt[3*ipnt+2];

            ubar = (data2[0] - MODL->body[jbody].arg[1].val[0]) / MODL->body[jbody].arg[4].val[0];
            vbar = (data2[1] - MODL->body[jbody].arg[2].val[0]) / MODL->body[jbody].arg[4].val[0];
            wbar = (data2[2] - MODL->body[jbody].arg[3].val[0]) / MODL->body[jbody].arg[4].val[0];

            dxyz[3*ipnt  ] = MODL->body[jbody].arg[1].dot[0] + MODL->body[jbody].arg[4].dot[0] * ubar;
            dxyz[3*ipnt+1] = MODL->body[jbody].arg[2].dot[0] + MODL->body[jbody].arg[4].dot[0] * vbar;
            dxyz[3*ipnt+2] = MODL->body[jbody].arg[3].dot[0] + MODL->body[jbody].arg[4].dot[0] * wbar;
        }

    /* get the velocities for a cone */
    } else if (MODL->body[jbody].brtype == OCSM_CONE) {
        SPRINT0(2, "        -> analytical for CONE");

        xvrtx = MODL->body[jbody].arg[1].val[0];
        yvrtx = MODL->body[jbody].arg[2].val[0];
        zvrtx = MODL->body[jbody].arg[3].val[0];

        xbase = MODL->body[jbody].arg[4].val[0];
        ybase = MODL->body[jbody].arg[5].val[0];
        zbase = MODL->body[jbody].arg[6].val[0];

        rad   = MODL->body[jbody].arg[7].val[0];

        dx    = xbase - xvrtx;
        dy    = ybase - yvrtx;
        dz    = zbase - zvrtx;

        tht = atan2(dy, dx);
        phi = atan2(dz, sqrt(dy*dy + dx*dx));
        len = sqrt(dx*dx + dy*dy + dz*dz);
        psi = atan2(rad, len);

        num = dy;
        den = dx;
        if (fabs(num) > EPS06 && fabs(den) > EPS06) {
            numdot = MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0];
            dendot = MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0];
            thtdot = (dendot * num - numdot * den) / (num * num + den * den);
        } else {
            thtdot = 0;
        }

        num = dz;
        den = sqrt(dy*dy + dx*dx);
        if (fabs(num) > EPS06 && fabs(den) > EPS06) {
            numdot = MODL->body[jbody].arg[6].dot[0] - MODL->body[jbody].arg[3].dot[0];
            dendot = ( dx * (MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0])
                     + dy * (MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0])) / den;
            phidot = (dendot * num - numdot * den) / (num * num + den * den);
        } else {
            phidot = 0;
        }

        lendot = ( dx * (MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0])
                 + dy * (MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0])
                 + dz * (MODL->body[jbody].arg[6].dot[0] - MODL->body[jbody].arg[3].dot[0])) / len;

        psidot = (len * MODL->body[jbody].arg[7].dot[0] - lendot * rad) / (len * len * rad * rad);

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            x1 = xyz_pnt[3*ipnt  ];
            y1 = xyz_pnt[3*ipnt+1];
            z1 = xyz_pnt[3*ipnt+2];

            /* move to origin */
            x2 = x1 - xvrtx;
            y2 = y1 - yvrtx;
            z2 = z1 - zvrtx;

            /* rotate around z2 axis so that cone lies in (x3,z3) plane */
            x3 = x2 * cos(tht) + y2 * sin(tht);
            y3 = y2 * cos(tht) - x2 * sin(tht);
            z3 = z2;

            /* rotate around y3 axis so that cone lies along x3 axis */
            x4 = x3 * cos(phi) + z3 * sin(phi);
            y4 = y3;
            z4 = z3 * cos(phi) - x3 * sin(phi);

            /* conical coordinates (x, psi, theta) */
            x5 = x4;
            s5 = atan2(sqrt(y4*y4 + z4*z4), x4);   // psi5
            t5 = atan2(z4, y4);                    // theta5

            /* set velocities on the cone in its canonical orientation */
            x5dot = x5 / len * lendot;
            t5dot = 0;
            s5dot = s5 / psi * psidot;

            /* transform velocities back to original orientation */
            x4dot = x5dot;
            y4dot = x5dot * cos(t5)      * tan(s5)
                  + s5dot * cos(t5) * x5 / cos(s5) / cos(s5)
                  - t5dot * sin(t5) * x5 * tan(s5);
            z4dot = x5dot * sin(t5)      * tan(s5)
                  + s5dot * sin(t5) * x5 / cos(s5) / cos(s5)
                  + t5dot * cos(t5) * x5 * tan(s5);

            x3dot = x4dot * cos(phi) - x4 * phidot * sin(phi)
                  - z4dot * sin(phi) - z4 * phidot * cos(phi);
            y3dot = y4dot;
            z3dot = z4dot * cos(phi) - z4 * phidot * sin(phi)
                  + x4dot * sin(phi) + x4 * phidot * cos(phi);

            x2dot = x3dot * cos(tht) - x3 * thtdot * sin(tht)
                  - y3dot * sin(tht) - y3 * thtdot * cos(tht);
            y2dot = y3dot * cos(tht) - y3 * thtdot * sin(tht)
                  + x3dot * sin(tht) + x3 * thtdot * cos(tht);
            z2dot = z3dot;

            x1dot = x2dot + MODL->body[jbody].arg[1].dot[0];
            y1dot = y2dot + MODL->body[jbody].arg[2].dot[0];
            z1dot = z2dot + MODL->body[jbody].arg[3].dot[0];

            dxyz[3*ipnt  ] = x1dot;
            dxyz[3*ipnt+1] = y1dot;
            dxyz[3*ipnt+2] = z1dot;
        }

    /* get the velocities for a cylinder */
    } else if (MODL->body[jbody].brtype == OCSM_CYLINDER) {
        SPRINT0(2, "        -> analytical for CYLINDER");

        xbeg = MODL->body[jbody].arg[1].val[0];
        ybeg = MODL->body[jbody].arg[2].val[0];
        zbeg = MODL->body[jbody].arg[3].val[0];

        xend = MODL->body[jbody].arg[4].val[0];
        yend = MODL->body[jbody].arg[5].val[0];
        zend = MODL->body[jbody].arg[6].val[0];

        rad  = MODL->body[jbody].arg[7].val[0];

        dx   = xend - xbeg;
        dy   = yend - ybeg;
        dz   = zend - zbeg;

        tht = atan2(dy, dx);
        phi = atan2(dz, sqrt(dy*dy + dx*dx));
        len = sqrt(dx*dx + dy*dy + dz*dz);

        num = dy;
        den = dx;
        if (fabs(num) > EPS06 && fabs(den) > EPS06) {
            numdot = MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0];
            dendot = MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0];
            thtdot = (dendot * num - numdot * den) / (num * num + den * den);
        } else {
            thtdot = 0;
        }

        num = dz;
        den = sqrt(dy*dy + dx*dx);
        if (fabs(num) > EPS06 && fabs(den) > EPS06) {
            numdot = MODL->body[jbody].arg[6].dot[0] - MODL->body[jbody].arg[3].dot[0];
            dendot = ( dx * (MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0])
                     + dy * (MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0])) / den;
            phidot = (dendot * num - numdot * den) / (num * num + den * den);
        } else {
            phidot = 0;
        }

        lendot = ( dx * (MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0])
                 + dy * (MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0])
                 + dz * (MODL->body[jbody].arg[6].dot[0] - MODL->body[jbody].arg[3].dot[0])) / len;

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            x1 = xyz_pnt[3*ipnt  ];
            y1 = xyz_pnt[3*ipnt+1];
            z1 = xyz_pnt[3*ipnt+2];

            /* move to origin */
            x2 = x1 - xbeg;
            y2 = y1 - ybeg;
            z2 = z1 - zbeg;

            /* rotate around z2 axis so that cylinder lies in (x3,z3) plane */
            x3 = x2 * cos(tht) + y2 * sin(tht);
            y3 = y2 * cos(tht) - x2 * sin(tht);
            z3 = z2;

            /* rotate around y3 axis so that cylinder lies along x3 axis */
            x4 = x3 * cos(phi) + z3 * sin(phi);
            y4 = y3;
            z4 = z3 * cos(phi) - x3 * sin(phi);

            /* cylindrical coordinates (x, r, theta) */
            x5 = x4;
            r5 = sqrt(y4*y4 + z4*z4);   // rad5
            t5 = atan2(z4, y4);         // theta5

            /* set velocities on the cylinder in its canonical orientation */
            x5dot = x5 / len * lendot;
            r5dot = r5 / rad * MODL->body[jbody].arg[7].dot[0];
            t5dot = 0;

            /* transform velocities back to original orientation */
            x4dot = x5dot;
            y4dot = r5dot * cos(t5) - r5 * t5dot * sin(t5);
            z4dot = r5dot * sin(t5) + r5 * t5dot * cos(t5);

            x3dot = x4dot * cos(phi) - x4 * phidot * sin(phi)
                  - z4dot * sin(phi) - z4 * phidot * cos(phi);
            y3dot = y4dot;
            z3dot = z4dot * cos(phi) - z4 * phidot * sin(phi)
                  + x4dot * sin(phi) + x4 * phidot * cos(phi);

            x2dot = x3dot * cos(tht) - x3 * thtdot * sin(tht)
                  - y3dot * sin(tht) - y3 * thtdot * cos(tht);
            y2dot = y3dot * cos(tht) - y3 * thtdot * sin(tht)
                  + x3dot * sin(tht) + x3 * thtdot * cos(tht);
            z2dot = z3dot;

            x1dot = x2dot + MODL->body[jbody].arg[1].dot[0];
            y1dot = y2dot + MODL->body[jbody].arg[2].dot[0];
            z1dot = z2dot + MODL->body[jbody].arg[3].dot[0];

            dxyz[3*ipnt  ] = x1dot;
            dxyz[3*ipnt+1] = y1dot;
            dxyz[3*ipnt+2] = z1dot;
        }

    /* get the velocities for a udprim */
    } else if (MODL->body[jbody].brtype == OCSM_UDPRIM) {
        SPRINT0(2, "        -> analytical for UDPRIM");

        /* load and execute the user-defined primitive */
        ibrch = MODL->body[jbody].ibrch;

        STRNCPY(primtype, &(MODL->brch[ibrch].arg1[1]), MAX_EXPR_LEN);

        status = udp_initialize(primtype, &udp_num, &udp_names, &udp_types, &udp_idef, &udp_ddef);
        if (status == EGADS_NOLOAD) {
            status = EGADS_SUCCESS;
        }
        CHECK_STATUS(udp_initialize);

        status = udp_clrArguments(primtype);
        CHECK_STATUS(udp_clrArguments);

        needfd = 0;

        /* go back to first associated udparg statement whose primtype matches the
           udprim's primtype */
        kbody = jbody - 1;
        while(kbody > 0 && MODL->body[kbody].brtype == OCSM_UDPARG &&
              strcmp(MODL->body[kbody].arg[1].str, MODL->body[jbody].arg[1].str) == 0) {
            kbody--;
        }
        kbody++;

        while (kbody <= jbody) {
            for (iarg = 3; iarg < 10; iarg+=2) {
                if (MODL->body[kbody].arg[iarg].nval < 0) break;

                if (MODL->body[kbody].arg[iarg].nval > 0) {
                    status = udp_setVelocity(MODL->body[kbody].arg[1].str,
                                             MODL->body[jbody].ebody,
                                             MODL->body[kbody].arg[iarg-1].str,
                                             MODL->body[kbody].arg[iarg  ].dot,
                                             MODL->body[kbody].arg[iarg  ].nval);
                    if (status != EGADS_SUCCESS) {
                        /* indexerr indicates that a variable is not available for sensitivity,
                           (for example an integer), so we will just get zeroes */
                        if (status == EGADS_INDEXERR) {
                            SPRINT2(0, "WARNING:: sensitivity cannot be computed for \"%s\" in %s",
                                    MODL->body[kbody].arg[iarg-1].str, MODL->body[kbody].arg[1].str);
                        } else {
                            needfd++;
                        }
                    }
                }
            }

            kbody++;
        }

        /* if any of the velocities could not be set, we need finite differences */
        if (needfd > 0) {
            SPRINT0(2, "        -> analytical derivaitives not available, so reverting to finiteDifferences");

            status = finiteDifference(MODL, ibody, OCSM_FACE, iface, npnt, uv, dxyz);
            CHECK_STATUS(finiteDifference);
            goto cleanup;
        }

        MALLOC(uvs, double, 2*npnt);

        if (uv == NULL) {
            for (ipnt = 0 ; ipnt < npnt; ipnt++) {
                uvs[2*ipnt  ]  = uv_tess[2*ipnt  ];
                uvs[2*ipnt+1]  = uv_tess[2*ipnt+1];
            }
        } else {
            for (ipnt = 0 ; ipnt < npnt; ipnt++) {
                uvs[2*ipnt  ]  = uv[2*ipnt  ];
                uvs[2*ipnt+1]  = uv[2*ipnt+1];
            }
        }

        /* find the Face in jbody that maps to face iface in ibody */
        jface = -1;
        for (kface = 1; kface <= MODL->body[jbody].nface; kface++) {
            if (MODL->body[jbody].face[kface].iford == MODL->body[ibody].face[iface].iford) {
                jface = kface;
                break;
            }
        }
        SPRINT4(2, "        ibody=%d,iface=%d maps to jbody=%d,jface=%d", ibody, iface, jbody, jface);

        if (jface > 0) {
            status = udp_sensitivity(primtype, MODL->body[jbody].ebody,
                                     npnt, OCSM_FACE, jface, uvs, dxyz);
            CHECK_STATUS(udp_sensitivity);
        } else {
            signalError(MODL, OCSM_INTERNAL_ERROR, 0,
                        "cannot find Face %d:%d in jbody=%d", ibody, iface, jbody);
            status = OCSM_INTERNAL_ERROR;
            goto cleanup;
        }

    /* get the velocities for an extrusion */
    } else if (MODL->body[jbody].brtype == OCSM_EXTRUDE) {
        SPRINT0(2, "        -> analytical for EXTRUDE");

        MALLOC(ts,   double,   npnt);
        MALLOC(tn,   double,   npnt);
        MALLOC(vels, double, 3*npnt);

        /* count number of Edges in isketch[0] */
        nedge = MODL->body[isketch[0]].nedge;

        /* analytical derivatives from Sketch */
        if (iface <= nedge) {
            status = EG_getRange(MODL->body[ibody].face[iface].eface, trange, &periodic);
            CHECK_STATUS(EG_getRange);

            /* assume that iface=1 maps to iedge=1 in Sketch */
            for (ipnt = 0; ipnt < npnt; ipnt++) {
                if (uv == NULL) {
                    ts[ipnt] = uv_tess[2*ipnt  ];
                    tn[ipnt] = uv_tess[2*ipnt+1];
                } else {
                    ts[ipnt] = uv[2*ipnt  ];
                    tn[ipnt] = uv[2*ipnt+1];
                }
            }

            /* get velocities from Sketch */
            status = velocityOfEdge(MODL, isketch[0], iface, npnt, ts, vels);
            CHECK_STATUS(velocityOfEdge);

            /* apply velocities to SOLID Body */
            for (ipnt = 0; ipnt < npnt; ipnt++) {
                frac = (tn[ipnt] - trange[3]) / (trange[2] - trange[3]);

                dxyz[3*ipnt  ] = vels[3*ipnt  ] + frac * MODL->body[jbody].arg[1].dot[0];
                dxyz[3*ipnt+1] = vels[3*ipnt+1] + frac * MODL->body[jbody].arg[2].dot[0];
                dxyz[3*ipnt+2] = vels[3*ipnt+2] + frac * MODL->body[jbody].arg[3].dot[0];
            }

        /* analytical derivatives for ends */
        } else if (iface == nedge+1) {
            for (ipnt = 0; ipnt < npnt; ipnt++) {
                dxyz[3*ipnt  ] = 0;
                dxyz[3*ipnt+1] = 0;
                dxyz[3*ipnt+2] = 0;
            }

        } else if (iface == nedge+2) {
            for (ipnt = 0; ipnt < npnt; ipnt++) {
                dxyz[3*ipnt  ] = MODL->body[jbody].arg[1].dot[0];
                dxyz[3*ipnt+1] = MODL->body[jbody].arg[2].dot[0];
                dxyz[3*ipnt+2] = MODL->body[jbody].arg[3].dot[0];
            }

        } else {
            status = OCSM_FACE_NOT_FOUND;
            goto cleanup;
        }

        FREE(vels);
        FREE(tn  );
        FREE(ts  );

    /* get the velocities for a rule */
    } else if (MODL->body[jbody].brtype == OCSM_RULE) {
        SPRINT0(2, "        -> analytical for RULE");

        MALLOC(ts,   double,   npnt);
        MALLOC(tn,   double,   npnt);
        MALLOC(vels, double, 3*npnt);
        MALLOC(veln, double, 3*npnt);

        /* count number of Edges in isketch[0] */
        nedge = MODL->body[isketch[0]].nedge;

        /* analytical derivatives as blending of adjoining Edge velocities */
        if (iface <= nedge*(nsketch-1)) {
            isketchs = isketch[    (iface - 1) % (nsketch-1)];
            isketchn = isketch[1 + (iface - 1) % (nsketch-1)];
            iedge    =         1 + (iface - 1) / (nsketch-1);

            /* get face and edge limits */
            status = EG_getRange(MODL->body[jbody].face[iface].eface, uvlimits, &periodic);
            CHECK_STATUS(EG_getRange);

            if (MODL->body[isketchs].botype != OCSM_NODE_BODY) {
                status = EG_getRange(MODL->body[isketchs].edge[iedge].eedge, ulimits, &periodic);
                CHECK_STATUS(EG_getRange);
            }

            if (MODL->body[isketchn].botype != OCSM_NODE_BODY) {
                status = EG_getRange(MODL->body[isketchn].edge[iedge].eedge, ulimitn, &periodic);
                CHECK_STATUS(EG_getRange);
            }

            /* convert given uv to ubar and vbar */
            for (ipnt = 0; ipnt < npnt; ipnt++) {
                if (uv == NULL) {
                    ubar = (uv_tess[2*ipnt] - uvlimits[0]) / (uvlimits[1] - uvlimits[0]);
                } else {
                    ubar = (uv[     2*ipnt] - uvlimits[0]) / (uvlimits[1] - uvlimits[0]);
                }

                /* evaluate velocities on south and north boundaries */
                if (MODL->body[isketchs].botype != OCSM_NODE_BODY) {
                    ts[ipnt] = ulimits[0] + (ulimits[1] - ulimits[0]) * ubar;
                }

                if (MODL->body[isketchn].botype != OCSM_NODE_BODY) {
                    tn[ipnt] = ulimitn[0] + (ulimitn[1] - ulimitn[0]) * ubar;
                }
            }

            /* evaluate velocities on south and north Sketches */
            if (MODL->body[isketchs].botype != OCSM_NODE_BODY) {
                status = velocityOfEdge(MODL, isketchs, iedge, npnt, ts, vels);
                CHECK_STATUS(velocityOfEdge);
            } else {
                status = velocityOfNode(MODL, isketchs, iedge, vels);
                CHECK_STATUS(velocityOfNode);
            }

            if (MODL->body[isketchn].botype != OCSM_NODE_BODY) {
                status = velocityOfEdge(MODL, isketchn, iedge, npnt, tn, veln);
                CHECK_STATUS(velocityOfEdge);
            } else {
                status = velocityOfNode(MODL, isketchn, iedge, veln);
                CHECK_STATUS(velocityOfNode);
            }

            /* linearly interpolate between south and north */
            if (MODL->body[isketchs].botype != OCSM_NODE_BODY &&
                MODL->body[isketchn].botype != OCSM_NODE_BODY   ) {
                for (ipnt = 0; ipnt < npnt; ipnt++) {
                    if (uv == NULL) {
                        vbar = (uv_tess[2*ipnt+1] - uvlimits[2]) / (uvlimits[3] - uvlimits[2]);
                    } else {
                        vbar = (uv[     2*ipnt+1] - uvlimits[2]) / (uvlimits[3] - uvlimits[2]);
                    }

                    dxyz[3*ipnt  ] = (1-vbar) * vels[3*ipnt  ] + vbar * veln[3*ipnt  ];
                    dxyz[3*ipnt+1] = (1-vbar) * vels[3*ipnt+1] + vbar * veln[3*ipnt+1];
                    dxyz[3*ipnt+2] = (1-vbar) * vels[3*ipnt+2] + vbar * veln[3*ipnt+2];
                }
            } else if (MODL->body[isketchs].botype != OCSM_NODE_BODY) {
                for (ipnt = 0; ipnt < npnt; ipnt++) {
                    if (uv == NULL) {
                        vbar = (uv_tess[2*ipnt+1] - uvlimits[2]) / (uvlimits[3] - uvlimits[2]);
                    } else {
                        vbar = (uv[     2*ipnt+1] - uvlimits[2]) / (uvlimits[3] - uvlimits[2]);
                    }

                    dxyz[3*ipnt  ] = (1-vbar) * vels[3*ipnt  ] + vbar * veln[       0];
                    dxyz[3*ipnt+1] = (1-vbar) * vels[3*ipnt+1] + vbar * veln[       1];
                    dxyz[3*ipnt+2] = (1-vbar) * vels[3*ipnt+2] + vbar * veln[       2];
                }
            } else {
                for (ipnt = 0; ipnt < npnt; ipnt++) {
                    if (uv == NULL) {
                        vbar = (uv_tess[2*ipnt+1] - uvlimits[2]) / (uvlimits[3] - uvlimits[2]);
                    } else {
                        vbar = (uv[     2*ipnt+1] - uvlimits[2]) / (uvlimits[3] - uvlimits[2]);
                    }

                    dxyz[3*ipnt  ] = (1-vbar) * vels[       0] + vbar * veln[3*ipnt  ];
                    dxyz[3*ipnt+1] = (1-vbar) * vels[       1] + vbar * veln[3*ipnt+1];
                    dxyz[3*ipnt+2] = (1-vbar) * vels[       2] + vbar * veln[3*ipnt+2];
                }
            }

        /* analytical derivatives for ends */
        } else if (iface == nedge*(nsketch-1)+1) {
            if (MODL->body[isketch[0]].botype == OCSM_NODE_BODY) {
                status = velocityOfNode(MODL, isketch[0], 1, dxyz);
                CHECK_STATUS(velocityOfNode);
            } else if (uv == NULL) {
                status = velocityOfFace(MODL, isketch[0], 1, npnt, (double*)uv_tess, dxyz);
                CHECK_STATUS(velocityOfFace);
            } else {
                status = velocityOfFace(MODL, isketch[0], 1, npnt, (double*)uv,      dxyz);
                CHECK_STATUS(velocityOfFace);
            }

        } else if (iface == nedge*(nsketch-1)+2) {
            if (MODL->body[isketch[nsketch-1]].botype == OCSM_NODE_BODY) {
                status = velocityOfNode(MODL, isketch[nsketch-1], 1, dxyz);
                CHECK_STATUS(velocityOfNode);
            } else if (uv == NULL) {
                status = velocityOfFace(MODL, isketch[nsketch-1], 1, npnt, (double*)uv_tess, dxyz);
                CHECK_STATUS(velocityOfFace);
            } else {
                status = velocityOfFace(MODL, isketch[nsketch-1], 1, npnt, (double*)uv,      dxyz);
                CHECK_STATUS(velocityOfFace);
            }

        } else {
            status = OCSM_FACE_NOT_FOUND;
            goto cleanup;
        }

        FREE(veln);
        FREE(vels);
        FREE(tn  );
        FREE(ts  );

//$$$    /* get the velocities for a blend */
//$$$    } else if (MODL->body[jbody].brtype == OCSM_BLEND) {
//$$$        SPRINT0(2, "        -> analytical for BLEND");
//$$$
//$$$        int     atype, nsample;
//$$$        CINT    *tempIlist;
//$$$        double  *rc1, *rcN, *velmap;
//$$$        CDOUBLE *sample;
//$$$        CCHAR   *tempClist;
//$$$        void    *results;
//$$$
//$$$        extern int EG_blend_dot(ego Face, double *rc1, double *rcN,  void **results);
//$$$        extern int EG_blend_seval(void *results, double *uv, double *velmap, double *rc1_dot, double *rcN_dot, double *vel);
//$$$        extern int EG_sens_free(void *results);
//$$$
//$$$        /* Face is in blend direction */
//$$$        if (iface < nsketch) {
//$$$
//$$$            /* compute data required to get design velocities */
//$$$            if (MODL->body[ibody].arg[1].nval < 8) {
//$$$                rc1 = NULL;
//$$$            } else {
//$$$                rc1 = MODL->body[ibody].arg[1].val;
//$$$            }
//$$$
//$$$            if (MODL->body[ibody].arg[2].nval < 8) {
//$$$                rcN= NULL;
//$$$            } else {
//$$$                rcN = MODL->body[ibody].arg[2].val;
//$$$            }
//$$$
//$$$            status = EG_blend_dot(MODL->body[ibody].face[iface].eface, rc1, rcN, &results);
//$$$            CHECK_STATUS(EG_blend_dot);
//$$$
//$$$            /* set up velmap (velocities in the Sketches) */
//$$$            status = EG_attributeRet(MODL->body[ibody].face[iface].eface,
//$$$                                     ".blend", &atype, &nsample,
//$$$                                     &tempIlist, &sample, &tempClist);
//$$$            CHECK_STATUS(EG_attributeRet);
//$$$
//$$$            if (atype != ATTRREAL) {
//$$$                status = OCSM_INTERNAL_ERROR;
//$$$                CHECK_STATUS(EG_attributeRet);
//$$$            }
//$$$
//$$$            MALLOC(velmap, double, 3*nsample*nsketch);
//$$$            MALLOC(ts,     double,   nsample        );
//$$$
//$$$            for (i = 0; i < nsketch; i++) {
//$$$//$$$ check for sense of Edge
//$$$                for (j = 0; j < nsample; j++) {
//$$$                    ts[j] = (1-sample[j]) * ulimits[0] + sample[j] * ulimits[1];
//$$$                }
//$$$
//$$$                status = velocityOfEdge(MODL, isketch[i], iedge, nsample, ts, &(velmap[3*i]));
//$$$                CHECK_STATUS(velocityOfEdge);
//$$$            }
//$$$
//$$$            /* evaluate the sensitivities */
//$$$            if (uv == NULL) {
//$$$                for (ipnt = 0; ipnt < npnt; ipnt++) {
//$$$                    status = EG_blend_seval(results, &(uv[2*ipnt]), velmap,
//$$$                                            MODL->body[jbody].arg[1].dot, MODL->body[jbody].arg[2].dot,
//$$$                                            &(dxyz[3*ipnt]));
//$$$                    CHECK_STATUS(EG_blend_seval);
//$$$                }
//$$$            } else {
//$$$                for (ipnt = 0; ipnt < npnt; ipnt++) {
//$$$                    status = EG_blend_seval(results, &(uv_tess[2*ipnt]), velmap,
//$$$                                            MODL->body[jbody].arg[1].dot, MODL->body[jbody].arg[2].dot,
//$$$                                            &(dxyz[3*ipnt]));
//$$$                    CHECK_STATUS(EG_blend_seval);
//$$$                }
//$$$            }
//$$$
//$$$//$$$       figure out a better place to do this
//$$$            FREE(velmap);
//$$$            FREE(ts    );
//$$$
//$$$            status = EG_sens_free(results);
//$$$            CHECK_STATUS(EG_send_free);
//$$$
//$$$        /* analytical derivatives for ends */
//$$$        } else if (iface == nsketch) {
//$$$            if (MODL->body[isketch[0]].botype == OCSM_NODE_BODY) {
//$$$                status = velocityOfNode(MODL, isketch[0], 1, dxyz);
//$$$                CHECK_STATUS(velocityOfNode);
//$$$            } else if (uv == NULL) {
//$$$                status = velocityOfFace(MODL, isketch[0], 1, npnt, (double*)uv_tess, dxyz);
//$$$                CHECK_STATUS(velocityOfFace);
//$$$            } else {
//$$$                status = velocityOfFace(MODL, isketch[0], 1, npnt, (double*)uv,      dxyz);
//$$$                CHECK_STATUS(velocityOfFace);
//$$$            }
//$$$
//$$$        } else if (iface == nsketch+1) {
//$$$            if (MODL->body[isketch[nsketch-1]].botype == OCSM_NODE_BODY) {
//$$$                status = velocityOfNode(MODL, isketch[nsketch-1], 1, dxyz);
//$$$                CHECK_STATUS(velocityOfNode);
//$$$            } else if (uv == NULL) {
//$$$                status = velocityOfFace(MODL, isketch[nsketch-1], 1, npnt, (double*)uv_tess, dxyz);
//$$$                CHECK_STATUS(velocityOfFace);
//$$$            } else {
//$$$                status = velocityOfFace(MODL, isketch[nsketch-1], 1, npnt, (double*)uv,      dxyz);
//$$$                CHECK_STATUS(velocityOfFace);
//$$$            }
//$$$
//$$$        } else {
//$$$            status = OCSM_FACE_NOT_FOUND;
//$$$            goto cleanup;
//$$$        }

    /* unknown primitive type, so get velocities by finite differences */
    } else {
        SPRINT1(2, "        -> using finiteDifference (brtype=%s)",
                ocsmGetText(MODL->body[jbody].brtype));

        status = finiteDifference(MODL, ibody, OCSM_FACE, iface, npnt, uv, dxyz);
        CHECK_STATUS(finiteDifference);
        goto cleanup;
    }

    /* follow the children (up to the root) and modify the velocities
       for any OCSM_TRANSFORM that is visited */
    while (jbody != 0) {
        status = xformVelocity(MODL, jbody, npnt_tess, xyz_pnt, npnt, dxyz);
        CHECK_STATUS(xformVelocity);

        jbody = MODL->body[jbody].ichld;
    }

cleanup:
    FREE(veln);
    FREE(vels);
    FREE(tn  );
    FREE(ts  );

    FREE(xyz_pnt);
    FREE(uvs);

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   velocityOfNode - compute Node velocities analytically              *
 *                                                                      *
 ************************************************************************
 */

int
velocityOfNode(modl_T *MODL,            /* (in)  pointer to MODL */
               int    ibody,            /* (in)  Body index (bias-1) */
               int    inode,            /* (in)  Node index (bias-1) */
               double dxyz[])           /* (out) velocities (3*npnt in length) */
{
    int       status = 0;               /* (out) return status */

    int       iedge, jbody, j;
    double    tang[3], tang1[3], tang2[3], A[4], b[2], x[2], t;
    double    velo[3], velo1[3], velo2[3], trange[4];
    double    data[18], data2[3], mat[12], ubar, vbar, wbar;
    double    dx, dy, dz, num, numdot, den, dendot, xyz_pnt[3];
    double    xbeg, ybeg, zbeg, xend, yend, zend, xbase, ybase, zbase, xvrtx, yvrtx, zvrtx, rad;
    double    tht, thtdot, phi, phidot, len, lendot, psi, psidot;
    double    x1, x1dot, y1, y1dot, z1, z1dot;
    double    x2, x2dot, y2, y2dot, z2, z2dot;
    double    x3, x3dot, y3, y3dot, z3, z3dot;
    double    x4, x4dot, y4, y4dot, z4, z4dot;
    double    x5, x5dot, r5, r5dot, t5, t5dot, s5, s5dot;
    double    *uvs=NULL;

    int       periodic;

    ROUTINE(velocityOfNode);
    DPRINT3("%s(ibody=%d, inode=%d) {",
            routine, ibody, inode);

    /* --------------------------------------------------------------- */

    SPRINT2(2, "    velocityOfNode(ibody=%d, inode=%d)", ibody, inode);

    if        (ibody < 1 || ibody > MODL->nbody) {
        status = OCSM_ILLEGAL_BODY_INDEX;
        goto cleanup;
    } else if (inode < 1 || inode > MODL->body[ibody].nnode) {
        status = OCSM_NODE_NOT_FOUND;
        goto cleanup;
    }

    /* if there is an active perturbation, get velocity by finite
       differences */
    if (MODL->perturb != NULL) {
        status = finiteDifference(MODL, ibody, OCSM_NODE, inode, 1, NULL, dxyz);
        CHECK_STATUS(finiteDifference);

        goto cleanup;
    }

    /* if the Node was generated by a boolean operation or an OCSM_EXTRUDE,
       OCSM_RULE, OCSM_LOFT, OCSM_BLEND, or OCSM_UDPRIM velocity can be obtained
       from Edge velocities */
    jbody = MODL->body[ibody].node[inode].ibody;
    if (MODL->body[jbody].brtype == OCSM_INTERSECT ||
        MODL->body[jbody].brtype == OCSM_SUBTRACT  ||
        MODL->body[jbody].brtype == OCSM_UNION     ||
        MODL->body[jbody].brtype == OCSM_JOIN      ||
        MODL->body[jbody].brtype == OCSM_EXTRUDE   ||
        MODL->body[jbody].brtype == OCSM_RULE      ||
        MODL->body[jbody].brtype == OCSM_LOFT      ||
        MODL->body[jbody].brtype == OCSM_BLEND     ||
        MODL->body[jbody].brtype == OCSM_UDPRIM      ) {

        velo1[0] = velo1[1] = velo1[2] = 0;
        tang1[0] = tang1[1] = tang1[2] = 0;

        /* find two non-parallel Edges */
        j = 0;
        for (iedge = 1; iedge <= MODL->body[ibody].nedge; iedge++) {
            if (MODL->body[ibody].edge[iedge].ibeg == inode ||
                MODL->body[ibody].edge[iedge].iend == inode   ) {

                status = EG_getRange(MODL->body[ibody].edge[iedge].eedge, trange, &periodic);
                CHECK_STATUS(EG_getRange);

                if (MODL->body[ibody].edge[iedge].ibeg == inode) {
                    t = trange[0];
                } else {
                    t = trange[1];
                }

                /* velocity normal to Edge */
                status = velocityOfEdge(MODL, ibody, iedge, 1, &t, velo);
                CHECK_STATUS(velocityOfEdge);

                /* normalize tangent such that largest element is one */
                status = EG_evaluate(MODL->body[ibody].edge[iedge].eedge, &t, data);
                CHECK_STATUS(EG_evaluate);

                if (fabs(data[3]) >= MAX(fabs(data[4]), fabs(data[5]))) {
                    tang[0] = 1;
                    tang[1] = data[4] / data[3];
                    tang[2] = data[5] / data[3];
                } else if (fabs(data[4]) >= fabs(data[5])) {
                    tang[0] = data[3] / data[4];
                    tang[1] = 1;
                    tang[2] = data[5] / data[4];
                } else {
                    tang[0] = data[3] / data[5];
                    tang[1] = data[4] / data[5];
                    tang[2] = 1;
                }

                /* first Edge */
                if        (j == 0) {
                    tang1[0] = tang[0];
                    tang1[1] = tang[1];
                    tang1[2] = tang[2];
                    velo1[0] = velo[0];
                    velo1[1] = velo[1];
                    velo1[2] = velo[2];
                    j++;

                /* second Edge */
                } else if (j == 1) {
                    tang2[0] = tang[0];
                    tang2[1] = tang[1];
                    tang2[2] = tang[2];
                    velo2[0] = velo[0];
                    velo2[1] = velo[1];
                    velo2[2] = velo[2];

                    /* process second Edge if not parallel to first Edge */
                    if (fabs(tang1[0]-tang2[0]) > EPS03 ||
                        fabs(tang1[1]-tang2[1]) > EPS03 ||
                        fabs(tang1[2]-tang2[2]) > EPS03   ) {

                        /* minimize (velo1+x[0]*tang1) - (velo2+x[1]*tang2) */
                        A[0] =   tang1[0]*tang1[0] + tang1[1]*tang1[1] + tang1[2]*tang1[2];
                        A[1] = - tang1[0]*tang2[0] - tang1[1]*tang2[1] - tang1[2]*tang2[2];
                        A[2] = - tang2[0]*tang1[0] - tang2[1]*tang1[1] - tang2[2]*tang1[2];
                        A[3] =   tang2[0]*tang2[0] + tang2[1]*tang2[1] + tang2[2]*tang2[2];

                        b[0] = (velo2[0]-velo1[0]) * tang1[0]
                             + (velo2[1]-velo1[1]) * tang1[1]
                             + (velo2[2]-velo1[2]) * tang1[2];
                        b[1] = (velo1[0]-velo2[0]) * tang2[0]
                             + (velo1[1]-velo2[1]) * tang2[1]
                             + (velo1[2]-velo2[2]) * tang2[2];

                        status = matsol(A, b, 2, x);
                        CHECK_STATUS(matsol);

                        /* reconstruct velocity by velo+x[0]*tang1 */
                        dxyz[0] = velo1[0] + x[0] * tang1[0];
                        dxyz[1] = velo1[1] + x[0] * tang1[1];
                        dxyz[2] = velo1[2] + x[0] * tang1[2];

                        goto cleanup;
                    }
                }
            }
        }

        /* getting here means that we did not find two non-parallel Edges,
           so just return the (local) Edge velocity */
        SPRINT0(2, "WARNING:: could not find two non-parallel Edges");

        dxyz[0] = velo1[0];
        dxyz[1] = velo1[1];
        dxyz[2] = velo1[2];

        goto cleanup;
    }

    /* jbody is the Body in which this Node first appeared */
    jbody = MODL->body[ibody].node[inode].ibody;

    /* find the transformtion matrix which is required
       to track a point from its location in ibody back to
       its corresponding location in jbody */
    status = xformToOriginal(MODL, ibody, jbody, mat);
    CHECK_STATUS(xformToOriginal);

    /* xyz_pnt is the Node location transformed back
       to the its location in jbody */
    xyz_pnt[0] = MODL->body[ibody].node[inode].x * mat[ 0]
               + MODL->body[ibody].node[inode].y * mat[ 1]
               + MODL->body[ibody].node[inode].z * mat[ 2] + mat[ 3];
    xyz_pnt[1] = MODL->body[ibody].node[inode].x * mat[ 4]
               + MODL->body[ibody].node[inode].y * mat[ 5]
               + MODL->body[ibody].node[inode].z * mat[ 6] + mat[ 7];
    xyz_pnt[2] = MODL->body[ibody].node[inode].x * mat[ 8]
               + MODL->body[ibody].node[inode].y * mat[ 9]
               + MODL->body[ibody].node[inode].z * mat[10] + mat[11];

    /* if the .dots are all zero, then the velocities are 0 (and do not
       bother computing the derivatives below) */
    if (MODL->body[jbody].hasdots == 0) {
        SPRINT0(2, "        -> setting to zero");
        dxyz[0] = 0;
        dxyz[1] = 0;
        dxyz[2] = 0;

    /* get the velocities for a box */
    } else if (MODL->body[jbody].brtype == OCSM_BOX) {
        SPRINT0(2, "        -> analytical for BOX");

        data2[0] = xyz_pnt[0];
        data2[1] = xyz_pnt[1];
        data2[2] = xyz_pnt[2];

        ubar = (data2[0] - MODL->body[jbody].arg[1].val[0]) / MODL->body[jbody].arg[4].val[0];
        vbar = (data2[1] - MODL->body[jbody].arg[2].val[0]) / MODL->body[jbody].arg[5].val[0];
        wbar = (data2[2] - MODL->body[jbody].arg[3].val[0]) / MODL->body[jbody].arg[6].val[0];

        dxyz[0] = MODL->body[jbody].arg[1].dot[0] + MODL->body[jbody].arg[4].dot[0] * ubar;
        dxyz[1] = MODL->body[jbody].arg[2].dot[0] + MODL->body[jbody].arg[5].dot[0] * vbar;
        dxyz[2] = MODL->body[jbody].arg[3].dot[0] + MODL->body[jbody].arg[6].dot[0] * wbar;

    /* get the velocities for a sphere */
    } else if (MODL->body[jbody].brtype == OCSM_SPHERE) {
        SPRINT0(2, "        -> analytical for SPHERE");

        data2[ 0] = xyz_pnt[0];
        data2[ 1] = xyz_pnt[1];
        data2[ 2] = xyz_pnt[2];

        ubar = (data2[0] - MODL->body[jbody].arg[1].val[0]) / MODL->body[jbody].arg[4].val[0];
        vbar = (data2[1] - MODL->body[jbody].arg[2].val[0]) / MODL->body[jbody].arg[4].val[0];
        wbar = (data2[2] - MODL->body[jbody].arg[3].val[0]) / MODL->body[jbody].arg[4].val[0];

        dxyz[0] = MODL->body[jbody].arg[1].dot[0] + MODL->body[jbody].arg[4].dot[0] * ubar;
        dxyz[1] = MODL->body[jbody].arg[2].dot[0] + MODL->body[jbody].arg[4].dot[0] * vbar;
        dxyz[2] = MODL->body[jbody].arg[3].dot[0] + MODL->body[jbody].arg[4].dot[0] * wbar;

    /* get the velocities for a cone */
    } else if (MODL->body[jbody].brtype == OCSM_CONE) {
        SPRINT0(2, "        analytical for CONE");

        xvrtx = MODL->body[jbody].arg[1].val[0];
        yvrtx = MODL->body[jbody].arg[2].val[0];
        zvrtx = MODL->body[jbody].arg[3].val[0];

        xbase = MODL->body[jbody].arg[4].val[0];
        ybase = MODL->body[jbody].arg[5].val[0];
        zbase = MODL->body[jbody].arg[6].val[0];

        rad   = MODL->body[jbody].arg[7].val[0];

        dx    = xbase - xvrtx;
        dy    = ybase - yvrtx;
        dz    = zbase - zvrtx;

        tht = atan2(dy, dx);
        phi = atan2(dz, sqrt(dy*dy + dx*dx));
        len = sqrt(dx*dx + dy*dy + dz*dz);
        psi = atan2(rad, len);

        num = dy;
        den = dx;
        if (fabs(num) > EPS06 && fabs(den) > EPS06) {
            numdot = MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0];
            dendot = MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0];
            thtdot = (dendot * num - numdot * den) / (num * num + den * den);
        } else {
            thtdot = 0;
        }

        num = dz;
        den = sqrt(dy*dy + dx*dx);
        if (fabs(num) > EPS06 && fabs(den) > EPS06) {
            numdot = MODL->body[jbody].arg[6].dot[0] - MODL->body[jbody].arg[3].dot[0];
            dendot = ( dx * (MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0])
                     + dy * (MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0])) / den;
            phidot = (dendot * num - numdot * den) / (num * num + den * den);
        } else {
            phidot = 0;
        }

        lendot = ( dx * (MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0])
                 + dy * (MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0])
                 + dz * (MODL->body[jbody].arg[6].dot[0] - MODL->body[jbody].arg[3].dot[0])) / len;

        psidot = (len * MODL->body[jbody].arg[7].dot[0] - lendot * rad) / (len * len * rad * rad);

        x1 = xyz_pnt[0];
        y1 = xyz_pnt[1];
        z1 = xyz_pnt[2];

        /* move to origin */
        x2 = x1 - xvrtx;
        y2 = y1 - yvrtx;
        z2 = z1 - zvrtx;

        /* rotate around z2 axis so that cone lies in (x3,z3) plane */
        x3 = x2 * cos(tht) + y2 * sin(tht);
        y3 = y2 * cos(tht) - x2 * sin(tht);
        z3 = z2;

        /* rotate around y3 axis so that cone lies along x3 axis */
        x4 = x3 * cos(phi) + z3 * sin(phi);
        y4 = y3;
        z4 = z3 * cos(phi) - x3 * sin(phi);

        /* conical coordinates (x, psi, theta) */
        x5 = x4;
        s5 = atan2(sqrt(y4*y4 + z4*z4), x4);   // psi5
        t5 = atan2(z4, y4);                    // theta5

        /* set velocities on the cone in its canonical orientation */
        x5dot = x5 / len * lendot;
        t5dot = 0;
        s5dot = s5 / psi * psidot;

        /* transform velocities back to original orientation */
        x4dot = x5dot;
        y4dot = x5dot * cos(t5)      * tan(s5)
              + s5dot * cos(t5) * x5 / cos(s5) / cos(s5)
              - t5dot * sin(t5) * x5 * tan(s5);
        z4dot = x5dot * sin(t5)      * tan(s5)
              + s5dot * sin(t5) * x5 / cos(s5) / cos(s5)
              + t5dot * cos(t5) * x5 * tan(s5);

        x3dot = x4dot * cos(phi) - x4 * phidot * sin(phi)
              - z4dot * sin(phi) - z4 * phidot * cos(phi);
        y3dot = y4dot;
        z3dot = z4dot * cos(phi) - z4 * phidot * sin(phi)
              + x4dot * sin(phi) + x4 * phidot * cos(phi);

        x2dot = x3dot * cos(tht) - x3 * thtdot * sin(tht)
              - y3dot * sin(tht) - y3 * thtdot * cos(tht);
        y2dot = y3dot * cos(tht) - y3 * thtdot * sin(tht)
              + x3dot * sin(tht) + x3 * thtdot * cos(tht);
        z2dot = z3dot;

        x1dot = x2dot + MODL->body[jbody].arg[1].dot[0];
        y1dot = y2dot + MODL->body[jbody].arg[2].dot[0];
        z1dot = z2dot + MODL->body[jbody].arg[3].dot[0];

        dxyz[0] = x1dot;
        dxyz[1] = y1dot;
        dxyz[2] = z1dot;

    /* get the velocities for a cylinder */
    } else if (MODL->body[jbody].brtype == OCSM_CYLINDER) {
        SPRINT0(2, "        -> analytical for CYLINDER");

        xbeg = MODL->body[jbody].arg[1].val[0];
        ybeg = MODL->body[jbody].arg[2].val[0];
        zbeg = MODL->body[jbody].arg[3].val[0];

        xend = MODL->body[jbody].arg[4].val[0];
        yend = MODL->body[jbody].arg[5].val[0];
        zend = MODL->body[jbody].arg[6].val[0];

        rad  = MODL->body[jbody].arg[7].val[0];

        dx   = xend - xbeg;
        dy   = yend - ybeg;
        dz   = zend - zbeg;

        tht = atan2(dy, dx);
        phi = atan2(dz, sqrt(dy*dy + dx*dx));
        len = sqrt(dx*dx + dy*dy + dz*dz);

        num = dy;
        den = dx;
        if (fabs(num) > EPS06 && fabs(den) > EPS06) {
            numdot = MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0];
            dendot = MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0];
            thtdot = (dendot * num - numdot * den) / (num * num + den * den);
        } else {
            thtdot = 0;
        }

        num = dz;
        den = sqrt(dy*dy + dx*dx);
        if (fabs(num) > EPS06 && fabs(den) > EPS06) {
            numdot = MODL->body[jbody].arg[6].dot[0] - MODL->body[jbody].arg[3].dot[0];
            dendot = ( dx * (MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0])
                     + dy * (MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0])) / den;
            phidot = (dendot * num - numdot * den) / (num * num + den * den);
        } else {
            phidot = 0;
        }

        lendot = ( dx * (MODL->body[jbody].arg[4].dot[0] - MODL->body[jbody].arg[1].dot[0])
                 + dy * (MODL->body[jbody].arg[5].dot[0] - MODL->body[jbody].arg[2].dot[0])
                 + dz * (MODL->body[jbody].arg[6].dot[0] - MODL->body[jbody].arg[3].dot[0])) / len;

        x1 = xyz_pnt[0];
        y1 = xyz_pnt[1];
        z1 = xyz_pnt[2];

        /* move to origin */
        x2 = x1 - xbeg;
        y2 = y1 - ybeg;
        z2 = z1 - zbeg;

        /* rotate around z2 axis so that cylinder lies in (x3,z3) plane */
        x3 = x2 * cos(tht) + y2 * sin(tht);
        y3 = y2 * cos(tht) - x2 * sin(tht);
        z3 = z2;

        /* rotate around y3 axis so that cylinder lies along x3 axis */
        x4 = x3 * cos(phi) + z3 * sin(phi);
        y4 = y3;
        z4 = z3 * cos(phi) - x3 * sin(phi);

        /* cylindrical coordinates (x, r, theta) */
        x5 = x4;
        r5 = sqrt(y4*y4 + z4*z4);   // rad5
        t5 = atan2(z4, y4);         // theta5

        /* set velocities on the cylinder in its canonical orientation */
        x5dot = x5 / len * lendot;
        r5dot = r5 / rad * MODL->body[jbody].arg[7].dot[0];
        t5dot = 0;

        /* transform velocities back to original orientation */
        x4dot = x5dot;
        y4dot = r5dot * cos(t5) - r5 * t5dot * sin(t5);
        z4dot = r5dot * sin(t5) + r5 * t5dot * cos(t5);

        x3dot = x4dot * cos(phi) - x4 * phidot * sin(phi)
              - z4dot * sin(phi) - z4 * phidot * cos(phi);
        y3dot = y4dot;
        z3dot = z4dot * cos(phi) - z4 * phidot * sin(phi)
              + x4dot * sin(phi) + x4 * phidot * cos(phi);

        x2dot = x3dot * cos(tht) - x3 * thtdot * sin(tht)
              - y3dot * sin(tht) - y3 * thtdot * cos(tht);
        y2dot = y3dot * cos(tht) - y3 * thtdot * sin(tht)
              + x3dot * sin(tht) + x3 * thtdot * cos(tht);
        z2dot = z3dot;

        x1dot = x2dot + MODL->body[jbody].arg[1].dot[0];
        y1dot = y2dot + MODL->body[jbody].arg[2].dot[0];
        z1dot = z2dot + MODL->body[jbody].arg[3].dot[0];

        dxyz[0] = x1dot;
        dxyz[1] = y1dot;
        dxyz[2] = z1dot;

    /* unknown primitive type, so get velocities by finite differences */
    } else {
        SPRINT1(2, "        -> using finiteDifference (brtype=%s)",
                ocsmGetText(MODL->body[jbody].brtype));

        status = finiteDifference(MODL, ibody, OCSM_NODE, inode, 1, NULL, dxyz);
        CHECK_STATUS(finiteDifference);
        goto cleanup;
    }

    /* follow the children (up to the root) and modify the velocities
       for any OCSM_TRANSFORM that is visited */
    while (jbody != 0) {

        status = xformVelocity(MODL, jbody, 1, xyz_pnt, 1, dxyz);
        CHECK_STATUS(xformVelocity);

        jbody = MODL->body[jbody].ichld;
    }

cleanup:
    FREE(uvs);

    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   xformToOriginal - collect transformation from final to original    *
 *                                                                      *
 ************************************************************************
 */

static int
xformToOriginal(modl_T *MODL,           /* (in)  pointer to MODL */
                int    ibody,           /* (in)  Body index of source (bias-1) */
                int    jbody,           /* (in)  Body index of target (bias-1) */
                double mat[])           /* (out) transformation matrix */
{
    int      status = 0;                /* (out) return status */

    int      kbody, ktgt, j;
    double   dx, dy, dz, costht, sintht, temp0, temp1, temp2;
    double   fact, nx, ny, nz, dist;

    ROUTINE(xformToOriginal);
    DPRINT2("%s(ibody=%d) {",
            routine, ibody);

    /* --------------------------------------------------------------- */

    /* find the transformtion matrix which is required
       to track a point from its location in ibody back to
       its corresponding location in jbody */
    mat[ 0] = 1;   mat[ 1] = 0;   mat[ 2] = 0;   mat[ 3] = 0;
    mat[ 4] = 0;   mat[ 5] = 1;   mat[ 6] = 0;   mat[ 7] = 0;
    mat[ 8] = 0;   mat[ 9] = 0;   mat[10] = 1;   mat[11] = 0;

    /* walk from ibody back to jbody (which is the first Body
       in which inode appeared).  see comment at bottom
       of loop to see how we make sure that we choose the
       correct parent */
    kbody = ibody;
    while (kbody != jbody) {

        /* translate */
        if        (MODL->body[kbody].brtype == OCSM_TRANSLATE) {
            dx = MODL->body[kbody].arg[1].val[0];
            dy = MODL->body[kbody].arg[2].val[0];
            dz = MODL->body[kbody].arg[3].val[0];

            mat[ 3] -= dx;
            mat[ 7] -= dy;
            mat[11] -= dz;

        /* rotatex */
        } else if (MODL->body[kbody].brtype == OCSM_ROTATEX) {
            costht = cos(MODL->body[kbody].arg[1].val[0] * PIo180);
            sintht = sin(MODL->body[kbody].arg[1].val[0] * PIo180);
            dy     =     MODL->body[kbody].arg[2].val[0];
            dz     =     MODL->body[kbody].arg[3].val[0];

            for (j = 0; j < 4; j++) {
                temp1 = mat[4+j];
                temp2 = mat[8+j];

                mat[4+j] =   temp1 * costht + temp2 * sintht;
                mat[8+j] = - temp1 * sintht + temp2 * costht;
            }

            mat[ 7] += dy - dy * costht - dz * sintht;
            mat[11] += dz + dy * sintht - dz * costht;

        /* rotatey */
        } else if (MODL->body[kbody].brtype == OCSM_ROTATEY) {
            costht = cos(MODL->body[kbody].arg[1].val[0] * PIo180);
            sintht = sin(MODL->body[kbody].arg[1].val[0] * PIo180);
            dz     =     MODL->body[kbody].arg[2].val[0];
            dx     =     MODL->body[kbody].arg[3].val[0];

            for (j = 0; j < 4; j++) {
                temp2 = mat[8+j];
                temp0 = mat[  j];

                mat[8+j] =   temp2 * costht + temp0 * sintht;
                mat[  j] = - temp2 * sintht + temp0 * costht;
            }

            mat[11] += dz - dz * costht - dx * sintht;
            mat[ 3] += dx + dz * sintht - dx * costht;

        /* rotatez */
        } else if (MODL->body[kbody].brtype == OCSM_ROTATEZ) {
            costht = cos(MODL->body[kbody].arg[1].val[0] * PIo180);
            sintht = sin(MODL->body[kbody].arg[1].val[0] * PIo180);
            dx     =     MODL->body[kbody].arg[2].val[0];
            dy     =     MODL->body[kbody].arg[3].val[0];

            for (j = 0; j < 4; j++) {
                temp0 = mat[  j];
                temp1 = mat[4+j];

                mat[  j] =   temp0 * costht + temp1 * sintht;
                mat[4+j] = - temp0 * sintht + temp1 * costht;
            }

            mat[ 3] += dx - dx * costht - dy * sintht;
            mat[ 7] += dy + dx * sintht - dy * costht;

        /* scale */
        } else if (MODL->body[kbody].brtype == OCSM_SCALE) {
            fact = MODL->body[kbody].arg[1].val[0];

            mat[ 0] /= fact;
            mat[ 1] /= fact;
            mat[ 2] /= fact;
            mat[ 3] /= fact;
            mat[ 4] /= fact;
            mat[ 5] /= fact;
            mat[ 6] /= fact;
            mat[ 7] /= fact;
            mat[ 8] /= fact;
            mat[ 9] /= fact;
            mat[10] /= fact;
            mat[11] /= fact;

        /* mirror */
        } else if (MODL->body[kbody].brtype == OCSM_MIRROR) {
            nx    = MODL->body[kbody].arg[1].val[0];
            ny    = MODL->body[kbody].arg[2].val[0];
            nz    = MODL->body[kbody].arg[3].val[0];
            dist  = MODL->body[kbody].arg[4].val[0];

            fact  = sqrt(nx*nx + ny*ny + nz*nz);
            dx    = nx / fact;
            dy    = ny / fact;
            dz    = nz / fact;

            for (j = 0; j < 4; j++) {
                temp0 = mat[  j];
                temp1 = mat[4+j];
                temp2 = mat[8+j];

                mat[  j] = temp0 * (1-2*dx*dx) + temp1 * ( -2*dy*dx) + temp2 * ( -2*dz*dx);
                mat[4+j] = temp0 * ( -2*dx*dy) + temp1 * (1-2*dy*dy) + temp2 * ( -2*dz*dy);
                mat[8+j] = temp0 * ( -2*dx*dz) + temp1 * ( -2*dy*dz) + temp2 * (1-2*dz*dz);
            }

            mat[ 3] += 2 * dist * dx;
            mat[ 7] += 2 * dist * dy;
            mat[11] += 2 * dist * dz;
        }

        /* go to kbody's left or rite parent.  the only way to get
           the correct parent is to start over from jbody until
           the parent is the old kbody (which we call ktgt) */
        ktgt  = kbody;
        kbody = jbody;
        while (MODL->body[kbody].ichld != ktgt) {
            kbody = MODL->body[kbody].ichld;
        }
    }


//cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   xformVelocity - transform velocity and xyz_pnt by following tree   *
 *                                                                      *
 ************************************************************************
 */

static int
xformVelocity(modl_T *MODL,             /* (in)  pointer to MODL */
              int    jbody,             /* (in)  Body index (bias-1) */
              int    nxyz,              /* (in)  number of xyz_pnt */
              double xyz_pnt[],         /* (both)coordinates to be transformed */
              int    ndxyz,             /* (in)  number of dxyz */
              double dxyz[])            /* (both)velocities  to be transformed */
{
    int       status = 0;               /* (out) return status */

    int       ipnt;
    double    dx, dy, dz, sintht, costht, fact, nx, ny, nz, dist;
    double    xold, xolddot, yold, yolddot, zold, zolddot, thtdot;
    double    xbase, ybase, zbase;
    double    dxdot, dydot, dzdot, factdot, nxdot, nydot, nzdot, distdot;

    ROUTINE(xformVelocity);
    DPRINT4("%s(jbody=%d, nxyz=%d, ndxyz=%d) {",
            routine, jbody, nxyz, ndxyz);

    /* --------------------------------------------------------------- */

    /* translate */
    if (MODL->body[jbody].brtype == OCSM_TRANSLATE) {
        SPRINT0(2, "            -> translate");

        dx    = MODL->body[jbody].arg[1].val[0];
        dy    = MODL->body[jbody].arg[2].val[0];
        dz    = MODL->body[jbody].arg[3].val[0];

        dxdot = MODL->body[jbody].arg[1].dot[0];
        dydot = MODL->body[jbody].arg[2].dot[0];
        dzdot = MODL->body[jbody].arg[3].dot[0];

        for (ipnt = 0; ipnt < ndxyz; ipnt++) {
            dxyz[3*ipnt  ] += dxdot;
            dxyz[3*ipnt+1] += dydot;
            dxyz[3*ipnt+2] += dzdot;
        }

        for (ipnt = 0; ipnt < nxyz; ipnt++) {
            xyz_pnt[3*ipnt  ] += dx;
            xyz_pnt[3*ipnt+1] += dy;
            xyz_pnt[3*ipnt+2] += dz;
        }

    /* rotatex */
    } else if (MODL->body[jbody].brtype == OCSM_ROTATEX) {
        SPRINT0(2, "            -> rotatex");

        sintht = sin(MODL->body[jbody].arg[1].val[0] * PIo180);
        costht = cos(MODL->body[jbody].arg[1].val[0] * PIo180);
        thtdot = PIo180 * MODL->body[jbody].arg[1].dot[0];

        for (ipnt = 0; ipnt < ndxyz; ipnt++) {
            yold    = xyz_pnt[3*ipnt+1];
            zold    = xyz_pnt[3*ipnt+2];

            yolddot = dxyz[   3*ipnt+1];
            zolddot = dxyz[   3*ipnt+2];

            dxyz[3*ipnt+1] = MODL->body[jbody].arg[2].dot[0]
                + (yolddot - MODL->body[jbody].arg[2].dot[0]) * costht
                - (yold    - MODL->body[jbody].arg[2].val[0]) * sintht * thtdot
                - (zolddot - MODL->body[jbody].arg[3].dot[0]) * sintht
                - (zold    - MODL->body[jbody].arg[3].val[0]) * costht * thtdot;
            dxyz[3*ipnt+2] = MODL->body[jbody].arg[3].dot[0]
                + (yolddot - MODL->body[jbody].arg[2].dot[0]) * sintht
                + (yold    - MODL->body[jbody].arg[2].val[0]) * costht * thtdot
                + (zolddot - MODL->body[jbody].arg[3].dot[0]) * costht
                - (zold    - MODL->body[jbody].arg[3].val[0]) * sintht * thtdot;
        }

        ybase = MODL->body[jbody].arg[2].val[0];
        zbase = MODL->body[jbody].arg[3].val[0];

        for (ipnt = 0; ipnt < nxyz; ipnt++) {
            yold = xyz_pnt[3*ipnt+1];
            zold = xyz_pnt[3*ipnt+2];

            xyz_pnt[3*ipnt+1] = ybase + (yold-ybase) * costht - (zold-zbase) * sintht;
            xyz_pnt[3*ipnt+2] = zbase + (yold-ybase) * sintht + (zold-zbase) * costht;
        }

    /* rotatey */
    } else if (MODL->body[jbody].brtype == OCSM_ROTATEY) {
        SPRINT0(2, "            -> rotatey");

        sintht = sin(MODL->body[jbody].arg[1].val[0] * PIo180);
        costht = cos(MODL->body[jbody].arg[1].val[0] * PIo180);
        thtdot = PIo180 * MODL->body[jbody].arg[1].dot[0];

        for (ipnt = 0; ipnt < ndxyz; ipnt++) {
            xold    = xyz_pnt[3*ipnt  ];
            zold    = xyz_pnt[3*ipnt+2];

            xolddot = dxyz[   3*ipnt  ];
            zolddot = dxyz[   3*ipnt+2];

            dxyz[3*ipnt+2] = MODL->body[jbody].arg[2].dot[0]
                + (zolddot - MODL->body[jbody].arg[2].dot[0]) * costht
                - (zold    - MODL->body[jbody].arg[2].val[0]) * sintht * thtdot
                - (xolddot - MODL->body[jbody].arg[3].dot[0]) * sintht
                - (xold    - MODL->body[jbody].arg[3].val[0]) * costht * thtdot;
            dxyz[3*ipnt  ] = MODL->body[jbody].arg[3].dot[0]
                + (zolddot - MODL->body[jbody].arg[2].dot[0]) * sintht
                + (zold    - MODL->body[jbody].arg[2].val[0]) * costht * thtdot
                + (xolddot - MODL->body[jbody].arg[3].dot[0]) * costht
                - (xold    - MODL->body[jbody].arg[3].val[0]) * sintht * thtdot;
        }

        zbase = MODL->body[jbody].arg[2].val[0];
        xbase = MODL->body[jbody].arg[3].val[0];

        for (ipnt = 0; ipnt < nxyz; ipnt++) {
            zold = xyz_pnt[3*ipnt+2];
            xold = xyz_pnt[3*ipnt  ];

            xyz_pnt[3*ipnt+2] = zbase + (zold-zbase) * costht - (xold-xbase) * sintht;
            xyz_pnt[3*ipnt  ] = xbase + (zold-zbase) * sintht + (xold-xbase) * costht;
        }

    /* rotatez */
    } else if (MODL->body[jbody].brtype == OCSM_ROTATEZ) {
        SPRINT0(2, "            -> rotatez");

        sintht = sin(MODL->body[jbody].arg[1].val[0] * PIo180);
        costht = cos(MODL->body[jbody].arg[1].val[0] * PIo180);
        thtdot = PIo180 * MODL->body[jbody].arg[1].dot[0];

        for (ipnt = 0; ipnt < ndxyz; ipnt++) {
            xold    = xyz_pnt[3*ipnt  ];
            yold    = xyz_pnt[3*ipnt+1];

            xolddot = dxyz[   3*ipnt  ];
            yolddot = dxyz[   3*ipnt+1];

            dxyz[3*ipnt  ] = MODL->body[jbody].arg[2].dot[0]
                + (xolddot - MODL->body[jbody].arg[2].dot[0]) * costht
                - (xold    - MODL->body[jbody].arg[2].val[0]) * sintht * thtdot
                - (yolddot - MODL->body[jbody].arg[3].dot[0]) * sintht
                - (yold    - MODL->body[jbody].arg[3].val[0]) * costht * thtdot;
            dxyz[3*ipnt+1] = MODL->body[jbody].arg[3].dot[0]
                + (xolddot - MODL->body[jbody].arg[2].dot[0]) * sintht
                + (xold    - MODL->body[jbody].arg[2].val[0]) * costht * thtdot
                + (yolddot - MODL->body[jbody].arg[3].dot[0]) * costht
                - (yold    - MODL->body[jbody].arg[3].val[0]) * sintht * thtdot;
        }

        xbase = MODL->body[jbody].arg[2].val[0];
        ybase = MODL->body[jbody].arg[3].val[0];

        for (ipnt = 0; ipnt < nxyz; ipnt++) {
            xold = xyz_pnt[3*ipnt  ];
            yold = xyz_pnt[3*ipnt+1];

            xyz_pnt[3*ipnt  ] = xbase + (xold-xbase) * costht - (yold-ybase) * sintht;
            xyz_pnt[3*ipnt+1] = ybase + (xold-xbase) * sintht + (yold-ybase) * costht;
        }

    /* scale */
    } else if (MODL->body[jbody].brtype == OCSM_SCALE) {
        SPRINT0(2, "            -> scale");

        fact    = MODL->body[jbody].arg[1].val[0];
        factdot = MODL->body[jbody].arg[1].dot[0];

        for (ipnt = 0; ipnt < ndxyz; ipnt++) {
            xold    = xyz_pnt[3*ipnt  ];
            yold    = xyz_pnt[3*ipnt+1];
            zold    = xyz_pnt[3*ipnt+2];

            xolddot = dxyz[   3*ipnt  ];
            yolddot = dxyz[   3*ipnt+1];
            zolddot = dxyz[   3*ipnt+2];

            dxyz[3*ipnt  ] = xolddot * fact + xold * factdot;
            dxyz[3*ipnt+1] = yolddot * fact + yold * factdot;
            dxyz[3*ipnt+2] = zolddot * fact + zold * factdot;
        }

        for (ipnt = 0; ipnt < nxyz; ipnt++) {
            xyz_pnt[3*ipnt  ] *= fact;
            xyz_pnt[3*ipnt+1] *= fact;
            xyz_pnt[3*ipnt+2] *= fact;
        }

    /* mirror */
    } else if (MODL->body[jbody].brtype == OCSM_MIRROR) {
        SPRINT0(2, "            -> mirror");

        nx       = MODL->body[jbody].arg[1].val[0];
        ny       = MODL->body[jbody].arg[2].val[0];
        nz       = MODL->body[jbody].arg[3].val[0];
        dist     = MODL->body[jbody].arg[4].val[0];

        nxdot    = MODL->body[jbody].arg[1].dot[0];
        nydot    = MODL->body[jbody].arg[2].dot[0];
        nzdot    = MODL->body[jbody].arg[3].dot[0];
        distdot  = MODL->body[jbody].arg[4].dot[0];

        fact     = sqrt(nx*nx    + ny*ny    + nz*nz   );
        factdot  =     (nx*nxdot + ny*nydot + nz*nzdot) / fact;

        dx       = nx / fact;
        dy       = ny / fact;
        dz       = nz / fact;

        dxdot    = (fact * nxdot - nx * factdot) / fact / fact;
        dydot    = (fact * nydot - ny * factdot) / fact / fact;
        dzdot    = (fact * nzdot - nz * factdot) / fact / fact;

        for (ipnt = 0; ipnt < ndxyz; ipnt++) {
            xold    = xyz_pnt[3*ipnt  ];
            yold    = xyz_pnt[3*ipnt+1];
            zold    = xyz_pnt[3*ipnt+2];

            xolddot = dxyz[   3*ipnt  ];
            yolddot = dxyz[   3*ipnt+1];
            zolddot = dxyz[   3*ipnt+2];

            dxyz[3*ipnt  ] = xolddot * (1-2*dx*dx) + xold * (-4*dx*dxdot           )
                           + yolddot * ( -2*dx*dy) + yold * (-2*dx*dydot-2*dy*dxdot)
                           + zolddot * ( -2*dx*dz) + zold * (-2*dx*dzdot-2*dz*dxdot)
                           + distdot * (  2*dx   ) + dist * (            2*   dxdot);
            dxyz[3*ipnt+1] = xolddot * ( -2*dy*dx) + xold * (-2*dy*dxdot-2*dx*dydot)
                           + yolddot * (1-2*dy*dy) + yold * (-4*dy*dydot           )
                           + zolddot * ( -2*dy*dz) + zold * (-2*dy*dzdot-2*dz*dydot)
                           + distdot * (  2*dy   ) + dist * (            2*   dydot);
            dxyz[3*ipnt+2] = xolddot * ( -2*dz*dx) + xold * (-2*dz*dxdot-2*dx*dzdot)
                           + yolddot * ( -2*dz*dy) + yold * (-2*dz*dydot-2*dy*dzdot)
                           + zolddot * (1-2*dz*dz) + zold * (-4*dz*dzdot           )
                           + distdot * (  2*dz   ) + dist * (            2*   dzdot);
        }

        for (ipnt = 0; ipnt < nxyz; ipnt++) {
            xold = xyz_pnt[3*ipnt  ];
            yold = xyz_pnt[3*ipnt+1];
            zold = xyz_pnt[3*ipnt+2];

            xyz_pnt[3*ipnt  ] = xold * (1-2*dx*dx) + yold * ( -2*dy*dx) + zold * ( -2*dz*dx) + 2*dist*dx;
            xyz_pnt[3*ipnt+1] = xold * ( -2*dx*dy) + yold * (1-2*dy*dy) + zold * ( -2*dz*dy) + 2*dist*dy;
            xyz_pnt[3*ipnt+2] = xold * ( -2*dx*dz) + yold * ( -2*dy*dz) + zold * (1-2*dz*dz) + 2*dist*dz;
        }
    }

//cleanup:
    DPRINT2("%s --> status=%d}", routine, status);
    return status;
}
